/* $RCSfile: multiActuator.cc,v $ */
/* CM Info: $
*/
/* Russ Brill - Jet Propulsion Laboratory */

#include "isit.h"
#include "multiActuator.h"
#include "outputDest.h"
#include "stCommands.h"
#include "stError.h"

/*******************************************************************
*
* tMultiActuator::tMultiActuator -- tMultiActuator constructor
*
*  Allocates and enables actuators
*
* SPEC:
      [[U tMultiActuator::tMultiActuator(tOutputChannel *const ChannelToUse[],
                                         int iActuators,double dScaleFactor) ]]
*
* RETURNS:
*    nothing
*
****/
tMultiActuator::tMultiActuator(tOutputChannel *const ChannelToUse[],
                                 int iActuators,double dScaleFactor) :
     XformDetToAct_(iActuators,2),
     XformActToDet_(iActuators,2),
     bXformValid(FALSE),
     CiActuators(iActuators),
     pActuatorArray(new tModulatedActuator*[iActuators])
{
   int iActuator;
   
   for (iActuator = 0; iActuator < iActuators; iActuator++) {

      /****
      [[U tModulatedActuator::tModulatedActuator_td(*ChannelToUse[iActuator],
                                                    dScaleFactor) ]]
      ****/ 
      this->pActuatorArray[iActuator] = 
         new tModulatedActuator(*ChannelToUse[iActuator],dScaleFactor);

      /****
      [[U this->pActuatorArray[iActuator]->tModulatedActuator::EnableModulation() ]]
      ****/ 
      this->pActuatorArray[iActuator]->EnableModulation();
   }

   //currently the reverse transform is valid only for 2 actuators
   XformActToDet_.clear(); 
}


/*******************************************************************
*
* tMultiActuator::~tMultiActuator -- tMultiActuator destructor
*
*  Deallocates pActuatorArray[]
*
* SPEC:
      [[U tMultiActuator::~tMultiActuator()]]
*
* RETURNS:
*    nothing
*
****/
tMultiActuator::~tMultiActuator()
{
   int iActuator;
   
   for (iActuator = 0; iActuator < this->CiActuators; iActuator++) {

      /****
      [[U tModulatedActuator::~tModulatedActuator() ]]
      ****/ 
      delete this->pActuatorArray[iActuator];
   }

   delete[] pActuatorArray;
}



/*******************************************************************
*
* tMultiActuator::SetXform
*
*  Sets forward and reverse transform matrices.  Reverse transform is currently the inverse 
* of XformDetToAct, which only works for a 2-actuator device.  Otherwise, the reverse transform
* will be all zeros, as initialized.  Later software
* to do automatic calibration, will generate the reverse transform for other numbers of 
* actuators.
*
* SPEC:
      [[U tMultiActuator::SetXform(const tMatrix& MatrixToUse)]]
*
* RETURNS:
*    nothing
*
****/
void tMultiActuator::SetXform(const tMatrix& MatrixToUse)
{
   this->XformDetToAct_ = MatrixToUse;
   if (CiActuators == 2) {
      this->XformActToDet_ = MatrixToUse.inverse();
   }  
   this->bXformValid = TRUE;
}



/*******************************************************************
*
* tMultiActuator::SetActuatorModulation
*
*  Sets the dither parameters
*
* SPEC:
      [[U tMultiActuator::SetActuatorModulation(int iActuator,
                                                const tModulationParams& Params)]]
*
* RETURNS:
*    ERROR_CODE
*
****/
ERROR_CODE tMultiActuator::SetActuatorModulation(int iActuator,
                                                 const tModulationParams& Params)
{
   tModulationParams CountsParams = Params;
   
   //convert physical units to counts
   /****
   [[U this->pActuatorArray[iActuator]->
      tActuator::PhysicalUnitsToCounts(Params.dAmplitude) ]]
   ****/
   CountsParams.dAmplitude = 
      this->pActuatorArray[iActuator]->PhysicalUnitsToCounts(Params.dAmplitude);

   this->iCalibrationActuator = iActuator;

   //pass the specified parameters to the specified actuator
   /****
   [[U this->Actuator[Params.CalActuator].tModulatedActuator::
      SetModulationParameters(CountsParams) ]]
   ****/ 
   
   ERROR_CODE ec = this->pActuatorArray[iActuator]->SetModulationParameters(CountsParams);

   return ec;
}

   

/*******************************************************************
*
* tMultiActuator::Move
*
*  Moves the FSM to the specified detector coordinates
*
* SPEC:
      [[U tMultiActuator::Move(TIME tTime,const tVector& Value)]]
*
* RETURNS:
*    ERROR_CODE
*
****/
ERROR_CODE tMultiActuator::Move(TIME tTime,const tVector& Value)
{
   tVector  ActuatorValue(3);
   ERROR_CODE ec;
   int iActuator;
   double dCounts;
   
   if (!this->bXformValid) 
      return S_stgizmo_TRANSFORM_UNDEFINED;
   
   //transform (x,y) to actuator positions
   ActuatorValue = XformDetToAct_ * Value;

   for (iActuator = 0; iActuator < this->CiActuators; iActuator++) {

      //convert physical units to counts
      /****
      [[D this->pActuatorArray[iActuator]->
         tActuator::PhysicalUnitsToCounts(ActuatorValue[iActuator]) ]]
      ****/
      dCounts = this->pActuatorArray[iActuator]->
         PhysicalUnitsToCounts(ActuatorValue[iActuator]);

      /****
      [[U this->pActuatorArray[iActuator]->tModulatedActuator::
         UnmodulatedOutputValue(0,dCounts) ]]
      ****/ 
      ec = this->pActuatorArray[iActuator]->UnmodulatedOutputValue(0,dCounts);

      if (ec != OK) return ec;
   }

   return OK;
}
   
   
/*******************************************************************
*
* tMultiActuator::Dither
*
*  Moves the calibration actuator to the current dither position
*
* SPEC:
      [[U tMultiActuator::Dither(TIME tTime)]]
*
* RETURNS:
*    ERROR_CODE
*
****/
ERROR_CODE tMultiActuator::Dither(TIME tTime)
{

   //move the calibrating actuator to the position corresponding to the 
   //specified time
   /****
   [[U this->pActuatorArray[this->CalibrationActuator]->tOutputDestination::
      OutputValue(tTime,0) ]]
   ****/ 
   ERROR_CODE ec = this->pActuatorArray[this->iCalibrationActuator]->
             OutputValue(tTime,0);     


   return ec;
}

   
/*******************************************************************
*
* tPositionMultiActuator::tPositionMultiActuator -- 
* tPositonMultiActuator constructor
*
*  Allocates and enables actuators
*
* SPEC:
      [[U tPositionMultiActuator::tPositionMultiActuator(tOutputChannel *const ChannelToUse[],
                                         int iActuators,double dScaleFactor) ]]
*
* RETURNS:
*    nothing
*
****/
tPositionMultiActuator::tPositionMultiActuator(tOutputChannel *const ChannelToUse[],
                                 int iActuators,double dScaleFactor) :
    tMultiActuator(ChannelToUse,iActuators,dScaleFactor),
    bHomeValid(FALSE)
{
}

   
/*******************************************************************
*
* tPositionMultiActuator::Home
*
*  Moves the actuators to their home positions, without dither
*
* SPEC:
      [[U tPositionMultiActuator::Home()]]
*
* RETURNS:
*    error code returned by any actuator, or
*    S_stgizmo_TRANSFORM_UNDEFINED, if home has not been specified, or
*    OK, if successful
*
****/
ERROR_CODE tPositionMultiActuator::Home()
{

   ERROR_CODE ec;
   int iActuator;
   
   if (!this->bHomeValid)       
      return S_stgizmo_TRANSFORM_UNDEFINED;
   
    
   for (iActuator = 0; iActuator < this->CiActuators; iActuator++) {

      /****
      [[U this->pActuatorArray[iActuator]->
         tModulatedActuator::UnmodulatedOutputValue(0,0) ]]
      ****/ 
      
      ec = this->pActuatorArray[iActuator]->UnmodulatedOutputValue(0,0);
      if (ec != OK) return ec;
   }

   return OK;
}

   
/*******************************************************************
*
* tPositionMultiActuator::SetHome
*
*  Sets the FSM HOME position
*
* SPEC:
      [[U tPositionMultiActuator::SetHome(const tVector& VectorToUse)]]
*
* RETURNS:
*    nothing
*
****/
void tPositionMultiActuator::SetHome(const tVector& VectorToUse)
{
   int iActuator;
   double dCounts;
   
   for (iActuator = 0; iActuator < this->CiActuators; iActuator++) {
   
      //convert physical units to counts
      /****
      [[D this->pActuatorArray[iActuator]->
         tActuator::PhysicalUnitsToCounts(VectorToUse[iActuator]) ]]
      ****/
      dCounts = this->pActuatorArray[iActuator]->
         PhysicalUnitsToCounts(VectorToUse[iActuator]);

      /****
      [[U pActuatorArray[iActuator]->
      tOutputDestination::SetOffset(dCounts) ]]
      ****/
      pActuatorArray[iActuator]->SetOffset(dCounts);
   }
   this->bHomeValid = TRUE;
}



/*******************************************************************
*
* tPositionMultiActuator::Move
*
*  Moves the FSM to the specified detector coordinates from the HOME position
*
* SPEC:
      [[U tPositionMultiActuator::Move(TIME tTime,const tPoint& Position)]]
*
* RETURNS:
*    ERROR_CODE
*
****/
ERROR_CODE tPositionMultiActuator::Move(TIME tTime,const tPoint& Position)
{
   
   if (!this->bHomeValid) 
      return S_stgizmo_TRANSFORM_UNDEFINED;
   
   else
      return tMultiActuator::Move(tTime,Position);
      
}
   
