/* recLibB.c  -  buffered data recording subroutine library */
/* sccs info:  @(#)recLibB.c	1.3  11/29/93  */

/*
modification history
--------------------
01l,29nov93,mmc	  	fixed size check on recBytes
01k,01nov93,mmc	  	fixed some includes for 5.1
01j,08aug93,mmc     sped up recBytes
01i,15jul93,mmc     increased MARGIN to 500k
01h,23feb93,grw     modified recOpenArg to work with off board buffer
01h,15nov92,mmc     added recLibArg 
01g,09aug91,mmc     rewrote without ring buffers for speed  
01f,30jul91,mmc     added cast to avoid 5.0 warning 
01e,03jun91,mmc     do a memFindMax to set buffer size 
01d,31may91,mmc     delete rng buffer on recClose, cleanup
01c,14oct90,mmc   under SCCS, add recLibB.h for prototypes 
01b,20dec89,mmc   random cleanup--documentation, msgs, globals made static
01a,31oct89,mmc   written
*/

/*
DESCRIPTION
This module provides simple routines for
recording real-time data to an NFS file

This module buffers the data in a local
array.  The data are only written when the file is closed.

recOpen opens the remote file, creates the local buffer,
and spawns a buffer-checking daemon.  The routines recInt,
etc., store data into the buffer.  The daemon announces when
the buffer is full.  recClose flushes the buffer to the remote
file, closes the file, deletes the buffer, and kills the daemon.

*/

#include <vxWorks.h>
#include <stdioLib.h>
#include <memLib.h>
#include <sysLib.h>
#include <string.h>
#include <taskLib.h>
#include <stdlib.h>
#include <vme.h>
#include "recLibB.h"

#define MARGIN  500000  /* amount of memory to leave free when creating buffer*/
#define REC_PRIORITY 100  /* priority of buffer-checking daemon task */
#define REC_DELAY_TICKS (sysClkRateGet()) /* delay time for daemon */
#define BSIZE 4096      /* block size for writing data */

static FILE *recFile = NULL;
static int recTaskID = ERROR;
static enum {OFF,ON,OVERFLOW} recMode = OFF;

static char * pbuf;
static int nbuf;
static int bufSize;
static enum {BUF_LOCAL, BUF_EXTERNAL} bufFlag;

/************************************************
*
* recOpen - open a file for data recording
*
* This routine opens a file for data recording, 
* creates the local buffer, and spawns a task which
* announces when the buffer is full.
*
* recOpen also enables sysClk.  sysClk can be disabled
* after the file is open, although there won't be any
* overflow message.
*
* use recClose to write the data to the file.
*
* RETURNS:
*   ERROR if file already exists, cannot be opened, not enough
* memory for the buffer, task cannot be spawned, or data recording
* already active;
*   OK otherwise.
*/

STATUS recOpen(char * fileName)
{
  return recOpenArg (fileName, 0, 0);
}

/************************************************
*
* recOpenArg - open a file for data recording
*
* This routine is like recOpen, except that it takes additional
* arguments:
*
* precBuf, if NULL, signifies that space for the recording buffer should be
* malloced from on-board memory.  Otherwise, precBuf is interpreted
* as a pointer to a buffer in the VME_AM_EXT_SUP_DATA space; address
* remapping will be done automatically.
*
* recBufSize, if positive, is the maximum size of the recording buffer, and
* must be specified if an off-board buffer is specified.  If recBufSize is
* zero, the buffer size is limited to the maximum free space minus a margin.
* If recBufSize is negative, its absolute value overrides the default margin
* for a malloc from local memory.
*
*/
STATUS recOpenArg(char * fileName, char * precBuf, int recBufSize)
{
  void recDaemon();

  /* check if already recording */
  if(recMode != OFF){
    printErr("data recording already active \n");
    return ERROR;
  }

    /* force sysClk on */
    sysClkEnable();
    printf("sysClk enabled \n");

  /* check if file exists */
  recFile = fopen(fileName,"rb");
  if(recFile != NULL){
    printErr("file already exists \n");
    fclose(recFile);
    return ERROR;
  }
  
  /* create buffer */
  if (precBuf != NULL && recBufSize <= 0) {
    printErr ("Must specify buffer size with off-board memory \n");
    return ERROR;
  }

  if (recBufSize > 0)
    bufSize = recBufSize;
  else if (recBufSize == 0)
    bufSize = memFindMax() - MARGIN;
  else
    bufSize = memFindMax() + recBufSize;

  if(bufSize < BSIZE) {
    printErr ("Not enough memory for buffer \n");
    return ERROR;
  }

  if (precBuf == NULL) {
    if ((pbuf = (char *)malloc(bufSize)) == NULL) {
      printErr ("Error allocating memory \n");
      return ERROR;
    }
    bufFlag = BUF_LOCAL;
  } 
  else {
    if(sysBusToLocalAdrs((int) VME_AM_EXT_SUP_DATA, precBuf, &pbuf)){
      printErr ("Address conversion error \n");
      return ERROR;
    }
    bufFlag = BUF_EXTERNAL;
  }

  nbuf = 0;
  printf ("%d byte buffer created \n", bufSize);

  /* open file */
  recFile = fopen(fileName,"wb");
  if(recFile == NULL){
    printErr("Error opening file \n");
    recClose();
    return ERROR;
  }
  
  /* create buffer-watching task */
  recTaskID = taskSpawn("recDaemon",REC_PRIORITY,0,4000,(FUNCPTR)recDaemon,
		0,0,0,0,0,0,0,0,0,0);
  if(recTaskID == ERROR){
    printErr("Error creating task \n");
    recClose();
    return ERROR;
  } 

  /* enable recording */
  recMode = ON; 
  
  return OK;
}

/*buffer-watching daemon--started by recOpen*/
void recDaemon()
{
  for(;;){
    /* check for overflow */
    if(recMode == OVERFLOW){
      printErr("data recording buffer full\n");
      taskDelete(recTaskID);
      recTaskID = ERROR;
    }
    /* delay task */
    taskDelay(REC_DELAY_TICKS);
  }
}

/************************************************
*
* recClose - close data recording file
*
* This routine disables recording, writes the data to disk,
* closes the data recording file, and deletes the buffer-watching
* task
*
* sysClk is also enabled.
*
* RETURNS:
*   OK.
*/

STATUS recClose()
{
  int bytes = 0;

  /* write out buffer */
  if (recMode != OFF) {
    recMode = OFF;

      /* force sysClk on */
    sysClkEnable();
      printf("sysClk enabled \n");

    printf("writing %d bytes \n", nbuf);
    while (bytes < nbuf) {
      if(fwrite(pbuf+bytes, 1, min(BSIZE,(nbuf-bytes)), recFile)
         != min(BSIZE, (nbuf-bytes))){
        printErr("data recording write error \n");
        break;
      }
      bytes += BSIZE;
    }
  }

  /* close file */
  if(recFile != NULL){
    fclose(recFile);
    recFile = NULL;
  }

  /* delete task */
  if(recTaskID != ERROR)
    taskDelete(recTaskID);
    recTaskID = ERROR;

  /* delete buffer */
  if(bufFlag == BUF_LOCAL)
    free (pbuf);

  return OK;
}

/************************************************
*
* recInt - record a int
*
* This routine pushes an int into the buffer.
* It's a no-op if recording is disabled.
*
* RETURNS:
*   OK if recording enabled; ERROR otherwise.
*/

STATUS recInt(int rec)
{
  FAST int nbufNext;

  if(recMode == ON){
      if((nbufNext = nbuf + sizeof(int)) <= bufSize) {
      *((int *)(pbuf + nbuf)) = rec;
      nbuf = nbufNext;
      return OK;
    }
    else {
      recMode = OVERFLOW;
      return ERROR;
    }
  }
  return ERROR;
}

/************************************************
*
* recLong - record a long
*
* This routine pushes a long into the buffer.
* It's a no-op if recording is disabled.
*
* RETURNS:
*   OK if recording enabled; ERROR otherwise.
*/

STATUS recLong(long rec)
{
  FAST int nbufNext;

  if(recMode == ON){
      if((nbufNext = nbuf + sizeof(long)) <= bufSize) {
      *((long *)(pbuf + nbuf)) = rec;
      nbuf = nbufNext;
      return OK;
    }
    else {
      recMode = OVERFLOW;
      return ERROR;
    }
  }
  return ERROR;
}

/************************************************
*
* recShort - record a short 
*
* This routine pushes an short into the buffer.
* It's a no-op if recording is disabled.
*
* RETURNS:
*   OK if recording enabled; ERROR otherwise.
*/

STATUS recShort(short rec)
{
  FAST int nbufNext;

  if(recMode == ON){
      if((nbufNext = nbuf + sizeof(short)) <= bufSize) {
      *((short *)(pbuf + nbuf)) = rec;
      nbuf = nbufNext;
      return OK;
    }
    else {
      recMode = OVERFLOW;
      return ERROR;
    }
  }
  return ERROR;
}

/************************************************
*
* recChar - record a char 
*
* This routine pushes an char into the buffer.
* It's a no-op if recording is disabled.
*
* RETURNS:
*   OK if recording enabled; ERROR otherwise.
*/

STATUS recChar(char rec)
{
  FAST int nbufNext;

  if(recMode == ON){
      if((nbufNext = nbuf + sizeof(char)) <= bufSize) {
      *((char *)(pbuf + nbuf)) = rec;
      nbuf = nbufNext;
      return OK;
    }
    else {
      recMode = OVERFLOW;
      return ERROR;
    }
  }
  return ERROR;
}

/************************************************
*
* recBytes - record an arbitrary number of bytes
*
* This routine pushes random data into the rng buffer
* for recording by the data recording task.
* It's a no-op if recording is disabled.
*
* RETURNS:
*   OK if recording enabled; ERROR otherwise.
*/

STATUS recBytes(char * rec, int nbytes)
{
  FAST int nbufNext;

  if(recMode == ON){
    if((nbufNext = nbuf + nbytes) <= bufSize) {
      bcopy(rec, pbuf + nbuf, nbytes);
      nbuf = nbufNext;
      return OK;
    }
    else {
      recMode = OVERFLOW;
      return ERROR;
    }
  }
  return ERROR;
}

