/* $RCSfile: search.cc,v $  */
/* CM Info:$Header: /home/chara/cvsroot/ople/sys/comm/io/search.cc,v 1.1.1.1 1998/06/11 21:36:19 chara Exp $*/
/* Ken Starr - Jet Propulsion Laboratory */
 
/*
modification history
--------------------
01a,kms,14Nov97    written
*/
 
/*
DESCRIPTION

*/

#include "search.h"
#include "hwTypes.h"
#include "channel.h"
#include "math.h"



/*******************************************************************/
/***************************  tSearch  *****************************/
/*******************************************************************/
/******************************************************************
*
* tSearch::tSearch
*
* DESCRIPTION: Constructor
*
* RETURNS: Nothing
*

  [[U tSearch::tSearch_ddddst(dSpiralRadiusInitialToUse, dSpiralRadiusLimitToUse,		 
			 dSpiralStepToUse, dProgressionConstantToUse, 
			 stSearchTypeToUse)]]

*
*/

tSearch::tSearch(const double dSpiralRadiusInitialToUse,		 
	   	const double dSpiralRadiusLimitToUse,		 
	  	const double dSpiralStepToUse,
	   	const double dProgressionConstantToUse,		 
	   	const SEARCH_TYPE stSearchTypeToUse) 
{
  dSpiralRadiusInitial = dSpiralRadiusInitialToUse;			   
  dSpiralRadiusLimit = dSpiralRadiusLimitToUse;
  dSpiralStep = dSpiralStepToUse;
  dProgressionConstant = dProgressionConstantToUse;
  stSearchType = stSearchTypeToUse;
	
  dLastStepValue = 0.0;
  iCurrentStepDirection = 0 ;
  dCurrentSpiralMaximum = dSpiralRadiusInitial;
  iCurrentSpiralCycle = 0;
 
}


/*
*
  [[U tSearch::tSearch_t(FtSearchParametersToUse)]]

*
*/

tSearch::tSearch(const tFtSearchParameters& FtSearchParametersToUse) 

{
  dSpiralRadiusInitial = FtSearchParametersToUse.dSpiralRadiusInitial;			   
  dSpiralRadiusLimit = FtSearchParametersToUse.dSpiralRadiusLimit;
  dSpiralStep = FtSearchParametersToUse.dSpiralStep;
  dProgressionConstant = FtSearchParametersToUse.dProgressionConstant;
  stSearchType = FtSearchParametersToUse.stSearchType;
	
  dLastStepValue = 0.0;
  iCurrentStepDirection = 0 ;
  dCurrentSpiralMaximum = dSpiralRadiusInitial;
  iCurrentSpiralCycle = 0;
 
}

/*******************************************************************
*
* tSearch::~tSearch
*
* DESCRIPTION: Compiler-supplied destructor
*
* RETURNS: Nothing
*

[[U tSearch::~tSearch()]]

*/





/*******************************************************************
*
* tSearch::UpdateTarget
*
* DESCRIPTION: Computes new Integrated value using current and last
*		   values.
* DESCRIPTION: Computes new Search Target value using last values.
*		   
* New Search Target is calulated from last value and stSearchType, 
	dSpiralRadiusInitial, dSpiralRadiusLimit, dSpiralStep, 
	dProgressionConstant, dLastStepValue, iCurrentStepDirection, 
	dCurrentSpiralMaximum, iCurrentSpiralCycle.

  Update will return the next value in the search pattern taking steps of size
  dSpiralStep until the dCurrentSpiralMaximum is reached. Once the 
  dCurrentSpiralMaximum	is reached a new value for dCurrentSpiralMaximum is set 
  based on the Search Type and the  CurrentStepDirection is reversed.
	
  Currently, two search patterns are supported SPIRAL_GEOMETRIC (0,1,2,4,8...)
  and SPIRAL_ARITHMETRIC (0,1,2,3,4...).
 
* RETURNS: double
*

  [[U dSearchTarget := tSearch::UpdateTarget(void)]]

*
*/
double tSearch::UpdateTarget()
{
 dLastStepValue = dLastStepValue + (iCurrentStepDirection * dSpiralStep);

  if (iCurrentStepDirection == 0) 
    {
     dLastStepValue = 0;
     iCurrentStepDirection = 1 ;
     dCurrentSpiralMaximum = dSpiralRadiusInitial;
     iCurrentSpiralCycle = 0;
    }
  else if (dLastStepValue >= dSpiralRadiusLimit)
       
       dLastStepValue = dSpiralRadiusLimit;
              
  else if (dLastStepValue <= - dSpiralRadiusLimit)
       
       dLastStepValue = -dSpiralRadiusLimit;
            
  else if (fabs(dLastStepValue) >= dCurrentSpiralMaximum)  { 
    iCurrentSpiralCycle = iCurrentSpiralCycle + 1 ; 
     	  
    if (stSearchType == SPIRAL_GEOMETRIC)
     	  
      dCurrentSpiralMaximum =  
    	    pow(dProgressionConstant,iCurrentSpiralCycle) *
    	    dSpiralRadiusInitial;
     				   
    else /* stSearchType == SPIRAL_ARITHMETRIC */ 
            
      dCurrentSpiralMaximum = iCurrentSpiralCycle *
     			       dProgressionConstant *
     			       dSpiralRadiusInitial;
     				   				   
      iCurrentStepDirection = -(iCurrentStepDirection);  
     	  
  }
  
    return(dLastStepValue);
}


/******************************************************************
*
* tSearch::Reset 
*
* DESCRIPTION: Resets the Search pattern
*
* RETURNS: ERROR_CODE
*

  [[U tSearch::Reset (void)]]

*
*/

ERROR_CODE tSearch::Reset (void)
{
  iCurrentStepDirection = 0;
  dLastStepValue = 0;
  dCurrentSpiralMaximum = dSpiralRadiusInitial;
  iCurrentSpiralCycle = 0;
     
  return OK;
}




/*
*******************************************************************
*
* tSearch::tSetParameters
*
* DESCRIPTION: Set search's controlling parameters
*
* RETURNS: ERROR_CODE
*
*

[[U tSearch::SetParameters_ddddst (const double dSpiralRadiusInitialToUse,		 
			 const double dSpiralRadiusLimitToUse,		 
			 const double dSpiralStepToUse,
			 const double dProgressionConstantToUse,		 
			 const SEARCH_TYPE stSearchTypeToUse ) 
 ]]
*/

ERROR_CODE tSearch::SetParameters(const double dSpiralRadiusInitialToUse,		 
	   	const double dSpiralRadiusLimitToUse,		 
	  	const double dSpiralStepToUse,
	   	const double dProgressionConstantToUse,		 
	   	const SEARCH_TYPE stSearchTypeToUse) 
{
  
  dSpiralRadiusInitial = dSpiralRadiusInitialToUse;			   
  dSpiralRadiusLimit = dSpiralRadiusLimitToUse;
  dSpiralStep = dSpiralStepToUse;
  dProgressionConstant = dProgressionConstantToUse;
  stSearchType = stSearchTypeToUse;
	
  return OK;
 	
}



/*
*******************************************************************
*
* tSearch::tSetParameters
*
* DESCRIPTION: Set search's controlling parameters
*
* RETURNS: ERROR_CODE
*
*

  [[U ec:= tSearch::SetParameters_t(FtSearchParametersToUse)]]

*
*/

ERROR_CODE tSearch::SetParameters(const tFtSearchParameters& 
						FtSearchParametersToUse) 

{
  
  dSpiralRadiusInitial = FtSearchParametersToUse.dSpiralRadiusInitial;			   
  dSpiralRadiusLimit = FtSearchParametersToUse.dSpiralRadiusLimit;
  dSpiralStep = FtSearchParametersToUse.dSpiralStep;
  dProgressionConstant = FtSearchParametersToUse.dProgressionConstant;
  stSearchType = FtSearchParametersToUse.stSearchType;
	
  return OK;
 
}


/*
*
*******************************************************************
*
* tSearch::SearchParameters
*
* DESCRIPTION: Return current values of search parameters
*
* RETURNS: tFtSearchParameters 
*
*

[[U SearchParameters := tSearch::SearchParameters_t () ]]
  
  
*/
tFtSearchParameters tSearch::SearchParameters () const 
  {
    tFtSearchParameters Temp_FtSearchParameters;
    
    Temp_FtSearchParameters.dSpiralRadiusInitial = dSpiralRadiusInitial ;			   
    Temp_FtSearchParameters.dSpiralRadiusLimit = dSpiralRadiusLimit; 
    Temp_FtSearchParameters.dSpiralStep = dSpiralStep ;
    Temp_FtSearchParameters.dProgressionConstant = dProgressionConstant ;
    Temp_FtSearchParameters.stSearchType = stSearchType ;
    
    return Temp_FtSearchParameters;
  
  }