/* $RCSfile: mutex.cc,v $ - implements mutually exclusive access to a resource */
/* CM Info:
$Header: /home/chara/cvsroot/ople/sys/comm/mutex/mutex.cc,v 1.1.1.1 1998/06/11 21:36:20 chara Exp $
*/
/* Brad Hines - Jet Propulsion Laboratory */
 
/*
modification history
--------------------
01a,beh,28Sep96    written
*/
 
/*
DESCRIPTION

*/

#include <vxWorks.h>
#include <semLib.h>
#include <errnoLib.h>
#include <objLib.h>
#include <strLib.h>
#include "mutex.h"
#include "gizmo.h"
#include "errorStream.h"


/******************************************************************
*
* tMutex State Box State Invariants
*
* 1) [ siSemaphore != NULL ->     // Violated in tMutex constructor
*        WatchDog.bInit == TRUE
*    ]
*
* 2) [ siSemaphore valid ->       // Helps to simplify some spec
*        *siSemaphore.Type == SEM_BINARY
*    ]
*/

/******************************************************************
*
* tMutexUser State Box State Invariants
*
* 1) [ siAbortSemaphore == NULL ->
*        Command associated with the object is not currently active
*    | siAbortSemaphore is invalid ->
*        Command abort completed, object about to be destroyed 
*    | *siAbortSemaphore full ->
*        iTaskId valid and task iTaskid is spawned.
*    | *siAbortSemaphore empty ->
*        Task creation, termination, or abortion in process.  You
*                       should wait until the process is complete
*                       before attempting to manipulate the task.
*    ]
*/

/*******************************************************************
*
* SPEC FUNCTIONS
*
*********
*
* This function returns true if the semaphore can be taken within the
* timeout specified by iTimeoutInTicks.  Copied from semLib.spec
*
* BOOL CanTakeSemaphoreWithinSpecifiedTimeout(SEM_ID siSemaphore,
*                                               int iTimeoutInTicks)
*       [ iTimeoutInTicks != WAIT_FOREVER &&
*            (iTimeoutInTicks == NO_WAIT &&
*               ! SemAvailable(siSemaphore) ||
*               SemAvailable(siSemaphore) not true within
*                                     iTimeoutInTicks ticks) ->
*           return FALSE
*       | true ->
*           return TRUE
*       ]
*
*********
*
* This function returns a possibly-modified copy of the watchdog
* passed into it.  If the timeout is not WAIT_FOREVER, the copy is
* an activated watchdog with the specified timeout.  If the timeout
* is WAIT_FOREVER, the copy is unchanged (a non-activated watchdog)
*
* tWatchDog ActivateWatchDogIfNotForeverTimeout(tWatchDog wd, int iTimeout) :==
*      [ iTimeout == WAIT_FOREVER ->
*            return a copy of wd
*      | true ->
*            return a copy of wd, with
*                            .iTimeLeft := iTimeout
*                            .bActive := TRUE
*      ]
*/


/******************************************************************
* A
* tMutex::tMutex - create a mutex object
*
* Creates a mutex object.  This constructor creates a binary semaphore
* and a tWatchDog set up to call our own WatchDogHandler method.
*
* IN:
*   tMutex   *this
*   MUTEX_DEFAULT_TIMEOUT   - default time to wait for mutex, in seconds
* 
* INOUT:
*   *this.WatchDog
*   OS.SemaphoreSet
*   OS.WdTimersAvailable
*   OS.SpawnedTasks
*   available memory
*
* OUT:
*   *this.siSemaphore
*   *this.*siSempahore
*   *this.WatchDog.fpHandler
*   *this.WatchDog.iHandlerArg
*   *this.WatchDog.iTimeout
*   *this.WatchDog.bActive
*   *this.WatchDog.bInit
*   *this.pOwner
*
* SPEC:
    [ this invalid -> UNDEFINED
    | memory for semaphore cannot be allocated -> UNDEFINED
    | a pointer cannot be stored in an int on this machine -> UNDEFINED
    | MUTEX_DEFAULT_TIMEOUT <= 0 -> UNDEFINED
    | VxWorks.WdTimersAvailable <= 0 -> UNDEFINED
    | MUTEX_DEFAULT_TIMEOUT <= 0 ||
                TIME_TO_VXWORKS_TICKS(MUTEX_DEFAULT_TIMEOUT) < 1 -> UNDEFINED
    | Task cannot be created || Insufficient memory to spawn task -> UNDEFINED
    | true ->
       OS.SemaphoreSet := OS.SemphoreSet U
                        NewBinarySemaphore(SEM_Q_FIFO, SEM_FULL)
       : *this.siSemaphore := NewBinarySemaphore(SEM_Q_FIFO, SEM_FULL).Id
       : *this.WatchDog.fpHandler := &tMutex::WatchDogHandler
       : *this.WatchDog.iHandlerArg := this
       : *this.WatchDog.iTimeout := MUTEX_DEFAULT_TIMEOUT
       : *this.WatchDog.bActive := FALSE
       : *this.WatchDog.bInit := TRUE
       : VxWorks.WdTimersAvailable :-= 1
       : OS.SpawnedTasks := OS.SpawnedTasks U
                    { *this.WatchDog.TaskLevelHandler(), with
                      priority WATCHDOG_HANDLER_DEFAULT_PRIORITY,
                      options WATCHDOG_HANDLER_DEFAULT_TASK_OPTIONS,
                      stack size WATCHDOG_HANDLER_DEFAULT_STACK_SIZE,
                      and name "wdHandlerTask" }
       : *this.pOwner := NULL
    ]
*
* RETURNS:
*   Nothing.
*/

tMutex::tMutex(void)
{
  ERROR_CODE ec;

  /******    (tTaskLevelWatchDog Constructor)
    [ &*this.WatchDog invalid -> UNDEFINED
    | true ->
        *this.WatchDog.fpHandler := NULL
        : *this.WatchDog.iHandlerArg := 0
        : *this.WatchDog.iTimeout := 0
        : *this.WatchDog.bInit := FALSE
    ] 
  *****/
  
  /******
    [ this invalid -> UNDEFINED
    | memory for semaphore cannot be allocated ->
       *this.siSemaphore = NULL
    | true ->
       OS.SemaphoreSet := OS.SemphoreSet U
                        NewBinarySemaphore(SEM_Q_FIFO, SEM_FULL)
       : *this.siSemaphore := NewBinarySemaphore(SEM_Q_FIFO, SEM_FULL).Id
    ]
   ******/
  // Create the VxWorks semaphore associated with this object
  this->siSemaphore = semBCreate(SEM_Q_FIFO, SEM_FULL);
  
  /*****
    [ this invalid -> UNDEFINED
    | &(*this.WatchDog) invalid -> UNDEFINED
    | a pointer cannot be stored in an int on this machine -> UNDEFINED
    | *this.WatchDog.bInit == TRUE -> ec = S_watchDog_ALREADY_INITIALIZED
    | &tMutex::WatchDogHandler == NULL ->
        ec = S_watchDog_NULL_FUNCTION_POINTER
    | !(&tMutex::WatchDogHandler is a pointer to a valid C function ||
       (&tMutex::WatchDogHandler a pointer to a valid C++ method &&
        this is a pointer to an object of the type for
                             which the method is defined)) -> UNDEFINED
    | &tMutex::WatchDogHandler a method of tTaskLevelWatchDog -> UNDEFINED
    | MUTEX_DEFAULT_TIMEOUT <= 0 ||
                     TIME_TO_VXWORKS_TICKS(MUTEX_DEFAULT_TIMEOUT) < 1 ->
        ec = S_watchDog_INVALID_TIMEOUT_VALUE
    | VxWorks.WdTimersAvailable <= 0 -> ec = VxWorks error code (wdCreate)
    | Task cannot be created || Insufficient memory to spawn task ->
           *this.iHandlerTaskId := ERROR
           : ec = VxWorks errno (taskSpawn)
    | true ->
        *this.WatchDog.fpHandler := &tMutex::WatchDogHandler
        : *this.WatchDog.iHandlerArg := this
        : *this.WatchDog.iTimeout := MUTEX_DEFAULT_TIMEOUT
        : *this.WatchDog.bActive := FALSE
        : *this.WatchDog.bInit := TRUE
        : VxWorks.WdTimersAvailable :-= 1
        : OS.SpawnedTasks := OS.SpawnedTasks U
                    { *this.WatchDog.TaskLevelHandler(), with
                      priority WATCHDOG_HANDLER_DEFAULT_PRIORITY,
                      options WATCHDOG_HANDLER_DEFAULT_TASK_OPTIONS,
                      stack size WATCHDOG_HANDLER_DEFAULT_STACK_SIZE,
                      and name "wdHandlerTask" }
        : ec = OK
    ]
  *****/
  ec = this->WatchDog.Initialize((FUNCPTR)&tMutex::WatchDogHandler,
                                 (int) this, MUTEX_DEFAULT_TIMEOUT);

  /*****
    [ this invalid -> UNDEFINED
    | *this.siSemaphore == NULL || ec != OK -> NOP
    | true -> NOP
    ]
  *****/
  if (this->siSemaphore == NULL || ec != OK) {
    // Do something!
    #ifndef INCR1
      #error Increment2ChangeNeeded
    #endif
  }

  /*****
    [ this invalid -> UNDEFINED
    | true->          *this.pOwner := NULL ]
  *****/
  pOwner = NULL;
}


/******************************************************************
* A
* tMutex::Take - acquire the mutex
*
* The tMutexUser class calls this method in order to acquire the
* mutex to gain access to the shared resource protected by the
* mutex.  The caller specifies how long he's willing to wait
* for the mutex as well, and whether to steal the mutex from the
* current owner if it's not available at the end of that time.
* Take also sets up a watchdog timer for the caller.  If the caller
* has not returned the mutex within iTicksToAllowForOperation
* system timer ticks, the operation will be aborted and the
* mutex reclaimed.
*
* The time to wait is specified in system timer ticks, whose
* rate can be determined via sysClkRateGet(), usually 60 Hz.
* In lieu of numbers of ticks, the values WAIT_FOREVER
* or NO_WAIT can be specified instead if desired.  WAIT_FOREVER
* is also a valid value for iTicksToAllowForOperation.  Specifying
* NO_WAIT for iTicksToAllowForOperation could be done, but
* the operation would be aborted immediately.
*
* The "maximally impatient" way to call this function is with
* iTicksToWaitForMutex set to NO_WAIT and bKillCurrentOwnerIfTimeout
* set to TRUE.  In this case, if the mutex is in use, the Take will
* time out immediately, and Take will respond by killing the current
* owner of the mutex and awarding it to the caller.
*
* The maximally friendly way to call this function is with
* iTicksToWaitForMutex set to WAIT_FOREVER.
*
* Typically, a caller would set iTicksToWaitForMutex to some
* reasonable value, perhaps a few seconds, and would set bKillCurrentOwner
* to FALSE.
*
* Other typical use might be for a command processor to be able
* to process commands of two flavors, one flavor in which commands
* are queued to follow one another in sequence, and another in which
* any new command immediately halts any previous command and starts
* anew.  These are the cases with bKillCurrentOwner set to FALSE and
* to TRUE, respectively.
*
* If the mutex is not returned in iTicksToAllowForOperation timer
* ticks, the operation will be aborted.
*
******
*
* OOPS!  The current version of this function merely Aborts the current
* owner if bKillCurrentOwner is TRUE.  If other tasks are pended on
* the same mutex, they will still get the mutex before this one, because
* they did the semTake first!  We'll leave it this way for inc. 1,
* but it will need to be fixed in inc. 2.  It makes one ask what the
* desired behavior actually is.  I see three scenarios, kill the current
* task and begin executing immediately (this could be accomplished
* by simply deleting and recreating the semaphore, causing all pended
* tasks to return with an error code), wait for the current task to
* complete and then be next in line to get the mutex (more difficult,
* probably requires a message queue and some sort of request server),
* or just go to the end of the line and wait.
*
* P.S.  Believe it or not, it was the gory details of the spec that
* uncovered this bug for me...
*
******
*
* OOPS again!  In increment 1, we need the ability to abort commands.
* For example, processing of a Track command (Track "mode" at the
* system spec level) could go on indefinitely and must be terminated
* by another command (e.g. Off).  So, our new solution is to
* always set bKillCurrentOwner to TRUE and iTicksToWaitForMutex to
* NO_WAIT.  This avoids the problem of things queuing up, because
* any new command always kills the previous one.
*
* I'm not going to bother updating the spec for that.  Mike inspected
* it and says it's okay.  11/14/96 - BEH.
*
******
*
* IN:
* 
* INOUT:
*
* OUT:
*
* SPEC:
*
    [ this invalid -> UNDEFINED
    | called from interrupt level -> UNDEFINED
    | *this.siSemaphore invalid || *this.WatchDog.bInit != TRUE -> UNDEFINED
    | ! CanTakeSemaphoreWithinSpecifiedTimeout(*this.siSemaphore,
                                            iTicksToWaitForMutex) ->
        return S_objLib_OBJ_TIMEOUT or S_objLib_OBJ_UNAVAILABLE
    | true ->
        Calling task is blocked until SemAvailable(*this.siSemaphore) is true
        : *this.*siSemaphore.State := SEM_EMPTY
        : *this.pOwner := pNewOwner
        : [ iTicksToAllowForOperation == WAIT_FOREVER ->
              return OK
          | iTicksToAllowForOperation < 1 ->
              return S_watchDog_INVALID_TIMEOUT_VALUE
          | true ->
              *this.WatchDog.iTimeLeft = iTicksToAllowForOperation
              : *this.WatchDog.bActive := TRUE
              : return OK
          ]
    ]
*
************
*
* Ignore the spec below.  It is inc-2 related, not in inc 1.
*
    | !bKillCurrentOwnerIfTimeout || *this.pOwner == NULL ->
        return S_objLib_OBJ_TIMEOUT or S_objLib_OBJ_UNAVAILABLE
    | *this.pOwner invalid -> UNDEFINED
    | *this.*pOwner.pMutex != this -> UNDEFINED   // If the current owner
                 // of the mutex doesn't think this is the mutex he's using,
                 // this is a big inconsistency.  Shouldn't happen.  Asserting
                 // this makes some of the following spec simpler.
    | *this.*pOwner.siAbortSemaphore NULL or invalid ->
        Do nothing
    | ! CanTakeSemaphoreWithinSpecifiedTimeout(
             *this.*pOwner.siAbortSemaphore, iTicksToWaitForAbortSemaphore) ->
         return S_objLib_OBJ_TIMEOUT or S_objLib_OBJ_UNAVAILABLE
    | true ->
        OS.SemaphoreSet := OS.SemaphoreSet
                                  with *this.*pOwner.*siAbortSemaphore removed
        : *this.pOwner := NULL     // These 2 steps free pMutex
        : *this.WatchDog.bActive := FALSE
        : [ *this.*pOwner.sTaskName != NULL ->
                  return *this.*pOwner.*sTaskName storage to system heap ]
        : return *this.*pOwner storage to system heap
        : *this.pOwner := INVALID
        : // Delete the task associated with ExecuteMethod
          [ *this.*pOwner.iTaskId != NULL or ERROR,
                                            and taskDelete succeeds ->
              OS.SpawnedTasks := OS.SpawnedTasks with the task with
                                           id *this.*pOwner.iTaskId removed
              : results from BB spec of *this.*pOwner.AbortMethod()
          ]
        Calling task is blocked until SemAvailable(*this.siSemaphore) is true
        : *this.*siSemaphore.State := SEM_EMPTY
        : *this.pOwner := pNewOwner
        : [ iTicksToAllowForOperation == WAIT_FOREVER ->
              return OK
          | iTicksToAllowForOperation < 1 ->
              return S_watchDog_INVALID_TIMEOUT_VALUE
          | true ->
              *this.WatchDog.iTimeLeft = iTicksToAllowForOperation
              : *this.WatchDog.bActive := TRUE
              : return OK
          ]
    ]
    
*
* RETURNS:
*   OK if successful, or an ERROR_CODE if it fails.
*/

ERROR_CODE tMutex::Take(tMutexUser *pNewOwner, int iTicksToWaitForMutex,
                        BOOL bKillCurrentOwnerIfTimeout,
                        int iTicksToAllowForOperation)
{
  ERROR_CODE ec;

  /*******
    [ this invalid -> UNDEFINED
    | called from interrupt level -> UNDEFINED
    | *this.siSemaphore invalid ->
        OS.Task.errno := S_objLib_OBJ_ID_ERROR
        : ec := ERROR
    | iTicksToWaitForMutex != WAIT_FOREVER &&
         (iTicksToWaitForMutex == NO_WAIT && !SemAvailable(*this.siSemaphore)
          || SemAvailable(*this.siSemaphore) not true within
                                            iTicksToWaitForMutex ticks) ->
        OS.Task.errno := S_objLib_OBJ_TIMEOUT or S_objLib_OBJ_UNAVAILABLE
        : ec := ERROR
             (prior to timeout, the task is inserted into
              *this.*siSemaphore.PendedTaskList as specified
              by *this.*siSemaphore.QueuingMethod)
    | true ->
        Calling task is blocked until SemAvailable(*this.siSemaphore) is true
             (while blocked, the task is inserted into
              *this.*siSemaphore.PendedTaskList as specified
              by *this.*siSemaphore.QueuingMethod)
        : [ *this.*siSemaphore.Type == SEM_BINARY ->
               *this.*siSemaphore.State := SEM_EMPTY
          | true ->
              unfinished at this time
          ]
        : ec := OK
    ]
  ********/
  // Inc2+: Add taskIdSelf to list of tasks pending on semaphore
  ec = semTake(this->siSemaphore, iTicksToWaitForMutex);
  // Inc2+: Delete taskIdSelf from list of tasks pending on semaphore
  #ifndef INCR1
    #error Increment2UpgradeNeeded
  #endif

  /*******
    [ ec == ERROR ->
      [ (VxWorks error code (semTake) !=
                          S_objLib_OBJ_TIMEOUT or S_objLib_OBJ_UNAVAILABLE ||
                !bKillCurrentOwnerIfTimeout || *this.pOwner == NULL) ->
          return VxWorks error code (semTake)
      ]
      : [ this invalid -> UNDEFINED
        | *this.pOwner invalid -> UNDEFINED
        | true ->
            [*** *this.*pOwner.Abort() (substituted above) ***]
            : [ this invalid -> UNDEFINED
              | called from interrupt level -> UNDEFINED
              | *this.siSemaphore invalid ->
                  OS.Task.errno := S_objLib_OBJ_ID_ERROR
                  : return S_objLib_OBJ_ID_ERROR
              | NO_WAIT != WAIT_FOREVER &&
                    (NO_WAIT == NO_WAIT &&
                                 !SemAvailable(*this.siSemaphore)
                     || SemAvailable(*this.siSemaphore) not true within
                                               NO_WAIT ticks) ->
                  OS.Task.errno := S_objLib_OBJ_TIMEOUT or
                                          S_objLib_OBJ_UNAVAILABLE
                  : return S_objLib_OBJ_TIMEOUT or S_objLib_OBJ_UNAVAILABLE
              | true ->
                 Calling task is blocked until
                            SemAvailable(*this.siSemaphore) is true
                 : [ *this.*siSemaphore.Type == SEM_BINARY ->
                       *this.*siSemaphore.State := SEM_EMPTY
                   | true ->
                       unfinished at this time
                   ]
              ]
        ]
    ]
  *******/
  if (ec == ERROR) {
    ec = errnoGet();
    // Return the error unless it was a timeout, and we're supposed
    // to kill the current owner, and there is a current owner.
    if ( (ec != S_objLib_OBJ_TIMEOUT && ec != S_objLib_OBJ_UNAVAILABLE)
         || !bKillCurrentOwnerIfTimeout || this->pOwner == NULL)
      return Traceback(ec, "tMutex::Take", pNewOwner, iTicksToWaitForMutex,
                       bKillCurrentOwnerIfTimeout, iTicksToAllowForOperation);

    // Someone must own the semaphore; let's abort them.  Ignore any errors
    // except timeout, which indicates that the Abort actually didn't
    // complete in the specified time.
    ec = this->pOwner->Abort();
    if (ec == S_objLib_OBJ_TIMEOUT || ec == S_objLib_OBJ_UNAVAILABLE)
      return ec;
    
    // Now the semaphore should be available; take it again.
    ec = semTake(this->siSemaphore, NO_WAIT);
    if (ec == ERROR) {
      ec = errnoGet();
      return Traceback(ec, "tMutex::Take - Part 2", pNewOwner,
                       iTicksToWaitForMutex, bKillCurrentOwnerIfTimeout,
                       iTicksToAllowForOperation);
    }
  }

  /*******
    [ this invalid -> UNDEFINED
    | true         -> *this.pOwner := pNewOwner ]
  *******/
  // Is pOwner valid for sure here???
  this->pOwner = pNewOwner;

  /*=====
  SYNCHRONIZATION ISSUE - What if the iTicksToAllowForOperation is
         very short, and the Abort method gets called before the
         execute method has a chance to get off the ground.
         Perhaps tMutexUsers need to always be in a state of
         readiness for deletion, from the time they're first init'ed.
  DESIGN RESPONSE - use a semaphore to guard the abort process, as
         implemented in tMutexUser::Execute.  Aborts cannot take place
         now until the task running ExecuteMethod is at least spawned.
  =====*/

  /*******
     [ iTicksToAllowForOperation != WAIT_FOREVER ->
          [ this invalid -> UNDEFINED
          | &*this.WatchDog invalid -> UNDEFINED
          | *this.WatchDog.bInit == FALSE -> return S_watchDog_NOT_INITIALIZED
          | iTicksToAllowForOperation < 1 ->
                                    return S_watchDog_INVALID_TIMEOUT_VALUE
          | true ->
              *this.WatchDog.iTimeLeft = iTicksToAllowForOperation
              : *this.WatchDog.bActive := TRUE
          ]
     | true ->
          return OK
     ]
  *******/
  if (iTicksToAllowForOperation != WAIT_FOREVER) {
    ec = this->WatchDog.StartUsingTicks(iTicksToAllowForOperation);
    if (ec != OK)
      return Traceback(ec, "tMutex::Take", pNewOwner,
                       iTicksToWaitForMutex, bKillCurrentOwnerIfTimeout,
                       iTicksToAllowForOperation);
  }

  return OK;
}


/******************************************************************
* A
* tMutex::Give
*
*
* IN:
*   this
*   *this.siSempahore
*   *this.WatchDog.bInit
* 
* INOUT:
*   *this.*siSemaphore
*
* OUT:
*   *this.WatchDog.bActive
*
* SPEC:
    [ this invalid -> UNDEFINED
    | *this.siSemaphore invalid ->
                return S_objLib_OBJ_ID_ERROR
    | *this.WatchDog.bInit == FALSE ->
                return S_watchDog_NOT_INITIALIZED
    | true ->
        *this.pOwner := NULL
        : *this.WatchDog.bActive := FALSE
        : [ *this.*siSemaphore.PendedTaskList not empty ->
              First task in *this.*siSemaphore.PendedTaskList unblocks
              : First task removed from *this.*siSemaphore.PendedTaskList
          | true ->
              *this.*siSemaphore.State := SEM_FULL
          ]
        : return OK
    ]
*
*
* RETURNS:
* OK if successful, or an ERROR_CODE if it fails.
*/

ERROR_CODE tMutex::Give(void)
{
  ERROR_CODE ec, ec1;

  /********
     [ this invalid -> UNDEFINED
     | &*this.WatchDog invalid -> UNDEFINED
     | *this.WatchDog.bInit == FALSE -> ec1 := S_watchDog_NOT_INITIALIZED
     | true ->
         *this.WatchDog.bActive := FALSE
         : ec1 := OK
     ]
  ********/
  //===== Does it all work if the WD times out during the reset or Give call?
  ec1 = this->WatchDog.Reset();

  /********
    [ this invalid -> UNDEFINED
    | *this.siSemaphore invalid ->
        OS.Task.errno := S_objLib_OBJ_ID_ERROR
        : ec := ERROR
    | *this.*siSemaphore.PendedTaskList not empty ->
        First task in *this.*siSemaphore.PendedTaskList unblocks
        : First task removed from *this.*siSemaphore.PendedTaskList
        : ec := OK
    | true ->
	[ *this.*siSemaphore.Type == SEM_BINARY ->
            *this.*siSemaphore.State := SEM_FULL
        | true ->
            unfinished at this time
        ]
        : return OK
    ]
  ********/ 
  ec = semGive(this->siSemaphore);

  /*******
    [ this invalid -> UNDEFINED
    | true         -> *this.pOwner := NULL ]
  *******/
  this->pOwner = NULL;

  /*******
    [ ec == OK -> return ec1
    | true -> return OS.Task.errno
    ]
  ********/
  ec = (ec == OK ? ec1 : errnoGet());
  
  if (ec != OK)
    return Traceback(ec, "tMutex::Give");

  return OK;
}


/******************************************************************
* A
* tMutex::IsOwner - Is the specified tMutexUser this tMutex's owner?
*
* IN:
*   this
*   *this.pOwner
*   pPossibleOwner
* 
* SPEC:
    [ this invalid -> UNDEFINED
    | *this.pOwner == NULL -> return FALSE
    | *this.pOwner != pPossibleOwner -> return FALSE
    | true -> return TRUE
    ]
*
* RETURNS:
*   TRUE or FALSE
*/

BOOL tMutex::IsOwner(tMutexUser *pPossibleOwner)
{
  if (this->pOwner == NULL)
    return FALSE;
  if (this->pOwner != pPossibleOwner)
    return FALSE;

  return TRUE;
}


/******************************************************************
* A
* tMutex::WatchDogHandler
*
* Note that this->pOwner->pMutex should equal "this" if all is
* well.  This fact is used in a number of places to simplify the
* spec, but only after making sure it is a valid assumption.
*
*
* IN:
*   this
*   *this.pOwner
*   *this.*pOwner.pMutex
*   *this.*pOwner.siAbortSempahore
*   *this.*pOwner.*siAbortSempahore
*   iTicksToWaitForAbortSemaphore  (defaults to MUTEX_DEFAULT_TIMEOUT)
*   *this.siSemaphore
*   *this.WatchDog.bInit
*   *this.*pOwner.sTaskName
*   *this.*pOwner.iTaskId
*
* INOUT:
*   OS.SemaphoreSet
*   OS.SpawnedTasks
*   *this.*pOwner.*pMutex.pOwner ( :== *this.pOwner)
*   *this.*siSemaphore
*   System heap
*
* OUT:
*   *this.pOwner (becomes invalid if it works!)
*   *this.WatchDog.Active
*
* OTHER INOUTS:
*   Ins/outs due to the call to AbortMethod.
* SPEC:
*
    [ this invalid -> UNDEFINED
    | *this.siSemaphore invalid || *this.WatchDog.bInit != TRUE -> UNDEFINED
    | *this.pOwner == NULL -> return OK
    | *this.pOwner invalid -> UNDEFINED
    | called from interrupt level -> UNDEFINED
    | *this.*pOwner.pMutex invalid -> UNDEFINED
    | *this.*pOwner.siAbortSemaphore == NULL ->
        return S_mutexUser_ABORT_ATTEMPTED_ON_INACTIVE_MUTEX_USER_OBJECT
    | *this.*pOwner.siAbortSemaphore invalid (i.e., has been deleted or
                                is deleted while waiting on semaphore) ->
        return OK
    | ! CanTakeSemaphoreWithinSpecifiedTimeout(*this.*pOwner.siAbortSemaphore,
                                            iTicksToWaitForAbortSemaphore) ->
          return S_objLib_OBJ_TIMEOUT or S_objLib_OBJ_UNAVAILABLE
    | true ->
        OS.SemaphoreSet := OS.SemaphoreSet
                                 with *this.*pOwner.*siAbortSemaphore removed
        : // Free pMutex if we are owner, else report an error
          [ *this.*pOwner.*pMutex.pOwner == *this.pOwner -> 
              // Give back the mutex
              *this.pOwner := NULL
              : *this.WatchDog.bActive := FALSE
              : // Start up anyone blocked on mutex, or mark mutex as avail
                [ *this.*siSemaphore.PendedTaskList not empty ->
                    First task in *this.*siSemaphore.PendedTaskList unblocks
                    : First task removed from *this.*siSemaphore.PendedTaskList
                | true ->
                    *this.*siSemaphore.State := SEM_FULL
                ]
          ]
        : [ *this.*pOwner.sTaskName != NULL ->
                return *this.*pOwner.*sTaskName storage to system heap ]
        : return *this.*pOwner storage to system heap
        : *this.pOwner := INVALID
        : // Delete the task associated with ExecuteMethod
          [ *this.*pOwner.iTaskId != NULL or ERROR, and taskDelete fails ->
              return appropriate error code
          | true ->
              OS.SpawnedTasks := OS.SpawnedTasks with the task with
                                              id *this.*pOwner.iTaskId removed
              : results from BB spec of *this.*pOwner.AbortMethod()
              : return, with return code = return code from
                                              *this.*pOwner.AbortMethod()
          ]
    ]   
* 
* RETURNS:
*
*/

ERROR_CODE tMutex::WatchDogHandler(void)
{
  ERROR_CODE ec = OK;

  if (this->pOwner != NULL)
     ec = this->pOwner->Abort();
  
  if (ec != OK) {
    // Do Something!
    return Traceback(ec, "tMutex::WatchDogHandler");
  }

  return OK;
}


/******************************************************************
* A
* tMutex::~tMutex
*
* IN:
* 
* INOUT:
*
* OUT:
*
* SPEC:  (inludes spec for tTaskLevelWatchDog::~tTaskLevelWatchDog)
*        This spec assumes the function is called via delete -
*        therefore, any modifications to *this have been removed
*
    [ this invalid -> UNDEFINED
    | *this.siSemaphore invalid || *this.WatchDog.bInit != TRUE -> UNDEFINED
    | true ->
      // Abort stuff
      [ *this.pOwner != NULL ->
        [ *this.pOwner invalid -> UNDEFINED
        | called from interrupt level -> UNDEFINED
        | *this.*pOwner.pMutex invalid -> UNDEFINED
        | ! (*this.*pOwner.siAbortSemaphore == NULL ||
                  *this.*pOwner.siAbortSemaphore invalid (i.e., has been
                          deleted or is deleted while waiting on semaphore) ||
                  ! CanTakeSemaphoreWithinSpecifiedTimeout(
                          *this.*pOwner.siAbortSemaphore,
                                     iTicksToWaitForAbortSemaphore)) ->
            OS.SemaphoreSet := OS.SemaphoreSet
                                 with *this.*pOwner.*siAbortSemaphore removed
            : [ *this.*pOwner.sTaskName != NULL ->
                    return *this.*pOwner.*sTaskName storage to system heap ]
            : // Delete the task associated with ExecuteMethod
              [ *this.*pOwner.iTaskId != NULL or ERROR,
                                               and taskDelete succeeds ->
                  OS.SpawnedTasks := OS.SpawnedTasks with the task with
                                              id *this.*pOwner.iTaskId removed
                  : results from BB spec of *this.*pOwner.AbortMethod()
              ]
        | true ->
           Do nothing
        ]
      ]
      : All tasks in *this.*siSemaphore.PendedTaskList unblock (with error
                                                                return codes)
      : OS.SemaphoreSet := OS.SemaphoreSet with *this.*siSemaphore removed
      : VxWorks.WdTimersAvailable :+= 1
    ]
*
* RETURNS:
* 
*/

tMutex::~tMutex(void)
{
  ERROR_CODE ec;

  // Abort anyone using this mutex first
  if (this->pOwner != NULL)
    this->pOwner->Abort();
  
  if (siSemaphore != NULL) {
    ec = semDelete(siSemaphore);
    if (ec != OK) {
      // Do something!
    }
  }
}


/******************************************************************
* A
* tMutexUser::tMutexUser
*
*
* IN:
*   this
*   pMutexToUse
*   pErrorClientToUse
*   sTaskNameToUse
*   *sTaskNameToUse (if sTaskName not NULL)
* INOUT:
*   system heap
* OUT:
*   *this.pMutex
*   *this.pErrorClient
*   *this.iTaskId
*   *this.sTaskName
*   *this.iTaskOptions
*   *this.iTaskStackSize
*   *this.iTaskPriority
*   *this.siAbortSemaphore
*
* SPEC:
*
   [ this or pMutexToUse or pGizmoToUse invalid -> UNDEFINED
   | sTaskNameToUse != NULL && sTaskName invalid or not a pointer to a
                             valid (null-terminated) string -> UNDEFINED
   | true ->
       *this.pMutex := pMutexToUse
       : *this.pErrorClient := *pGizmoToUse.pRteError
       : *this.iTaskId := NULL
       : [ sTaskNameToUse == NULL ->
             *this.sTaskName := address of a new copy of "MutexTask"
         | insufficient heap for string ->
             *this.sTaskName := NULL
         | true ->
             *this.sTaskName := address of a new copy of *sTaskNameToUse
         ]
       : *this.iTaskOptions     := MUTEX_TASK_OPTIONS
       : *this.iTaskStackSize   := MUTEX_TASK_STACK_SIZE
       : *this.iTaskPriority    := MUTEX_TASK_PRIORITY
       : *this.siAbortSemaphore := NULL
   ] 
*
* RETURNS:
*   Nothing
*/

tMutexUser::tMutexUser(tMutex *pMutexToUse, tGizmo *pGizmoToUse,
                       char *sTaskNameToUse)
{

  /*****
    [[U ::GetErrorStream()]] 
  *****/
  tErrorStreamSender &ErrorStream = GetErrorStream();

  /*****
    [ sDefaultName := address of "MutexTask" ]
  *****/
  char sDefaultName[] = "MutexTask";

  /*****
    [ this or pMutexToUse or pGizmoToUse invalid -> UNDEFINED
    | true ->
       *this.pMutex := pMutexToUse
       : *this.pErrorClient := GetErrorStream()
       : *this.iTaskId := NULL
    ]
  *****/
  this->pMutex = pMutexToUse;
  this->pErrorClient = &ErrorStream;
  this->iTaskId = NULL;

  /*****
    [ sTaskNameToUse == NULL ->
        sTaskNameToUse := sDefaultName ]
  *****/
  if (sTaskNameToUse == NULL)
    sTaskNameToUse = sDefaultName;

  /*****
    [ sTaskNameToUse invalid or not a pointer to a
                    valid (null-terminated) string -> UNDEFINED
    | insufficient heap for string ->
        *this.sTaskName := NULL
    | true ->
        *this.sTaskName := address of a new copy of *sTaskNameToUse
    ]
  *****/
  this->sTaskName = (char *) malloc(strlen(sTaskNameToUse)+1);
  if (this->sTaskName == NULL) {
    // Do something
  }
  else
    strcpy(this->sTaskName, sTaskNameToUse);

  /*****
     [ this invalid -> UNDEFINED
     | true ->
         *this.iTaskOptions     := MUTEX_TASK_OPTIONS
         : *this.iTaskStackSize := MUTEX_TASK_STACK_SIZE
         : *this.iTaskPriority  := MUTEX_TASK_PRIORITY
     ]
  *****/
  this->SetTaskOptions();   // Has default arguments!

  /*****
    [ this invalid -> UNDEFINED
    | true ->
        *this.siAbortSemaphore := NULL
    ]
  *****/
  this->siAbortSemaphore = NULL;
}


/******************************************************************
* A
* tMutexUser::SetTaskOptions - set options for taskSpawn to Use
*
* This function sets the options that will be used by taskSpawn
* when a tMutexUser ExecuteMethod is spawned.  It is normally
* called automatically and is not required to be called by the
* user.  However, the user may wish to call this method if he
* wants something other than the default task options.
*
* IN:
*   this
*   options    (defaults to MUTEX_TASK_OPTIONS)
*   stackSize  (defaults to MUTEX_TASK_STACK_SIZE)
*   priority   (defaults to MUTEX_TASK_PRIORITY)
* INOUT:
* OUT:
*   *this.iTaskOptions
*   *this.iTaskStackSize
*   *this.iTaskPriority
*
* SPEC:
*
   [ this invalid -> UNDEFINED
   | true ->
       *this.iTaskOptions     := options
       : *this.iTaskStackSize := stackSize
       : *this.iTaskPriority  := priority
       : return OK
   ] 
*
* RETURNS:
*   OK
*/

ERROR_CODE tMutexUser::SetTaskOptions(int options,
                                      int stackSize,
                                      int priority)
{
  iTaskOptions   = options;
  iTaskStackSize = stackSize;
  iTaskPriority  = priority;

  return OK;
}


/******************************************************************
* A
* tMutexUser::Execute - perform an operation requiring a mutex
*
* The purpose of this function is to spawn a task to execute the
* virtual function ExecuteMethod.  This requires acquiring the mutual
* exclusion object specified as pMutexToUse when this object was
* created.  This function will wait a maximum of iTicksToWaitForMutex
* timer ticks for the mutex to become available.  Valid values for
* iTicksToWait include the special VxWorks values NO_WAIT and
* WAIT_FOREVER, in addition to positive integers.
*
* If the mutex is not acquired in the specified time, this function
* will return with an error, unless bKillCurrentOwnerIfTimeout is
* TRUE.  In this case, the mutex will be "stolen" from its current
* owner, resulting in the deletion of the owner's task that was
* using the mutex.
*
* Once the mutex has been acquired, a task is spawned to execute
* ExecuteMethod and return the mutex when done.  Before the task is spawned,
* a watchdog timer is started, with a timeout to occur after
* iTicksToAllowForOperation timer ticks (again, WAIT_FOREVER is a
* valid value).  If the watchdog timer expires before DoIt completes,
* the task running ExecuteMethod will be deleted and the mutex will be
* made available for another user.
*
* The calling task's priority must be higher (have a lower value)
* than MUTEX_TASK_PRIORITY in order to ensure that the caller gets to
* complete his call to Execute prior to the spawned task's being allowed
* to run.
*
* IN:
*   iTicksToWaitForMutex       (defaults to NO_WAIT)
*   bKillCurrentOwnerIfTimeout (defaults to TRUE)
*   iTicksToAllowForOperation  (defaults to MUTEX_DEFAULT_TIMEOUT)
*   MUTEX_TASK_PRIORITY
*   priority of calling task
*   this
*   (available memory)
*
* INOUT:
*   OS.SemaphoreSet
*   OS.SpawnedTasks
*
* OUT:
*   *this.siAbortSemaphore
*   *this.iTaskId
*
* SPEC: (Note, on failure, the function may perform slightly differently
*        from what's spec'ed, in that there may be additional assignments,
*        allocations, etc.)
*
    [ MUTEX_TASK_PRIORITY <= priority of calling task ->
        return S_mutexUser_MUTEX_TASK_PRIORITY_TOO_HIGH
    | this invalid -> UNDEFINED
    | memory for semaphore cannot be allocated ->
        *this.siAbortSemaphore = NULL
        : return VxWorks errno (semBCreate)
    | Task cannot be created || Insufficient memory to spawn task ->
        *this.iTaskId := ERROR
        : *this.siAbortSemaphore := INVALID
        : return VxWorks errno (taskSpawn)   
    | true ->
        OS.SemaphoreSet := OS.SemaphoreSet U
                         NewBinarySemaphore(SEM_Q_FIFO, SEM_FULL)
        : *this.siAbortSemaphore := NewBinarySemaphore(SEM_Q_FIFO, SEM_FULL).Id
        : OS.SpawnedTasks := OS.SpawnedTasks U
                    { *this.DoTheOperation(iTicksToWaitForMutex,
                                           bKillCurrentOwnerIfTimeout,
                                           iTicksToAllowForOperation), with
                      priority *this.iTaskPriority,
                      options *this.iTaskOptions,
                      stack size *this, iTaskStackSize,
                      and name *this.sTaskName }
        : *this.iTaskId := new task ID from VxWorks
        : return OK
    ]
*
* RETURNS:
*   OK or S_mutexUser_MUTEX_TASK_PRIORITY_TOO_HIGH
*      or VxWorks errno (semBCreate) or VxWorks errno (taskSpawn)
*/

ERROR_CODE tMutexUser::Execute(int iTicksToWaitForMutex,
                               BOOL bKillCurrentOwnerIfTimeout,
                               int iTicksToAllowForOperation)
{
  int iThisTaskPriority;
  ERROR_CODE ec;

  /*****
    [ iThisTaskPriority := priority of calling task ]
    // Note: According to manual, the if below should never be true
    // (result is never ERROR).  That's why there's no spec for
    // that case.  There are no known cases where it returns ERROR.
    // This is just defensive programming.
  *****/
  // Ensure that spawned task is at a lower priority than this task
  if (taskPriorityGet(taskIdSelf(), &iThisTaskPriority) == ERROR)
    return Traceback(errnoGet(), "tMutexUser::Execute",
                     iTicksToWaitForMutex, bKillCurrentOwnerIfTimeout);
  
  /*****
    [ MUTEX_TASK_PRIORITY <= iThisTaskPriority ->
        return S_mutexUser_MUTEX_TASK_PRIORITY_TOO_HIGH ]
  *****/
  if (MUTEX_TASK_PRIORITY <= iThisTaskPriority)
    return Traceback(S_mutexUser_MUTEX_TASK_PRIORITY_TOO_HIGH,
                     "tMutexUser::Execute",
                     iTicksToWaitForMutex, bKillCurrentOwnerIfTimeout);
      
  /******
    [ this invalid -> UNDEFINED
    | memory for semaphore cannot be allocated ->
       *this.siAbortSemaphore = NULL
       : OS.Task.errno := VxWorks errno (semBCreate)
    | true ->
       OS.SemaphoreSet := OS.SemaphoreSet U
                        NewBinarySemaphore(SEM_Q_FIFO, SEM_EMPTY)
       : *this.siAbortSemaphore := NewBinarySemaphore(SEM_Q_FIFO, SEM_EMPTY).Id
    ]
   ******/
  // Create the VxWorks semaphore to be used to control access to the Abort
  // routine.
  //
  // The semaphore is created EMPTY to preserve State Invariant 1
  // What about priority inversion?  Are we immune?
  this->siAbortSemaphore = semBCreate(SEM_Q_FIFO, SEM_EMPTY);

  /*****
    [ this invalid -> UNDEFINED
    | *this.siAbortSemaphore == NULL ->
         return VxWorks errno (semBCreate)
    ]
  *****/
  if (this->siAbortSemaphore == NULL)
    return Traceback(errnoGet(), "tMutexUser::Execute",
                     iTicksToWaitForMutex, bKillCurrentOwnerIfTimeout);

  /*****
    [ ! (&*this.DoTheOperation is a pointer to valid C function ||
          (&*this.DoTheOperation is a valid pointer to a C++ method &&
           this is a valid pointer to a C++ object with method
           DoTheOperation)) -> UNDEFINED
    | Task cannot be created || Insufficient memory to spawn task ->
           OS.Task.errno := VxWorks errno (taskSpawn)
           : *this.iTaskId := ERROR
    | true ->
        OS.SpawnedTasks := OS.SpawnedTasks U
                    { *this.DoTheOperation(iTicksToWaitForMutex,
                                           bKillCurrentOwnerIfTimeout,
                                           iTicksToAllowForOperation), with
                      priority *this.iTaskPriority,
                      options *this.iTaskOptions,
                      stack size *this, iTaskStackSize,
                      and name *this.sTaskName }
        : *this.iTaskId := new task ID from VxWorks
    ]
  *****/
  this->iTaskId = taskSpawn(sTaskName, this->iTaskPriority,
                            this->iTaskOptions, this->iTaskStackSize,
                            (FUNCPTR) &DoTheOperation, (int) this,
                            iTicksToWaitForMutex,
                            bKillCurrentOwnerIfTimeout,
                            iTicksToAllowForOperation,
                            0,0,0,0,0,0);

  /*****
    [ this invalid -> UNDEFINED
    | *this.iTaskId == ERROR ->
        : [ *this.siAbortSempahore invalid ->
              OS.Task.errno := S_objLib_OBJ_ID_ERROR
          | true ->
              All tasks in *this.*siAbortSempahore.PendedTaskList unblock
              : OS.SemaphoreSet := OS.SemaphoreSet with
                                        *this.*siAbortSempahore removed
              : *this.siAbortSempahore := invalid
          ]
        : return VxWorks errno (taskSpawn)
    ]
  *****/
  if (this->iTaskId == ERROR) {
    ec = errnoGet();
    semDelete(this->siAbortSemaphore);
    return Traceback(ec, "tMutexUser::Execute",
                     iTicksToWaitForMutex, bKillCurrentOwnerIfTimeout);
  }

  /*****
    [ *this.siAbortSemaphore invalid ->
        OS.Task.errno := S_objLib_OBJ_ID_ERROR
        : return VxWorks errno (semGive)
    | *this.*siAbortSempahore.PendedTaskList not empty
        First task in *this.*siAbortSempahore.PendedTaskList unblocks
        : First task removed from *this.*siAbortSempahore.PendedTaskList
        : return OK
    | true ->
        [ *this.*siAbortSempahore.Type == SEM_BINARY ->
            *this.*siAbortSempahore.State := SEM_FULL
        | true ->
            unfinished at this time
        ]
        : return OK
    ]
  *****/
  // Mark the semaphore as full.  We're now in the "full" case of
  //                                            State Invariant 1.
  if (semGive(this->siAbortSemaphore) != OK)
    return Traceback(errnoGet(), "tMutexUser::Execute",
                     iTicksToWaitForMutex, bKillCurrentOwnerIfTimeout);
  
  return OK;
}


/******************************************************************
* A
* tMutexUser::DoTheOperation
*
*
* IN:
*   iTicksToWaitForMutex
*   iTicksToAllowForOperation
*   *this.pMutex
*   Calling context (task or interrupt)
*   *this.pErrorClient
*   *this.*pMutex.siSemaphore
*   *this.*pMutex.WatchDog.bInit
*   *this.siAbortSemaphore
*   *this.*pErrorClient.miQid
*   RELAY_Q_SEND_MAXWAIT
*   *this.sTaskName
*   <Spec for ExecuteMethod>
*
* INOUT:
*   this
*   VxWorks queue *this.*pErrorClient.miQid
*   system heap
*   OS.SpawnedTasks
*   *this.*siAbortSemaphore.PendedTaskList
*   OS.SemaphoreSet
*
* OUT:
*   *this.*pMutex.*siSemaphore.State
*   *this.*pMutex.pOwner
*   *this.*pMutex.WatchDog.bActive
*   *this.*pMutex.WatchDog.iTimeLeft
*
*
* SPEC:
*
* We list two sets of spec here.  One is the net effect of the
* whole function.  The second is actually still a clear box in two
* parts.  Part 1 is the part of the function up to starting the
* watchdog and calling ExecuteMethod; Part 2 is Cleanup and exit.
*
* The reason for this is that we expect that this function may be
* terminated in between parts 1 and 2 by a watchdog timeout.  So,
* in order to verify that the sequence Part1/Watchdog Timeout performs
* as we desire, we need the spec broken into these two parts.
*
* NET EFFECT:
*
    [ this invalid -> UNDEFINED
    | *this.pMutex invalid -> UNDEFINED
    | called from interrupt level -> UNDEFINED
    | *this.pErrorClient invalid -> UNDEFINED
    | *this.*pMutex.siSemaphore invalid ||
                    *this.*pMutex.WatchDog.bInit != TRUE -> UNDEFINED
    | *this.siAbortSemaphore invalid -> UNDEFINED
    | not started via taskSpawn -> UNDEFINED
    | ! CanTakeSemaphoreWithinSpecifiedTimeout(*this.*pMutex.siSemaphore,
                                                     iTicksToWaitForMutex) ->
           [ ErrorQueueOkay(*this.*pErrorClient.miQid,
                           RELAY_Q_SEND_MAXWAIT, sizeof(ERROR_INFO)) ->
                Append S_objLib_OBJ_TIMEOUT or S_objLib_OBJ_UNAVAILABLE
                           message to VxWorks queue *this.*pErrorClient.miQid
           ]
           : [ *this.sTaskName != NULL ->
               return *this.*sTaskName storage to system heap
             ]
           : return *this storage to system heap
           : this := invalid
           : OS.SpawnedTasks := OS.SpawnedTasks with this task removed
     | true ->
         [ iTicksToAllowForOperation != WAIT_FOREVER &&
                                  iTicksToAllowForOperation < 1 ->
             [ ErrorQueueOkay(*this.*pErrorClient.miQid,
                            RELAY_Q_SEND_MAXWAIT, sizeof(ERROR_INFO)) ->
                  Append S_watchDog_INVALID_TIMEOUT_VALUE message to VxWorks
                                           queue *this.*pErrorClient.miQid
             ]
         | true ->
             : <Spec for *this.ExecuteMethod()>
             : [ return value of *this.ExecuteMethod() != OK &&
                        ErrorQueueOkay(*this.*pErrorClient.miQid,
                              RELAY_Q_SEND_MAXWAIT, sizeof(ERROR_INFO)) ->
                   Append appropriate message to VxWorks queue
                                            *this.*pErrorClient.miQid
               ]
         ]
         : *this.*pMutex.*siSemaphore.State := SEM_FULL
         : *this.*pMutex.pOwner := NULL
         : *this.*pMutex.WatchDog.bActive := FALSE
         : *this.*pMutex.WatchDog.iTimeLeft := anything
         : All tasks in *this.*siAbortSemaphore.PendedTaskList unblock
         : OS.SemaphoreSet := OS.SemaphoreSet
                                       with *this.*siAbortSemaphore removed
         : [ *this.sTaskName != NULL ->
               return *this.*sTaskName storage to system heap
           ]
         : return *this storage to system heap
         : this := invalid
         : OS.SpawnedTasks := OS.SpawnedTasks with this task removed
     ]
*
* PART 1:
*
    [ this invalid -> UNDEFINED
    | *this.pMutex invalid -> UNDEFINED
    | called from interrupt level -> UNDEFINED
    | *this.pErrorClient invalid -> UNDEFINED
    | *this.*pMutex.siSemaphore invalid ||
                    *this.*pMutex.WatchDog.bInit != TRUE -> UNDEFINED
    | not started via taskSpawn -> UNDEFINED
    | ! CanTakeSemaphoreWithinSpecifiedTimeout(*this.*pMutex.siSemaphore,
                                                     iTicksToWaitForMutex) ->
          [ ErrorQueueOkay(*this.*pErrorClient.miQid,
                          RELAY_Q_SEND_MAXWAIT, sizeof(ERROR_INFO)) ->
               Append S_objLib_OBJ_TIMEOUT or S_objLib_OBJ_UNAVAILABLE
                          message to VxWorks queue *this.*pErrorClient.miQid
          ]
          : OS.SpawnedTasks := OS.SpawnedTasks with this task removed
    | true ->
        This task is blocked until
                              SemAvailable(*this.*pMutex.siSemaphore) is true
        : *this.*pMutex.*siSemaphore.State := SEM_EMPTY
        : *this.*pMutex.pOwner := this
        : [ iTicksToAllowForOperation != WAIT_FOREVER &&
                                  iTicksToAllowForOperation < 1 ->
              : [ ErrorQueueOkay(*this.*pErrorClient.miQid,
                            RELAY_Q_SEND_MAXWAIT, sizeof(ERROR_INFO)) ->
                   Append S_watchDog_INVALID_TIMEOUT_VALUE message to VxWorks
                                             queue *this.*pErrorClient.miQid
                ]
          | true ->
              *this.*pMutex.WatchDog :=
                 ActivateWatchDogIfNotForeverTimeout(*this.*pMutex.WatchDog,
                                                     iTicksToAllowForOperation)
              : <Spec for *this.ExecuteMethod()>
              : [ return value of *this.ExecuteMethod() != OK &&
                       ErrorQueueOkay(*this.*pErrorClient.miQid,
                                RELAY_Q_SEND_MAXWAIT, sizeof(ERROR_INFO)) ->
                   Append appropriate message to VxWorks queue
                                            *this.*pErrorClient.miQid
          ]
    ]
*
* PART 2:
*   Actually, it's not necessary to separately spec this part to do
*   the verification we need, so I'm going to be lazy and not include it.
*
* RETURNS:
*   OK if successful, or an ERROR_CODE if it fails.
*   However, if spawned as a task as designed, the return code is ignored.
*/

ERROR_CODE tMutexUser::DoTheOperation(int iTicksToWaitForMutex,
                               BOOL bKillCurrentOwnerIfTimeout,
                               int iTicksToAllowForOperation)
{
  ERROR_CODE ecTake, ecExecute, ecCleanup;

  /***** [ ecExecute := OK ] *****/
  ecExecute = OK;

  /*****
    [ this invalid -> UNDEFINED
    | *this.pMutex invalid -> UNDEFINED
    | called from interrupt level -> UNDEFINED
    | *this.*pMutex.siSemaphore invalid ||
                    *this.*pMutex.WatchDog.bInit != TRUE -> UNDEFINED
    | ! CanTakeSemaphoreWithinSpecifiedTimeout(*this.*pMutex.siSemaphore,
                                                     iTicksToWaitForMutex) ->
        ecTake := S_objLib_OBJ_TIMEOUT or S_objLib_OBJ_UNAVAILABLE
    | true ->
        Calling task is blocked until
                              SemAvailable(*this.*pMutex.siSemaphore) is true
        : *this.*pMutex.*siSemaphore.State := SEM_EMPTY
        : *this.*pMutex.pOwner := this
        : [ iTicksToAllowForOperation == WAIT_FOREVER ->
              ecTake := OK
          | iTicksToAllowForOperation < 1 ->
              ecTake := S_watchDog_INVALID_TIMEOUT_VALUE
          | true ->
              *this.*pMutex.WatchDog.iTimeLeft = iTicksToAllowForOperation
              : *this.*pMutex.WatchDog.bActive := TRUE
              : ecTake := OK
          ]
    ]
  *****/
  ecTake = this->pMutex->Take(this, iTicksToWaitForMutex,
                          bKillCurrentOwnerIfTimeout,
                          iTicksToAllowForOperation);

  /*****
    [ this invalid -> UNDEFINED
    | *this.pErrorClient invalid -> UNDEFINED
    | ecTake != OK ->
        [ ErrorQueueOkay(*this.*pErrorClient.miQid,
                         RELAY_Q_SEND_MAXWAIT, sizeof(ERROR_INFO)) ->
          Append appropriate message to VxWorks queue *this.*pErrorClient.miQid
        ]
        : [ *this.sTaskName != NULL ->
              return *this.*sTaskName storage to system heap
          ]
        : return *this storage to system heap
        : this := invalid
        : [ spawned as a task ->
              OS.SpawnedTasks := OS.SpawnedTasks with this task removed
          | true ->
              return ecTake
          ]
    ]
  *****/
  // Use tErrorClient::Put instead of traceback
  if (ecTake != OK) {
    this->pErrorClient->Put(ecTake);
    delete this;
    return ecTake;
  }

  /*****
    [ this invalid -> UNDEFINED
    | *this.pErrorClient invalid -> UNDEFINED
    | true ->
        <Spec for *this.ExecuteMethod()>
        : ecExecute := return value of *this.ExecuteMethod()
        : [ return value of *this.ExecuteMethod() != OK &&
                   ErrorQueueOkay(*this.*pErrorClient.miQid,
                              RELAY_Q_SEND_MAXWAIT, sizeof(ERROR_INFO)) ->
             Append appropriate message to VxWorks queue
                                          *this.*pErrorClient.miQid
          ]
    ]
  *****/
  // We've acquired the semaphore, now do the requested operation
  // Inc 2+: Put in an exception handler that calls cleanup
  ecExecute = this->ExecuteMethod();
  if (ecExecute != OK) {
    this->pErrorClient->Put(ecExecute);
  }

  /*****
    [ this invalid -> UNDEFINED
    | *this.pMutex invalid -> UNDEFINED
    | *this.*pMutex.siSemaphore invalid ||
                       *this.*pMutex.WatchDog.bInit != TRUE -> UNDEFINED
    | true->
       [ *this.*pMutex.pOwner == this && != NULL ->  // Give back the mutex
           *this.*pMutex.pOwner := NULL
           : *this.*pMutex.WatchDog.bActive := FALSE
           : [ *this.*pMutex.*siSemaphore.PendedTaskList not empty ->
                 // Note that, if we succeed in acquiring the semaphore,
                 // the PendedTaskList was, in fact, empty.  If we did
                 // not succeed, then we will not be the pOwner, so we
                 // never get here.  Net effect will be to set SEM_FULL
                 // iff we succeeded in acquiring the semaphore, so this
                 // block will simplify in the net effect.
                 First task in *this.*pMutex.*siSemaphore.PendedTaskList
                                                                  unblocks
                 : First task removed from
                                 *this.*pMutex.*siSemaphore.PendedTaskList
             | true ->
                 *this.*pMutex.*siSemaphore.State := SEM_FULL
             ]
      ]
      : [ *this.siAbortSemaphore not invalid ->
            All tasks in *this.*siAbortSemaphore.PendedTaskList unblock
            : OS.SemaphoreSet := OS.SemaphoreSet
                                      with *this.*siAbortSemaphore removed
        ]
      : [ *this.sTaskName != NULL ->
            return *this.*sTaskName storage to system heap
        ]
      : return *this storage to system heap
      : this := invalid
      : ecCleanup := OK
    ]
  *****/
  // Return the mutex and turn off the watchdog
  ecCleanup = this->CleanUp();

  /*****
    [ spawned as a task ->
        OS.SpawnedTasks := OS.SpawnedTasks with this task removed
    | ecTake != OK -> return ecTake
    | ecExecute != OK -> return ecExecute
    | true -> return ecCleanup
    ]
  *****/
  return (ecTake != OK ? ecTake : (ecExecute != OK ? ecExecute : ecCleanup));
}


/******************************************************************
* A
* tMutexUser::Abort
*
* This function cleanly performs a hostile shutdown of a mutex
* user.  This shutdown could come as a result of a watchdog timeout
* or of another mutex user doing a Take with bKillCurrentOwnerIfTimeout
* set to TRUE.
*
* Because multiple sources might simultaneously attempt an abort, it
* is necessary to allow for concurrent execution.  The abort semaphore
* does this.  The first task to call Abort takes the semaphore and then
* deletes in (in CleanUp) when the Abort is complete.  This causes any
* other Aborters that have started an Abort (and have blocked on the
* semaphore) to unblock with an error return code, telling them that
* someone else successfully completed the abort.
*
* Once Abort has the semaphore in hand, it deletes the spawned
* task which has been using the mutex, and then it calls CleanUp
* to return the mutex, delete the abort semaphore, and finally delete
* the previous mutex user's object itself.
*
* IN:
*   this
*   *this.pMutex
*   *this.siAbortSempahore
*   *this.*siAbortSempahore
*   iTicksToWaitForAbortSemaphore  (defaults to MUTEX_DEFAULT_TIMEOUT)
*   *this.*pMutex.siSemaphore
*   *this.*pMutex.WatchDog.bInit
*   *this.sTaskName
*   *this.iTaskId
* 
* INOUT:
*   Message queue *this.*pErrorClient.miQid
*   OS.SemaphoreSet
*   OS.SpawnedTasks
*   *this.*pMutex.pOwner
*   *this.*pMutex.*siSemaphore
*   System heap
*
* OUT:
*   this (becomes invalid if it works!)
*   *this.*pMutex.WatchDog.Active
*
* OTHER INOUTS:
*   Ins/outs due to the call to AbortMethod.
*
* SPEC:
*
    [ this invalid -> UNDEFINED
    | *this.pMutex invalid -> UNDEFINED
    | *this.*pMutex.siSemaphore invalid ||
                       *this.*pMutex.WatchDog.bInit != TRUE -> UNDEFINED
    | called from interrupt level -> UNDEFINED
    | *this.siAbortSemaphore == NULL ->
        return S_mutexUser_ABORT_ATTEMPTED_ON_INACTIVE_MUTEX_USER_OBJECT
    | *this.siAbortSemaphore invalid (i.e., has been deleted or is deleted
                                      while waiting on semaphore) ->
        return OK
    | iTicksToWaitForAbortSemaphore != WAIT_FOREVER &&
            (iTicksToWaitForAbortSemaphore == NO_WAIT &&
             ! SemAvailable(*this.siAbortSemaphore) ||
             SemAvailable(*this.siAbortSemaphore) not true within
                                iTicksToWaitForAbortSemaphore ticks) ->
        return S_objLib_OBJ_TIMEOUT or S_objLib_OBJ_UNAVAILABLE
    | true ->
        OS.SemaphoreSet := OS.SemaphoreSet
                                 with *this.*siAbortSemaphore removed
        : // Free pMutex if we are owner, else report an error
          [ *this.*pMutex.pOwner == this && != NULL ->  // Give back the mutex
              *this.*pMutex.pOwner := NULL
              : *this.*pMutex.WatchDog.bActive := FALSE
              : // Start up anyone blocked on mutex, or mark mutex as avail
                [ *this.*pMutex.*siSemaphore.PendedTaskList not empty ->
                     First task in *this.*pMutex.*siSemaphore.PendedTaskList
                                                                     unblocks
                     : First task removed from
                                *this.*pMutex.*siSemaphore.PendedTaskList
                | true ->
                    *this.*pMutex.*siSemaphore.State := SEM_FULL
                ]
          ]
        : [ *this.sTaskName != NULL -> return *this.*sTaskName storage
                                                         to system heap ]
        : return *this storage to system heap
        : this := invalid
        : // Delete the task associated with ExecuteMethod
          [ *this.iTaskId != NULL or ERROR, and taskDelete fails ->
              return appropriate error code
          | true ->
              OS.SpawnedTasks := OS.SpawnedTasks with the task with
                                                  id *this.iTaskId removed
              : results from BB spec of *this.AbortMethod()
              : return, with return code = return code from *this.AbortMethod()
          ]
    ]  
*
*
* RETURNS:
*   OK if successful, or an ERROR_CODE if it fails.
*/

ERROR_CODE tMutexUser::Abort(int iTicksToWaitForAbortSemaphore)
{
  ERROR_CODE ecCleanup, ec = OK;

  // Take the abort semaphore.  If the semaphore id is NULL (as in
  // the case where the object is inactive or where Execute is just
  // starting and the semaphore hasn't been created yet), we exit
  // with an error.  If not NULL and the semTake fails, it means
  // someone else got here first and took the semaphore and did
  // the abort and then deleted the semaphore (in CleanUp).
  //
  // In the latter case, the task is successfully deleted, so all
  // is well.  For this reason, we do *not* return an error if
  // the semTake fails, but return OK instead.

  /*****
    [ this invalid -> UNDEFINED
    | *this.siAbortSemaphore == NULL ->
        return S_mutexUser_ABORT_ATTEMPTED_ON_INACTIVE_MUTEX_USER_OBJECT
    | true ->
        Do nothing
    ]
  *****/
  if (this->siAbortSemaphore == NULL)        // see state box invariant 1
    return Traceback(S_mutexUser_ABORT_ATTEMPTED_ON_INACTIVE_MUTEX_USER_OBJECT,
                     "tMutexUser::Abort");

  /*****
    [ this invalid -> UNDEFINED
    | called from interrupt level -> UNDEFINED
    | *this.siAbortSemaphore invalid ->
        OS.Task.errno := S_objLib_OBJ_ID_ERROR
        : return OK
    | iTicksToWaitForAbortSemaphore != WAIT_FOREVER &&
            (iTicksToWaitForAbortSemaphore == NO_WAIT &&
             ! SemAvailable(*this.siAbortSemaphore) ||
             SemAvailable(*this.siAbortSemaphore) not true within
                                iTicksToWaitForAbortSemaphore ticks) ->
        OS.Task.errno := S_objLib_OBJ_TIMEOUT or S_objLib_OBJ_UNAVAILABLE
        : return S_objLib_OBJ_TIMEOUT or S_objLib_OBJ_UNAVAILABLE
    | true ->
        Calling task is blocked until SemAvailable(*this.siAbortSemaphore)
                                                                   is true
           (while blocked, the task is inserted into
            *this.*siAbortSemaphore.PendedTaskList as specified by
            *this.*siAbortSemaphore.QueuingMethod)
        : [ *this.*siAbortSemaphore.Type == SEM_BINARY ->
              *this.*siAbortSemaphore.State := SEM_EMPTY
          | true ->
              unfinished at this time
          ]
    ]
  *****/
  if (semTake(this->siAbortSemaphore, iTicksToWaitForAbortSemaphore)== ERROR) {
    if (errnoGet() == S_objLib_OBJ_TIMEOUT)
      return Traceback(S_objLib_OBJ_TIMEOUT,"tMutexUser::Abort");
    else if (errnoGet() == S_objLib_OBJ_UNAVAILABLE)
      return Traceback(S_objLib_OBJ_UNAVAILABLE,"tMutexUser::Abort");
    else
      return OK;
  }

  /*****
    [ this invalid -> UNDEFINED
    | *this.iTaskId == NULL or ERROR ->
        Do nothing
    | *this.iTaskId not a valid task id ->
        ec := S_objLib_OBJ_ID_ERROR
    | task cannot be deleted ->
        ec := VxWorks errno (taskDelete)
    | true ->
        OS.SpawnedTasks := OS.SpawnedTasks with the task with id
                                               *this.iTaskId removed
        : results from BB spec of *this.AbortMethod()
        : ec := return code from *this.AbortMethod()
    ]
  *****/
  // Delete the task
  // If we can't delete the task, something has probably gone wrong,
  // and we don't want to call the abort method.
  if (this->iTaskId != NULL && this->iTaskId != ERROR) {
    if (taskDelete(this->iTaskId) == OK)
     // Clean up the task
      ec = this->AbortMethod();
    else
      ec = errnoGet();
  }

  // What now?  What does it mean if the iTaskId was invalid?
  
  /*********
    [ this invalid -> UNDEFINED
    | *this.pMutex invalid -> UNDEFINED
    | *this.*pMutex.siSemaphore invalid ||
                       *this.*pMutex.WatchDog.bInit != TRUE -> UNDEFINED
    | true->
       [ *this.*pMutex.pOwner == this && != NULL ->   // Give back the mutex
           *this.*pMutex.pOwner := NULL
           : *this.*pMutex.WatchDog.bActive := FALSE
           : [ *this.*pMutex.*siSemaphore.PendedTaskList not empty ->
                 First task in *this.*pMutex.*siSemaphore.PendedTaskList
                                                                   unblocks
                 : First task removed from
                             *this.*pMutex.*siSemaphore.PendedTaskList
             | true ->
                 *this.*pMutex.*siSemaphore.State := SEM_FULL
             ]
      ]
      : [ *this.siAbortSemaphore not invalid ->
            All tasks in *this.*siAbortSemaphore.PendedTaskList unblock
                // Note - the above result will go away in the    *
                // all-at-once abstraction, because the fact that *
                // we were able to take the semaphore means that  *
                // there were no tasks on the PendedTaskList      *
            : OS.SemaphoreSet := OS.SemaphoreSet
                                      with *this.*siAbortSemaphore removed
            : *this.siAbortSemaphore := INVALID
        ]
      : [ *this.sTaskName != NULL ->
            return *this.*sTaskName storage to system heap
        ]
      : return *this storage to system heap
      : this := invalid
      : ecCleanup := OK
    ]
  *********/
  // Return the semaphore and turn off the watchdog
  ecCleanup = this->CleanUp();
  
  ec = (ec == OK ? ecCleanup : ec);

  if (ec != OK)
    return Traceback(ec, "tMutexUser::Abort");

  return OK;
}

  
/******************************************************************
* A
* tMutexUser::CleanUp - cleans up resources used by the mutex user
*
* This routine gives back the mutex and turns off the watchdog.
*
*
* IN:
*   this
*   *this.*pMutex
*   *this.*pMutex.siSempahore
*   *this.*pMutex.WatchDog.bInit
* 
* INOUT:
*   *this.*pMutex.*siSemaphore
*
* OUT:
*   *this.*pMutex.WatchDog.bActive
*
* SPEC:
    [ this invalid -> UNDEFINED
    | *this.pMutex invalid -> UNDEFINED
    | *this.*pMutex.siSemaphore invalid ||
                       *this.*pMutex.WatchDog.bInit != TRUE -> UNDEFINED
    | true->
       [ *this.*pMutex.pOwner == this && != NULL ->  // Give back the mutex
           *this.*pMutex.pOwner := NULL
           : *this.*pMutex.WatchDog.bActive := FALSE
           : [ *this.*pMutex.*siSemaphore.PendedTaskList not empty ->
                 First task in *this.*pMutex.*siSemaphore.PendedTaskList
                                                                  unblocks
                 : First task removed from
                                 *this.*pMutex.*siSemaphore.PendedTaskList
             | true ->
                 *this.*pMutex.*siSemaphore.State := SEM_FULL
             ]
      ]
      : [ *this.siAbortSemaphore not invalid ->
            All tasks in *this.*siAbortSemaphore.PendedTaskList unblock
            : OS.SemaphoreSet := OS.SemaphoreSet
                                      with *this.*siAbortSemaphore removed
            : *this.siAbortSemaphore := INVALID
        ]
      : [ *this.sTaskName != NULL ->
            return *this.*sTaskName storage to system heap
        ]
      : return *this storage to system heap
      : this := invalid
      : return OK
    ]
*
* RETURNS:
* OK if successful, or an ERROR_CODE if it fails.
*/

ERROR_CODE tMutexUser::CleanUp(void)
{
  //TRACEBACK ec("tMutexUser::CleanUp");

  ERROR_CODE ec;
  

  /********
    [ this invalid -> UNDEFINED
    | &*this.*pMutex invalid -> UNDEFINED
    | *this.*pMutex.pOwner == NULL -> Do nothing
    | *this.*pMutex.pOwner != this -> Do nothing
    | true ->
        *** Spec from below for this->pMutex->Give ***
    ]
  ********/
  // Return the semaphore if there is one and we are the owner
  if (this->pMutex->IsOwner(this)) {
    /*****
      [ *this.pMutex invalid -> UNDEFINED
      | *this.*pMutex.siSemaphore invalid ->
                  ec := S_objLib_OBJ_ID_ERROR
      | *this.*pMutex.WatchDog.bInit == FALSE ->
                  ec := S_watchDog_NOT_INITIALIZED
      | true ->
          *this.*pMutex.pOwner := NULL
          : *this.*pMutex.WatchDog.bActive := FALSE
          : [ *this.*pMutex.*siSemaphore.PendedTaskList not empty ->
                First task in *this.*pMutex.*siSemaphore.PendedTaskList
                                                                 unblocks
                : First task removed from
                                *this.*pMutex.*siSemaphore.PendedTaskList
            | true ->
                *this.*pMutex.*siSemaphore.State := SEM_FULL
            ]
          : ec := OK
      ]
    *****/
    ec = this->pMutex->Give();

    /*****
      [ ec == OK -> Do nothing
      | this invalid -> UNDEFINED
      | *this.pErrorClient invalid -> UNDEFINED
      | ! *this.*pErrorClient.miQid invalid &&
        !  (called from ISR and (RELAY_Q_SEND_MAXWAIT != NO_WAIT)) &&
        ! *this.*pErrorClient.miQid queue deleted &&
        ! ((RELAY_Q_SEND_MAXWAIT == NO_WAIT) &&
                             queue *this.*pErrorClient.miQid is full) &&
        ! (queue *this.*pErrorClient.miQid is full &&
                              queue *this.*pErrorClient.miQid remains full
                              for more than RELAY_Q_SEND_MAXWAIT ticks) &&
        ! (sizeof(ERROR_INFO) greater than queue *this.*pErrorClient.miQid's
                                                        max msg length) ->
             message containing 
                 ( *this.*pErrorClient.*pRte.iMyRteNum,
                   *this.*pErrorClient.*pRte.iMyCpuNum,
                   *this.*pErrorClient.*pRte.tSchedTime, ec ) is added to
                   end of VxWorks queue *this.*pErrorClient.miQid 
      ]
    *****/
    if (ec != OK)
      this->pErrorClient->Put(ec);
      // return ec;
  }

  /*****
    [ this invalid -> UNDEFINED
    | true -> *this.iTaskId := NULL ]
  *****/
  // This is important for proper synchronization, but has no net effect
  // since later, this := INVALID
  this->iTaskId = NULL;
  
  /*****
    [ *this.siAbortSemaphore invalid ->
        OS.Task.errno := S_objLib_OBJ_ID_ERROR
    | true ->
        All tasks in *this.*siAbortSemaphore.PendedTaskList unblock
        : OS.SemaphoreSet := OS.SemaphoreSet
                                  with *this.*siAbortSemaphore removed
        : *this.siAbortSemaphore := invalid      // Again, no net effect
    ]    
  *****/
  semDelete(this->siAbortSemaphore);

  /*****
    [ this invalid -> UNDEFINED
    | true ->
      [ *this.sTaskName != NULL ->
         return *this.*sTaskName storage to system heap
      ]
      : return *this storage to system heap
      : this := invalid
    ]
  *****/
  delete this;

  /***** [ return OK ] *****/
  return OK;
}


/******************************************************************
* A
* tMutexUser::~tMutexUser
*
*
* IN:
*   this
* INOUT:
*   system heap
*   *this.sTaskName
* OUT:
*   
* SPEC:
*
   [ this invalid -> UNDEFINED
   | true ->
     [ *this.sTaskName != NULL ->
        : return *this.*sTaskName storage to system heap
        : *this.*sTaskName := invalid
     ]
     : *this.sTaskName := NULL
   ]
*
***********
*   
   If called via delete:

   [ this invalid -> UNDEFINED
   | true ->
     [ *this.sTaskName != NULL ->
        return *this.*sTaskName storage to system heap
     ]
     : return *this storage to system heap
     : this := invalid
   ]
*
* RETURNS:
*   Nothing.
*/

tMutexUser::~tMutexUser(void)
{
  ERROR_CODE ec;
  
  if (this->sTaskName != NULL)
    free(this->sTaskName);
  this->sTaskName = NULL;

  #if 0
    // Not in this increment - this is too hard!  Circular call loop!
    if (iTaskId != NULL) {
      ec = Abort();
      if (ec != OK) {
        // Do something!
      }
    }
  #endif
}


