/* $RCSfile: Scheduler.cc,v $ - Scheduler routines */
/* CM Info:   

$Header: /home/chara/cvsroot/ople/sys/comm/scheduler/Scheduler.cc,v 1.4 1999/05/04 20:57:11 theo Exp $

*/
/* Art Sims - Jet Propulsion Laboratory */
 
/*
modification history
--------------------
01b,29Oct96,eamck    all sched routines collected into single file
01a,23Oct96,asims    routines written
01a,06Mar97,asims    Reviewed and verified: GetTaskPtr, GetTaskPtrFromIndex,
					    RemoveTask, and EnableTask.
*/
 
/*
DESCRIPTION -- Scheduler routines:

 * tScheduler::AddTask - Add a task into the periodic scheduling system.
 * tScheduler::Disable - Disable the Scheduler task
 * tScheduler::Enable - Enable the Scheduler task
 * tScheduler::DisableTask - Disable a task in Scheduler's TaskSet
 * tScheduler::EnableTask - Enable a task in Scheduler's TaskSet
 * tScheduler::DeleteTask - Delete a task in Scheduler's TaskSet
 * tScheduler::Initialize - Initialize the Scheduler

 * tScheduler::StartNextTask - Goes through task list, finds highest priority 
 *                      TS_READY or TS_RUN task and starts it. 
 * tScheduler::SchedIsr - Runs on every scheduler interrupt stimulus, examines 
 *                      and updates states of all tasks in task list. 
 * tScheduler::Suspender - start next user's task and suspend current one

 * tTask::Initialize(void) - Initialize a task control block with dummy values
 * tTask::Initialize - Initialize a task control block with user's values
 * tTask::taskSkel - task skeleton for user's tasks

*/
 

#include <isit.h>
#include <usrLib.h>
#include <stdioLib.h>
#include <errnoLib.h>
#include <vxLib.h>         /* for vxFpscrGet/Set floating point routines */
#include <logLib.h>
#include "Scheduler.h"
#include "ppcFpExceptions.h" 
#include "errorStream.h" 
 
//#define A_TASK_IS_STARVING_AND_WE_CANT_FIGURE_OUT_WHICH_ONE


/************************************************************************
 * 
 * tScheduler::GetTaskPtr - Return pointer corresponding to tTask object
 *                            whose .Id member == iTaskId
 *
 * SPEC:
     [[D tScheduler::GetTaskPtr(in: this,
				    iTaskId,
				    *this.TaskSet,
				    *this.iNumTasks;
				out: return; )]

     [ this invalid => UNDEFINED
     | *this.TaskSet is not pointer to array of at least 
	  (*this.iNumTasks) tTasks => UNDEFINED
     | there is no tTask in *this.TaskSet whose .Id == iTaskId => 
	  return NULL
     | true => return pointer to tTask whose .Id == iTaskId 
     ]]
 *
 * RETURNS: (Type tTask *)
 *    pointer to a tTask in Scheduler's task set 
 *    NULL
 *
 *************************************************************************/
tTask *tScheduler::GetTaskPtr( int iTaskId )
{
   int i;

   for (i=0; i<this->iNumTasks; i++)
      if (this->TaskSet[i].Id == iTaskId)
         return ( &(this->TaskSet[i]) );

   return NULL;

}  /* tScheduler::GetTaskPtr */



/************************************************************************
 * 
 * tScheduler::GetTaskPtrFromIndex - Return pointer found in
 *                                   hTaskSetPtr[iIndex].
 *
 * IN:
    iIndex
    this
    *this.hTaskSetPtr
 * INOUT:
 * OUT:
 *
 * SPEC:
     [ this invalid -> UNDEFINED
     | *this.hTaskSetPtr is not an array of pointers, each of which points 
	    to a tTasks in TaskSet -> UNDEFINED
     | iIndex < 0 || iIndex > *this.iNumTasks -> return UNDEFINED
     | true -> return pointer found in hTaskSetPtr[iIndex]
     ]
 *
 * RETURNS: (Type tTask *)
 *    pointer to a tTask in Scheduler's task set 
 *
 *************************************************************************/
tTask *tScheduler::GetTaskPtrFromIndex( int iIndex )
{
  return this->hTaskSetPtr[iIndex];
}  /* tScheduler::GetTaskPtrFromIndex */



/************************************************************************
 * 
 * tScheduler::RemoveTask - Remove the tTask object whose 
 *                           .Id member == iTid from the TaskSet.
 *
 * IN:
    iTid
    this
    *this.iNumTasks
    *this.hTaskSetPtr
 * INOUT:
    *this.TaskSet
 * OUT:
 *
 * SPEC:
     [ this invalid -> UNDEFINED
     | *this.TaskSet is not a pointer to an array of at least
	  (*this.iNumTasks) tTasks -> UNDEFINED
     | *this.iNumTasks <= 0 -> UNDEFINED
     | there is no tTask in *this.TaskSet whose .Id == iTid ->
          return S_sched_INVALID_TASK_ID
     | true -> 
	  : *this.TaskSet := *this.TaskSet - the task iTid
	  : where i == *this.TaskSet[*this.iNumTask-1].iTaskPosition ->
	       *this.hTaskSetPtr[i] := NULL
	  : *this.iNumTasks := *this.iNumTasks - 1
     ]
 *
 * RETURNS: 
 *   S_sched_INVALID_TASK_ID
 *   return OK
 *
 *************************************************************************/
ERROR_CODE tScheduler::RemoveTask(int iTid)
{
  int i;
  int iPosition;
  ERROR_CODE ec;

  TraceMark(S_sched_ENTERING_REMOVETASK);


  /*
   * Turn off scheduling
   */

  /*********
     [ this invalid -> UNDEFINED
     | *this.bInit != TRUE -> return S_sched_SCHEDULER_NOT_INITIALIZED
     | *this.TaskSet invalid -> UNDEFINED
     | *this.TaskSet is not a pointer to an array of tTasks of 
         dimension (*this.pEndTaskSet - *this.TaskSet)/sizeof(tTask) 
          -> UNDEFINED
     | there is no task in the TaskSet array whose .Mode is 
       TM_ENABLED and whose .ExecState is TS_READY or TS_RUN ->
         *this.Level := SL_DISABLED
     | there exists a task in the TaskSet array whose .Mode is 
       TM_ENABLED and whose .ExecState is TS_READY or TS_RUN that 
       cannot be resumed ->
         *this.Level := SL_DISABLED
         : some higher priority tasks may be resumed
         : return VxWorks errno (taskResume)
     | true ->
         *this.Level := SL_DISABLED
         : for each task whose .Mode is TM_ENABLED and whose 
            .ExecState is TS_READY or TS_RUN, that task's .ExecState 
              := TS_RUN
         : for each task whose .Mode is TM_ENABLED and whose 
            .ExecState is TS_READY or TS_RUN, that task is resumed
         : ec := OK
     ]
   ********/
   if ( (ec = this->Disable()) != OK)
        return Traceback ( ec,
                           "tScheduler::RemoveTask",
                            iTid);


   /********
      [ prevent other tasks from preempting the DisableTask ]
   ********/
   taskLock();

   // find where in TaskSet the task iTid resides.
   /*******
      [ this invalid -> UNDEFINED
      | *this.TaskSet is not a pointer to array of at least
           (*this.iNumTasks) tTasks -> UNDEFINED
      | if there does not exist an i in (0..*this.iNumTasks) such that
          *this.TaskSet[i].Id == iTid -> iPosition :=  -1.
      | true -> iPosition = i
      ]
   *******/
   iPosition = -1;
   for (i=0; i< this->iNumTasks; i++)
   {
     if (this->TaskSet[i].Id == iTid)
     {
          iPosition = i;
          break;
     }
   }

   // if the task to be deleted was not found, report error
   if (iPosition == -1)
       return Traceback ( S_sched_INVALID_TASK_ID,
                          "tScheduler::RemoveTask",
                           iTid);

   // set the pointer inside of hTaskSetPtr which points to the task
   // being deleted to NULL.
   this->hTaskSetPtr[this->TaskSet[this->iNumTasks-1].iTaskPosition] = NULL;

   /*****
      [ this invalid -> UNDEFINED
      | *this.iNumTasks <= 0 || *this.iNumTasks > *this.iMaxTasks 
           -> UNDEFINED
      | iPosition < 0 || iPosition >= *this.NumTasks -> UNDEFINED
      | true ->
	   for every i in iPosition..*this.iNumTasks-1 ->
           TaskSet[i] := TaskSet[i+1]
           : hTaskSetPtr[TaskSet[i].iTaskPosition] := *This.TaskSet[i]
      ]
   *****/
   for (i=iPosition; i<this->iNumTasks-1; i++)
   {
     this->TaskSet[i] = this->TaskSet[i+1];

     // at this point, TaskSet[i].iTaskPosition and TaskSet[i+1].iTaskPosition
     //  are the same.
     this->hTaskSetPtr[this->TaskSet[i].iTaskPosition] =
	 &(this->TaskSet[i]);
   }

   // Decrement the number of task in the tTaskSet.
   this->iNumTasks--;
      
    // Enable the Scheduler and preemptions again
    /********
       [ allow other tasks to preempt this one ]
    ********/
    taskUnlock();

    /*************
       [ this invalid -> UNDEFINED
       | *this.bInit != TRUE -> return S_sched_SCHEDULER_NOT_INITIALIZED
       | true ->
           *this.Level := SL_ENABLED
           : ec := OK
       ]
    ********/
    ec = this->Enable();

    if (ec != OK)
        return Traceback ( ec,
                         "tScheduler::RemoveTask",
                          iTid);


   /* [return OK] */
   return (OK);

}  /* RemoveTask() */
    

/************************************************************************
 
Long form:
      [[D tScheduler::AddTask_long(int                *piTaskId,
                                 char                *sName,
                                 FUNCPTR              pExec,
                                 FUNCPTR              pInit, 
                                 FUNCPTR              pDelete,
                                 void                *pArgument,
                                 int                  iPriority,
                                 TIME                *ptPeriod,
                                 TIME                *ptOffset, 
                                 TASK_ERROR_BEHAVIOR  tebErrorBehavior,
                                 BOOL                 bNopanic)]

      [ this invalid => UNDEFINED
      | *this.bInit !=TRUE => return := S_sched_SCHEDULER_NOT_INITIALIZED
      | *this.TaskSet is not a pointer to an array of tTasks of dimension >=
              (*this.pEndTaskSet - *this.TaskSet) / sizeof(tTask) => UNDEFINED
      | *this.iNumTasks >= *this.iMaxTasks => return := S_sched_MAX_TASKS_EXCEEDED
      | pExec == NULL => return := S_sched_NULL_TASK_EXEC_FUNCTION_POINTER 
      | (iPriority < MIN_SCHED_TASK_PRIO) || 
           (iPriority > MAX_SCHED_TASK_PRIO) => return := S_sched_INVALID_PRIORITY
      | tPeriod <= 0 => return := S_sched_ILLEGAL_TASK_PERIOD
      | *this.iRate != scheduler interrupt rate in Hz => UNDEFINED
      | *ptOffset < 0 or *ptOffset >= tPeriod => 
	    return := S_sched_ILLEGAL_TASK_OFFSET
      | *ptPeriod is not an integer multiple of scheduler interrupt rate =>
	  print warning message
	  : ptPeriod := next lower value that is integer multiple of rate
      | *ptOffset is not an integer multiple of scheduler interrupt rate =>
	  print warning message
	  : *tOffset := 0
      | *this.hTaskSetPtr not a pointer to an array of 
             (tTask *) of dimension >= *this.iMaxTasks => UNDEFINED
      | there does not exist an i in (0..*this.iMaxTasks-1) such that
             *this.hTaskSetPtr[i] == NULL => return := S_sched_OUT_OF_TCB_POINTERS
      | there exists a task in the TaskSet array whose .Mode is TM_ENABLED and
              whose .ExecState is TS_READY or TS_RUN that cannot be resumed =>
          some higher priority tasks may be resumed
          : return := VxWorks errno (taskResume)
      | *this.iNumTasks <= 0 => UNDEFINED
      | *this.iNumTasks > *this.iMaxTasks => UNDEFINED
      | pExec, pDelete, or pInit not NULL and not a valid C/C++ fn => UNDEFINED
      | VxWorks priority of calling task 
                           numerically <= MAX_SCHED_TASK_PRIO => UNDEFINED
      | true => 
          : for each task whose .Mode is TM_ENABLED and whose .ExecState is
              TS_READY or TS_RUN, that task's .ExecState := TS_RUN
          : for each task whose .Mode is TM_ENABLED and whose .ExecState is
              TS_READY or TS_RUN, that task is resumed
          : new tTask object is inserted into *this.TaskSet in order of incr.
              .iPriority, with the following values:
                   .bNoPanic := bNopanic
                   .BehaviorOnError := tebErrorBehavior
                   .PendingOp := TPO_NONE
                   .ExecState := TS_DONE
                   .tScheduledTime := 0  
                   .pExecFunc := pExec
                   .pInitFunc := pInit
                   .pDeleteFunc := pDelete
                   .iPriority := iPriority
                   .tPeriod := *ptPeriod
                   .tOffset := *ptOffset
                   .ErrorCode := OK
                   .iTaskPosition := index in *this.hTaskSetPtr of first NULL
                   .ptScheduler := this
          : [ Task cannot be spawned => 
               *piTaskId := ERROR
               : new tTask's .Id := ERROR
               : new tTask's .Mode := TM_ERROR
               : return := VxWorks errno (taskSpawn)
            | true => 
               *piTaskId := new task Id from VxWorks
               : new tTask's .Id := new task Id from VxWorks
               : new tTask's .Mode := TM_INIT
               : OS.SpawnedTasks := OS.SpawnedTasks U { new task }
            ]
          : *this.hTaskSetPtr array modified so that pointers point to correct
              elements of *this.TaskSet
          : *this.iNumTasks := *this.iNumTasks + 1
          : *this.pEndTaskSet := &(*this.TaskSet[*this.iNumTasks])
          : return := OK
      ]]
 *
 * RETURNS: (Type ERROR_CODE)
 *    S_sched_SCHEDULER_NOT_INITIALIZED
 *    S_sched_MAX_TASKS_EXCEEDED
 *    S_sched_INVALID_PRIORITY
 *    S_sched_ILLEGAL_TASK_PERIOD
 *    S_sched_ILLEGAL_TASK_OFFSET
 *    S_sched_OUT_OF_TCB_POINTERS
 *    VxWorks errno (taskResume)
 *    VxWorks errno (taskSpawn)
 *    OK
 *
 *************************************************************************/

 ERROR_CODE tScheduler::AddTask (int                *piTaskId,
                                 char                *sName,
                                 FUNCPTR              pExec,
                                 FUNCPTR              pInit, 
                                 FUNCPTR              pDelete,
                                 void                *pArgument,
                                 int                  iPriority,
                                 TIME                *ptPeriod,
                                 TIME                *ptOffset, 
                                 TASK_ERROR_BEHAVIOR  tebErrorBehavior,
                                 BOOL                 bNopanic)
{
   ERROR_CODE  ec, ecSpawn = OK;
   int         i;
   int         iNewTaskId;
   int         iTemp;
   int         iCurTask;
   int         iNewTask;
   int         iOldSchedValue;
   int         iPointerNum;
   unsigned    iTempNumber=0;
   TIME        tSchedTicks;
   TIME        tTemp;
   char        sMsgOut[256];

   TraceMark(S_sched_ENTERING_ADD_TASK);

   if (sName == NULL) sName = "Unnamed";
   sprintf(sMsgOut, "AddTask: Task %s, period %d, priority %d, arg 0x%x",
          sName, (int)*ptPeriod, iPriority, (int) pArgument);
   TraceMsg(tmt_INFO,OK,sMsgOut);  


   /*******
      [ this invalid -> UNDEFINED
      | *this.bInit !=TRUE -> return S_sched_SCHEDULER_NOT_INITIALIZED
      | true -> nothing happens
      ]
   *******/
   if (this->bInit != TRUE)
      return Traceback (S_sched_SCHEDULER_NOT_INITIALIZED,
                         "tScheduler::AddTask",
                          piTaskId,
                          sName,
                          pExec,
                          pInit,
                          pDelete,
                          pArgument,
                          iPriority,
                          ptPeriod,
                          ptOffset,
                          tebErrorBehavior,
                          bNopanic);
   

   /*******
      [ this invalid -> UNDEFINED
      | *this.iNumTasks >= *this.iMaxTasks -> return S_sched_MAX_TASKS_EXCEEDED
      | true -> nothing happens
      ]
   *******/
   if (this->iNumTasks >= this->iMaxTasks)
      return Traceback ( S_sched_MAX_TASKS_EXCEEDED,
                         "tScheduler::AddTask",
                          piTaskId,
                          sName,
                          pExec,
                          pInit,
                          pDelete,
                          pArgument,
                          iPriority,
                          ptPeriod,
                          ptOffset,
                          tebErrorBehavior,
                          bNopanic);
 

   /*******
      [ pExec == NULL -> return S_sched_NULL_TASK_EXEC_FUNCTION_POINTER 
      | true -> nothing happens
      ]
   *******/
    if ( pExec == NULL)
        return Traceback ( S_sched_NULL_TASK_EXEC_FUNCTION_POINTER,
                           "tScheduler::AddTask",
                            piTaskId,
                            sName,
                            pExec,
                            pInit,
                            pDelete,
                            pArgument,
                            iPriority,
                            ptPeriod,
                            ptOffset,
                            tebErrorBehavior,
                            bNopanic);

   /*******
      [ (iPriority < MIN_SCHED_TASK_PRIO) || 
           (iPriority > MAX_SCHED_TASK_PRIO) -> return S_sched_INVALID_PRIORITY
      | true -> nothing happens
      ]
   *******/
   if ((iPriority < MIN_SCHED_TASK_PRIO) || (iPriority > MAX_SCHED_TASK_PRIO))
      return Traceback ( S_sched_INVALID_PRIORITY,    
                         "tScheduler::AddTask",
                          piTaskId,
                          sName,
                          pExec,
                          pInit,
                          pDelete,
                          pArgument,
                          iPriority,
                          ptPeriod,
                          ptOffset,
                          tebErrorBehavior,
                          bNopanic);


   /*******
      [ *ptPeriod <= 0 -> return S_sched_ILLEGAL_TASK_PERIOD
      | true -> nothing happens
      ]
   *******/
    if ( *ptPeriod <= 0 ) 
        return Traceback ( S_sched_ILLEGAL_TASK_PERIOD,            
                         "tScheduler::AddTask",
                          piTaskId,
                          sName,
                          pExec,
                          pInit,
                          pDelete,
                          pArgument,
                          iPriority,
                          ptPeriod,
                          ptOffset,
                          tebErrorBehavior,
                          bNopanic);


        /* change ptPeriod so it's an integer number of sched. interrupts */

   #ifndef INCR1
   #error Pass ptPeriod by reference so we can change it for user
   #endif
   /*******
      [ this invalid -> UNDEFINED
      | *this.iRate != scheduler interrupt rate in Hz -> UNDEFINED
      | true -> tSchedTicks := time units between scheduler interrupts
      ]
   *******/
    tSchedTicks = TIME_TICKS_PER_SECOND / this->iRate ;


       /* NOTE: if ptPeriod < tSchedTicks then *ptPeriod will become 0 */
   /*******
      [ *ptPeriod is not an integer multiple of tSchedTicks ->
          iTemp := *ptPeriod/tSchedTicks rounded down to integer > 0
          : print warning message
          : *ptPeriod := next lower value that is integer multiple of tSchedTicks
      | true -> nothing happens
      ]
   *******/
    if ( *ptPeriod % tSchedTicks != 0)
    {
       iTemp = ((int)*ptPeriod) / (int)tSchedTicks; /* force integer division */
       if (iTemp <= 0)
         iTemp = 1;

       sprintf(sMsgOut,
	      "WARNING: (tScheduler::AddTask) Invalid ptPeriod %d, now %d.  tSchedTicks=%d",
                                   (int) *ptPeriod, (int) (iTemp*tSchedTicks), (int) tSchedTicks );
       TraceMsg(tmt_WARN,OK,sMsgOut);

       *ptPeriod = iTemp * tSchedTicks;
    }

   TraceMark(S_sched_TASK_PERIOD_SET);

   /*******
      [ (*ptOffset < 0) || (*ptOffset >= *ptPeriod) -> 
           return S_sched_ILLEGAL_TASK_OFFSET
      | true -> nothing happens
      ]
   *******/
    if ((*ptOffset < iTempNumber) || ( *ptOffset >= *ptPeriod ))

        return Traceback ( S_sched_ILLEGAL_TASK_OFFSET,
                         "tScheduler::AddTask",
                          piTaskId,
                          sName,
                          pExec,
                          pInit,
                          pDelete,
                          pArgument,
                          iPriority,
                          ptPeriod,
                          ptOffset,
                          tebErrorBehavior,
                          bNopanic);
 
   #ifndef INCR1
   #error incr2: make *ptOffset an integer multiple of tSchedTicks
   #endif

   // incr2
   // make *ptOffset an interger multiple of tSchedTicks
   /*******
      [ *ptOffset is not an integer multiple of tSchedTicks ->
          *ptOffset := max (*ptOffset - (*ptOffset % tSchedTicks), 0)
          : print warning message
      | true -> nothing happens
      ]
   *******/
    if ( *ptOffset % tSchedTicks != 0)
    {
       tTemp = *ptOffset;
       *ptOffset -= (*ptOffset % tSchedTicks);
       if (*ptOffset <=0)
	  *ptOffset = 0;

       sprintf(sMsgOut,
           "WARNING: (tScheduler::AddTask) Invalid ptOffset %d, set to %d",
                  (int)tTemp, (int)*ptOffset);
       TraceMsg(tmt_WARN,OK,sMsgOut);
    }

   TraceMark(S_sched_TASK_OFFSET_SET);

   //**** Start processing


    // Find an empty task pointer for the task to use 
   /****
      [ this invalid -> UNDEFINED
      | *this.hTaskSetPtr not pointer to an array of 
             (tTask *) of dimension >= *this.iMaxTasks -> UNDEFINED
      | if there doesn't exist an i in (0..*this.iMaxTasks-1) such that
             *this.hTaskSetPtr[i] == NULL -> return S_sched_OUT_OF_TCB_POINTERS
      | true -> 
           iPointerNum := index of first NULL element in *this.hTaskSetPtr
      ]
   ****/
   iPointerNum = -1;
   for (i=0; i < this->iMaxTasks; i++) 
   {
      if (this->hTaskSetPtr[i] == NULL) 
      {
         iPointerNum = i;
         break;
      }
   }

    // if an empty task was not found, report error 
    if (iPointerNum == -1) 
        return Traceback ( S_sched_OUT_OF_TCB_POINTERS,
                         "tScheduler::AddTask",
                          piTaskId,
                          sName,
                          pExec,
                          pInit,
                          pDelete,
                          pArgument,
                          iPriority,
                          ptPeriod,
                          ptOffset,
                          tebErrorBehavior,
                          bNopanic);

   TraceMark(S_sched_FOUND_EMPTY_TASK_PTR_TO_USE);

    /*
     * Turn off scheduling 
     */

    /*********
      [ this invalid -> UNDEFINED
      | *this.bInit != TRUE -> return S_sched_SCHEDULER_NOT_INITIALIZED
      | *this.TaskSet invalid -> UNDEFINED
      | *this.TaskSet is not a pointer to an array of tTasks of dimension
              (*this.pEndTaskSet - *this.TaskSet)/sizeof(tTask) -> UNDEFINED
      | there is no task in the TaskSet array whose .Mode is TM_ENABLED and
              whose .ExecState is TS_READY or TS_RUN ->
          *this.Level := SL_DISABLED
      | there exists a task in the TaskSet array whose .Mode is TM_ENABLED and
              whose .ExecState is TS_READY or TS_RUN that cannot be resumed ->
          *this.Level := SL_DISABLED
          : some higher priority tasks may be resumed
          : return VxWorks errno (taskResume)
      | true ->
          *this.Level := SL_DISABLED
          : for each task whose .Mode is TM_ENABLED and whose .ExecState is
              TS_READY or TS_RUN, that task's .ExecState := TS_RUN
          : for each task whose .Mode is TM_ENABLED and whose .ExecState is
              TS_READY or TS_RUN, that task is resumed
          : ec := OK
      ]
    ********/
    if ( (ec = this->Disable()) != OK)
       return Traceback ( ec,
                         "tScheduler::AddTask",
                          piTaskId,
                          sName,
                          pExec,
                          pInit,
                          pDelete,
                          pArgument,
                          iPriority,
                          ptPeriod,
                          ptOffset,
                          tebErrorBehavior,
                          bNopanic);

   TraceMark(S_sched_DISABLED_SCHEDULER_IN_ADDTASK);


    // Prevent anyone else from preempting the DisableTask
    /********
       [ prevent other tasks from preempting this one ]
    ********/
    taskLock();

    TraceMark(S_sched_JUST_ISSUED_TASK_LOCK);

    // Find where to insert new task
    /*****
       [ this invalid -> UNDEFINED
       | there doesn't exist a iNewTask in (0..*this.iNumTasks-1) such 
                    that *this.TaskSet[iNewTask].iPriority > iPriority -> 
          iNewTask := *this.iNumTasks
       | true ->
          iNewTask := the index of the first element of *this.TaskSet whose 
                              .iPriority is greater than iPriority.
       ]
    *****/
    for (iNewTask= 0; iNewTask< this->iNumTasks; iNewTask++)
    {
        if (this->TaskSet[iNewTask].iPriority > iPriority)
            break;
    }

   TraceMark(S_sched_FOUND_WHERE_TO_INSERT_TCB);

    // Increment number of Tasks
    /*****
      [ this invalid -> UNDEFINED
      | true -> *this.iNumTasks := *this.iNumTasks + 1
      ]
    *****/
    this->iNumTasks++;


   /*****
      [ this invalid -> UNDEFINED
      | *this.iNumTasks <= 0 -> UNDEFINED
      | *this.iNumTasks > *this.iMaxTasks -> UNDEFINED
      | true ->
           each TCB in TaskSet[iNewTask..*this.iNumTasks-2] moved 
                                                to next higher index
           : *this.hTaskSetPtr rearranged correspondingly 
      ]
   *****/
    // Move tasks up to make space for the new task beginning
    // with *this.iNumTasks-1.
   iCurTask = this->iNumTasks - 1;
   while (iCurTask-- > iNewTask)        /* drops out if iNewTask is at end */
   {
      this->TaskSet[iCurTask+1] = this->TaskSet[iCurTask];
      this->hTaskSetPtr[this->TaskSet[iCurTask].iTaskPosition] =
                &(this->TaskSet[iCurTask+1]);
   }

   TraceMark(S_sched_MADE_SPACE_TO_INSERT_TCB);


    /******
       [ this invalid -> UNDEFINED
       | iNewTask > *this.iMaxTasks -> UNDEFINED
       | true ->
            insert new task information into the tcb.
            : *this.pEndTaskSet := &(*this.TaskSet[*this.iNumTasks])
            : *this.hTaskSetPtr[iPointerNum] := &(*this.TaskSet[*this.iNewTask])
       ]
    *****/
    // Insert the new task's information into the tcb 
    this->TaskSet[iNewTask].Initialize( ERROR, 
                                        FALSE, 
                                        bNopanic, 
                                        tebErrorBehavior, 
                                        TPO_NONE, 
                                        TS_DONE, 
                                        TM_INIT, 
                                        (TIME)0, 
                                        pExec, 
                                        pInit, 
                                        pDelete, 
                                        iPriority, 
                                        *ptPeriod, 
                                        *ptOffset, 
                                        OK, 
                                        iPointerNum, 
                                        this );

    this->pEndTaskSet = this->TaskSet + this->iNumTasks;
    this->hTaskSetPtr[iPointerNum] = this->TaskSet + iNewTask;

    TraceMark(S_sched_JUST_INSERTED_NEW_TCB);

 
   /****
      [ this invalid -> UNDEFINED
      | *this.TaskSet not a ptr to an array of (tTask) 
                         of dimension >= iNewTask+1 -> UNDEFINED
      | Task cannot be spawned -> 
           iNewTaskId := ERROR
           : *piTaskId := ERROR
           : *this.TaskSet[iNewTask].Id := ERROR
           : *this.TaskSet[iNewTask].Mode := TM_ERROR
           : ecSpawn := VxWorks errno (taskSpawn)
      | true -> 
           iNewTaskId := new task Id from VxWorks
           : *piTaskId := iNewTaskId
           : *this.TaskSet[iNewTask].Id := new task Id from VxWorks
      ]
   ***/
    // Spawn a new task
    iNewTaskId = taskSpawn (sName, 
                          SCHED_INIT_PRIORITY,  
                          SCHED_TASK_OPTIONS,
                          SCHED_STACK_SIZE,
                          (FUNCPTR) taskSkel,
                          (int) pArgument, (int) this, iPointerNum,
                          0, 0, 0, 0, 0, 0, 0);

    *piTaskId = iNewTaskId;
    this->TaskSet[iNewTask].Id = iNewTaskId;
 
    if (iNewTaskId == ERROR)
    {
        this->TaskSet[iNewTask].Mode = TM_ERROR;
        ecSpawn = errnoGet();
    }


    TraceMark(S_sched_JUST_SPAWNED_NEW_TASKSKEL);


    // Enable the Scheduler and preemptions again
    /********
       [ allow other tasks to preempt this one ]
    ********/
    taskUnlock();

    /*************
       [ this invalid -> UNDEFINED
       | iNewTaskId == ERROR -> return VxWorks errno (taskSpawn)
       | *this.bInit != TRUE -> return S_sched_SCHEDULER_NOT_INITIALIZED
       | true ->
           *this.Level := SL_ENABLED
           : ec := OK
       ]
    ********/
    ec = this->Enable();
    if (iNewTaskId == ERROR)
        return Traceback ( ecSpawn,
                         "tScheduler::AddTask",
                          piTaskId,
                          sName,
                          pExec,
                          pInit,
                          pDelete,
                          pArgument,
                          iPriority,
                          ptPeriod,
                          ptOffset,
                          tebErrorBehavior,
                          bNopanic);
    else if (ec != OK)
        return Traceback ( ec,
                         "tScheduler::AddTask",
                          piTaskId,
                          sName,
                          pExec,
                          pInit,
                          pDelete,
                          pArgument,
                          iPriority,
                          ptPeriod,
                          ptOffset,
                          tebErrorBehavior,
                          bNopanic);

    TraceMark(S_sched_LEAVING_ADD_TASK);

    return (OK);

} /* tScheduler::AddTask */



/******************************************************************
 *
 * tScheduler::Disable - Disable scheduling
 * 
 * 
 * IN:
 *   this
 *   *this.bInit
 *   *this.TaskSet[].Mode
 *   *this.pEndTaskSet
 * INOUT:
 *   *this.TaskSet[].ExecState
 *   OS.RunningTasks
 * OUT:
 *   *this.Level
 *
 * SPEC:
      [ this invalid -> UNDEFINED
      | *this.bInit != TRUE -> return S_sched_SCHEDULER_NOT_INITIALIZED
      | *this.TaskSet invalid -> UNDEFINED
      | VxWorks priority of calling task numerically <= MAX_SCHED_TASK_PRIO ->
          UNDEFINED
      | *this.TaskSet is not a pointer to an array of tTasks of dimension
                    (*this.pEndTaskSet - *this.TaskSet)/sizeof(tTask) -> UNDEFINED
      | there is no task in the TaskSet array whose .Mode is TM_ENABLED and
              whose .ExecState is TS_READY or TS_RUN -> 
          *this.Level := SL_DISABLED
          : return OK
      | there exists a task in the TaskSet array whose .Mode is TM_ENABLED and
              whose .ExecState is TS_READY or TS_RUN that cannot be resumed ->
          *this.Level := SL_DISABLED
          : some higher priority tasks may be resumed
          : return VxWorks errno (taskResume)
      | true ->
          *this.Level := SL_DISABLED
	  : for each task whose .Mode == TM_ENABLED && .ExecStat is
	      TS_READY || TS_RUN ->  .ExecState := TS_RUN
          : all due tasks in *this.TaskSet are resumed
          : return OK
      ]
 *   
 * RETURNS: (Type ERROR_CODE)
 *    S_sched_SCHEDULER_NOT_INITIALIZED
 *    OK
 *
 ******************************************************************/

ERROR_CODE tScheduler::Disable( void )
{
   ERROR_CODE ec;
   tTask *pThisTask;
   tTask *pStop;
   
   TraceMark(S_sched_ABOUT_TO_DISABLE_SCHEDULER);

   /*******
      [this invalid -> UNDEFINED
      | *this.bInit != TRUE -> return S_sched_SCHEDULER_NOT_INITIALIZED
      ]
   *******/
   if (this->bInit != TRUE)
      return Traceback (S_sched_SCHEDULER_NOT_INITIALIZED,
                      "tScheduler::Disable");

   /*******
      [this invalid -> UNDEFINED
      | true ->
         *this.Level = SL_DISABLED
      ]
   *******/
   this->Level = SL_DISABLED;

   /*****
      [ this invalid -> UNDEFINED
      | true -> pStop := *this.pEndTaskSet
      ]
   *****/
   pStop = this->pEndTaskSet;   // points just past end of the TaskSet array 

   /*****
      [ this invalid -> UNDEFINED
      | *this.TaskSet invalid -> UNDEFINED
      | *this.TaskSet is not a pointer to an array of tTasks of dimension 
              (pStop - *this.TaskSet)/sizeof(tTask) -> UNDEFINED
      | there is no task in the TaskSet array whose .Mode is TM_ENABLED and 
              whose .ExecState is TS_READY or TS_RUN -> nothing happens
      | there exists a task in the TaskSet array whose .Mode is TM_ENABLED and 
              whose .ExecState is TS_READY or TS_RUN that cannot be resumed ->  
          return VxWorks errno (taskResume)
      | true -> 
          for each task whose .Mode is TM_ENABLED and whose .ExecState is 
              TS_READY or TS_RUN, that task's .ExecState := TS_RUN
          : for each task whose .Mode is TM_ENABLED and whose .ExecState is 
              TS_READY or TS_RUN, that task is resumed
          : ec := OK
      ]
   *******/
   for  ( pThisTask = this->TaskSet;   pThisTask < pStop;   pThisTask++ )
   {
      /*********
         [ pThisTask invalid -> UNDEFINED
         | pThisTask >= pStop -> do nothing  (i.e. exit loop)
         | pThisTask->ExecState != TS_READY -> nothing happens
         | cannot resume task *pThisTask.Id -> return VxWorks error (taskResume)
         | true -> 
             *pThisTask.ExecState := TS_RUN
             : *pThisTask.ExecState == READY -> task *pThisTask.Id is resumed
             : ec := OK
             : pThisTask := pThisTask + 1*sizeof(*pThisTask)
         ]
      ******/
      if ( pThisTask->ExecState == TS_READY )
      {
         pThisTask->ExecState = TS_RUN;

         ec = taskResume(pThisTask->Id);
         if (ec != OK)
	 {
           logMsg("taskResume failed in tScheduler::Disable, taskId=0x%08x\n",
                  pThisTask->Id,0,0,0,0,0);
            return Traceback( errnoGet(), "tScheduler::Disable" );
	 }
      }
   }

   /* [return OK] */
   return OK;

} /* tScheduler::Disable () */



/******************************************************************
 *
 * tScheduler::Enable - Enable scheduling
 * 
 * 
 * IN:
 *   this
 *   *this.bInit
 * INOUT:
 * OUT:
 *   *this.Level
 *
 * SPEC:
      [ this invalid -> UNDEFINED
      | *this.bInit != TRUE -> return S_sched_SCHEDULER_NOT_INITIALIZED
      | true ->
           *this.Level := SL_ENABLED
         : return OK
      ]
 *
 * RETURNS: (Type ERROR_CODE)
 *    OK
 *
 ******************************************************************/

ERROR_CODE tScheduler::Enable( void )
{
    
   TraceMark(S_sched_ABOUT_TO_ENABLE_SCHEDULER);

    /*******
       [this invalid -> UNDEFINED
       | *this.bInit != TRUE -> return S_sched_SCHEDULER_NOT_INITIALIZED
       ]
    *******/
    if (this->bInit != TRUE)
        return Traceback (S_sched_SCHEDULER_NOT_INITIALIZED,
                          "tScheduler::Enable");

    /*******
       [this invalid -> UNDEFINED
       | true ->
           *this.Level = SL_ENABLED
       ]
    *******/
    this->Level = SL_ENABLED;

    /* [return OK] */
        return OK;

} /* tScheduler::Enable */


 
/****
[[U tScheduler::DisableTask (iTaskId) ]]
****/
ERROR_CODE tScheduler::DisableTask(int iTaskId)
{
   tTask *pTcb;
   ERROR_CODE ec;
 
   TraceMark(S_sched_ABOUT_TO_DISABLE_A_TASK);

   /*******
      [this invalid -> UNDEFINED
      | *this.bInit != TRUE -> return S_sched_SCHEDULER_NOT_INITIALIZED
      ]
   *******/
   if (this->bInit != TRUE)
       return Traceback (S_sched_SCHEDULER_NOT_INITIALIZED,
                         "tScheduler::DisableTask", iTaskId);

   /*******
     [ this invalid -> UNDEFINED
     | *this.TaskSet is not pointer to array of (*this.iNumTasks) tTasks -> 
          UNDEFINED
     | there is no tTask in *this.TaskSet whose .Id == iTaskId -> 
          return S_sched_INVALID_TASK_ID
     | true -> pTcb := pointer to tTask whose .Id == iTaskId 
     ]
   *******/
   pTcb = this->GetTaskPtr(iTaskId);
   if (pTcb==NULL)
      return Traceback(S_sched_INVALID_TASK_ID,
                  "tScheduler::DisableTask",iTaskId);
 
#ifndef INCR1
#error "Should lock out suspender here so mode doesn't change"
#endif

    /*
     * Turn off scheduling 
     */

    /*********
      [ this invalid -> UNDEFINED
      | *this.bInit != TRUE -> return S_sched_SCHEDULER_NOT_INITIALIZED
      | *this.TaskSet invalid -> UNDEFINED
      | *this.TaskSet is not a pointer to an array of tTasks of dimension
              (*this.pEndTaskSet - *this.TaskSet)/sizeof(tTask) -> UNDEFINED
      | there is no task in the TaskSet array whose .Mode is TM_ENABLED and
              whose .ExecState is TS_READY or TS_RUN ->
          *this.Level := SL_DISABLED
      | there exists a task in the TaskSet array whose .Mode is TM_ENABLED a
nd
              whose .ExecState is TS_READY or TS_RUN that cannot be resumed
->
          *this.Level := SL_DISABLED
          : some higher priority tasks may be resumed
          : return VxWorks errno (taskResume)
      | true ->
          *this.Level := SL_DISABLED
          : for each task whose .Mode is TM_ENABLED and whose .ExecState is
              TS_READY or TS_RUN, that task's .ExecState := TS_RUN
          : for each task whose .Mode is TM_ENABLED and whose .ExecState is
              TS_READY or TS_RUN, that task is resumed
          : ec := OK
      ]
    ********/
    if ( (ec = this->Disable()) != OK)
       return Traceback ( ec,
                         "tScheduler::DisableTask",
                          iTaskId);


   // Prevent anyone else from preempting the DisableTask
   /********
      [ prevent other tasks from preempting this one ]
   ********/
   taskLock();

   /********
      [ pTcb invalid -> UNDEFINED
      | *pTcb.Mode == TM_INIT -> *pTcb.PendingOp := TPO_DISABLE_TASK
      | *pTcb.Mode == TM_ENABLED && *pTcb.ExecState == TS_DONE ->
	 *pTcb.Mode := TM_DISABLED
      | *pTcb.Mode == TM_ENABLED && *pTcb.ExecState != TS_DONE ->
	 *pTcb.PendingOp := TPO_DISABLE_TASK
      | true -> nothing happens
      ]
   *******/
   switch (pTcb->Mode)
   {
      case TM_INIT:
         pTcb->PendingOp = TPO_DISABLE_TASK;
         break;
         
      case TM_ENABLED:
         if (pTcb->ExecState==TS_DONE) 
            pTcb->Mode = TM_DISABLED;
         else
            pTcb->PendingOp = TPO_DISABLE_TASK;
         break;

      default:   /* any other mode, do nothing */
         break;

   }  /* switch(Mode) */
 
    /********
       [ allow other tasks to preempt this one ]
    ********/
    taskUnlock();

    /*************
       [ this invalid -> UNDEFINED
       | *this.bInit != TRUE -> return S_sched_SCHEDULER_NOT_INITIALIZED
       | true ->
           *this.Level := SL_ENABLED
           : ec := OK
       ]
    ********/
    ec = this->Enable();

    if (ec != OK)
        return Traceback ( ec,
                         "tScheduler::DisableTask",
                          iTaskId);

    /* [return OK] */
    return OK;
 
} /* tScheduler::DisableTask */
 


 
/****
[[U tScheduler::EnableTask (iTaskId) ]]
****/
ERROR_CODE tScheduler::EnableTask(int iTaskId)
{
   tTask *pTcb;
 
   TraceMark(S_sched_ABOUT_TO_ENABLE_A_TASK);

   /*******
      [this invalid -> UNDEFINED
      | *this.bInit != TRUE -> return S_sched_SCHEDULER_NOT_INITIALIZED
      ]
   *******/
   if (this->bInit != TRUE)
       return Traceback (S_sched_SCHEDULER_NOT_INITIALIZED,
                         "tScheduler::EnableTask", iTaskId);

   /*******
     [ this invalid -> UNDEFINED
     | *this.TaskSet is not pointer to array of at least (*this.iNumTasks) 
	  tTasks -> UNDEFINED
     | there is no tTask in *this.TaskSet whose .Id == iTaskId -> return NULL
     | true -> pTcb := pointer to tTask whose .Id == iTaskId 
     ]
   *******/
   pTcb = this->GetTaskPtr(iTaskId);
   if (pTcb==NULL)
       return Traceback(S_sched_INVALID_TASK_ID,
             "tScheduler::EnableTask",iTaskId);
 
   /*******
      [ pTcb invalid -> UNDEFINED
      | *pTcb.Mode==TM_DISABLED || TM_PANIC || TM_ERROR ->
          *pTcb.Mode := TM_ENABLED
          : *pTcb.ExecState := TS_DONE
      | *pTcb.Mode == TM_INIT || TM_ENABLED  && *pTcb.PendingOp == 
	 TPO_DISABLE_TASK ->
	  *pTcb.PendingOp := TPO_NONE
      | true -> nothing happens
      ]
   *******/
   switch (pTcb->Mode)
   {
      case TM_DISABLED:
      case TM_PANIC:
      case TM_ERROR:
         pTcb->Mode = TM_ENABLED;
         pTcb->ExecState = TS_DONE;
         break;
 
      case TM_INIT:
      case TM_ENABLED:
          if (pTcb->PendingOp == TPO_DISABLE_TASK)
             pTcb->PendingOp = TPO_NONE;
          break;

      default:   /* any other mode, do nothing */
         break;
 
   }  /* switch(Mode) */
 
 
    /* [return OK] */
    return OK;
 
} /* tScheduler::EnableTask */
 
 

 
/****
[[U tScheduler::DeleteTask (iTaskId) ]]
****/
ERROR_CODE tScheduler::DeleteTask(int iTaskId)
{
   ERROR_CODE ec;
   tTask *pTcb;
  
   TraceMark(S_sched_ABOUT_TO_DELETE_A_TASK);

   /*******
      [this invalid -> UNDEFINED
      | *this.bInit != TRUE -> return S_sched_SCHEDULER_NOT_INITIALIZED
      ]
   *******/
   if (this->bInit != TRUE)
       return Traceback (S_sched_SCHEDULER_NOT_INITIALIZED,
                         "tScheduler::DeleteTask", iTaskId);
 
   /*******
     [ this invalid -> UNDEFINED
     | *this.TaskSet is not pointer to array of (*this.iNumTasks) tTasks -> 
          UNDEFINED
     | there is no tTask in *this.TaskSet whose .Id == iTaskId -> return NULL
     | true -> pTcb := pointer to tTask whose .Id == iTaskId 
     ]
   *******/
   pTcb = this->GetTaskPtr(iTaskId);
   if (pTcb==NULL)
      return Traceback(S_sched_INVALID_TASK_ID,
                         "tScheduler::DeleteTask",iTaskId);
   
   /*******
      [ pTcb invalid -> UNDEFINED
      | *pTcb.Id not a valid VxWorks task ID -> return errno (taskPrioritySet)
      | task *pTcb.Id cannot be resumed -> return errno (taskResume)
      | *pTcb.Mode==TM_ENABLED && *pTcb.ExecState==TS_DONE ->
          *pTcb.Mode := TM_DELETE
      | *pTcb.Mode==TM_ENABLED && *pTcb.ExecState!=TS_DONE ->
          *pTcb.PendingOp := TPO_DELETE_TASK
      | *pTcb.Mode==TM_INIT -> *pTcb.PendingOp := TPO_DELETE_TASK
      | true -> nothing happens
      ]
   *******/
   switch (pTcb->Mode)
   {
      case TM_INIT:
         pTcb->PendingOp = TPO_DELETE_TASK;
         break;
         
      case TM_ENABLED:
         if (pTcb->ExecState==TS_DONE) 
         {
            pTcb->Mode = TM_DELETE;
            ec = taskPrioritySet(pTcb->Id, SCHED_INIT_PRIORITY);  
            if (ec != OK)
               return Traceback(errnoGet(), "tScheduler::DeleteTask", iTaskId);
            ec = taskResume(pTcb->Id);
            if (ec != OK)
	    {
              logMsg("taskResume failed in tScheduler::DeleteTask, taskId=0x%08x\n",
                                                  pTcb->Id,0,0,0,0,0);
               return Traceback(errnoGet(), "tScheduler::DeleteTask", iTaskId);
	    }
         }
         else
            pTcb->PendingOp = TPO_DELETE_TASK;
         break;
 
      default:   /* any other mode, do nothing */
         break;
 
   }  /* switch(Mode) */
 
 
    /* [return OK] */
    return OK;
 
}  /* tScheduler::DeleteTask */
 
 

 
/********************************************************************
 * 
 * tScheduler::Initialize - Initialize the Scheduler
 * 
 * IN:
 *   this
 *   iMaxTasks
 *   iRate
 *   pRteErrorStream
 *   ptInterruptObject
 *   pRteTime
 *   SCHED_TELNETD_PRIORITY
 *   SCHED_RLOGIND_PRIORITY
 *   SCHED_SHELL_PRIORITY
 * INOUT:
 *  *this.bInit 
 *  *this.iMaxTasks
 * OUT:
 *  *this.Level
 *  *this.iRate
 *  *this.TaskSet 
 *  *this.NumTasks 
 *  *this.ptInterrupt 
 *  *this.pRteTime
 *  *this.pRteError
 *  *this.pEndTaskSet
 *  *this.TaskSet[].bStarving 
 *  *this.TaskSet[].bNoPanic 
 *  *this.TaskSet[].BehaviorOnError 
 *  *this.TaskSet[].PendingOp 
 *  *this.TaskSet[].State 
 *  *this.TaskSet[].pExecFunc 
 *  *this.TaskSet[].pInitFunc 
 *  *this.TaskSet[].pDeleteFunc 
 *  *this.TaskSet[].iPriority 
 *  *this.TaskSet[].tPeriod 
 *  *this.TaskSet[].tOffset 
 *  *this.TaskSet[].ErrorCode 
 *  *this.TaskSet[].ptScheduler 
 *  *this.schedTelnetdPriority 
 *  *this.schedRlogindPriority 
 *  *this.schedShellPriority 
 *  *this.TaskSet[].Mode
 *  *this.Taskset[].iTaskPosition
 *
 *
 * SPEC:
      [[D  tScheduler::Initialize (int        iPriority,
                                   int        iMaxTasks,
                                   INT_RATE   iRate,
                                   TIME       *pRteTime,
                                   tInterrupt *ptInterruptObject) ]
      [ this invalid => UNDEFINED
      | *this.bInit == TRUE => return := S_sched_SCHEDULER_ALREADY_INITIALIZED
      | iMaxTasks < 1 => return := S_sched_INVALID_VALUE_FOR_MAXTASKS
      | task "tTelnetd" not found || task "tRlogind" not found || 
                        task "tShell" not found => 
          put VxWorks errno (taskPriorityGet or Set) on RTE error stream
      | unable to allocate memory for array of tTask of dimension iMaxTasks => 
          return := S_sched_INSUFFICIENT_MEMORY
      | unable to allocate memory for array of (tTask *) of dim. iMaxTasks => 
           return := S_sched_INSUFFICIENT_MEMORY 
      | ptInterruptObject invalid => UNDEFINED
      | *ptInterruptObject.bInit == FALSE => 
           return := S_interrupt_E_INT_NOT_INITIALIZED
      | *ptInterruptObject.pBoard is invalid => UNDEFINED
      | *ptInterruptObject.*pBoard is not initialized => 
           return := uninit error for *ptInterruptObject.pBoard
      | *ptInterruptObject.iWhich is not valid interrupt source on 
              *ptInterruptObject.*pBoard || *ptInterruptObject.iWhich is not 
              clock-driven interrupt source for *ptInterruptObject.*pBoard =>
           return := invalid interrupt source error for *ptInterruptObject.*pBoard
      | *ptInterruptObject.pBoard.pFunc is invalid => UNDEFINED
      | *ptInterruptObject.bConnected == TRUE => 
           return := S_interrupt_E_INT_ALREADY_CONNECTED
      | *ptInterruptObject.bEnabled == TRUE =>
           return := S_interrupt_E_INT_ALREADY_ENABLED
      | &(*this.SchedIsr) == NULL => return := S_interrupt_E_NULL_FUNCTION_POINTER
      | true =>
          *this.schedTelnetdPriority := tTelnetd's priority
          : *this.schedRlogindPriority := tRlogind's priority
          : *this.schedShellPriority := tShell's priority
          : priority of tTelnetd task := SCHED_TELNETD_PRIORITY
          : priority of tRlogind task := SCHED_RLOGIND_PRIORITY
          : priority of tShell task := SCHED_SHELL_PRIORITY
          : *this.hTaskSetPtr[0..*this.iMaxTasks-1] := NULL
          : *this.TaskSet[0..*this.iMaxTasks-1].Id := ERROR
          : *this.TaskSet[0..*this.iMaxTasks-1].bStarving := FALSE
          : *this.TaskSet[0..*this.iMaxTasks-1].bNoPanic := FALSE
          : *this.TaskSet[0..*this.iMaxTasks-1].BehaviorOnError := TEB_DISABLE
          : *this.TaskSet[0..*this.iMaxTasks-1].PendingOp := TPO_NONE
          : *this.TaskSet[0..*this.iMaxTasks-1].Exec_State := TS_DONE
          : *this.TaskSet[0..*this.iMaxTasks-1].Mode := TM_INIT
          : *this.TaskSet[0..*this.iMaxTasks-1].tScheduledTime := 0
          : *this.TaskSet[0..*this.iMaxTasks-1].pExecFunc := NULL
          : *this.TaskSet[0..*this.iMaxTasks-1].pInitFunc := NULL
          : *this.TaskSet[0..*this.iMaxTasks-1].pDeleteFunc := NULL
          : *this.TaskSet[0..*this.iMaxTasks-1].iPriority := 255
          : *this.TaskSet[0..*this.iMaxTasks-1].tPeriod := 0
          : *this.TaskSet[0..*this.iMaxTasks-1].tOffset := 0
          : *this.TaskSet[0..*this.iMaxTasks-1].ErrorCode := OK
          : *this.TaskSet[0..*this.iMaxTasks-1].iTaskPosition := -1
          : *this.TaskSet[0..*this.iMaxTasks-1].ptScheduler := NULL
          : *this.bInit := TRUE
          : *this.Level := SL_ENABLED
          : *this.iRate := iRate
          : *this.pRteError := pRteErrorStream
          : *this.pRteTime := pRteTime
          : *this.ptInterrupt := ptInterruptObject
          : *this.iMaxTasks := iMaxTasks
          : *this.iNumTasks := 0 
          : *this.pEndTaskSet := *this.TaskSet
          [ *ptInterruptObject.*pBoard.bInit == FALSE =>
               return := error code for *ptInterruptObject.pBoard
          | *ptInterruptObject.iWhich is out of range for 
                                               *ptInterruptObject.*pBoard => 
               return := error for *ptInterruptObject.*pBoard
          | *ptInterruptObject.uiIntNum < MIN_BUS_INT_PRIORITY ||
             *ptInterruptObject.uiIntNum > MAX_BUS_INT_PRIORITY =>
               return := error for *ptInterruptObject.*pBoard
          | true => 
               enable board interrupt source *ptInterruptObject.iWhich
               : connect board interrupt source *ptInterruptObject.iWhich to
                 VME interrupt *ptInterruptObject.uiIntNum at 
                                               *ptInterruptObject.uiIntVector
          ]
          : iRate := nearest value to iRate that is valid for 
                                                   *ptInterruptObject.*pBoard
          : *ptInterruptObject.*pBoard.pRate[*ptInterruptObject.iWhich] := 
                                   nearest value to iRate that is valid for 
                                   *ptInterruptObject.*pBoard
          : *ptInterruptObject.bRateSetYet := TRUE
          : *ptInterruptObject.bConnected := TRUE
          : *ptInterruptObject.bEnabled := TRUE
          : *ptInterruptObject.pBoard interrupt clock is enabled
          : *ptInterruptObject.pBoard.pFunc starts getting called at 
                                          *ptInterruptObject.rate
          : return := OK
      ]]
 *   
 * RETURNS:  (type ERROR_CODE)
 *     S_sched_SCHEDULER_ALREADY_INITIALIZED
 *     S_sched_INVALID_VALUE_FOR_MAXTASKS
 *     VxWorks errno (taskPriorityGet or taskPrioritySet)
 *     S_sched_INSUFFICIENT_MEMORY
 *     S_interrupt_E_INT_NOT_INITIALIZED
 *     uninit error for *ptInterruptObject.pBoard
 *     invalid interrupt source error for *ptInterruptObject.*pBoard
 *     S_interrupt_E_INT_ALREADY_CONNECTED
 *     S_interrupt_E_INT_ALREADY_ENABLED
 *     S_interrupt_E_NULL_FUNCTION_POINTER
 *     error codes for *ptInterruptObject.pBoard
 *     OK  
 * 
 */
 
ERROR_CODE tScheduler::Initialize (int        iPriority,
                                   int        iMaxTasks,
                                   INT_RATE   iRate,
                                   TIME       *pRteTime,
                                   tPeriodicTimeInterrupt *ptInterruptObject)
{
   ERROR_CODE ec;
   int i,
       iTempTaskId;
   char sMsgOut[256];
   
    TraceMark(S_sched_ENTERING_INIT);


    /*******
      [ this invalid -> UNDEFINED
      | *this.bInit == TRUE -> return S_sched_SCHEDULER_ALREADY_INITIALIZED
      | true -> nothing happens
      ]
   ********/

   if  (this->bInit == TRUE)
       return Traceback (S_sched_SCHEDULER_ALREADY_INITIALIZED, 
                         "tScheduler::Initialize", 
                         iPriority,
                         iMaxTasks, 
                         iRate, 
                         pRteTime,
                         ptInterruptObject);



   /*******
      [ iMaxTasks < 1 -> return S_sched_INVALID_VALUE_FOR_MAXTASKS
      | true -> nothing happens
      ]
   ********/
   if (iMaxTasks < 1)
       return Traceback ( S_sched_INVALID_VALUE_FOR_MAXTASKS,
                          "tScheduler::Initialize", 
                          iPriority,
                          iMaxTasks, 
                          iRate, 
                          pRteTime,
                          ptInterruptObject);


    /*
     *  First adjust the priorities of certain VxWorks tasks (if they exist)
     */


            /***** change priority of tTelnetd task, if any *****/

    /*****
       [ this invalid -> UNDEFINED
       | task "tTelnetd" not found -> 
            put VxWorks errno (taskPriorityGet/Set) on RTE error stream
       | true ->
            *this.schedTelnetdPriority := tTelnetd's priority
            : priority of tTelnetd task := SCHED_TELNETD_PRIORITY
       ]
    ****/
    if ( (iTempTaskId = taskNameToId("tTelnetd")) == ERROR )
       TraceMsg(tmt_WARN, OK, "tTelnetd does not exist");
    else if ( (ec = taskPriorityGet(iTempTaskId,
                               &(this->schedTelnetdPriority)) ) == ERROR) 
    {
       TraceMsg(tmt_WARN, OK, "taskPriorityGet failed for tTelnetd");
       GetErrorStream().Put( errnoGet() );
    }
    else if ( (ec = taskPrioritySet(iTempTaskId,
                                       SCHED_TELNETD_PRIORITY)) == ERROR) 
    {
       TraceMsg(tmt_WARN, OK, "taskPrioritySet failed for tTelnetd");
       GetErrorStream().Put( errnoGet() );
    }


            /***** change priority of tRlogind task, if any *****/

    /*****
       [ this invalid -> UNDEFINED
       | task "tRlogind" not found -> 
            put VxWorks errno (taskPriorityGet/Set) on RTE error stream
       | true ->
            *this.schedRlogindPriority := tRlogind's priority
            : priority of tRlogind task := SCHED_RLOGIND_PRIORITY
       ]
    ****/
    if ( (iTempTaskId = taskNameToId("tRlogind")) == ERROR )
       TraceMsg(tmt_WARN, OK, "tRlogind does not exist");
    else if ( (ec = taskPriorityGet(iTempTaskId,
                               &(this->schedRlogindPriority)) ) == ERROR) 
    {
       TraceMsg(tmt_WARN, OK, "taskPriorityGet failed for tRlogind");
       GetErrorStream().Put( errnoGet() );
    }
    else if ( (ec = taskPrioritySet(iTempTaskId,
                                       SCHED_RLOGIND_PRIORITY)) == ERROR) 
    {
       TraceMsg(tmt_WARN, OK, "taskPrioritySet failed for tRlogind");
       GetErrorStream().Put( errnoGet() );
    }


            /***** change priority of tShell task, if any *****/

    /*****
       [ this invalid -> UNDEFINED
       | task "tShell" not found -> 
            put VxWorks errno (taskPriorityGet/Set) on RTE error stream
       | true ->
            *this.schedShellPriority := tShell's priority
            : priority of tShell task := SCHED_SHELL_PRIORITY
       ]
    ****/

         /* this is a kludge due to tShell weirdities */
    if ( (iTempTaskId = taskNameToId("tShell")) == ERROR )
    {
       TraceMsg(tmt_WARN, OK, "Waiting for tShell...");
       taskDelay(300);
    }

    if ( (iTempTaskId = taskNameToId("tShell")) == ERROR )
       TraceMsg(tmt_WARN, OK, "tShell does not exist");
    else if ( (ec = taskPriorityGet(taskNameToId("tShell"),
                               &(this->schedShellPriority)) ) == ERROR)
    {
       TraceMsg(tmt_WARN, OK, "taskPriorityGet failed for tShell");
       GetErrorStream().Put( errnoGet() );
    }
    else if ( (ec = taskPrioritySet(taskNameToId("tShell"),
                                       SCHED_SHELL_PRIORITY)) == ERROR)
    {
       TraceMsg(tmt_WARN, OK, "taskPrioritySet failed for tShell");
       GetErrorStream().Put( errnoGet() );
    }


            /***** change priority of tWdbTask, if any *****/

    if ( (iTempTaskId = taskNameToId("tWdbTask")) == ERROR )
       TraceMsg(tmt_WARN, OK, "tWdbTask does not exist");
    else if ( (ec = taskPrioritySet(iTempTaskId, SCHED_SHELL_PRIORITY)) == ERROR)
    {
       TraceMsg(tmt_WARN, OK, "taskPriorityGet failed for tWdbTask");
       GetErrorStream().Put( errnoGet() );
    }

    else if ( (ec = taskPrioritySet(taskNameToId("tWdbTask"),
                                       SCHED_SHELL_PRIORITY)) == ERROR)
    {
       TraceMsg(tmt_WARN, OK, "taskPrioritySet failed for tWdbTask");
       GetErrorStream().Put( errnoGet() );
    }


   TraceMark(S_sched_ADJUSTED_VXWORKS_TASK_PRIOS);


 
   /********
      [ this invalid -> UNDEFINED
      | unable to allocate memory for array of tTask of dimension iMaxTasks -> 
          return S_sched_INSUFFICIENT_MEMORY
      | true -> memory allocated for *this.TaskSet[]
      ]
   ********/
   this->TaskSet = new tTask[iMaxTasks];

   if (this->TaskSet == NULL)
   {
      return Traceback (S_sched_INSUFFICIENT_MEMORY,
                        "tScheduler::Initialize",
                         iPriority,
                        iMaxTasks,
                        iRate,
                        pRteTime,
                        ptInterruptObject);
   }

   /*******
      [ this invalid -> UNDEFINED
      | unable to allocate memory for array of (tTask *) of dim. iMaxTasks -> 
           return S_sched_INSUFFICIENT_MEMORY 
      | true -> memory allocated for *this.hTaskSetPtr[]
      ]
   ********/
   this->hTaskSetPtr = (tTask **) malloc (iMaxTasks * sizeof(tTask *) );

   if ( this->hTaskSetPtr == NULL)
       return Traceback (S_sched_INSUFFICIENT_MEMORY,
                         "tScheduler::Initialize",
                         iPriority,
                         iMaxTasks,
                         iRate,
                         pRteTime,
                         ptInterruptObject);



   
     /**********
        [ this invalid -> UNDEFINED
        | true ->
           for every i in [0..*this.iMaxTasks] ->
             *this.hTaskSetPtr[i] = NULL
             : call the initialize method for every task such
               that *this.TaskSet[i]
        ]
     ********/
     for (i=0; i<iMaxTasks; i++)
     {
         this->hTaskSetPtr[i] = NULL;
         /********
         [ this invalid -> UNDEFINED
         | true ->
             *this.Id := ERROR
             : *this.bStarving := FALSE
             : *this.bNoPanic := FALSE
             : *this.BehaviorOnError := TEB_DISABLE
             : *this.PendingOp := TPO_NONE
             : *this.ExecState := TS_DONE
             : *this.Mode := TM_INIT
             : *this.tScheduledTime := 0
             : *this.pExecFunc := NULL
             : *this.pInitFunc := NULL
             : *this.pDeleteFunc := NULL
             : *this.iPriority := 255
             : *this.tPeriod := 0
             : *this.tOffset := 0
             : *this.ErrorCode := OK
             : *this.iTaskPosition := -1
             : *this.ptScheduler := NULL
         ]
         ******/
         this->TaskSet[i].Initialize();
     }
      
   TraceMark(S_sched_TASK_SET_INITIALIZED);

   /*******
      [ this invalid -> UNDEFINED
      | true ->
          *this.bInit := TRUE
          : *this.Level := SL_ENABLED
          : *this.iRate := iRate
          : *this.pRteTime := pRteTime
          : *this.ptInterrupt := ptInterruptObject
          : *this.iMaxTasks := iMaxTasks
          : *this.iNumTasks := 0 
          : *this.pEndTaskSet := *this.TaskSet
      ]
   *******/

   this->bInit = TRUE;
   this->Level = SL_ENABLED;
   this->iRate = iRate;

   if (iRate!=0) 
      this->tSchedPeriod = TIME_TICKS_PER_SECOND / iRate;    /* int div */
   else 
      this->tSchedPeriod = 0;                   /* should never happen! */

   this->pRteTime = pRteTime;
   this->ptInterrupt = ptInterruptObject;
   this->iMaxTasks = iMaxTasks;
   this->iNumTasks = 0;
   this->pEndTaskSet = this->TaskSet;
   this->iCountInterrupts = 0;


    /*******
       [ ptInterruptObject invalid -> UNDEFINED
       | *ptInterruptObject.bInit == FALSE -> 
            return S_interrupt_E_INT_NOT_INITIALIZED
       | &iRate invalid -> UNDEFINED
       | *ptInterruptObject.pBoard is invalid -> UNDEFINED
       | *ptInterruptObject.*pBoard is not initialized -> 
            return uninit error for *ptInterruptObject.pBoard
       | *ptInterruptObject.iWhich is not valid interrupt source on 
                                                 *ptInterruptObject.*pBoard ||
         *ptInterruptObject.iWhich is not clock-driven interrupt source for 
                                                 *ptInterruptObject.*pBoard ->
           return invalid interrupt source error for *ptInterruptObject.*pBoard
       | *&iRate > highest possible rate for *ptInterruptObject.*pBoard ->
           *&iRate := highest possible rate for *ptInterruptObject.*pBoard
           : *ptInterruptObject.*pBoard.pRate[*ptInterruptObject.iWhich] := 
                  higest possible rate for *ptInterruptObject.*pBoard
           : *ptInterruptObject.bRateSetYet := TRUE
           : *this.iRate := iRate  (reassignment, since iRate was changed)
           :  ec := S_interrupt_W_INVALID_RATE 
       | *&iRate not valid for *ptInterruptObject.*pBoard -> 
           *&iRate := valid rate nearest and higher to *&iRate
           : *ptInterruptObject.*pBoard.pRate[*ptInterruptObject.iWhich] := 
                                         valid rate nearest/higher to *&iRate
           : *ptInterruptObject.bRateSetYet := TRUE
           : *this.iRate := iRate  (reassignment, since iRate was changed)
           :  ec := S_interrupt_W_INVALID_RATE 
       | true -> *ptInterruptObject.*pBoard.pRate[*ptInterruptObject.iWhich] := 
                                                          *&iRate
           *ptInterruptObject.bRateSetYet := TRUE
           :  ec := OK 
       ]
    *******/
    ec = ptInterruptObject->RateSet( &iRate );

    if ( ec != OK )
    {
       if (ec == S_interrupt_W_INVALID_RATE)
       {
          this->iRate = iRate;    /* RateSet() changed iRate to valid rate */

          sprintf(sMsgOut,
                "WARNING (tScheduler::Initialize): iRate set to %d",iRate);
         
          TraceMsg(tmt_WARN,OK,sMsgOut);
       }
       else return Traceback (ec,
                         "tScheduler::Initialize", 
                         iPriority,
                         iMaxTasks, 
                         iRate,    /* <--- note this value may have changed */
                         pRteTime,
                         ptInterruptObject);
    }
   


    /* Now connect the scheduler to the interrupt. */
    /*******
       [ this is invalid -> UNDEFINED
       | ptInterruptObject is invalid -> UNDEFINED
       | *ptInterruptObject.bInit == FALSE -> 
            return S_interrupt_E_INT_NOT_INITIALIZED
       | *ptInterruptObject.bConnected == TRUE -> 
            return S_interrupt_E_INT_ALREADY_CONNECTED
       | *ptInterruptObject.bEnabled == TRUE ->
            return S_interrupt_E_INT_ALREADY_ENABLED
       | *ptInterruptObject.pBoard is invalid -> UNDEFINED
       | &(*this.SchedIsr) is invalid -> UNDEFINED
       | ( &(*this.SchedIsr) is not a pointer to valid C function ||
                ! (&(*this.SchedIsr) is a pointer to valid C++ function && 
                                                 (this == this)) ) -> UNDEFINED
       | &(*this.SchedIsr) == NULL -> 
            return S_interrupt_E_NULL_FUNCTION_POINTER
       | true -> 
           [ *ptInterruptObject.*pBoard.bInit == FALSE ->
               return error code for *ptInterruptObject.pBoard
           | *ptInterruptObject.iWhich is out of range for 
                                              *ptInterruptObject.*pBoard -> 
               return error for *ptInterruptObject.*pBoard
           | *ptInterruptObject.uiIntNum < MIN_BUS_INT_PRIORITY ||
             *ptInterruptObject.uiIntNum > MAX_BUS_INT_PRIORITY ->
               return error for *ptInterruptObject.*pBoard
           | true -> 
               enable board interrupt source *ptInterruptObject.iWhich
               : connect board interrupt source *ptInterruptObject.iWhich to
                 VME interrupt *ptInterruptObject.uiIntNum at 
                                               *ptInterruptObject.uiIntVector
           ]
           : *ptInterruptObject.bConnected := TRUE
           : ec := OK
       ]
    *********/
    ec = ptInterruptObject->Connect( (FUNCPTR) &(this->SchedIsr), (int)this );
   
    if (ec != OK)
       return Traceback (ec,
                         "tScheduler::Initialize",
                         iPriority,
                         iMaxTasks,
                         iRate,
                         pRteTime,
                         ptInterruptObject);

      

   TraceMark(S_sched_CONNECTED_INTERRUPT_TO_SCHEDISR);


    /* enable interrupt */
    /********
       [ ptInterruptObject is invalid -> UNDEFINED 
       | More than one CPU is connected to *ptInterruptObject.uiIntNum -> 
             UNDEFINED
       | *ptInterruptObject.bInit == FALSE -> 
             return S_hm_ E_INT_NOT_INITIALIZED
       | *ptInterruptObject.bEnabled == TRUE -> 
             return S_hm_E_INT_ALREADY_ENABLED
       | *ptInterruptObject.bRateSetYet == FALSE ->  
             return S_hm_E_INT_RATE_NOT_SET
       | *ptInterruptObject.bConnected == FALSE -> 
             return S_ hm_E_INT_NOT_CONNECTED
       | true -> 
         [ *ptInterruptObject.*pBoard.bInit == FALSE -> 
               return init error for *ptInterruptObject.*pBoard
         | *ptInterruptObject.iWhich not valid range for 
                                         *ptInterruptObject.*pBoard -> 
               return error for *ptInterruptObject.*pBoard
         | true -> 
               enable interrupt *ptInterruptObject.iWhich on 
                                                    *ptInterruptObject.*pBoard
               : Begin sending interrupt number 
           *ptInterruptObject.*pBoard.puwIntNum[*ptInterruptObject.iWhich] 
                    to the VME bus.  The ISR on the CPU connected to
                    *ptInterruptObject.*pBoard interrupt source 
                    *ptInterruptObject.iWhich will be called at a rate of 
                    *ptInterruptObject.*pBoard.pRate[*ptInterruptObject.iWhich]
                    until a Disable command is given.
         ]
         : *ptInterruptObject.bEnabled := TRUE
         : ec := OK
       ]
    ********/
    ec = ptInterruptObject->Enable();
    if (ec != OK)
       return Traceback (ec,
                         "tScheduler::Initialize",
                         iPriority,
                         iMaxTasks,
                         iRate,
                         pRteTime,
                         ptInterruptObject);


    TraceMark(S_sched_LEAVING_INIT);

    return OK;
   
}  /* tScheduler::Initialize */



/******************************************************************************
*
* tScheduler::StartNextTask- Goes through task list, finds highest priority 
*                       TS_READY or TS_RUN task and starts it. 
*
* IN : 
*   this
*   *this.bInit
*   *this.Level
*   *this.TaskSet[].Mode
* INOUT :
*   *this.TaskSet[].ExecState
*   OS.RunningTasks
* OUT :
*
* SPEC: 
         [ this invalid -> UNDEFINED
         | *this.bInit != TRUE -> UNDEFINED
         | *this.Level != SL_ENABLED -> return S_sched_NOT_ENABLED
         | *this.TaskSet is not pointer to array of tTasks of dimension 
              (*this.pEndTaskSet - *this.TaskSet) / sizeof(tTask) -> UNDEFINED
         | TaskSet array not arranged in incr. order of .iPriority -> UNDEFINED 
         | there is no task in the TaskSet array whose .Mode is TM_ENABLED and 
                 whose .ExecState is TS_READY or TS_RUN -> 
              return OK 
         | the first task in the TaskSet array whose .Mode is TM_ENABLED and 
                 whose .ExecState is TS_READY or TS_RUN cannot be resumed ->  
              return VxWorks errno (taskResume)
         | true -> 
              that task's .ExecState := TS_RUN
              : that task is resumed
              : return OK
         ] 
*
* RETURNS: 
*    S_sched_NOT_ENABLED
*    VxWorks errno (taskResume)
*    OK
*
************************************************************************/

ERROR_CODE  tScheduler::StartNextTask( void )
{
   FAST ERROR_CODE ec;
   FAST tTask *pThisTask, *pStop;
   FAST TASK_STATE *pTaskState;
   char errString[256];

   // TraceMark(S_sched_ENTERING_STARTNEXTTASK);

   if (this->Level != SL_ENABLED)
         return Traceback(S_sched_NOT_ENABLED, "tScheduler::StartNextTask");

   pStop = this->pEndTaskSet;      /* points just past end of TaskSet array */

   for  ( pThisTask = this->TaskSet;   pThisTask < pStop;   pThisTask++ )
      if ( pThisTask->Mode == TM_ENABLED )
      {
         pTaskState = &(pThisTask->ExecState);

         if ( (*pTaskState==TS_READY) || (*pTaskState==TS_RUN) )
         {
            *pTaskState = TS_RUN;
            ec = taskResume(pThisTask->Id);
            if (ec != OK) 
	    {
	       TraceMsg(tmt_ERROR,S_sched_INVALID_TASK_ID,
			   "tScheduler::StartNextTask: taskResume failed");
               return Traceback( errnoGet(), "tScheduler::StartNextTask" );
	    }
            else
               return OK;
         }

      } /* for () if () */

   return OK;   /* only reach this if there are no tasks ready to run */

}               /* tScheduler::StartNextTask() */


/******************************************************************************
*
* tScheduler::SchedIsr - Runs on every scheduler interrupt stimulus, examines 
*               and updates states of all tasks in task list. 
*
* IN : 
*   this
*   *this.bInit
*   *this.Level
*   *this.TaskSet[].Id
*   *this.TaskSet[].bNoPanic
*   *this.pEndTaskSet
*   SCHED_PANIC_PRIORITY
* INOUT :
*   *this.TaskSet[].ExecState
*   *this.TaskSet[].Mode
* OUT :
*   *this.TaskSet[].tScheduledTime
*   *this.TaskSet[].bStarving
*   *this.*pRteTime
*
* SPEC: 
*     Definition:   due
*        Task *pThisTask is called "due" whenever 
*                 (current time % *pThisTask.tPeriod) == *pThisTask.tOffset).
*
      [ this invalid -> UNDEFINED
      | *this.bInit != TRUE -> UNDEFINED
      | *this.Level != SL_ENABLED -> put S_sched_NOT_ENABLED on RTE error stream
      | manual capture (GetTimeNow) performed during this function -> UNDEFINED
      | *this.TaskSet invalid -> UNDEFINED
      | *this.TaskSet not a pointer to an array of tTasks of dimension 
                 (*this.pEndTaskSet - *this.TaskSet) / sizeof(tTask) ->UNDEFINED
      | true -> 
            *this.*pRteTime :=  time of last interrupt (in 25us since midnight)
            : all due tasks in array with .Mode == TM_ENABLED and 
                  .ExecState == TS_DONE get .ExecState := TS_READY, 
                  .bStarving := FALSE, and .tScheduledTime :=  time of last 
                          interrupt (in 25us units since midnight)
            : all due tasks in array with .Mode == TM_ENABLED and 
                  .ExecState == TS_READY || TS_RUN get .bStarving := TRUE
            : all due tasks in array with .Mode == TM_ENABLED and 
                  .ExecState == TS_READY || TS_RUN and .bNoPanic==TRUE get
                  .tScheduledTime :=  time of last interrupt (in 25us units 
                  since midnight),  .Mode := TM_PANIC, .bStarving := TRUE, 
                  and VxWorks priority of task is changed to 
                  SCHED_PANIC_PRIORITY (unless .Id not a valid VxWorks taskId)
      ] 
*
* RETURNS: nothing; puts msgs on RTE error stream instead
*    S_sched_NOT_ENABLED
*
************************************************************************/

// ********************** kludgey debug stuff  **********************
long int iCountMissedInterrupts = 0;
int ShowMissedInterrupts( void )
{
   printf("Have missed %lu interrupts\n",iCountMissedInterrupts);
   return iCountMissedInterrupts;
}
void ResetMissedInterrupts( void ) { iCountMissedInterrupts = 0; }
// ********************** kludgey debug stuff  **********************

int iNumTimesSchedRan = 0;

void  tScheduler::SchedIsr( void )
{
   FAST ERROR_CODE ec;
   FAST TIME tTimeNow;
   FAST tTask *pThisTask, *pStop;


//********* D/A Bit Twiddle ***********
//   *((int16_t *) 0xefffc9a8) = 5000;


   if (this->Level != SL_ENABLED)
      return;

   tTimeNow = this->ptInterrupt->GetTimeLastInt();

// ********************** kludgey debug stuff  **********************
// if ( tTimeNow != (*(this->pRteTime) + this->tSchedPeriod) ) {
//    extern int iGlobalSmallTime;
//    logMsg("tTimeNow = %lu, SmallTime = %d\n",tTimeNow,iGlobalSmallTime,0,0,0,0);
//    iCountMissedInterrupts++;
// }
// ********************** kludgey debug stuff  **********************

   *(this->pRteTime) += this->tSchedPeriod;

   iNumTimesSchedRan++; 
   if ((tTimeNow % TIME_TICKS_PER_SECOND) == 0) {
      if (iNumTimesSchedRan != (int)(TIME_TICKS_PER_SECOND/this->tSchedPeriod))
         logMsg("Scheduler Ran %d times last second\n", iNumTimesSchedRan,0,0,0,0,0);

      iNumTimesSchedRan = 0;

      if (*(this->pRteTime) != tTimeNow) {
        logMsg("Time sync error: IRIG time = %lu, RTE time = %lu, Synchronizing.\n",
               tTimeNow, *(this->pRteTime),0,0,0,0);
        *(this->pRteTime) = tTimeNow;                /* update Rte time */
      }
   }

   pStop = this->pEndTaskSet;     /* points just past end of TaskSet array */

   for  ( pThisTask = this->TaskSet;    pThisTask < pStop;    pThisTask++ ) 
   {
      // PCI: added iNumStarvingCycles to look at number of times task starved
      //if ((tTimeNow % 40000) == 0) pThisTask->iNumStarvingCycles = 0;

      if ( (pThisTask->Mode==TM_ENABLED) && 
           (tTimeNow % pThisTask->tPeriod == pThisTask->tOffset) )

         switch (pThisTask->ExecState)
         {
            case TS_DONE:
                pThisTask->ExecState = TS_READY;
                pThisTask->bStarving = FALSE;
                pThisTask->tScheduledTime = tTimeNow;
                break;

            case TS_READY:
            case TS_RUN:
                pThisTask->bStarving = TRUE;
                pThisTask->iNumStarvingCycles++;

                if ( pThisTask->bNoPanic )
                {
                   TraceMark(S_sched_TASK_STARVING);

#ifdef A_TASK_IS_STARVING_AND_WE_CANT_FIGURE_OUT_WHICH_ONE
                   logMsg("tScheduler::SchedIsr : Task 0x%x (%s) is starving [Time=%lu]\n",
			   pThisTask->Id,(int)taskName(pThisTask->Id),tTimeNow,0,0,0);
#endif
//                   GetErrorStream().PutNoWait(S_sched_TASK_STARVING,tTimeNow);
                }
                else
                {
                   pThisTask->Mode = TM_PANIC;
                   pThisTask->tScheduledTime = tTimeNow;
                   TraceMark(S_sched_TASK_HAS_PANICKED);
//                   GetErrorStream().PutNoWait(S_sched_TASK_HAS_PANICKED,
//					      tTimeNow);
                   logMsg("TASK_HAS_PANICKED %d tScheduler::SchedIsr\n",pThisTask->Id,0,0,0,0,0);

                   ec = taskPrioritySet( pThisTask->Id, SCHED_PANIC_PRIORITY );
                   if (ec != OK)
//                      GetErrorStream().PutNoWait( errnoGet(), tTimeNow );
                      logMsg("SCHED_PANIC_PRIORITY %d tScheduler::SchedIsr\n",pThisTask->Id,0,0,0,0,0);
                }
                break;

            default:
                break;

         }      /* switch */
   } /* for loop */


   ec = StartNextTask();
   if (ec != OK)
//      GetErrorStream().PutNoWait( ec, tTimeNow );
      logMsg("%d tScheduler::SchedIsr",ec,0,0,0,0,0);

//********* D/A Bit Twiddle *********
//   *((int16_t *) 0xefffc9a8) = 0;

   return;

}               /* tScheduler::SchedIsr() */





/**************************************************************************
*
* tScheduler::Suspender -- start next user's task and suspend current one
*
* IN :
*   this
*   pThisTcb
*   ecTaskErr
*   *this.*pThisTcb.BehaviorOnError
*   *pThisTcb.iTaskPosition
*   SCHED_INIT_PRIORITY
* INOUT :
*   *this.TaskSet[]
*   *this.hTaskSetPtr[]
*   *this.iNumTasks
*   *this.*pThisTcb.Mode
*   *this.*pThisTcb.PendingOp
*   *this.*pThisTcb.ExecState
*   *this.TaskSet[].ExecState??
*   *this.*pRteError
* OUT :
*   *this.*pThisTcb.ErrorCode 
*   *this.*pThisTcb.iPriority
*
* SPEC: 
    [ this invalid -> UNDEFINED
    | pThisTcb invalid -> UNDEFINED
    | *this.TaskSet is not a pointer to  array of (*this.iNumTasks)
         tTasks -> UNDEFINED
    | *pThisTcb.iTaskPosition < 0 -> UNDEFINED
    | *this.iNumTasks <= 0 -> UNDEFINED
    | TaskSet array not arranged in incr. order of .iPriority ->
        UNDEFINED
     | *this.hTaskSetPtr is not an array of pointers which points
	 to an array of tTasks -> UNDEFINED
    | *this.pRteError invalid -> UNDEFINED
    | *pThisTcb.Mode not an element of the set {TM_DELETE, TM_DISABLE,
          TM_INIT, TM_ENABLED, TM_ERROR, TM_PANIC} -> UNDEFINED
    | *pThisTcb.Mode not an element of the set {TPO_DISABLE_TASK,
          TPO_DELETE_TASK, TPO_NONE} - > UNDEFINED
    | *pThisTcb.Mode == TM_DELETE ->
          *this.TaskSet := *this.TaskSet with *pThisTcb removed.
        : *this.hTaskSetPtr[*this.(*pThisTcb.iTaskPosition)] := NULL
        : *this.iNumTasks := *this.iNumTasks - 1
        : [[U Scheduler::StartNextTask()]]
        : delete current task
	: append error codes resulting from task activities to
	   RTE error stream as appropriate
    | ecTaskErr != OK && *pThisTcb.BehaviorOnError == TEB_DISABLE ->
	*this.pRteError := *this.pRteError with ecTaskErr added to it.
        : [[U Scheduler::NewPendingOp]]
        : *pThisTcb.Mode := TM_ERROR
        : [[U Scheduler::Suspend1]]
        : delete current task
    | ecTaskErr != OK && *pThisTcb.BehaviorOnError == TEB_DELETE ->
	*this.pRteError := *this.pRteError with ecTaskErr added to it.
        : [[U Scheduler::NewPendingOp]]
        : *pThisTcb.Mode := TM_DELETE
        : VxWorks task *pThisTcb.Id priority := SCHED_INIT_PRIORITY
        : [[U Scheduler::Suspend1]]
        : return OK
    | ecTaskErr != OK && *pThisTcb.BehaviorOnError == TEB_CONTINUE ->
	*this.pRteError := *this.pRteError with ecTaskErr added to it.
        : [[U Scheduler::SetPendingOpMode]]
        : [[U Scheduler::Suspend1]]
        : return OK
    | true ->
        [[U Scheduler::SetPendingOpMode]]
        : [[ Scheduler::Suspend1]]
    ]]
*    
*
* RETURNS: nothing (puts errors on RTE error stream)
*  except OK if success
*
**************************************************************************/
ERROR_CODE  tScheduler::Suspender(tTask *pThisTcb, ERROR_CODE ecTaskErr)
{
   ERROR_CODE ec;
   int        iTaskSetIndex;

/**************************************************************************/
/******* do we care whether sched is (a) initialized or (b) enabled? ******/
/**************************************************************************/

/* 
 * In every case, we want to set these two values as follows: 
 */

//   TraceMark(S_sched_ENTERING_SUSPENDER);

   /*****
      [ this invalid -> UNDEFINED
      | pThisTcb invalid -> UNDEFINED
      | true -> *pThisTcb.ErrorCode := ecTaskErr
      ]
   *****/
   pThisTcb->ErrorCode = ecTaskErr;
   
   /*****
      [ this invalid -> UNDEFINED
      | pThisTcb invalid -> UNDEFINED
      | true -> *pThisTcb.ExecState := TS_DONE
      ]
   *****/
   pThisTcb->ExecState = TS_DONE;

   /*
    * If .Mode is TM_DELETE then we've already run *pDeleteFunc (if any)
    * so we finish deleting the task by removing its entry in TaskSet now.  
    * Then start next task and kill this one.
    *      [Note that if .Mode is TM_DELETE then we drop out here.]
    */
   /*****
      [ this invalid -> UNDEFINED
      | pThisTcb invalid -> UNDEFINED
      | *pThisTcb.Mode != TM_DELETE -> nothing happens // no longer true
      | *this.Level != SL_ENABLED -> *this.pRteError := *this.pRteError 
  	    with S_sched_NOT_ENABLED added to it.
      | *this.TaskSet invalid -> UNDEFINED
      | *this.TaskSet is not pointer to array of tTasks of dimension
             (*this.pEndTaskSet - *this.TaskSet) / sizeof(tTask) -> UNDEFINED
      | TaskSet array not arranged in incr. order of .iPriority -> UNDEFINED
      | there is no task in the TaskSet array whose .Mode is
           TM_ENABLED and whose .ExecState is TS_READY or TS_RUN -> ec := OK
      | the first task in the TaskSet array whose .Mode is TM_ENABLED and
              whose .ExecState is TS_READY or TS_RUN cannot be resumed ->
 	  *this.pRteError := *this.pRteError with VxWorks errno (taskResume) 
			      added to it
	  *this.pRteError := *this.pRteError with VxWorks errno (taskResume) 
			     added to it.
      | this task cannot be suspended -> 
	 *this.pRteError := *this.pRteError with VxWorks errno (taskResume) 
			    added to it.
      | true ->
          .ExecState of first ready task in task set := TS_RUN
          : first ready task in task set is resumed
          : suspend this task (now gets ignored forever by scheduler)
          :  return OK
      ]
   *****/
   if (pThisTcb->Mode==TM_DELETE)     /* i.e. we already ran *pDeleteFunc() */
   {
      // incr 2
      // remove task from TaskSet

      /******
          [[U Scheduler::RemoveTask(iiId=*pThisTcb.Id)]]
      *****/
      RemoveTask(pThisTcb->Id);


       /****
	  [[U Scheduler::StartNextTask()]]
       ****/
       ec = this->StartNextTask(); 
       if (ec != OK)
         GetErrorStream().Put( ec );


       ec = taskDelete(0);          /* now it gets ignored forever */
      
       /*******
 	  [ this invalid -> UNDEFINED
 	  | true -> *this.pRteError := *this.pRteError with formatted 
		       erronoGet repsonse added to it
 	  ]
       *******/
       if (ec != OK)
          GetErrorStream().Put( errnoGet() );
       
       return OK;   /* just in case we do come back, we exit anyway */
   }

   /*
    * Now check to see whether there were any pending operations, such
    * as Delete orDisable, that were waiting until now to take effect:
    */

   /*****
      [ pThisTcb invalid -> UNDEFINED
      | *pThisTcb.PendingOp == TPO_DISABLE_TASK ->
          *pThisTcb.Mode := TM_DISABLED
          : *pThisTcb.PendingOp := TPO_NONE
      | *pThisTcb.PendingOp == TPO_DELETE_TASK && 
                *pThisTcb.Id not a valid VxWorks task id ->
          *pThisTcb.Mode := TM_DELETE
          : *pThisTcb.PendingOp := TPO_NONE
          : 
      | *pThisTcb.PendingOp == TPO_DELETE_TASK ->
          *pThisTcb.Mode := TM_DELETE
          : *pThisTcb.PendingOp := TPO_NONE
          : VxWorks task *pThisTcb.Id priority := SCHED_INIT_PRIORITY
          : ec := OK
      | true -> nothing happens
      ]
   *****/
   switch (pThisTcb->PendingOp)
   {
      case TPO_DISABLE_TASK:
         if (pThisTcb->Mode == TM_INIT)
         {
            ec = taskPrioritySet(pThisTcb->Id, pThisTcb->iPriority);
            if (ec != OK) {
              logMsg("Error 0x%x changing priority of task 0x%x to %d\n",
                     ec, pThisTcb->Id, pThisTcb->iPriority,0,0,0);
               GetErrorStream().Put( errnoGet() );
            }
         }  /* if (TM_INIT) */

         pThisTcb->Mode = TM_DISABLED;
         pThisTcb->PendingOp = TPO_NONE;
         break;

      case TPO_DELETE_TASK:
         pThisTcb->Mode = TM_DELETE;
         pThisTcb->PendingOp = TPO_NONE;

         ec = taskPrioritySet(pThisTcb->Id, SCHED_INIT_PRIORITY);  
         if (ec != OK)
            GetErrorStream().Put( errnoGet() );

         break;

      default:
         break;
   }

   /*
    * If user's function returned a non-OK error code, respond according to
    * how user set .BehaviorOnError in AddTask.  Function's error code is
    * always put on the RTE error stream.  If .BehOnError is TEB_DISABLE,
    * .Mode is set to TM_ERROR, which will act like TM_DISABLED since the
    * Scheduler will ignore it.  If TEB_CONTINUE, then the .Mode is not
    * changed so the task will continue to get scheduled.  If TEB_DELETE,
    * then it becomes TM_DELETE so we can run *pDeleteFunc().
    */

   /*******
      [ this invalid -> UNDEFINED
      | pThisTcb invalid -> UNDEFINED
      | ecTaskErr != OK && *pThisTcb.BehaviorOnError == TEB_DISABLE ->
	  *this.pRteError := *this.pRteError with ecTaskErr added to it.
          : *pThisTcb.Mode := TM_ERROR
      | ecTaskErr != OK && *pThisTcb.BehaviorOnError == TEB_DELETE &&
                *pThisTcb.Id not a valid VxWorks task id ->
	  *this.pRteError := *this.pRteError with ecTaskErr added to it.
          : *pThisTcb.Mode := TM_DELETE
	  : *this.pRteError := *this.pRteError with VxWorks error 
	                       (taskPrioritySet) added to it.
          : ec := VxWorks errno (taskPrioritySet) 
      | ecTaskErr != OK && *pThisTcb.BehaviorOnError == TEB_DELETE ->
	  *this.pRteError := *this.pRteError with ecTaskErr added to it.
          : *pThisTcb.Mode := TM_DELETE
          : VxWorks task *pThisTcb.Id priority := SCHED_INIT_PRIORITY
          : ec := OK
      | ecTaskErr != OK -> 
	  *this.pRteError := *this.pRteError with ecTaskErr added to it.
      | true -> nothing happens
      ]
   ******/
   if (ecTaskErr != OK)
   {
       GetErrorStream().Put( ecTaskErr );

       switch (pThisTcb->BehaviorOnError)
       {
          case TEB_DISABLE:
             pThisTcb->Mode = TM_ERROR;           /* acts like TM_DISABLED */
             break;

          case TEB_DELETE:
             pThisTcb->Mode = TM_DELETE;

             ec = taskPrioritySet(pThisTcb->Id, SCHED_INIT_PRIORITY);  
             if (ec != OK)
                GetErrorStream().Put( errnoGet() );

             break;

          case TEB_CONTINUE:
             /* do nothing; user wants function to continue being scheduled */
             break;

          default: 
                break;
        }
   }

   /*
    * After all this, if .Mode is still TM_INIT, then we want to move task
    * into its normal operation mode.  Set the .Mode to TM_ENABLED and 
    * change the VxWorks priority of the task to the user-requested value.
    */

   /*****
      [ pThisTcb invalid -> UNDEFINED
      | *pThisTcb.Mode ==TM_INIT && *pThisTcb.Id not a valid VxWorks task id 
	 -> *pThisTcb.Mode := TM_ENABLED 
          : *this.pRteError := *this.pRteError with VxWorks errno 
			       (taskPrioritySet) added to it.
      | *pThisTcb.Mode == TM_INIT ->
          *pThisTcb.Mode := TM_ENABLED
          : priority of task *pThisTcb.Id := *pThisTcb.iPriority
      | true -> nothing happens
      ]
   *****/
   if (pThisTcb->Mode == TM_INIT)
   {
      pThisTcb->Mode = TM_ENABLED;
      ec = taskPrioritySet(pThisTcb->Id, pThisTcb->iPriority);  
      if (ec != OK) {
        logMsg("Error 0x%x changing priority of task 0x%x to %d\n",
               ec, pThisTcb->Id, pThisTcb->iPriority,0,0,0);
         GetErrorStream().Put( errnoGet() );
      }

   }  /* if (TM_INIT) */


   /*
    * Now that we've modified the task information so that it will be treated
    * correctly by the Scheduler, call StartNextTask to resume next scheduled
    * task in the set (if any).
    */

   /******
      [ this invalid -> UNDEFINED
          *this.pRteError := *this.pRteError with S_sched_NOT_ENABLED error
			     added to it.
      | *this.TaskSet invalid -> UNDEFINED
      | *this.TaskSet is not pointer to array of tTasks of dimension 
             (*this.pEndTaskSet - *this.TaskSet) / sizeof(tTask) -> UNDEFINED
      | TaskSet array not arranged in incr. order of .iPriority -> UNDEFINED 
      | there is no task in the TaskSet array whose .Mode is 
           TM_ENABLED and whose .ExecState is TS_READY or TS_RUN -> ec := OK
      | the first task in the TaskSet array whose .Mode is TM_ENABLED and
              whose .ExecState is TS_READY or TS_RUN cannot be resumed ->
          *this.pRteError := *this.pRteError with VxWorks errno 
			       (taskResume) added to it.
      true ->
          that task's .ExecState := TS_RUN
          : that task is resumed
          : ec := OK
      ]
   *****/
   ec = this->StartNextTask(); 
   if (ec != OK)
      GetErrorStream().Put( ec );
       

   /*
    * Now we suspend the current task, and wait until the Scheduler resumes
    * it next time it comes due... unless the .Mode is TM_DELETE, in which case
    * we are ready to go ahead and run *pDeleteFunc(), so we return to
    * taskSkel directly instead of suspending.
    */

   /*****
      [ this invalid -> UNDEFINED
      | *this.pRteError invalid -> UNDEFINED
      | pThisTcb invalid -> UNDEFINED
      | *pThisTcb.Mode != TM_DELETE && current task cannot be suspended -> 
          *this.pRteError := *this.pRteError with formatted erronoGet 
			     repsonse added to it
      | *pThisTcb.Mode != TM_DELETE -> 
           current task is suspended
           : ec := OK
      | true -> nothing happens (i.e. return to taskSkel)
      ]
   *****/ 
   if (pThisTcb->Mode != TM_DELETE)  /* if deleting, don't suspend now */
   {
      ec = taskSuspend(0);
      if (ec != OK)
         GetErrorStream().Put( errnoGet() );
   }
       
   /* [return OK] */
   return(OK);

}  /* tScheduler::Suspender */



/********************************************************************
 * 
 * tTask::Initialize - Initialize the task control block object
 *                                 (with dummy values)
 * 
 * IN:
 *   this
 * INOUT:
 * OUT:
 *  *this.Id
 *  *this.bStarving 
 *  *this.bNoPanic 
 *  *this.BehaviorOnError 
 *  *this.PendingOp 
 *  *this.ExecState 
 *  *this.Mode 
 *  *this.pExecFunc 
 *  *this.pInitFunc 
 *  *this.pDeleteFunc 
 *  *this.iPriority 
 *  *this.tPeriod 
 *  *this.tOffset 
 *  *this.tScheduledTime
 *  *this.ErrorCode 
 *  *this.ptScheduler 
 *
 *
 * SPEC:
      [ this invalid -> UNDEFINED
      true -> 
          *this.Id := ERROR
          : *this.bStarving := FALSE
          : *this.bNoPanic := FALSE
          : *this.BehaviorOnError := TEB_DISABLE
          : *this.PendingOp := TPO_NONE
          : *this.ExecState := TS_DONE
          : *this.Mode := TM_INIT
          : *this.tScheduledTime := 0
          : *this.pExecFunc := NULL
          : *this.pInitFunc := NULL
          : *this.pDeleteFunc := NULL
          : *this.iPriority := 255
          : *this.tPeriod := 0
          : *this.tOffset := 0
          : *this.ErrorCode := OK
          : *this.iTaskPosition := -1
          : *this.ptScheduler := NULL
          : *this.iTaskPosition
          : return OK
      ]
 *   
 * RETURNS:  (type ERROR_CODE)
 *     OK  
 * 
 ************************************************************************/

ERROR_CODE tTask::Initialize ( void )
{
   this->Initialize( ERROR, FALSE, FALSE, TEB_DISABLE, TPO_NONE, TS_DONE,
           TM_INIT, 0, NULL, NULL, NULL, 255, 0, 0, OK, -1, NULL);

   return OK;

}   /* tTask::Initialize( void ) */


/********************************************************************
 *
 * tTask::Initialize - Initialize the task control block object
 *                               (with user-supplied values)
 *
 * IN:
 *   this
 * INOUT:
 * OUT:
 *  *this.Id
 *  *this.bStarving
 *  *this.bNoPanic
 *  *this.BehaviorOnError
 *  *this.PendingOp
 *  *this.ExecState
 *  *this.Mode
 *  *this.pExecFunc
 *  *this.pInitFunc
 *  *this.pDeleteFunc
 *  *this.iPriority
 *  *this.tPeriod
 *  *this.tOffset
 *  *this.ErrorCode
 *  *this.ptScheduler
 *  *this.iTaskPosition
 *  *this.tScheduledTime
 *
 *
 * SPEC:
      [ this invalid -> UNDEFINED
      true ->
	 [ for every task in TaskSet[0..iMaxTasks-1] ->
            *this.Id := aId
            : *this.bStarving := abStarving
            : *this.bNoPanic := abNoPanic
            : *this.BehaviorOnError := aBehaviorOnError
            : *this.PendingOp := aPendingOp
            : *this.ExecState := aExecState
            : *this.Mode := aMode
            : *this.tScheduledTime := atScheduledTime
            : *this.pExecFunc := apExecFunc
            : *this.pInitFunc := apInitFunc
            : *this.pDeleteFunc := apDeleteFunc
            : *this.iPriority := aiPriority
            : *this.tPeriod := atPeriod
            : *this.tOffset := atOffset
            : *this.ErrorCode := aErrorCode
            : *this.iTaskPosition :=  aiTaskPosition
            : *this.ptScheduler := aptScheduler
	  ]
	  hTaskSetPtr[0..iMaxTasks] := NULL
          : return OK
      ]
 *
 * RETURNS:  (type ERROR_CODE)
 *     OK 
 *
 ************************************************************************/
 
ERROR_CODE tTask::Initialize ( int aId,
                               BOOL abStarving,
                               BOOL abNoPanic,
                               TASK_ERROR_BEHAVIOR aBehaviorOnError,
                               TASK_PENDING_OPERATION aPendingOp,
                               TASK_STATE aExecState,
                               TASK_MODE aMode,
                               TIME atScheduledTime,
                               FUNCPTR apExecFunc,
                               FUNCPTR apInitFunc,
                               FUNCPTR apDeleteFunc,
                               int aiPriority,
                               TIME atPeriod,
                               TIME atOffset,
                               ERROR_CODE aErrorCode,
                               int aiTaskPosition,
                               tScheduler *aptScheduler)
{

    this->Id = aId;
    this->bStarving = abStarving;
    this->iNumStarvingCycles = 0;
    this->bNoPanic = abNoPanic;
    this->BehaviorOnError = aBehaviorOnError;
    this->PendingOp = aPendingOp;
    this->ExecState = aExecState;
    this->Mode = aMode;
    this->tScheduledTime = atScheduledTime;
    this->pExecFunc = apExecFunc;
    this->pInitFunc = apInitFunc;
    this->pDeleteFunc = apDeleteFunc;
    this->iPriority = aiPriority;
    this->tPeriod = atPeriod;
    this->tOffset = atOffset;
    this->ErrorCode = aErrorCode;
    this->iTaskPosition =  aiTaskPosition;
    this->ptScheduler = aptScheduler;

    return OK;

}    /* tTask::Initialize */



/**************************************************************************
*
* tTask::taskSkel() -- task skeleton for user's tasks
*
* IN :
*   this
*   pArgument
*   *this.pInitFunc
*   *this.pExecFunc
*   *this.pDeleteFunc
*   *this.ptScheduler
*   *this.Id
*   *this.TaskSet[].Mode
*   *this.BehaviorOnError
*   SCHED_INIT_PRIORITY
* INOUT :
*   *this.Mode
*   *this.PendingOp
*   *this.ExecState
*   *this.TaskSet[].ExecState
* OUT :
*   *this.ErrorCode
*   *this.iPriority
*   RTE error stream
*
* SPEC:
*
   [ this invalid -> UNDEFINED
   | *this.pInitFunc not valid -> UNDEFINED
   | *this.pExecFunc not valid -> UNDEFINED
   | *this.pDeleteFunc not valid -> UNDEFINED
   | *this.ptScheduler not valid -> UNDEFINED
   | *this.Mode not a valid Task_Mode -> return S_sched_INVALID_TASK_MODE
   | *this.pInitFunc == NULL -> break to the beginning of the loop
   | *this.pExecFunc == NULL -> break to the beginning of the loop
   | *this.pDeleteFunc == NULL -> break to the beginning of the loop
   | *this.Mode == TM_INIT  && *this.pInitFunc != NULL -> 
               return (*this.pInitFunc)(pArgument)
   | *this.Mode == TM_ENABLED  && *this.pEnableFunc != NULL -> 
               return (*this.pEnableFunc)(pArgument)
   | *this.Mode == TM_DELETE  && *this.pDeleteFunc != NULL -> 
               return (*this.pDeleteFunc)(pArgument)
   | true ->  (spec for suspender goes here)
   ]
*
* RETURNS: (normally, doesn't return; stays in FOREVER loop until task killed)
*  S_sched_INVALID_TASK_MODE
*
**************************************************************************/

ERROR_CODE  taskSkel( void *pArgument, tScheduler *pSched, int iTaskPosition )
{
   ERROR_CODE ec, ecSusp;
   FAST tTask *pTheTask;
   uint32_t FpScr;

   /*******
      [ this invalid -> UNDEFINED
      | *this.pInitFunc not valid -> UNDEFINED
      | *this.pExecFunc not valid -> UNDEFINED
      | *this.pDeleteFunc not valid -> UNDEFINED
      | *this.ptScheduler not valid -> UNDEFINED
      | *this.Mode not a valid Task_Mode -> 
               ec := S_sched_INVALID_TASK_MODE
      | *this.Mode == TM_INIT  && *this.pInitFunc != NULL -> 
               ec := (*this.pInitFunc)(pArgument)
      | *this.Mode == TM_ENABLED  && *this.pEnableFunc != NULL -> 
               ec := (*this.pExecFunc)(pArgument, *this.tScheduledTime)
      | *this.Mode == TM_DELETE  && *this.pDeleteFunc != NULL -> 
               ec := (*this.pDeleteFunc)(pArgument)
      | *this.Mode == TM_DISABLED || TM_ERROR || TM_PANIC -> nothing happens
      | true -> return S_sched_INVALID_TASK_MODE  
      ]
   *******/ 

   /* this is to disable exceptions in the case where a floating  */
   /* point number shrinks to zero.  processor will set it to 0.0 */
   //FpScr = vxFpscrGet();
   //vxFpscrSet( FpScr & (~_PPC_FPSCR_UE));  /* disable FP underflow exception */

   /********
   * added - 5/3/1999 PcI  for complete FP exception disable
   ********/
   // disable all floating point exception handler and exception flags for this task
   DisableFpExceptionHandler ();
   DisableFpExceptions ();

   // Set rounding mode to -Infinity - It is theorized that this (or + Infinity) is
   // the fastest mode compared to nearest and zero. (nearest is the default set by
   // vxWorks.)
   SetFpRoundMode (FP_RND_NEG_INF);
   /*********** end FP exception block ***********/


   FOREVER
   {
      ec = OK;   /* just in case all our func ptrs are null */

      // Since the task set may have been reordered, we have to compute
      // the correct entry in the task array to use.
      pTheTask = pSched->GetTaskPtrFromIndex(iTaskPosition);

      TraceMark(S_sched_IN_TASKSKEL);

      switch (pTheTask->Mode)
      {
         case TM_INIT:
            /*******
               [ pTheTask invalid -> UNDEFINED
               | *pTheTask.pInitFunc != NULL -> 
                    ec := (*pTheTask.pInitFunc)(pArgument)
               | true -> nothing happens
               ]
            *******/
            if (pTheTask->pInitFunc != NULL)
                ec = (pTheTask->pInitFunc)(pArgument);
            break;

        case TM_ENABLED:
            /*******
               [ pTheTask invalid -> UNDEFINED
               | *pTheTask.pExecFunc != NULL -> 
                    ec := (*pTheTask.pExecFunc)(pArgument,
                                                *pTheTask.tScheduledTime)
               | true -> nothing happens
               ]
            ******/
            if (pTheTask->pExecFunc != NULL)
                ec = (pTheTask->pExecFunc)(pArgument,pTheTask->tScheduledTime);
            break;
         
         case TM_DELETE:
            /*******
               [ pTheTask invalid -> UNDEFINED
               | *pTheTask.pDeleteFunc != NULL -> 
                    ec := (*pTheTask.pDeleteFunc)(pArgument)
               | true -> nothing happens
               ]
            *******/
            if (pTheTask->pDeleteFunc != NULL)
                ec = (pTheTask->pDeleteFunc)(pArgument);
            break;
            
         case TM_PANIC :   /* The task is finishing its panicked execution */
           ec = OK;
           break;

         case TM_DISABLED :   /* actually we shouldn't see these here */
         case TM_ERROR :
            ec = S_sched_UNEXPECTED_TASK_MODE;
            break;

         default:
            ec = S_sched_INVALID_TASK_MODE;
      }

      if (ec != OK) {
         GetErrorStream().Put( ec );
      }

      /*****

         ****** spec for Suspender() goes here ******

      *****/
      ecSusp = pSched->Suspender(pTheTask, ec);
      if (ecSusp != OK)
         GetErrorStream().Put( ecSusp );

   } /* FOREVER */

}  /* taskSkel */



#ifdef NOT_NOW
/******************************************************************
 *
 * tScheduler::Shutdown - Shutdown the Scheduler
 *
 *
 * IN:
 *   this
 *   *this.bInit
 *   *this.TaskSet[].Id
 *   *this.TaskSet[].ExecState
 *   *this.iMaxTasks
 * INOUT:
 *   *this.TaskSet[].Mode
 *   *this.iNumTasks
 *   *this.hTaskSetPtr
 * OUT:
 *
 * SPEC:
      [ this invalid -> UNDEFINED
      | *this.bInit != TRUE -> return S_sched_SCHEDULER_NOT_INITIALIZED
      | *this.TaskSet is not pointer to array of at least (*this.iNumTasks) 
	  tTasks -> UNDEFINED
      | for each task whose .Mode == TM_ENABLED && .ExecState == TS_DONE -> 
	    *pTcb.Mode := pointer to tTask found in hTaskSetPtr[i]
	    : *pTcb.Mode := TM_DELETE
	    : *this.iNumTasks := *this.iNumTasks - 1
      | for each task whose .Mode != TM_INIT ->
	    *pTcb.Mode := pointer to tTask found in hTaskSetPtr[i]
	    : *this.iNumTasks := *this.iNumTasks - 1
	    : [[U Scheduler::RemoveT()]

      | true -> 
	return OK
      ]
 *
 * RETURNS: (Type ERROR_CODE)
 *    S_sched_SCHEDULER_NOT_INITIALIZED
 *    S_sched_INVALID_TASK_ID
 *    OK
 *
 ******************************************************************/
ERROR_CODE tScheduler::Shutdown(void)
{
   tTask *pTcb;
   int i;
   ERROR_CODE ec;
 
   /*******
      [this invalid -> UNDEFINED
      | *this.bInit != TRUE -> return S_sched_SCHEDULER_NOT_INITIALIZED
      ]
   *******/
   if (this->bInit != TRUE)
       return Traceback (S_sched_SCHEDULER_NOT_INITIALIZED, 
			 "tScheduler::Shutdown");

   /******
       [ this invalid -> UNDEFINED
       | *this.TaskSet is not pointer to an array of at least
	     (*this.iNumTasks) tTasks -> UNDEFINED
       | *this.hTaskSetPtr is not an array of pointers of size 
                (*this.iNumTasks)
       | for each task whose .Mode == TM_ENABLED && .ExecState == TS_DONE -> 
	    *pTcb.Mode := pointer to tTask found in hTaskSetPtr[i]
	    : *pTcb.Mode := TM_DELETE
	    : *this.iNumTasks := *this.iNumTasks - 1
       | for each task whose .Mode != TM_INIT ->
	    *pTcb.Mode := pointer to tTask found in hTaskSetPtr[i]
	    : *this.iNumTasks := *this.iNumTasks - 1
	    : [[U Scheduler::RemoveT()]
	   ]
       true -> do nothing
       ]
    *****/
   while (this->iMaxTasks > 0)
   {
       /*******
	 [ this invalid -> UNDEFINED
	 | *this.TaskSet is not pointer to array of at least
              (*this.iNumTasks) tTasks -> UNDEFINED
         | *this.hTaskSetPtr is not a pointer to an array of 
             (tTask *) of dimension >= *this.iMaxTasks -> UNDEFINED
	 | true -> pTcb := pointer to tTask found in hTaskSetPtr[i]
	 ]
       *******/
       pTcb = this->GetTaskPtrFromIndex(i);
       if (pTcb==NULL)
	  continue;
     
       switch (pTcb->Mode)
       {
	  case TM_INIT:
	     break;
	  case TM_ENABLED:
	     if (pTcp->ExecState == TS_DONE)
	     {
		 pTcb->Mode = TM_DELETE;
		 this->iNumTasks--;
		 ec = taskPrioritySet (pTcb->Id, SCHED_INIT_PRIORITY);
		 if (ec != OK)
                     GetErrorStream().Put( errnoGet() );

		 ec = taskResume(pTcb->Id);
		 if (ec!= OK)
		 {
                   logMsg("taskResume failed in tScheduler::Shutdown, taskId=0x%08x\n",
                          pTcb->Id,0,0,0,0,0);
		      GetErrorStream().Put( errnoGet() );
		 }
	      }
	      break;
	  default:     
             /***** Spec for RemoveTask is defined above *****/ 
	     RemoveTask(pTcb->Id);

       }  /* switch(Mode) */
    }
 

    /* [return OK] */
    return OK;
 
} /* tScheduler::Shutdown */
 
#endif



