/* $RCSfile: lagLib.cc,v $ - lag compensator */
/*
$Header: /home/chara/cvsroot/ople/sys/comm/servo/lagLib.cc,v 1.1.1.1 1998/06/11 21:36:21 chara Exp $
*/
/* Brad Hines - Jet Propulsion Laboratory */

/*
modification history
--------------------
01a,21May97,beh   adapted from Palomar code
*/

/*
DESCRIPTION
This module contains a simple lag compensator with unity
high-frequency gain.  The calculation is performed in
double precision with no overflow checking.  A setup routine
precalculates the filter coefficients for the given
time constants and sample rate.
*/

#include <vxWorks.h>
#include "lagLib.h"
#include "pidLibF.h"

/***********************************************
*
* tLagCompensator::tLagCompensator
*
* DESCRIPTION: 
*
[[U2 tLagCompensator::tLagCompensator_ddd (taup, tauz, tsamp) ]]
*
[[U3 tLagCompensator::tLagCompensator_ddd (taup, tauz, tsamp) ]]
*/
tLagCompensator::tLagCompensator (double taup, double tauz,
                                 double tsamp)
{
  /****
  [[U tLagCompensator::Initialize_ddd (taup, tauz, tsamp) ]]
  ****/
  this->Initialize (taup, tauz, tsamp);
}


/***********************************************
*
* tLagCompensator::tLagCompensator
*
* DESCRIPTION: 
*
[[U2 tLagCompensator::tLagCompensator_dddd (taup, tauz, tsamp, limit) ]]
*
[[U3 tLagCompensator::tLagCompensator_dddd (taup, tauz, tsamp, limit) ]]
*/
tLagCompensator::tLagCompensator (double taup, double tauz,
                                 double tsamp, double limit)
{
  /****
  [[U tLagCompensator::Initialize_dddd (taup, tauz, tsamp, limit) ]]
  ****/
  this->Initialize (taup, tauz, tsamp, limit);
}

/***********************************************
*
* tLagCompensator::Initialize 
*
* DESCRIPTION:
[[U2 tLagCompensator::Initialize_ddd (taup, tauz, tsamp) ]]
*
[[U3 tLagCompensator::Initialize_ddd (taup, tauz, tsamp) ]]
*/
void tLagCompensator::Initialize (double taup, double tauz,
                                  double tsamp)
{
  /****
  [[U tLagCompensator::Initialize_dddd (taup, tauz, tsamp, 32766.99) ]]
  ****/
  this->Initialize (taup, tauz, tsamp, 32766.99);
}

/***********************************************
*
* tLagCompensator::Initialize 
*
* DESCRIPTION:
[[U2 tLagCompensator::Initialize_dddd (taup, tauz, tsamp, lim) ]]
*
[[U3 tLagCompensator::Initialize_dddd (taup, tauz, tsamp, lim) ]]
*/
void tLagCompensator::Initialize (double taup, double tauz,
                                  double tsamp, double lim)
{
  this->xLast = 0;
  this->yLast = 0;
  this->a = (1.0 + tsamp / tauz) / (1.0 + tsamp / taup);
  this->b = 0;
  this->c = (1.0 + tsamp / taup);
  this->lim = lim;
}

/***********************************************
*
* tLagCompensator::TauP
*
* DESCRIPTION: 
[[U2 tLagCompensator::TauP () ]]
*
[[U3 tLagCompensator::TauP () ]]
*/
double tLagCompensator::TauP () const
{
  return this->dSamplePeriod / (this->a * this->c - 1);
}


/***********************************************
*
* tLagCompensator::TauZ
*
* DESCRIPTION: 
[[U2 tLagCompensator::TauZ () ]]
*
[[U3 tLagCompensator::TauZ () ]]
*/
double tLagCompensator::TauZ () const
{
  return this->dSamplePeriod / (this->c - 1);
}

/***********************************************
*
* tLagCompensator::SamplePeriod
*
* DESCRIPTION: 
[[U2 tLagCompensator::SamplePeriod () ]]
*
[[U3 tLagCompensator::SamplePeriod () ]]
*/
double tLagCompensator::SamplePeriod () const
{
  return this->dSamplePeriod;
}



/***********************************************
*
* tLagCompensator::Reset
*
* DESCRIPTION: 
[[U2 tLagCompensator::Reset (x) ]]
*
[[U3 tLagCompensator::Reset (x) ]]
*/
void tLagCompensator::Reset (double x)
{
  this->yLast = 0;
  this->xLast = x;
}

/***********************************************
*
* tLagCompensator::Calculate 
*
* DESCRIPTION:
*
[[U2 tLagCompensator::Calculate (x) ]]
*
[[U3 tLagCompensator::Calculate (x) ]]
*/
double tLagCompensator::Calculate (double x)
{
  FAST double yTmp;
  
  yTmp = this->a * x + (this->yLast - this->xLast) / this->c ;

  /****
  [[U this->yLast := ::limit (yTmp, this->lim) ]]
  ****/
  this->yLast = limit (yTmp, this->lim);

  this->xLast = x;
  
  return yTmp;
}

/***********************************************
*
* tLagCompensator::CopyLagMemory
*
* DESCRIPTION: 
[[U2 ::CopyLagMemory (Destination, Source) ]]
*
[[U3 ::CopyLagMemory (Destination, Source) ]]
*/
void CopyLagMemory (tLagCompensator& Destination, 
         const tLagCompensator& Source)
{
  Destination.xLast = Source.xLast;
  Destination.yLast = Source.yLast;
}