/* $RCSfile: watchdog.cc,v $ - Watchdog timer class */
/* CM Info:
$Header: /home/chara/cvsroot/ople/sys/comm/watchdog/watchdog.cc,v 1.1.1.1 1998/06/11 21:36:22 chara Exp $
*/
/* Brad Hines - Jet Propulsion Laboratory */
 
/*
modification history
--------------------
01a,beh,03Oct96    written
*/
 
/*
DESCRIPTION
This file contains a C++ wrapper for VxWorks watchdog timers.
*/


#include "isit.h"
#include <sysLib.h>
#include <errnoLib.h>
#include <logLib.h>
#include "watchdog.h"

/**************
  State Box State Invariants

  1) [ wdId != NULL ->
        wdId is the id of a valid watchdog created by wdCreate that can
        be cancelled ]
  
 **************/


/******************************************************************
*
* tWatchDog::tWatchDog - create a WatchDog object
*
* IN:
*   tWatchDog   *this
* 
* INOUT:
*
* OUT:
*   *this.fpHandler
*   *this.iHandlerArg
*   *this.iDefaultTimeoutInTicks
*   *this.wdId
*   
*
* SPEC:
*
  [ this invalid -> UNDEFINED
  | true ->
      *this.fpHandler := NULL
      : *this.iHandlerArg := 0
      : *this.iDefaultTimeoutInTicks := 0
      : *this.wdId := NULL
  ] 
*
* SEE ALSO:
*
*
* RETURNS:
*   Nothing.
*/

tWatchDog::tWatchDog(void)
{
  fpHandler = NULL;
  iHandlerArg = 0;
  iDefaultTimeoutInTicks = 0;
  wdId = NULL;
}


/******************************************************************
*
* tWatchDog::~tWatchDog - destroy a tWatchDog object
*
*
*
* IN:
*   tWatchDog   *this
* 
* INOUT:
*   VxWorks.WdTimersAvailable
*
* OUT:
*   *this.wdId
*   
*
* SPEC:
  [ this invalid -> UNDEFINED
  | *this.wdId != NULL ->
      VxWorks.WdTimersAvailable :+= 1
      : *this.wdId := NULL
  | true ->
      *this.wdId := NULL
  ] 
*
* RETURNS:
*   Nothing.
*/

tWatchDog::~tWatchDog(void)
{
  WDOG_ID wdTemp;  // Used to preserve the invariant mentioned at top of file
  
  /**********
     [ *this.wdId != NULL && watchdog timer *this.wdId can be deallocated ->
         VxWorks.WdTimersAvailable :+= 1
         : *this.wdId := NULL
     | true ->
         *this.wdId := NULL
     ]
  **********/
  wdTemp = this->wdId;
  this->wdId = NULL;    // Set to NULL prior to delete to preserve invariant
  if (wdTemp != NULL)
    wdDelete(wdTemp);
}


#ifdef TRACEBACK
#undef TRACEBACK
#endif

/******************************************************************
*
* tWatchDog::Initialize - initialize a WatchDog object
*
*
*
* IN:
*   this
*   fpHandler
*   iHandlerArg
*   iDefaultTimeout
*   *this.wdId
* 
* INOUT:
*   VxWorks.WdTimersAvailable
*
* OUT:
*   *this.fpHandler
*   *this.iHandlerArg
*   *this.iDefaultTimeoutInTicks
*   *this.wdId
*   *this.bActive
*   
*
* SPEC:
    [ this invalid -> UNDEFINED
    | *this.wdId != NULL -> return S_watchDog_ALREADY_INITIALIZED
    | fpHandler == NULL -> return S_watchDog_NULL_FUNCTION_POINTER
    | ! (fpHandler is a pointer to valid C function ||
            (fpHandler is a valid pointer to a C++ method &&
            iHandlerArg is a valid pointer to a C++ object
            with method *fpHandler)) -> UNDEFINED
    | iDefaultTimeout < 1 -> return S_watchDog_INVALID_TIMEOUT_VALUE
    | VxWorks.WdTimersAvailable <= 0 -> return VxWorks error code (wdCreate)
    | true ->
        *this.fpHandler := fpHandler
        : *this.iHandlerArg := iHandlerArg
        : *this.iDefaultTimeoutInTicks := iDefaultTimeout
        : *this.wdId := id of new VxWorks watchdog timer
        : *this.bActive := FALSE
        : VxWorks.WdTimersAvailable :-= 1
        : return OK
    ]
*
* SEE ALSO:
*
*
* RETURNS:
*   An ERROR_CODE
*/

ERROR_CODE tWatchDog::Initialize(FUNCPTR fpHandler, int iHandlerArg,
                                 int iDefaultTimeout)
{
  #define TRACEBACK(ec)    Traceback(ec, "tWatchDog::Initialize", \
                                     fpHandler, iHandlerArg, iDefaultTimeout)
  /************
    [ this invalid -> UNDEFINED
    | *this.wdId != NULL -> return S_watchDog_ALREADY_INITIALIZED
    | fpHandler == NULL -> return S_watchDog_NULL_FUNCTION_POINTER
    | iDefaultTimeout < 1 -> return S_watchDog_INVALID_TIMEOUT_VALUE
    ]
  ************/
  if (this->wdId != NULL)
    return TRACEBACK(S_watchDog_ALREADY_INITIALIZED);

  if (fpHandler == NULL)
    return TRACEBACK(S_watchDog_NULL_FUNCTION_POINTER);

  if (iDefaultTimeout <= 1)
    return TRACEBACK(S_watchDog_INVALID_TIMEOUT_VALUE);

  
  /************
    [ this invalid -> UNDEFINED
    | true ->
        *this.fpHandler := fpHandler
        : *this.iHandlerArg := iHandlerArg
        : *this.iDefaultTimeoutInTicks := iDefaultTimeout
    ]
  ************/
  this->fpHandler = fpHandler;
  this->iHandlerArg = iHandlerArg;
  this->iDefaultTimeoutInTicks = iDefaultTimeout;
  
  /************
    [ this invalid -> UNDEFINED
    | VxWorks.WdTimersAvailable <= 0 -> return VxWorks error code (wdCreate)
    | true ->
        VxWorks.WdTimersAvailable :-= 1
        : *this.wdId := id of new VxWorks watchdog timer
        : *this.bActive := FALSE
    ]
   ************/
  wdId = wdCreate();
  if (wdId == NULL)
    return TRACEBACK(errnoGet());

  
  /************ [ return OK ] ***********/
  return OK;
  
}


#ifdef TRACEBACK
#undef TRACEBACK
#endif

/******************************************************************
*
* tWatchDog::Start(void) - start a watchdog going
*
*
*
* IN:
*   this
*   *this.wdId
*   *this.fpHandler
*   *this.iHandlerArg
*   *this.iDefaultTimeoutTicks
* 
* INOUT:
*
* OUT:
*   *this.bActive
*   *this.iTimeLeft
*   
*
* SPEC:
    [ this invalid -> UNDEFINED
    | *this.wdId == NULL -> return S_watchDog_NOT_INITIALIZED
    | *this.iDefaultTimeoutInTicks < 1 ->
         return S_watchDog_INVALID_TIMEOUT_VALUE
    | *this.wdId not the id of a valid watchdog created by wdCreate ->
                                   UNDEFINED
    | !(*this.fpHandler not a pointer to a valid C function ||
         (*this.fpHandler a pointer to a valid C++ method &&
          *this.iHandlerArg is a pointer to an object of the type for
                             which the method is defined)) -> UNDEFINED
    | this is a pointer to a tTaskLevelWatchDog and *this.iHandlerTaskId
         is not the id of a task executing the code
         in tTaskLevelWatchDog::TaskLevelHandler -> UNDEFINED
    | true ->
        *this.iTimeLeft := iDefaultTimeoutInTicks
        : *this.bActive := TRUE
        : return OK
    ]
*
* SEE ALSO:
*
*
* RETURNS:
*   An ERROR_CODE
*/

ERROR_CODE tWatchDog::Start(void)
{
  #define TRACEBACK(ec)    Traceback(ec, "tWatchDog::Start(void)")

  ERROR_CODE ec;
  
  /************
    [ this invalid -> UNDEFINED
    | *this.wdId == NULL -> return S_watchDog_NOT_INITIALIZED
    | *this.iDefaultTimeoutInTicks < 1 ->
         return S_watchDog_INVALID_TIMEOUT_VALUE
    | *this.wdId not the id of a valid watchdog created by wdCreate ->
                                   UNDEFINED
     | !(*this.Handler not a pointer to a valid C function ||
         (*this.Handler a pointer to a valid C++ method &&
          this is a pointer to an object of the type for
                             which the method is defined)) -> UNDEFINED
    | !(*this.fpHandler not a pointer to a valid C function ||
         (*this.fpHandler a pointer to a valid C++ method &&
          *this.iHandlerArg is a pointer to an object of the type for
                             which the method is defined)) -> UNDEFINED
    | this is a pointer to a tTaskLevelWatchDog and *this.iHandlerTaskId
         is not the id of a task executing the code
         in tTaskLevelWatchDog::TaskLevelHandler -> UNDEFINED
    | true ->
        *this.iTimeLeft := iDefaultTimeoutInTicks
        : *this.bActive := TRUE
        : return OK
    ]
  ************/
  ec = this->StartUsingTicks(this->iDefaultTimeoutInTicks);
  if (ec != OK)
    return TRACEBACK(ec);

  return OK;
}


#ifdef TRACEBACK
#undef TRACEBACK
#endif


/******************************************************************
*
* tWatchDog::StartUsingTIME(TIME) - start a watchdog going, in TIME units
*
*
*
* IN:
*   this
*   *this.wdId
*   *this.fpHandler
*   *this.iHandlerArg
*   tTimeout
* 
* INOUT:
*
* OUT:
*   *this.bActive
*   *this.iTimeLeft
*   
*   
*
* SPEC:
    [ this invalid -> UNDEFINED
    | *this.wdId == NULL -> return S_watchDog_NOT_INITIALIZED
    | TIME_TO_VXWORKS_TICKS(tTimeout) < 1 ->
         return S_watchDog_INVALID_TIMEOUT_VALUE
    | *this.wdId not the id of a valid watchdog created by wdCreate ->
                                   UNDEFINED
    | !(*this.fpHandler not a pointer to a valid C function ||
         (*this.fpHandler a pointer to a valid C++ method &&
          *this.iHandlerArg is a pointer to an object of the type for
                             which the method is defined)) -> UNDEFINED
    | this is a pointer to a tTaskLevelWatchDog and *this.iHandlerTaskId
         is not the id of a task executing the code
         in tTaskLevelWatchDog::TaskLevelHandler -> UNDEFINED
    | true ->
        *this.iTimeLeft := TIME_TO_VXWORKS_TICKS(tTimeout)
        : *this.bActive := TRUE
        : return OK
    ]
*
* SEE ALSO:
*
*
* RETURNS:
*   An ERROR_CODE
*/

ERROR_CODE tWatchDog::StartUsingTIME(TIME tTimeout)
{
  #define TRACEBACK(ec)  Traceback(ec, "tWatchDog::Start(TIME)", tTimeout)
  int iTimeout;
  ERROR_CODE ec;
  
  /************
     [ this invalid -> UNDEFINED
     | true ->
        iTimeout := TIME_TO_VXWORKS_TICKS(tTimeout)
     ]
   ************/
  iTimeout = TIME_TO_VXWORKS_TICKS(tTimeout);
      
  /************
    [ this invalid -> UNDEFINED
    | *this.wdId == NULL -> return S_watchDog_NOT_INITIALIZED
    | iTimeout < 1 ->
         return S_watchDog_INVALID_TIMEOUT_VALUE
    | *this.wdId not the id of a valid watchdog created by wdCreate ->
                                   UNDEFINED
     | !(*this.Handler not a pointer to a valid C function ||
         (*this.Handler a pointer to a valid C++ method &&
          this is a pointer to an object of the type for
                             which the method is defined)) -> UNDEFINED
    | !(*this.fpHandler not a pointer to a valid C function ||
         (*this.fpHandler a pointer to a valid C++ method &&
          *this.iHandlerArg is a pointer to an object of the type for
                             which the method is defined)) -> UNDEFINED
    | this is a pointer to a tTaskLevelWatchDog and *this.iHandlerTaskId
         is not the id of a task executing the code
         in tTaskLevelWatchDog::TaskLevelHandler -> UNDEFINED
    | true ->
        *this.iTimeLeft := iTimeout
        : *this.bActive := TRUE
        : return OK
    ]
  ************/
  ec = this->StartUsingTicks(iTimeout);
  if (ec != OK)
    return TRACEBACK(ec);

  return OK;
}


#ifdef TRACEBACK
#undef TRACEBACK
#endif


/******************************************************************
*
* tWatchDog::StartUsingTicks(int) - start a watchdog going, timeout in ticks
*
*
*
* IN:
*   this
*   *this.wdId
*   *this.fpHandler
*   *this.iHandlerArg
*   iTimeoutInTicks
* 
* INOUT:
*
* OUT:
*   *this.bActive
*   *this.iTimeLeft
*   
*   
*
* SPEC:
    [ this invalid -> UNDEFINED
    | *this.wdId == NULL -> return S_watchDog_NOT_INITIALIZED
    | iTimeoutInTicks < 1 ->
         return S_watchDog_INVALID_TIMEOUT_VALUE
    | *this.wdId not the id of a valid watchdog created by wdCreate ->
                                   UNDEFINED
    | !(*this.fpHandler not a pointer to a valid C function ||
         (*this.fpHandler a pointer to a valid C++ method &&
          *this.iHandlerArg is a pointer to an object of the type for
                             which the method is defined)) -> UNDEFINED
    | this is a pointer to a tTaskLevelWatchDog and *this.iHandlerTaskId
         is not the id of a task executing the code
         in tTaskLevelWatchDog::TaskLevelHandler -> UNDEFINED
    | true ->
        *this.iTimeLeft := iTimeoutInTicks
        : *this.bActive := TRUE
        : return OK
    ]
*
* SEE ALSO:
*
*
* RETURNS:
*   An ERROR_CODE
*/

ERROR_CODE tWatchDog::StartUsingTicks(int iTimeoutInTicks)
{
  #define TRACEBACK(ec)  Traceback(ec, "tWatchDog::Start(int)",iTimeoutInTicks)
  
  /************
     [ this invalid -> UNDEFINED
     | *this.wdId == NULL -> return S_watchDog_NOT_INITIALIZED
     | iTimeoutInTicks < 1 ->
         return S_watchDog_INVALID_TIMEOUT_VALUE
     | true -> NOP
     ]
   ************/
  if (this->wdId == NULL)
    return TRACEBACK(S_watchDog_NOT_INITIALIZED);
  if (iTimeoutInTicks < 1)
    return S_watchDog_INVALID_TIMEOUT_VALUE;

  // NOTE! The following function eventually may result in a call
  // to tWatchDog::Handler or tTaskLevelWatchDog::Handler, 
  // so the UNDEFINED conditions from those functions are copied here.
  /************ 
     [ this invalid -> UNDEFINED
     | *this.wdId is not a valid watchdog -> UNDEFINED
     | *this.iTimeoutInTicks < 1 -> UNDEFINED
     | !(*this.Handler not a pointer to a valid C function ||
         (*this.Handler a pointer to a valid C++ method &&
          this is a pointer to an object of the type for
                             which the method is defined)) -> UNDEFINED
     | !(*this.fpHandler not a pointer to a valid C function ||
         (*this.fpHandler a pointer to a valid C++ method &&
          *this.iHandlerArg is a pointer to an object of the type for
                             which the method is defined)) -> UNDEFINED
     | this is a pointer to a tTaskLevelWatchDog and *this.iHandlerTaskId
         is not the id of a task executing the code
         in tTaskLevelWatchDog::TaskLevelHandler -> UNDEFINED
     | true ->
         *this.iTimeLeft := iTimeoutInTicks
         : *this.bActive := TRUE
     ]
   ************/
  if (wdStart(this->wdId, iTimeoutInTicks, (FUNCPTR) &(this->WdTimeoutHandler),
              (int) this) != OK)
    return TRACEBACK(errnoGet());

  //********** [ return OK ]
  return OK;
}


#ifdef TRACEBACK
#undef TRACEBACK
#endif

/******************************************************************
*
* tWatchDog::Reset(void) - Reset a watchdog to inactive
*
*
*
* IN:
*   this
*   *this.wdId
* 
* INOUT:
*
* OUT:
*   *this.bActive
*   
*
* SPEC:
    [ this invalid -> UNDEFINED
    | *this.wdId == NULL -> return S_watchDog_NOT_INITIALIZED
    | *this.wdId not the id of a valid watchdog created by wdCreate ->
                                   UNDEFINED
    | *this.wdId cannot be cancelled -> return VxWorks error (wdCancel)
    | true ->
        *this.bActive := TRUE
        : return OK
    ]
*
* SEE ALSO:
*
*
* RETURNS:
*   An ERROR_CODE
*/

ERROR_CODE tWatchDog::Reset(void)
{
  #define TRACEBACK(ec)    Traceback(ec, "tWatchDog::Reset(void)")
  
  /************
     [ this invalid -> UNDEFINED
     | *this.wdId == NULL -> return S_watchDog_NOT_INITIALIZED
     ]
   ************/
  if (this->wdId == NULL)
    return TRACEBACK(S_watchDog_NOT_INITIALIZED);
      
  /************
     [ this invalid -> UNDEFINED
     | *this.wdId is not a valid watchdog created by wdCreate -> UNDEFINED
     | *this.wdId cannot be cancelled -> return VxWorks error (wdCancel)
     | true ->
         *this.bActive := FALSE
     ]
   ************/
  if (wdCancel(this->wdId) == ERROR)
    return TRACEBACK(errnoGet());

  //********** [ return OK ]
  return OK;
}



/******************************************************************
*
* tWatchDog::WdTimeoutHandler(void) - Handle a watchdog timeout
*
* This function simply turns around and calls the virtual
* handler function for the class.  Why not just have the
* watchdog timeout call this->Handler() directly?
*
* For some reason, the equivalence of C calls to C++ calls
* (i.e., the fact that the C++ this->Method() is equivalent
* to the C call Method(this)) breaks down when the C++ function
* is virtual.  Therefore we must have a non-virtual handler
* to handle the VxWorks watchdog timeout which can then turn
* around and call the actual handler function.
*
* IN:
*   this
* 
* INOUT:
*
* OUT:
*
*   
*
* SPEC:
    [ this invalid -> UNDEFINED
    | true->
       Refer to the spec for the Handler method of the object represented
           by "this".
    ]
*
*
* RETURNS:
*   Nothing
*/

void tWatchDog::WdTimeoutHandler(void)
{
  logMsg("Watchdog Timeout!! this = 0x%x\n", (int) this, 0,0,0,0,0);
  this->Handler();
}


/******************************************************************
*
* tWatchDog::Handler(void) - Handle a watchdog timeout
*
* This function simply turns around and calls the handler
* function specified by the user.  Why not just have the
* watchdog timeout call the user's function directly?
*
* It turns out that doing is this way makes it easier to
* overload the behavior of the watchdog in the derived
* class tTaskLevelWatchDog, which replaces this handler
* function with its own.
*
* IN:
*   this
*
* 
* INOUT:
*
* OUT:
*
*   
*
* SPEC:
    [ this invalid -> UNDEFINED
    | !(*this.fpHandler not a pointer to a valid C function ||
         (*this.fpHandler a pointer to a valid C++ method &&
          *this.iHandlerArg is a pointer to an object of the type for
                             which the method is defined)) -> UNDEFINED
    | true->
       Refer to the spec for the function/method for the fpHandler
         that was specified during the Initialize call.
    ]
*
*
* RETURNS:
*   Nothing
*/

void tWatchDog::Handler(void)
{
  (*(this->fpHandler)) (this->iHandlerArg);
}


#ifdef TRACEBACK
#undef TRACEBACK
#endif

/******************************************************************
*
* tTaskLevelWatchDog::Initialize - initialize a TaskLevelWatchDog object
*
*
*
* IN:
*   this
*   fpHandler
*   iHandlerArg
*   iDefaultTimeout
*   *this.wdId
* 
* INOUT:
*   VxWorks.WdTimersAvailable
*   OS.SpawnedTasks
*   Available memory
*
* OUT:
*   *this.fpHandler
*   *this.iHandlerArg
*   *this.iDefaultTimeoutInTicks
*   *this.wdId
*   *this.bActive
*   *this.iHandlerTaskId
*   
*   
*
* SPEC:
*
    [ this invalid -> UNDEFINED
    | *this.wdId != NULL -> return S_watchDog_ALREADY_INITIALIZED
    | fpHandler == NULL -> return S_watchDog_NULL_FUNCTION_POINTER
    | ! (fpHandler is a pointer to valid C function ||
            (fpHandler is a valid pointer to a C++ method &&
            iHandlerArg is a valid pointer to a C++ object
            with method *fpHandler)) -> UNDEFINED
    | iDefaultTimeout < 1 -> return S_watchDog_INVALID_TIMEOUT_VALUE
    | VxWorks.WdTimersAvailable <= 0 -> return VxWorks error code (wdCreate)
    | Task cannot be created || Insufficient memory to spawn task ->
           *this.iHandlerTaskId := ERROR
           : return VxWorks errno (taskSpawn)
    | true ->
        *this.fpHandler := fpHandler
        : *this.iHandlerArg := iHandlerArg
        : *this.iDefaultTimeoutInTicks := iDefaultTimeout
        : *this.wdId := id of new VxWorks watchdog timer
        : *this.bActive := FALSE
        : VxWorks.WdTimersAvailable :-= 1
        : OS.SpawnedTasks := OS.SpawnedTasks U
                    { *this.TaskLevelHandler(), with
                      priority iHandlerTaskPriority,
                      options iHandlerTaskOptions,
                      stack size iHandlerTaskStackSize,
                      and name "wdHandlerTask" }
        : *this.iHandlerTaskId := new task ID from VxWorks
        : return OK
    ]
*
* RETURNS:
*   An ERROR_CODE
*/

ERROR_CODE tTaskLevelWatchDog::Initialize(FUNCPTR fpHandler, int iHandlerArg,
                                 int iDefaultTimeout,
                                 int iHandlerTaskPriority,
                                 int iHandlerTaskOptions,
                                 int iHandlerTaskStackSize)
{
#define TRACEBACK(ec)    Traceback(ec, "tTaskLevelWatchDog::Initialize", \
                                fpHandler, iHandlerArg, iDefaultTimeout, \
                                iHandlerTaskPriority, iHandlerTaskOptions, \
                                iHandlerTaskStackSize)
  ERROR_CODE ec;

  /******
    [ this invalid -> UNDEFINED
    | *this.wdId != NULL -> return S_watchDog_ALREADY_INITIALIZED
    | fpHandler == NULL -> return S_watchDog_NULL_FUNCTION_POINTER
    | ! (fpHandler is a pointer to valid C function ||
            (fpHandler is a valid pointer to a C++ method &&
            iHandlerArg is a valid pointer to a C++ object
            with method *fpHandler)) -> UNDEFINED
    | iDefaultTimeout < 1 -> return S_watchDog_INVALID_TIMEOUT_VALUE
    | VxWorks.WdTimersAvailable <= 0 -> return VxWorks error code (wdCreate)
    | true ->
        *this.fpHandler := fpHandler
        : *this.iHandlerArg := iHandlerArg
        : *this.iDefaultTimeoutInTicks := iDefaultTimeout
        : *this.wdId := id of new VxWorks watchdog timer
        : *this.bActive := FALSE
        : VxWorks.WdTimersAvailable :-= 1
        : return OK
    ]
  ******/
  ec = this->tWatchDog::Initialize(fpHandler, iHandlerArg, iDefaultTimeout);
  if (ec != OK)
    return TRACEBACK(ec);


  /*****
    [ ! (&*this.TaskLevelHandler is a pointer to valid C function ||
          (&*this.TaskLevelHandler is a valid pointer to a C++ method &&
           this is a valid pointer to a C++ object with method
           TaskLevelHandler)) -> UNDEFINED
    | Task cannot be created || Insufficient memory to spawn task ->
           OS.Task.errno := VxWorks errno (taskSpawn)
           : *this.iHandlerTaskId := ERROR
    | true ->
        OS.SpawnedTasks := OS.SpawnedTasks U
                    { *this.TaskLevelHandler(), with
                      priority iHandlerTaskPriority,
                      options iHandlerTaskOptions,
                      stack size iHandlerTaskStackSize,
                      and name "wdHandlerTask" }
        : *this.iHandlerTaskId := new task ID from VxWorks
    ]
  *****/
  this->iHandlerTaskId = taskSpawn("wdHandlerTask",
                                   iHandlerTaskPriority,
                                   iHandlerTaskOptions,
                                   iHandlerTaskStackSize,
                                   (FUNCPTR) &(this->TaskLevelHandler),
                                   (int) this,
                                   0,0,0,0,0,0,0,0,0);

  /*****
    [ *this.iHandlerTaskId == ERROR -> return OS.Task.errno
    | true -> return OK
    ]
  *****/
  if (this->iHandlerTaskId == ERROR) {
    return TRACEBACK(errnoGet());
  }

  return OK;
  
}


/******************************************************************
* 
* tTaskLevelWatchDog::Handler - respond to a watchdog timeout
*
* This function simply does a taskResume on the task-level watchdog
* handler, which then takes care of dispatching to the user-defined
* watchdog handler from task level.
*
* IN:
*   this
*   *this.iHandlerTaskId
*   *this.fpHandler
*   *this.iHandlerArg
*
* INOUT:
* OUT:
*
* OTHER: See spec for method specified by fpHandler
*   
*
* SPEC: (in essence, the net effect is the same as a call to the portion
*        of TaskLevelHandler inside the FOREVER loop)
*
    [ this invalid -> UNDEFINED
    | *this.iHandlerTaskId is not the id of a task executing the code
         in tTaskLevelWatchDog::TaskLevelHandler
    | !(*this.fpHandler not a pointer to a valid C function ||
         (*this.fpHandler a pointer to a valid C++ method &&
          *this.iHandlerArg is a pointer to an object of the type for
                             which the method is defined)) -> UNDEFINED
    | true->
       Refer to the spec for the function/method for the fpHandler
         that was specified during the Initialize call.
    ]
*
* RETURNS:
*   Nothing
*/

void tTaskLevelWatchDog::Handler(void)
{
  taskResume(this->iHandlerTaskId);
}




/******************************************************************
*
* tTaskLevelWatchDog::TaskLevelHandler - handle a watchdog timeout
*
* This function is a task which waits for a watchdog timeout.
* When a timeout occurs, tTaskLevelWatchDog::Handler will be
* called from interrupt level and will issue a taskResume to
* this task, which will be pended on the taskSuspend call.
*
* At this point, this function simply responds to the stimulus by
* calling tWatchDog::Handler.  This gives the same net effect as
* is exhibited by the tWatchDog class, but with the difference that
* tWatchDog::Handler is called from task level instead of interrupt
* level.
*
* IN:
*   this
*   *this.fpHandler
*   *this.iHandlerArg
* 
* INOUT:
* OUT:
*   
* OTHER: See spec for method specified by fpHandler
*
*
* SPEC: (in response to a taskResume - just the spec for tWatchDog::Handler)
*
    [ this invalid -> UNDEFINED
    | !(*this.fpHandler not a pointer to a valid C function ||
         (*this.fpHandler a pointer to a valid C++ method &&
          *this.iHandlerArg is a pointer to an object of the type for
                             which the method is defined)) -> UNDEFINED
    | true->
       Refer to the spec for the function/method for the fpHandler
         that was specified during the Initialize call.
    ]
*
* RETURNS:
*   Nothing
*/

void tTaskLevelWatchDog::TaskLevelHandler(void)
{
  FOREVER {
    taskSuspend(0);
    tWatchDog::Handler();
  }
}


/******************************************************************
*
* tTaskLevelWatchDog::~tTaskLevelWatchDog - destructor
*
* Deletes the spawned task and calls the base class destructor
*
* IN:
*   this
* 
* INOUT:
*
* OUT:
*   
*
* SPEC:
    [ this invalid -> UNDEFINED
    | true ->
       [ *this iHandlerTaskId != ERROR ->
           OS.SpawnedTasks := OS.SpawnedTasks with the task corresponding
                  to *this.iHandlerTaskId removed
           : *this.iHandlerTaskId := ERROR
       ]
       : [ *this.wdId != NULL ->
           VxWorks.WdTimersAvailable :+= 1
           : *this.wdId := NULL
         ]
    ]
*
* SEE ALSO:
*
*
* RETURNS:
*   An ERROR_CODE
*/

tTaskLevelWatchDog::~tTaskLevelWatchDog(void)
{
  /*****
    [ this invalid -> UNDEFINED
    | *this.iHandlerTaskId != ERROR ->
        OS.SpawnedTasks := OS.SpawnedTasks with the task corresponding
                  to *this.iHandlerTaskId removed
    | true ->
        *this.iHandlerTaskId := ERROR
  *****/
  if (this->iHandlerTaskId != ERROR)
    taskDelete(this->iHandlerTaskId);
  this->iHandlerTaskId = ERROR;

  // Spec for tWatchDog::~tWatchDog, called implicitly
  /*****
    [ this invalid -> UNDEFINED
    | *this.wdId != NULL ->
        VxWorks.WdTimersAvailable :+= 1
        : *this.wdId := NULL
    | true ->
        *this.wdId := NULL
    ]
  *****/
}





#ifdef TRACEBACK
#undef TRACEBACK
#endif
