/* $RCSfile: dlController.cc,v $ */
/* CM Info:
$Header: /home/chara/cvsroot/ople/sys/gizmo/dl/dlController.cc,v 1.1.1.1 1998/06/11 21:36:22 chara Exp $
*/
/* Brad Hines - Jet Propulsion Laboratory */
 
/*
modification history
--------------------
02a,beh,11May97    written
*/
 
/*
DESCRIPTION

*/

#include <math.h>
#include "controllerSet.h"
#include "inputSource.h"
#include "outputDest.h"
#include "targetSource.h"
#include "dl/dlIoSet.h"
#include "dl/dlController.h"

static inline double sqr(double d) { return d*d; }



/*******************************************************************/
/***********************  tDlControllerData ************************/
/*******************************************************************/

/******************************************************************
*
* tDlControllerData::tControllerData
*
* DESCRIPTION: Sets up important DlController quantities
*
* RETURNS: Nothing
*
[[U tDlControllerData::tControllerData (LimitsToUse,
         MotorToUse,Vc1,PztReadback, Laser,
         dDecelerationAggressiveness,iMotorServoRate,
         iDlDataDumpPriority, iDlDataBufferSize, iDlJuitterHpBlockSize
         DlDataTag, DlStatusToUse)]]
*
*/

tDlControllerData::tDlControllerData (tDlLimits &LimitsToUse,
           tModulatedVelocityActuator &MotorToUse,
           tModulatedPositionActuator &Vc1,
           tModulatedSensor &PztReadback, tLaserMetrologySensor &Laser,
           double dDecelerationAggressiveness, int iMotorServoRate,
           int iDlDataDumpPriority, int iDlDataBufferSize, int iDlJitterHpBlockSize,
           const tTypeTag& DlDataTag, tDlStatus& DlStatusToUse) :
  Motor (MotorToUse),
  Limits (LimitsToUse),
  DlData (iDlDataDumpPriority, iDlDataBufferSize, iDlJitterHpBlockSize, DlDataTag, DlStatusToUse)

{
  TraceMark(S_dlController_CONTROLLER_DATA_INITIALIZE_START);

  
  // Get max velocity and acceleration in physical units
  
  /****
  [[U this->CdMotorVelLimitInMotorUnits := 
      Motor.tModulatedVelocityActuator::MaxVelocity (void)]]
  ****/
  this->CdMotorVelLimitInMotorUnits =     // motor counts/sec
    Motor.MaxVelocityInChannelUnits();
    
  /****
  [[U this->Motor.tModulatedVelocityActuator::MaxAcceleration (void) ]]
  [[U this->Motor.tModulatedVelocityActuator::PhysicalUnitsPerCount (void) ]]
  ****/
  this->CdMotorAccelLimitInMotorUnits =   // (motor counts/sec) / sec
    this->Motor.MaxAcceleration() /          // (m / s) / sec
    this->Motor.PhysicalUnitsPerCount();        // (m / s) / motor count

  // Compute laser to PZT scale factor
  // (converts from laser counts to PZT counts)
  /****
    [[U Laser.tSensor::OpdUnitsPerCount (void)]]
    [[U PztReadback.tSensor::PhysicalUnitsPerCount (void)]]
  ****/
  this->CdLaserToPztUnits =        // pzt counts / laser count
    Laser.OpdUnitsPerCount() /               // m OPD / laser count
    PztReadback.PhysicalUnitsPerCount();     // m physical / pzt count
  
  // Compute laser to motor scale factor
  // (converts from laser counts/sec to motor counts/sec)
  /****
    [[U Laser.tSensor::OpdUnitsPerCount (void) ]]
    [[U this->Motor.tModulatedVelocityActuator::PhysicalUnitsPerCount(void) ]]
  ****/
  this->CdLaserVelocityToMotorUnits =     // motor counts / (laser count/sec)
    Laser.OpdUnitsPerCount() /            // m OPD / laser count
    this->Motor.PhysicalUnitsPerCount();        // (m/s) / motor count
  
  this->CdMotorVelLimitInLaserUnits =
    this->CdMotorVelLimitInMotorUnits / this->CdLaserVelocityToMotorUnits;
  this->CdMotorAccelLimitInLaserUnits =
    this->CdMotorAccelLimitInMotorUnits / this->CdLaserVelocityToMotorUnits;

  // No VC position limits for nominal delay line
  this->CdVcLowerLimitInVcUnits = Vc1.MinPositionInChannelUnits();
  this->CdVcUpperLimitInVcUnits = Vc1.MaxPositionInChannelUnits();

  // between 0 and 1, 1 most aggressive
  this->CdDecelerationAggressiveness = dDecelerationAggressiveness;
  this->CdMotorServoRate = iMotorServoRate;

  this->CdLaserToOpd = Laser.OpdUnitsPerCount();
  
  this->dLaserError = 0.0;
  this->dMotorVelocityErrorInLaserUnits = 0.0;
 
  TraceMark(S_dlController_CONTROLLER_DATA_INITIALIZE_EXIT);
}


/******************************************************************
*
* tDlControllerData::UpdateCompensatorCoefficients
*
* DESCRIPTION: Updates control law coefficients based on servo
* parameters
*
* RETURNS: Nothing
*
  [[U tDlControllerData::UpdateCompensatorCoefficients()]]
*
*/

ERROR_CODE tDlControllerData::UpdateCompensatorCoefficients()
{
  TraceMark(S_dlController_UPDATE_COMPENSATOR_COEFFS_START);

  const tDlServoParams &Params = this->GetServoParams();
  
  /****
  [[U this->PztLag.tLagCompensator::Initialize (Params.dPztLagPole,
                           Params.dPztLagZero,
                           1.0/Params.dPztServoRate) ]]
  ****/                        
  this->PztLag.Initialize (Params.dPztLagPole,
                           Params.dPztLagZero,
                           1.0/Params.dPztServoRate);
  
  /****
  [[U this->Vc1TrackPid.tPidDouble::InitializeUsingPolesAndZeros (
                         Params.dVc1TrackPidGain,
                         Params.dVc1TrackPidZero1,
                         Params.dVc1TrackPidZero2,
                         Params.dVc1TrackPidPole,
                         1.0/Params.dVc1TrackServoRate)  ]]
  ****/                        
  this->Vc1TrackPid.InitializeUsingPolesAndZeros(Params.dVc1TrackPidGain,
                         Params.dVc1TrackPidZero1,
                         Params.dVc1TrackPidZero2,
                         Params.dVc1TrackPidPole,
                         1.0/Params.dVc1TrackServoRate);

  /****
  [[U  this->Vc1TrackPztFeedforward.tFeedforward::Initialize(
         Params.dVc1LaserFeedforwardGain * this->CdLaserToPztUnits) ]]
  ****/                        
  this->Vc1TrackPztFeedforward.Initialize(
                   Params.dVc1LaserFeedforwardGain * this->CdLaserToPztUnits);

  /****
  [[U  this->Vc1TrackPztFeedforwardFilter.tFilter::Initialize(
         Params.dVc1LaserFeedforwardGain * this->CdLaserToPztUnits) ]]
  ****/                        
  this->Vc1TrackPztFeedforwardFilter.Initialize(
                   Params.dVc1LaserFeedforwardFilterMemory);
  
  /****
  [[U  this->Vc1SlewPid.tPidDouble::InitializeUsingPolesAndZeros (
                         Params.dVc1SlewPidGain,
                         Params.dVc1SlewPidZero1,
                         Params.dVc1SlewPidZero2,
                         Params.dVc1SlewPidPole,
                         1.0/Params.dVc1SlewServoRate) ]]
  ****/                        
  this->Vc1SlewPid.InitializeUsingPolesAndZeros(Params.dVc1SlewPidGain,
                         Params.dVc1SlewPidZero1,
                         Params.dVc1SlewPidZero2,
                         Params.dVc1SlewPidPole,
                         1.0/Params.dVc1SlewServoRate);

  /****
  [[U  this->MotorPid.tPidDouble::InitializeUsingPolesAndZeros (
          Params.dMotorGain, 0.0, 0.0, 0.0, 1.0/Params.dMotorServoRate) ]]
  ****/                        
  this->MotorPid.InitializeUsingPolesAndZeros(Params.dMotorGain,
                      0.0,
                      0.0,
                      0.0,
                      1.0/Params.dMotorServoRate);

  // Motor feedforward is in laser units, so include conversion factor


  /****
  [[U  this->MotorVelocityFeedforward.tFeedforward::Initialize (
         Params.dMotorFeedForwardGain * this->CdLaserVelocityToMotorUnits) ]]
  ****/                        
  this->MotorVelocityFeedforward.Initialize(
                   Params.dMotorFeedForwardGain *
                   this->CdLaserVelocityToMotorUnits);

  this->bPztServoEnabled       = (BOOL) Params.bPztServoEnabled;
  this->bPztLagEnabled         = (BOOL) Params.bLagEnabled;
  this->bVc1TrackServoEnabled  = (BOOL) Params.bVc1TrackServoEnabled;
  this->bVc1SlewServoEnabled   = (BOOL) Params.bVc1SlewServoEnabled;
  this->bMotorServoEnabled     = (BOOL) Params.bMotorServoEnabled;

  this->dPztTrackGain          = Params.dPztGain;

  // Slew/Track transition thresholds, in laser units
  this->dTrackToSlewErrorThreshold =
    Params.dTrackToSlewThresholdInMm / 1000.0 / this->CdLaserToOpd;
  this->dSlewToTrackErrorThreshold =
    Params.dSlewToTrackThresholdInMm / 1000.0 / this->CdLaserToOpd;

  this->dSlewToTrackMotorVelocityErrorThreshold = 
    Params.dSlewToTrackThresholdInMm / 1000.0 / this->CdLaserToOpd;
    
  this->dTrackToSlewMotorVelocityErrorThreshold = 
    Params.dTrackToSlewThresholdInMm / 1000.0 / this->CdLaserToOpd;

  TraceMark(S_dlController_UPDATE_COMPENSATOR_COEFFS_EXIT);

  return OK;
}


/*******************************************************************/
/*************************  tDlController **********************/
/*******************************************************************/

/******************************************************************
*
* tDlController::tDlController
*
* DESCRIPTION: Constructor
*
* RETURNS: Nothing
*
[[U tDlController::tDlController (ControllerDataToUse) ]]
*
*/

tDlController::tDlController (tDlControllerData &ControllerDataToUse) :

  /****
  [ this->ControllerData := ControllerDataToUse ]
  ****/
  ControllerData(ControllerDataToUse)
{
  TraceMark(S_dlController_DL_CONTROLLER_CONSTRUCTOR);
}





/*******************************************************************
*
* tDlCompleteControllerSet::InstallNewServoParams
*
* DESCRIPTION: Sets servo parameters and updates compensator coefficients
*
* RETURNS: ERROR_CODE
*
  [[U tDlCompleteControllerSet::InstallNewServoParams (ServoParams)]]
*
*/

ERROR_CODE tDlCompleteControllerSet::InstallNewServoParams (
       const TRACK_SERVO_PARAMS &ServoParams) 
{
  ERROR_CODE ec;
  
  ec = this->SetServoParams (ServoParams);
  if (ec != OK)
    return Traceback(ec, "tDlCompleteSetOfControllerSets::DoSetServoParams",
                     ServoParams);
  
  return GetControllerData().UpdateCompensatorCoefficients();
}
















































































