/* $RCSfile: dlDither.cc,v $ - Delay Line Dither Controller */
/* CM Info:
$Header: /home/chara/cvsroot/ople/sys/gizmo/dl/dlDither.cc,v 1.1.1.1 1998/06/11 21:36:22 chara Exp $
*/
/* Beth McKenney - Jet Propulsion Laboratory */
 
/*
modification history
--------------------
01a,eamck,22Apr97    written 
*/
 
/*
DESCRIPTION

*/

#include "isit.h"
#include "errorStream.h"
#include "dlTypes.h"
#include "dlDither.h"
#include "hwTypes.h"

#include "vxworks/semLib.spec"
#include <semLib.h>
#include <logLib.h>
#include <errnoLib.h>
#include <math.h>

#include "configMgr.h"
#include "dlGizmo.h"

/******************************************************************
* Dither metrics criteria
*
* CalibrateDither will not complete until the phase, amplitude, and
* distortion errors are all below these thresholds.
*/

#define MAX_PHASE_ERROR (.01)     /* Max phase error in fractions of a wave   */
#define MAX_AMPLITUDE_ERROR (.02) /* Max amplitude error in frac of full scl */
#define MAX_DISTORTION (0.10)     /* when is dither distortion "good enough"? */
#define MAX_CALIB_CYCLES (5)      /* Max number of times to adjust */



/******************************************************************
*              Global declaration of static members
****/
TIME       tDlDither::stDitherPeriod = 0;
double     tDlDither::sdDitherSamplePeriod = 0.0;
int        tDlDither::siDitherBufSize = 0;
int        tDlDither::siDoubleDitherBufSize = 0;
int        tDlDither::siMaxDitherBufSize = 0;
int        tDlDither::siDoubleMaxDitherBufSize = 0;
int        tDlDither::siMaxWaveOffset = 0;  
int        tDlDither::siSweepBufSize = 0;    





tDlDither::tDlDither (tWaveformChannel& DitherPztChannel, 
               double dDitherChannelUnitsIn, 
               TIME tDlFastServoPeriodIn, TIME tDitherPeriodIn, 
               double dDitherAmplitudeIn) :

  bDoingCheckOrCalib (FALSE),               
  pDitherBank (DitherPztChannel.GetChannelBank ()),
  pDitherChannel (&DitherPztChannel),
  dDitherChannelUnits (dDitherChannelUnitsIn), 
  tDlServoTaskPeriod (tDlFastServoPeriodIn),
  bDithering (FALSE),
  
  bCalibrating (FALSE),
  bWasDitherOn (FALSE),
  bStoreDither (FALSE),
  iWaveOffset (0),
  iWavesToAvg (NUM_WAVES_TO_AVG),
  iSweepSampleCount (0),
  iSweepErrorCount (0),
  iTotalSweepCount (0),

  dCurrentAmplFactor (1.0),
  dProposedAmplFactor (1.0),
  iCurrentPhaseShiftIndex (0),
  iProposedPhaseShiftIndex (0),
  CalibInfo (0.0, 0.0, 0.0)

{
   int i, ec;
   double dSampOutFreq;

   cout << "Dither Channel Unit " << dDitherChannelUnits << endl;

   if (this->pDitherBank==NULL) {
     TraceMsg (tmt_ERROR, S_dlDither_NULL_BANK_POINTER,"tDlDither::tDlDither");
     return;
   }

   /***
   [[U ec := this->tDlDither::DitherOff (void) ]]
   ***/
   ec = this->DitherOff();   
   if (ec != OK) {
     TraceMsg (tmt_ERROR, ec,"tDlDither::Initialize");
     return;
   }
   
   /***
     [[U dSampOutFreq := 
          this->pDitherBank->tWaveformChannelBank::GetSampleFreq()]]
    ***/
   dSampOutFreq = this->pDitherBank->GetSampleFreq();
   
   if (dSampOutFreq <= 0.0) {
     TraceMsg (tmt_ERROR, S_dlDither_INVALID_SAMPLE_FREQ,"tDlDither::tDlDither");
     return;
   }			       

   this->sdDitherSamplePeriod = TIME_TICKS_PER_SECOND / dSampOutFreq;


   /***
   [[U this->siMaxDitherBufSize := 
       this->pDitherChannel->tWaveformChannel::GetMemSize()]]
   ***/
   this->siMaxDitherBufSize  = this->pDitherChannel->GetMemSize();
   this->siDoubleMaxDitherBufSize = this->siMaxDitherBufSize * 2;
 
   /* for simplicity, allocate each array to max possible size */
   bGotSampleArray = new BOOL[this->siMaxDitherBufSize];
   dSweepSampleArray = new double[this->siMaxDitherBufSize];
   dCollectDitherTable = new double[this->siMaxDitherBufSize];
   dCollectSqDitherTable = new double[this->siMaxDitherBufSize];
   dWaveDistortionTable = new double[this->siMaxDitherBufSize];
   dIdealTriangleWave = new double[this->siMaxDitherBufSize];
   dDitherCorrTable = new double[this->siMaxDitherBufSize];
   dDitherOutputTable = new double[this->siDoubleMaxDitherBufSize];
   dDitherOutputTableInChannelUnits = new double[this->siMaxDitherBufSize];
   dSavedDitherCorrTable = new double[this->siMaxDitherBufSize];
   dSavedDitherOutputTable = new double [this->siDoubleMaxDitherBufSize];
  

   /***
   [[U this->siNextOffsetSem := ::semBCreate( SEM_Q_FIFO, SEM_EMPTY )]]
   ***/
   this->siNextOffsetSem = semBCreate( SEM_Q_FIFO, SEM_EMPTY );
   if (this->siNextOffsetSem==NULL)
      TraceMsg (tmt_ERROR, errnoGet(), "tDlDither::tDlDither");
   TraceMark(S_dlDither_SEM_CREATED);


   if (bGotSampleArray == NULL ||
       dSweepSampleArray == NULL ||
       dCollectDitherTable == NULL ||
       dCollectSqDitherTable == NULL ||
       dWaveDistortionTable == NULL ||
       dIdealTriangleWave == NULL ||
       dDitherCorrTable == NULL ||
       dDitherOutputTable == NULL ||
       dDitherOutputTableInChannelUnits == NULL ||
       dSavedDitherCorrTable == NULL ||
       dSavedDitherOutputTable == NULL) {
         TraceMsg (tmt_ERROR, ERROR, "tDlDither::tDlDither");
         return;
       }

   /* zero all arrays to start with: */

   for (i=0; i<this->siMaxDitherBufSize; i++)
   {
      this->bGotSampleArray[i] = FALSE;
      this->dSweepSampleArray[i] = 0.0;
      this->dCollectDitherTable[i] = 0.0;
      this->dCollectSqDitherTable[i] = 0.0;
      this->dWaveDistortionTable[i] = 0.0;
      this->dIdealTriangleWave[i] = 0.0;

      this->dDitherCorrTable[i] = 0.0;
      this->dDitherOutputTable[i] = 0.0;
      this->dDitherOutputTable[i+this->siMaxDitherBufSize] = 0.0;
      this->dDitherOutputTableInChannelUnits[i] = 0.0;
      this->dSavedDitherCorrTable[i] = 0.0;
      this->dSavedDitherOutputTable[i] = 0.0;
      this->dSavedDitherOutputTable[i+this->siMaxDitherBufSize] = 0.0;
   }

   // Set dither amplitude
   /***
   [[U ec := this->tDlDither::SetDitherParameters (
         dDitherAmplitudeIn, tDitherPeriodIn)]]
   ***/
   ec = this->SetDitherParameters (dDitherAmplitudeIn, tDitherPeriodIn);
   if (ec!=OK) {
     TraceMsg (tmt_ERROR, ec, "tDlDither::tDlDither");
     return;
   }  

          /* put 2 cycles in dDitherOutputTable[] */
   /***
   [[U ec := this->tDlDither::SetWaveform()]]
   ***/
   ec = this->SetWaveform ();
   if (ec!=OK) {
     TraceMsg (tmt_ERROR, ec, "tDlDither::tDlDither");
     return;
   }  

   TraceMark(S_dlDither_INITIAL_WAVEFORM_SET);

   for (i =0 ; i < this->siDoubleMaxDitherBufSize; i++)
      this->dSavedDitherOutputTable[i] = this->dDitherOutputTable[i];

}  





/******************************************************************
*
* int tDlDither::GetStoreIndex(TIME tTheTime)
*
* SPEC:
   [[D tDlDither::GetStoreIndex_CB(TIME tTheTime)]
   [ this invalid => UNDEFINED
   | true =>
       return := index into waveform measurement table that corresponds
                   to the output waveform sample that was output
                   at time tTheTime
   ]]

* RETURNS:
*    (int) index into table
*
*****/
/* inline code in header file */

/******************************************************************
*
* double tDlDither::GetDitherValue(TIME tCurrTime)
*
* SPEC:
   [[D tDlDither::GetDitherValue_CB(TIME tCurrTime)]
   [ this invalid => UNDEFINED
   | this->sdDitherSamplePeriod == 0 => UNDEFINED
   | this->bDithering == FALSE => return := 0.0
   | this->sdDitherSamplePeriod or this->iWaveOffset or
       this->siDitherBufSize not appropriate for current conditions =>
                                     return := anything
   | true =>
       return := waveform measurement table value that corresponds 
                   to the output waveform sample that was output 
                   at time tCurrTime 
   ]]

* RETURNS:
*    (double) dither value
*
*****/
double tDlDither::GetDitherValue(TIME tCurrTime)
{

   if (this->bDithering)
   {
      /***
        [[U tDlDither::GetCorrectionIndex(tCurrTime)]]
       ***/
      return this->dDitherCorrTable[ this->GetCorrectionIndex(tCurrTime) ];
   }
   else
      return 0.0;

}  /* tDlDither::GetDitherValue() */


/******************************************************************
*
* ERROR_CODE tDlDither::SetWaveform()
*
*     SetWaveform draws a triangle wave in this->dDitherOutputTable[],
*     as well as in this->dIdealTriangleWave[].
*     It should also take the metrics necessary for modifying a perfect
*     triangle wave so that outputting the modified wave will produce
*     a more perfect triangle wave modulation in the metrology.
*     For Incr 2, we get the new output wave by subtracting the diff 
*     betw the measured wave and the output wave from the output wave 
*     and iterate (suggestion from mmc).  this->dIdealTriangleWave[]
*     stays constant, providing a yardstick against which to measure
*     the quality of our measured dither waveform.
*
*     Currently, we ignore iPhaseOffset completely, except in calculating
*     the pseudo-"distortion" metric, this->CalibInfo.dMetric.
*
*     SetWaveform also sets the board-wide dither period; note this affects 
*     ALL DLs using this waveform generator board (in incr 2, ALL DLs).
*     The period gets set every time SetWaveform is called, since within 
*     SetWaveform we don't know whether or not tPeriod is a new value.
*
*     Dither amplitude is specified in meters of stellar Optical Path Delay
*
* SPEC:
   [[U tDlDither::SetWaveform()]]
*
* RETURNS:
*    OK
*    S_dlDither_INVALID_AMPL_FACTOR
*    S_dlDither_INVALID_DITHER_PERIOD
*    S_dlDither_INVALID_BUFSIZE
*
*****/
ERROR_CODE tDlDither::SetWaveform(void)
{
   int i, k, ec;
   int iPosPeak, iNegPeak;    /* use these for readability */
   int iSourceIndex;
   double dSlope, dSampOutFreq, dHalfAmplitude;
   double dLaserAmplitude;

   if ( this->dCurrentAmplFactor <= 0.0) 
      return Traceback(S_dlDither_INVALID_AMPL_FACTOR,"tDlDither::SetWaveform");

     /* if SetWaveformSize() returns error, call GetWS() to find */
     /* out what board thinks the waveform size is.              */
     /*  (this is why we don't have to check limits on iBufSize) */
 
   /***
     [[U ec := 
           this->pDitherBank->tWaveformChannelBank::SetWaveformSize(iBufSize)]]
    ***/
   ec = this->pDitherBank->SetWaveformSize(this->siDitherBufSize);

   if (ec != OK)
      /***
        [[U this->siDitherBufSize := 
                  this->pDitherBank->tWaveformChannelBank::GetWaveformSize()]]
       ***/
      this->siDitherBufSize = this->pDitherBank->GetWaveformSize();

   
         /* now set related quantities */
   this->siDoubleDitherBufSize = 2 * this->siDitherBufSize;
   this->siSweepBufSize = this->stDitherPeriod / this->tDlServoTaskPeriod;
   this->siMaxWaveOffset = this->siDitherBufSize / this->siSweepBufSize;

         /* If this isn't exact, this algorithm doesn't handle it; */
         /* change must be made in hardware setup (or configMgr)   */
   if (this->siSweepBufSize * this->siMaxWaveOffset != this->siDitherBufSize)
      return Traceback(S_dlDither_INVALID_DITHER_SAMPLE_RATE,
                       "tDlDither::SetWaveform");

   
        /* for this part, ignore phase offset and just make triangle wave */

        /* slope of triangle between samples (not incl. sign) */

      dSlope = 2 * this->dDitherAmplitude * this->dCurrentAmplFactor
               / ((double) this->siDitherBufSize);

      #if POSITIVE_SINUSOIDAL_STYLE_WAVEFORM

        iPosPeak = this->siDitherBufSize / 4;      /* array index of pos peak */
        iNegPeak = 3 * this->siDitherBufSize / 4;  /* array index of neg peak */

           /* go from amplitude -A/2 at i=0 to amplitude A/2 at i=iPosPeak */
        for (i = 0; i < iPosPeak; i++)
           dDitherOutputTable[i] = dIdealTriangleWave[i] = dSlope * i;

           /* go from A/2 at i=iPosPeak to -A/2 at i=iNegPeak */
        for (i = iPosPeak, k=0; i < iNegPeak; i++, k++)
           dDitherOutputTable[i] = 
	        dIdealTriangleWave[i] = dSlope * (iPosPeak - k);

           /* go from -A/2 at i=iNegPeak to 0 at end of wave */
        for (i = iNegPeak, k=0; i < this->siDitherBufSize; i++, k++)
           dDitherOutputTable[i] = 
	 	    dIdealTriangleWave[i] = dSlope * (k - iPosPeak);

      #else    // NEGATIVE_COSINE_STYLE_WAVEFORM
        
        iPosPeak = this->siDitherBufSize / 2;      /* array index of pos peak */
        dHalfAmplitude = this->dDitherAmplitude * this->dCurrentAmplFactor / 2.0;

           /* go from amplitude -A/2 at i=0 to amplitude A/2 at i=iPosPeak */
        for (i = 0; i < iPosPeak; i++)
           dDitherOutputTable[i] = dIdealTriangleWave[i] =
                         dSlope * i - dHalfAmplitude;

           /* go from A/2 at i=iPosPeak to -A/2 at i=this->siDitherBufSize */
        for (i = iPosPeak; i < this->siDitherBufSize; i++)
           dDitherOutputTable[i] = 
	  	    dIdealTriangleWave[i] =
                         dHalfAmplitude - dSlope * (i - iPosPeak);

      #endif


      // Apply phase shift
      /* Copy shifted waveform into second half of array.  Then later
       * we'll copy the second half of the array back into the first half.
       * (Note that the array is large enough to hold *two* complete copies
       * of the waveform.  We use the second copy here and then we don't
       * need a temporary buffer.)
       */
       for (i=0; i < this->siDitherBufSize; i++) {
         iSourceIndex = (i + this->iCurrentPhaseShiftIndex) % this->siDitherBufSize;
         /* Note: In a%b, if a<0, then a%b will be between -b<a%b<=0.  So we have
          * to add b back if the result is less than 0.
          */
         if (iSourceIndex < 0) 
           iSourceIndex += this->siDitherBufSize;
         
         dDitherOutputTable[this->siDitherBufSize + i] =
                                    dDitherOutputTable[iSourceIndex];
       }

   
   /* now copy second half of wave into first half of array so we have two cycles */
   for (i=0; i<this->siDitherBufSize; i++)
      dDitherOutputTable[i] =
        this->dDitherOutputTable[i + this->siDitherBufSize];

   return OK;

}  /* tDlDither::SetWaveform() */


/******************************************************************
*
* ERROR_CODE tDlDither::SetDitherParameters(double dNewAmplitude, TIME tNewPeriod)
*
*   This changes the dither wave amplitude, although it does not
*	actually get loaded into the waveform board RAM until
*	the next DitherOn command (which may come from a dither 
*	check or calibrate command engine).
*
* SPEC:
   [[D tDlDither::SetDitherParameters_CB(double dNewAmplitude)]
   [ this invalid => UNDEFINED
   | true => 
       finish the spec later
   ]]
* RETURNS:
*    OK
*    SetDitherWaveform error msgs
*
*****/
ERROR_CODE tDlDither::SetDitherParameters(double dNewAmplitude, TIME tNewPeriod)
{
   ERROR_CODE ec = OK;
   double dSampleOutputFreq;
   int iBufSize;
   char sMsgOut[1024];


   // Assume amplitude in laser counts.
   // (This will be converted to dither units
   // in LoadDitherWaveform.)  Laser counts are the internal native units
   // of the dither class.
   
   this->dDitherAmplitude = dNewAmplitude; 


   // Reset the amplification factor in case this set amplitude request
   // might be because we got it screwed up before
   this->dCurrentAmplFactor = 1.0;
   this->dProposedAmplFactor = 1.0;
   this->iCurrentPhaseShiftIndex = 0;
   this->iProposedPhaseShiftIndex = 0;

          /* if we don't like the dither period, balk (incr 2 only) */

      /* NOTE: Even in double precision, sdDitherSamplePeriod may be a value  */
      /* that gets represented inaccurately enough to cause this logic to fail. */
      /* So we actually check for the answer to be "near" zero instead of exact. */

   #define EPSILON (1e-10)

   if ( (( fmod(tNewPeriod,this->sdDitherSamplePeriod) > EPSILON ) &&
         ( fmod(tNewPeriod,this->sdDitherSamplePeriod) < 
			 (this->sdDitherSamplePeriod - EPSILON) )) ||
        ( (tNewPeriod % this->tDlServoTaskPeriod) != 0 ) )
   {
      sprintf(sMsgOut,
     "One of these is not 0:\n fmod(%lu,%0.20g) = %0.20g"
              "(< %0.20g or > %0.20g),\n\tor %lu %% %lu = %lu",
        tNewPeriod, this->sdDitherSamplePeriod,
              fmod(tNewPeriod,this->sdDitherSamplePeriod),
              EPSILON, this->sdDitherSamplePeriod-EPSILON, tNewPeriod,
              this->tDlServoTaskPeriod,tNewPeriod % this->tDlServoTaskPeriod );
      
      TraceMsg(tmt_ERROR,S_dlDither_INVALID_DITHER_PERIOD,sMsgOut);
      
      return Traceback(S_dlDither_INVALID_DITHER_PERIOD,
                       "tDlDither::SetDitherParameters", dNewAmplitude, tNewPeriod);
   }
 
   this->stDitherPeriod = tNewPeriod;
   
   /***
     [[U dSampleOutputFreq := 
                   this->pDitherBank->tWaveformChannelBank::GetSampleFreq()]]
    ***/
   dSampleOutputFreq = this->pDitherBank->GetSampleFreq();

   iBufSize = (int)(( tNewPeriod / (1.0*TIME_TICKS_PER_SECOND) 
                                       * dSampleOutputFreq ) + 0.5); /* round */ 
   if (iBufSize <= 0)
      return Traceback(S_dlDither_INVALID_BUFSIZE,
                       "tDlDither::SetDitherParameters", dNewAmplitude, tNewPeriod);

   this->siDitherBufSize = iBufSize;
   
   return ec;
  
}  /* tDlDither::SetDitherParameters() */

 
/******************************************************************
*
* ERROR_CODE tDlDither::LoadDitherWaveform(int iOffset)
*     LoadDitherWaveform takes the current waveform stored in 
*     this->dDitherOutputTable[], converts it to "channel units"
*     and outputs it to the waveform board's RAM.  
*     iOffset changes the starting point.  (Note that a positive
*     offset means the wave shifts *earlier* in time.) Always
*     loads this->siDitherBufSize samples to include a full cycle.
*
*     Note for spec purists:  there is a *little* sequentiality implied
*     by first setting this->dDitherOutputTableInChannelUnits and then 
*     using it, but I decided to leave it in rather than write a spec fn.
*
* SPEC:
   [[D tDlDither::LoadDitherWaveform_CB(int iOffset)]
   [ this invalid => UNDEFINED
   | this->pDitherChannel is invalid => UNDEFINED
   | iOffset >= this->siDitherBufSize => return := S_dlDither_INVALID_OFFSET
   | (iOffset + this->siDitherBufSize) >= this->siDoubleMaxDitherBufSize =>
           return := S_dlDither_INVALID_BUFSIZE
   | this->pDitherChannel->bInit == FALSE => 
       this->dDitherOutputTableInChannelUnits[] := 
             this->dDitherChannelUnits * (this->dDitherOutputTable+iOffset)[]
       : return := S_tWaveformChannel_CHANNEL_NOT_INITIALIZED
   | ABS(any element in this->dDitherOutputTableInChannelUnits) >=
          2^this->pDitherChannel->iBitWidth =>
       this->dDitherOutputTableInChannelUnits[] := 
             this->dDitherChannelUnits * (this->dDitherOutputTable+iOffset)[]
       : this->pDitherChannel->dActualChannelValues := anything 
                 (loaded values will reflect relevant bits)
       : return := OK
   | this->pDitherChannel->ccOuputCode == CC_COMPLEMENT_BINARY =>
       this->dDitherOutputTableInChannelUnits[] := 
             this->dDitherChannelUnits * (this->dDitherOutputTable+iOffset)[]
       : this->pDitherChannel->dActualChannelValue :=
            (2^this->pDitherChannel->iBitWidth - 1 -
              Unsigned(this->dDitherOutputTableInChannelUnits[0..waveform size],
                 this->pDitherChannel->iBitWidth) *
                        DaResolution(this->pDitherChannel->iOutputRange,
                                              this->pDitherChannel->iBitWidth))
       : return := OK
   | this->pDitherChannel->ccConvCode == CC_STRAIGHT_BINARY =>
       this->dDitherOutputTableInChannelUnits[] := 
             this->dDitherChannelUnits * (this->dDitherOutputTable+iOffset)[]
       : this->pDitherChannel->dActualChannelValue :=
              Unsigned(this->dDitherOutputTableInChannelUnits[0..waveform size],
                 this->pDitherChannel->iBitWidth) *
                        DaResolution(this->pDitherChannel->iOutputRange,
                                              this->pDitherChannel->iBitWidth)
       : return := OK
   | this->pDitherChannel->ccConvCode == CC_OFFSET_BINARY =>
       this->dDitherOutputTableInChannelUnits[] := 
             this->dDitherChannelUnits * (this->dDitherOutputTable+iOffset)[]
       : this->pDitherChannel->dActualChannelValue :=
              Unsigned(this->dDitherOutputTableInChannelUnits[0..waveform size],
                 this->pDitherChannel->iBitWidth) *
                       DaResolution(this->pDitherChannel->iOutputRange,
                           this->pDitherChannel->iBitWidth) -
                               (2^this->pDitherChannel->iBitWidth/2)
       : return := OK
   | this->pDitherChannel->ccConvCode == CC_TWOS_COMPLEMENT =>
       this->dDitherOutputTableInChannelUnits[] := 
             this->dDitherChannelUnits * (this->dDitherOutputTable+iOffset)[]
       : this->pDitherChannel->dActualChannelValue := 
                 this->dDitherOutputTableInChannelUnits[0..waveform size] *
                       DaResolution(this->pDitherChannel->iOutputRange,
                                             this->pDitherChannel->iBitWidth)
       : return := OK
   | true =>
       this->dDitherOutputTableInChannelUnits[] := 
             this->dDitherChannelUnits * (this->dDitherOutputTable+iOffset)[]
       : this->pDitherChannel->dActualChannelValues := 
                  this->dDitherOutputTableInChannelUnits[0..waveform size] ]]
       : return := OK
   ]]

* RETURNS:
*    OK
*    S_dlDither_INVALID_OFFSET
*    S_dlDither_INVALID_BUFSIZE
*    S_tWaveformChannel_CHANNEL_NOT_INITIALIZED
*
*****/

ERROR_CODE tDlDither::LoadDitherWaveform(int iOffset)
{
   ERROR_CODE ec = OK;
   int i;

   if (iOffset > this->siDitherBufSize) 
       return Traceback(S_dlDither_INVALID_OFFSET,
                           "tDlDither::LoadDitherWaveform",iOffset);

   if ((iOffset + this->siDitherBufSize) > this->siDoubleMaxDitherBufSize) 
       return Traceback(S_dlDither_INVALID_BUFSIZE,
                           "tDlDither::LoadDitherWaveform",iOffset);


   /* first convert waveform from laser counts to dither channel units */
   /* Note - Positive PZT motion (stack growth) results in a *decrease*
    * in optical path.  The user, however, wants a positive waveform
    * (starting at zero, then going positive, then returning to zero).
    * So we have to invert the drive signal for the PZT.
    */

   for (i=0; i<this->siDitherBufSize; i++)
     this->dDitherOutputTableInChannelUnits[i] = 
                                   - this->dDitherOutputTable[i+iOffset]
                                   * this->dDitherChannelUnits;

   /* now load converted waveform into wave board RAM */
   /***
     [[U ec := this->pDitherChannel->tWaveformChannel::LoadMemory(
                                   this->dDitherOutputTableInChannelUnits)]]
    ***/
   ec = this->pDitherChannel->LoadMemory(this->dDitherOutputTableInChannelUnits);
   if (ec!=OK)
       return Traceback(ec,"tDlDither::LoadDitherWaveform",iOffset);
   else
      return OK;

}  /* tDlDither::LoadDitherWaveform() */



/******************************************************************
*
* ERROR_CODE tDlDither::DitherOff(void)
*
* SPEC:
   [[D tDlDither::DitherOff_CB(void)]
   [ this invalid || this->pDitherChannel invalid => UNDEFINED
   | this->pDitherChannel->bInit == FALSE => UNDEFINED
   | ABS(this->pDitherChannel->iResetValue) >= 2^this->iBitWidth =>
        this->pDitherChannel->dActualChannelValues[] := anything 
	      (output will reflect relevant bits only)
        : return := OK
   | true =>
        this->pDitherChannel->dActualChannelValues[] := channel value
			 corresponding to this->pDitherChannel->iResetValue 
        : return := OK
   ]]

* RETURNS:
*    OK
*
*****/
ERROR_CODE tDlDither::DitherOff(void)
{
   int ec;

   /***
     [[U ec := this->pDitherChannel->tOutputChannel::Reset()]]
    ***/
   ec = this->pDitherChannel->Reset();
   if (ec != OK)
      return Traceback(ec, "tDlDither::DitherOff");

   this->bDithering = FALSE;

   return OK;

}  /* tDlDither::DitherOff() */



/******************************************************************
*
* ERROR_CODE tDlDither::DitherOn(void)
*
* SPEC:
   [[D tDlDither::DitherOn_CB(void)]
   [ this invalid => UNDEFINED
   | this->pDitherChannel is invalid => UNDEFINED
   | this->pDitherChannel->bInit == FALSE =>
       this->dDitherOutputTableInChannelUnits[] :=
             this->dDitherChannelUnits * this->dDitherOutputTable[]
       : return := S_tWaveformChannel_CHANNEL_NOT_INITIALIZED
   | ABS(any element in this->dDitherOutputTable) >=
         (2^this->pDitherChannel->iBitWidth/this->this->dDitherChannelUnits) =>
       this->dDitherOutputTableInChannelUnits[] :=
             this->dDitherChannelUnits * this->dDitherOutputTable[]
       : this->pDitherChannel->dActualChannelValues := anything
                 (loaded values will reflect relevant bits)
       : this->bDithering := TRUE
       : return := OK
   | this->pDitherChannel->ccOuputCode == CC_COMPLEMENT_BINARY =>
       this->dDitherOutputTableInChannelUnits[] :=
             this->dDitherChannelUnits * this->dDitherOutputTable[]
       : this->pDitherChannel->dActualChannelValue :=
            (2^this->pDitherChannel->iBitWidth - 1 -
              Unsigned(this->dDitherOutputTableInChannelUnits[0..waveform size],
                 this->pDitherChannel->iBitWidth) *
                        DaResolution(this->pDitherChannel->iOutputRange,
                                              this->pDitherChannel->iBitWidth))
       : this->bDithering := TRUE
       : return := OK
   | this->pDitherChannel->ccConvCode == CC_STRAIGHT_BINARY =>
       this->dDitherOutputTableInChannelUnits[] :=
             this->dDitherChannelUnits * this->dDitherOutputTable[]
       : this->pDitherChannel->dActualChannelValue :=
              Unsigned(this->dDitherOutputTableInChannelUnits[0..waveform size],
                 this->pDitherChannel->iBitWidth) *
                        DaResolution(this->pDitherChannel->iOutputRange,
                                              this->pDitherChannel->iBitWidth)
       : this->bDithering := TRUE
       : return := OK
   | this->pDitherChannel->ccConvCode == CC_OFFSET_BINARY =>
       this->dDitherOutputTableInChannelUnits[] :=
             this->dDitherChannelUnits * this->dDitherOutputTable[]
       : this->pDitherChannel->dActualChannelValue :=
              Unsigned(this->dDitherOutputTableInChannelUnits[0..waveform size],
                 this->pDitherChannel->iBitWidth) *
                       DaResolution(this->pDitherChannel->iOutputRange,
                           this->pDitherChannel->iBitWidth) -
                               (2^this->pDitherChannel->iBitWidth/2)
       : this->bDithering := TRUE
       : return := OK
   | this->pDitherChannel->ccConvCode == CC_TWOS_COMPLEMENT =>
       this->dDitherOutputTableInChannelUnits[] :=
             this->dDitherChannelUnits * this->dDitherOutputTable[]
       : this->pDitherChannel->dActualChannelValue :=
                 this->dDitherOutputTableInChannelUnits[0..waveform size] *
                       DaResolution(this->pDitherChannel->iOutputRange,
                                             this->pDitherChannel->iBitWidth)
       : this->bDithering := TRUE
       : return := OK
   | true =>
       this->dDitherOutputTableInChannelUnits[] :=
             this->dDitherChannelUnits * this->dDitherOutputTable[]
       : this->pDitherChannel->dActualChannelValues :=
                  this->dDitherOutputTableInChannelUnits[0..waveform size] ]]
       : this->bDithering := TRUE
       : return := OK
   ]]

* RETURNS:
*    OK
*    S_dlDither_INVALID_OFFSET
*    S_dlDither_INVALID_BUFSIZE
*    S_tWaveformChannel_CHANNEL_NOT_INITIALIZED
*
*****/
ERROR_CODE tDlDither::DitherOn(void)
{
   int ec;

   /***
       [[U ec := this->tDlDither::LoadDitherWaveform(0)]]
    ***/
   ec = this->LoadDitherWaveform(0);
   if (ec != OK)
      return Traceback(ec, "tDlDither::DitherOn");

   this->bDithering = TRUE;

   return OK;

}  /* tDlDither::DitherOn() */



/******************************************************************
*
* ERROR_CODE tDlDither::Initialize(tDlGizmo         *pGizmo, 
                                   TIME              tDlServoPeriodToUse,     
                                   tWaveformChannel *pDitherPztChannel,
                                   tActuator        *pWaveformActuatorToUse,
                                   tLaserMetrologySensor *pLaserSensorToUse,
                                   TIME              tInitPeriod,
                                   double            dInitAmplitude)
*
* 
* Note to spec purists: For readability reasons, I have left out
* "partial initialization" consequences.  I.e. if this function returns
* an error code, the spec states only the error code that is returned;
* in fact, some of the assignments or other conditions that appear under
* the "true" case may also have occurred.  Until and unless it proves
* worth the trouble to make every case *completely* correct, the reviewer
* (or debugger) will have to resort to code inspection to determine what
* the side effects of a given exception path will be.
* 
* 
* SPEC:
   [[D tDlDither::Initialize_CB(tDlGizmo *pGizmo, TIME tDlServoPeriodToUse,
                                  tWaveformChannel *pDitherPztChannel,
                                  TIME tInitPeriod, double dInitAmplitude)]
   [ this invalid || pGizmo invalid => UNDEFINED
   | waveform generator board not operational or not scanning => UNDEFINED
   | pWaveformActuatorToUse or pLaserSensorToUse invalid => UNDEFINED
   | something here about proper init of actuator and sensor => UNDEFINED
   | pDitherPztChannel invalid || pDitherPztChannel->bInit == FALSE => UNDEFINED
   | pDitherPztChannel->pChannelBank invalid => UNDEFINED
   | pGizmo == NULL => return := S_dlDither_NULL_GIZMO_POINTER
   | pDitherPztChannel == NULL => return := S_dlDither_NULL_CHANNEL_POINTER
   | pDitherPztChannel->pChannelBank == NULL => 
				  return := S_dlDither_NULL_BANK_POINTER
   | (1/(0.000001 * sample period of each channel)) <= 0.0 => 
				  return := S_dlDither_INVALID_SAMPLE_FREQ
   | cannot create semaphore this->siNextOffsetSem => 
				  return := VxWorks error msg (semBCreate)
   | insufficient system memory to allocate *all* arrays =>
				  return := VxWorks error msg (memPartAlloc)
   | tInitPeriod not an integer multiple of TIME_TICKS_PER_SECOND /
               (1/(0.000001 * sample period of each channel)) ||
                tInitPeriod not an integer multiple of tDlServoPeriodToUse =>
        return := S_dlDither_INVALID_DITHER_PERIOD
   | tInitPeriod <=0  || ROUND( tInitPeriod / (1.0*TIME_TICKS_PER_SECOND) *
	   (1/(0.000001 * sample period of each channel)) )  >
                                     this->pDitherBank->iTotalMemoryLoc =>
                this->siDitherBufSize := waveform size value from channel bank
   | true =>
       this->pDlStatus := &(pGizmo->DlStatus)
       : this->bDoingCheckOrCalib := FALSE
       : this->dDitherChannelUnits := conversion value from counts to 
							       channel units
       : this->pDitherBank := pDitherPztChannel->pChannelBank
       : this->pDitherChannel := pDitherPztChannel
       : this->tDlServoTaskPeriod := tDlServoPeriodToUse
       : pDitherPztChannel->dActualChannelValues[] := channel value
		          corresponding to pDitherPztChannel->iResetValue
       : this->sdDitherSamplePeriod := TIME_TICKS_PER_SECOND / 
			    (1/(0.000001 * sample period of each channel))
       : this->siMaxDitherBufSize := amount of wave board memory available to 
       : this->siDoubleMaxDitherBufSize := this->siMaxDitherBufSize * 2
       : this->bCalibrating := FALSE
       : this->bWasDitherOn := FALSE
       : this->bStoreDither := FALSE
       : this->siNextOffsetSem := ID of VxWorks binary semaphore created 
				       with options (SEM_Q_FIFO, SEM_EMPTY)
       : this->iWaveOffset := 0
       : this->iWavesToAvg := NUM_WAVES_TO_AVG
       : this->dCurrentAmplFactor := 1.0
       : this->iCurrentPhaseShiftIndex := 0
       : this->iProposedPhaseShiftIndex := 0
       : this->CalibInfo.dAmplitude  := 0.0
       : this->CalibInfo.dPhaseShift := 0.0
       : this->CalibInfo.dMetric     := 0.0
       : these arrays allocated and initialized to 0 values ...
           this->dCollectDitherTable [this->siMaxDitherBufSize]
	   this->dCollectSqDitherTable [this->siMaxDitherBufSize]
	   this->dWaveDistortionTable [this->siMaxDitherBufSize]
	   this->dIdealTriangleWave [this->siMaxDitherBufSize]
	   this->dDitherCorrTable [this->siMaxDitherBufSize]
	   this->dDitherOutputTable [this->siDoubleMaxDitherBufSize]
	   this->dDitherOutputTableInChannelUnits [this->siMaxDitherBufSize]
	   this->dSavedDitherCorrTable [this->siMaxDitherBufSize]
	   this->dSavedDitherOutputTable [this->siDoubleMaxDitherBufSize]
       : this->stDitherPeriod := tInitPeriod
       : this->siDitherBufSize := bufsize needed to store waveform
       : waveform size for this->pDitherBank := bufsize needed to store waveform
       : this->siDoubleDitherBufSize := 2 * this->siDitherBufSize
       : this->dDitherAmplitude := dInitAmplitude
       : this->dIdealTriangleWave[] gets a triangle wave of amplitude
                     this->dDitherAmplitude and period this->tDitherPeriod
       : this->dDitherOutputTable[] gets two cycles of a perfect triangle wave 
            of amplitude this->dDitherAmplitude and period this->tDitherPeriod
       : this->dSavedDitherOutputTable[] gets two cycles of a perfect triangle 
                           wave of amplitude this->dDitherAmplitude and 
                                   period this->tDitherPeriod
       : return := OK 
   ]]
*
* RETURNS:
*    OK
*
*****/



/******************************************************************
*
* double tDlDither::DoDither(TIME tCurrTime, double dLaserReading)
*
*              (See also long description in dlDither.h)
*
* SPEC:
   [[X tDlDither::DoDither_CB(TIME tCurrTime, double dLaserReading)]
   [ this invalid => UNDEFINED
   | this->*pbDoingCheckOrCalib == TRUE && this->bStoreDither == TRUE =>
       iStoreIndex := this->GetStoreIndex(tCurrTime)
       : this->iFirstIndex == -1 => this->iFirstIndex := iStoreIndex
       : this->iSumCountArray[iStoreIndex] == this->iWavesToAvg =>
                  this->iNumFullSamples := this->iNumFullSamples + 1
       : iStoreIndex == this->iFirstIndex &&
               this->iSumCountArray[iStoreIndex] > this->iWavesToAvg =>
           this->bStoreDither := FALSE
           : this->iFirstIndex := -1
           : semGive(this->siNextOffsetSem)
           : return := this->dDitherCorrTable[iStoreIndex]
       : this->iSumCountArray[iStoreIndex] < this->iWavesToAvg =>
          [
            this->dCollectDitherTable[iStoreIndex] :=
                   this->dCollectDitherTable[iStoreIndex] + dLaserReading
            : this->dCollectSqDitherTable[iStoreIndex] :=
                   this->dCollectSqDitherTable[iStoreIndex] + dLaserReading^2
            : this->iSumCountArray[iStoreIndex] := 
                                     this->iSumCountArray[iStoreIndex] + 1
          ]
       : return := this->dDitherCorrTable[iStoreIndex]
   | this->*pbDoingCheckOrCalib == TRUE =>
       return := this->dDitherCorrTable[ this->GetStoreIndex(tCurrTime) ]
   | this->bDithering == FALSE =>
       return := 0.0
   | true =>
       return := this->dDitherCorrTable[ this->GetStoreIndex(tCurrTime) ]
   ]]

* RETURNS:
*    (double) appropriate dither value for time tCurrTime
*
*****/
double tDlDither::DoDither(TIME tCurrTime, double dLaserReading)
{
   int iStoreIndex, iRealIndex;
   double dSweepMean;
   int i, ec;

        /* normal DoDither operation: return dither value */


   if ( ((this->bDoingCheckOrCalib) == FALSE) || (!this->bStoreDither) )
      /***
        [[U return := tDlDither::GetDitherValue(tCurrTime)]]
       ***/
      return this->GetDitherValue(tCurrTime);


          /* only get here if we are doing calib/check AND storing */
  /* (NOTE this logic assumes "bStoreDither == TRUE" means we are dithering) */


   /***
   [[U temp := this->tDlDither::GetStoreIndex (tCurrTime) ]]
   
   [ iStoreIndex := temp / this->siMaxWaveOffset ]
   ***/
   iStoreIndex = (int)(this->GetStoreIndex(tCurrTime)/this->siMaxWaveOffset);


/***** mega output, but need this info  ****/
//   if (this->iTotalSweepCount % 20 == 0)
//      printf("Sweep # %d, Time=%lu, iStoreIndex=%d, iSweepErrorCount=%d\n",
//	  this->iTotalSweepCount,tCurrTime,iStoreIndex,this->iSweepErrorCount);
/***** mega output, but need this info  ****/


     /* if bGotSampleArray[iStoreIndex] is FALSE then we're still doing  */
     /* the current sweep.  otherwise, we've come back to starting pt.   */

   if ( this->bGotSampleArray[iStoreIndex] == FALSE ) 
   {
      this->iSweepSampleCount++;   /* count # samples in this sweep */
      this->dSweepSampleArray[iStoreIndex] = dLaserReading;  
      this->bGotSampleArray[iStoreIndex] = TRUE;  
   }
   else 
   {      /* we're done with this sweep; did we get all the values? */  

      if (this->iSweepSampleCount == this->siSweepBufSize) /* good sweep */
      {
	 this->iTotalSweepCount++;    /* number of sweeps at this offset */


     /* subtract mean from sweep values and store in main arrays */

         dSweepMean = 0.0;
         for (i=0; i<this->siSweepBufSize; i++)
	    dSweepMean += this->dSweepSampleArray[i]; 
	 dSweepMean /= (double)this->siSweepBufSize;

// printf("Sweep # %d good: mean is %f\n",this->iTotalSweepCount-1,dSweepMean);

         for (i=0; i<this->siSweepBufSize; i++)
	 {
	    iRealIndex = (i*this->siMaxWaveOffset) + this->iWaveOffset;
	    this->dCollectDitherTable[iRealIndex] +=
	                         this->dSweepSampleArray[i] - dSweepMean;
	    this->dCollectSqDitherTable[iRealIndex] +=
	                   (this->dSweepSampleArray[i] - dSweepMean) *
	                         (this->dSweepSampleArray[i] - dSweepMean);
	 }
      }
      else    /* Error!! this sweep didn't get all its values */
      {   
	 this->iSweepErrorCount++;   
      }

         /* reset for next sweep (whether error or not) */

      this->iSweepSampleCount = 0;
      for (i=0; i<this->siSweepBufSize; i++)
         this->bGotSampleArray[i] = FALSE;

         /* if we're done with this offset... */

      if ( (this->iTotalSweepCount >= this->iWavesToAvg) ||
	          (this->iSweepErrorCount > MAX_SWEEP_ERRORS) )
      {
         this->bStoreDither = FALSE;              /* stop recording */

         /* tell MeasureDither that we're ready for the next offset */

         /***
           [[U ec := ::semGive( this->siNextOffsetSem )]]
          ***/
         ec = semGive( this->siNextOffsetSem );
         if (ec == ERROR)      
	 {
	    Traceback(errnoGet(),"tDlDither::DoDither",tCurrTime,dLaserReading);
            return -1.0;
	 }

         return this->dDitherCorrTable[iStoreIndex];
      }
   }

   return this->dDitherCorrTable[iStoreIndex];

}  /* tDlDither::DoDither() */



/******************************************************************
*
* ERROR_CODE tDlDither::MeasureDither(int iNumWaves)
*     MeasureDither fills up dCollectDitherTable[] with laser readings
*     from DoDither, offsetting the dither waveform in between (this 
*     requires reloading the waveform starting at a different point 
*     each time), then calculates the average waveform, subtracts the
*     mean, and stores the avg waveform's amplitude, phase, and 
*     distortion in the DL Status block.
*
* ***** check code to make sure we clean up on abnormal exit!! ***** *
*
* SPEC:
   [[X tDlDither::MeasureDither_CB(int iNumWaves)]
   [ this invalid => UNDEFINED
   | waveform parameters change during this function => UNDEFINED
   | (this->iNumWaves==0) || (this->siDitherBufSize<=0) ||
              (this->dDitherAmplitude<=0) => UNDEFINED
   | true =>
       this->iWavesToAvg := iNumWaves
       :\ for i == 0 to this->siDitherBufSize-1,
             this->dCollectDitherTable[i] := 0.0
       :\ for i == 0 to this->siDitherBufSize-1,
             this->dCollectSqDitherTable[i] := 0.0
       :\ for this->iWaveOffset == 0 to this->siDitherBufSize-1,
             if this->iNumFullSamples < this->siDitherBufSize (not all full)
             : [[U tDlDither::LoadDitherWaveform(iOffset=this->iWaveOffset)]]
             : this->bStoreDither := TRUE
             : semTake(this->siNextOffsetSem)
       :\ if this->iNumFullSamples < this->siDitherBufSize =>
                          return := S_dlDither_DITHER_MEASUREMENT_ERROR
       :\ for i == 0 to this->siDitherBufSize-1,
             this->dCollectDitherTable[i] :=
                       this->dCollectDitherTable[i] / iNumWaves
       :\ for i == 0 to this->siDitherBufSize-1,
             this->dCollectSqDitherTable[i] :=
                       this->dCollectSqDitherTable[i] / iNumWaves
       :\ calculate and subtract off mean value of avg waveform
       :\ calculate metrics
       :\ [[U tDlStatus::UpdateDitherCalibration(this=pDlStatus,
                                    dAmplitude=dDitherAmplitude,
                                    dPhase=dDitherPhase,
                                    dDistortion=dDitherDistortion) ]]
       : return := OK
   ]]

* RETURNS:
*    OK
*    S_dlDither_DITHER_MEASUREMENT_ERROR
*
*/

ERROR_CODE tDlDither::MeasureDither(int iNumWaves)
{
   int i, ec, iAdjIndex;
   double dMean, dMeanSq, dSum = 0.0, dSqSum = 0.0, dStdDev = 0.0;
   double dSumSqs = 0.0;
   double dMinAmplitude, dMaxAmplitude;
   int iMinAmplIndex, iMaxAmplIndex;
   double dPhaseShiftDegrees;
   double dDiffPerfect;

   TraceMark(S_dlDither_ENTERING_MEASURE_DITHER);

   if (iNumWaves <= 0)
      return Traceback(S_dlDither_INVALID_NUMWAVES,"tDlDither::MeasureDither",
							  iNumWaves);

   this->iWavesToAvg = iNumWaves;              /* how many waves to avg */
   for (i=0; i<this->siDitherBufSize; i++) {
      this->dCollectDitherTable[i] = 0.0;
      this->dCollectSqDitherTable[i] = 0.0;
   }
   

      /* shift output wave and read until we have read wave iNumWaves    */
      /* times.  then shift wave again.  repeat until all array elements */
      /* have accumulated iNumWaves measurements each.                   */

   for ( this->iWaveOffset=0; 
         (this->iWaveOffset < this->siMaxWaveOffset) ||
	     (this->iSweepErrorCount > MAX_SWEEP_ERRORS) ; 
	 this->iWaveOffset++ ) {
     
//      printf("\nSetting iWaveOffset to %d... \n",this->iWaveOffset);

              /* initialize sweep vars, to start with */
      this->iTotalSweepCount = 0;
      this->iSweepErrorCount = 0;
      this->iSweepSampleCount = 0;
      for (i=0; i<this->siSweepBufSize; i++)
         this->bGotSampleArray[i] = FALSE;

      /***
      [[U ec := this->tDlDither::LoadDitherWaveform(this->iWaveOffset)]]
      ***/
      ec = this->LoadDitherWaveform(this->iWaveOffset);
      if (ec!=OK)
         return Traceback(errnoGet(),"tDlDither::MeasureDither",iNumWaves);
	 
      this->bStoreDither = TRUE;

      TraceMark(S_dlDither_LOADED_WAVEFORM_NOW_WAIT);

      /***
      [[U ec := ::semTake( this->siNextOffsetSem, WAIT_FOREVER )]]
      ***/
      ec = semTake( this->siNextOffsetSem, WAIT_FOREVER );
      if (ec != OK)
         return Traceback(errnoGet(),"tDlDither::MeasureDither",iNumWaves);

// printf("\nMeasured iWaveOffset %d.\n",this->iWaveOffset);

      if (this->iSweepErrorCount > 0)
        printf("SWEEP COUNT ERRORS: iWaveOffset = %d, iSweepErrorCount = %d\n",
               this->iWaveOffset, this->iSweepErrorCount);
   }

      /* now get average wave, and calc mean value */

   for (i=0; i<this->siDitherBufSize; i++) {
     
      this->dCollectDitherTable[i] /= iNumWaves;
      this->dCollectSqDitherTable[i] /= iNumWaves;

      dSum += this->dCollectDitherTable[i];
      dSqSum += this->dCollectSqDitherTable[i];

    /* finally, this array stores std dev of each measurement location */
      this->dCollectSqDitherTable[i] = sqrt(this->dCollectSqDitherTable[i] - 
                   (this->dCollectDitherTable[i] * this->dCollectDitherTable[i]));
   }
   
   dMean = dSum / (1.0*this->siDitherBufSize);  /* mean value of measured wave */
   dMeanSq = dSqSum / (1.0*this->siDitherBufSize);  /* mean of squared meas. wave */
   dStdDev = sqrt(dMeanSq - (dMean*dMean));     /* std dev of whole wave (?) */

   dMaxAmplitude = dMinAmplitude = this->dCollectDitherTable[0] - dMean;
   iMaxAmplIndex = iMinAmplIndex = 0;


      /* now subtract mean, and calc metrics */

   for (i=0; i<this->siDitherBufSize; i++) {
      this->dCollectDitherTable[i] -= dMean;

      this->dWaveDistortionTable[i] = 
		 this->dDitherOutputTable[i] - this->dCollectDitherTable[i];

      if (this->dCollectDitherTable[i] < dMinAmplitude) {
         dMinAmplitude = this->dCollectDitherTable[i];
         iMinAmplIndex = i;
      }

      if (this->dCollectDitherTable[i] > dMaxAmplitude) {
         dMaxAmplitude = this->dCollectDitherTable[i];
         iMaxAmplIndex = i;
      }
   }

   this->dProposedAmplFactor = this->dCurrentAmplFactor *
                       this->dDitherAmplitude / (dMaxAmplitude-dMinAmplitude);

   
   /* if we get here and all array elements are *not* full, VERY BAD!! */
   /*  (but we wanted to do those mean/stddev calculations anyway)     */

   if (this->iSweepErrorCount > MAX_SWEEP_ERRORS)
      return Traceback(S_dlDither_DITHER_MEASUREMENT_ERROR,
                       "tDlDither::MeasureDither",iNumWaves);

   TraceMark(S_dlDither_MEASURED_DITHER_WAVE_SUCCESSFULLY);



   /* now measure "metrics": amplitude, phase, distortion 
            and store in this->pGizmo->Status block      */

    /* dumb way of measuring "phase": how far is peak from 1/2-wave pt? */
   this->iProposedPhaseShiftIndex = (this->iCurrentPhaseShiftIndex +
                                     iMaxAmplIndex - this->siDitherBufSize/2) %
                                    this->siDitherBufSize;
   dPhaseShiftDegrees = 360 * this->iProposedPhaseShiftIndex / 
                                      (1.0*this->siDitherBufSize);

   /**** actually for Incr 2 we don't measure distortion, ****/
   /**** just send it our "metric" of RMS difference betw ****/
   /**** measured wave and perfect triangle.              ****/

   for (i=0; i<this->siDitherBufSize; i++) {
     iAdjIndex = (i + this->iProposedPhaseShiftIndex
                    - this->iCurrentPhaseShiftIndex) % this->siDitherBufSize;
     /* Note: In a%b, if a<0, then a%b will be between -b<a%b<=0.  So we have
      * to add b back if the result is less than 0.
      */
     if (iAdjIndex < 0) 
       iAdjIndex += this->siDitherBufSize;
     dDiffPerfect = this->dIdealTriangleWave[i] - 
                                        this->dCollectDitherTable[iAdjIndex];
     dSumSqs += dDiffPerfect * dDiffPerfect;
   }
   
   this->CalibInfo.dMetric     = sqrt(dSumSqs / this->siDitherBufSize);
   this->CalibInfo.dPhaseShift = dPhaseShiftDegrees;

   // Amplitude in laser counts
   this->CalibInfo.dAmplitude  =   (dMaxAmplitude-dMinAmplitude);

   this->dAmpError = (dMaxAmplitude-dMinAmplitude) / this->dDitherAmplitude
                                                                      - 1.0;
   this->dPhaseError =
     (this->iProposedPhaseShiftIndex - this->iCurrentPhaseShiftIndex) /
                                              (double) this->siDitherBufSize;
   this->dDistortionError = this->CalibInfo.dMetric / this->dDitherAmplitude;


   printf("Leaving MeasureDither: Amp Factor = %0.4g, iPhaseShiftIndex = %d\n",
               this->dProposedAmplFactor, this->iProposedPhaseShiftIndex);
   printf("Metrics: Amp Error: %5.3f, Phase Error: %5.3f, Distortion: %5.3f\n",
          dAmpError, dPhaseError, dDistortionError);

   TraceMark(S_dlDither_LEAVING_MEASURE_DITHER);

   return OK;

}  /* tDlDither::MeasureDither() */



/******************************************************************
*
* ERROR_CODE tDlDither::CheckDither(void)
*    Called by tDlDitherCheckCommandEngine::ExecuteMethod();
*      assumes that the command engine calls tDlDither::CleanUp()
*
*
* ***** check code to make sure we clean up on abnormal exit!! ***** *
*
* SPEC:
   [[X tDlDither::CheckDither_CB(void)]
   [ this invalid => UNDEFINED
   | other bad things => UNDEFINED
   | true =>
       this->bWasDitherOn := this->bDithering
       :\ [[U tDlDither::DitherOn() ]]   (reset in CleanUp() )
       :\ [[U tDlDither::MeasureDither(iNumWaves=NUM_WAVES_TO_AVG) ]]
       :\ if this->MeasureDither returns S_dlDither_DITHER_MEASUREMENT_ERROR =>
            return := S_dlDither_DITHER_MEASUREMENT_ERROR
                      (assume command engine calls CleanUp())
       : return := OK
   ]]

* RETURNS:
*    OK
*    error msgs from these tDlDither methods
*
*****/
ERROR_CODE tDlDither::CheckDither(void)
{
   ERROR_CODE i, ec;

   TraceMark(S_dlDither_ENTERING_CHECK_DITHER);

   this->bDoingCheckOrCalib = TRUE;
   this->bWasDitherOn = this->bDithering;

   /***
     [[U ec := this->tDlDither::DitherOn()]]
    ***/
   ec=this->DitherOn();
   if (ec!=OK)
      return Traceback(ec,"tDlDither::CheckDither");

   /***
     [[U ec := this->tDlDither::MeasureDither(NUM_WAVES_TO_AVG)]]
    ***/
   ec = this->MeasureDither(NUM_WAVES_TO_AVG);
   if (ec!=OK)
      return Traceback(ec,"tDlDither::CheckDither");

   this->bDoingCheckOrCalib = FALSE;

   TraceMark(S_dlDither_SUCCESSFULLY_COMPLETED_CHECK_DITHER);

   return OK;

}  /* tDlDither::CheckDither() */


/******************************************************************
*
* ERROR_CODE tDlDither::CalibrateDither(void)
*    Called by tDlDitherCalibrateCommandEngine::ExecuteMethod()
*
*    note: :\ denotes sequentiality that I haven't gotten rid of yet
*
* ***** check code to make sure we clean up on abnormal exit!! ***** *
*
*
* SPEC:
   [[X tDlDither::CalibrateDither_CB(void)]
   [ this invalid => UNDEFINED
   | other bad things I haven't thought of yet => UNDEFINED
   | true =>
       this->bWasDitherOn := this->bDithering
       :\ [[U tDlDither::DitherOn() ]]
       :\ this->dSavedDitherOutputTable[] := this->dDitherOutputTable[]
       :\ this->dSavedDitherCorrTable[] := this->dDitherCorrTable[]
       :\      (save period and amplitude, too??)
       :\ [[U tDlDither::SetWaveform(tPeriod=this->stDitherPeriod,
                                      dAmplitude=this->dDitherAmplitude,
                                      dAmplFactor=1.0,iPhaseOffset=0,
                                                bPreDistortion=FALSE) ]]
       :\ [[U tDlDither::MeasureDither(iNumWaves=NUM_WAVES_TO_AVG) ]]
       :\ if this->MeasureDither returns S_dlDither_DITHER_MEASUREMENT_ERROR =>
            return := S_dlDither_DITHER_MEASUREMENT_ERROR
                      (assume command engine calls CleanUp())
       :\ while (this->dMetrics/this->dDitherAmplitude > CLOSE_ENOUGH)
            [[U tDlDither::SetWaveform()]]
            :\ [[U tDlDither::MeasureDither(iNumWaves=NUM_WAVES_TO_AVG) ]]
            :\ if this->MeasureDither returns
                                         S_dlDither_DITHER_MEASUREMENT_ERROR =>
                 return := S_dlDither_DITHER_MEASUREMENT_ERROR
                           (assume command engine calls CleanUp())
       :\ this->dDitherCorrTable[] := this->dCollectDitherTable[]
       : return := OK
   ]]
*
* RETURNS:
*    OK
*    error msgs from these tDlDither methods
*
*/

ERROR_CODE tDlDither::CalibrateDither(void)
{
   ERROR_CODE i, ec;
   int iNumCalibCycles = 0;

   TraceMark(S_dlDither_ENTERING_CALIBRATE_DITHER);

   this->bDoingCheckOrCalib = TRUE;
   this->bWasDitherOn = this->bDithering;

   // Reset waveform generation parameters, then output "raw" waveform
   this->dCurrentAmplFactor = 1.0;
   this->iCurrentPhaseShiftIndex = 0;
   
   /***
     [[U ec := this->tDlDither::SetWaveform()]]
    ***/
   ec = this->SetWaveform();
   if (ec!=OK) 
      return Traceback(ec,"tDlDither::CalibrateDither");

   TraceMark(S_dlDither_INITIAL_WAVEFORM_SET);

   /***
     [[U ec := this->tDlDither::DitherOn()]]
    ***/
   ec = this->DitherOn();
   if (ec!=OK) 
      return Traceback(ec,"tDlDither::CalibrateDither");

   for (i=0; i<this->siMaxDitherBufSize; i++)
      this->dSavedDitherCorrTable[i] = this->dDitherCorrTable[i];
   for (i=0; i<this->siDoubleMaxDitherBufSize; i++)
      this->dSavedDitherOutputTable[i] = this->dDitherOutputTable[i];

   /***
     [[U ec := this->tDlDither::MeasureDither(NUM_WAVES_TO_AVG) ]]
    ***/
   ec = this->MeasureDither(NUM_WAVES_TO_AVG);
   if (ec!=OK)
      return Traceback(ec,"tDlDither::CalibrateDither");

   TraceMark(S_dlDither_COMPLETED_FIRST_MEASURE_DITHER);

   while ( (fabs(this->dPhaseError) > MAX_PHASE_ERROR)   ||
           (fabs(this->dAmpError) > MAX_AMPLITUDE_ERROR) ||
           (this->dDistortionError > MAX_DISTORTION) )
   {
      this->dCurrentAmplFactor = this->dProposedAmplFactor;
      this->iCurrentPhaseShiftIndex = this->iProposedPhaseShiftIndex;
      
      /***
        [[U ec := this->tDlDither::SetWaveform()]]
       ***/
      ec = this->SetWaveform();
      if (ec!=OK) 
         return Traceback(ec,"tDlDither::CalibrateDither");

      /***
        [[U ec := this->tDlDither::MeasureDither(NUM_WAVES_TO_AVG) ]]
       ***/
      ec = this->MeasureDither(NUM_WAVES_TO_AVG);

      if (ec!=OK)
         return Traceback(ec,"tDlDither::CalibrateDither");

      TraceMark(S_dlDither_COMPLETED_ANOTHER_MEASURE_DITHER);

      if (++iNumCalibCycles > MAX_CALIB_CYCLES) {
        GetErrorStream().Put(S_dlDither_DITHER_CALIBRATE_FAILED);
        return Traceback(S_dlDither_DITHER_CALIBRATE_FAILED,
                         "tDlDither::CalibrateDither");
      }


   }

   /* if we get here then we're done */

   for (i=0; i<this->siMaxDitherBufSize; i++)
      this->dDitherCorrTable[i] = this->dCollectDitherTable[i];

   this->bDoingCheckOrCalib = FALSE;

   TraceMark(S_dlDither_SUCCESSFULLY_COMPLETED_CALIBRATE_DITHER);

   return OK;

}  /* tDlDither::CalibrateDither() */


/******************************************************************
*
* ERROR_CODE tDlDither::CleanUp(ERROR_CODE ecCondition)
*     CleanUp is called by CalibrateDither and CheckDither command 
*     engines' Execute and Abort methods, to make sure things get back
*     to normal after the calibration or check.  ecCondition is our
*     way of telling whether to reset to the previous settings or just
*     close up shop.
*
* SPEC:
   [[X tDlDither::CleanUp_CB(ERROR_CODE ecCondition)]
   [ this invalid => UNDEFINED
   | this->bCalibrating == TRUE && (ecCondition != OK) =>
       this->bStoreDither := FALSE
       : semGive( this->siNextOffsetSem )
       : this->dDitherOutputTable[] := this->dSavedDitherOutputTable[]
       : this->dDitherCorrTable[] := this->dSavedDitherCorrTable[]
       : [[U tDlDither::LoadDitherWaveform(iOffset=0)]]
       : turn dither off or on, according to this->bWasDitherOn
       : this->bCalibrating := FALSE
       : return := ecCondition
   | true =>
       this->bStoreDither := FALSE
       : semGive( this->siNextOffsetSem )
       : [[U tDlDither::LoadDitherWaveform(iOffset=0)]]
       : turn dither off or on, according to this->bWasDitherOn
       : this->bCalibrating := FALSE
       : return := ecCondition
   ]]

* RETURNS:
*    OK
*
*****/
ERROR_CODE tDlDither::CleanUp(ERROR_CODE ecCondition)
{
   int i, ec;

   this->bDoingCheckOrCalib = FALSE;
   this->bStoreDither = FALSE;

   /***
     [[U ec := ::semGive(this->siNextOffsetSem)]]
    ***/
   ec = semGive( this->siNextOffsetSem );  
   if (ec!=OK)
      return Traceback(ec,"tDlDither::CleanUp",ecCondition);

   if ( (this->bCalibrating) && (ecCondition != OK))
   {
      for (i=0; i<this->siDitherBufSize; i++)
         this->dDitherCorrTable[i] = this->dSavedDitherCorrTable[i];
      for (i=0; i<this->siDitherBufSize; i++)
         this->dDitherOutputTable[i] = this->dSavedDitherOutputTable[i];
   }

   this->bCalibrating = FALSE;        /* (works, even if it already was) */

   /***
     [[U ec := this->tDlDither::LoadDitherWaveform(0)]]
    ***/
   ec = this->LoadDitherWaveform(0);
   if (ec!=OK)
      return Traceback(ec,"tDlDither::CleanUp",ecCondition);

   if (this->bWasDitherOn)
   /***
     [[U ec := this->tDlDither::DitherOn()]]
    ***/
      ec = this->DitherOn();
   else
   /***
     [[U ec := this->tDlDither::DitherOff()]]
    ***/
      ec = this->DitherOff();

   if (ec!=OK)
      return Traceback(ec,"tDlDither::CleanUp",ecCondition);

   return ecCondition;

}  /* tDlDither::CleanUp() */



/******************************************************************
*
* ERROR_CODE tDlDither::SpoutMeasurementDiags(void)
*
* SPEC:
   [[D tDlDither::SpoutMeasurementDiags()]
   [ this invalid => UNDEFINED
   | this->siDitherBufSize==0 => UNDEFINED
   | true =>
       Brad said I didn't have to write spec for this fn
   ]]

* RETURNS:
*    OK
*
*****/
ERROR_CODE tDlDither::SpoutMeasurementDiags(void)
{
   int i;
   double dSum = 0.0,  dSumSq = 0.0;
   double dMean, dMeanSq;
   double dVar;

   if (this->bStoreDither)
   {
      TraceMsg(tmt_WARN,S_dlDither_CANNOT_DO_DIAGS_WHILE_MEASURING,
                             "tDlDither::SpoutMeasurementDiags");
      return OK;
   }

   printf("tDlServoTaskPeriod = %lu\n", this->tDlServoTaskPeriod);
   printf("bDoingCheckOrCalib = %d\n",(int)this->bDoingCheckOrCalib);
   printf("bDithering = %d\n",(int)this->bDithering);

         /* these vars are static since they reflect board properties */
   printf("stDitherPeriod = %lu\n", this->stDitherPeriod);
   printf("sdDitherSamplePeriod = %f\n", this->sdDitherSamplePeriod);
   printf("siDitherBufSize = %d\n", this->siDitherBufSize);
   printf("siDoubleDitherBufSize = %d\n", this->siDoubleDitherBufSize);
   printf("siMaxDitherBufSize = %d\n", this->siMaxDitherBufSize);
   printf("siDoubleMaxDitherBufSize = %d\n", this->siDoubleMaxDitherBufSize);
   printf("siMaxWaveOffset = %d\n", this->siMaxWaveOffset);
   printf("siSweepBufSize = %d\n", this->siSweepBufSize);

         /* amplitude however can be different for different DLs */
   printf("dDitherAmplitude = %f\n", this->dDitherAmplitude);

            /* vars used during dither calibrate operation */

   printf("bCalibrating = %d\n",(int)this->bCalibrating);
   printf("bWasDitherOn = %d\n",(int)this->bWasDitherOn);
   printf("bStoreDither = %d\n",(int)this->bStoreDither);

   printf("iWaveOffset = %d\n",this->iWaveOffset);
   printf("iWavesToAvg = %d\n",this->iWavesToAvg);

   printf("dCurrentAmplFactor = %f\n", this->dCurrentAmplFactor);
   printf("dProposedAmplFactor = %f\n", this->dProposedAmplFactor);
   printf("iCurrentPhaseShiftIndex = %d\n",this->iCurrentPhaseShiftIndex);
   printf("iProposedPhaseShiftIndex = %d\n",this->iProposedPhaseShiftIndex);

   printf("\nValues from most recent sweep:\n");
   for (i=0; i<this->siSweepBufSize; i++)
      printf("[%2d]:\t| %f\t| %5s\n", i, this->dSweepSampleArray[i],
			(this->bGotSampleArray[i] ? "TRUE " : "FALSE") );
   printf("iSweepSampleCount = %d\n",this->iSweepSampleCount);
   printf("iSweepErrorCount = %d\n",this->iSweepErrorCount);
   printf("iTotalSweepCount = %d\n",this->iTotalSweepCount);

   printf("\nCollected wave samples, with std dev for each sample:\n");

     /* col headers */
   printf("\nIndex\t| Wave        \t| Dither Units   \t| Measured    \t| StdDev\n"); 

   for (i=0; i<this->siDitherBufSize; i++)
   {
      printf("[%3d]:\t| %f\t| %f\t| %f\t| %f\n", i, this->dIdealTriangleWave[i],
                  this->dDitherOutputTable[i],
                  this->dCollectDitherTable[i], this->dCollectSqDitherTable[i]);

      dSum += this->dCollectDitherTable[i];
   }

   return OK;

}  /* tDlDither::SpoutMeasurementDiags() */



/******************************************************************
* ERROR_CODE DitherDiags()
*
*    compiled as an "extern C" fn; exists only to call 
*            tDlDither::SpoutMeasurementDiags
*
*****/
ERROR_CODE DitherDiags( int iDlNumber )
{
  GIZMO_TYPE gtTemp;
  int iDlId;

   GetEmbeddedSystemGizmoInfo (GT_DELAY_LINE, iDlNumber, gtTemp, iDlId);
   tDlGizmo *pGizmo = (tDlGizmo*) GetGizmoPointer(iDlId);

   


   if (pGizmo!=NULL) pGizmo->SpoutDitherDiags();

   return OK;
}
        



/******************************************************************
*
* double tDlDither::ComputeMeasuredWaveformAmplitude
*
* SPEC:
   [[X tDlDither::ComputeMeasuredWaveformAmplitude (void)
   [ this invalid => UNDEFINED
   ]]

* RETURNS:
*   Amplitude of the waveform in this->dCollectDitherTable
*
*/

#if 0

double tDlDither::ComputeMeasuredWaveformAmplitude(void)
{
  int i;
  double dSum = 0.0;

  // The net sum of the absolute value of a triangle wave is
  // the wave's amplitude/2 (average value) * the number of samples
  
  for (i=0; i<this->siDitherBufSize; i++)
    dSum += fabs(this->dCollectDitherTable[i]);


  // Multiply average value by 4.  This accounts for the factor
  // of two to convert from average value to amplitude and then
  // another factor of two to convert from amplitude to peak-to-
  // peak amplitude.
  
  return dSum / ((double) this->siDitherBufSize) * 4.0;
}

#endif
