/* $RCSfile: dlGizmo.cc,v $ - Type definitions for Delay Line gizmo stuff  */
/* CM Info:   
Header: /home/isit/cvsroot/isit/sys/dl/dlGizmo/dlGizmo.cc,v 1.26.2.2 1997/09/04 05:44:13 mckenn Exp $
*/
/* Richard Johnson - Jet Propulsion Laboratory */

/*
modification history
--------------------
04a,24Sep97,rlj	     converted to new architecture
02b,21May97,beh      Merged BB1 and BB2 into Quic
02a,07May97,pci      changed Initialize to constructor
01a,29Oct96,eamck    written
*/
 
/*
DESCRIPTION
  DL Gizmo init code, etc.
*/
   

#include "isit.h"
#include "dlGizmo.h"
#include "dl/dlIoSet.h"
#include "dl/dlControllerSet.h"
#include "dl/dlCommandEngine.h"


const int CiDlDataDumpPriority = 78;
const int CiDlJitterHpBlockSize = DL_DATA_BUF_SIZE / 2;


/******************************************************************************
*
* tDlGizmo::tDlGizmo (const tDlGizmoInfo &GizmoInfo,
                      tDlCompleteIoSet& CompleteIoSet,
                      tDlCompleteControllerSet& CompleteControllerSet)
*
* DESCRIPTION: 
  Constructs a delay line gizmo using configuration contained in GizmoInfo.
  Specific delay line complete I/O and controller sets are defined externally 
  as well (usually by a derived class).  This constructor sets up all common
  delay line functionality including command processing, servo configuration,
  dither, and target generation. (NOTE: Target generation is really part of 
  the I/O and should eventually be moved to a common I/O object) 
  The constructor fails upon first error which is recored in ecInitError.
  Upon completetion, only the target generator task is running.  Other tasks
  will not start until a subseqent call to tDlGizmo::StartTasks. (usually 
  called from the derived delay line class constructor)
*
  [[U tDlGizmo::tDlGizmo()]]
*
*/
 
tDlGizmo::tDlGizmo (const tDlGizmoInfo &GizmoInfo) :

  /****
  [[U tRealtimeGizmo::tRealtimeGizmo (GizmoInfo)]]
  ****/
  tRealtimeGizmo (GizmoInfo),
  
  /****
  [[U this->MutexDelayLine.tMutex::tMutex (void) ]]
  ****/
  MutexDelayLine (),

  iNumLaserPasses (GizmoInfo.PhysConst.iLaserNumPasses),
  iNumStellarPasses (GizmoInfo.PhysConst.iStellarNumPasses)
  
{

  ERROR_CODE ec;


  if (this->GetInitError() != OK) {
    TraceMsg (tmt_ERROR,this->GetInitError(),"tDlGizmo::tDlGizmo");
    return;
  }
  
  /****
  [[ temp := U GizmoInfo.tGizmoInfo::GetGizmoType (void) ]]
  [ temp != DT_DL_BB1 &&
    temp != DT_DL_BB2 &&
    temp != DT_DL_GSU => 
      this->ecInitError := S_dlgizmo_GIZMO_INFO_NOT_DL
      : console := console + appropriate error messages
  | true => no response]
  ****/
  
  if ( GizmoInfo.GetGizmoType () != GT_DL_BB1 &&
       GizmoInfo.GetGizmoType () != GT_DL_BB2 &&
       GizmoInfo.GetGizmoType () != GT_DL_GSU ) {

    this->SetInitError (S_dlgizmo_GIZMO_INFO_NOT_DL);
    TraceMsg (tmt_ERROR,S_dlgizmo_GIZMO_INFO_NOT_DL,
              "tDlGizmo::tDlGizmo" );
    return;
  }
  

  // Sanity Checks until we can better automate the controller setup.

  /*****
  [[U temp := GizmoInfo.tRealtimeGizmoInfo::GetNumberOfServoModes (void) ]]
  [ temp != 8 => 
     this->ecInitError := S_dlgizmo_INVALID_NUMBER_OF_SERVO_MODES
      : console := console + appropriate error messages
  | true => no response]
  ****/
  if ( GizmoInfo.GetNumberOfServoModes () != 8) {

    this->SetInitError (S_dlgizmo_INVALID_NUMBER_OF_SERVO_MODES);
    TraceMsg (tmt_ERROR,S_dlgizmo_INVALID_NUMBER_OF_SERVO_MODES,
              "tDlGizmo::tDlGizmo" );
      return;
    }
    
  /*****
  [[U temp := GizmoInfo.tRealtimeGizmoInfo::GetNumberOfServoPartitions (void) ]]
  [ temp != 3 => 
     this->ecInitError := S_dlgizmo_INVALID_NUMBER_OF_SERVO_PARTITIONS
      : console := console + appropriate error messages
  | true => no response]
  ****/
  if (GizmoInfo.GetNumberOfServoPartitions () != 3 ) {

    this->SetInitError (S_dlgizmo_INVALID_NUMBER_OF_SERVO_PARTITIONS);
    TraceMsg (tmt_ERROR,S_dlgizmo_INVALID_NUMBER_OF_SERVO_PARTITIONS,
              "tDlGizmo::tDlGizmo" );
    return;
  }

  // Select Off Mode

  /****
  [[U ec := this->Servo.tServo::SelectMode (DL_OFF) ]] 
  ****/ 
  ec = this->Servo.SelectMode (DSM_OFF);
    
  if (ec != OK) {
    TraceMsg (tmt_ERROR,ec,"tDLGizmo::tDLGizmo" );
    this->SetInitError (ec);
    return;
  }

}



/******************************************************************************
*
* tDlGizmo::~tDlGizmo (void)
*
* DESCRIPTION: 
  This method is incomplete and should delete all dynamically allocated
  objects, delete all periodic and event driven tasks owned by the gizmo etc.
  For now it only destroys the tDlLimits object.
*
*
[[U tDlGizmo::~tDlGizmo()]]*
*/
 
tDlGizmo::~tDlGizmo (void)
{
  TraceMark(S_dlGizmo_ENTERING_DL_GIZMO_DESTRUCTOR);
   

  TraceMark(S_dlGizmo_EXITING_DL_GIZMO_DESTRUCTOR);
}



/******************************************************************************
*
* tDlGizmo::Restart (void)
*
* DESCRIPTION: 
  This method is incomplete and should restart all objects owned by the 
  delay line, resume all event driven tasks, and enable all periodic tasks.
*
[[U  tDlGizmo::Restart (void) ]]
****/

ERROR_CODE tDlGizmo::Restart () 
{
  return this->tRealtimeGizmo::Restart ();
}

/******************************************************************************
*
* tDlGizmo::Shutdown (void)
*
* DESCRIPTION: 
  This method is incomplete and should shutdown all objects owned by the 
  delay line, suspend all event driven tasks, and disable all periodic tasks.
*
[[U  tDlGizmo::Shutdown (void) ]]
****/

ERROR_CODE tDlGizmo::Shutdown () 
{
  return this->tRealtimeGizmo::Shutdown ();
}

/******************************************************************************
*
* tDlGizmo::IsOk (void)
*
* DESCRIPTION: 
  This method is incomplete and should check the IsOk status of all objects 
  owned by the delay line, and return FALSE if any are malfunctioning.
  Eventually this method will be smarter and be able to make diagnosis and 
  prognosis of the particular malfunction.
*
[[U  tDlGizmo::IsOk (void) ]]
****/

BOOL tDlGizmo::IsOk () 
{
  return this->tRealtimeGizmo::IsOk ();
}


/******************************************************************************
*
* tDlGizmo::operator<< (void)
*
* DESCRIPTION: 
  This method is incomplete and should modify the stream to contain diagnostic
  information about all all objects owned by the delay line, as well as 
  information about each event driven and periodic task that it owns.
*
[[U ::operator_extractor (Stream,Gizmo) ]]
****/

ostream& operator<< (ostream& Stream, const tDlGizmo& Gizmo) 
{
  return Stream << (tRealtimeGizmo&) Gizmo;
}


ERROR_CODE tDlGizmo::SpoutDitherDiags () 
{
  tDlCompleteIoSet& IoSet = this->GetCompleteIoSet ();
  
  return  IoSet.Laser.SpoutMeasurementDiags ();
}


/******************************************************************************
*
* tDlGizmo::CollectStatus (void)
*
* DESCRIPTION: 
  Updates all status information that is common to all delay lines.
  Currently it is the responsibility of the derived delay line gizmos to
  actually send the status to the status stream. 
*
[[U  tDlGizmo::CollectStatus (void) ]]
****/

ERROR_CODE tDlGizmo::CollectStatus (void) {


  ERROR_CODE ec;
  LASER_VALUE laLaserValue, laTargetValue, laErrorValue;
  PZT_DOUBLE pdPztPos;
  PZT_VALUE pvPztVelocity;
  VC_VALUE vvVc1;
  MOTOR_VALUE mvMotorVelocity;

 
   // Update the status block common to all delay lines....

  /****
  [[U Status := this->tDlGizmo::GetStatus (void) ]]
  ****/
  tDlStatus& Status = this->GetStatus ();

  /****
  [[U IoSet := this->tDlGizmo::GetCompleteIoSet (void)]]
  ****/
  tDlCompleteIoSet &IoSet= this->GetCompleteIoSet ();  
 
  /****
  [[U Status.tTime := ::GetTime (void) ]] 
  ****/
  Status.tTime = ::GetTime ();
  

  /****
  [[U tempMode := this->Servo.tServo::GetMode (void) ]]
  
  [[U Status.tDlStatus::UpdateServoState (tempMode) ]] 
  ****/
  Status.UpdateServoState ( (DL_SERVO_MODE) this->Servo.GetMode ());

  /****
  [[U bTemp1 := IoSet.DlLimits.tDlLimits::IsOnFrontSoftLimit (void) ]] 

  [[U bTemp2 :=  IoSet.DlLimits.tDlLimits::IsOnFrontHardLimit (void) ]] 

  [[U bTemp3 :=  IoSet.DlLimits.tDlLimits::IsOnBackSoftLimit (void) ]] 

  [[U bTem4 := IoSet.DlLimits.tDlLimits::IsOnBackHardLimit (void) ]] 

  [[U bTemp5 :=  IoSet.DlLimits.tDlLimits::IsOnHomeSensor (void) ]] 

  [[U Status.tDlStatus::UpdateLimits (bTemp1,bTemp2,bTemp3,bTemp4,bTemp5) ]] 
  ****/
  Status.UpdateLimits (IoSet.DlLimits.IsOnFrontSoftLimit (),
                      IoSet.DlLimits.IsOnFrontHardLimit (),
                      IoSet.DlLimits.IsOnBackSoftLimit (),
                      IoSet.DlLimits.IsOnBackHardLimit (),
                      IoSet.DlLimits.IsOnHomeSensor () );

  /****
  [[U TempMod := this->tDlGizmo::GetLaserModulation (void) ]] 

  [[U Status.tDlStatus::UpdateModulation (TempMod) ]] 
  ****/
  Status.UpdateModulation (this->GetLaserModulation () );

  /****
  [[U Temp := IoSet.Laser.tLaserMetrologySensor::DitherCalibrationInfo (void) ]] 
  ****/
  tDitherCalibInfo Temp = IoSet.Laser.DitherCalibrationInfo ();
  

  // Convert amplitude from laser opd to stellar opd in nanometers 
  
  /****
  [[U Temp.dAmplitude := tDlGizmo::LaserOpdToStellarOpd (Temp.dAmplitude) ]] 
  ****/
  Temp.dAmplitude = this->LaserOpdToStellarOpd (Temp.dAmplitude) * 1.0e9; // in nanometers
  
  /****
  [[U Status.tDlStatus::UpdateDitherCalibration (Temp) ]] 
  ****/
  Status.UpdateDitherCalibration (Temp);

  /****
  [[U TempCheck := IoSet.Laser.tLaserMetrologySensor::DoingCheckOrCalib (void) ]] 

  [[U Status.tDlStatus::UpdateDitherCheckOrCalib (TempCheck) ]] 
  ****/
  Status.UpdateDitherCheckOrCalib (IoSet.Laser.DoingCheckOrCalib () );


  laLaserValue     = (LASER_VALUE) IoSet.Laser.Value ();
  pdPztPos         = (PZT_DOUBLE) IoSet.PztPosition.Update (0);
  pvPztVelocity    = (PZT_VALUE) IoSet.Pzt.Value ();
  vvVc1            = (VC_VALUE) IoSet.Vc1.Value ();
  mvMotorVelocity  = (MOTOR_VALUE) IoSet.Motor.Velocity ();
  laTargetValue = IoSet.LaserTarget.Value ();
  laErrorValue  = this->ServoError ();

  /****
  [[U Status.tDlStatus::UpdateServoInfo (laLaserValue, pdPztPos, pvPztVelocity,
                         vvVc1, mvMotorVelocity, laTargetValue, laErrorValue,
                         0) ]]
  ****/ 
  Status.UpdateServoInfo (laLaserValue, pdPztPos, pvPztVelocity,
                          vvVc1, mvMotorVelocity, laTargetValue, laErrorValue, 0);

  /* These are progress indicators and are updated by the
   * things making the progress as they go, so we don't need
   * to update them here.
   *
     Status.UpdateCommandStatus(dcsNewCommandStatus);
   */


  
  /* The following are things that are currently updated by other
   * objects elsewhere as they compute the new values.  However, these
   * should eventually be changed so that the tDlStatusObject pulls the
   * proper value from them, rather than having them asynchronously write
   * into the tDlStatusObject.
   *
     Status.UpdateJitter(dJitter, dHighPassJitter);
   */  

  return OK;
}


/******************************************************************************
*
* tDlGizmo::Dispatch (const tIsitCmd& Command)
*
* DESCRIPTION: 
  Calls appropriate delay line gizmo command methods based on the specific 
  type of command in Command.  This method should only be called by 
  tGizmo::ProcessCommand as virtual method call. Any errors are returned to 
  the caller.
*
[[U  tDlGizmo::Dispatch (Command) ]]
****/

ERROR_CODE tDlGizmo::Dispatch (const tIsitCmd& Command) {

  switch (Command.Command) {

      /****
      [[U  tDlGizmo::Ping (void) ]]
      ****/
    case DL_CMD_PING : return this->Ping ();

      /****
      [[U tDlGizmo::DataRecordOnOff()]]
      ****/
    case DL_CMD_DATA_RECORD_ON_OFF :
                      return this->DataRecordOnOff ((tDlDataRecordCmd &) Command);

      /****
      [[U  tDlGizmo::Off (void) ]]
      ****/
    case DL_CMD_OFF : return this->Off ();

      /****
      [[U  tDlGizmo::Track (void) ]]
      ****/
    case DL_CMD_TRACK : return this->Track ();

      /****
      [[U  tDlGizmo::Front (void) ]]
      ****/
    case DL_CMD_FRONT : return this->Front ();

      /****
      [[U  tDlGizmo::Front (void) ]]
      ****/
    case DL_CMD_BACK : return this->Back ();

      /****
      [[U  tDlGizmo::ClearLaserCounter (void) ]]
      ****/
    case DL_CMD_CLR_LASER_CTR : return this->ClearLaserCounter ();

      /****
      [[U  tDlGizmo::User1 (void) ]]
      ****/
    case DL_CMD_USER1 : return this->User1 ();

      /****
      [[U  tDlGizmo::User2 (void) ]]
      ****/
    case DL_CMD_USER2 : return this->User2 ();

      /****
      [[U  tDlGizmo::HomeCheck (void) ]]
      ****/
    case DL_CMD_HOME_CHECK : return this->HomeCheck ();

      /****
      [[U  tDlGizmo::HomeSet (void) ]]
      ****/
    case DL_CMD_HOME_SET : return this->HomeSet ();

      /****
      [[U  tDlGizmo::DitherOnOff (Command) ]]
      ****/
    case DL_CMD_DITHER_ON : return this->DitherOnOff (
                                 (tDlDitherOnCmd&) Command);

      /****
      [[U  tDlGizmo::SetActiveTarget (Command) ]]
      ****/
    case DL_CMD_SELECT_TARGET : return this->SetActiveTarget (
                                         (tDlSetActiveTargetCmd&) Command);
      /****
      [[U  tDlGizmo::CalibrateDither (void) ]]
      ****/
    case DL_CMD_DITHER_CALIBRATE : return this->CalibrateDither ();

      /****
      [[U  tDlGizmo::CheckDither (void) ]]
      ****/
    case DL_CMD_DITHER_CHECK : return this->CheckDither ();

      /****
      [[U  tDlGizmo::SetDither (void) ]]
      ****/
    case DL_CMD_SET_DITHER : return this->SetDither (
                                 (tDlSetDitherCmd&) Command);

      /****
      [[U  tDlGizmo::SetModulation (Command) ]]
      ****/
    case DL_CMD_MODULATE : return this->SetModulation (
                                   (tDlSetModulationCmd&) Command);

//      /****
//      [[U  tDlGizmo::SetServoParameters (Command) ]]
//      ****/
//    case DL_CMD_CHNG_SERVO_PARAM : return this->SetServoParameters (
//                                    (tDlSetServoParametersCmd&) Command);

    default : return S_dlgizmo_UNRECOGNIZED_COMMAND;
  };

  return OK;
}



/******************************************************************************
* tDlGizmo::StartTasks (const tDlGizmoInfo::GizmoInfo,tDlCompleteIoSet& CompleteIoSet)
*
* DESCRIPTION: 
  Starts all tasks necessary for command and status processing, as well as 
  all servo tasks.
*
[[U  tDlGizmo::StartTasks (GizmoInfo) ]]
****/
ERROR_CODE tDlGizmo::StartTasks (const tDlGizmoInfo& GizmoInfo,
                                 tDlCompleteIoSet& CompleteIoSet) 
{
  
  ERROR_CODE ec;


  /****
  [[U ec := tRealtimeGizmo::StartTasks (GizmoInfo) ]] 
  ****/ 
  ec = tRealtimeGizmo::StartTasks (GizmoInfo);

  if (ec != OK) return Traceback (ec,"tDlGizmo::tDlGizmo" );

  return OK;
}


ERROR_CODE tDlGizmo::SetOffGroup (tControllerGroup& OffGroup) 
{

  ERROR_CODE ec;

  /****
  [[U ec := this->ControllerSet.tControllerSet::AssignControllerGroup (
      DSM_OFF, OffGroup) ]]
  ****/
  ec = this->ControllerSet.AssignControllerGroup (DSM_OFF, OffGroup);

  if (ec != OK) return Traceback (ec,"tDlGizmo::SetOffGroup");
  return OK;
}
  
ERROR_CODE tDlGizmo::SetFrontGroup (tControllerGroup& FrontGroup)
{
  ERROR_CODE ec;

  /****
  [[U ec := this->ControllerSet.tControllerSet::AssignControllerGroup (
      DSM_FRONT, FrontGroup) ]]
  ****/
  ec = this->ControllerSet.AssignControllerGroup (DSM_FRONT,FrontGroup);

  if (ec != OK) return Traceback (ec,"tDlGizmo::SetFrontGroup");
  return OK;
}

ERROR_CODE tDlGizmo::SetBackGroup (tControllerGroup& BackGroup)
{
  ERROR_CODE ec;

  /****
  [[U ec := this->ControllerSet.tControllerSet::AssignControllerGroup (
      DSM_BACK, BackControllerGroup) ]]
  ****/
  ec = this->ControllerSet.AssignControllerGroup (DSM_BACK, BackGroup);
 
  if (ec != OK) return Traceback (ec,"tDlGizmo::SetBackGroup");
  return OK;
}

    

ERROR_CODE tDlGizmo::SetFrontCreepGroup (tControllerGroup& FrontCreepGroup)
{
  ERROR_CODE ec;

  /****
  [[U ec := this->ControllerSet.tControllerSet::AssignControllerGroup (
      DSM_FRONT_CREEP, FrontCreepGroup) ]]
  ****/
  ec = this->ControllerSet.AssignControllerGroup (DSM_CREEPFRONT, FrontCreepGroup);

  if (ec != OK) return Traceback (ec,"tDlGizmo::SetFrontCreepGroup");
  return OK;
}


    
ERROR_CODE tDlGizmo::SetBackCreepGroup (tControllerGroup& BackCreepGroup)
{
  ERROR_CODE ec;

  /****
  [[U ec := this->ControllerSet.tControllerSet::AssignControllerGroup (
      DSM_BACK_CREEP, BackCreepGroup) ]]
  ****/
  ec = this->ControllerSet.AssignControllerGroup (DSM_CREEPBACK, BackCreepGroup);

  if (ec != OK) return Traceback (ec,"tDlGizmo::SetBackCreepGroup");
  return OK;
}
    
ERROR_CODE tDlGizmo::SetTrackGroup (tControllerGroup& TrackGroup)
{
  ERROR_CODE ec;

  /****
  [[U ec := this->ControllerSet.tControllerSet::AssignControllerGroup (
      DSM_TRACK, TrackGroup) ]]
  ****/
  ec = this->ControllerSet.AssignControllerGroup (DSM_TRACK, TrackGroup);

  if (ec != OK) return Traceback (ec,"tDlGizmo::SetTrackGroup");
  return OK;
}

    
ERROR_CODE tDlGizmo::SetUser1Group (tControllerGroup& User1Group)
{
  ERROR_CODE ec;

  /****
  [[U ec := this->ControllerSet.tControllerSet::AssignControllerGroup (
      DSM_USER1, User1Group) ]]
  ****/
  ec = this->ControllerSet.AssignControllerGroup (DSM_USER1, User1Group);

  if (ec != OK) return Traceback (ec,"tDlGizmo::SetUser1Group");
  return OK;
}

    
ERROR_CODE tDlGizmo::SetUser2Group (tControllerGroup& User2Group)
{

  ERROR_CODE ec;

  /****
  [[U ec := this->ControllerSet.tControllerSet::AssignControllerGroup (
      DSM_USER2, User2Group) ]]
  ****/
  ec = this->ControllerSet.AssignControllerGroup (DSM_USER2, User2Group);

  if (ec != OK) return Traceback (ec,"tDlGizmo::SetUser2Group");
  return OK;
}




/******************************************************************************
*
* tDlGizmo::Ping (void)
*
* DESCRIPTION: 
  Sends a message to the ErrorStream indicating that the delay line is alive.
  This command requires no mutual exclusion object and runs to completion 
  before exiting.
*
[[U  tDlGizmo::Ping (void) ]]
****/

ERROR_CODE tDlGizmo::Ping () 
{ 
   /****
   [[U  tGizmo::SendError_ec (S_dlgizmo_DELAY_LINE_ALIVE) ]]
   ****/
   return SendError (S_dlgizmo_DELAY_LINE_ALIVE);
}
  


/******************************************************************************
*
* tDlGizmo::DataRecordOnOff (void)
*
[[U  tDlGizmo::DataRecordOnOff (void) ]]
****/

ERROR_CODE tDlGizmo::DataRecordOnOff (const tDlDataRecordCmd &Command) 
{
  this->GetControllerData().LaserDataStore.SetSuspendData(Command.bSuspendData);

  if(Command.bSuspendData)
    return SendError (S_dlgizmo_DATA_RECORDING_OFF);
  else 
    return SendError (S_dlgizmo_DATA_RECORDING_ON);
}


/******************************************************************************
*
* tDlGizmo::Off (void)
*
* DESCRIPTION: 
  Causes the delay line cart to cease previous operation and eventually come
  to a halt.
  This command requires the DelayLine mutex and spawns the appropriate
  command engine to perform off functionality. 
*
  [[U  tDlGizmo::Off (void) ]]
****/

ERROR_CODE tDlGizmo::Off () 
{

  tDlOffCommandEngine* pCommandEngine;

  /****
  [[U pCommandEngine->tDlOffCommandEngine::
        tDlOffCommandEngine (this->MutexDelayLine,*this) ]]
  ****/
  pCommandEngine = new tDlOffCommandEngine (this->MutexDelayLine,*this);

  /****
  [[U ec := pCommandEngine.tMutexUser::Execute (void) ]]
  
  [[U tDlOffCommandEngine::ExecuteMethod (void) ]]
  ****/
  return pCommandEngine->Execute ();

}



/******************************************************************************
*
* tDlGizmo::Track (void)
*
* DESCRIPTION: 
  Causes the delay line cart to start tracking the sum of all active target 
  sources to minimize the rms jitter.   
  Tracking performance is affected by the current servo 
  parameters as well as the set of control laws used to perform the servo 
  calculations.
  The delay line will continue to track until another command requiring the 
  delay line cart is issued.  If the target tracked goes beyond the limits of 
  the cart travel, the cart will stop physically tracking the target, but will
  continue to mathematically track the target until it comes within the delay 
  line travel.

  This command requires the DelayLine mutex and spawns the appropriate
  command engine to perform track functionality. 
*
  [[U  tDlGizmo::Track (void) ]]
****/

ERROR_CODE tDlGizmo::Track () 
{

  tDlTrackCommandEngine* pCommandEngine;

  /****
  [[U pCommandEngine->tDlTrackCommandEngine::
          tDlTrackCommandEngine (this->MutexDelayLine,*this) ]]
  ****/
  pCommandEngine = new tDlTrackCommandEngine (this->MutexDelayLine,*this);

  /****
  [[U ec := pCommandEngine.tMutexUser::Execute (void) ]]
  
  [[U tDlTrackCommandEngine::ExecuteMethod (void) ]]
  ****/
  return pCommandEngine->Execute ();

}
 

/******************************************************************************
*
* tDlGizmo::Front (void)
*
* DESCRIPTION: 
  Causes the delay line cart to begin moving to the front of the delay line
  track.  The delay line will continue to move forward until it 
  1) reaches a hard or soft front limit or
  2) is interrupted by another command requiring the delay line hardware.

  This command requires the DelayLine mutex and spawns the appropriate
  command engine to perform front functionality. 
*
  [[U  tDlGizmo::Front (void) ]]
****/

ERROR_CODE tDlGizmo::Front () 
{

  tDlFrontCommandEngine* pCommandEngine;

  /****
  [[U pCommandEngine->tDlFrontCommandEngine::
          tDlFrontCommandEngine (this->MutexDelayLine,*this) ]]
  ****/
  pCommandEngine = new tDlFrontCommandEngine (this->MutexDelayLine,*this);

  /****
  [[U ec := pCommandEngine.tMutexUser::Execute (void) ]]
  
  [[U tDlFrontCommandEngine::ExecuteMethod (void) ]]
  ****/
  return pCommandEngine->Execute ();

}


/******************************************************************************
*
* tDlGizmo::Back (void)
*
* DESCRIPTION: 
  Causes the delay line cart to begin moving to the back of the delay line
  track.  The delay line will continue to move backwards until it 
  1) reaches a hard or soft back limit or
  2) is interrupted by another command requiring the delay line hardware.

  This command requires the DelayLine mutex and spawns the appropriate
  command engine to perform back functionality. 
*
  [[U  tDlGizmo::Back (void) ]]
****/

ERROR_CODE tDlGizmo::Back () 
{

  tDlBackCommandEngine* pCommandEngine;

  /****
  [[U pCommandEngine->tDlBackCommandEngine::
          tDlBackCommandEngine (this->MutexDelayLine,*this) ]]
  ****/
  pCommandEngine = new tDlBackCommandEngine (this->MutexDelayLine,*this);

  /****
  [[U ec := pCommandEngine.tMutexUser::Execute (void) ]]
  
  [[U tDlBackCommandEngine::ExecuteMethod (void) ]]
  ****/
  return pCommandEngine->Execute ();

}


/******************************************************************************
*
* tDlGizmo::ClearLaserCounter (void)
*
* DESCRIPTION: 
  Resets the integer portion of the delay line laser counter register which 
  describes the current delay line position.

  This command requires the DelayLine mutex and spawns the appropriate
  command engine to perform clear laser counter functionality. 

  [[U  tDlGizmo::ClearLaserCounter (void) ]]
****/

ERROR_CODE tDlGizmo::ClearLaserCounter () 
{

  tDlClearLaserCounterCommandEngine* pCommandEngine;

  /****
  [[U pCommandEngine->tDlClearLaserCounterCommandEngine::
          tDlClearLaserCounterCommandEngine (this->MutexDelayLine,*this) ]]
  ****/
  pCommandEngine = new tDlClearLaserCounterCommandEngine (this->MutexDelayLine,
                                                          *this);

  /****
  [[U ec := pCommandEngine.tMutexUser::Execute (void) ]]
  
  [[U tDlClearLaserCounterCommandEngine::ExecuteMethod (void) ]]
  ****/
  return pCommandEngine->Execute ();

}


/******************************************************************************
*
* tDlGizmo::User1 (void)
*
* DESCRIPTION: 
  Causes the delay line cart to start tracking the sum of all active target 
  sources to minimize the rms jitter.   
  Tracking performance is affected by the current servo 
  parameters as well as the set of control laws used to perform the servo 
  calculations.
  The delay line will continue to track until another command requiring the 
  delay line cart is issued.  If the target tracked goes beyond the limits of 
  the cart travel, the cart will stop physically tracking the target, but will
  continue to mathematically track the target until it comes within the delay 
  line travel.

  This command requires the DelayLine mutex and spawns the appropriate
  command engine to perform track functionality. 
*
  [[U  tDlGizmo::User1 (void) ]]
****/

ERROR_CODE tDlGizmo::User1 () 
{

  tDlUser1CommandEngine* pCommandEngine;

  /****
  [[U pCommandEngine->tDlUser1CommandEngine::
          tDlUser1CommandEngine (this->MutexDelayLine,*this) ]]
  ****/
  pCommandEngine = new tDlUser1CommandEngine (this->MutexDelayLine,*this);

  /****
  [[U ec := pCommandEngine.tMutexUser::Execute (void) ]]
  
  [[U tDlUser1CommandEngine::ExecuteMethod (void) ]]
  ****/
  return pCommandEngine->Execute ();

}  


/******************************************************************************
*
* tDlGizmo::User2 (void)
*
* DESCRIPTION: 
  Causes the delay line cart to start tracking the sum of all active target 
  sources to minimize the rms jitter.   
  Tracking performance is affected by the current servo 
  parameters as well as the set of control laws used to perform the servo 
  calculations.
  The delay line will continue to track until another command requiring the 
  delay line cart is issued.  If the target tracked goes beyond the limits of 
  the cart travel, the cart will stop physically tracking the target, but will
  continue to mathematically track the target until it comes within the delay 
  line travel.

  This command requires the DelayLine mutex and spawns the appropriate
  command engine to perform track functionality. 
*
  [[U  tDlGizmo::User2 (void) ]]
****/

ERROR_CODE tDlGizmo::User2 () 
{

  tDlUser2CommandEngine* pCommandEngine;

  /****
  [[U pCommandEngine->tDlUser2CommandEngine::
          tDlUser2CommandEngine (this->MutexDelayLine,*this) ]]
  ****/
  pCommandEngine = new tDlUser2CommandEngine (this->MutexDelayLine,*this);

  /****
  [[U ec := pCommandEngine.tMutexUser::Execute (void) ]]
  
  [[U tDlUser2CommandEngine::ExecuteMethod (void) ]]
  ****/
  return pCommandEngine->Execute ();

}  


/******************************************************************************
*
* tDlGizmo::HomeCheck (void)
*
* DESCRIPTION: 
  Causes the delay line cart to search for the home sensor in a way
  that it always approaches the home sensor in a predetermined direction.
  This command allows the user to check the position of the home sensor.
  The delay line continues to move until
  1) Reaches the home sensor
  2) interupted by another command requiring the delay line hardware
  3) it traverses the track without finding the home sensor.
 
  This command requires the DelayLine mutex and spawns the appropriate
  command engine to perform home check functionality. 
*
  [[U  tDlGizmo::HomeCheck (void) ]]
****/

ERROR_CODE tDlGizmo::HomeCheck () 
{

  tDlHomeCheckCommandEngine* pCommandEngine;

  /****
  [[U pCommandEngine->tDlHomeCheckCommandEngine::
          tDlHomeCheckCommandEngine (this->MutexDelayLine,*this) ]]
  ****/
  pCommandEngine = new tDlHomeCheckCommandEngine (this->MutexDelayLine,*this);

  /****
  [[U ec := pCommandEngine.tMutexUser::Execute (void) ]]
  
  [[U tDlHomeCheckCommandEngine::ExecuteMethod (void) ]]
  ****/
  return pCommandEngine->Execute ();

}


/******************************************************************************
*
* tDlGizmo::HomeSet (void)
*
* DESCRIPTION: 
  Causes the delay line cart to search for the home sensor in a way
  that it always approaches the home sensor in a predetermined direction.
  In addition to finding the home sensor, this command also clears the 
  laser counter once the home sensor is found.
  The delay line continues to move until
  1) Reaches the home sensor
  2) interupted by another command requiring the delay line hardware
  3) it traverses the track without finding the home sensor.
 
  This command requires the DelayLine mutex and spawns the appropriate
  command engine to perform home set functionality. 
*
  [[U  tDlGizmo::HomeSet (void) ]]
****/

ERROR_CODE tDlGizmo::HomeSet () 
{

  tDlHomeSetCommandEngine* pCommandEngine;

  /****
  [[U pCommandEngine->tDlHomeSetCommandEngine::
          tDlHomeSetCommandEngine (this->MutexDelayLine,*this) ]]
  ****/
  pCommandEngine = new tDlHomeSetCommandEngine (this->MutexDelayLine,*this);

  /****
  [[U ec := pCommandEngine.tMutexUser::Execute (void) ]]
  
  [[U tDlHomeSetCommandEngine::ExecuteMethod (void) ]]
  ****/
  return pCommandEngine->Execute ();

}



/******************************************************************************
*
* tDlGizmo::DitherOnOff (const tDlDitherOnCmd& Command)
*
* DESCRIPTION: 
  This command turns the pzt dither on and off based on the boolean value set
  in Command.
 
  This command does not require a mutex object, and runs to completion before
  exiting.
*
  [[U tDlGizmo::DitherOnOff (Command) ]]
****/

ERROR_CODE tDlGizmo::DitherOnOff (const tDlDitherOnCmd& Command)
{
  ERROR_CODE ec;  
  
  tDlCompleteIoSet& IoSet = this->GetCompleteIoSet ();
  
  if (Command.bTurnDitherOn == TRUE)
  {

     /****
     [[U ec := IoSet.Laser.tLaserMetrologySensor::EnableDither (void) ]]
     ****/      
     ec = IoSet.Laser.EnableDither ();
     
     /****
     [[U Status := this->tDlGizmo::GetStatus (void) ]]
     
     [[U Status.tDlStatus::UpdateCommandStatus (DLCS_TURN_DITHER_ON_DONE) ]]
     ****/
     this->GetStatus ().UpdateCommandStatus (DLCS_TURN_DITHER_ON_DONE);

     /****
     [[U this->tGizmo::SendError_ec (
           S_dlSetDitheringCommand_TURN_DITHER_ON_DONE) ]]
     ****/
     this->SendError (S_dlSetDitheringCommand_TURN_DITHER_ON_DONE);
  }

  else 

 {
     /****
     [[U ec := IoSet.Laser.tLaserMetrologySensor::DisableDither (void) ]]
     ****/      
     ec = IoSet.Laser.DisableDither ();

   if (ec == OK) {

     /****
     [[U Status := this->tDlGizmo::GetStatus (void) ]]
      
     [[U Status.tDlStatus::UpdateCommandStatus (DLCS_TURN_DITHER_OFF_DONE) ]]
     ****/
     this->GetStatus ().UpdateCommandStatus (DLCS_TURN_DITHER_OFF_DONE);

     /****
     [[U this->tGizmo::SendError_ec (
           S_dlSetDitheringCommand_TURN_DITHER_OFF_DONE) ]]
     ****/
     this->SendError (S_dlSetDitheringCommand_TURN_DITHER_OFF_DONE);
   }
 }
  return ec;
}


/******************************************************************************
*
* tDlGizmo::SetActiveTarget (const tDlSetActiveTargetCmd& Command)
*
* DESCRIPTION: 
  Selects/deselects appropriate target source.  All selected target sources
  are subsequently used by the delay line target generator to compute the
  total target used during tracking.
 
  This command does not require a mutex object, and runs to completion before
  exiting.
*
  [[U tDlGizmo::SetActiveTarget (const tDlSetActiveTargetCmd& Command) ]]
****/

ERROR_CODE tDlGizmo::SetActiveTarget (const tDlSetActiveTargetCmd& Command)
{
  ERROR_CODE ec;
  
  tDlCompleteIoSet& IoSet = this->GetCompleteIoSet ();
  
  ec = IoSet.TargetGenerator.SetActiveTarget (Command.ttWhichTarget,
			                Command.bTargetSetActive);
  
  if (ec == OK) {

    /****
    [[U Status := this->tDlGizmo::GetStatus (void) ]]
     
    [[U Status.tDlStatus::UpdateCommandStatus (DLCS_SELECT_TARGET_DONE) ]]
    ****/
    this->GetStatus ().UpdateCommandStatus (DLCS_SELECT_TARGET_DONE);

    /****
    [[U this->tGizmo::SendError_ec (
         S_setActiveTargetCommand_SELECT_TARGET_COMPLETE) ]]
    ****/
    this->SendError (S_setActiveTargetCommand_SELECT_TARGET_COMPLETE);
  }

  return ec;  
}


/******************************************************************************
*
* tDlGizmo::CalibrateDither (void)
*
* DESCRIPTION: 
  Performs a calibration of the dither waveform by moving the delay line to 
  the front of the track, and then measures the response of the delay line pzt 
  to a triangle wave input.  From these measurements, a waveform is produced 
  which will cause the pzt to transcribe a nearly perfect triangle wave.

  This command requires the DelayLine mutex and spawns the appropriate
  command engine to perform home set functionality. 
  
*
  [[U tDlGizmo::CalibrateDither (void) ]]
****/

ERROR_CODE tDlGizmo::CalibrateDither ()
{

  tDlDitherCalibrateCommandEngine* pCommandEngine;

  /****
  [[U pCommandEngine->tDlDitherCalibrateCommandEngine::
          tDlDitherCalibrateCommandEngine (this->MutexDelayLine,*this) ]]
  ****/
  pCommandEngine = new tDlDitherCalibrateCommandEngine (this->MutexDelayLine,
                                                        *this);

  /****
  [[U ec := pCommandEngine->tMutexUser::Execute (void) ]]
  
  [[U pCommandEngine->tDlDitherCalibrateCommandEngine::ExecuteMethod (void) ]]
  ****/
  return pCommandEngine->Execute ();

}


/******************************************************************************
*
* tDlGizmo::CheckDither (void)
*
* DESCRIPTION: 
  Checks the calibrated dither waveform to measure the quality of the triangle
  wave. This is done by moving the delay line to the front of the track, 
  turn on dither, and measure the pzt response.

  This command does not require a mutex object, and runs to completion before
  exiting.  
*
  [[U tDlGizmo::CheckDither (void) ]]
****/

ERROR_CODE tDlGizmo::CheckDither ()
{

  tDlDitherCheckCommandEngine* pCommandEngine;

  /****
  [[U pCommandEngine->tDlDitherCheckCommandEngine::
          tDlDitherCheckCommandEngine (this->MutexDelayLine,*this) ]]
  ****/
  pCommandEngine = new tDlDitherCheckCommandEngine (this->MutexDelayLine,
                                                    *this);

  /****
  [[U ec := pCommandEngine->tMutexUser::Execute (void) ]]
  
  [[U pCommandEngine->tDlDitherCheckCommandEngine::ExecuteMethod (void) ]]
  ****/
  return pCommandEngine->Execute ();

}


/******************************************************************************
*
* tDlGizmo::SetDither (const tDlSetDitherCmd& Command)
*
* DESCRIPTION: 
  Sets the amplitude and period of the dither waveform.

  This command does not require a mutex object, and runs to completion before
  exiting.  
*
  [[U tDlGizmo::SetDither (Command) ]]
****/

ERROR_CODE tDlGizmo::SetDither (const tDlSetDitherCmd& Command)
{
  ERROR_CODE ec;
  double dAmplitudeInLaserCounts;
  tDlCompleteIoSet& IoSet = this->GetCompleteIoSet ();
  
  /****
  [[U dTemp := tDlGizmo::StellarOpdToLaserOpd (Command.dAmplitude) ]]
  
  [[U ec := IoSet.Laser.tLaserMetrologySensor::SetDitherParameters (dTemp , Command.dPeriod) ]]
  ****/
  ec =  IoSet.Laser.SetDitherParameters (StellarOpdToLaserOpd (Command.dAmplitude) , 
          (int) Command.dPeriod);

  if (ec == OK) {
    
    /****
    [[U Status := this->tDlGizmo::GetStatus (void) ]]
     
    [[U Status.tDlStatus::UpdateCommandStatus (DLCS_SET_DITHER_COMPLETE) ]]
    ****/
    this->GetStatus ().UpdateCommandStatus (DLCS_SET_DITHER_COMPLETE);

    /****
    [[U this->tGizmo::SendError_ec (
          S_dlSetDitheringCommand_SET_DITHER_COMPLETE) ]]
    ****/
    this->SendError (S_dlSetDitheringCommand_SET_DITHER_COMPLETE);
  }

  return ec;
}


/******************************************************************************
*
* tDlGizmo::SetModulation (const tDlSetModulationCmd& Command)
*
* DESCRIPTION: 
  Sets the modulation parameters of the laser channel used to measure
  delay line servo performance.

  This command does not require a mutex object, and runs to completion before
  exiting.  
*
  [[U tDlGizmo::SetModulation (Command) ]]
****/
   
ERROR_CODE tDlGizmo::SetModulation (const tDlSetModulationCmd& Command)
{
  ERROR_CODE ec;

  TraceMark(S_dlGizmo_ENTERING_SET_LASER_MODULATION);

  /*****
  [[U ModParams.tModulationParams::tModulationParams (
        Command.mtModulationType, Command.dFrequency,
        Command.dAmplitude, Command.dMeanVal, Command.tSyncTime)]]
  *****/

  tModulationParams ModParams (Command.mtModulationType, Command.dFrequency,
                               Command.dAmplitude, Command.dMeanVal,
                               Command.tSyncTime);
  
  /*****
  [[U ec := this->tDlGizmo::SetLaserModulation (ModParams) ]]
  ****/

  /****
  [[U this->CompleteIoSet.Laser.tLaserMetrologySensor::
          SetModulation (ModParams )]]
  ****/
  ec = this->GetCompleteIoSet ().Laser.SetModulationParameters (ModParams);

  /****
  [[U this->CompleteIoSet.Laser.tLaserMetrologySensor::EnableModulation () ]]
  ****/
  this->GetCompleteIoSet ().Laser.EnableModulation ();
  
  if (ec == OK) {

    /****
    [[U Status := this->tDlGizmo::GetStatus (void) ]]
     
    [[U Status.tDlStatus::UpdateCommandStatus (DLCS_SET_MODULATION_COMPLETE) ]]
    ****/
    this->GetStatus ().UpdateCommandStatus (DLCS_SET_MODULATION_COMPLETE);

    /****
    [[U this->tGizmo::SendError_ec (
         S_setModulationCommand_SET_MODULATION_COMPLETE) ]]
    ****/
    this->SendError (S_setModulationCommand_SET_MODULATION_COMPLETE);
  }

  return ec;  
}




/******************************************************************************
*
* tDlGizmo::GetLaserModulation (void)
*
* DESCRIPTION: 
  Returns the current laser modulation parameters.
*
  [[U tDlGizmo::GetLaserModulation (void)]]
*
*/

const tModulationParams &tDlGizmo::GetLaserModulation ()
{
  TraceMark(S_dlGizmo_ENTERING_GET_LASER_MODULATION);

  /****
  [[U this->CompleteIoSet.Laser.tModulatedSensor::GetModulation (void) ]]
  ****/
  return this->GetCompleteIoSet ().Laser.ModulationParameters ();
}












