/* $RCSfile: fringe.cc,v $ */
/* CM Info:
$Header: /home/chara/cvsroot/ople/sys/gizmo/ft/fringe.cc,v 1.1.1.1 1998/06/11 21:36:23 chara Exp $
*/

#include <math.h>
#include "isit.h"
#include "ft/fringe.h"
#include "ftError.h"

/******************************************************************
*
* class tFringe
*
*  DESCRIPTION:
   This class represents the fringe.  Its inputs are detector readings, and its
   outputs are  fringe statistics.  Messages are to be sent to this class at two 
   different rates:  AcceptCount() should be called for each detector reading,
   and Calculate() should be called at the end of every sweep (i.e., once for
   every 4 calls to AcceptCount()).  And for any given sweep, Calculate() 
   should be called before any of the other public methods, which return 
   data derived from that sweep.  It is assumed that these messages are 
   synchronized, such that for a given sweep, AcceptCount() is not called for 
   the last value of a sweep while Calculate() is processing.  It is also 
   assumed that sweeps are synchronized to the clock, such that each forward
   sweep starts at a time which is divisible by the dither period.  And wobble
   bin 0 occurs whenever the time is divisible by the wobble period.
****/ 


/******************************************************************
*
* tFringe::tFringe(int iSampleSize,int iWobbleBins,TIME tFTPeriod)
*
*  DESCRIPTION:
   The constructor.  iCountArray is initialized to zero to mitigate the effects
   of missing data in the first 8 APD readings.
*
[[U tFringe::tFringe(iSampleSize,iWobbleBins,tFTPeriod) ]]
****/ 
tFringe::tFringe(int iSampleSize,int iWobbleBins,TIME tFTPeriod) :
   iInputBuffer(0),
   PhotonsArray(new double[iSampleSize + 1]),
   PowerArray(new double[iSampleSize + 1]),
   iLastIndex(0),
   CiSampleSize(iSampleSize),
   WobbleSamplesArray(new unsigned long int[iWobbleBins]),
   WobblePhotonsArray(new double[iWobbleBins]),
   WobblePowerArray(new double[iWobbleBins]),
   PositionSamplesArray(new unsigned long int[iWobbleBins]),
   WobblePositionArray(new double[iWobbleBins]),
   CiWobbleBins(iWobbleBins),
   dDarkCurrent(0),
   CtReadingPeriod(MAX(tFTPeriod / QUADRANTS,1)) //avoid divide by zero
{
   int iSide,iQuadrant;

   for (iSide = 0; iSide < 2; iSide++) {
      for (iQuadrant = 0; iQuadrant < QUADRANTS; iQuadrant++) {
         iCountArray[iSide][iQuadrant] = 0;
      }
   }

   this->Reset();
}

/******************************************************************
*
* tFringe::tFringe(const tFringe& f)
*
*  DESCRIPTION:
   The default copy constructor is inadequate because of the member pointers.
   However, a copy constructor is unecessary for this class.  So it is declared
   private, and it is not properly defined.
*
****/ 
tFringe::tFringe(const tFringe& f) {}


/******************************************************************
*
* tFringe::~tFringe()
*
*  DESCRIPTION:
   The destructor deletes all of the dynamically allocated memory.   
*
[[U tFringe::~tFringe() ]]
****/ 
tFringe::~tFringe()
{
   delete[] this->PhotonsArray;
   delete[] this->PowerArray; 
   delete[] this->WobbleSamplesArray;
   delete[] this->WobblePhotonsArray;
   delete[] this->WobblePowerArray;
   delete[] this->PositionSamplesArray;
   delete[] this->WobblePositionArray;
}
   

/******************************************************************
*
* tFringe::SetDarkCurrent(double dDark)
*
*  DESCRIPTION:
   This methods sets the dark current. 
*
[[U tFringe::SetDarkCurrent(dDark) ]]
****/ 
ERROR_CODE tFringe::SetDarkCurrent(double dDark)
{
   this->dDarkCurrent = dDark;
   return OK;
}


/******************************************************************
*
* tFringe::Reset()
*
*  DESCRIPTION:
   This method resets all of the averages to zero samples.  
*
[[U tFringe::Reset() ]]
****/ 
ERROR_CODE tFringe::Reset()
{
   int iBin;
   int iSample;
   
   this->dTotalPhotons = 0;
   this->dTotalPower = 0;
   this->iSampleCount= 0;
   for (iBin = 0; iBin < CiWobbleBins; iBin++) {
      this->WobbleSamplesArray[iBin] = 0;
      this->WobblePhotonsArray[iBin] = 0;
      this->WobblePowerArray[iBin] = 0;
      this->PositionSamplesArray[iBin] = 0;
      this->WobblePositionArray[iBin] = 0;
   }

   //in case Visibility2(), etc. sent before data available
   for (iSample = 0; iSample < CiSampleSize; iSample++) {
      PhotonsArray[iSample] = 0;
      PowerArray[iSample] = 0;
   }

   tLastTime = 0;
   bNewData = FALSE;
   this->bDiscardCounts = TRUE;

   return OK;
}


/******************************************************************
*
* tFringe::DiscardCounts ()
*
*  DESCRIPTION:
*
[[U tFringe::DiscardCounts () ]]
****/ 
ERROR_CODE tFringe::DiscardCounts ()
{

  this->bDiscardCounts = TRUE;
  return OK;

}

/******************************************************************
*
* tFringe::StoreOnly ()
*
*  DESCRIPTION:
*
[[U tFringe::StoreOnly () ]]
****/ 
ERROR_CODE tFringe::StoreOnly ()
{

  this->bStoreOnly = TRUE;
  return OK;

}
/******************************************************************
*
* tFringe::CheckCounts ()
*
*  DESCRIPTION:*
[[U tFringe::CheckCounts () ]]
****/ 
ERROR_CODE tFringe::CheckCounts ()
{

  this->bStoreOnly = FALSE;
  return OK;

}


/******************************************************************
*
* tFringe::AcceptCount(TIME tTime,int iCount)
*
*  DESCRIPTION:
   The method saves the specified photon count.  It only saves the last 8 such 
   counts (in a double buffer).  Once a 4-count PZT sweep is complete, the input
   side of the double buffer is toggled.
*[[U tFringe::AcceptCount(tTime,iCount) ]]
****/ 
ERROR_CODE tFringe::AcceptCount(TIME tTime,int iCount)
{  
   int iReading;
   int iQuadrant;   

   //check for missing data
   if (this->bStoreOnly == FALSE && tLastTime != 0 && tTime != tLastTime + CtReadingPeriod) {
      TraceMsg(tmt_ERROR,S_ft_MISSING_APD_DATA,"tFringe::AcceptCount()");
      this->bDiscardCounts = TRUE;
   }
   tLastTime = tTime;

   //reading number
   iReading = (int)(tTime / CtReadingPeriod);
   
   //quadrant number of sweep
   iQuadrant = (iReading + QUADRANTS - 1) % QUADRANTS;

/*
if (iInputBuffer < 0 || iInputBuffer >1 ||
    iQuadrant < 0 || iQuadrant > 3) {
  logMsg("Buf, Quad: %d %d\n",iInputBuffer,iCount);
  //return OK;
}
*/

   //store data
   iCountArray[iInputBuffer][iQuadrant] = iCount;
   
   //if last quadrant of sweep
   if (iQuadrant == QUADRANTS - 1) {
      //check for unprocessed data in the other side of the buffer
      if (this->bStoreOnly == FALSE && bNewData) {
         TraceMsg(tmt_ERROR,S_ft_UNPROCESSED_SWEEP,"tFringe::AcceptCount()");
      }
      
      //switch to other side of the buffer for input
      if (this->bDiscardCounts == FALSE)            // Swap buffers only if not discarding data
         this->iInputBuffer = !this->iInputBuffer;
      else this->bDiscardCounts = FALSE;    // Discarding data is a one shot mechanism

      bNewData = TRUE;
   }
   return OK;
}

/******************************************************************
*
* tFringe::Calculate(TIME tTime)
*
*  DESCRIPTION:
   The method calculates N ("Photons") and NUM ("Power") for the last sweep.  
   The last CiSampleSize such numbers are saved (in PhotonsArray and PowerArray), 
   so that they may be removed from the respective averages when they have 
   become sufficiently mature.  The method also maintains several totals of 
   these two quantities, as follows:

      dTotalPhotons:		total of all entries in PhotonsArray
      dTotalPower:		total of all entries in PowerArray
      WobblePhotonsArray:	totals of *all* N's since last 
                                 WobbleVisibility2() message, sorted by 
                                 Wobble bin
      WobblePowerArray:		totals of *all* NUM's since last 
                                 WobbleVisibility2() message, sorted by 
                                 Wobble bin
*[[U tFringe::Calculate(tTime) ]]
****/ 
ERROR_CODE tFringe::Calculate(TIME tTime)
{  
   int iSweep;
   int iBin;
   int iX,iY;
   int iQuadrant;

   iLastIndex = (iLastIndex + 1) % CiSampleSize;

   //sweep number
   iSweep = (int)(tTime / (4 * CtReadingPeriod));
   
   //check for unprocessed data
   if (!bNewData) {
      TraceMsg(tmt_ERROR,S_ft_OVERPROCESSED_SWEEP,"tFringe::Calculate()");
   }

   //copy last sweep in DL order
   for (iQuadrant = 0; iQuadrant < QUADRANTS; iQuadrant++) {
      if (iSweep % 2 == 0) {
         iLastCountArray[iQuadrant] = iCountArray[!iInputBuffer][iQuadrant];
      } 
      else {
         iLastCountArray[iQuadrant] = 
            iCountArray[!iInputBuffer][QUADRANTS - 1 - iQuadrant];
      }
   }
      
   //data has been copied
   bNewData = FALSE;

   //remove excess samples from the totals
   if (this->iSampleCount == this->CiSampleSize) {
      this->dTotalPhotons -=this->PhotonsArray[iLastIndex];
      this->dTotalPower -= this->PowerArray[iLastIndex];
   }
   
   //N = A + B + C + D - Dark
   this->PhotonsArray[iLastIndex] = -this->dDarkCurrent;
   for (iQuadrant = 0; iQuadrant < QUADRANTS; iQuadrant++) {
      this->PhotonsArray[iLastIndex] += this->iLastCountArray[iQuadrant];
   }

   /***
   [[U iX := tFringe::InPhaseComponent() ]]
   [[U iY := tFringe::QuadratureComponent() ]]
   ***/
   iX = this->InPhaseComponent();
   iY = this->QuadratureComponent();

   //NUM = X^2 + Y^2 - N
   this->PowerArray[iLastIndex] = iX * iX + iY * iY - 
                                    this->PhotonsArray[iLastIndex];

   //increment total samples, min iSampleSize
   this->iSampleCount = MIN(this->iSampleCount++,this->CiSampleSize);

   //avoid divide by 0
   if (this->CiWobbleBins != 0) {
      //add new samples to totals
      iBin = tTime / (this->CtReadingPeriod * QUADRANTS) % this->CiWobbleBins; 
      this->WobblePhotonsArray[iBin] += this->PhotonsArray[iLastIndex];
      this->WobblePowerArray[iBin] += this->PowerArray[iLastIndex];
      this->WobbleSamplesArray[iBin]++;
      this->dTotalPhotons +=this->PhotonsArray[iLastIndex];
      this->dTotalPower += this->PowerArray[iLastIndex];
   }
   return OK;
}
   

/******************************************************************
*
* tFringe::AcceptWobble(TIME tTime,double dPosition)
*
*  DESCRIPTION:
   This method add the specified wobble position to the sum of wobble positions
   for the appropriate wobble bin.  
*
[[U tFringe::AcceptWobble(tTime,dPosition) ]]
****/ 
ERROR_CODE tFringe::AcceptWobble(TIME tTime,double dPosition)
{
   int iBin;   
   
   //avoid divide by 0
   if (this->CiWobbleBins == 0) return OK;

   iBin = tTime / (this->CtReadingPeriod * QUADRANTS) % this->CiWobbleBins; 
   this->WobblePositionArray[iBin] += dPosition;
   this->PositionSamplesArray[iBin]++;
   
   return OK;  
}



/******************************************************************
*
* tFringe::Photons()
*
*  DESCRIPTION:
   This method returns N for the last sweep. 
*
[[U tFringe::Photons() ]]
****/ 
double tFringe::Photons() const
{
   return this->PhotonsArray[iLastIndex];
}

double tFringe::A () const
{
 return this->iLastCountArray[0];
}

double tFringe::B () const
{
 return this->iLastCountArray[1];
}

double tFringe::C () const
{
 return this->iLastCountArray[2];
}


double tFringe::D () const
{
 return this->iLastCountArray[3];
}



/******************************************************************
*
* tFringe::Visibility2()
*
*  DESCRIPTION:
   This method returns V^2 for the last sweep.  
*
[[U tFringe::Visibility2() ]]
****/ 
double tFringe::Visibility2() const
{
   /***
   [[U tFringe::CalculateVisibility2(this->PowerArray[this->iLastIndex],
                                     this->PhotonsArray[this->iLastIndex]) ]]
   ***/
   return this->CalculateVisibility2(this->PowerArray[this->iLastIndex],
                                     this->PhotonsArray[this->iLastIndex]);
}


/******************************************************************
*
* tFringe::Snr2()
*
*  DESCRIPTION:
   This method returns SNR^2 for the last sweep.   
*
[[U tFringe::Snr2() ]]
****/ 
double tFringe::Snr2() const
{
   /***
   [[U tFringe::CalculateSnr2(this->PowerArray[this->iLastIndex],
                              this->PhotonsArray[this->iLastIndex]) ]]
   ***/
   return this->CalculateSnr2(this->PowerArray[this->iLastIndex],
                              this->PhotonsArray[this->iLastIndex]);
}


/******************************************************************
*
* tFringe::Phase()
*
*  DESCRIPTION:
   This method returns the simple phase for the last sweep.  The value is
   in radians, and in the range [-PI,PI].  
*
[[U tFringe::Phase() ]]
****/ 
double tFringe::Phase() const
{
   //avoid domain error
   if (this->QuadratureComponent() == 0.0 && this->InPhaseComponent() == 0.0)
      return 0.0;

   /***
   [[U X := tFringe::InPhaseComponent() ]]
   [[U Y := tFringe::QuadratureComponent() ]]
   [return := atan2 (Y,X) ]
   ***/
   return atan2 (this->QuadratureComponent (), this->InPhaseComponent () );

}


/******************************************************************
*
* tFringe::EstVisibility2()
*
*  DESCRIPTION:
   This method returns the estimated V^2, calculated from averages of the last
   iSampleCount N and NUM.  If the sample count is 0, the method returns 0.0. 
*
[[U tFringe::EstVisibility2() ]]
****/ 
double tFringe::EstVisibility2() const
{
   //avoid divide by zero
   if (this->iSampleCount == 0) return 0.0;
   
   /***
   [[U tFringe::CalculateVisibility2((this->dTotalPower / this->iSampleCount),
                                     (this->dTotalPhotons / this->iSampleCount)) ]]
   ***/
   return this->CalculateVisibility2(this->dTotalPower / this->iSampleCount,
                                     this->dTotalPhotons / this->iSampleCount);
}


/******************************************************************
*
* tFringe::EstSnr2()
*
*  DESCRIPTION:
   This method returns the estimated SNR^2, calculated from averages of the last
   iSampleCount N and NUM.  If the sample count is 0, the method returns 0.0. 
*
[[U tFringe::EstSnr2() ]]
****/ 
double tFringe::EstSnr2() const
{

   //avoid divide by zero
   if (this->iSampleCount == 0) return 0.0;
   
   /***
   [[U tFringe::CalculateSnr2((this->dTotalPower / this->iSampleCount),
                              (this->dTotalPhotons / this->iSampleCount)) ]]
   ***/
   return this->CalculateSnr2(this->dTotalPower / this->iSampleCount,
                              this->dTotalPhotons / this->iSampleCount);
}


/******************************************************************
*
* tFringe::WobbleVisibility2(double V2Array[],
                             double PositionArray[],
                             int iSize,
                             double* pSnr2)
*
*  DESCRIPTION:
   This method returns the estimate V^2 for each wobble bin, each one calculated
   from the averages of N and NUM for that bin, since the last call to this 
   method or to Reset().  It then resets all averages to zero samples.  If any
   sample count is 0, then 0.0 will be returned for that bin.  
*
[[U tFringe::WobbleVisibility2(V2Array,PositionArray,iSize,pSnr2) ]]
****/ 
ERROR_CODE tFringe::WobbleVisibility2(double V2Array[],
                                      double PositionArray[],
                                      int iSize,
                                      double* pSnr2) const
{
   double dPower,dPhotons;
   double dSumPhotons = 0;
   double dSumPower = 0;
   int iBin;
   
   if (iSize < this->CiWobbleBins) return S_ft_ARRAY_TOO_SMALL;
   if (CiWobbleBins == 0) return OK;
      
   //for each wobble bin
   for (iBin = 0; iBin < CiWobbleBins; iBin++) {

      //avoid divide by zero (results will all be zero)
      if (this->WobbleSamplesArray[iBin] == 0) 
         this->WobbleSamplesArray[iBin] = 1;
      if (this->PositionSamplesArray[iBin] == 0)
         this->PositionSamplesArray[iBin] = 1;
         
      dPower = this->WobblePowerArray[iBin] / this->WobbleSamplesArray[iBin];

      //get average Photons   
      dPhotons = this->WobblePhotonsArray[iBin] / this->WobbleSamplesArray[iBin];

      //get visibility squared
      /***
      [[U tFringe::CalculateVisibility2(dPower,dPhotons) ]]
      ***/
      V2Array[iBin] = this->CalculateVisibility2(dPower,dPhotons);

      //get average position
      PositionArray[iBin] = this->WobblePositionArray[iBin] / 
                            this->PositionSamplesArray[iBin];          

      //increment sums of averages                      
      dSumPower   += dPower;
      dSumPhotons += dPhotons;
         
      //reset wobble sums and sample numbers
      this->WobblePowerArray[iBin] = 0;
      this->WobblePhotonsArray[iBin] = 0;
      this->WobblePositionArray[iBin] = 0;
      this->WobbleSamplesArray[iBin] = 0;
      this->PositionSamplesArray[iBin] = 0;
   }

   //get SNR squared
   /***
   [[U tFringe::CalculateSnr2((dSumPhotons / this->CiWobbleBins),
                              (dSumPower / this->CiWobbleBins)) ]]
   ***/
   *pSnr2 = this->CalculateSnr2(dSumPhotons / this->CiWobbleBins,
                                dSumPower / this->CiWobbleBins);
      
   return OK;
}


/******************************************************************
*
* tFringe::CalculateVisibility2(double dPower,double dPhotons)
*
*  DESCRIPTION:
   This method returns V^2 for the specified NUM and N.  If dPhotons is 0, the
   method returns 0.0.
*
[[U tFringe::CalculateVisibility2(dPower,dPhotons) ]]
****/ 
double tFringe::CalculateVisibility2(double dPower,double dPhotons) const
{
   //avoid divide by 0
   if (dPhotons == 0) return 0.0;
   
   return (PI * PI / 2) * dPower / (dPhotons * dPhotons);
}


/******************************************************************
*
* tFringe::CalculateSnr2(double dPower,double dPhotons)
*
*  DESCRIPTION:
   This method returns SNR^2 for the specified NUM and N.  If dPhotons is 0, 
   the method returns 0.0. 
*
[[U tFringe::CalculateSnr2(dPower,dPhotons) ]]
****/ 
double tFringe::CalculateSnr2(double dPower,double dPhotons) const
{
   //avoid divide by 0
   if (dPhotons == 0) return 0.0;
   
   return 2 * dPower / dPhotons;
}

/******************************************************************
*
* tFringe::InPhaseComponent()
*
*  DESCRIPTION:
   This method returns the X component for the last sweep.  
*
[[U tFringe::InPhaseComponent() ]]
****/ 
int tFringe::InPhaseComponent() const
{
      //X = B - D
      return this->iLastCountArray[1] - this->iLastCountArray[3];
}


/******************************************************************
*
* tFringe::QuadratureComponent()
*
*  DESCRIPTION:
   This method returns the X component for the last sweep.  
*
[[U tFringe::QuadratureComponent() ]]
****/ 
int tFringe::QuadratureComponent() const
{
      //Y = A - C
      return this->iLastCountArray[0] - this->iLastCountArray[2];
}
