/* $RCSfile: stController.cc,v $ */
/* CM Info:   
$Header: /home/chara/cvsroot/ople/sys/gizmo/st/stController.cc,v 1.1.1.1 1998/06/11 21:36:24 chara Exp $
*/
/* Richard Johnson - Jet Propulsion Laboratory */

/*
modification history
--------------------
01b  22Sep97,eamck  taken over
01a  27Aug97,rlj    written
*/

#include "st/stController.h"
#include "st/stIoSet.h"
#include "ipcid.h"
#include "stGizmoInfo.h"
#include "stTypes.h"
#include "stError.h"
#include "servo.h"
#include "imageStatus.h"
#include "image.h"
#include "matrix.h"
#include "multiActuator.h"
#include "pidLibF.h"
#include "lagLib.h"
#include "ipcDblBufferAccessor.h"

const int CiStDataDumpPriority = 78;
const int CiStDataBufferSize = 100;     // #elements in data bag


/*******************************************************************
*
* tStControllerData::tStControllerData 
*
*  Constructs tStControllerData object.
*
* SPEC:
     [[U tStControllerData::tStControllerData(StarImageStatusToUse,
                                              StarPosDataTag,
                                              FsmPosDataTag,
                                              StarImageDataTag) ]]
*
* RETURNS:
*    nothing
*
****/

tStControllerData::tStControllerData(tImageStatus& StarImageStatusToUse,
                                      const tTypeTag& StarPosDataTag, 
                                      const tTypeTag& FsmPosDataTag, 
                                      const tTypeTag& StarImageDataTag) :
  /**** 
    [[U this->StarPositionData.tDataStore::tDataStore(CiStDataDumpPriority,
              CiStDataBufferSize,StarPosDataTag) ]]
  *****/
  StarPositionData(CiStDataDumpPriority,CiStDataBufferSize,StarPosDataTag),

  /**** 
    [[U this->FsmPositionData.tDataStore::tDataStore(CiStDataDumpPriority,
              CiStDataBufferSize,FsmPosDataTag) ]]
  *****/
  FsmPositionData(CiStDataDumpPriority,CiStDataBufferSize,FsmPosDataTag),

  /**** 
    [[U this->StarImageData.tDataStore::tDataStore(CiStDataDumpPriority,
              CiStDataBufferSize,StarImageDataTag) ]]
  *****/
  StarImageData(CiStDataDumpPriority,CiStDataBufferSize,StarImageDataTag),

  /**** 
     [ this->StarImageStatus := StarImageStatusToUse ]
  *****/
  StarImageStatus(StarImageStatusToUse) 
{
   TraceMark(S_stgizmo_ST_CONTROLLER_DATA_CONSTRUCTOR);
}


/*******************************************************************
*
* tStController::tStController
*
*  Constructs tStController object.
*
* SPEC:
     [[U tStController::tStController(CompleteIoSet, ControllerDataToUse) ]]
*
* RETURNS:
*    nothing
*
****/

tStController::tStController(tStCompleteIoSet& CompleteIoSet,
                             tStControllerData& ControllerDataToUse) :

  /**** 
     [ this->ControllerData := ControllerDataToUse ]
  *****/
  ControllerData(ControllerDataToUse),
  
  /**** 
     [ this->StarSensor := CompleteIoSet.tStCompleteIoSet::StarSensor ]
  *****/
  StarSensor(CompleteIoSet.StarSensor) 
{
   TraceMark(S_stgizmo_ST_CONTROLLER_CONSTRUCTOR);
}


/*******************************************************************
*
* tStController::~tStController
*
*  Destroys tStController object.
*
* SPEC:
     [[U tStController::~tStController(void) ]]
*
* RETURNS:
*    nothing
*
****/

tStController::~tStController(void)
{
   TraceMark(S_stgizmo_ST_CONTROLLER_DESTRUCTOR);
}

/*******************************************************************
*
* tStController::Calculate
*
*  Calculate function for all Star Tracker modes.  Called by servo
*     task at whatever rate it is running.  Reads in an image from
*     the "star sensor" object, which is the vehicle for receiving
*     images from the Fringe Tracker Gizmo.  Calculates the centroid
*     of intensity in the image, which is presumed to be the position
*     of the star.  Stores coordinates of centroid in the controller
*     data block. 
*
* SPEC:
     [[U tStController::Calculate(tServoTime) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/

ERROR_CODE tStController::Calculate(TIME tServoTime) 
{
  ERROR_CODE ec;
  double dUpdateError;

  tPoint StarPosition;
  
   TraceMark(S_stgizmo_ENTERING_ST_CONTROLLER_CALCULATE);

   TraceMark(S_stgizmo_UPDATE_STAR_SENSOR);

  // Get an image from Fringe Tracker, cache in StarSensor object
  /**** 
    [[U dUpdateError := this->StarSensor.tStarSensor::Update(tServoTime) ]]
  *****/
   
  dUpdateError = this->StarSensor.Update(tServoTime);
  if (dUpdateError != 0.0)
     return Traceback(S_stgizmo_ERROR_UPDATING_STAR_SENSOR,
              "tStTrackController::Calculate",tServoTime);
              
   
  // Get reference to image cached in StarSensor object
  /**** 
    [[U StarImage := this->StarSensor.tStarSensor::GetStarImage(void) ]]
  *****/
  tImage& StarImage = this->StarSensor.GetStarImage();


   TraceMark(S_stgizmo_CALCULATE_CENTROID);

  // Calculate the centroid of intensity in the image
  /**** 
    [[U tStController::CalculateCentroid(StarImage,StarPosition) ]]
  *****/
  this->CalculateCentroid(StarImage,this->ControllerData.StarPosition);
 
   TraceMark(S_stgizmo_UPDATE_CONTROLLER_DATA_BLOCK);

  /**** 
    [[U ec := 
       this->ControllerData.StarImageStatus.tImageStatus::Update_t(StarImage)]]
  *****/
  ec = this->ControllerData.StarImageStatus.Update(StarImage); 

  if (ec!=OK)
     return Traceback(ec,"tStTrackController::Calculate");


   TraceMark(S_stgizmo_LEAVING_ST_CONTROLLER_CALCULATE);

  return OK;

}


/*******************************************************************
*
* tStController::CalculateCentroid
*
*  Calculates centroid of intensity in image Image, stores coordinates
*      in StarPosition.  
*      Note: this algorithm should be re-examined in a future increment 
*      when data quality and customer requirements are better known.
*
* SPEC:
    [[U tStController::CalculateCentroid(Image, StarPosition) ]]
*
* RETURNS:
*    nothing
*
****/

void tStController::CalculateCentroid(const tImage&  Image, 
                                      tPoint& StarPosition) 
{
  int x,y, iHSize, iVSize;
  int iTotalIntensity = 0;
  int iWeightedXSum = 0, iWeightedYSum = 0;
  int PixelValue;
  double dStarXPosition, dStarYPosition;

   TraceMark(S_stgizmo_ENTERING_CALCULATE_CENTROID);

  Image.GetDimensions(iHSize,iVSize);
 
  for (x = 0; x < iHSize; x++) {
    for (y = 0; y < iVSize; y++) {
        PixelValue = Image.GetPixel(x,y);
        iTotalIntensity += PixelValue;
        iWeightedXSum += x * PixelValue;
        iWeightedYSum += y * PixelValue;
    }
  }
   
  if (iTotalIntensity <= 0)
  {
     dStarXPosition = 0.0;
     dStarYPosition = 0.0;
  }
  else
  {
     dStarXPosition = ((double)iWeightedXSum / (double)iTotalIntensity) 
                  - ((double)(iHSize-1)/2.0);
     dStarYPosition = ((double)iWeightedYSum / (double)iTotalIntensity) 
                  - ((double)(iVSize-1)/2.0);
  }

  StarPosition.x() = dStarXPosition;
  StarPosition.y() = dStarYPosition;

   TraceMark(S_stgizmo_LEAVING_CALCULATE_CENTROID);

} 



/*******************************************************************
*
* tStTrackController::tStTrackController
*
*  Constructs tStTrackController.
*
* SPEC:
    [[U tStTrackController::tStTrackController(CompleteIoSet,
                       ControllerDataToUse,tServoPeriod) ]]
*
* RETURNS:
*    nothing
*
****/

tStTrackController::tStTrackController(tStCompleteIoSet& CompleteIoSet,
                                       tStControllerData& ControllerDataToUse,
                                       TIME tServoPeriod) :
  
  /**** 
  [[U this->tStController::tStController(CompleteIoSet, ControllerDataToUse)]]
  *****/
   tStController(CompleteIoSet, ControllerDataToUse),

   Fsm(CompleteIoSet.Fsm),   
   bLagEnabled(FALSE),
   CtServoPeriod(tServoPeriod)///,

   ///Wap(CompleteIoSet.Wap)
{  
   TraceMark(S_stgizmo_ST_TRACK_CONTROLLER_CONSTRUCTOR);
}


/*******************************************************************
*
* tStTrackController::~tStTrackController
*
*  Destroys tStTrackController.
*
* SPEC:
    [[U tStTrackController::~tStTrackController(void) ]]
*
* RETURNS:
*    nothing
*
****/

tStTrackController::~tStTrackController(void) 
{
   TraceMark(S_stgizmo_ST_TRACK_CONTROLLER_DESTRUCTOR);
}


/*******************************************************************
*
* tStTrackController::Calculate
*
*  Calculate function for Star Tracker TRACK mode.  Called by servo
*     task at whatever rate it is running.  Reads in an image from
*     the "star sensor" object, which is the vehicle for receiving
*     images from the Fringe Tracker Gizmo.  Calculates the centroid
*     of intensity in the image, which is presumed to be the position
*     of the star.  Stores coordinates of centroid in the controller
*     data block.  [Next Incr: Send coords and image to RTE data stream.]
*
* SPEC:
     [[U tStTrackController::Calculate(tServoTime) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/

ERROR_CODE tStTrackController::Calculate(TIME tServoTime) 
{
  ERROR_CODE ec;
  double dUpdateError;
  tIpcPoint IpcOutput;

   TraceMark(S_stgizmo_ENTERING_ST_TRACK_CONTROLLER_CALCULATE);

   //calculate the star position
   /**** 
   [[U ec := this->tStController::Calculate(tServoTime)]]
   *****/
   ec = this->tStController::Calculate(tServoTime);
   if (ec != OK) return ec;

   //get the (negative) servo error
   /**** 
   [[U this->tStController::StarPosition()]]
   *****/
   ControllerData.FsmPosition = this->StarTarget - this->StarPosition();

   //if the lag compensator is enabled
   if (bLagEnabled) {
      //process with the lag compensator
      /**** 
      [[U this->Lag[X].tLagCompensator::Calculate(ControllerData.FsmPosition[X])]]
      [[U this->Lag[Y].tLagCompensator::Calculate(ControllerData.FsmPosition[Y])]]
      *****/
      ControllerData.FsmPosition[X] = this->Lag[X].Calculate(ControllerData.FsmPosition[X]);
      ControllerData.FsmPosition[Y] = this->Lag[Y].Calculate(ControllerData.FsmPosition[Y]);
   }
   
   //process with the PID controller
   /**** 
   [[U this->Pid[X].tPidDouble::Calculate(ControllerData.FsmPosition[X])]]
   [[U this->Pid[Y].tPidDouble::Calculate(ControllerData.FsmPosition[Y])]]
   *****/
   ControllerData.FsmPosition[X] = this->Pid[X].Calculate(ControllerData.FsmPosition[X]);
   ControllerData.FsmPosition[Y] = this->Pid[Y].Calculate(ControllerData.FsmPosition[Y]);

   //move the actuators according to the result
   /**** 
   [[U this->Fsm.tMultiActuator::Move(tServoTime,ControllerData.FsmPosition)]]
   *****/
   this->Fsm.Move(tServoTime,ControllerData.FsmPosition);
   
   //IPC output must be a simple structure
   IpcOutput.dX = ControllerData.FsmPosition.x();
   IpcOutput.dY = ControllerData.FsmPosition.y();
  
   //put commanded position in data stream
   /**** 
   [[U this->ControllerData.FsmPositionData.tDataStore::StoreData(IpcOutput)]]
   *****/
   this->ControllerData.FsmPositionData.StoreData(IpcOutput);
     
   //send commanded position to WAP
   /**** 
   [[U this->Wap.tIpcDblBufferAccessor::Put(IcpOutput) ]]
   *****/
   ///ec = Wap.Put(IpcOutput);
   ///if (ec != OK) return ec;

   TraceMark(S_stgizmo_LEAVING_ST_TRACK_CONTROLLER_CALCULATE);

  return OK;
}


/*******************************************************************
*
* tStTrackController::Init
*
*  Resets the controller
*
* SPEC:
     [[U tStTrackController::Init(TIME tServoTime) ]]
*
* RETURNS:
*    OK
*
****/

ERROR_CODE tStTrackController::Init(TIME tServoTime)
{
   int iCoord;
   
   for (iCoord = 0; iCoord < 2; iCoord++) {
      /**** 
      [[U this->Pid[iCoord].tPidDouble::Reset(0)]]
      [[U this->Lag[iCoord].tLagCompensator::Reset(0)]]
      *****/
      this->Pid[iCoord].Reset(0);
      this->Lag[iCoord].Reset(0);
   }
  
   return OK; 
}

/*******************************************************************
*
* tStTrackController::Reset
*
* Sends FSM position (0,0) to the Wap, to disable that controller
*
* SPEC:
     [[U tStTrackController::Reset(TIME tServoTime) ]]
*
* RETURNS:
*    Result of write to buffer
*
****/

ERROR_CODE tStTrackController::Reset(TIME tServoTime)
{
   tIpcPoint IpcOutput = {0,0};

   /**** 
   [[U this->Wap.tIpcDblBufferAccessor::Put(IcpOutput) ]]
   *****/
   ///return Wap.Put(IpcOutput); [delete next line]
         return OK;
}

   
/*******************************************************************
*
* tStTrackController::SetControllerParams
*
*  Sets the controller parameters
*
* SPEC:
     [[U tStTrackController::SetControllerParams(const tStControllerParams& Params) ]]
*
* RETURNS:
*    nothing
*
****/

void tStTrackController::SetControllerParams(const tStControllerParams& Params)
{
   int iCoord;
   //servo period in seconds
   double dPeriod = (double) this->CtServoPeriod / TIME_TICKS_PER_SECOND;
   
   this->bLagEnabled = Params.LagEnabled;
   
   this->StarTarget[X] = Params.StarTarget[X];
   this->StarTarget[Y] = Params.StarTarget[Y];
   
   //set those parameters for both x and y
   for (iCoord = 0; iCoord < 2; iCoord++) {
      /**** 
      [[U Error := this->Lag[iCoord].
         tLagCompensator::Initialize_ddd(Params.LagTauPole,
                                          Params.LagTauZero,dPeriod)]]
      *****/
      this->Lag[iCoord].Initialize(Params.LagTauPole,Params.LagTauZero,dPeriod);
   }  

   //set PID parameters for both x and y
   for (iCoord = 0; iCoord < 2; iCoord++) {
      /**** 
      [[U this->Pid[iCoord].tPidDouble::InitializeUsingPolesAndZeros(
                                                 Params.Gain,Params.PidTauZero1,
                                                 Params.PidTauZero2,
                                                 Params.PidTauPole,dPeriod)]]
      *****/
      this->Pid[iCoord].InitializeUsingPolesAndZeros(Params.Gain,Params.PidTauZero1,
                                                      Params.PidTauZero2,
                                                      Params.PidTauPole,dPeriod);

   }
      
}

        
/*******************************************************************
*tStCalibrateController::tStCalibrateController
*
*  Constructor.  Saves a reference to the actuator
*
* SPEC:
    [[U tStCalibrateController::tStCalibrateController(CompleteIoSet, ControllerDataToUse)]]
*
* RETURNS:
*    nothing
*
****/
tStCalibrateController::tStCalibrateController(tStCompleteIoSet& CompleteIoSet,
                                       tStControllerData& ControllerDataToUse) : 

   /**** 
   [[U this->tStController::tStController(CompleteIoSet, ControllerDataToUse)]]
   *****/
   tStController(CompleteIoSet, ControllerDataToUse),
                                          
   Fsm(CompleteIoSet.Fsm) {}
   

/*******************************************************************
*tStCalibrateController::Calculate(TIME tTime)
*
*  Moves the calibrating actuator to the positio corresponding to tTime
*
* SPEC:
    [[U tStCalibrateController::Calculate(TIME tTime)]]
*
* RETURNS:
*    ERROR_CODE
*
****/
                             
ERROR_CODE tStCalibrateController::Calculate(TIME tTime)
{

  ERROR_CODE ec;


  /**** 
  [[U ec := this->tStController::Calculate(tTime)]]
  *****/
   ec = this->tStController::Calculate(tTime);
   if (ec != OK) return ec;
   
  /**** 
  [[U ec := this->Fsm.tMultiActuator::Dither(tTime)]]
  *****/
  ec = this->Fsm.Dither(tTime);

  return ec;  
}

  
  
/*******************************************************************
*
* tStCompleteControllerSet::tStCompleteControllerSet
*
*  Constructs tStCompleteControllerSet.  This includes constructing
*     the controller data block, the OFF and TRACK controllers, and
*     storing a reference to the complete I/O set.
*
* SPEC:
*    [[U tStCompleteControllerSet::tStCompleteControllerSet(CompleteIoSetToUse,
*                                                        ImageStatusToUse, 
*                                                        StarImageDataTag,
*                                                        StarPosDataTag,
*                                                        FsmPosDataTag,
*                                                        tServoPeriod) ]]
*
* RETURNS:
*    nothing
*
****/

tStCompleteControllerSet::tStCompleteControllerSet(
                                         tStCompleteIoSet& CompleteIoSetToUse,
                                         tImageStatus&     ImageStatusToUse,
                                         const tTypeTag&   StarPosDataTag, 
                                         const tTypeTag&   FsmPosDataTag, 
                                         const tTypeTag&   StarImageDataTag,
                                         TIME tServoPeriod) :

  /****
    [[U this->ControllerData.tStControllerData::tStControllerData(
            ImageStatusToUse,StarPosDataTag,FsmPosDataTag,StarImageDataTag) ]]
  ****/
  ControllerData(ImageStatusToUse,StarPosDataTag,FsmPosDataTag,StarImageDataTag),

  /****
    [[U this->OffController.tStController::tStController(
    CompleteIoSetToUse,ControllerData) ]]
  ****/
  OffController(CompleteIoSetToUse,ControllerData),

  /****
    [[U this->TrackController.tStTrackController::tStTrackController(
                 CompleteIoSetToUse,ControllerData,tServoPeriod) ]]
  ****/
  TrackController(CompleteIoSetToUse,ControllerData,tServoPeriod),

  /****
    [[U this->CalibrateController.tStCalibrateController::tStCalibrateController(
                 CompleteIoSetToUse,ControllerData) ]]
  ****/
  CalibrateController(CompleteIoSetToUse,ControllerData),

  /****
  [[U this->StOffControllerGroup.tSimpleControllerGroup::tSimpleControllerGroup (1) ]]
  [[U this->StTrackControllerGroup.tSimpleControllerGroup::tSimpleControllerGroup (1) ]]
  [[U this->StCalibrateControllerGroup.tSimpleControllerGroup::tSimpleControllerGroup (1) ]]
  ****/
  StOffControllerGroup (1),
  StTrackControllerGroup (1),
  StCalibrateControllerGroup (1)
  
{
   ERROR_CODE ec;

    TraceMark(S_stgizmo_ST_COMPLETE_CONTROLLER_SET_CONSTRUCTOR);
    TraceMark(S_stgizmo_ASSIGNING_CTLRS_TO_PARTITIONS);


    // Assign Controllers to Servo Partitions in Controller Group
    /****
       [[U ec := this->StOffControllerGroup.tSimpleControllerGroup::
                 AssignController(0,OffController) ]]
     ****/ 
     ec = this->StOffControllerGroup.AssignController(0,OffController);

    if (ec != OK) {
      TraceMsg(tmt_ERROR,ec,"tStGizmo::tStGizmo" );
      return;
    }

    /****
       [[U ec := this->StTrackControllerGroup.tSimpleControllerGroup::
                 AssignController(0,TrackController) ]]
     ****/ 
    ec = this->StTrackControllerGroup.AssignController(0,TrackController);
 
    if (ec != OK) {
      TraceMsg(tmt_ERROR,ec,"tStGizmo::tStGizmo" );
      return;
    }

    /****
       [[U ec := this->StCalibrateControllerGroup.tSimpleControllerGroup::
                 AssignController(0,CalibrateController) ]]
     ****/ 
    ec = this->StCalibrateControllerGroup.AssignController(0,CalibrateController);
 
    if (ec != OK) {
      TraceMsg(tmt_ERROR,ec,"tStGizmo::tStGizmo" );
      return;
    }
}


/*******************************************************************
*
* tStCompleteControllerSet::~tStCompleteControllerSet
*
*  Destroys the tStCompleteControllerSet.
*
* SPEC:
    [[U tStCompleteControllerSet::~tStCompleteControllerSet(void) ]]
*
* RETURNS:
*    nothing
*
****/

tStCompleteControllerSet::~tStCompleteControllerSet(void) 
{
   TraceMark(S_stgizmo_ST_COMPLETE_CONTROLLER_SET_DESTRUCTOR);
}

