/* $RCSfile: stGizmo.cc,v $ */
/* CM Info:   
$Header: /home/chara/cvsroot/ople/sys/gizmo/st/stGizmo.cc,v 1.2 1998/09/23 23:57:10 phil Exp $
*/
/* Richard Johnson - Jet Propulsion Laboratory */

/*
modification history
--------------------
01b  22Sep97,eamck  taken over
01a  27Aug97,rlj    written
*/

#include "iostream.h"
#include "isit.h"
#include "baseObject.h"
#include "stError.h"
#include "stGizmo.h"
#include "stGizmoInfo.h"
#include "st/stIoSet.h"
#include "st/stController.h"
#include "servo.h"
#include "stCommands.h"
#include "controllerSet.h"
#include "trace.h"

/*******************************************************************
*
* tStGizmo::tStGizmo -- tStGizmo constructor
*
*  Calls tRealtimeGizmo(base class) constructor, plus constructors
*     for I/O and controllers and status.  Does minimal error
*     checking on GizmoInfo parameters.  Sets up controllers,
*     assigning controller groups to servo modes and controllers
*     to controller groups.  Initializes servo mode to OFF and then
*     calls tRealtimeGizmo::StartTasks() to complete initialization.
*
* SPEC:
      [[U tStGizmo::tStGizmo(GizmoInfo) ]]
*
* RETURNS:
*    nothing
*
****/

tStGizmo::tStGizmo(const tStGizmoInfo& GizmoInfo) :

  // First call the base class constructor
  /****
     [[U tRealtimeGizmo::tRealtimeGizmo(GizmoInfo) ]]
   ****/ 
  tRealtimeGizmo(GizmoInfo),
  
  // Now construct the status block type tag
  /****
     [[U this->CStStatusTag.tTypeTag::tTypeTag_STTNisz(ST_STATUS, 
                    TN_ST_STATUS, GizmoInfo.iGizmoId, sizeof(tStStatus) ) ]]
   ****/ 
  CStStatusTag(ST_STATUS, TN_ST_STATUS, GizmoInfo.GetGizmoId(), 
                sizeof(tStStatus) ),



  // Now construct the image status type tag
  /****
     [[U this->CStStarImageStatusTag.tTypeTag::tTypeTag_STTNisz(ST_STATUS, 
                TN_ST_IMAGE, GizmoInfo.iGizmoId, GizmoInfo.iStarImageHSize *
      GizmoInfo.iStarImageVSize) ]]
   ****/ 
  CStStarImageStatusTag(ST_STATUS, TN_ST_IMAGE, GizmoInfo.GetGizmoId(), 
             GizmoInfo.GetStarImageHSize() * GizmoInfo.GetStarImageVSize() ), 



  // Now construct the star position data block type tag
  /****
     [[U this->CStDataTag.tTypeTag::tTypeTag_STTNisz(ST_DATA, 
                TN_ST_STAR_POSITION, GizmoInfo.iGizmoId, sizeof(tStData) ) ]]
   ****/ 
  CStStarPosDataTag(ST_DATA, TN_ST_STAR_POSITION, GizmoInfo.GetGizmoId(), 
                 sizeof(tPoint)),  



  // Now construct the FSM position data block type tag
  /****
     [[U this->CStDataTag.tTypeTag::tTypeTag_STTNisz(ST_DATA, 
                TN_ST_FSM_POSITION, GizmoInfo.iGizmoId, sizeof(tPoint) ) ]]
   ****/ 
  CStFsmPosDataTag(ST_DATA, TN_ST_FSM_POSITION, GizmoInfo.GetGizmoId(), 
                 sizeof(tPoint)),



  // Now construct the image data type tag
  /****
     [[U this->CStDataTag.tTypeTag::tTypeTag_STTNisz(ST_DATA, TN_ST_IMAGE, 
                GizmoInfo.iGizmoId, GizmoInfo.iStarImageHSize *
      GizmoInfo.iStarImageVSize ) ]]
   ****/ 
  CStStarImageDataTag(ST_DATA, TN_ST_IMAGE, GizmoInfo.GetGizmoId(), 
                GizmoInfo.GetStarImageHSize() * GizmoInfo.GetStarImageVSize()),
 
  // Now construct the status block
  /****
     [[U this->Status.tStStatus::tStStatus(GizmoInfo.iGizmoId) ]]
   ****/ 
  Status(GizmoInfo.GetGizmoId() ),

  // Now construct the image status block
  /****
     [[U this->StarImageStatus.tImageStatus::tImageStatus(
            this->CStStarImageStatusTag,
            GizmoInfo.iStarImageHSize, GizmoInfo.iStarImageVSize) ]]
   ****/ 
  StarImageStatus( this->CStStarImageStatusTag,
                   GizmoInfo.GetStarImageHSize(),
         GizmoInfo.GetStarImageVSize() ),

  /****
     [[U this->CompleteIoSet.tStCompleteIoSet::tStCompleteIoSet(GizmoInfo) ]]
   ****/ 
  CompleteIoSet(GizmoInfo),

  // Now construct the controller set  (if new returns NULL, what happens here?)
  /****
  [[U   ServoInfo := this->tRealtimeGizmoInfo::GetServoInfo() ]]
  [[U   ServoTaskInfo := ServoInfo.tServoInfo::GetServoTaskInfo(0) ]]
  [[U   this->CompleteControllerSet.
        tStCompleteControllerSet::tStCompleteControllerSet(CompleteIoSet,
        StarImageStatus, 
        CStStarPosDataTag, CStFsmPosDataTag, CStStarImageDataTag,
        ServoTaskInfo.tTaskPeriod) ]]
   ****/ 
  CompleteControllerSet(CompleteIoSet,
    StarImageStatus, CStStarPosDataTag, CStFsmPosDataTag, CStStarImageDataTag,
    GizmoInfo.GetServoInfo().GetServoTaskInfo(STP_BASIC).tTaskPeriod)

 {
    ERROR_CODE ec;
  
    TraceMark(S_stgizmo_ENTERING_ST_GIZMO_CONSTRUCTOR);

    // If we had a error in base class constructor, exit with error 

    if ((ec = this->tBaseObject::GetInitError()) != OK) {
      TraceMsg(tmt_ERROR,ec,"tStGizmo::tStGizmo");
      return;
    }

    // If we had a error in allocating members, exit with error

    if ((&CompleteIoSet==NULL) || (&CompleteControllerSet==NULL))
    {
      this->SetInitError(S_stgizmo_CANT_CONSTRUCT_IO_OR_CTLR_SET);
      TraceMsg(tmt_ERROR,S_stgizmo_CANT_CONSTRUCT_IO_OR_CTLR_SET,
                       "tStGizmo::tStGizmo");
      return;
    }


    // If GizmoInfo doesn't think we're a STAR_TRACKER, exit with error 

    /****
    [[U this->GizmoInfo.tGizmoInfo::GetGizmoType() ]]
    ****/ 
    if ( GizmoInfo.GetGizmoType() != GT_STAR_TRACKER ) {

      this->SetInitError(S_stgizmo_GIZMO_INFO_NOT_ST);
      TraceMsg(tmt_ERROR,S_stgizmo_GIZMO_INFO_NOT_ST,
                "tStGizmo::tStGizmo" );
      return;
    }
  

    // Sanity Checks until we can better automate the controller setup.


    // Code currently expects 3 servo modes 

    /****
    [[U this->GizmoInfo.tRealtimeGizmoInfo::GetNumberOfServoModes() ]]
    ****/ 
    if ( GizmoInfo.GetNumberOfServoModes() != SSS_NUM_SERVO_MODES) {

      this->SetInitError(S_stgizmo_INVALID_NUMBER_OF_SERVO_MODES);
      TraceMsg(tmt_ERROR,S_stgizmo_INVALID_NUMBER_OF_SERVO_MODES,
                "tStGizmo::tStGizmo" );
      return;
    }
    
    /****
    [[U this->GizmoInfo.tRealtimeGizmoInfo::GetNumberOfServoPartitions() ]]
    ****/ 
    // Code currently expects only 1 servo partition 

    if (GizmoInfo.GetNumberOfServoPartitions() != STP_NUM_PARTITIONS ) {

      this->SetInitError(S_stgizmo_INVALID_NUMBER_OF_SERVO_PARTITIONS);
      TraceMsg(tmt_ERROR,S_stgizmo_INVALID_NUMBER_OF_SERVO_PARTITIONS,
                "tStGizmo::tStGizmo" );
      return;
    }
    
    TraceMark(S_stgizmo_ASSIGNING_CTLR_GROUPS_TO_MODES);


    // Assign Controller Groups to Servo Modes in Controller Set
    /****
       [[U OffGroup := this->CompleteControllerSet.
         tStCompleteControllerSet::OffControllerGroup() ]]
       [[U ec := this->ControllerSet.tControllerSet::AssignControllerGroup 
              (SSM_OFF,OffGroup) ]] 
     ****/ 
    tSimpleControllerGroup& OffGroup = 
      this->CompleteControllerSet.OffControllerGroup();
    ec = this->ControllerSet.AssignControllerGroup(SSM_OFF,OffGroup);
    if (ec != OK) {
      TraceMsg(tmt_ERROR,ec,"tStGizmo::tStGizmo" );
      this->SetInitError(ec);
      return;
    }

    /****
       [[U TrackGroup := this->CompleteControllerSet.
         tStCompleteControllerSet::TrackControllerGroup() ]]
       [[U ec := this->ControllerSet.tControllerSet::AssignControllerGroup 
              (SSM_TRACK,TrackGroup) ]]
     ****/ 
    tSimpleControllerGroup& TrackGroup = 
      this->CompleteControllerSet.TrackControllerGroup();
    ec = this->ControllerSet.AssignControllerGroup(SSM_TRACK,TrackGroup);
              
    if (ec != OK) {
      TraceMsg(tmt_ERROR,ec,"tStGizmo::tStGizmo" );
      this->SetInitError(ec);
      return;
    }
  
    /****
       [[U CalibrateGroup := this->CompleteControllerSet.
         tStCompleteControllerSet::CalibrateControllerGroup() ]]
       [[U ec := this->ControllerSet.tControllerSet::AssignControllerGroup 
              (SSM_CALIBRATE,CalibraterGroup) ]]
     ****/ 
    tSimpleControllerGroup& CalibrateGroup = 
      this->CompleteControllerSet.CalibrateControllerGroup();
    ec = this->ControllerSet.AssignControllerGroup(SSM_CALIBRATE,CalibrateGroup);

    if (ec != OK) {
      TraceMsg(tmt_ERROR,ec,"tStGizmo::tStGizmo" );
      this->SetInitError(ec);
      return;
    } 

    // Select Off Mode
    /****
       [[U ec := this->Servo.tServo::SelectMode(SSM_OFF) ]] 
     ****/ 
    ec = this->Servo.SelectMode(SSM_OFF);
    
    if (ec != OK) {
      TraceMsg(tmt_ERROR,ec,"tStGizmo::tStGizmo" );
      this->SetInitError(ec);
      return;
    }

    /****
    [[U ServoParams := GizmoInfo.tStGizmoInfo::ServoParams() ]]
    [[U this->tStGizmo::ChangeServoParams(ServoParams)]]
    ****/
    this->ChangeServoParams(GizmoInfo.ServoParams());

    TraceMark(S_stgizmo_ENABLING_BASE_CLASS_TASKS);

    // Enable Base class tasks ...
    /****
       [[U ec := this->tRealtimeGizmo::StartTasks(GizmoInfo) ]] 
     ****/ 
    this->tRealtimeGizmo::StartTasks(GizmoInfo);

    if (ec != OK) {
      TraceMsg(tmt_ERROR,ec,"tStGizmo::tStGizmo" );
      this->SetInitError(ec);
      return;
    }

    TraceMark(S_stgizmo_LEAVING_ST_GIZMO_CONSTRUCTOR);

}

/******************************************************************
*
* tStGizmo::~tStGizmo -- tStGizmo destructor
*
*  Destroys tStGizmo object.  (calls base class destructor)
*
* SPEC:
      [[U tStGizmo::~tStGizmo(void) ]]
*
* RETURNS:
*    nothing
*
****/
tStGizmo::~tStGizmo(void) 

  /****
     [[U tRealtimeGizmo::~tRealtimeGizmo(void) ]]
  ****/
{ 
   TraceMark(S_stgizmo_ST_GIZMO_DESTRUCTOR);
}

/******************************************************************
*
* tStGizmo::Restart 
*
*  Calls base class Restart().
*
* SPEC:
      [[U tStGizmo::Restart(void) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/
ERROR_CODE tStGizmo::Restart(void) 
{
   TraceMark(S_stgizmo_RESTART);

  /****
     [[U return := this->tRealtimeGizmo::Restart(void) ]]
  ****/
  return this->tRealtimeGizmo::Restart();
}

/******************************************************************
*
* tStGizmo::Shutdown 
*
*  Calls base class Shutdown().
*
* SPEC:
      [[U tStGizmo::Shutdown(void) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/

ERROR_CODE tStGizmo::Shutdown(void) 
{
   TraceMark(S_stgizmo_SHUTDOWN);

  /****
     [[U return := this->tRealtimeGizmo::Shutdown(void) ]]
  ****/
  return this->tRealtimeGizmo::Shutdown();
}
 
/******************************************************************
*
* tStGizmo::IsOk 
*
*  Calls base class IsOk().
*
* SPEC:
      [[U tStGizmo::IsOk(void) ]]
*
* RETURNS:
*    BOOLEAN
*
****/

BOOL tStGizmo::IsOk(void) 
{
  /****
     [[U return := this->tRealtimeGizmo::IsOk(void) ]]
  ****/
  return this->tRealtimeGizmo::IsOk();
}

/******************************************************************
*
* tStGizmo::operator<< 
*
*  Prints out "Star Tracker Gizmo" followed by results of
*     calling base class operator<<
*
* SPEC:
      [[U tStGizmo::operator_extractor(ostream& stream, tStGizmo& Gizmo) ]]
*
* RETURNS:
*     stream
*
****/

ostream& operator<<(ostream& stream, tStGizmo& Gizmo) 
{
  stream << "Star Tracker Gizmo" << endl;

  /****
     [[U this->tRealtimeGizmo::operator_extractor(stream,Gizmo) ]]
  ****/
  stream << (tRealtimeGizmo &) Gizmo;
  return stream;
}


/******************************************************************
*
* tStGizmo::CollectStatus 
*
* Gets status info(star position, image) from rest of Star Tracker 
*    Gizmo, then sends it to RTE status stream
*
* SPEC:
      [[U tStGizmo::CollectStatus(void) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/

ERROR_CODE tStGizmo::CollectStatus(void) 
{
  ERROR_CODE ec;
  tImage StarImage;

   TraceMark(S_stgizmo_ENTERING_COLLECTSTATUS);

  // Get current time stamp - do first; more likely to match image
  this->Status.tTime = ::GetTime();


   TraceMark(S_stgizmo_GETTING_IMAGE_STATUS);

  // Get copy of latest image from Image Status object
  /****
     [[U this->StarImageStatus.tImageStatus::GetImage(StarImage) ]]
  ****/
  this->StarImageStatus.GetImage(StarImage);



  // Put latest star position coords into ST status block
  /****
     [[U this->Status.StarPosition := 
                  this->CompleteControllerSet.
          tStCompleteControllerSet::GetStarPosition(void) ]]
  ****/
  this->Status.StarPosition.dXPosition = 
    this->CompleteControllerSet.GetStarPosition().x(); 
  this->Status.StarPosition.dYPosition = 
    this->CompleteControllerSet.GetStarPosition().y();

   //put latest FSM position into ST status block
   this->Status.FsmPosition[X] = 
      this->CompleteControllerSet.GetFsmPosition().x(); 
   this->Status.FsmPosition[Y] = 
      this->CompleteControllerSet.GetFsmPosition().y(); 

   TraceMark(S_stgizmo_SENDING_STATUS_BLOCK);

  // Send the status block to RTE status stream
  /****
     [[U ec := tGizmo::SendStatus_tp(CStStatusTag, &this->Status) ]]
  ****/
  ec = this->SendStatus(CStStatusTag,(BYTE *) &this->Status);
  
  // Handle error if any
  /****
     [[U ec := tGizmo::SendError_ec (S_stgizmo_UNABLE_TO_SEND_STATUS) ]]
  ****/
  if (ec != OK) 
     this->SendError (S_stgizmo_UNABLE_TO_SEND_STATUS);



   TraceMark(S_stgizmo_SENDING_STATUS_IMAGE);

  // Send image to RTE status stream
  /****
     [[U ec := tGizmo::SendStatus_tt(CStStarImageStatusTag,StarImage) ]]
  ****/
  ec = this->SendStatus(CStStarImageStatusTag,StarImage);

  // Handle error if any
  /****
     [[U ec := tGizmo::SendError_ec(S_stgizmo_UNABLE_TO_SEND_STATUS) ]]
  ****/
  if (ec != OK) 
     this->SendError(S_stgizmo_UNABLE_TO_SEND_STATUS);

   TraceMark(S_stgizmo_LEAVING_COLLECTSTATUS);

  return OK;
}
  
/******************************************************************
*
* tStGizmo::Dispatch
*
* Calls the appropriate tStGizmo method that corresponds to the
*    command icNewCmd.  Note that this early version of the ST
*    does not use command engines, just function calls, since 
*    none of the commands requires any mutual exclusion because
*    all current commands execute and return immediately.
*
* SPEC:
*    [[U tStGizmo::Dispatch(const tIsitCmd& icNewCmd) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/

ERROR_CODE tStGizmo::Dispatch(const tIsitCmd& icNewCmd) 
{

   TraceMark(S_stgizmo_ENTERING_DISPATCH);

  switch(icNewCmd.Command) {

    case ST_CMD_PING : 
       /****
          [[U return := tStGizmo::Ping(void) ]]
       ****/
       return this->Ping();

    case ST_CMD_OFF : 
       /****
          [[U return := tStGizmo::Off(void) ]]
       ****/
       return this->Off();

    case ST_CMD_TRACK : 
       /****
          [[U return := tStGizmo::Track(void) ]]
       ****/
       return this->Track();

    case ST_CMD_CALIBRATE : 
       /****
          [[U return := tStGizmo::Calibrate((tStCalibrateCmd&)icNewCmd) ]]
       ****/
       return this->Calibrate((tStCalibrateCmd&)icNewCmd);

    case ST_CMD_SETSS : 
       /****
          [[U return := tStGizmo::SetServoParams((tStSetServoParamsCmd&)icNewCmd) ]]
       ****/
       return this->SetServoParams((tStSetServoParamsCmd&)icNewCmd);

    case ST_CMD_HOME : 
       /****
          [[U return := tStGizmo::Home() ]]
       ****/
       return this->Home();

    default : 
       return S_stgizmo_UNRECOGNIZED_COMMAND;
  }

   TraceMark(S_stgizmo_LEAVING_DISPATCH);

   return OK;
}
  
    
/******************************************************************
*
* tStGizmo::Ping
*
* Called in response to the ST receiving a PING command.  Sends a 
*     friendly message to the RTE error stream.
*
* SPEC:
     [[U  tStGizmo::Ping(void) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/

ERROR_CODE tStGizmo::Ping(void) 
{
   TraceMark(S_stgizmo_PING);

   /****
   [[U  return := tGizmo::SendError_ec(S_stgizmo_STAR_TRACKER_ALIVE) ]]
   ****/
   return SendError(S_stgizmo_STAR_TRACKER_ALIVE);
}

   
/******************************************************************
*
* tStGizmo::Off
*
* Called in response to the ST receiving an OFF command.  Sets the
*     servo mode to OFF, as well as updating information in status
*     and error streams about executing the command.
*
* SPEC:
     [[U tStGizmo::Off(void) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/

ERROR_CODE tStGizmo::Off(void) 
{
  ERROR_CODE ec;

   TraceMark(S_stgizmo_ENTERING_OFF_METHOD);


  // Tell the servo to change mode to OFF
  /****
  [[U ec := this->Servo.tServo::SelectMode(SSM_OFF) ]]
  ****/
  ec = this->Servo.SelectMode(SSM_OFF);

  if (ec == OK) {

    /****
    [[U this->Status.tStStatus::UpdateServoState (SSS_OFF) ]]
    ****/
    this->Status.UpdateServoState (SSS_OFF);


    // Report status: we finished the command 
    /****
    [[U this->Status.tStStatus::UpdateCommandStatus(STCS_OFF) ]]
    ****/
    this->Status.UpdateCommandStatus(STCS_OFF);
  
    /****
    [[U this.tGizmo::SendError_ec(S_stgizmo_OFF_COMPLETE) ]]
    ****/
    this->SendError(S_stgizmo_OFF_COMPLETE);
   
  }
  SendError(S_stgizmo_LEAVING_OFF_METHOD);
  
  TraceMark(S_stgizmo_LEAVING_OFF_METHOD);

  return ec; 
}

  
/******************************************************************
*
* tStGizmo::Track
*
* Called in response to the ST receiving an TRACK command.  Sets the
*     servo mode to TRACK, as well as updating information in status
*     and error streams about executing the command.
*
* SPEC:
     [[U  tStGizmo::Track(void) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/

ERROR_CODE tStGizmo::Track(void) 
{
  ERROR_CODE ec;

   TraceMark(S_stgizmo_ENTERING_TRACK_METHOD);

  // Tell the servo to change mode to TRACK

  /****
  [[U ec := this->Servo.tServo::SelectMode(SSM_TRACK) ]]
  ****/
  ec = this->Servo.SelectMode(SSM_TRACK);
  
  if (ec == OK) {

    /****
    [[U this->Status.tStStatus::UpdateServoState (SSS_TRACK) ]]
    ****/
    this->Status.UpdateServoState (SSS_TRACK);

    // Report status: we finished the command(but stay in TRACK mode) 
    /****
    [[U this->Status.tStStatus::UpdateCommandStatus(STCS_TRACK_ACTIVE) ]]
    ****/
    this->Status.UpdateCommandStatus(STCS_TRACK_ACTIVE);

    /****
    [[U this->tGizmo::SendError_ec(S_stgizmo_TRACK_IN_PROGRESS) ]]
    ****/
    this->SendError(S_stgizmo_TRACK_IN_PROGRESS);
  }
   SendError(S_stgizmo_LEAVING_TRACK_METHOD);
   
   TraceMark(S_stgizmo_LEAVING_TRACK_METHOD);

  return ec; 
}



/******************************************************************
*
* tStGizmo::Calibrate
*
* Called in response to the ST receiving a CALIBRATE command.  Sets the
*     servo mode to CALIBRATE, as well as updating information in status
*     and error streams about executing the command.
*
* SPEC:
     [[U  tStGizmo::Calibrate(const tFtCalibrateCmd& icNewCmd) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/

ERROR_CODE tStGizmo::Calibrate(const tStCalibrateCmd& icNewCmd) 
{
  ERROR_CODE ec;
   
   TraceMark(S_stgizmo_ENTERING_CALIBRATE_METHOD);


  /****
  [[U this->CompleteIoSet.tStCompleteIoSet::
       SetCalParams(icNewCmd) ]]
  ****/
  this->CompleteIoSet.SetCalParams(icNewCmd);

  // Tell the servo to change mode to CALIBRATE

  /****
  [[U ec := this->Servo.tServo::SelectMode(SSM_CALIBRATE) ]]
  ****/
  ec = this->Servo.SelectMode(SSM_CALIBRATE);
  
  if (ec == OK) {

    /****
    [[U this->Status.tStStatus::UpdateServoState (SSS_CALIBRATE) ]]
    ****/
    this->Status.UpdateServoState (SSS_CALIBRATE);

    // Report status: we finished the command(but stay in CALIBRATE mode) 
    /****
    [[U this->Status.tStStatus::UpdateCommandStatus(STCS_CALIBRATE_ACTIVE) ]]
    ****/
    this->Status.UpdateCommandStatus(STCS_CALIBRATE_ACTIVE);

    /****
    [[U this->tGizmo::SendError_ec(S_stgizmo_CALIBRATE_IN_PROGRESS) ]]
    ****/
    this->SendError(S_stgizmo_CALIBRATE_IN_PROGRESS);

  }
   SendError(S_stgizmo_LEAVING_CALIBRATE_METHOD);
   
   TraceMark(S_stgizmo_LEAVING_CALIBRATE_METHOD);

  return ec; 
}



/******************************************************************
*
* tStGizmo::SetServoParams
*
* Called in response to the ST receiving a SETSERVOPARAMETERS command. It 
* updates information in status
* and error streams about executing the command.
*
* SPEC:
     [[U  tStGizmo::SetServoParams(const tStSetServoParametersCmd& icNewCmd) ]]
*
* RETURNS:
*    OK
*
****/

ERROR_CODE tStGizmo::SetServoParams(const tStSetServoParamsCmd& icNewCmd) 
{

  TraceMark(S_stgizmo_ENTERING_SSP_METHOD);

  /****
  [[U  this->tStGizmo::ChangeServoParams(icNewCmd.Params)]]
  ****/
  this->ChangeServoParams(icNewCmd.Params);
  
  SendError(S_stgizmo_LEAVING_SSP_METHOD);
  
  TraceMark(S_stgizmo_LEAVING_SSP_METHOD);
  
  return OK;
}



/******************************************************************
*
* tStGizmo::ChangeServoParams
*
* It updates information in status and error streams about executing the 
* command.  N.B., Xform is the transpose of Params.XformDetToFsm, 
* for historical reasons.
*
* SPEC:
     [[U  tStGizmo::ChangeServoParams(const tStServoParameters& Params) ]]
*
* RETURNS:
*    nothing
*
****/

void tStGizmo::ChangeServoParams(const tStServoParams& Params) 
{
  tVector Home(NUM_FSM_CHANNELS);
  tMatrix Xform(NUM_FSM_CHANNELS,2);
  int iRow,iColumn;
  
   //send parameters to status
   this->Status.ServoParams = Params;

    //copy from command structure to tVector, tMatrix (transpose latter)
    for (iRow = 0; iRow < NUM_FSM_CHANNELS; iRow++) {
      Home[iRow] = Params.FsmHome[iRow];
      for (iColumn = 0; iColumn < 2; iColumn++) {
         Xform[iRow][iColumn] = Params.XformDetToFsm[iColumn][iRow];
      }
    }
        
    /****
    [[U this->CompleteIoSet.tStCompleteIoSet::
      SetXformParams(Xform,Home) ]]
    ****/
    this->CompleteIoSet.SetXformParams(Xform,Home);

    /****
    [[U this->CompleteControllerSet.tStCompleteControllerSet::
      SetControllerParams(Params.ControllerParams) ]]
    ****/
    this->CompleteControllerSet.SetControllerParams(Params.ControllerParams);
}



/******************************************************************
*
* tStGizmo::Home
*
* Called in response to the ST receiving a HOME command.  Moves
* the FSM to the home position.
*
* SPEC:
     [[U  tStGizmo::Home() ]]
*
* RETURNS:
*    OK
*
****/

ERROR_CODE tStGizmo::Home() 
{
   ERROR_CODE ec;
   
   TraceMark(S_stgizmo_ENTERING_HOME_METHOD);

   // Tell the servo to change mode to OFF
   /****
   [[U ec := this->Servo.tServo::SelectMode(SSM_OFF) ]]
   ****/
   ec = this->Servo.SelectMode(SSM_OFF);
   
 
   if (ec == OK) {

       /****
       [[U this->Status.tStStatus::UpdateServoState (SSS_OFF) ]]
       ****/
       this->Status.UpdateServoState (SSS_OFF); 
   
       /****
       [[U this->CompleteIoSet.Fsm.tMultiActuator::Home() ]]
       ****/
       this->CompleteIoSet.Fsm.Home();
   }
   
  SendError(S_stgizmo_LEAVING_HOME_METHOD); 
    
  TraceMark(S_stgizmo_LEAVING_HOME_METHOD);

  return OK; 
}



