/* $RCSfile: stepperMultiActuator.cc,v $ */
/* CM Info$
*/

#include <float.h>

#include "taskLib.h"
#include "time.h"
#include "isit.h"
#include "wap/stepperMultiActuator.h"
#include "maError.h"

/*******************************************************************
*
* tStepperMultiActuator::tStepperMultiActuator()
*
*  Constructor.  Initializes the base class, and creates and initializes sensors.
*
* SPEC:
      [[U tStepperMultiActuator::tStepperMultiActuator(int iActuators,
                            tInputChannel *const InputChannelArray[2],
                            double dInputScaleFactor,
                            tOutputChannel *const OutputChannelArray[2],
                            double dOutputScaleFactor) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/
tStepperMultiActuator::tStepperMultiActuator(int iActuators,
                            tInputChannel *const InputChannelArray[2],
                            double dInputScaleFactor,
                            tOutputChannel *const OutputChannelArray[2],
                            double dOutputScaleFactor) :
   tMultiActuator(OutputChannelArray,iActuators,dOutputScaleFactor)
{
   int iSensor;
   
   for (iSensor = 0; iSensor < iActuators; iSensor++) {

      /****
      [[U tSensor::tSensor_td(*InputChannelArray[iSensor],dInputScaleFactor) ]]
      ****/ 
      this->pSensorArray[iSensor] = 
         new tSensor(*InputChannelArray[iSensor],dInputScaleFactor);
   }
}
   

/*******************************************************************
*
* tStepperMultiActuator::~tStepperMultiActuator -- tStepperMultiActuator destructor
*
*  Deallocates pSensorArray[]
*
* SPEC:
      [[U tStepperMultiActuator::~tStepperMultiActuator()]]
*
* RETURNS:
*    nothing
*
****/
tStepperMultiActuator::~tStepperMultiActuator()
{
   int iSensor;
   
   for (iSensor = 0; iSensor < this->CiActuators; iSensor++) {

      /****
      [[U tSensor::~tSensor() ]]
      ****/ 
      delete this->pSensorArray[iSensor];
   }

   delete[] this-> pSensorArray;
}


/*******************************************************************
*
* tStepperMultiActuator::SetHomeParams
*
*  Saves the specified Home parameters
*
* SPEC:
      [[U tStepperMultiActuator::SetHomeParams(const tVector& NewHardStop,
                        double dNewHardStopTime,double dtNewHomeTime) ]]
*
* RETURNS:
*    nothing
*
****/
void tStepperMultiActuator::SetHomeParams(const tVector& NewHardStop,
                         double dNewHardStopTime,double dNewHomeTime)
{
   this->HardStop = NewHardStop;
   this->dHardStopTime = dNewHardStopTime;
   this->dHomeTime = dNewHomeTime;
}

/*******************************************************************
*
* tStepperMultiActuator::SetHome
*
*  Saves the specified Home position
*
* SPEC:
      [[U tStepperMultiActuator::SetHome(const tVector& NewHomePosition) ]]
*
* RETURNS:
*    nothing
*
****/
void tStepperMultiActuator::SetHome(const tVector& NewHomePosition)
{
   this->HomePosition = NewHomePosition;
}


/*******************************************************************
*
* tStepperMultiActuator::Position
*  Returns the position of the actuators, from the hardware register
*
* SPEC:
      [[U tStepperMultiActuator::Position() ]]
*
* RETURNS:
*    nothing
*
****/
tPoint tStepperMultiActuator::Position(void) const
{  
   tPoint Position;
   
   /****
   [[U this-> pSensorArray[X]->tSensor::Update(0) ]]
   [[U this-> pSensorArray[Y]->tSensor::Update(0) ]]
   ****/
   Position[X] = this->pSensorArray[X]->Update(0);   
   Position[Y] = this->pSensorArray[Y]->Update(0);   

   //transform from actuator position to detector position
   /****
   [[U this->tMultiActuator::XformActToDet() ]]
   ****/
   return this->XformActToDet() * Position;
}


/*******************************************************************
*
* tStepperMultiActuator::Home
*   Moves to the hardstops and then home
*
* SPEC:
      [[U tStepperMultiActuator::Home() ]]
*
* RETURNS:
*    Error if move or taskDelay fails
*
****/
ERROR_CODE tStepperMultiActuator::Home()
{  
   tVelocity Velocity;
   double dDuration;
   ERROR_CODE ec;
   STATUS Status;
   int iActuator;
   
   if (this->dHardStopTime < DBL_EPSILON || this->dHomeTime < DBL_EPSILON) 
      return S_multiact_DIVIDE_BY_ZERO;

   //move to hard stop at required velocity
   /****
   [[U this->tMultiActuator::Move(0,this->HardStop / this->dHardStopTime) ]]
   ****/
   ec = this->Move(0,this->HardStop / this->dHardStopTime);
   
   if (!ec) return ec;

   //wait for specified duration
   Status = taskDelay((int)(this->dHardStopTime * CLOCKS_PER_SEC));

   if (Status != OK) return S_multiact_TASK_DELAY_FAILED;
   
   //move to home position at required velocity
   /****
   [[U this->tMultiActuator::Move(0,this->HomePosition / this->dHomeTime) ]]
   ****/
   ec = this->Move(0,this->HomePosition / this->dHomeTime);
   
   if (!ec) return ec;

   //wait for specified duration
   Status = taskDelay((int)(this->dHomeTime * CLOCKS_PER_SEC));

   if (Status != OK) return S_multiact_TASK_DELAY_FAILED;

   //stop the collector
   /****
   [[U this->tMultiActuator::Move(0,tVelocity(0,0)) ]]
   ****/
   ec = this->Move(0,tVelocity(0,0));   
 
   if (!ec) return ec;
   
   //zero the hardware position registers
   for (iActuator = 0; iActuator < this->CiActuators; iActuator++) {
      /****
      [[U this->pActuatorArray[iActuator]->tModulatedActuator::Reset() ]]
      ****/
      ec = this->pActuatorArray[iActuator]->Reset();

      if (!ec) return ec;
   }

   return OK;
}


