/* $RCSfile: wapController.cc,v $ */
/* CM Info:   
$Header: /home/chara/cvsroot/ople/sys/gizmo/wap/wapController.cc,v 1.1.1.1 1998/06/11 21:36:25 chara Exp $
*/
/* Russ Brill - Jet Propulsion Laboratory */

#include <float.h>

#include "isit.h"
#include "isitExt.h"
#include "wap/wapController.h"
#include "wap/wapControllerSet.h"
#include "wap/wapIoSet.h"
#include "ipcid.h"
#include "wapGizmoInfo.h"
#include "wapTypes.h"
#include "wapError.h"
#include "servo.h"
#include "imageStatus.h"
#include "image.h"
#include "matrix.h"
#include "multiActuator.h"
#include "pidLibF.h"
#include "lagLib.h"
#include "ipcDblBufferAccessor.h"


/*******************************************************************
*
* tWapController::tWapController
*
*  Constructs tWapController object.
*
* SPEC:
     [[U tWapController::tWapController(CompleteIoSet, ControllerDataToUse) ]]
*
* RETURNS:
*    nothing
*
****/

tWapController::tWapController(tWapCompleteIoSet& CompleteIoSet,
                               tWapControllerData& ControllerDataToUse) :

  ControllerData(ControllerDataToUse),
  Collector(CompleteIoSet.Collector)
  
{
   TraceMark(S_wapgizmo_WAP_CONTROLLER_CONSTRUCTOR);
}


/*******************************************************************
*
* tWapController::~tWapController
*
*  Destroys tWapController object.
*
* SPEC:
     [[U tWapController::~tWapController(void) ]]
*
* RETURNS:
*    nothing
*
****/

tWapController::~tWapController(void)
{
   TraceMark(S_wapgizmo_WAP_CONTROLLER_DESTRUCTOR);
}


/*******************************************************************
*tWapController::Calculate(TIME tTime)
*
*  Sends the Collector position only
*
* SPEC:
    [[U tWapController::Calculate(TIME tTime)]]
*
* RETURNS:
*    OK
*
****/
                             
ERROR_CODE tWapController::Calculate(TIME tTime)
{
   
   /****
   [[U  this->tWapController::SendPosition() ]]
   ****/
   this->SendPosition();

   return OK;    
}


/*******************************************************************
*
* tWapController::Reset
*
*  Stops collector motion
*
* SPEC:
     [[U tWapController::Reset(TIME tServoTime) ]]
*
* RETURNS:
*    return value from tStepperMultiActuator::Move()
*
****/

ERROR_CODE tWapController::Reset(TIME tServoTime)
{  

   TraceMark(S_wapgizmo_WAP_CONTROLLER_RESET);
   
   /****
   [[U return := this->Collector.tMultiActuator::Move(tServoTime,tVelocity(0,0)) ]]   
   ****/
   return this->Collector.Move(tServoTime,tVelocity(0,0));   
}



/*******************************************************************
*
* tWapController::SendPosition
*
*  Stops collector motion
*
* SPEC:
     [[U tWapController::SendPosition() ]]
*
* RETURNS:
*    nothing
*
****/

void tWapController::SendPosition()
{
   tIpcPoint IpcOutput;

  
   //IPC output must be a simple structure
   /****
   [[U Position := this->Collector.tStepperMultiActuator::Position() ]]   
   ****/
   IpcOutput.dX = this->Collector.Position()[X];
   IpcOutput.dY = this->Collector.Position()[Y];

   //put position in data stream
   /**** 
   [[U this->ControllerData.
      CollectorPositionData.tDataStore::StoreData(IpcOutput)]]
   *****/
   this->ControllerData.CollectorPositionData.StoreData(IpcOutput);
}



/*******************************************************************
*
* tWapTrackController::tWapTrackController
*
*  Constructs tWapTrackController.
*
* SPEC:
    [[U tWapTrackController::tWapTrackController(CompleteIoSet,
                       ControllerDataToUse,tServoPeriod) ]]
*
* RETURNS:
*    nothing
*
****/

tWapTrackController::tWapTrackController(tWapCompleteIoSet& CompleteIoSet,
                                       tWapControllerData& ControllerDataToUse,
                                       TIME tServoPeriod) :
  
   /**** 
   [[U this->tWapController::tWapController(CompleteIoSet, ControllerDataToUse)]]
   *****/
   tWapController(CompleteIoSet, ControllerDataToUse),

   dGain(0),

   //save servo period in seconds
   CdServoPeriod(tServoPeriod / TIME_TICKS_PER_SECOND),

   Fsm(CompleteIoSet.Fsm)   
{  
   TraceMark(S_wapgizmo_WAP_TRACK_CONTROLLER_CONSTRUCTOR);
}


/*******************************************************************
*
* tWapTrackController::~tWapTrackController
*
*  Destroys tWapTrackController.
*
* SPEC:
    [[U tWapTrackController::~tWapTrackController(void) ]]
*
* RETURNS:
*    nothing
*
****/

tWapTrackController::~tWapTrackController(void) 
{
   TraceMark(S_wapgizmo_WAP_TRACK_CONTROLLER_DESTRUCTOR);
}


/*******************************************************************
*
* tWapTrackController::Calculate
*
*  Calculate function for WAP TRACK mode.  Called by servo
*     task at whatever rate it is running.  Attempts to drive the FSM to (0,0).
*     Reads FSM position from 
*     the star tracker, multiplies that by the negative of the gain, and uses
*     that number to command collector velocity.
*
* SPEC:
     [[U tWapTrackController::Calculate(tServoTime) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/

ERROR_CODE tWapTrackController::Calculate(TIME tServoTime) 
{
  ERROR_CODE ec;
  tVelocity Velocity;
  tIpcPoint IpcInput;

   if (this->CdServoPeriod < DBL_EPSILON) return S_wapgizmo_DIVIDE_BY_ZERO;

   TraceMark(S_wapgizmo_ENTERING_WAP_TRACK_CONTROLLER_CALCULATE);

   //get FSM position (which is error)
   /****
   [[U  this->Fsm.tIpcDblBufferAccessor::Get(IpcInput) ]]
   ****/
   ec = this->Fsm.Get(IpcInput);

   if (!ec) return ec;
   
   //move negative of error in the following servo period
   Velocity[X] = -dGain * IpcInput.dX / this->CdServoPeriod;   
   Velocity[Y] = -dGain * IpcInput.dY / this->CdServoPeriod;
   
   /****
   [[U this->Collector.tMultiActuator::Move(tServoTime,Velocity) ]]
   ****/
   ec = this->Collector.Move(tServoTime,Velocity);

   /****
   [[U  this->tWapController::SendPosition() ]]
   ****/
   this->SendPosition();
   
   TraceMark(S_wapgizmo_LEAVING_WAP_TRACK_CONTROLLER_CALCULATE);

  return ec;
}


           
/*******************************************************************
*tWapCalibrateController::tWapCalibrateController
*
*  Constructor.  Initializes the base class
*
* SPEC:
    [[U tWapCalibrateController::tWapCalibrateController(CompleteIoSet, ControllerDataToUse)]]
*
* RETURNS:
*    nothing
*
****/
tWapCalibrateController::tWapCalibrateController(tWapCompleteIoSet& CompleteIoSet,
                                       tWapControllerData& ControllerDataToUse) : 

   /**** 
   [[U this->tWapController::tWapController(CompleteIoSet, ControllerDataToUse)]]
   *****/
   tWapController(CompleteIoSet, ControllerDataToUse) {}
   

/*******************************************************************
*tWapCalibrateController::Calculate(TIME tTime)
*
*  Moves the calibrating actuator to the positio corresponding to tTime
*
* SPEC:
    [[U tWapCalibrateController::Calculate(TIME tTime)]]
*
* RETURNS:
*    ERROR_CODE
*
****/
                             
ERROR_CODE tWapCalibrateController::Calculate(TIME tTime)
{
   ERROR_CODE ec;
   
   /**** 
   [[U this->Collector.tMultiActuator::Dither(tTime)]]
   *****/
   ec =  this->Collector.Dither(tTime);

   /****
   [[U  this->tWapController::SendPosition() ]]
   ****/
   this->SendPosition();

   return ec;    
}

/*******************************************************************
*tWapSlewController::tWapSlewController
*
*  Constructor.  Initializes the base class, and initializes the paddle reference
*
* SPEC:
    [[U tWapSlewController::tWapSlewController(CompleteIoSet, ControllerDataToUse)]]
*
* RETURNS:
*    nothing
*
****/
tWapSlewController::tWapSlewController(tWapCompleteIoSet& CompleteIoSet,
                                       tWapControllerData& ControllerDataToUse) : 

   /**** 
   [[U this->tWapController::tWapController(CompleteIoSet, ControllerDataToUse)]]
   *****/
   tWapController(CompleteIoSet, ControllerDataToUse),
   
   Paddle(CompleteIoSet.Paddle) {}
   
  
/*******************************************************************
*tWapSlewController::Calculate(TIME tTime)
*
*  Moves the collector as specified by the paddle
*
* SPEC:
    [[U tWapSlewController::Calculate(TIME tTime)]]
*
* RETURNS:
*    ERROR_CODE
*
****/
                             
ERROR_CODE tWapSlewController::Calculate(TIME tTime)
{
   ERROR_CODE ec;
   tVelocity Velocity;
   
   /****
   [[U Velocity := this->Paddle.tPaddle::Update(tTime) ]]
   [[U this->Collector.tMultiActuator::Move(tTime,Velocity) ]]
   *****/
   ec =  this->Collector.Move(tTime,this->Paddle.Update(tTime));

   /****
   [[U  this->tWapController::SendPosition() ]]
   ****/
   this->SendPosition();

   return ec;  
}

  
