/* $RCSfile: wapGizmo.cc,v $ */
/* CM Info:   
$Header: /home/chara/cvsroot/ople/sys/gizmo/wap/wapGizmo.cc,v 1.1.1.1 1998/06/11 21:36:25 chara Exp $
*/
/* Russ Brill - Jet Propulsion Laboratory */


#include <float.h>

#include "iostream.h"
#include "time.h"
#include "taskLib.h"
#include "isit.h"
#include "baseObject.h"
#include "wapError.h"
#include "wapGizmo.h"
#include "wapGizmoInfo.h"
#include "wap/wapIoSet.h"
#include "wap/wapController.h"
#include "servo.h"
#include "wapCommands.h"
#include "controllerSet.h"
#include "trace.h"


/*******************************************************************
*
* tWapGizmo::tWapGizmo -- tWapGizmo constructor
*
*  Calls tRealtimeGizmo(base class) constructor, plus constructors
*     for I/O and controllers and status.  Does minimal error
*     checking on GizmoInfo parameters.  Sets up controllers,
*     assigning controller groups to servo modes and controllers
*     to controller groups.  Initializes servo mode to SLEW and then
*     calls tRealtimeGizmo::StartTasks() to complete initialization.
*
* SPEC:
      [[U tWapGizmo::tWapGizmo(GizmoInfo) ]]
*
* RETURNS:
*    nothing
*
****/

tWapGizmo::tWapGizmo(const tWapGizmoInfo& GizmoInfo) :

  // First call the base class constructor
  /****
     [[U tRealtimeGizmo::tRealtimeGizmo(GizmoInfo) ]]
   ****/ 
  tRealtimeGizmo(GizmoInfo),
  
  // Now construct the status block type tag
  /****
     [[U this->CWapStatusTag.tTypeTag::tTypeTag_STTNisz(WAP_STATUS, 
                    TN_WAP_STATUS, GizmoInfo.iGizmoId, sizeof(tWapStatus) ) ]]
   ****/ 
  CWapStatusTag(ST_STATUS, TN_WAP_STATUS, GizmoInfo.GetGizmoId(), 
                sizeof(tWapStatus) ),

  // Now construct the star position data block type tag
  /****
     [[U this->CWapDataTag.tTypeTag::tTypeTag_STTNisz(WAP_DATA, 
                TN_WAP_STAR_POSITION, GizmoInfo.iGizmoId, sizeof(tWapData) ) ]]
   ****/ 
  CWapStarPosDataTag(ST_DATA, TN_WAP_STAR_POSITION, GizmoInfo.GetGizmoId(), 
                 sizeof(tPoint) ),  



  // Now construct the Collector position data block type tag
  /****
     [[U this->CWapDataTag.tTypeTag::tTypeTag_STTNisz(WAP_DATA, 
                TN_WAP_COLLECTOR_POSITION, GizmoInfo.iGizmoId, sizeof(tPoint) ) ]]
   ****/ 
  CWapCollectorPosDataTag(ST_DATA, TN_WAP_COLLECTOR_POSITION, GizmoInfo.GetGizmoId(), 
                 sizeof(tPoint) ),  

  // Now construct the status block
  /****
     [[U this->Status.tWapStatus::tWapStatus(GizmoInfo.iGizmoId) ]]
   ****/ 
  Status(GizmoInfo.GetGizmoId() ),

  /****
     [[U this->CompleteIoSet.tWapCompleteIoSet::tWapCompleteIoSet(GizmoInfo) ]]
   ****/ 
  CompleteIoSet(GizmoInfo),

  // Now construct the controller set  (if new returns NULL, what happens here?)
  /****
  [[U   ServoInfo := this->tRealtimeGizmoInfo::GetServoInfo() ]]
  [[U   ServoTaskInfo := ServoInfo.tServoInfo::GetServoTaskInfo(0) ]]
  [[U   this->CompleteControllerSet.
        tWapCompleteControllerSet::tWapCompleteControllerSet(CompleteIoSet,
        CWapStatusTag,CWapStarPosDataTag, CWapCollectorPosDataTag,
        ServoTaskInfo.tTaskPeriod) ]]
   ****/ 
  CompleteControllerSet(CompleteIoSet,
    CWapStatusTag,CWapStarPosDataTag,CWapCollectorPosDataTag, 
    GizmoInfo.GetServoInfo().GetServoTaskInfo(STP_BASIC).tTaskPeriod),
    
    WapGizmoInfo(GizmoInfo)

 {
    ERROR_CODE ec;
    tPoint HardStop;

    /****
    [[U GizmoInfo.tWapGizmoInfo::PhysicalConstants() ]]
    ****/
    const tWapPhysicalConstants& PhysicalConstants = GizmoInfo.PhysicalConstants();
    
    TraceMark(S_wapgizmo_ENTERING_WAP_GIZMO_CONSTRUCTOR);

    // If we had a error in base class constructor, exit with error 

    if ((ec = this->tBaseObject::GetInitError()) != OK) {
      TraceMsg(tmt_ERROR,ec,"tWapGizmo::tWapGizmo");
      return;
    }

    // If we had a error in allocating members, exit with error

    if ((&CompleteIoSet==NULL) || (&CompleteControllerSet==NULL))
    {
      this->SetInitError(S_wapgizmo_CANT_CONSTRUCT_IO_OR_CTLR_SET);
      TraceMsg(tmt_ERROR,S_wapgizmo_CANT_CONSTRUCT_IO_OR_CTLR_SET,
                       "tWapGizmo::tWapGizmo");
      return;
    }


    // If GizmoInfo doesn't think we're a WAP, exit with error 

    /****
    [[U this->GizmoInfo.tGizmoInfo::GetGizmoType() ]]
    ****/ 
    if ( GizmoInfo.GetGizmoType() != GT_WAP ) {

      this->SetInitError(S_wapgizmo_GIZMO_INFO_NOT_WAP);
      TraceMsg(tmt_ERROR,S_wapgizmo_GIZMO_INFO_NOT_WAP,
                "tWapGizmo::tWapGizmo" );
      return;
    }
  

    // Sanity Checks until we can better automate the controller setup.


    // Code currently expects 3 servo modes 

    /****
    [[U this->GizmoInfo.tRealtimeGizmoInfo::GetNumberOfServoModes() ]]
    ****/ 
    if ( GizmoInfo.GetNumberOfServoModes() != WSS_NUM_SERVO_MODES) {

      this->SetInitError(S_wapgizmo_INVALID_NUMBER_OF_SERVO_MODES);
      TraceMsg(tmt_ERROR,S_wapgizmo_INVALID_NUMBER_OF_SERVO_MODES,
                "tWapGizmo::tWapGizmo" );
      return;
    }
    
    /****
    [[U this->GizmoInfo.tRealtimeGizmoInfo::GetNumberOfServoPartitions() ]]
    ****/ 
    // Code currently expects only 1 servo partition 

    if (GizmoInfo.GetNumberOfServoPartitions() != STP_NUM_PARTITIONS ) {

      this->SetInitError(S_wapgizmo_INVALID_NUMBER_OF_SERVO_PARTITIONS);
      TraceMsg(tmt_ERROR,S_wapgizmo_INVALID_NUMBER_OF_SERVO_PARTITIONS,
                "tWapGizmo::tWapGizmo" );
      return;
    }
    
    TraceMark(S_wapgizmo_ASSIGNING_CTLR_GROUPS_TO_MODES);


    // Assign Controller Groups to Servo Modes in Controller Set
    /****
       [[U OffGroup := this->CompleteControllerSet.
         tWapCompleteControllerSet::OffControllerGroup() ]]
       [[U ec := this->ControllerSet.tControllerSet::AssignControllerGroup 
              (WSM_OFF,OffGroup) ]] 
     ****/ 
    tSimpleControllerGroup& OffGroup = 
      this->CompleteControllerSet.OffControllerGroup();
    ec = this->ControllerSet.AssignControllerGroup(WSM_OFF,OffGroup);
    if (ec != OK) {
      TraceMsg(tmt_ERROR,ec,"tWapGizmo::tWapGizmo" );
      this->SetInitError(ec);
      return;
    }

    /****
       [[U SlewGroup := this->CompleteControllerSet.
         tWapCompleteControllerSet::SlewControllerGroup() ]]
       [[U ec := this->ControllerSet.tControllerSet::AssignControllerGroup 
              (WSM_OFF,SlewGroup) ]] 
     ****/ 
    tSimpleControllerGroup& SlewGroup = 
      this->CompleteControllerSet.SlewControllerGroup();
    ec = this->ControllerSet.AssignControllerGroup(WSM_OFF,SlewGroup);
    if (ec != OK) {
      TraceMsg(tmt_ERROR,ec,"tWapGizmo::tWapGizmo" );
      this->SetInitError(ec);
      return;
    }

    /****
       [[U TrackGroup := this->CompleteControllerSet.
         tWapCompleteControllerSet::TrackControllerGroup() ]]
       [[U ec := this->ControllerSet.tControllerSet::AssignControllerGroup 
              (WSM_TRACK,TrackGroup) ]]
     ****/ 
    tSimpleControllerGroup& TrackGroup = 
      this->CompleteControllerSet.TrackControllerGroup();
    ec = this->ControllerSet.AssignControllerGroup(WSM_TRACK,TrackGroup);
              
    if (ec != OK) {
      TraceMsg(tmt_ERROR,ec,"tWapGizmo::tWapGizmo" );
      this->SetInitError(ec);
      return;
    }
  
    /****
       [[U CalibrateGroup := this->CompleteControllerSet.
         tWapCompleteControllerSet::CalibrateControllerGroup() ]]
       [[U ec := this->ControllerSet.tControllerSet::AssignControllerGroup 
              (WSM_CALIBRATE,CalibraterGroup) ]]
     ****/ 
    tSimpleControllerGroup& CalibrateGroup = 
      this->CompleteControllerSet.CalibrateControllerGroup();
    ec = this->ControllerSet.AssignControllerGroup(WSM_CALIBRATE,CalibrateGroup);

    if (ec != OK) {
      TraceMsg(tmt_ERROR,ec,"tWapGizmo::tWapGizmo" );
      this->SetInitError(ec);
      return;
    } 

    // Select Off Mode
    /****
       [[U ec := this->Servo.tServo::SelectMode(WSM_OFF) ]] 
     ****/ 
    ec = this->Servo.SelectMode(WSM_OFF);
    
    if (ec != OK) {
      TraceMsg(tmt_ERROR,ec,"tWapGizmo::tWapGizmo" );
      this->SetInitError(ec);
      return;
    }

    //command Collector with default servo parameters
    /****
    [[U ServoParams := GizmoInfo.tWapGizmoInfo::ServoParams() ]]
    [[U this->tWapGizmo::ChangeServoParams(ServoParams)]]
    ****/
    this->ChangeServoParams(GizmoInfo.ServoParams());

    //send home parameters to Collector
    HardStop[X] = PhysicalConstants.dHardStopPosition[X];
    HardStop[Y] = PhysicalConstants.dHardStopPosition[Y];
    
    /****
    this->CompleteIoSet.Collector.tStepperMultiActuator::SetHomeParams(HardStop,
                                                PhysicalConstants.dHardStopDuration,
                                                PhysicalConstants.dHomeDuration) ]]
    ****/                                             
    this->CompleteIoSet.Collector.SetHomeParams(HardStop,
                                                PhysicalConstants.dHardStopDuration,
                                                PhysicalConstants.dHomeDuration);
    
    TraceMark(S_wapgizmo_ENABLING_BASE_CLASS_TASKS);

    // Enable Base class tasks ...
    /****
       [[U ec := this->tRealtimeGizmo::StartTasks(GizmoInfo) ]] 
     ****/ 
    this->tRealtimeGizmo::StartTasks(GizmoInfo);

    if (ec != OK) {
      TraceMsg(tmt_ERROR,ec,"tWapGizmo::tWapGizmo" );
      this->SetInitError(ec);
      return;
    }

    TraceMark(S_wapgizmo_LEAVING_WAP_GIZMO_CONSTRUCTOR);

}

/******************************************************************
*
* tWapGizmo::~tWapGizmo -- tWapGizmo destructor
*
*  Destroys tWapGizmo object.  (calls base class destructor)
*
* SPEC:
      [[U tWapGizmo::~tWapGizmo(void) ]]
*
* RETURNS:
*    nothing
*
****/
tWapGizmo::~tWapGizmo(void) 

  /****
     [[U tRealtimeGizmo::~tRealtimeGizmo(void) ]]
  ****/
{ 
   TraceMark(S_wapgizmo_WAP_GIZMO_DESTRUCTOR);
}

/******************************************************************
*
* tWapGizmo::Restart 
*
*  Calls base class Restart().
*
* SPEC:
      [[U tWapGizmo::Restart(void) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/
ERROR_CODE tWapGizmo::Restart(void) 
{
   TraceMark(S_wapgizmo_RESTART);

  /****
     [[U return := this->tRealtimeGizmo::Restart(void) ]]
  ****/
  return this->tRealtimeGizmo::Restart();
}

/******************************************************************
*
* tWapGizmo::Shutdown 
*
*  Calls base class Shutdown().
*
* SPEC:
      [[U tWapGizmo::Shutdown(void) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/

ERROR_CODE tWapGizmo::Shutdown(void) 
{
   TraceMark(S_wapgizmo_SHUTDOWN);

  /****
     [[U return := this->tRealtimeGizmo::Shutdown(void) ]]
  ****/
  return this->tRealtimeGizmo::Shutdown();
}
 
/******************************************************************
*
* tWapGizmo::IsOk 
*
*  Calls base class IsOk().
*
* SPEC:
      [[U tWapGizmo::IsOk(void) ]]
*
* RETURNS:
*    BOOLEAN
*
****/

BOOL tWapGizmo::IsOk(void) 
{
  /****
     [[U return := this->tRealtimeGizmo::IsOk(void) ]]
  ****/
  return this->tRealtimeGizmo::IsOk();
}

/******************************************************************
*
* tWapGizmo::operator<< 
*
*  Prints out "Star Tracker Gizmo" followed by results of
*     calling base class operator<<
*
* SPEC:
      [[U tWapGizmo::operator_extractor(ostream& stream, tWapGizmo& Gizmo) ]]
*
* RETURNS:
*     stream
*
****/

ostream& operator<<(ostream& stream, tWapGizmo& Gizmo) 
{
  stream << "Wide Angle Pointing Gizmo" << endl;

  /****
     [[U this->tRealtimeGizmo::operator_extractor(stream,Gizmo) ]]
  ****/
  stream << (tRealtimeGizmo &) Gizmo;
  return stream;
}


/******************************************************************
*
* tWapGizmo::CollectStatus 
*
* Gets status info(star position, image) from rest of Star Tracker 
*    Gizmo, then sends it to RTE status stream
*
* SPEC:
      [[U tWapGizmo::CollectStatus(void) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/

ERROR_CODE tWapGizmo::CollectStatus(void) 
{
  ERROR_CODE ec;
  tImage StarImage;

   TraceMark(S_wapgizmo_ENTERING_COLLECTSTATUS);

  // Get current time stamp - do first; more likely to match image
  this->Status.tTime = ::GetTime();

  // Put latest star position coords into ST status block
  /****
     [[U this->Status.StarPosition := 
                  this->CompleteControllerSet.
          tWapCompleteControllerSet::GetStarPosition(void) ]]
  ****/
  this->Status.StarPosition.dXPosition = 
    this->CompleteControllerSet.GetStarPosition().x(); 
  this->Status.StarPosition.dYPosition = 
    this->CompleteControllerSet.GetStarPosition().y();


   TraceMark(S_wapgizmo_SENDING_STATUS_BLOCK);

  // Send the status block to RTE status stream
  /****
     [[U ec := tGizmo::SendStatus_tp(CWapStatusTag, &this->Status) ]]
  ****/
  ec = this->SendStatus(CWapStatusTag,(BYTE *) &this->Status);
  
  // Handle error if any
  /****
     [[U ec := tGizmo::SendError_ec (S_wapgizmo_UNABLE_TO_SEND_STATUS) ]]
  ****/
  if (ec != OK) 
     this->SendError (S_wapgizmo_UNABLE_TO_SEND_STATUS);

   TraceMark(S_wapgizmo_LEAVING_COLLECTSTATUS);

  return OK;
}
  
/******************************************************************
*
* tWapGizmo::Dispatch
*
* Calls the appropriate tWapGizmo method that corresponds to the
*    command icNewCmd.  Note that this early version of the ST
*    does not use command engines, just function calls, since 
*    none of the commands requires any mutual exclusion because
*    all current commands execute and return immediately.
*
* SPEC:
*    [[U tWapGizmo::Dispatch(const tIsitCmd& icNewCmd) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/

ERROR_CODE tWapGizmo::Dispatch(const tIsitCmd& icNewCmd) 
{

   TraceMark(S_wapgizmo_ENTERING_DISPATCH);

  switch(icNewCmd.Command) {

    case WAP_CMD_PING : 
       /****
          [[U return := tWapGizmo::Ping(void) ]]
       ****/
       return this->Ping();

    case WAP_CMD_SLEW : 
       /****
          [[U return := tWapGizmo::Slew(void) ]]
       ****/
       return this->Slew();

    case WAP_CMD_OFF : 
       /****
          [[U return := tWapGizmo::Off(void) ]]
       ****/
       return this->Off();

    case WAP_CMD_TRACK : 
       /****
          [[U return := tWapGizmo::Track(void) ]]
       ****/
       return this->Track();

    case WAP_CMD_CALIBRATE : 
       /****
          [[U return := tWapGizmo::Calibrate((tWapCalibrateCmd&)icNewCmd) ]]
       ****/
       return this->Calibrate((tWapCalibrateCmd&)icNewCmd);

    case WAP_CMD_SETSS : 
       /****
          [[U return := tWapGizmo::SetServoParams((tWapSetServoParamsCmd&)icNewCmd) ]]
       ****/
       return this->SetServoParams((tWapSetServoParamsCmd&)icNewCmd);

    case WAP_CMD_HOME : 
       /****
          [[U return := tWapGizmo::Home() ]]
       ****/
       return this->Home();

    case WAP_CMD_MOVE : 
       /****
          [[U return := tWapGizmo::Move((tWapMoveCmd&)icNewCmd) ]]
       ****/
       return this->Move((tWapMoveCmd&)icNewCmd);

    default : 
       return S_wapgizmo_UNRECOGNIZED_COMMAND;
  }

   TraceMark(S_wapgizmo_LEAVING_DISPATCH);

   return OK;
}
  
    
/******************************************************************
*
* tWapGizmo::Ping
*
* Called in response to the ST receiving a PING command.  Sends a 
*     friendly message to the RTE error stream.
*
* SPEC:
     [[U  tWapGizmo::Ping(void) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/

ERROR_CODE tWapGizmo::Ping(void) 
{
   TraceMark(S_wapgizmo_PING);

   /****
   [[U  return := tGizmo::SendError_ec(S_wapgizmo_STAR_TRACKER_ALIVE) ]]
   ****/
   return SendError(S_wapgizmo_WAP_ALIVE);
}

   
/******************************************************************
*
* tWapGizmo::Off
*
* Called in response to the ST receiving an OFF command.  Sets the
*     servo mode to OFF, as well as updating information in status
*     and error streams about executing the command.
*
* SPEC:
     [[U tWapGizmo::Off(void) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/

ERROR_CODE tWapGizmo::Off(void) 
{
  ERROR_CODE ec;

   TraceMark(S_wapgizmo_ENTERING_OFF_METHOD);


  // Tell the servo to change mode to OFF
  /****
  [[U ec := this->Servo.tServo::SelectMode(WSM_OFF) ]]
  ****/
  ec = this->Servo.SelectMode(WSM_OFF);

  if (ec == OK) {

    /****
    [[U this->Status.tWapStatus::UpdateServoState (WSS_OFF) ]]
    ****/
    this->Status.UpdateServoState (WSS_OFF);


    // Report status: we finished the command 
    /****
    [[U this->Status.tWapStatus::UpdateCommandStatus(WAPCS_OFF) ]]
    ****/
    this->Status.UpdateCommandStatus(WAPCS_OFF);
  
    /****
    [[U this.tGizmo::SendError_ec(S_wapgizmo_OFF_COMPLETE) ]]
    ****/
    this->SendError(S_wapgizmo_OFF_COMPLETE);
   
  }
  
  TraceMark(S_wapgizmo_LEAVING_OFF_METHOD);

  return ec; 
}

/******************************************************************
*
* tWapGizmo::Slew
*
* Called in response to the ST receiving an SLEW command.  Sets the
*     servo mode to SLEW, as well as updating information in status
*     and error streams about executing the command.
*
* SPEC:
     [[U tWapGizmo::Slew(void) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/

ERROR_CODE tWapGizmo::Slew(void) 
{
  ERROR_CODE ec;
   tWapCmdEngine *pCmd;

   TraceMark(S_wapgizmo_ENTERING_SLEW_METHOD);

  /****
   [[U tWapCmdEngine::tWapCmdEngine(this,"WAP Command Engine") ]]
   ****/
   pCmd = new tWapCmdEngine(this,"WAP Command Engine");

   //take the mutex
   /****
   [[U this->tMutexUser::Execute(NO_WAIT,TRUE,MUTEX_DEFAULT_TIMEOUT) ]]
   ****/
   ec = pCmd->Execute();
   
   if (!ec) return ec;
   
  // Tell the servo to change mode to SLEW
  /****
  [[U ec := this->Servo.tServo::SelectMode(WSM_SLEW) ]]
  ****/
  ec = this->Servo.SelectMode(WSM_SLEW);

  if (ec == OK) {

    /****
    [[U this->Status.tWapStatus::UpdateServoState (WSS_SLEW) ]]
    ****/
    this->Status.UpdateServoState (WSS_SLEW);


    // Report status: we finished the command 
    /****
    [[U this->Status.tWapStatus::UpdateCommandStatus(WAPCS_SLEW) ]]
    ****/
    this->Status.UpdateCommandStatus(WAPCS_SLEW_ACTIVE);
  
    /****
    [[U this.tGizmo::SendError_ec(S_wapgizmo_SLEW_COMPLETE) ]]
    ****/
    this->SendError(S_wapgizmo_SLEW_COMPLETE);
   
  }
  
  TraceMark(S_wapgizmo_LEAVING_SLEW_METHOD);

  return ec; 
}

  
/******************************************************************
*
* tWapGizmo::Track
*
* Called in response to the ST receiving an TRACK command.  Sets the
*     servo mode to TRACK, as well as updating information in status
*     and error streams about executing the command.
*
* SPEC:
     [[U  tWapGizmo::Track(void) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/

ERROR_CODE tWapGizmo::Track(void) 
{
  ERROR_CODE ec;
   tWapCmdEngine *pCmd;

   TraceMark(S_wapgizmo_ENTERING_TRACK_METHOD);

  /****
   [[U tWapCmdEngine::tWapCmdEngine(this,"WAP Command Engine") ]]
   ****/
   pCmd = new tWapCmdEngine(this,"WAP Command Engine");

   //take the mutex
   /****
   [[U this->tMutexUser::Execute(NO_WAIT,TRUE,MUTEX_DEFAULT_TIMEOUT) ]]
   ****/
   ec = pCmd->Execute();
   
   if (!ec) return ec;
   
  // Tell the servo to change mode to TRACK
  /****
  [[U ec := this->Servo.tServo::SelectMode(WSM_TRACK) ]]
  ****/
  ec = this->Servo.SelectMode(WSM_TRACK);
  
  if (ec == OK) {

    /****
    [[U this->Status.tWapStatus::UpdateServoState (WSS_TRACK) ]]
    ****/
    this->Status.UpdateServoState (WSS_TRACK);

    // Report status: we finished the command(but stay in TRACK mode) 
    /****
    [[U this->Status.tWapStatus::UpdateCommandStatus(WAPCS_TRACK_ACTIVE) ]]
    ****/
    this->Status.UpdateCommandStatus(WAPCS_TRACK_ACTIVE);

    /****
    [[U this->tGizmo::SendError_ec(S_wapgizmo_TRACK_IN_PROGRESS) ]]
    ****/
    this->SendError(S_wapgizmo_TRACK_IN_PROGRESS);
  }

   TraceMark(S_wapgizmo_LEAVING_TRACK_METHOD);

  return ec; 
}



/******************************************************************
*
* tWapGizmo::Calibrate
*
* Called in response to the ST receiving a CALIBRATE command.  Sets the
*     servo mode to CALIBRATE, as well as updating information in status
*     and error streams about executing the command.
*
* SPEC:
     [[U  tWapGizmo::Calibrate(const tFtCalibrateCmd& icNewCmd) ]]
*
* RETURNS:
*    ERROR_CODE
*
****/

ERROR_CODE tWapGizmo::Calibrate(const tWapCalibrateCmd& icNewCmd) 
{
  ERROR_CODE ec;
   tWapCmdEngine *pCmd;
   
   TraceMark(S_wapgizmo_ENTERING_CALIBRATE_METHOD);

  /****
   [[U tWapCmdEngine::tWapCmdEngine(this,"WAP Command Engine") ]]
   ****/
   pCmd = new tWapCmdEngine(this,"WAP Command Engine");

   //take the mutex
   /****
   [[U this->tMutexUser::Execute(NO_WAIT,TRUE,MUTEX_DEFAULT_TIMEOUT) ]]
   ****/
   ec = pCmd->Execute();
   
   if (!ec) return ec;
   
  // Tell the servo to change mode to CALIBRATE
  /****
  [[U ec := this->Servo.tServo::SelectMode(WSM_CALIBRATE) ]]
  ****/
  ec = this->Servo.SelectMode(WSM_CALIBRATE);
  
  if (ec == OK) {

    /****
    [[U this->Status.tWapStatus::UpdateServoState (WSS_CALIBRATE) ]]
    ****/
    this->Status.UpdateServoState (WSS_CALIBRATE);

    // Report status: we finished the command(but stay in CALIBRATE mode) 
    /****
    [[U this->Status.tWapStatus::UpdateCommandStatus(WAPCS_CALIBRATE_ACTIVE) ]]
    ****/
    this->Status.UpdateCommandStatus(WAPCS_CALIBRATE_ACTIVE);

    /****
    [[U this->tGizmo::SendError_ec(S_wapgizmo_CALIBRATE_IN_PROGRESS) ]]
    ****/
    this->SendError(S_wapgizmo_CALIBRATE_IN_PROGRESS);

    /****
    [[U this->CompleteIoSet.tWapCompleteIoSet::
      SetCalParams(icNewCmd) ]]
    ****/
    this->CompleteIoSet.SetCalParams(icNewCmd);
  }

   TraceMark(S_wapgizmo_LEAVING_CALIBRATE_METHOD);

  return ec; 
}



/******************************************************************
*
* tWapGizmo::SetServoParams
*
* Called in response to the ST receiving a SETSERVOPARAMETERS command. It 
* updates information in status
* and error streams about executing the command.
*
* SPEC:
     [[U  tWapGizmo::SetServoParams(const tWapSetServoParametersCmd& icNewCmd) ]]
*
* RETURNS:
*    OK
*
****/

ERROR_CODE tWapGizmo::SetServoParams(const tWapSetServoParamsCmd& icNewCmd) 
{

  TraceMark(S_wapgizmo_ENTERING_SSP_METHOD);

  /****
  [[U  this->tWapGizmo::ChangeServoParams(icNewCmd.Params)]]
  ****/
  this->ChangeServoParams(icNewCmd.Params);
  
  TraceMark(S_wapgizmo_LEAVING_SSP_METHOD);
  
  return OK;
}



/******************************************************************
*
* tWapGizmo::ChangeServoParams
*
* It updates information in status and error streams about executing the 
* command.
*
* SPEC:
     [[U  tWapGizmo::ChangeServoParams(const tWapServoParameters& Params) ]]
*
* RETURNS:
*    nothing
*
****/

void tWapGizmo::ChangeServoParams(const tWapServoParams& Params) 
{
  tVector Home(3);
  tMatrix Xform(2,3);
  int iRow,iColumn;
  
    //copy from command structure to tVector, tMatrix
    for (iColumn = 0; iColumn < 3; iColumn++) {
      Home[iColumn] = Params.Home[iColumn];
      for (iRow = 0; iRow < 2; iRow++) {
         Xform[iRow][iColumn] = Params.XformDetToFsm[iRow][iColumn];
      }
    }
         
    /****
    [[U this->CompleteIoSet.tWapCompleteIoSet::
      SetXformParams(Xform,Home) ]]
    ****/
    this->CompleteIoSet.SetXformParams(Xform,Home);

    /****
    [[U this->CompleteControllerSet.tWapCompleteControllerSet::
      SetControllerParams(Params.ControllerParams) ]]
    ****/
    this->CompleteControllerSet.SetControllerParams(Params.ControllerParams);
}



/******************************************************************
*
* tWapGizmo::Home
*
* Called in response to the ST receiving a HOME command.  Moves
* the FSM to the home position.
*
* SPEC:
     [[U  tWapGizmo::Home() ]]
*
* RETURNS:
*    error if mode selection or motion fails
*
****/

ERROR_CODE tWapGizmo::Home() 
{
   ERROR_CODE ec,ecMode;
   tWapHomeCmdEngine *pCmd;
   
   TraceMark(S_wapgizmo_ENTERING_HOME_METHOD);

  // Tell the servo to change mode to OFF
   /****
   [[U ec := this->Servo.tServo::SelectMode(WSM_OFF) ]]
   ****/
   ec = this->Servo.SelectMode(WSM_OFF);

   if (!ec) return ec;
   
   /****
   [[U this->Status.tWapStatus::UpdateServoState (WSS_OFF) ]]
   ****/
   this->Status.UpdateServoState (WSS_OFF); 
   
  /****
   [[U tWapHomeCmdEngine::tWapHomeCmdEngine(this) ]]
   ****/
   pCmd = new tWapHomeCmdEngine(this);

   //move the collector
   /****
   [[U this->tMutexUser::Execute(NO_WAIT,TRUE,MUTEX_DEFAULT_TIMEOUT) ]]
   ****/
   ec = pCmd->Execute();
      
  // Tell the servo to change mode to SLEW, even if command engine error
   /****
   [[U ecMode := this->Servo.tServo::SelectMode(WSM_SLEW) ]]
   ****/
   ecMode = this->Servo.SelectMode(WSM_SLEW);

   if (!ec) return ec;
   if (!ecMode) return ecMode;
   
   /****
   [[U this->Status.tWapStatus::UpdateServoState (WSS_SLEW) ]]
   ****/
   this->Status.UpdateServoState (WSS_SLEW); 
    
  TraceMark(S_wapgizmo_LEAVING_HOME_METHOD);

  return OK; 
}



/******************************************************************
*
* tWapGizmo::Move
*
* Called in response to the ST receiving a MOVE command.  Moves
* the Collector to the specified position.
*
* SPEC:
     [[U  tWapGizmo::Move(icNewCmd) ]]
*
* RETURNS:
*    error if mode selection fails or motion fails
*
****/

ERROR_CODE tWapGizmo::Move(const tWapMoveCmd& icNewCmd) 
{
   ERROR_CODE ec,ecMode;
   tWapMoveCmdEngine *pCmd;
   
   TraceMark(S_wapgizmo_ENTERING_MOVE_METHOD);

   // Tell the servo to change mode to OFF
   /****
   [[U ec := this->Servo.tServo::SelectMode(WSM_OFF) ]]
   ****/
   ec = this->Servo.SelectMode(WSM_OFF);

   if (!ec) return ec;
   
   /****
   [[U this->Status.tWapStatus::UpdateServoState (WSS_OFF) ]]
   ****/
   this->Status.UpdateServoState (WSS_OFF); 
       
   /****
   [[U tWapMoveCmdEngine::tWapMoveCmdEngine(this,
      tPoint(icNewCmd.Delta[X],icNewCmd.Delta[Y])) ]]
   ****/
   pCmd = new tWapMoveCmdEngine(this,
                                tPoint(icNewCmd.Delta[X],icNewCmd.Delta[Y]));

   //move the collector
   /****
   [[U this->tMutexUser::Execute(NO_WAIT,TRUE,MUTEX_DEFAULT_TIMEOUT) ]]
   ****/
   ec = pCmd->Execute();
   
   // Tell the servo to change mode to SLEW, even if command engine error
   /****
   [[U ecMode := this->Servo.tServo::SelectMode(WSM_SLEW) ]]
   ****/
   ecMode = this->Servo.SelectMode(WSM_SLEW);

   if (!ec) return ec;
   if (!ecMode) return ecMode;
   
   /****
   [[U this->Status.tWapStatus::UpdateServoState (WSS_SLEW) ]]
   ****/
   this->Status.UpdateServoState (WSS_SLEW); 
          
   TraceMark(S_wapgizmo_LEAVING_HOME_METHOD);

  return OK; 
}


/******************************************************************
*
* tWapCmdEngine::tWapCmdEngine
*
* Constructor.  It initializes the base class
*
* SPEC:
     [[U tWapCmdEngine::tWapCmdEngine(tWapGizmo *pGizmoToUse,
                                       char* sTaskName) ]]
*
* RETURNS:
*    nothing
*
****/
tWapCmdEngine::tWapCmdEngine(tWapGizmo *pGizmoToUse,
                             char* sTaskName) :

   /****
   [[U tMutexUser::tMutexUser_tts(&pGizmoToUse->CollectorMutex,pGizmoToUse,
                                  sTaskName) ]]
   ****/
   tMutexUser(&pGizmoToUse->CollectorMutex,pGizmoToUse,sTaskName)
{}


/******************************************************************
*
* tWapMoveCmdEngine::tWapMoveCmdEngine
*
* Constructor.  It initializes the base class and saves DeltaToUse.
*
* SPEC:
     [[U tWapMoveCmdEngine::tWapMoveCmdEngine(tWapGizmo *pGizmoToUse,
                            const tPoint DeltaToUse) ]]
*
* RETURNS:
*    nothing
*
****/
tWapMoveCmdEngine::tWapMoveCmdEngine(tWapGizmo *pGizmoToUse,
                                     const tPoint& DeltaToUse) :

   /****
   [[U tWapCmdEngine::tWapCmdEngine(pGizmoToUse,"WAP MOVE Command Engine") ]]
   ****/
   tWapCmdEngine(pGizmoToUse,"WAP MOVE Command Engine"),
   Delta(DeltaToUse)
{}


/******************************************************************
*
* tWapMoveCmdEngine::ExecuteMethod()
*
* Moves collector to the previously specified position
*
* SPEC:
     [[U tWapMoveCmdEngine::ExecuteMethod()]]
*
* RETURNS:
*    error if Move() or taskDelay fails
*
****/
ERROR_CODE tWapMoveCmdEngine::ExecuteMethod()
{
   tVelocity Velocity;
   double dDuration;
   ERROR_CODE ec;
   STATUS Status;
   
   //get duration of move
   /****
   [[U pWapGizmo->WapGizmoInfo.tWapGizmoInfo::PhysicalConstants() ]]
   ****/
   dDuration = pWapGizmo->WapGizmoInfo.PhysicalConstants().dMoveDuration;
   
   if (dDuration < DBL_EPSILON) return S_wapgizmo_DIVIDE_BY_ZERO;

   Velocity = Delta / dDuration; 
   
   //command collector at required velocity
   /****
   [[U pWapGizmo->CompleteIoSet.Collector.tMultiActuator::Move(0,Velocity) ]]
   ****/
   ec = pWapGizmo->CompleteIoSet.Collector.Move(0,Velocity);
   
   if (!ec) return ec;

   //wait for specified duration
   Status = taskDelay((int)(dDuration * CLOCKS_PER_SEC));

   if (Status != OK) return S_wapgizmo_TASK_DELAY_FAILED;

   Velocity.clear();

   //stop the collector
   /****
   pWapGizmo->CompleteIoSet.Collector.tMultiActuator::Move(0,Velocity.clear()) ]]
   ****/
   return pWapGizmo->CompleteIoSet.Collector.Move(0,Velocity);   
}

   
/******************************************************************
*
* tWapHomeCmdEngine::tWapHomeCmdEngine
*
* Constructor.  It initializes the base class
*
* SPEC:
     [[U tWapHomeCmdEngine::tWapHomeCmdEngine(tWapGizmo *pGizmoToUse) ]]
*
* RETURNS:
*    nothing
*
****/
tWapHomeCmdEngine::tWapHomeCmdEngine(tWapGizmo *pGizmoToUse) :

   /****
   [[U tWapCmdEngine::tWapCmdEngine(pGizmoToUse,"WAP HOME Command Engine") ]]
   ****/
   tWapCmdEngine(pGizmoToUse,"WAP HOME Command Engine")
{}
   
   
/******************************************************************
*
* tWapHomeCmdEngine::ExecuteMethod()
*
* Moves collector to the home position
*
* SPEC:
     [[U tWapHomeCmdEngine::ExecuteMethod()]]
*
* RETURNS:
*    error if Home() fails
*
****/
ERROR_CODE tWapHomeCmdEngine::ExecuteMethod()
{
   /****
   pWapGizmo->CompleteIoSet.Collector.tStepperMultiActuator::Home()
   ****/
   return pWapGizmo->CompleteIoSet.Collector.Home();
}

   
   
