#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
int xvAllocColor( Display *dp, Colormap cm, XColor *cdef);

#define XV 
#include "xvUtils.h"

/***************************************************/
void xvbcopy(char *src, char *dst, size_t len)
{
  /* Modern OS's (Solaris, etc.) frown upon the use of bcopy(),
   * and only want you to use memcpy().  However, memcpy() is broken,
   * in the sense that it doesn't properly handle overlapped regions
   * of memory.  This routine does, and also has its arguments in
   * the same order as bcopy() did, without using bcopy().
   */

  /* determine if the regions overlap
   *
   * 3 cases:  src=dst, src<dst, src>dst
   *
   * if src=dst, they overlap completely, but nothing needs to be moved
   * if src<dst and src+len>dst then they overlap
   * if src>dst and src<dst+len then they overlap
   */

  if (src==dst || len<=0) return;    /* nothin' to do */

  if (src<dst && src+len>dst) {  /* do a backward copy */
    src = src + len - 1;
    dst = dst + len - 1;
    for ( ; len>0; len--, src--, dst--) *dst = *src;
  }

  else {  /* they either overlap (src>dst) or they don't overlap */
    /* do a forward copy */
    for ( ; len>0; len--, src++, dst++) *dst = *src;
  }
}
   
void xvFreeColors( Display *dp, 
     Colormap cm, 
     unsigned long *pixels,
     int npixels,
     unsigned long planes)
{

  // if ( (theVisual->class != TrueColor) && (theVisual->class != DirectColor) )
    XFreeColors(dp, cm, pixels, npixels, planes);
}


void SortColormap(byte *pic, int pwide, int phigh, int *pnumcols, 
                  byte *rmap,byte *gmap,byte *bmap, byte *order, byte *trans)
{
  /* operates on 8-bit images.  sorts the colormap into 'best' order
   * 'order' is the 'best' order to allocate the colors.  'trans' is a 
   * transformation to be done to pic, cpic, and epic (in PIC8 mode) to
   * compress the colormap
   */

  byte *p;
  int   i, j, mdist, entry, d, hist[256], ncols;
  static CMAPENT c[256], c1[256], *cp, *cj, *ck;

  /* init some stuff */
  for (i=0; i<256; i++) order[i]=i;

  /* initialize histogram and compute it */
  for (i=0; i<256; i++) hist[i]=0;
  for (i=pwide*phigh, p=pic; i; i--, p++) hist[*p]++;
  
  
  /* put the actually-used colors into the 'c' array in the order they occur
     also, while we're at it, calculate ncols, and close up gaps in
     colortable */
  
  for (i=ncols=0; i<256; i++) {
    if (hist[i]) { 
      rmap[ncols] = rmap[i];
      gmap[ncols] = gmap[i];
      bmap[ncols] = bmap[i];
      trans[i]    = (byte) ncols;

      cp = &c[ncols];
      cp->r = rmap[i];    cp->g = gmap[i];    cp->b = bmap[i];
      cp->use = hist[i];  cp->oldindex = ncols;
      cp->mindist = 1000000; /* 255^2 * 3 = 195075 */
      ncols++;
    }
  }


  /* find most-used color, put that in c1[0] */
  entry = -1;  mdist = -1;
  for (i=0; i<ncols; i++) {
    if (c[i].use > mdist) { mdist = c[i].use;  entry=i; }
  }
  xvbcopy((char *) &c[entry], (char *) &c1[0], sizeof(CMAPENT));
  c[entry].use = 0;   /* dealt with */
  
  
  /* sort rest of colormap.  Half of the entries are allocated on the
     basis of distance from already allocated colors, and half on the
     basis of usage.  (NB: 'taxicab' distance is used throughout this file.)

     Mod:  pick first 10 colors based on maximum distance.  pick remaining
     colors half by distance and half by usage   -- JHB

     To obtain O(n^2) performance, we keep each unselected color
     (in c[], with use>0) marked with the minimum distance to any of
     the selected colors (in c1[]).  Each time we select a color, we
     can update the minimum distances in O(n) time. 

     mod by Tom Lane   Tom.Lane@g.gp.cs.cmu.edu */

  for (i=1; i<ncols; i++) {
    int ckR, ckG, ckB;
    /* Get RGB of color last selected  and choose selection method */
    ck = &c1[i-1];            /* point to just-selected color */
    ckR = ck->r; ckG = ck->g; ckB = ck->b;

    /* Now find the i'th most different color */

    if (i&1 || i<10) {
      /* select the unused color that has the greatest mindist */
      entry = -1;  mdist = -1;
      for (j=0, cj=c; j<ncols; j++,cj++) {
	if (cj->use) {      /* this color has not been marked already */
	  /* update mindist */
          d = (cj->r - ckR)*(cj->r - ckR) + 
	      (cj->g - ckG)*(cj->g - ckG) + 
	      (cj->b - ckB)*(cj->b - ckB);
          if (cj->mindist > d) cj->mindist = d;
	  if (cj->mindist > mdist) { mdist = cj->mindist;  entry = j; }
	}
      }
    }
    else {
      /* select the unused color that has the greatest usage */
      entry = -1;  mdist = -1;
      for (j=0, cj=c; j<ncols; j++,cj++) {
	if (cj->use) {  /* this color has not been marked already */
	  /* update mindist */
          d = (cj->r - ckR)*(cj->r - ckR) + 
	      (cj->g - ckG)*(cj->g - ckG) + 
    	      (cj->b - ckB)*(cj->b - ckB);
          if (cj->mindist > d) cj->mindist = d;
	  if (cj->use > mdist) { mdist = cj->use;  entry = j; }
	}
      }
    }


    /* c[entry] is the next color to put in the map.  do so */
    xvbcopy((char *) &c[entry], (char *) &c1[i], sizeof(CMAPENT));
    c[entry].use = 0;
  }
  

  for (i=0; i<ncols; i++) order[i] = (byte) c1[i].oldindex;


  *pnumcols = ncols;
}


/***********************************/
void ColorCompress8(byte *trans)
{
  /* modify pic,cpic,epic to reflect new (compressed) colormap */

  int i;
  byte *p, j;

  for (i=pWIDE*pHIGH, p=pic; i; i--, p++) { j = trans[*p];  *p = j; }

}


/********************************/
void FreeColors()
{
  int i;

  /* frees all colors allocated by 'AllocColors()'.  Doesn't touch stdcmap */
  /* Note:  might be called multiple times.  Must not free colors once it
     has done so */

    for (i=0; i<nfcols; i++) 
      xvFreeColors(theDisp, theCmap, &freecols[i], 1, 0L);

    nfcols = 0;
    XFlush(theDisp);
}


/***********************************/
void allocROColors()
{
  int      i, j, c, unique, p2alloc;
  Colormap cmap;
  XColor   defs[256];
  XColor   ctab[256];
  int      failed[256];
  int      dc;


  unique = p2alloc = 0;

  /* FIRST PASS COLOR ALLOCATION:  
     for each color in the 'desired colormap', try to get it via
     xvAllocColor().  If for any reason it fails, mark that pixel
     'unallocated' and worry about it later.  Repeat. */

  /* attempt to allocate first ncols entries in colormap 
     note: On displays with less than 8 bits per RGB gun, it's quite
     possible that different colors in the original picture will be
     mapped to the same color on the screen.  X does this for you
     silently.  However, this is not-desirable for this application, 
     because when I say 'allocate me 32 colors' I want it to allocate
     32 different colors, not 32 instances of the same 4 shades... */
  

  for (i=0; i<256; i++) failed[i] = 1;

  cmap = theCmap;
  ncols = 256;

  for (i=0; i<numcols && unique<ncols; i++) {
    c = colAllocOrder[i];
    defs[c].red   = rMap[c]<<8;
    defs[c].green = gMap[c]<<8;
    defs[c].blue  = bMap[c]<<8;

    defs[c].flags = DoRed | DoGreen | DoBlue;

    if ( xvAllocColor(theDisp,cmap,&defs[c])) {

      unsigned long pixel, *fcptr;


      pixel = cols[c] = defs[c].pixel;
      rdisp[c] = defs[c].red   >> 8;
      gdisp[c] = defs[c].green >> 8;
      bdisp[c] = defs[c].blue  >> 8;
      failed[c]= 0;
      
      /* see if the newly allocated color is new and different */
      for (j=0, fcptr=freecols; j<nfcols && *fcptr!=pixel; j++,fcptr++);
      if (j==nfcols) unique++;

      fc2pcol[nfcols] = c;
      freecols[nfcols++] = pixel;
    }

  }  /* FIRST PASS */
  
  
  
  if (nfcols==numcols) {
    return;
  }
  


  /* SECOND PASS COLOR ALLOCATION:
     Allocating 'exact' colors failed.  Now try to allocate 'closest'
     colors.

     Read entire X colormap (or first 256 entries) in from display.
     for each unallocated pixel, find the closest color that actually
     is in the X colormap.  Try to allocate that color (read only).
     If that fails, the THIRD PASS will deal with it */

  /* read entire colormap (or first 256 entries) into 'ctab' */
  dc = (ncells<256) ? ncells : 256;

  if (dc>0) {  /* only do SECOND PASS if there IS a colormap to read */
    for (i=0; i<dc; i++) ctab[i].pixel = (unsigned long) i;
    XQueryColors(theDisp, cmap, ctab, dc);
    
    for (i=0; i<numcols && unique<ncols; i++) {
      c = colAllocOrder[i];
      
      if (failed[c]) {  /* an unallocated pixel */
	int d, mdist, close;
	int rd, gd, bd, ri, gi, bi;
	
	mdist = 1000000;   close = -1;
	ri = rMap[c];  gi = gMap[c];  bi = bMap[c];
	
	for (j=0; j<dc; j++) {
	  rd = ri - (ctab[j].red  >>8);
	  gd = gi - (ctab[j].green>>8);
	  bd = bi - (ctab[j].blue >>8);
	  
	  d = rd*rd + gd*gd + bd*bd;
	  if (d<mdist) { mdist=d; close=j; }
	}
	
	if (close<0) printf("This Can't Happen! (How reassuring.)");
	if (xvAllocColor(theDisp, cmap, &ctab[close])) { 
	  xvbcopy((char *) &ctab[close], (char *) &defs[c], sizeof(XColor));
	  failed[c]= 0;
	  cols[c]  = ctab[close].pixel;
	  rdisp[c] = ctab[close].red   >> 8;
	  gdisp[c] = ctab[close].green >> 8;
	  bdisp[c] = ctab[close].blue  >> 8;
	  fc2pcol[nfcols] = c;
	  freecols[nfcols++] = cols[c];
	  p2alloc++;
	  unique++;
	}
      }
    }
  }


  /* THIRD PASS COLOR ALLOCATION:
     We've alloc'ed all the colors we can.  Now, we have to map any
     remaining unalloced pixels into either the colors that we DID get */

  for (i=0; i<numcols; i++) {
    c = colAllocOrder[i];

    if (failed[c]) {  /* an unallocated pixel */
      int d, k, mdist, close;
      int rd,gd,bd, ri,gi,bi;

      mdist = 1000000;   close = -1;
      ri = rMap[c];  gi = gMap[c];  bi = bMap[c];
      
      /* search the alloc'd colors */
      for (j=0; j<nfcols; j++) {
	k = fc2pcol[j];
	rd = ri - (defs[k].red  >>8);
	gd = gi - (defs[k].green>>8);
	bd = bi - (defs[k].blue >>8);

	d = rd*rd + gd*gd + bd*bd;

	if (d<mdist) { mdist=d;  close=k; }
      }

      if (close<0) printf("Pass3: Failed to alloc ANY colors!\n");
      xvbcopy((char *) &defs[close], (char *) &defs[c], sizeof(XColor));
      failed[c]= 0;
      cols[c]  = defs[c].pixel;
      rdisp[c] = defs[c].red   >> 8;
      gdisp[c] = defs[c].green >> 8;
      bdisp[c] = defs[c].blue  >> 8;
    }
  }  /* THIRD PASS */



}


int xvAllocColor( Display *dp, Colormap cm, XColor *cdef)
{
  return (XAllocColor(dp,cm,cdef));
}



void NewPicGetColors(Display *display)
{
  int i;
  byte trans[256];
  char *fgstr, *bgstr;
  XColor ecdef;
  XVisualInfo *vinfo, rvinfo;
  int          best,  numvis;
  long         flags;
  unsigned char dflt[256];



    /* init default colors */
    fgstr = "#000000";  bgstr = "#B2C0DC";

    theDisp   = display;
    theScreen = DefaultScreen(theDisp);
    theCmap   = DefaultColormap(theDisp, theScreen);
    theGC     = DefaultGC(theDisp,theScreen);
    theVisual = DefaultVisual(theDisp,theScreen);
    ncells    = DisplayCells(theDisp, theScreen);
    dispDEEP  = DisplayPlanes(theDisp,theScreen);

    FreeColors();

    for (i=0; i<256; i++) 
      dflt[i] = rMap[i] = gMap[i] = bMap[i] = i;

    white = WhitePixel(theDisp,theScreen);
    black = BlackPixel(theDisp,theScreen);

    fg = black;   bg = white;
    if (fgstr && XParseColor(theDisp, theCmap, fgstr, &ecdef) &&
        xvAllocColor(theDisp, theCmap, &ecdef)) {
      fg = ecdef.pixel;
    }

    if (bgstr && XParseColor(theDisp, theCmap, bgstr, &ecdef) &&
        xvAllocColor(theDisp, theCmap, &ecdef))  {
      bg = ecdef.pixel;
    }

    XSetForeground(theDisp,theGC,fg);
    XSetBackground(theDisp,theGC,bg);

    infofg = fg;  infobg = bg;

    /* some stuff that necessary whenever running an algorithm or
       installing a new 'pic' (or switching 824 modes) */

    numcols = 0;   /* gets set by SortColormap:  set to 0 for PIC24 images */
    for (i=0; i<256; i++) cols[i]=infobg;

  pic = dflt;
  pWIDE = 16;
  pHIGH = 16;

  SortColormap(pic,pWIDE,pHIGH,&numcols,rMap,gMap,bMap,colAllocOrder,trans);
  ColorCompress8(trans);

  allocROColors();

}

XImage *Pic8ToXImage(
     byte          *pic8, 
     unsigned int   wide, 
     unsigned int   high,
     unsigned long *xcolors,
     byte          *rmap, 
     byte          *gmap, 
     byte          *bmap)
{
  /*
   * this has to do the tricky bit of converting the data in 'pic8'
   * into something usable for X.
   *
   */


  int     i;
  unsigned long xcol;
  XImage *xim;

  xim = (XImage *) NULL;

  if (!pic8) return xim;  /* shouldn't happen */

  switch (dispDEEP) {

  case 8: {
    byte  *imagedata, *ip, *pp;
    int   j, imWIDE, nullCount;
  
    nullCount = (4 - (wide % 4)) & 0x03;  /* # of padding bytes per line */
    imWIDE = wide + nullCount;
 
    /* Now create the image data - pad each scanline as necessary */
    imagedata = (byte *) malloc((size_t) (imWIDE * high));
    if (!imagedata) printf("couldn't malloc imagedata");
    
    pp = pic8;

    for (i=0, ip=imagedata; i<high; i++) {

      for (j=0; j<wide; j++, ip++, pp++) *ip = (byte) xcolors[*pp];

      for (j=0; j<nullCount; j++, ip++) *ip = 0;
    }
      
    xim = XCreateImage(theDisp,theVisual,dispDEEP,ZPixmap,0,
		       (char *) imagedata,  wide,  high, 
		       32, imWIDE);
    if (!xim) printf("couldn't create xim!");
  }
    break;

}

  return(xim);
}

int setPixmapImage(unsigned char *img, int h, int w,Pixmap *thePixmap, 
                   Widget draw_area, Boolean do_copy)
{
   int i;

   pic = img;
   pWIDE = w;
   pHIGH = h;

   ximage = Pic8ToXImage(pic, pWIDE, pHIGH, cols, rMap, gMap, bMap);

   /* clear pixmap with white */
   // XFillRectangle (theDisp, *thePixmap, theGC, 0, 0, pWIDE, pHIGH);

   XPutImage (theDisp, *thePixmap, theGC, ximage, 0,0,0,0, w, h);

   XFree((caddr_t)ximage);
   if (do_copy)
      XCopyArea (XtDisplay(draw_area),
                 *thePixmap, XtWindow(draw_area), theGC,
                 0, 0,
                 w,
                 h,
                 0, 0);


}
