/* $RCSfile: dvme626.cc,v $ - board and channel class implementaions for the Datel DVME-626 D/A
               board */
/*
$Header: /home/chara/cvsroot/ople/sys/hw/drivers/dvme626.cc,v 1.2 1998/07/13 22:11:18 phil Exp $
*/

/* 
modification history
--------------------
02a,16Apr97,pci  Modified for use in the RICST project
01h,19Feb94,beh  Changed order of calls in dacFullInit to prevent bus error
01g,31aug91,mmc  added 2 includes
01f,30Jul91,mmc  changed includes for 5.0
01e,16Jul91,grw  added function dacFullInit
01d,15Oct90,arg  changed subroutine comments for mangen
01c,28Sep90,arg  changed dacOut to dacOutMV and added dacOutRaw
01b,20Sep90,arg  added sccs info, vxProto.h, dacConfig.h
01a,07Mar90,arg  written.
*/

/*
* DESCRIPTION
* This module is an interface to the Datel DVME-626 digital-to-analog board.
* The following methods are defined here:
*          tDvme626Board::SoftInit
*          tDvme626Board::HardInit
*          tDvme626Board::ResetAllChannels
*          tDvme626DaChannel::Initialize
*          tDvme626DaChannel::Output
*/


#include "isit.h"
#include "hwTypesP.h"
#include "board.h"
#include "outputChannel.h"
#include "dvme626.h"


/*******************************************************
*
* tDvme626Board::tDvme626Board - constructor
*
* This is the constructor for the PZT/Motor class.
*
[[U tDvme626Board::tDvme626Board(iBoardNum)]]
* 
******/
tDvme626Board::tDvme626Board(int iBoardNum)
{

  TraceMark(S_dvme626_ENTERING_CONSTRUCTOR);

  switch(iBoardNum) {
   case 0:
    this->vmeAddress = tVmeAddress(VME_AM_SUP_SHORT_IO, (uint8_t *) 0xc900);
    break;
   case 1:
    this->vmeAddress = tVmeAddress(VME_AM_SUP_SHORT_IO, (uint8_t *) 0xca00);
    break;
   default:
    this->vmeAddress = tVmeAddress();
    break;
  }
  this->boardName  = DVME626_BOARD_NAME + iBoardNum;

  TraceMark(S_dvme626_FINISHED_CONSTRUCTING);
}

/******************************************************************************
*
* tDvme626Board::SoftInit - parameterized version
*                           sets board address, initializes pointers to board
*                           variables, and sets class variables
*
* [[U tDvme626Board::SoftInitParams()]]
*/
ERROR_CODE tDvme626Board::SoftInit(int         iNumChannelsToUse,
                                   int         iResetValueToUse,
                                   CONV_CODE   ccConvCodeToUse)
{
  ERROR_CODE ec;
  int i;

  TraceMark(S_dvme626_ENTERING_SOFTINIT);

  ec=tBoard::Initialize();
  if(ec != OK) {
    return Traceback(ec,"tDvme626Board::SoftInit", iNumChannelsToUse,
                       iResetValueToUse, ccConvCodeToUse);
  }

  if((iNumChannelsToUse < 1) || (iNumChannelsToUse > this->CiMaxChannels)) {
    return Traceback(S_dvme626_INVALID_NUMBER_OF_CHANNELS,
                     "tDvme626Board::SoftInit",
                     iNumChannelsToUse, iResetValueToUse, ccConvCodeToUse);
  }

  this->iNumActiveChannels = MAX(this->iNumActiveChannels, iNumChannelsToUse);

  for(i=0; i < iNumActiveChannels; i++) {
    ec=this->ChannelArray[i].Initialize(iResetValueToUse,
                                        NULL,
                                        (int16_t *) (this->vu8LocalAddress + i*2 +
                                                     this->CiChannelRegOffset),
                                        ccConvCodeToUse);
    if(ec != OK) {
      return Traceback(ec,"tDvme626Board::SoftInit", iNumChannelsToUse,
                       iResetValueToUse, ccConvCodeToUse);
    }
  }
  this->SetInit(TRUE);

  TraceMark(S_dvme626_INITIALIZED_SOFTLY);

  return OK;
}
 
/******************************************************************************
*
* tDvme626Board::SoftInit - No parameter version, uses defaults.
*                           sets board address, initializes pointers to board
*                           variables, and sets class variables
*
* [[U tDvme626Board::SoftInitVoid()]]
*/
ERROR_CODE tDvme626Board::SoftInit(void)
{

  return(this->SoftInit(this->CiMaxChannels,
                        this->CiDefaultResetValue,
                        this->CccDefaultConvCode));
}

/******************************************************************************
*
* tDvme626Board::HardInit - 3 parameter version
*                           sets board address, initializes pointers to board
*                           variables, sets class variables, and sets channel
*                           outputs to zero
*
* [[U tDvme626Board::HardInitParams()]]
*/
ERROR_CODE tDvme626Board::HardInit(int         iNumChannelsToUse,
                                   int         iResetValueToUse,
                                   CONV_CODE   ccConvCodeToUse)
{
  ERROR_CODE ec;
  int i;

  TraceMark(S_dvme626_ENTERING_HARDINIT);

  ec=this->SoftInit(iNumChannelsToUse, iResetValueToUse, ccConvCodeToUse);
  if(ec != OK) {
    return Traceback(ec,"tDvme626Board::HardInit", iNumChannelsToUse,
                      iResetValueToUse, ccConvCodeToUse);
  }

  ec=this->ResetAllChannels();
  if(ec != OK) {
    return Traceback(ec,"tDvme626Board::HardInit", iNumChannelsToUse,
                       iResetValueToUse, ccConvCodeToUse);
  }

  TraceMark(S_dvme626_INITIALIZED_HARDLY);

  return OK;
}
 
/******************************************************************************
*
* tDvme626Board::HardInit - No parameter version, uses defaults.
*                           sets board address, initializes pointers to board
*                           variables, and sets class variables
*
* [[U tDvme626Board::HardInitVoid()]]
*/
ERROR_CODE tDvme626Board::HardInit(void)
{
  return(this->HardInit(this->CiMaxChannels,
                        this->CiDefaultResetValue,
                        this->CccDefaultConvCode));
}


/******************************************************************************
*
* tDvme626Board::CreateChannel - gets channel specified by iChannelNum (global
*                             board numbering)
*
[[U tDvme626Board::CreateChannel()]]
*/
tChannel *tDvme626Board::CreateChannel(int iChannelNum)
{
  if((this->IsInit() && iChannelNum >= this->iNumActiveChannels) ||
      iChannelNum < 0 || iChannelNum >= this->CiMaxChannels) {
    TraceMsg(tmt_ERROR, S_dvme626_RETURNING_NULL_CHANNEL_POINTER,
             "ERROR: Returning NULL channel pointer for DVME626 Channel #%d",
           iChannelNum); 
    return NULL;
  }
  this->iNumActiveChannels = MAX(this->iNumActiveChannels, (iChannelNum+1));
  return (&(this->ChannelArray[iChannelNum]));
}


/******************************************************************************
*
* tDvme626Board::ResetAllChannels - sets the output of all active channels to
*                                   the current reset value.
*
[[U tDvme626Board::ResetAllChannels()]]
*/

ERROR_CODE tDvme626Board::ResetAllChannels(void)
{
  ERROR_CODE ec;
  int i;

  for(i = 0; i < this->iNumActiveChannels; i++) {
    ec = this->ChannelArray[i].Reset();
    if(ec != OK) {
      return Traceback(ec,"tDvme626Board::ResetAllChannels");
    }
  }
  return OK;
}


/******************************************************************************
*
* tDvme626Board::ShowID - Prints out the ID code of the board
*
* The ID code of the board number board is printed to the screen.  This is used
* to verify that this class does indeed point to a dvme626 board.
* 
*    *** I did not write spec for this function because I don't think it will
*        be used in Increment 2.  It was in the Palomar code, so I included it ***
*/

void tDvme626Board::ShowId(void)
{
  int a;

  for (a = 1; a < 32; a+=2) {
    printf("%c",  *(this->vu8LocalAddress + a));
  }
  printf("\n");

}


/******************************************************************************
*
* tDvme626DaChannel::Initialize - Initializes an individual D/A channel
*
* [[U tDvme626DaChannel::Initialize() ]]
*
*/
ERROR_CODE tDvme626DaChannel::Initialize(int              iResetValueToUse,
                                         tInputChannel    *pReadbackChannelToUse,
                                         volatile int16_t *vwOutputToUse,
                                         CONV_CODE        ccConvCodeToUse)
{
  TraceMark(S_dvme626_ENTERING_CHANNEL_INITIALIZE);

  this->iResetValue = iResetValueToUse;
  this->pReadbackChannel = pReadbackChannelToUse;
  this->iBitWidth = this->CiBitWidth;
  this->dChannelRange = (double) CiVoltageRange;
  this->vwOutput  = vwOutputToUse;
  this->ccConvCode = ccConvCodeToUse;
  this->SetInit(TRUE);

  TraceMark(S_dvme626_CHANNEL_INITIALIZED);

  return OK;
}


/******************************************************************************
*
* tDvme626DaChannel::Output - set output of a channel to value passed
*
* Outputs the lower 16 bits of value to the specified channel.  No conversion
* takes place.
*
* [[U tDvme626DaChannel::Output() ]]
*
*/
ERROR_CODE tDvme626DaChannel::Output(int iValue)
{
  *this->vwOutput = (int16_t) iValue;

  return OK;
}
