/* $RCSfile: ip16Dac.cc,v $ - Greenspring IP-16DAC digital-to-analog board driver */

/* $Header: /home/chara/cvsroot/ople/sys/hw/drivers/ip16Dac.cc,v 1.1.1.1 1998/06/11 21:36:33 chara Exp $ */

/* Description:
 * This module is an interface to the Greenspring IP-16DAC digital-to-analog
 *   converter board hosted on a VME bus.
 * The following methods are defined here:
 *          tIp16DacBoard::SoftInit
 *          tIp16DacBoard::HardInit
 *          tIp16DacChannel::Initialize
 *          tIp16DacChannel::Output
 *          tIp16DacChannel::Reset
 *          tIp16DacChannel::SetReset
 *
 * This is a 16-bit board, so most register pointers are for a 16-bit quantity.
 * The PROM parameters are accessed in 8-bit chunks and assembled for use.
 */

/* $Log: ip16Dac.cc,v $
 * Revision 1.1.1.1  1998/06/11 21:36:33  chara
 * First import of RICST Code
 *
 * Revision 1.7.2.4  1998/05/08 01:21:28  phil
 * interrupt functionality for I5b
 *
 * Revision 1.7.2.3  1998/04/21 00:13:23  tgl
 * Set channel range to voltage range of class, rather than nominal value
 *  calculation stolen from ipQuad code.
 *  This makes the scaling assumptions in the Star Tracker code match up.
 *
 * Revision 1.7.2.2  1998/04/17 16:50:12  tgl
 * Change output channel characteristics from "straight binary"
 *  to "offset binary". There was confusion on which module should
 *  do the conversion...
 *
 * Revision 1.7.2.1  1998/04/04 00:31:49  phil
 * commit before build_5a_002
 *
 * Revision 1.7  1998/04/02 21:01:07  phil
 * Build I5a near complete; still some configMgr stuff left
 *
 * Revision 1.6  1998/03/31 20:40:07  tgl
 * Fix up a few lines to get within an 80 column page.
 *
 * Revision 1.5  1998/03/31 20:12:31  tgl
 * Ensure board identification string read from PROM is printable.
 * Check board model number from PROM in VerifyId().
 *
 * Revision 1.4  1998/03/13 18:48:27  tgl
 * Decouple tBoard methods from tChannel methods; board resets shouldn't
 *  specify a reset value for all channels.
 * Match spec to code.
 * Fix address arithmetic for reading gain and offset calibration values
 *  from board.
 * Set default reset value in Initialize code, which reads calibration
 *  values from board.
 * Add comments regarding use of hardware calibration values.
 *
 * Revision 1.3  1998/03/12 00:39:50  tgl
 * Update spec to allow a clean pass through the spec tool.
 * Make some changes to calling parameters to match spec.
 *
 * Revision 1.2  1998/03/11 23:47:29  tgl
 * Accumulate changes to accomplish a clean compilation.
 * Redefine the Initialize method to include parameters for all relevant
 *  registers, including offset and gain calibrations.
 * Use 32-bit integer inside Output method to allow range checking.
 *
 * Revision 1.1  1998/03/11 18:19:20  tgl
 * Support for IP-16DAC digital to analog board.
 * First check-in; code not yet compiled and definitely not ready.
 * */


#include "isit.h"
#include "hwTypesP.h"
#include "board.h"
#include "outputChannel.h"
#include "ip16Dac.h"


/*******************************************************
*
*    Hardware Constants
*
* Currently, VME board addresses are hardcoded here.
* Later, this will move to the Configuration Manager.
*
* Two IP-16DAC daughterboards are plugged in to slots C and D of
*  the 616 carrier board. This determines the addresses for access:
*
*  board address = (carrier base address)+((slot # [0-3])*0x100)
*
***/

#define IP16DAC_BOARD_BASE_0 (0x6200)
#define IP16DAC_BOARD_BASE_1 (0x6300)

#define IP16DAC_MINIMUM_VALUE (0)
#define IP16DAC_MAXIMUM_VALUE (0xFFFF)


/*******************************************************
*
*    Register Masks
*
***/

/* None right now... */


/*******************************************************
*
* tIp16DacBoard::tIp16DacBoard - constructor
*
* This is the constructor for a PZT/Motor class with DAC interface.
*
* Currently, VME board addresses are hardcoded here.
* Later, this will move to the Configuration Manager.
*
* Two IP-16DAC daughterboards are plugged in to slots C and D of
*  the 616 carrier board. This determines the addresses for access.
*
  [[U tIp16DacBoard::tIp16DacBoard(iBoardNum)]]
* 
******/
tIp16DacBoard::tIp16DacBoard(int iBoardNum)

/****
[[U tBoard::tBoard()]]
****/

{
  TraceMark(S_ip16Dac_ENTERING_CONSTRUCTOR);

  switch(iBoardNum) {
   case 0:
    this->vmeAddress = tVmeAddress(VME_AM_SUP_SHORT_IO,
                                   (uint8_t *) IP16DAC_BOARD_BASE_0);
    break;
   case 1:
    this->vmeAddress = tVmeAddress(VME_AM_SUP_SHORT_IO,
                                   (uint8_t *) IP16DAC_BOARD_BASE_1);
    break;
   default:
    this->vmeAddress = tVmeAddress();
    this->SetInitError(S_ip16Dac_INVALID_BOARD_NUM);
    break;
  }

  this->uiSize     = 0x400;
  this->boardName  = IP16DAC_BOARD_NAME + iBoardNum;

  TraceMark(S_ip16Dac_FINISHED_CONSTRUCTING);
}

/*******************************************************
*
* tIp16DacBoard::VerifyId - verify identity of board
*
  [[U tIp16DacBoard::VerifyId()]]
* 
******/
ERROR_CODE tIp16DacBoard::VerifyId(void)

{
  volatile uint8_t *vu8IdProm;
  int               i, a;
  uint8_t           boardModel;
  char              boardId[5];

  TraceMark(S_ip16Dac_ENTERING_VERIFIER);

  vu8IdProm = this->vu8LocalAddress + this->CiIdPromOffset;
  for (i = 0, a = 1; i < 4; i++, a+=2) {
    boardId[i] = *(vu8IdProm + a);
    // fix up non-printable characters; isprint() not available?
    if ((boardId[i] < ' ') || (boardId[i] > 'z')) boardId[i] = '?';
  }
  boardId[4] = '\0';
  boardModel = *(vu8IdProm + 0x0b);

  if ((strcmp(boardId, "IPAC") == 0) && (boardModel == 0x25)) {
    TraceMsg(tmt_INFO, S_ip16Dac_BOARD_ID,
             "tIp16DacBoard::VerifyId %08X board ID is '%s' 0x%x",
             this->boardName, boardId, boardModel);

  } else {
    return (Traceback(S_ip16Dac_INVALID_BOARD_ID,
                      "tIp16DacBoard::VerifyId %08X board ID is '%s' 0x%x"
                      "; expected 'IPAC' 0x25",
                      this->boardName, boardId, boardModel));
  }

  TraceMark(S_ip16Dac_FINISHED_VERIFYING);
  return OK;
}

/******************************************************************************
*
* tIp16DacBoard::SoftInit - parameterized version
*                           sets board address, initializes pointers to board
*                           variables, and sets class variables
*
  [[U tIp16DacBoard::SoftInitParams(int iNumChannelsToUse)]]
*/
ERROR_CODE tIp16DacBoard::SoftInit(int iNumChannelsToUse)
{
  ERROR_CODE ec;
  int i;

  TraceMark(S_ip16Dac_ENTERING_SOFTINIT);

  ec=tBoard::Initialize();
  if(ec != OK) {
    return Traceback(ec, "tIp16DacBoard::SoftInit tBoard failed"
                     " for %d channels", iNumChannelsToUse);
  }

  if((iNumChannelsToUse < 1) || (iNumChannelsToUse > this->CiMaxChannels)) {
    return Traceback(S_ip16Dac_INVALID_NUMBER_OF_CHANNELS,
                     "tIp16DacBoard::SoftInit %d channels requested"
                     " out of allowed range [%d,%d]",
                     iNumChannelsToUse, 1, this->CiMaxChannels);
  }

  this->iNumActiveChannels = MAX(this->iNumActiveChannels, iNumChannelsToUse);

  for(i=0; i < iNumActiveChannels; i++) {
    /****
    [[U tIp16DacChannel::Initialize(
           (uint16_t *) (this->vu8LocalAddress + i*2),
           (this->vu8LocalAddress + this->CiIdPromOffset + 0x18 + i*0x04),
           (this->vu8LocalAddress + this->CiIdPromOffset + 0x24 + i*0x04),
           0)]]
    ****/
    ec = this->ChannelArray[i].Initialize(
           (uint16_t *) (this->vu8LocalAddress + i*2),
           (this->vu8LocalAddress + this->CiIdPromOffset + 0x18 + i*0x04),
           (this->vu8LocalAddress + this->CiIdPromOffset + 0x24 + i*0x04),
           0);
    if(ec != OK) {
      return Traceback(ec,"tIp16DacBoard::SoftInit channel #%d/%d"
                       " failed to initialize",
                       i, this->iNumActiveChannels);
    }
  }
  this->SetInit(TRUE);

  TraceMark(S_ip16Dac_INITIALIZED_SOFTLY);

  return OK;
}
 
/******************************************************************************
*
* tIp16DacBoard::SoftInit - No parameter version, uses defaults.
*                           sets board address, initializes pointers to board
*                           variables, and sets class variables
*
  [[U tIp16DacBoard::SoftInitVoid()]]
*/
ERROR_CODE tIp16DacBoard::SoftInit(void)
{
  /****
  [[U tIp16DacBoard::SoftInitParams(this->CiMaxChannels)]]
  ****/
  return(this->SoftInit(this->CiMaxChannels));
}

/******************************************************************************
*
* tIp16DacBoard::HardInit - 1 parameter version
*                           sets board address, initializes pointers to board
*                           variables, sets class variables, and sets channel
*                           outputs to zero
*
  [[U tIp16DacBoard::HardInitParams(int iNumChannelsToUse)]]
*/
ERROR_CODE tIp16DacBoard::HardInit(int iNumChannelsToUse)
{
  ERROR_CODE ec;
  int i;

  TraceMark(S_ip16Dac_ENTERING_HARDINIT);

  /****
  [[U tIp16DacBoard::SoftInitParams(iNumChannelsToUse)]]
  ****/
  ec=this->SoftInit(iNumChannelsToUse);
  if(ec != OK) {
    return Traceback(ec, "tIp16DacBoard::HardInit failed SoftInit"
                     " for %d channels", iNumChannelsToUse);
  }

  /****
  [[U tIp16DacBoard::VerifyId()]]
  ****/
  ec=this->VerifyId();
  if(ec != OK) {
    return Traceback(ec, "tIp16DacBoard::HardInit failed board verification");
  }

  /****
  [[U tIp16DacBoard::ResetAllChannels()]]
  ****/
  ec=this->ResetAllChannels();
  if(ec != OK) {
    return Traceback(ec, "tIp16DacBoard::HardInit failed channel reset");
  }

  TraceMark(S_ip16Dac_INITIALIZED_HARDLY);

  return OK;
}
 
/******************************************************************************
*
* tIp16DacBoard::HardInit - No parameter version, uses defaults.
*                           sets board address, initializes pointers to board
*                           variables, and sets class variables
*
  [[U tIp16DacBoard::HardInitVoid()]]
*/
ERROR_CODE tIp16DacBoard::HardInit(void)
{
  /****
  [[U tIp16DacBoard::HardInitParams(this->CiMaxChannels)]]
  ****/
  return (this->HardInit(this->CiMaxChannels));
}


/******************************************************************************
*
* tIp16DacBoard::CreateChannel - gets channel specified by iChannelNum (global
*                             board numbering)
*
[[U tIp16DacBoard::CreateChannel(int iChannelNum)]]
*/
tChannel *tIp16DacBoard::CreateChannel(int iChannelNum)
{
  if((this->IsInit() && iChannelNum >= this->iNumActiveChannels) ||
      iChannelNum < 0 || iChannelNum >= this->CiMaxChannels) {
    TraceMsg(tmt_ERROR, S_ip16Dac_RETURNING_NULL_CHANNEL_POINTER,
             "ERROR: Returning NULL channel pointer for IP-16DAC Channel #%d",
             iChannelNum); 
    return NULL;
  }
  this->iNumActiveChannels = MAX(this->iNumActiveChannels, (iChannelNum+1));
  return (&(this->ChannelArray[iChannelNum]));
}


/******************************************************************************
*
* tIp16DacBoard::ResetAllChannels - sets the output of all active channels to
*                                   the current reset value.
*
[[U tIp16DacBoard::ResetAllChannels()]]
*/

ERROR_CODE tIp16DacBoard::ResetAllChannels(void)
{
  ERROR_CODE ec;
  int i;

  for(i=0; i < this->iNumActiveChannels; i++) {
    ec = this->ChannelArray[i].Reset();
    if(ec != OK) {
      return Traceback(ec,"tIp16DacBoard::ResetAllChannels");
    }
  }
  return OK;
}


/******************************************************************************
*
* tIp16DacBoard::ShowId - Prints out the ID code of the board
*
* The ID code of the board number board is printed to the screen.  This is used
* to verify that this class does indeed point to a ip16Dac board.
*
[[U tIp16DacBoard::ShowId(void)]]
*/

void tIp16DacBoard::ShowId(void)
{
  volatile uint8_t *vu8IdProm = this->vu8LocalAddress + this->CiIdPromOffset;
  int a;

  for (a = 1; a < 0x09; a+=2) {
    printf("%c", *(vu8IdProm + a));
  }
  printf("\n");

  printf("Manufacturer ID: %d\n", *(vu8IdProm + 0x09));
  printf("Model %d, Rev. %d\n", *(vu8IdProm + 0x0b), *(vu8IdProm + 0x0d));

  printf("\n");
}




/******************************************************************************
*
* tIp16DacChannel::Initialize - Initializes an individual dac channel
*
* Calibration values are available in PROM for this model of DAC board.
* Read these calibration values so they are available
*  for use in some future version of the tOutputChannel class.
*
  [[U tIp16DacChannel::Initialize(volatile uint16_t *vuwDacRegToUse,
                                  volatile uint8_t  *vu8OffsetCorrRegToUse,
                                  volatile uint8_t  *vu8GainCorrRegToUse,
                                  int                iResetValueToUse)]]
*
*/
ERROR_CODE tIp16DacChannel::Initialize(volatile uint16_t *vuwDacRegToUse,
                                       volatile uint8_t  *vu8OffsetCorrRegToUse,
                                       volatile uint8_t  *vu8GainCorrRegToUse,
                                       int                iResetValueToUse)
{
  int16_t wOffsetCorr;
  int16_t wGainCorr;
//  double  dOffsetCorr;
//  double  dGainCorr;

  TraceMark(S_ip16Dac_ENTERING_CHANNEL_INITIALIZE);

  wOffsetCorr = ((*(vu8OffsetCorrRegToUse + 0x01) << 8)
                 | *(vu8OffsetCorrRegToUse + 0x03));
  wGainCorr = ((*(vu8GainCorrRegToUse + 0x01) << 8)
                 | *(vu8GainCorrRegToUse + 0x03));

  this->iResetValue   = (1 << (this->CiBitWidth/2));
  this->iBitWidth     = this->CiBitWidth;
  this->dChannelRange = this->CiVoltageRange;
  this->ccConvCode    = CC_OFFSET_BINARY;
  this->vuwOutput     = vuwDacRegToUse;

  /****
  [[U tIp16DacChannel::ResetParams(iResetValueToUse)]]
  ****/
  this->Reset(iResetValueToUse);

  this->SetInitError(OK);
  this->SetInit(TRUE);

  TraceMark(S_ip16Dac_CHANNEL_INITIALIZED);

  return OK;
}


/******************************************************************************
*
* tIp16DacChannel::Output - sets DAC output to specified value
*
  [[U tIp16DacChannel::Output(int iValue) ]]

[ invalid_pointer(this) ||
  another Output call is made on this channel during execution => UNDEFINED
| true =>
    *(this->vuwOutput) := normalized integer value with min/max range checking
    : return := OK
]
*
*/
ERROR_CODE tIp16DacChannel::Output(int iValue)
{
  iValue = GetOutputValue(iValue);

  if (iValue > IP16DAC_MAXIMUM_VALUE) iValue = IP16DAC_MAXIMUM_VALUE;
  if (iValue < IP16DAC_MINIMUM_VALUE) iValue = IP16DAC_MINIMUM_VALUE;
  *(this->vuwOutput) = iValue;

  return OK;
}


/******************************************************************************
*
* tIp16DacChannel::Reset - set DAC output to nominal value
*
  [[U tIp16DacChannel::ResetVoid() ]]
*
*/
ERROR_CODE tIp16DacChannel::Reset(void)
{
  *(this->vuwOutput) = GetOutputValue(this->iResetValue);

  return OK;
}


/******************************************************************************
*
* tIp16DacChannel::Reset - set DAC output to nominal value
*
  [[U tIp16DacChannel::ResetParams(int iResetValueToUse) ]]
*
*/
ERROR_CODE tIp16DacChannel::Reset(int iResetValueToUse)
{
  *(this->vuwOutput) = GetOutputValue(iResetValueToUse);

  return OK;
}

