/* $RCSfile: ipQuad.cc,v $ - board/channel class for the Greenspring IP-QUADRATURE counter board */
/*
$Header: /home/chara/cvsroot/ople/sys/hw/drivers/ipQuad.cc,v 1.1.1.1 1998/06/11 21:36:34 chara Exp $
*/

/* 
modification history
--------------------
02a,19Nov97,pci  written
*/

/*
* DESCRIPTION
* This module is an interface to the Greenspring IP-QUADRATURE counter board.
* The following methods are defined here:
*          tIpQuadBoard::SoftInit
*          tIpQuadBoard::HardInit
*          tIpQuadCounterChannel::Initialize
*          tIpQuadCounterChannel::Input
*/


#include "isit.h"
#include "hwTypesP.h"
#include "board.h"
#include "inputChannel.h"
#include "ipQuad.h"


/*******************************************************
*
*    Register Masks
*
***/
/* Masks for IP-QUADRATURE Board control status register */
#define BOARD_CSR_MASK 0xf8
#define TERM_1_ENABLE  0x01
#define TERM_1_DISABLE 0x00
#define TERM_2_ENABLE  0x02
#define TERM_2_DISABLE 0x00

/* Masks for IP-QUADRATURE Channel control register */
#define CR_MASK        0x00
#define CR_INIT        0x20
#define CR_RESET       0x04
#define CR_READ        0x03

/* Masks for IP-QUADRATURE Channel input control register */
#define ICR_MASK       0x40
#define ICR_ENABLE     0x08
#define ICR_DISABLE    0x00
#define ICR_COUNT_DIR  0x01

/* Masks for IP-QUADRATURE Channel output control register */
#define OCR_MASK          0x80
#define OCR_BINARY_MODE   0x00
#define OCR_BCD_MODE      0x01
#define OCR_RECYCLE       0x00
#define OCR_NON_RECYCLE   0x02
#define OCR_DIVIDE_BY_N   0x04
#define OCR_24HR_CLOCK    0x08
#define OCR_INT_ON_BORROW 0x10
#define OCR_INT_ON_MATCH  0x30

/* Masks for IP-QUADRATURE Channel Quadrature register */
#define QR_MASK           0xc0
#define QR_QUAD_DISABLED  0x00
#define QR_X1_MODE        0x01
#define QR_X2_MODE        0x02
#define QR_X4_MODE        0x03

/* Masks for IP-QUADRATURE Channel Configuration register */
#define CHCR_MASK         0x20
#define CHCR_X_INVERT     0x01
#define CHCR_Y_INVERT     0x02
#define CHCR_Z_INVERT     0x04
#define CHCR_Z_FCT_SELECT 0x08
#define CHCR_Z_FCT_ENABLE 0x10
#define CHCR_INT_ENABLE   0x40


/*******************************************************
*
* tIpQuadBoard::tIpQuadBoard - constructor
*
* This is the constructor for the PZT/Motor class.
*
  [[U tIpQuadBoard::tIpQuadBoard(iBoardNum)]]
* 
******/
tIpQuadBoard::tIpQuadBoard(int iBoardNum)

/****
[[U tBoard::tBoard()]]
****/

{

  TraceMark(S_ipQuad_ENTERING_CONSTRUCTOR);

  switch(iBoardNum) {
   case 0:
    this->vmeAddress = tVmeAddress(VME_AM_SUP_SHORT_IO, (uint8_t *) 0x6000);
    break;
   default:
    this->vmeAddress = tVmeAddress();
    this->SetInitError(S_ipQuad_INVALID_BOARD_NUM);
    break;
  }
  this->uiSize     = 0x100;
  this->boardName  = IPQUAD_BOARD_NAME + iBoardNum;

  TraceMark(S_ipQuad_FINISHED_CONSTRUCTING);
}

/******************************************************************************
*
* tIpQuadBoard::SoftInit - parameterized version
*                           sets board address, initializes pointers to board
*                           variables, and sets class variables
*
  [[U tIpQuadBoard::SoftInitParams()]]
*/
ERROR_CODE tIpQuadBoard::SoftInit(int iNumChannelsToUse)
{
  ERROR_CODE ec;
  int i;

  TraceMark(S_ipQuad_ENTERING_SOFTINIT);

  ec=tBoard::Initialize();
  if(ec != OK) {
    return Traceback(ec,"tIpQuadBoard::SoftInit", iNumChannelsToUse);
  }

  if((iNumChannelsToUse < 1) || (iNumChannelsToUse > this->CiMaxChannels)) {
    return Traceback(S_ipQuad_INVALID_NUMBER_OF_CHANNELS,
                     "tIpQuadBoard::SoftInit", iNumChannelsToUse);
  }

  this->iNumActiveChannels = MAX(this->iNumActiveChannels, iNumChannelsToUse);
  this->vu8BoardCsr  = vu8LocalAddress + 0x31;
  this->vu8VectorReg = vu8LocalAddress + 0x35;

  for(i=0; i < iNumActiveChannels; i++) {
    /****
    [[U tIpQuadCounterChannel::Initialize((this->vu8LocalAddress + i*4  + 1),
                                   (this->vu8LocalAddress + 0x21 + i*2))]]
    ****/
    ec=this->ChannelArray[i].Initialize((this->vu8LocalAddress + i*4  + 1),
                                        (this->vu8LocalAddress + 0x21 + i*2));
    if(ec != OK) {
      return Traceback(ec,"tIpQuadBoard::SoftInit", iNumChannelsToUse);
    }
  }
  this->SetInit(TRUE);

  TraceMark(S_ipQuad_INITIALIZED_SOFTLY);

  return OK;
}
 
/******************************************************************************
*
* tIpQuadBoard::SoftInit - No parameter version, uses defaults.
*                           sets board address, initializes pointers to board
*                           variables, and sets class variables
*
  [[U tIpQuadBoard::SoftInitVoid()]]
*/
ERROR_CODE tIpQuadBoard::SoftInit(void)
{
  /****
  [[U tIpQuadBoard::SoftInitParams(this->CiMaxChannels)]]
  ****/
  return(this->SoftInit(this->CiMaxChannels));
}

/******************************************************************************
*
* tIpQuadBoard::HardInit - 3 parameter version
*                           sets board address, initializes pointers to board
*                           variables, sets class variables, and sets channel
*                           outputs to zero
*
  [[U tIpQuadBoard::HardInitParams()]]
*/
ERROR_CODE tIpQuadBoard::HardInit(int         iNumChannelsToUse)
{
  ERROR_CODE ec;
  int i;

  TraceMark(S_ipQuad_ENTERING_HARDINIT);

  /****
  [[U tIpQuadBoard::SoftInitParams(iNumChannelsToUse)]]
  ****/
  ec=this->SoftInit(iNumChannelsToUse);
  if(ec != OK) {
    return Traceback(ec,"tIpQuadBoard::HardInit", iNumChannelsToUse);
  }

  /****
  [[U tIpQuadBoard::ResetAllChannels()]]
  ****/
  ec=this->ResetAllChannels();
  if(ec != OK) {
    return Traceback(ec,"tIpQuadBoard::HardInit", iNumChannelsToUse);
  }

  /* disable termination resistors for use with TTL signal */
  *(this->vu8BoardCsr) = BOARD_CSR_MASK | TERM_1_DISABLE | TERM_2_DISABLE;

  TraceMark(S_ipQuad_INITIALIZED_HARDLY);

  return OK;
}
 
/******************************************************************************
*
* tIpQuadBoard::HardInit - No parameter version, uses defaults.
*                           sets board address, initializes pointers to board
*                           variables, and sets class variables
*
  [[U tIpQuadBoard::HardInitVoid()]]
*/
ERROR_CODE tIpQuadBoard::HardInit(void)
{
  /****
  [[U tIpQuadBoard::HardInitParams(this->CiMaxChannels)]]
  ****/
  return (this->HardInit(this->CiMaxChannels));
}


/******************************************************************************
*
* tIpQuadBoard::CreateChannel - gets channel specified by iChannelNum (global
*                             board numbering)
*
[[U tIpQuadBoard::CreateChannel()]]
*/
tChannel *tIpQuadBoard::CreateChannel(int iChannelNum)
{
  if((this->IsInit() && iChannelNum >= this->iNumActiveChannels) ||
      iChannelNum < 0 || iChannelNum >= this->CiMaxChannels) {
    TraceMsg(tmt_ERROR, S_ipQuad_RETURNING_NULL_CHANNEL_POINTER,
             "ERROR: Returning NULL channel pointer for IP-QUADRATURE Channel #%d",
             iChannelNum); 
    return NULL;
  }
  this->iNumActiveChannels = MAX(this->iNumActiveChannels, (iChannelNum+1));
  return (&(this->ChannelArray[iChannelNum]));
}


/******************************************************************************
*
* tIpQuadBoard::ResetAllChannels - sets the output of all active channels to
*                                   the current reset value.
*
[[U tIpQuadBoard::ResetAllChannels()]]
*/

ERROR_CODE tIpQuadBoard::ResetAllChannels(void)
{
  ERROR_CODE ec;
  int i;

  for(i=0; i < this->iNumActiveChannels; i++) {
    ec = this->ChannelArray[i].Reset();
    if(ec != OK) {
      return Traceback(ec,"tIpQuadBoard::ResetAllChannels");
    }
  }
  return OK;
}


/******************************************************************************
*
* tIpQuadBoard::ShowId - Prints out the ID code of the board
*
* The ID code of the board number board is printed to the screen.  This is used
* to verify that this class does indeed point to a ipQuad board.
*
[[U tIpQuadBoard::ShowId(void)]]
*/

void tIpQuadBoard::ShowId(void)
{
  volatile uint8_t *vu8IdProm = this->vu8LocalAddress + this->CiIdPromOffset;
  int a;

  for (a = 1; a < 0x09; a+=2) {
    printf("%c", *(vu8IdProm + a));
  }
  printf("\n");

  printf("Manufacturer ID: %d\n", *(vu8IdProm + 0x09));
  printf("Model %d, Rev. %d\n", *(vu8IdProm + 0x0b), *(vu8IdProm + 0x0d));

  printf("\n");
}




/******************************************************************************
*
* tIpQuadCounterChannel::Initialize - Initializes an individual counter channel
*
  [[U tIpQuadCounterChannel::Initialize() ]]
*
*/
ERROR_CODE tIpQuadCounterChannel::Initialize(volatile uint8_t *vu8CounterRegToUse,
                                        volatile uint8_t *vu8ConfigRegToUse)
{
  TraceMark(S_ipQuad_ENTERING_CHANNEL_INITIALIZE);

  this->iBitWidth     = this->CiBitWidth;
  this->dChannelRange = pow(2.0, (double) this->CiBitWidth);
  this->ccConvCode    = CC_STRAIGHT_BINARY;
  this->vu8Input      = vu8CounterRegToUse;
  this->vu8ChannelCsr = vu8CounterRegToUse + 2;
  this->vu8ConfigReg  = vu8ConfigRegToUse;

  *(this->vu8ChannelCsr) = CR_MASK   | CR_INIT ;
  *(this->vu8ChannelCsr) = OCR_MASK  | OCR_INT_ON_BORROW ;
  *(this->vu8ChannelCsr) = QR_MASK   | QR_QUAD_DISABLED ;

  /****
  [[U tIpQuadCounterChannel::Enable()]]
  ****/
  this->Enable();

  this->SetInitError(OK);
  this->SetInit(TRUE);

  TraceMark(S_ipQuad_CHANNEL_INITIALIZED);

  return OK;
}


/******************************************************************************
*
* tIpQuadCounterChannel::Input - gets recent value from counter
*
  [[U tIpQuadCounterChannel::Input() ]]

[ invalid_pointer(this) => UNDEFINED
| another Input call is made on this channel during execution ||
  sizeof(int) is not larger than 24 bits =>
    *piValue := anything
    : return := OK
| true =>
    *piValue := most recent value of counter
    : return := OK
]
*
*/
ERROR_CODE tIpQuadCounterChannel::Input(int *piValue) const
{
  uint8_t byte0, byte1, byte2; 

  *(this->vu8ChannelCsr) = CR_MASK | CR_READ ;

  byte0 = *(this->vu8Input);
  byte1 = *(this->vu8Input);
  byte2 = *(this->vu8Input);

  *piValue = byte0 + (byte1<<8) + (byte2<<16);
 
  return OK;
}


/******************************************************************************
*
* tIpQuadCounterChannel::Reset - starts counter
*
  [[U tIpQuadCounterChannel::Reset() ]]
*
*/
ERROR_CODE tIpQuadCounterChannel::Reset(void) const
{
  *(this->vu8ChannelCsr) = CR_MASK | CR_RESET ;
  return OK;
}


/******************************************************************************
*
* tIpQuadCounterChannel::Enable - starts counter
*
  [[U tIpQuadCounterChannel::Enable() ]]
*
*/
void tIpQuadCounterChannel::Enable(void)
{
  *(this->vu8ChannelCsr) = ICR_MASK | ICR_ENABLE ;
}


/******************************************************************************
*
* tIpQuadCounterChannel::Disable - stops counter
*
  [[U tIpQuadCounterChannel::Disable() ]]
*
*/
void tIpQuadCounterChannel::Disable(void)
{
  *(this->vu8ChannelCsr) = ICR_MASK | ICR_DISABLE ;
}

