/* $RCSfile: laserLib.cc,v $ - routines to access the JPL laser fringe counter board.  */
/*
$Header: /home/chara/cvsroot/ople/sys/hw/drivers/laserLib.cc,v 1.4 1998/09/13 19:43:45 phil Exp $
*/

/*
modification history
--------------------
02a,pci,22Oct96     Added C++ support for ISIT (removed non-Inc1 functions)
01k,beh,25Sep91     LaserFullInit no longer tests or clears laser counter
01j,mmc,31Aug91     Added 3 vxWorks includes
01i,beh,29Aug91     Made changes to lboardTemp stuff to avoid gcc warnings
01h,beh,06Jul90     Added small delays in a couple of spots to laserTest.
01g,beh,06Jul90     Changed a couple of things to work with DEBUG=30
01f,beh,05Jul90     Changed laserTest to print diagnostics in hex, clean up properly
01e,beh,01Jun90     Fixed laserTest to be robust, added 64-bit locked read stuff
01d,beh,11May90     Completed cosmetics
01c,beh,03May90     Put in final form - changes to init, test routines
01b,beh,19Feb90     Put source code in standard format
01a,beh,24Jan90     written, debugged
*/

/*
DESCRIPTION
This module contains all the routines necessary to manipulate the laser counter
board hardware.  These routines use D32 accesses to the board.  If you do D16 or D8
reads or writes, you will need to make a lot of revisions to this library.  The
detailed operation of the laser board hardware is explained in detail in the
"Laser Counter Board Hardware Manual."  The use of these drivers is described
in the "Laser Counter Board Software User's Guide."
*/

#ifndef SPECTOOL
#include <stdio.h>
#include <math.h>
#endif

#include "isit.h"
#include "hwTypesP.h"
#include "hardware.h"
#include "board.h"
#include "inputChannel.h"
#include "laserLib.h"
#include "hwError.h"

// This define determines how many ref/unk pairs will be used.
// If the board is used with a single common reference, this should
// be 0
#define LASER_NUM_CHANNEL_PAIRS (5)

// Format: { {REF,UNK}, {REF,UNK}, ... }
const int tFastLaserBoard::iChannelPair2DArray[LASER_NUM_CHANNEL_PAIRS][2] = {
  {2,3}, {4,5}, {6,7}, {8,9}, {10,11}
};
  

/**************************************************
*
* tFastLaserBoard - class constructor
*
[[U tFastLaserBoard::tFastLaserBoard()]]
*
* sets register pointers to NULL
*
*/
tFastLaserBoard::tFastLaserBoard(int iBoardNum)
{
  TraceMark(S_laser_ENTERING_CONSTRUCTOR);

  switch(iBoardNum) {
   case 0:
    this->vmeAddress = tVmeAddress(VME_AM_STD_SUP_DATA, (uint8_t *) 0xfc0000);
    break;
   default:
    this->vmeAddress = tVmeAddress();
    break;
  }
  this->boardName  = FAST_LASER_BOARD_NAME + iBoardNum;
  this->viTestSigStatusInt = NULL;
  this->viTestSigStatusFracInt = NULL;
  this->vwClockStatus = NULL;
  this->viSyncModeHi = NULL;
  this->viSyncModeLo = NULL;
  this->vuwTestCntHi = NULL;
  this->vuwTestCntLo = NULL;
  this->viTestSigStatHi = NULL;
  this->viTestSigStatLo = NULL;
  
  this->iNumActiveCounters = 0;

  TraceMark(S_laser_DONE_CONSTRUCTING);
}


/**************************************************
*
* SoftInit  -  Initialize laser counter board software
*
* Initializes the software driver by setting up pointers to the various
* pieces of hardware on the laser board, but it does not reset or test the
* hardware.  This routine is useful if, for example, you have had to reboot
* your CPU but don't want to clear the laser counters when you initialize the
* board.  The
* argument numCounters is the number of counters on the board to initialize;
* you may specify a number smaller than the number actually installed if you
* wish.
*
[[U tFastLaserBoard::SoftInitParams()]]
*
*/

ERROR_CODE tFastLaserBoard::SoftInit(int iNumCountersToUse)
{
  int i, iRefChannelNum;
  ERROR_CODE ec;

  TraceMark(S_laser_ENTERING_SOFTINIT);

  if ((iNumCountersToUse < 1) || (iNumCountersToUse > this->CiMaxNumCounters)) {
    return S_laser_INIT_INVALID_NUM_COUNTERS;
  }

  /*
  [this is invalid || vmeBoardAddress is invalid -> UNDEFINED
  | vmeBoardAddress not reachable by CPU ->
      return VxWorks error code from sysBusToLocalAdrs
  | *this.bInit == TRUE -> return S_laser_BOARD_ALREADY_INITIALIZED
  | true -> *this.vmeAddress := vmeBoardAddress
    : *this.vu8LocalAddress := local address corresponding to vmeBoardAddress
    : *this.boardName := boardName
    : *this.bInit := TRUE
  ] */ 
  ec = tBoard::Initialize();
  if(ec!=OK) {
    return Traceback(ec, "tFastLaserBoard::SoftInit", iNumCountersToUse);
  }

  /*
  [ true -> this->iNumChannels := MAX(this->iNumActiveCounters, iNumCountersToUse);
    : [[U tFastLaserChannel::Initialize
                           (this=this->ChannelArray[0..iNumActiveCounters],
                            vu8Base=appropriate offset)]]
  ] */
  this->iNumActiveCounters = MAX(this->iNumActiveCounters, iNumCountersToUse);

  for (i=0; i < this->iNumActiveCounters; i++) {
    // if i is not an unknown channel of a pair, initialize a normal channel
    if((iRefChannelNum = this->GetRefChannelNum(i)) < 0) {
      TraceMsg(tmt_INFO, OK, "Creating Normal Channel %d", i);
      this->ChannelArray[i].Initialize(vu8LocalAddress,
                                       vu8LocalAddress + GetChannelOffset(i), i);
    }
    else {   // i is the unknown channel of a pair
      TraceMsg(tmt_INFO, OK, "Creating Channel Pair %d, %d", iRefChannelNum, i);
      this->ChannelArray[i].Initialize(vu8LocalAddress,
                                       vu8LocalAddress + GetChannelOffset(i),
                                       &(this->ChannelArray[iRefChannelNum]),
                                       i);
    }
    this->FidChannelArray[i].Initialize(&(ChannelArray[i]));
  }

   /*
  [ true -> set remainder of protected variables to the appropriate values
    : *this.bInit := TRUE
    : return OK
  ] */
  viTestSigStatusInt = (volatile int32_t *) (this->vu8LocalAddress +
                                         this->CiTestSignalStatusIntOffset);
  viTestSigStatusFracInt = (volatile int32_t *) (this->vu8LocalAddress +
                                         this->CiTestSignalStatusFracIntOffset);
  vwClockStatus   = (volatile int16_t *) (this->vu8LocalAddress + 
                                         this->CiClockStatusOffset);
  viSyncModeHi    = (volatile int32_t *) (this->vu8LocalAddress + 
                                         this->CiSyncModeHiOffset);
  viSyncModeLo    = (volatile int32_t *) (this->vu8LocalAddress + 
                                         this->CiSyncModeLoOffset);
  vuwTestCntHi    = (volatile uint16_t *) (this->vu8LocalAddress + 
                                         this->CiTestCntHiOffset);
  vuwTestCntLo    = (volatile uint16_t *) (this->vu8LocalAddress + 
                                         this->CiTestCntLoOffset);
  viTestSigStatHi = (volatile int32_t *) (this->vu8LocalAddress + 
                                         this->CiTestSigStatusHiOffset);
  viTestSigStatLo = (volatile int32_t *)  (this->vu8LocalAddress + 
                                         this->CiTestSigStatusLoOffset);
  this->SetInit(TRUE);

  TraceMark(S_laser_INITIALIZED_SOFTLY);

  return OK;
}


/**************************************************
*
* SoftInit  -  Void version.  Calls parameterized version
*              with default values.
* 
[[U tFastLaserBoard::SoftInitVoid()]]
*
*/

ERROR_CODE tFastLaserBoard::SoftInit(void)
{
  return (this->SoftInit(this->CiDefaultNumCounters));
}

/**************************************************
*
* HardInit  -  Initialize and test laser counter board hardware and software
*
* Initializes the hardware and performs some simple self-tests by calling
* laserCheckClocks. Argument iNumActiveCounters is the number of counters on the
* board to initialize.
* You may specify a number smaller than the number actually installed if you
* wish.
*
[[U tFastLaserBoard::HardInit()]]
*
SPEC :
  [this is invalid || vmeBoardAddress is invalid -> UNDEFINED
  | *this.bInit == TRUE -> return S_laser_BOARD_ALREADY_INITIALIZED
  | vmeBoardAddress not reachable by CPU ->
      return VxWorks error code from sysBusToLocalAdrs
  | iNumActiveCounters is not in range of board ->
      return S_laser_INIT_INVALID_ARGUMENT
  | no reference signal || PLL is out of lock ||
    at least one board clock is not present || external clock not present ||
    *this.*clockStatus has invalid value ||
    LASER_MODE_CPR_SYNC_TO_EXT is out of range ||
    unknown signal is not present for all installed counters 
      print error message on console and proceed with true case
  | true -> *this.vmeAddress := vmeBoardAddress
    : *this.vu8LocalAddress := local address corresponding to vmeBoardAddress
    : *this.boardName := boardName
    : *this.bInit := TRUE
    : initialize board register pointers and protected variables
    : set all channels to exit test mode (enter operational mode)
    : Set Sync mode to external sync for all installed counters
    : return OK
  ]
RETURNS:
S_laser_INIT_INVALID_ARGUMENT
OK
*/

ERROR_CODE tFastLaserBoard::HardInit(int iNumCountersToUse)
{
  int i;
  ERROR_CODE ec;
 
  TraceMark(S_laser_ENTERING_HARDTINIT);

  /****
  [[U tFastLaserBoard::SoftInit()]]
  *****/
  ec = this->SoftInit(iNumCountersToUse);
  if(ec!=OK) { 
    return Traceback(ec, "tFastLaserBoard::HardInit", iNumCountersToUse);
  }

  /*
  [ this is invalid -> UNDEFINED
  | *this.bInit == FALSE || no reference signal || PLL is out of lock ||
    at least one board clock is not present || external clock not present ||
    *this.*clockStatus has invalid value ->
      print error message on console and proceed
  | true -> NOP
  ] */
  ec = this->CheckClocks();
  if(ec!=OK) {
    TraceMsg(tmt_WARN, ec,  "tFastLaserBoard::HardInit: CheckClocks failed\n");
  }

  /*
  [ this is invalid -> UNDEFINED
  | *this.bInit == FALSE || i is out of range ||
    LASER_MODE_CPR_SYNC_TO_EXT is out of range ||
    unknown signal is not present for all installed counters ->
      return appropriate error
  | true -> set all channels to exit test mode (enter operational mode)
    : Set Sync mode to external sync for all installed counters
    : return OK
    ] */
  for (i=0; i<this->iNumActiveCounters; i++) {
    ec = this->ChannelArray[i].TestModeDisable();
    if(ec!=OK) {
      return Traceback(ec, "tFastLaserBoard::HardInit", iNumCountersToUse);
    }

    TraceMsg(tmt_INFO, OK, "Setting Sync Mode for Counter %d", i);
    ec = this->ChannelArray[i].SetSyncMode();
    if(ec!=OK) {
      return Traceback(ec, "tFastLaserBoard::HardInit", iNumCountersToUse);
    }

    #ifndef   DO_NOT_CHECK_LASER_UNKNOWNS
    if(!this->ChannelArray[i].UnknownStatus()) {
      return Traceback(S_laser_UNKNOWN_NOT_PRESENT,
                 "tFastLaserBoard::HardInit", iNumCountersToUse);
    }
    #endif
  }

  TraceMark(S_laser_INITIALIZED_HARDLY);

  return OK;
}


/**************************************************
*
* HardInit  -  Void version.  Calls parameterized version
*              with default values.
* 
[[U tFastLaserBoard::HardInitVoid()]]
*
*/

ERROR_CODE tFastLaserBoard::HardInit(void)
{
  return (this->HardInit(this->CiDefaultNumCounters));
}


/**************************************************
*
* tFastLaserBoard::GetRefChannelNum()
*
*/
int tFastLaserBoard::GetRefChannelNum(int iUnkChannelNum)
{
  int i,
      iBadReturnValue = -1;  /* must be < 0 */

  for(i=0; i < LASER_NUM_CHANNEL_PAIRS; i++) {
    if(this->iChannelPair2DArray[i][1] == iUnkChannelNum) {
      return this->iChannelPair2DArray[i][0];
    }
  }
  
  return iBadReturnValue;
}



/******************************************************************************
*
* tFastLaserBoard::CreateChannel - gets channel specified by iNumChannel (global
*                             board numbering).  Updates iNumActiveChannels,
*                             iNumActiveCounters
*
[[U tFastLaserBoard::CreateChannel()]]
*/
tChannel *tFastLaserBoard::CreateChannel(int iChannelNum)
{
  int iCounterNumToCreate = iChannelNum % CiMaxNumCounters;

  if((this->IsInit() && iCounterNumToCreate >= this->iNumActiveCounters) ||
     iCounterNumToCreate >= this->CiMaxNumCounters || iChannelNum < 0  ||
     iChannelNum >= CiMaxNumChannels) {
    TraceMsg(tmt_ERROR, S_laser_RETURNING_NULL_CHANNEL_POINTER,
             "Returning NULL channel pointer for Fast Laser Board - channel #%d\n",
           iChannelNum); 
    return NULL;
  }

  this->iNumActiveChannels = MAX(this->iNumActiveChannels, iChannelNum+1);
  this->iNumActiveCounters = MAX(this->iNumActiveCounters, iCounterNumToCreate+1);

  if(iChannelNum < CiMaxNumCounters) {
    return (&(this->ChannelArray[iCounterNumToCreate]));
  }
  else {
    return (&(this->FidChannelArray[iCounterNumToCreate]));
  }
}

/************************************************************
*
* CheckClocks - examines the Status of Reference and Clock Signals
*
* CheckClocks checks for the presence of valid clock signals, reference,
* external, and internal.  In order for the hardware
* (in its default configuration).
* To reliably detect the presence of these signals, the external sync signal
* must be at least 2kHz, and all others (reference, unknowns, and clocks) must
* be above 300 kHz.
*
[[U tFastLaserBoard::CheckClocks()]]
*
SPEC :
  [ this is invalid -> UNDEFINED
  | *this.bInit == FALSE || no reference signal || PLL is out of lock ||
    at least one board clock is not present || external clock not present ||
    *this.*clockStatus has invalid value ->
      return appropriate error
  | true -> return OK
  ]
*
*
RETURNS:
S_laser_BOARD_NOT_INITIALIZED
S_laser_NO_REFERENCE
S_laser_PLL_OUT_OF_LOCK
S_laser_NO_64_MHZ
S_laser_NO_16_MHZ
S_laser_NO_2_MHZ
S_laser_NO_EXTERNAL_CLOCK
S_laser_UNKNOWN_CLOCK_ERROR
OK
*
*/
ERROR_CODE tFastLaserBoard::CheckClocks(void)
{
  int32_t curStatus;

  TraceMark(S_laser_CHECKING_CLOCKS);

  /*
  [ this is invalid -> UNDEFINED
  | *this.bInit == FALSE -> return uninit error
  | true -> NOP
  ] */
  if (!this->IsInit()) {
    return S_laser_BOARD_NOT_INITIALIZED;
  }

  /*
  [ clocks are healthy -> return OK
  | true -> NOP
  ] */
  if (((curStatus = *(this->vwClockStatus)) & LASER_HEALTHY_CLOCKS) ==
           LASER_HEALTHY_CLOCKS) {
    return OK;
  }

  /*
  [ no reference signal -> return no reference error
  | true -> NOP
  ] */
  if ( (LASER_HEALTHY_CLOCKS & LASER_HEALTHY_REF) &&
       (curStatus & LASER_HEALTHY_REF) != LASER_HEALTHY_REF ) {
    return S_laser_NO_REFERENCE;
  }

  /*
  [ PLL is out of lock -> return PLL out of lock error
  | true -> NOP
  ] */
  if ( (LASER_HEALTHY_CLOCKS & LASER_HEALTHY_PLL) &&
       (curStatus & LASER_HEALTHY_PLL) != LASER_HEALTHY_PLL ) {
    return S_laser_PLL_OUT_OF_LOCK;
  }

  /*
  [ 64MHz clock not present -> return appropriate error
  | true -> NOP
  ] */
  if ( (LASER_HEALTHY_CLOCKS & LASER_HEALTHY_64_MHZ) &&
       (curStatus & LASER_HEALTHY_64_MHZ) != LASER_HEALTHY_64_MHZ ) {
    return S_laser_NO_64_MHZ;
  }

  /*
  [ 16MHz clock not present -> return appropriate error
  | true -> NOP
  ] */
  if ( (LASER_HEALTHY_CLOCKS & LASER_HEALTHY_16_MHZ) &&
       (curStatus & LASER_HEALTHY_16_MHZ) != LASER_HEALTHY_16_MHZ ) {
    return S_laser_NO_16_MHZ;
  }

  /*
  [ 2MHz clock not present -> return appropriate error
  | true -> NOP
  ] */
  if ( (LASER_HEALTHY_CLOCKS & LASER_HEALTHY_2_MHZ) &&
       (curStatus & LASER_HEALTHY_2_MHZ) != LASER_HEALTHY_2_MHZ ) {
    return S_laser_NO_2_MHZ;
  }

  /*
  [ external clock not present -> return appropriate error
  | true -> NOP
  ] */
  if ( (LASER_HEALTHY_CLOCKS & LASER_HEALTHY_EXT) &&
       (curStatus & LASER_HEALTHY_EXT) != LASER_HEALTHY_EXT ) {
    return S_laser_NO_EXTERNAL_CLOCK;
  }

  /*
  [ true -> return unknown error message
  ] */
  return S_laser_UNKNOWN_CLOCK_ERROR;
}


/****************************************************************
*
* tFastLaserChannel::tFastLaserChannel()
*
[[U tFastLaserChannel::tFastLaserChannel()]]
*
*****/
tFastLaserChannel::tFastLaserChannel(void)
{
  this->viCprHi         = NULL;
  this->viFprHi         = NULL;
  this->viCprLo         = NULL;
  this->viFprLo         = NULL;
  this->vwUnknownStatus = NULL;
  this->vwFidWaitStates = NULL;
  this->vwFidStates     = NULL;
  this->viCommand       = NULL;
  this->viMode          = NULL;
  this->SetInit(FALSE);
}



/****************************************************************
*
ERROR_CODE tFastLaserChannel::Initialize(...)
*
[[U tFastLaserChannel::Initialize()]]
*
*****/
ERROR_CODE tFastLaserChannel::Initialize(volatile uint8_t *vu8BoardBase,
                                         volatile uint8_t *vu8ChannelBase,
                                         int              iChannelNumToUse)
{

  TraceMark(S_laser_ENTERING_CHANNEL_INITIALIZE);

  this->viCprHi         = (volatile int32_t *) (vu8ChannelBase + CiCprMswOffset);
  this->viCprLo         = (volatile int32_t *) (vu8ChannelBase + CiCprLswOffset);
  this->viFprHi         = (volatile int32_t *) (vu8ChannelBase + CiFprMswOffset);
  this->viFprLo         = (volatile int32_t *) (vu8ChannelBase + CiFprLswOffset);
  this->vwUnknownStatus = (volatile int16_t *)
                            (vu8ChannelBase + CiUnknownStatusOffset);
  this->viCommand       = (volatile int32_t *) (vu8ChannelBase + CiCommandOffset);
  this->viMode          = (volatile int32_t *) (vu8ChannelBase + CiModeOffset);
    
  this->vwFidWaitStates = (volatile int16_t *) (vu8BoardBase + CiFidWaitStatesOffset);
  this->vwFidStates     = (volatile int16_t *) (vu8BoardBase + CiFidStatesOffset);
  
  this->dChannelRange = pow(2.0, 32.0);
  this->iBitWidth     = this->CiBitWidth;
  this->ccConvCode    = this->CccConvCode;
  this->iChannelNum   = iChannelNumToUse;

  this->pRefChannel   = NULL;
  this->SetInit(TRUE);

  TraceMark(S_laser_CHANNEL_INITIALIZED);

  return OK;
}


/****************************************************************
*
ERROR_CODE tFastLaserChannel::Initialize(...)
*
[[U tFastLaserChannel::Initialize()]]
*
*****/
ERROR_CODE tFastLaserChannel::Initialize(volatile uint8_t *vu8BoardBase,
                                         volatile uint8_t *vu8ChannelBase,
                                         tFastLaserChannel *pRefChannelToUse,
                                         int              iChannelNumToUse)
{
  this->Initialize(vu8BoardBase, vu8ChannelBase, iChannelNumToUse);

  if(pRefChannelToUse != NULL)   this->pRefChannel = pRefChannelToUse;
  else  TraceMsg(tmt_WARN, S_laser_REF_CHANNEL_POINTER_SET_TO_NULL,
                 "tFastLaserChannel::Initialize");

  return OK;
}


/****************************************************************
*
* tFastLaserChannel::Input(double *pdValue) - reads all 40 bits of the counter
*
* The fast laser card counter is 40 bits in length.  This method
*  reads in the 40 bits and returns the value as a double.
*
[[U tFastLaserChannel::InputDouble()]]
*
*/

ERROR_CODE tFastLaserChannel::Input(double *pdValue) const
{
  /* create FAST (register) variables for the addresses to 
  * improve performance in accessing the two values */
  FAST volatile int32_t *viCprLoAddr,
                        *viCprHiAddr;

  FAST int32_t iCprLoValue,
          iCprHiValue;

  FAST double dValueFromRefChannel = 0.0;

  /*
  [ this is invalid -> UNDEFINED
  | *this.bInit == FALSE -> UNDEFINED
  | true -> NOP
  */ 
#ifdef DEBUG
  if (!this->bInit) {
    return S_laser_BOARD_NOT_INITIALIZED;
  }
#endif

  /*
  [ true -> read lo word of counter first, then hi word
            (board requirement). In order to ensure proper
            latching, read lo word a second time.
  ] */ 
  viCprLoAddr = this->viCprLo;
  viCprHiAddr = this->viCprHi;

  iCprLoValue = *viCprLoAddr;
  iCprHiValue = *viCprHiAddr;
  // Shouldn't need to read lo word a second time with new board
  // iCprLoValue = *viCprLoAddr;
  
  /*
  [ bit 7 of cprHiValue is set -> cprHiValue |= 0xffffff00
  | true -> cprHiValue &= 0x000000ff
  ] */
  if (iCprHiValue & 0x00000080) {
    iCprHiValue |= 0xffffff00;
  }
  else {
    iCprHiValue &= 0x000000ff;
  }

  /*
  [ true ->  *this.*pdValue := (cprHiValue << 32) + cprLoValue
    : return OK
  ] */
  int64_t i64Temp((uint32_t) iCprHiValue, (uint32_t) iCprLoValue);

  if(this->pRefChannel != NULL)  this->pRefChannel->Input(&dValueFromRefChannel);

  *pdValue = (double) (i64Temp) - dValueFromRefChannel;
  
  return OK;
}


/****************************************************************
*
* tFastLaserChannel::Input(int *piValue) - reads low 32 bits of the counter
*
* The fast laser card counter is 40 bits in length.  This method
*  reads the lower 32 bits, which consists of 24 bits of integer
*  and the full 8 bits of fraction.
*
[[U tFastLaserChannel::InputInt()]]
*
*/

ERROR_CODE tFastLaserChannel::Input(int *piValue) const
{
  FAST iValueFromRefChannel;

  /*
  [ this is invalid -> UNDEFINED
  | *this.bInit == FALSE -> UNDEFINED
  | true -> NOP
  */ 
#ifdef DEBUG
  if (!this->bInit) {
    return S_laser_BOARD_NOT_INITIALIZED;
  }
#endif

  if(this->pRefChannel != NULL)  this->pRefChannel->Input(&iValueFromRefChannel);

  /*
  [ true -> *piValue := low 32 bits of counter
  |  return OK
  ] */ 
  *piValue = *(this->viCprLo) - iValueFromRefChannel;
  return OK;
}


/*****************************************************
*
* tFastLaserChannel::TestModeEnable - Turn on test mode for a counter
*
* Allows you to turn on test mode for each counter.
* In test mode, the counter will use a pair of internally
* generated test signals rather than the externally supplied
* reference and unknown.
*
[[U tFastLaserChannel::TestModeEnable()]]
*
SPEC :
  [ this is invalid -> UNDEFINED
  | *this.bInit == FALSE =>
      return appropriate error
  | true -> set counter channel to enter test mode
    : return OK
  ]
RETURNS:
S_laser_BOARD_NOT_INITIALIZED
S_laser_INVALID_COUNTER
OK
*/

ERROR_CODE tFastLaserChannel::TestModeEnable(void)
{
  /*
  [ this is invalid -> UNDEFINED
  | *this.bInit == FALSE -> return uninit error
  | true -> NOP
  ] */
  if (!this->IsInit()) {
    return S_laser_BOARD_NOT_INITIALIZED;
  }

  /*
  [ true -> set counter channel to enter test mode
    : return OK
  ] */    
  *(this->viCommand) = CiCmdEnterTestMode;
  return OK;
}


/*****************************************************
*
* TestModeDisable - Turn off test mode for a counter
*
* Allows you to turn off test mode for each counter.
* In test mode, the counter will use a pair of internally
* generated test signals rather than the externally supplied
* reference and unknown.
*
[[U tFastLaserChannel::TestModeDisable()]]
*
SPEC :
  [ this is invalid -> UNDEFINED
  | *this.bInit == FALSE =>
      return appropriate error
  | true -> set counter channel to exit test mode (enter operational mode)
    : return OK
  ]
RETURNS:
S_laser_BOARD_NOT_INITIALIZED
S_laser_INVALID_COUNTER
OK
*/
ERROR_CODE tFastLaserChannel::TestModeDisable(void)
{ 
  /*
  [ this is invalid -> UNDEFINED
  | *this.bInit == FALSE -> return uninit error
  | true -> NOP
  ] */
  if (!this->IsInit()) {
    return S_laser_BOARD_NOT_INITIALIZED;
  }

  /*
  [ true -> set counter channel to exit test mode (enter operational mode)
    : return OK
  ] */
  *(this->viCommand) = CiCmdLeaveTestMode;
  return OK;
}

/*****************************************************
*
* SetSyncMode - sets the register synchronization mode for a counter.
*
* In addition to setting the synchronization mode for the counter,
* SetSyncMode is also used to set the sync mode prior to
* issuing a ClearCounter command in order to select the
* desired clear mode.
* The desired mode for the counter is specified by the argument
* newMode.  For sync mode selection, it must be one of:
* LASER_MODE_CPR_FLOWTHROUGH,
* LASER_MODE_CPR_SYNC_TO_REF, LASER_MODE_CPR_SYNC_TO_EXT,
* LASER_MODE_FPR_LOAD_ON_FIDUCIAL,
*    OR
* for selecting a clear mode, it must be one of:
* LASER_CLEAR_IMMED, LASER_CLEAR_ON_REF, LASER_CLEAR_ON_EXT,
* LASER_CLEAR_ON_FIDUCIAL.
* Note that the various LASER_MODE_... and LASER_CLEAR_... modes are
* actually the same.  For example, calling SetSyncMode with
* an argument of LASER_CLEAR_IMMED will prepare the counter to be
* cleared immediately on the next clearCounter call, but it
* will also cause the counter to switch to LASER_MODE_CPR_FLOWTHROUGH
* register synchronization mode.
* See the source code for SoftInit for an example of how to use this procedure.
*
[[U tFastLaserChannel::SetSyncMode()]]
*
SPEC :
  [ this is invalid -> UNDEFINED
  | *this.bInit == FALSE || counter<1 || counter>*this.laserNumInstalled ||
    newMode<LASER_MODE_LOWEST_VALUE || newMode>LASER_MODE_HIGHEST_VALUE ->
      return appropriate error
  | true -> Set Sync mode of counter to newMode
    : return OK
  ]
RETURNS:
S_laser_BOARD_NOT_INITIALIZED
S_laser_INVALID_COUNTER
S_laser_INVALID_SYNC_MODE
OK
*
*/

ERROR_CODE tFastLaserChannel::SetSyncMode(int iNewMode=CiDefaultSyncMode)
{
  /* 
  [ this is invalid -> UNDEFINED
  | *this.bInit == FALSE -> return uninit error
  | true -> NOP
  ] */
  if (!this->IsInit()) {
    return S_laser_BOARD_NOT_INITIALIZED;
  }

  /*
  [ newMode<LASER_MODE_LOWEST_VALUE || newMode>LASER_MODE_HIGHEST_VALUE ->
      return invalid sync mode error
  | true -> NOP
  ] */
  if (iNewMode<CiModeLowestValue || iNewMode>CiModeHighestValue) {
    return S_laser_INVALID_SYNC_MODE;
  }
 
  if(this->pRefChannel != NULL)  this->pRefChannel->SetSyncMode(iNewMode);

  /*
  [ true -> Set Sync mode of counter to external sync
    : return OK
  ] */ 
  *(this->viMode) = (int32_t) iNewMode;
  return OK;
}

/************************************************************
*
* tFastLaserChannel::ClearCounter - clears (zeros) this counter.
*
* Clears the counter using the
* clear mode selected in the most recent SetSyncMode call.
*
[[U tFastLaserChannel::ClearCounter()]]
* 
SPEC :
  [ this is invalid -> UNDEFINED
  | *this.bInit == FALSE || counter<1 || counter>*this.laserNumInstalled -> 
      UNDEFINED
  | true -> Clear specified laser counter
    : return OK
  ]
RETURNS:
S_laser_BOARD_NOT_INITIALIZED
S_laser_INVALID_COUNTER
OK
*
*/
ERROR_CODE tFastLaserChannel::ClearCounter(void)
{
  /*
  [ this is invalid -> UNDEFINED
  | *this.bInit == FALSE -> UNDEFINED
  | true -> NOP
  ] */
#ifdef DEBUG
  if (!this->bInit) {
    return S_laser_BOARD_NOT_INITIALIZED;
  }
#endif

  if(this->pRefChannel != NULL)  this->pRefChannel->ClearCounter();

  /*
  [ true -> Clear specified laser counter
    : return OK
  ] */
  *(this->viCommand) = CiCmdClearCounter;
  return OK;
}


/************************************************************
*
* tFastLaserChannel::LoadOnFiducial - 
*
* This routine simply commands the counter to load the value
* of the CPR into the FPR the next time the fiducial is crossed
* (effectively storing the fiducial position in the FPR).  Note that the
* synchronization mode of the CPR is unaffected by this routine.
* If you desire to also clear the CPR on the next fiducial crossing, use
* the laserClearCounterOnFiducial procedure.
*
[[U tFastLaserChannel::LoadOnFiducial()]]
*
*/

ERROR_CODE tFastLaserChannel::LoadOnFiducial(void)
{
  ERROR_CODE ec;
  
  ec = this->SetSyncMode(CiModeFprLoadOnFiducial);
  if(ec != OK) {
    return Traceback(ec, "tFastLaserChannel::LoadOnFiducial");
  }
  
  return OK;
}


/************************************************************
*
* tFastLaserChannel::ClearCounterOnFiducial - 
*
* This routine sets up the counter so that the next fiducial
* crossing will cause the CPR to be loaded into the FPR and then the
* counter will be cleared.  Thus, after the next fiducial crossing, the
* FPR contains the position of the fiducial, and the CPR will contain
* zero.  Note that the CPR synchronization mode will continue to be
* the same as it was before this call.  The only difference is that,
* when the fiducial is crossed, the CPR will suddenly jump to zero and
* then continue as before.
*
[[U tFastLaserChannel::ClearCounterOnFiducial()]]
*
*/

ERROR_CODE tFastLaserChannel::ClearCounterOnFiducial(void)
{
  /*
  [ this is invalid -> UNDEFINED
  | *this.bInit == FALSE -> UNDEFINED
  | true -> NOP
  ] */
#ifdef DEBUG
  if (!this->bInit) {
    return S_laser_BOARD_NOT_INITIALIZED;
  }
#endif

  if (this->LoadOnFiducial() != OK || this->ClearCounter() != OK) {
    return Traceback(S_laser_ERROR_IN_CLEAR_OR_LOAD,
                     "tFastLaserChannel::ClearCounterOnFiducial");
  }
  return(OK);
}

/************************************************************
*
* tFastLaserChannel::ResetFiducialWait
*
* This routine will cancel the effect of any
* previous call to laserLoadFprOnFiducial or laserClearCounterOnFiducial (assuming
* the fiducial hasn't been crossed since that call).
*
[[U tFastLaserChannel::ResetFiducialWait()]]
*
*/

ERROR_CODE tFastLaserChannel::ResetFiducialWait(void)
{
  /*
  [ this is invalid -> UNDEFINED
  | *this.bInit == FALSE -> UNDEFINED
  | true -> NOP
  ] */
#ifdef DEBUG
  if (!this->bInit) {
    return S_laser_BOARD_NOT_INITIALIZED;
  }
#endif

  if(this->pRefChannel != NULL)  this->pRefChannel->ResetFiducialWait();

  *viCommand = this->CiCmdResetWaitStates;
  return(OK);
}


/************************************************************
*
* tFastLaserChannel::FiducialWaitState
*
* This routine returns TRUE if any part of
* the specified counter is waiting for a fiducial crossing (i.e.,
* if there is a pending Load-FPR-on-fiducial-crossing or Clear-
* CPR-on-fiducial-crossing condition).  You can use this function
* to monitor whether your commanded action has occurred yet (i.e.,
* whether the fiducial has been crossed since you issued the
* command).  If there is no pending wait-for-fiducial state, this
* function will return FALSE.
*
[[U tFastLaserChannel::FiducialWaitState()]]
*
*/
BOOL tFastLaserChannel::FiducialWaitState(void)
{
  /*
  [ this is invalid -> UNDEFINED
  | *this.bInit == FALSE -> UNDEFINED
  | true -> NOP
  ] */
#ifdef DEBUG
  if (!this->bInit) {
    return S_laser_BOARD_NOT_INITIALIZED;
  }
#endif

  if(this->pRefChannel != NULL)  this->pRefChannel->FiducialWaitState();

  if ( (0x01 & (*(this->vwFidWaitStates) >> this->iChannelNum)) == 0 ) {
    return(FALSE);
  }
   
  return(TRUE);
}


/************************************************************
*
* tFastLaserChannel::IsFiducialOn - 
*
* This routine returns TRUE if the specified
* counter's fiducial is currently tripped (in other
* words, if the cart is at the fiducial) and returns FALSE
* otherwise.
*
[[U tFastLaserChannel::IsFiducialOn()]]
*
*/

BOOL tFastLaserChannel::IsFiducialOn(void)
{
  /*
  [ this is invalid -> UNDEFINED
  | *this.bInit == FALSE -> UNDEFINED
  | true -> NOP
  ] */
#ifdef DEBUG
  if (!this->bInit) {
    return S_laser_BOARD_NOT_INITIALIZED;
  }
#endif

  if(this->pRefChannel != NULL)  this->pRefChannel->IsFiducialOn();

  if ( (0x01 & (*(this->vwFidStates) >> this->iChannelNum)) == 0 ) {
    return(FALSE);
  }
  
  return(TRUE);
}


/****************************************************************
*
* tFastLaserChannel::UnknownStatus - check if Unknown signal is present
*
SPEC :
  [ this is invalid -> UNDEFINED
  | *this.bInit == FALSE || counter<1 || counter>*this.laserNumInstalled ||
    unknown signal is not present for counter -> 
      return FALSE
  | true -> return TRUE
  ] 
*
[[U tFastLaserChannel::UnknownStatus()]]
*
*/

BOOL tFastLaserChannel::UnknownStatus(void)
{
  /*
  [ this is invalid -> UNDEFINED
  | *this.bInit == FALSE -> return uninit error
  | true -> NOP
  ] */
  if (!this->IsInit()) {
    return S_laser_BOARD_NOT_INITIALIZED;
  }

  if(this->pRefChannel != NULL)  this->pRefChannel->UnknownStatus();

  /*
  [ bit number (this->iChannelNum) of *this.*vwUnknownStatus is not set ->
      return FALSE
  ] */
  if ( (0x01 & (*(this->vwUnknownStatus) >> this->iChannelNum)) == 0 ) {
    return FALSE;
  }

  return TRUE;
}


