/* $RCSfile: mpv906.cc,v $ - Driver software for MPV906 A/D board */
/* 
$Header: /home/chara/cvsroot/ople/sys/hw/drivers/mpv906.cc,v 1.1.1.1 1998/06/11 21:36:34 chara Exp $
*/
/*
* This file contains the state box spec and methods necessary to initialize and use
*  the MPV906 A/D board.  
*/

/*
modification history
--------------------
01a,pci,28Apr97               written
*/

#include "isit.h"
#include "hwTypesP.h"
#include "board.h"
#include "inputChannel.h"
#include "channelBank.h"
#include "mpv906.h"

/*******************************************************************************
*
* tMpv906Board BB -> SB mapping
*
*    BB Representation           SB Representation
*    -----------------           -----------------
*    sample period               200ns * (*vuwPeriodReg(bits7-0) + 1) *
*                                        (iBaseSamplePeriod + 1)
*    sample frequency            1/(sample period)
*    amAdMode                    *vuwIntrCtrlReg   (bit 11)
*    smScanMode                  *vuwIoCtrlReg     (bit 9)
*    tsTriggerSelect             *vuwTrigSourceReg (bits 2-1)
******/

/*******************************************************************************
*
* ERROR_CODE tMpv906Board::tMpv906Board(void) 
*
[[U tMpv906Board::tMpv906Board()]]
*    
*/

tMpv906Board::tMpv906Board(int iBoardNum) : ChannelBank(this)
{
  TraceMark(S_mpv906_ENTERING_CONSTRUCTOR);

  switch(iBoardNum) {
   case 0:
    this->vmeAddress = tVmeAddress(VME_AM_SUP_SHORT_IO, (uint8_t *) 0xfc00);
    break;
   default:
    this->vmeAddress = tVmeAddress();
    break;
  }
  this->boardName  = MPV906_BOARD_NAME + iBoardNum;
  vuwIntrCtrlReg = NULL;
  vuwTrigSourceReg = NULL;
  vuwPeriodReg = NULL;

  TraceMark(S_mpv906_DONE_CONSTRUCTING);
}


/*******************************************************************************
*
* ERROR_CODE tMpv906Board::SoftInit(...)
*
[[U tMpv906Board::SoftInitParams()]]
*    
*/

ERROR_CODE tMpv906Board::SoftInit(double       dSampleFreqToUse,
                                  int          iBaseSamplePeriodToUse,
                                  int          iStartChannelToUse,
                                  CONV_CODE    ccConvCodeToUse,
                                  int          iVoltageRangeToUse,
                                  int          iGainToUse)
{
  ERROR_CODE ec;

  TraceMark(S_mpv906_ENTERING_SOFTINIT);

  ec = tBoard::Initialize();
  if(ec!=OK) {
    return Traceback(ec, "tMpv906Board::SoftInit", dSampleFreqToUse,
                      iBaseSamplePeriodToUse,
                      ccConvCodeToUse, iVoltageRangeToUse, iGainToUse);
  }

  vuwIntrCtrlReg    = (volatile uint16_t *) 
                        (this->vu8LocalAddress + CiIntrCtrlRegOffset);
  vuwTrigSourceReg  = (volatile uint16_t *)
                        (this->vu8LocalAddress + CiTrigSourceRegOffset);
  vuwPeriodReg      = (volatile uint16_t *)
                        (this->vu8LocalAddress + CiPeriodRegOffset);
  vuwIoCtrlReg      = (volatile uint16_t *)
                        (this->vu8LocalAddress + CiIoCtrlRegOffset);

  this->iBaseSamplePeriod  = iBaseSamplePeriodToUse;
  this->iBoardGain         = iGainToUse;

  if((this->iNumActiveChannels < 1) ||
     (this->iNumActiveChannels > this->GetMaxNumChannelsFromAdMode())) {
    TraceMsg(tmt_WARN, OK,
              "Invalid number of channels for MPV906, setting to default");
    this->iNumActiveChannels = this->CiDefaultNumChannels;
  }

  ec = ChannelBank.Initialize(
            (volatile int16_t *) (this->vu8LocalAddress + this->CiDataOffset),
            (volatile uint8_t *) (this->vu8LocalAddress + this->CiScanListOffset),
            (volatile uint16_t *) (this->vu8LocalAddress + this->CiIoCtrlRegOffset),
            this->iNumActiveChannels,
            iStartChannelToUse,
            ccConvCodeToUse,
            iVoltageRangeToUse);

  if(ec != OK) {
     return Traceback(ec, "tMpv906Board::SoftInit", dSampleFreqToUse,
                      iBaseSamplePeriodToUse, ccConvCodeToUse, iVoltageRangeToUse);
  }

  this->SetInit(TRUE);

  TraceMark(S_mpv906_INITIALIZED_SOFTLY);
  return OK;
}


/*******************************************************************************
*
* ERROR_CODE tMpv906Board::SoftInit(void)
*
[[U tMpv906Board::SoftInitVoid()]]
*    
******/
ERROR_CODE tMpv906Board::SoftInit(void)
{
  return (this->SoftInit(CdDefaultSampleFreq,
                        CiDefaultBaseSamplePeriod,
                        CiDefaultStartChannel,
                        CccDefaultConvCode,
                        CiDefaultVoltageRange,
                        CiDefaultBoardGain));
}


/*******************************************************************************
*
* ERROR_CODE tMpv906Board::HardInit(...)
*
[[U tMpv906Board::HardInitParams()]]
*    
******/
ERROR_CODE tMpv906Board::HardInit(double         dSampleFreqToUse,
                                  int            iBaseSamplePeriodToUse,
                                  int            iStartChannelToUse,
                                  CONV_CODE      ccConvCodeToUse,
                                  int            iVoltageRangeToUse,
                                  int            iGainToUse,
                                  SCAN_MODE      smScanModeToUse,
                                  TRIGGER_SELECT tsTriggerSelectToUse)
{
  ERROR_CODE ec, ec1, ec2;
  double dActualSampleFreq;
  char sMsg[256];

  TraceMark(S_mpv906_ENTERING_HARDINIT);

  ec = this->SoftInit(dSampleFreqToUse,
                      iBaseSamplePeriodToUse, iStartChannelToUse,
                      ccConvCodeToUse, iVoltageRangeToUse, iGainToUse);
  if(ec != OK) {
    return Traceback(ec, "tMpv906Board::HardInit",
                     dSampleFreqToUse, iBaseSamplePeriodToUse,
                     ccConvCodeToUse, iVoltageRangeToUse, iGainToUse,
                     smScanModeToUse, smScanModeToUse, tsTriggerSelectToUse);
  }

  dActualSampleFreq = this->SetSampleFrequency(dSampleFreqToUse);
  TraceMsg(tmt_INFO, OK,
       "tMpv906Board::HardInit:  The actual sample frequency is %f\n",
       dActualSampleFreq);

  ec  = this->SetScanMode(smScanModeToUse);
  ec1 = this->SetTriggerSelect(tsTriggerSelectToUse);
  if(ec!=OK || ec1!=OK) {
    return Traceback(ec, "tMpv906Board::HardInit",
                     dSampleFreqToUse, iBaseSamplePeriodToUse,
                     ccConvCodeToUse, iVoltageRangeToUse, iGainToUse,
                     smScanModeToUse, smScanModeToUse, tsTriggerSelectToUse);
  }

  ec = this->ChannelBank.StartScan();      // okay, all done...  let's go!
  if(ec!=OK) 
    return Traceback(ec, "tMpv906Board::HardInit",
                dSampleFreqToUse, iBaseSamplePeriodToUse,
                ccConvCodeToUse, iVoltageRangeToUse, iGainToUse,
                smScanModeToUse, smScanModeToUse, tsTriggerSelectToUse);

  TraceMark(S_mpv906_INITIALZED_HARDLY);

  return OK;
}


/*******************************************************************************
*
* ERROR_CODE tMpv906Board::HardInit(void)
*
[[U tMpv906Board::HardInitVoid()]]
*    
******/
ERROR_CODE tMpv906Board::HardInit(void)
{
  return (this->HardInit(CdDefaultSampleFreq,
                         CiDefaultBaseSamplePeriod,
                         CiDefaultStartChannel,
                         CccDefaultConvCode,
                         CiDefaultVoltageRange,
                         CiDefaultBoardGain,
                         CsmDefaultScanMode,
                         CtsDefaultTriggerSelect));
}


/******************************************************************************
*
* tMpv906Board::CreateChannel - gets channel specified by iChannelNum (global
*                             board numbering)
*
[[U tMpv906Board::CreateChannel()]]
*/
tChannel *tMpv906Board::CreateChannel(int iChannelNum)
{
  char msg[256];

  if((this->IsInit() && iChannelNum >= this->iNumActiveChannels) ||
     iChannelNum < 0 || iChannelNum >= this->CiMaxChannels) {
    sprintf(msg, "ERROR: Returning NULL channel pointer for MPV906,channel #%d",
            iChannelNum);
    TraceMsg(tmt_ERROR, S_dvme626_RETURNING_NULL_CHANNEL_POINTER, msg);
    return NULL;
  }

  this->iNumActiveChannels = MAX(this->iNumActiveChannels, (iChannelNum+1));

  /****
  [[U tMpv906AdChannelBank::CreateChannel()]]
  ****/
  return (this->ChannelBank.CreateChannel(iChannelNum));
}


/*******************************************************************************
*
* ERROR_CODE tMpv906Board::SetSampleFrequency(...)
* 
*    dSampleFreq is in Hz
*
[[U tMpv906Board::SetSampleFrequency()]]
*    
******/
double tMpv906Board::SetSampleFrequency(double dSampleFreq)
{
  int iNewPeriod; /* period in ns */

  if((dSampleFreq < CdMinSampleFreq) || (dSampleFreq > CdMaxSampleFreq)) {
    return this->GetSampleFrequency();
  }

  /* get nearest integer value for the period register */
  iNewPeriod = (int) (5000000 / (dSampleFreq*(this->iBaseSamplePeriod+1))) - 1;
  *this->vuwPeriodReg = (uint16_t) iNewPeriod;

  /* return the actual frequency value set */
  return (1.0/(200e-9 * (this->iBaseSamplePeriod + 1) * (iNewPeriod + 1)));
}

/*******************************************************************************
*
* ERROR_CODE tMpv906Board::SetScanMode(void)
*
[[U tMpv906Board::SetScanMode()]]
*    
******/
ERROR_CODE tMpv906Board::SetScanMode(SCAN_MODE smScanModeToUse)
{
  switch(smScanModeToUse) {
   case(SM_CONTINUOUS):
     *this->vuwIntrCtrlReg |= 0x0800;
     break;
   case(SM_ONE_SHOT):
     *this->vuwIntrCtrlReg &= 0xf7ff;
     break;
   default:
     return S_mpv906_INVALID_SCAN_MODE;
  }
  return OK;
}


/*******************************************************************************
*
* ERROR_CODE tMpv906Board::SetTriggerSelect(void)
*
[[U tMpv906Board::SetTriggerSelect()]]
*    
******/
ERROR_CODE tMpv906Board::SetTriggerSelect(TRIGGER_SELECT tsTriggerSelectToUse)
{
  switch(tsTriggerSelectToUse) {
   case(TS_SOFTWARE):
     *this->vuwTrigSourceReg = 0x0000;
     break;
   case(TS_INTERNAL):
     *this->vuwTrigSourceReg = 0x0001;
     break;
   case(TS_EXTERNAL):
     *this->vuwTrigSourceReg = 0x0002;
     break;
   case(TS_EVENT):
     *this->vuwTrigSourceReg = 0x0003;
     break;
   default:
     return S_mpv906_INVALID_TRIGGER_SELECT_VALUE;
  }
  return OK;
}

/*******************************************************************************
*
* ERROR_CODE tMpv906Board::SetNumChannels(...)
*
[[U tMpv906Board::SetNumChannels()]]
*    
******/
ERROR_CODE tMpv906Board::SetNumChannels(int iNumChannelsToUse)
{
  if((iNumChannelsToUse < 0) ||
     (iNumChannelsToUse > this->GetMaxNumChannelsFromAdMode())) {
    return Traceback(S_mpv906_CHANNEL_NUMBER_OUT_OF_RANGE,
                     "tMpv906Board::SetNumChannels", iNumChannelsToUse);
  }

  this->iNumActiveChannels = iNumChannelsToUse;
  return OK;
}
  

/*******************************************************************************
*
* ERROR_CODE tMpv906Board::GetMaxNumChannelsFromAdMode(void)
*
[[U tMpv906Board::GetMaxNumChannelsFromAdMode()]]
*    
******/
int tMpv906Board::GetMaxNumChannelsFromAdMode(void)
{
  if(*(this->vuwIoCtrlReg) & 0x0200) {
    return (64);                      /* single_ended mode */
  }
  else {
    return (32);                      /* differential mode */
  }
}


/*******************************************************************************
*
* tMpv906AdChannelBank BB -> SB mapping
*
*    BB Representation           SB Representation
*    -----------------           ----------------- 
*    smScanMode                  uwIoCtrlReg     (bit 9)
*    iChannelGain                *vu8ScanList[0..63] (bits 7-6)
*
******/

/*******************************************************************************
*
* ERROR_CODE tMpv906AdChannelBank::tMpv906AdChannelBank(void)
*
[[U tMpv906AdChannelBank::tMpv906AdChannelBank()]]
*    
******/
tMpv906AdChannelBank::tMpv906AdChannelBank(tMpv906Board *pBoardToUse)
{
  this->pBoard = pBoardToUse;
  this->vu8ScanList = NULL;
  this->vuwIoCtrlReg = NULL;
  this->SetInit(FALSE);
}

/*******************************************************************************
*
* ERROR_CODE tMpv906AdChannelBank::Initialize(...)
*
[[U tMpv906AdChannelBank::Initialize()]]
*    
******/
ERROR_CODE tMpv906AdChannelBank::Initialize(volatile int16_t      *vwDataToUse,
                                          volatile uint8_t      *vu8ScanListToUse,
                                          volatile uint16_t     *vuwIoCtrlRegToUse,
                                          int                   iNumChannelsToUse,
                                          int                   iStartChannelToUse,
                                          CONV_CODE             ccConvCodeToUse,
                                          int                   iVoltageRangeToUse)
{
  ERROR_CODE ec;
  int        i,
             iTempNumChannels,
             iTempOffsetChannel;

  TraceMark(S_mpv906_ENTERING_CHANNELBANK_INITIALIZE);

  /*****
  * Make this->iNumChannels an iteger multiple of 64 (number of scan list locations)
  *****/
  while(64 % (iTempNumChannels = this->pBoard->GetNumChannels())) {
    this->pBoard->SetNumChannels(++iTempNumChannels);
  }

  this->vu8ScanList   = vu8ScanListToUse;
  this->vuwIoCtrlReg  = vuwIoCtrlRegToUse;
  this->iStartChannel = iStartChannelToUse;
  this->bScanning = FALSE;   // eamck
  this->vwChannelData = vwDataToUse;  //beh

  for(i=0; i<iNumChannelsToUse; i++) {
    this->ChannelArray[i].Initialize(this, (vwDataToUse + i), i, ccConvCodeToUse,
                             iVoltageRangeToUse);
  }
 
  this->SetInit(TRUE);

  /* initialize scan list and set all channel gains to 1 */ 
  this->LoadScanList(TRUE);

  TraceMark(S_mpv906_DONE_INITIALIZING_CHANNELBANK);

  return OK;
}


/******************************************************************************
*
* tMpv906ChannelBank::CreateChannel - gets channel specified by iNumChannel
*                                     (global numbering)
*
[[U tMpv906ChannelBank::CreateChannel()]]
*/
tChannel *tMpv906AdChannelBank::CreateChannel(int iChannelNum)
{
  char msg[256];

  if(iChannelNum < 0 || iChannelNum >= this->pBoard->GetNumChannels()) {
    sprintf(msg, "ERROR: Returning NULL channel pointer for MPV906,channel #%d",
            iChannelNum);
    TraceMsg(tmt_ERROR, S_dvme626_RETURNING_NULL_CHANNEL_POINTER, msg);
  }

  return (&(ChannelArray[iChannelNum]));
}


/*******************************************************************************
*
* ERROR_CODE tMpv906AdChannelBank::StartScan(void)
*
[[U tMpv906AdChannelBank::StartScan()]]
*    
******/
ERROR_CODE tMpv906AdChannelBank::StartScan(void)
{
  if(this->IsInit() == FALSE) {
    return Traceback(S_mpv906_CHANNEL_BANK_NOT_INITIALIZED, 
                     "tMpv906AdChannelBank::StartScan");
  }
  
  if(this->bScanning) {
    TraceMsg(tmt_WARN, OK,
       "tMpv906AdChannelBank::StartScan called with bScanning TRUE");
    return OK;
  }

  *(this->vuwIoCtrlReg) = this->CuwStartScanValue;
  this->bScanning = TRUE;

  return OK;
}
  
/*******************************************************************************
*
* ERROR_CODE tMpv906AdChannelBank::SetGain(...)
*
[[U tMpv906AdChannelBank::SetGain()]]
*    
******/
ERROR_CODE tMpv906AdChannelBank::SetGain(int iChannel, int iNewGain)
{
  uint8_t u8GainValue;
  int i;

  if(iChannel >= this->pBoard->GetNumChannels()) {
    return Traceback(S_mpv906_CHANNEL_NUMBER_OUT_OF_RANGE,
                     "tMpv906AdChannelBank::SetGain", iChannel, iNewGain);
  }

  switch(iNewGain) {
   case 1:
     u8GainValue = 0x00;
     break;
   case 10:
     u8GainValue = 0x40;
     break;
   case 100:
     u8GainValue = 0x80;
     break;
   default:
     return Traceback(S_mpv906_INVALID_GAIN,
                      "tMpv906AdChannelBank::SetGain", iChannel, iNewGain);
     break;
  }
     
  for(i=0; i < this->CiNumScanListLoc; i++) { 
    if((*(this->vu8ScanList + i) & 0x3f) == iChannel) {
      *(this->vu8ScanList + i) &= 0x3f;
      *(this->vu8ScanList + i) |= u8GainValue;
    }
  }
  
  return OK;
}


/*******************************************************************************
*
* uint8_t tMpv906AdChannelBank::GetScanListPtr()
*
* Returns 8-bit scan list pointer (max value = 63).  This pointer indicates
*  which register location the input will be stored.  It was discovered
*  through testing that this value is not entirely accurate.  It is
*  actually necessary to subract an offset from the pointer to get a valid
*  value.  Otherwise, you may get old data.
*
*/
#define SL_OFFSET 4

uint8_t tMpv906AdChannelBank::GetScanListPtr()
{
return (uint8_t) (((*this->vuwIoCtrlReg & 0x003f) - SL_OFFSET) & 0x003f);
}


/*******************************************************************************
*
* ERROR_CODE tMpv906AdChannelBank::LoadScanList(...)
*
*  Loads the scan list continuously with values from this->iStartChannel to
*   (this->iStartChannel+this->iNumActiveChannels-1).
*
*  NOTE:
*    This code assumes that the number of active channels on the board is an
*      integer multiple of 64.
*      If it is not, strange things
*      may happen because the active channels will not fit evenly into the 
*      scan list, and when the scan list wraps around, the wrong channel info
*      might be read on a call to tMpv906AdChannel::Input().
*
[[U tMpv906AdChannelBank::LoadScanList()]]
*    
******/
ERROR_CODE tMpv906AdChannelBank::LoadScanList(BOOL bResetGain = FALSE)
{
  int i, iCurrentOffset,
      iCurrentMaxChannel = this->pBoard->GetMaxNumChannelsFromAdMode() - 1,
      iCurrentNumChannels = this->pBoard->GetNumChannels();

  if(this->IsInit() == FALSE) {
    return Traceback(S_mpv906_CHANNEL_BANK_NOT_INITIALIZED,
                     "tMpv906AdChannelBank::LoadScanList");
  }

  if(this->iStartChannel < 0 || this->iStartChannel > iCurrentMaxChannel ||
    (this->iStartChannel + this->pBoard->GetNumChannels()-1) > iCurrentMaxChannel) {
    return Traceback(S_mpv906_INVALID_START_CHANNEL_VALUE,
                     "tMpv906AdChannelBank::LoadScanList");
  }

  /*******
  * Load the list - if bResetGain is TRUE, the upper 2 bits get cleared,
  *                  otherwise, they are preserved 
  *
  *   Scan list is little endian, so flip offset.
  *         Scan List Element 0  @ Offset 0x81
  *         Scan List Element 1  @ Offset 0x80
  *         Scan List Element 2  @ Offset 0x83
  *         Scan List Element 3  @ Offset 0x82    etc...
  *******/
  for(i=this->iStartChannel; i<this->CiNumScanListLoc; i++) {
    if(i%2 == 0) 
          iCurrentOffset = i+1;
    else  
          iCurrentOffset = i-1;

    if(bResetGain) {
      *(this->vu8ScanList + iCurrentOffset) = i % iCurrentNumChannels;
    }
    else {
      *(this->vu8ScanList + iCurrentOffset) &= 0xc0;
      *(this->vu8ScanList + iCurrentOffset) += i % iCurrentNumChannels;
    }
  }

  return OK;    
}
   
  
/*******************************************************************************
*
* ERROR_CODE tMpv906AdChannelBank::GetCurrentChannelAddress
*
* We want to find the most recently read value of our data.
* If we are using fewer than 64 channels there may be more than one
* value to choose from.  For example, with 32 channels (the nominal
* case when reading in all channels in differential mode),
* each value is stored in two places.
*
* To do this computation, we think of the conversions as being
* organized in "segments", with each segment representing a
* complete set of conversions from all channels (32 in our example,
* for a total of two segments).
*
[[U tMpv906AdChannelBank::GetCurrentChannelAddress()]]
*    
*/

volatile int16_t *tMpv906AdChannelBank::GetCurrentChannelAddress(int iChannel)
{
  uint8_t   u8CurrentSlPtr  = this->GetScanListPtr();
  int iChannelsPerSegment, iSegmentNumber;
  int iFirstChannelInCurrentSegment, iOurLocationInCurrentSegment;
  int iIndexToRead;
  
  /* Find out which segment (0-based) the current conversion pointer is in */
  iChannelsPerSegment = this->pBoard->GetNumChannels();
  iSegmentNumber = u8CurrentSlPtr / iChannelsPerSegment;

  /* In the current segment, has the channel pointer passed our
   * location yet?  If so, we should read from the current segment.
   * If not, read from the previous segment.
   */
  iFirstChannelInCurrentSegment = iSegmentNumber * iChannelsPerSegment;
  iOurLocationInCurrentSegment = iFirstChannelInCurrentSegment + iChannel;
  if ( u8CurrentSlPtr <= iOurLocationInCurrentSegment )
    iSegmentNumber--;

  /* If the segment we want to read from is negative, that means
   * we read from the last segment, which was recorded earlier.
   */
  if (iSegmentNumber < 0)
    iSegmentNumber = this->pBoard->GetMaxChannels() / iChannelsPerSegment - 1;

  iIndexToRead = iSegmentNumber * iChannelsPerSegment + iChannel;

  return (this->vwChannelData + iIndexToRead);
}



/*******************************************************************************
*   
* tMpv906AdChannel  BB -> SB mapping
*
*    BB Representation           SB Representation
*    -----------------           -----------------
*    ccConvCode                  Jumper J11
*    iChannelRange               Jumpers J9 & J10
*
******/

/*******************************************************************************
*   
* tMpv906AdChannel::tMpv906AdChannel
*
******/
tMpv906AdChannel::tMpv906AdChannel(void)
{
  this->vwData = NULL;
}

/*******************************************************************************
*
* ERROR_CODE tMpv906AdChannel::Initialize(...)
*
[[U tMpv906AdChannel::Initialize()]]
*    
******/
ERROR_CODE tMpv906AdChannel::Initialize(tMpv906AdChannelBank *pChannelBankToUse,
                                        volatile int16_t     *vwDataToUse,
                                        int                  iDataOffsetToUse,
                                        CONV_CODE            ccConvCodeToUse,
                                        int                  iChannelRangeToUse)
{ 
  TraceMark(S_mpv906_ENTERING_CHANNEL_INITIALIZE);

  this->pChannelBank  = pChannelBankToUse;
  this->vwData        = vwDataToUse;
  this->iDataOffset   = iDataOffsetToUse;
  this->ccConvCode    = ccConvCodeToUse;
  this->dChannelRange = (double) iChannelRangeToUse;
  this->iBitWidth     = this->CiBitWidth;
  this->SetInit(TRUE);

  TraceMark(S_mpv906_DONE_INITIALIZING_CHANNEL);

  return OK;
}

/*******************************************************************************
*
* ERROR_CODE tMpv906AdChannel::SetGain(int iGainToUse)      -- eamck
*
[[X tMpv906AdChannel::SetGain()]]
*
******/

ERROR_CODE tMpv906AdChannel::SetGain(int iGainToUse)
{
   return this->pChannelBank->SetGain(this->iDataOffset,iGainToUse); 
}


/*******************************************************************************
*
* ERROR_CODE tMpv906AdChannel::Input(...)
*
[[U tMpv906AdChannel::Input()]]
*
******/

ERROR_CODE tMpv906AdChannel::Input(int *piValue) const
{
  int16_t wReadValue;
  volatile int16_t *pChannelData =
    this->pChannelBank->GetCurrentChannelAddress(this->iDataOffset);

  wReadValue = *pChannelData;

  *piValue = this->GetInputValue(wReadValue);

  return OK;
}
