/*  $RCSfile: mpv955.cc,v $ - library for accessing MPV955 card */

/*
$Header: /home/chara/cvsroot/ople/sys/hw/drivers/mpv955.cc,v 1.2 1998/08/17 21:01:07 phil Exp $
*/

/*
modification history
--------------------
01a,23Apr97,pci       written 
*/

/*
DESCRIPTION
This module contains subroutines to interface with the digital to
analog mpv955 card. The functions defined within, enable the user to
store the waveforms in MPV955 memory and start/stop the output of the
waveforms.  The generated waveforms of the MPV955 are multiples of 1 Hz.
The strokeFullInit routine should be called before using any of the
other routines.
*/


#include "isit.h"
#include "hwTypesP.h"
#include "hardware.h"
#include "board.h"
#include "outputChannel.h"
#include "channelBank.h"
#include "mpv955.h"

/////////// From Palomar ///////////////////////////////////////////////////////
//#define SL_SAWTOOTH
//
//#define NORMALIZER 10000 /* num of values which can represent predistorted
//                        waveform (values range from 0 - 9999) */
//#define BIAS (NORMALIZER / 2)  /** triangle wave values must always vary
//                                   from +/- (NORMALIZER / 2) **/
//
//#define STATIC static
//#define NO_PREDISTORTION -1
//#define NANOSECONDS_PER_SECOND 1000000000
//#define MAX_AMPLITUDE 4095
//#define ZERO_VOLTAGE 4095
//
//
//STATIC UWORD  internal_trigger   =  0x0000;
//STATIC UWORD  event_trigger       =   0x0002;
//STATIC UWORD  timeout_disable     = 0x0008;
//STATIC UWORD  continuous_mode_select = 0x0000;
//STATIC int tmp;
//STATIC char *localAdrs; /* convert bus to local addresses */
//STATIC int num_samples_per_channel;
//STATIC int num_samples_in_memory;       
//
//static double predistortion(double sample_index, double amplitude,
//       int type, double num_samples_in_memory);
//
//#define CONVERSION_FREQUENCY  500000 /* conversion rate of external timer in Hz */
//#define MAX_NUM_SAMPLES 10000        /* number which is a multiple of
//                                        CONVERSION_FREQUENCY *and*
//                                        NUM_OF_CHANNELS and less than
//            size of mpv954 data memory */
//#define NUM_CHANNELS 8               /* number of D/A channels to be driven by
//                                        mpv954 */
//#define DATA_MEMORY_SIZE 16384       /* size of data memory */
////////////////////////////////////////////////////////////////////////////////

/*******************************************************************************
*
* class tMpv955Board - BB->SB mapping
* 
*  BB Representation           SB Representation
*  --------------------        -----------------
*  tsTriggerSelectR            *vuwCtrlStatusReg (bit 0)
*  ttTriggerTypeR              *vuwCtrlStatusReg (bit 1)
*  smScanModeR                 *vuwCtrlStatusReg (bit 2)
*  bTimeoutEnableR             *vuwCtrlStatusReg (bit 4)
*  uwNumChannelsR - 1    *vuwCtrlStatusReg (bit 6-4, 6=MSB)
*  bHaltR                      *vuwCtrlStatusReg (bit 8) (read-only)
*  bTimeoutR                   *vuwCtrlStatusReg (bit 9) (read-only)
*  bCycleFinishedR             *vuwCtrlStatusReg (bit 10) (read-only)
*  bOversampling               *vuwCtrlStatusReg (bit 11) (read-only)
*
*/


/*******************************************************************************
*
* tMpv955Board::tMpv955Board(...) - constructor.  Sets up default address/name
*                              given the board number.
*
[[U tMpv955Board::tMpv955Board()]]
*
*****/
tMpv955Board::tMpv955Board(int iBoardNum) : ChannelBank(this)
{
  TraceMark(S_mpv955_ENTERING_CONSTRUCTOR);

  switch(iBoardNum) {
   case 0:
    this->vmeAddress = tVmeAddress(VME_AM_STD_SUP_DATA, (uint8_t *) 0xf00000);
    break;
   default:
    this->vmeAddress = tVmeAddress();
    break;
  }
  this->boardName  = MPV955_BOARD_NAME + iBoardNum;

  TraceMark(S_mpv955_DONE_CONSTRUCTING);
}

/*******************************************************************************
*
* tMpv955Board::SoftInit - Parameterized version
*
* Description: Parameterized version.  Initializes state variables,
*              pointers to the MPV955's registers.  For default values,
*              call tMpv955Board::SoftInit(void) 
*
[[D tMpv955Board::SoftInitParamsSB(in: iResetValueToUse;
                                inout: this;)]
[ this is invalid  => UNDEFINED
| true =>
  [[U tBoard::InitializeParams()]]
  : [[U tMpv955WaveformChannelBank::Initialize(
                pBoardToUse=this,
                vwDataToUse=(volatile uint16_t *)
                     (pvmeAddressToUse->GetLocalAddress() + CiDataOffset),
                vuwStartOutputToUse=(volatile uint16_t *)
                     (pvmeAddressToUse->GetLocalAddress() + CiStartOutputOffset),
                vuwStopOutputToUse=(volatile uint16_t)
                     (pvmeAddressToUse->GetLocalAddress() + CiStopOutputOffset),
                vuwRateTimerCtrlRegToUse=(volatile uint16_t)
                     (pvmeAddressToUse->GetLocalAddress() + CiRateTimerCtrlOffset2),
                iResetValueToUse=iResetValueToUse,
                this=ChannelBank)]]
  : vuwCtrlStatusReg = (volatile uint16_t *) this->vu8LocalAddress + 
                                                   CiCtrlStatusOffset2;
  : vuwInterruptCtrlReg = (volatile uint16_t *) this->vu8LocalAddress +
                                                   CiInterruptCtrlOffset2;
  : vuwTimeoutCtrlReg = (volatile uint16_t *) this->vu8LocalAddress + 
                                                   CiTimeoutCtrlOffset2;
  : vuwDisableOutputReg = (volatile uint16_t *) this->vu8LocalAddress + 
                                                   CiDisableOutputOffset2;
  : return := OK
]]
******/
ERROR_CODE tMpv955Board::SoftInit(int         iResetValueToUse,
                                  int         iSamplePeriodToUse,
                                  CONV_CODE   ccOutputCodeToUse,
                                  int         iVoltageRangeToUse,
                                  uint16_t    uwNumChannelsToUse)
{
  ERROR_CODE ec;
  char sTraceBuf[512];

  TraceMark(S_mpv955_ENTERING_SOFTINIT);

  ec = tBoard::Initialize();
  if (ec != OK) {
    return Traceback(ec, "tMpv955Board::SoftInit", iResetValueToUse, iSamplePeriodToUse,
                     ccOutputCodeToUse, iVoltageRangeToUse, uwNumChannelsToUse);
  }

  this->iNumActiveChannels = MAX(this->iNumActiveChannels, (int)uwNumChannelsToUse);

  ec = ChannelBank.Initialize(
            (volatile int16_t *)  (this->vu8LocalAddress + CiDataOffset),
            (volatile uint16_t *) (this->vu8LocalAddress + CiStartOutputOffset),
            (volatile uint16_t *) (this->vu8LocalAddress + CiStopOutputOffset),
            (volatile uint16_t *) (this->vu8LocalAddress + CiStartAddressOffset2),
            (volatile uint16_t *) (this->vu8LocalAddress + CiStopAddressOffset2),
            (volatile uint16_t *) (this->vu8LocalAddress + CiRateTimerCtrlOffset2),
            iResetValueToUse,
            ccOutputCodeToUse,
            iVoltageRangeToUse,
            (uint16_t) this->iNumActiveChannels);
  if (ec != OK) {
    return Traceback(ec, "tMpv955Board::SoftInit", iResetValueToUse, iSamplePeriodToUse,
                     ccOutputCodeToUse, iVoltageRangeToUse, uwNumChannelsToUse);
  }

  this->vuwCtrlStatusReg = (volatile uint16_t *)
                          (this->vu8LocalAddress + CiCtrlStatusOffset2);
  this->vuwInterruptCtrlReg = (volatile uint16_t *) 
                          (this->vu8LocalAddress + CiInterruptCtrlOffset2);
  this->vuwTimeoutCtrlReg = (volatile uint16_t *) 
                          (this->vu8LocalAddress + CiTimeoutCtrlOffset2);
  this->vuwDisableOutputReg = (volatile uint16_t *) 
                          (this->vu8LocalAddress + CiDisableOutputOffset2); 

  ec = this->ChannelBank.SetSamplePeriod(iSamplePeriodToUse);
  if(ec != OK) {
    return Traceback(ec, "tMpv955Board::SoftInit", iResetValueToUse, iSamplePeriodToUse,
                     ccOutputCodeToUse, iVoltageRangeToUse, uwNumChannelsToUse);
  }

  this->DumpRegs();

  this->SetInit(TRUE);
  TraceMark(S_mpv955_INITIALIZED_SOFTLY);

  return OK; 
}

/*******************************************************************************
*
* ERROR_CODE tMpv955Board::SoftInit(void)
* Description:  Default version of SoftInit. Sets board address, initializes
*               the pointers to board variables, and sets class variables using
*               the default values.
[[D tMpv955Board::SoftInitVoid(in: this;)]
[ this is invalid => UNDEFINED
| true =>
  [[U tBoard::InitializeVoid()]]
  : [[U tMpv955WaveformChannelBank::Initialize(
                pBoardToUse=this,
                vwDataToUse=(volatile uint16_t *)
                     (this->vmeAddress.GetLocalAddress() + CiDataOffset),
                vuwStartOutputToUse=(volatile uint16_t *)
                     (this->vmeAddress.GetLocalAddress() + CiStartOutputOffset),
                vuwStopOutputToUse=(volatile uint16_t *)
                     (this->vmeAddress.GetLocalAddress() + CiStopOutputOffset),
                vuwRateTimerCtrlRegToUse=(volatile uint16_t)
                     (this->vmeAddress->GetLocalAddress() + CiRateTimerCtrlOffset2),
                iResetValueToUse=iResetValueToUse,
                this=ChannelBank)]]
  : vuwCtrlStatusReg = (volatile uint16_t *) this->vu8LocalAddress + 
                                                   CiCtrlStatusOffset2;
  : vuwInterruptCtrlReg = (volatile uint16_t *) this->vu8LocalAddress +
                                                   CiInterruptCtrlOffset2;
  : vuwTimeoutCtrlReg = (volatile uint16_t *) this->vu8LocalAddress + 
                                                   CiTimeoutCtrlOffset2;
  : vuwDisableOutputReg = (volatile uint16_t *) this->vu8LocalAddress + 
                                                   CiDisableOutputOffset2;
  : return := OK
]]
*
*****/
ERROR_CODE tMpv955Board::SoftInit(void)
{
  return (this->SoftInit(this->CiDefaultResetValue,
                         this->CiDefaultSamplePeriod,
                         this->CccDefaultOutputCode,
                         this->CiDefaultVoltageRange,
                         this->CuwDefaultNumChannels));
}


/*******************************************************************************
*
* ERROR_CODE tMpv955Board::HardInit(...)
* Description: Parameterized version of HardInit. Calls SoftInit with values passed
*              in and sets board registers with values passed in.
[[D tMpv955Board::HardInitParamsSB(in: tsTriggerSelectToUse,
                                     ttTriggerTypeToUse,
                                     smScanModeToUse,
                                     bTimeoutEnableToUse,
                                     uwNumChannelsToUse,
                                     iResetValueToUse;
                              inout: this;)]
[ true =>
  [[U tMpv955Board::SoftInitParams()]]
  : set control register values per parameters
]]
*
******/
ERROR_CODE tMpv955Board::HardInit(TRIGGER_SELECT  tsTriggerSelectToUse,
                             TRIGGER_TYPE    ttTriggerTypeToUse,
                             SCAN_MODE       smScanModeToUse,
                             BOOL            bTimeoutEnableToUse,
                             int             iResetValueToUse,
                             int             iSamplePeriodToUse,
                             CONV_CODE       ccOutputCodeToUse,
                             int             iVoltageRangeToUse,
                             uint16_t        uwNumChannelsToUse)
{
  ERROR_CODE ec, ec1, ec2, ec3, ec4;
  char sTraceBuf[512];

  TraceMark(S_mpv955_ENTERING_HARDINIT);

  ec = this->SoftInit(iResetValueToUse, iSamplePeriodToUse, ccOutputCodeToUse,
                      iVoltageRangeToUse, uwNumChannelsToUse);
  if(ec != OK) {
    return Traceback(ec, "tMpv955Board::HardInit",
                   tsTriggerSelectToUse, ttTriggerTypeToUse, smScanModeToUse,
                   bTimeoutEnableToUse, ccOutputCodeToUse, iVoltageRangeToUse,
                   iSamplePeriodToUse, uwNumChannelsToUse, iResetValueToUse);
  }

  ec  = this->SetTriggerSelect(tsTriggerSelectToUse);
  ec1 = this->SetTriggerType(ttTriggerTypeToUse);
  ec2 = this->SetScanMode(smScanModeToUse);
  ec3 = this->SetTimeoutEnable(bTimeoutEnableToUse);
  ec4 = this->SetNumChannels();
  if(ec!=OK || ec1!=OK || ec2!=OK ||  ec3!=OK || ec4!=OK) { 
    return Traceback(ec, "tMpv955Board::HardInit",
                 tsTriggerSelectToUse, ttTriggerTypeToUse, smScanModeToUse,
                 bTimeoutEnableToUse, ccOutputCodeToUse, iVoltageRangeToUse,
                 iSamplePeriodToUse, uwNumChannelsToUse, iResetValueToUse);
  }

  this->ChannelBank.ZeroMemory();
  this->ChannelBank.StartOutput();

  this->DumpRegs();

  TraceMark(S_mpv955_INITIALIZED_HARDLY);

  return OK;
}

/*******************************************************************************
*
* ERROR_CODE tMpv955Board::HardInit(void)
* Description: Default version of HardInit.  Calls default version of SoftInit and
*              sets default values in hardware.
[[D tMpv955Board::HardInitVoidSB(inout: this;)]
[ true =>
  [[U tMpv955Board::SoftInitVoid()]]
  : set control register values per parameters
]]
*
******/
ERROR_CODE tMpv955Board::HardInit(void)
{
  return (this->HardInit(this->CtsDefaultTriggerSelect,
                         this->CttDefaultTriggerType,
                         this->CsmDefaultScanMode,
                         this->CbDefaultTimeoutEnable,
                         this->CiDefaultResetValue,
                         this->CiDefaultSamplePeriod,
                         this->CccDefaultOutputCode,
                         this->CiDefaultVoltageRange,
                         this->CuwDefaultNumChannels));
}


/*******************************************************************************
*
* void tMpv955Board::DumpRegs(void)
* Description: Register Dump
*
******/
void tMpv955Board::DumpRegs(void)
{
  char sTraceBuf[1024];

  sprintf(sTraceBuf, "MPV955 Board Data \n \
                      vuwCtrlStatusReg    = 0x%x, value = 0x%x \n \
                      vuwInterruptCtrlReg = 0x%x, value = 0x%x \n \
                      vuwTimeoutCtrlReg   = 0x%x, value = 0x%x \n \
                      vuwDisableOutputReg = 0x%x, value = 0x%x \n",
           (int) this->vuwCtrlStatusReg, *this->vuwCtrlStatusReg,
           (int) this->vuwInterruptCtrlReg, *this->vuwInterruptCtrlReg,
           (int) this->vuwTimeoutCtrlReg, *this->vuwTimeoutCtrlReg,
           (int) this->vuwDisableOutputReg, *this->vuwDisableOutputReg);
  TraceMsg(tmt_INFO, OK, sTraceBuf);
}


/******************************************************************************
*
* tMpv955Board::CreateChannel - gets channel specified by iNumChannel (global
*                             board numbering)
*
[[U tMpv955Board::CreateChannel()]]
*/
tChannel *tMpv955Board::CreateChannel(int iChannelNum)
{
  if((this->IsInit() && iChannelNum >= this->iNumActiveChannels) ||
     iChannelNum < 0 || iChannelNum >= this->ChannelBank.CiMaxNumChannels) {
    TraceMsg(tmt_ERROR, S_mpv955_RETURNING_NULL_CHANNEL_POINTER,
             "ERROR: Returning NULL channel pointer for MPV955 - channel #%d\n",
             iChannelNum); 
    return NULL;
  }
  this->iNumActiveChannels = MAX(this->iNumActiveChannels, (iChannelNum+1));
  /****
  [[U tMpv955AdChannelBank::CreateChannel()]]
  ****/
  return (this->ChannelBank.CreateChannel(iChannelNum));
}


/*******************************************************************************
*
* void tMpv955Board::DisableOutput(BOOL bOn)
* Description: Immediately disables all output channels if TRUE is passed in.
*              Immediately enables all output channels if FALSE is passed in.
*      NOTE: vuwDisableOutputReg *must* point to the Area 1 version of the register
*            in order for StopOutput to work properly.
[[U tMpv955Board::OutputEnable(bOn)]]
*
******/
void tMpv955Board::DisableOutput(BOOL bOn)
{
  switch(bOn) {
   case(TRUE):
     *vuwDisableOutputReg = 1;
     break;
   case(FALSE):
     *vuwDisableOutputReg = 0;
     break;
  }
}
     
/*******************************************************************************
* SS Sorry, no spec 
*   If setting to external, this function also forces the board into "normal"
*   trigger mode
*/

ERROR_CODE tMpv955Board::SetTriggerSelect(TRIGGER_SELECT tsTriggerSelectToUse) {
  switch(tsTriggerSelectToUse) {
   case(TS_INTERNAL):
     *(this->vuwCtrlStatusReg) &= 0xFFFE;
     TraceMsg(tmt_INFO, OK, "MPV955 Trigger Select set to TS_INTERNAL");
     break;
   case(TS_EXTERNAL):
     /* make sure Trigger Type is TT_NORMAL */
     this->SetTriggerType(TT_NORMAL);
     *(this->vuwCtrlStatusReg) |= 0x0001;
     TraceMsg(tmt_INFO, OK, "MPV955 Trigger Select set to TS_EXTERNAL");
     break;
   default:
     return Traceback(S_mpv955_INVALID_TRIGGER_SELECT_VALUE,
                      "tMpv955Board::SetTriggerSelect", tsTriggerSelectToUse);
  }
  return OK;
}


/*******************************************************************************
* SS Sorry, no spec 
*   If setting to event, this function also forces the board into "internal"
*   trigger select mode 
*/

ERROR_CODE tMpv955Board::SetTriggerType(TRIGGER_TYPE ttTriggerTypeToUse) {
  switch(ttTriggerTypeToUse) {
   case(TT_NORMAL):
     *(this->vuwCtrlStatusReg) &= 0xFFFD;
     TraceMsg(tmt_INFO, OK, "MPV955 Trigger Type set to TT_NORMAL");
     break;
   case(TT_EVENT):
     /* make sure Trigger Select is TS_INTERNAL */
     this->SetTriggerSelect(TS_INTERNAL);
     *(this->vuwCtrlStatusReg) |= 0x0002;
     TraceMsg(tmt_INFO, OK, "MPV955 Trigger Type set to TT_EVENT");
     break;
   default:
     return Traceback(S_mpv955_INVALID_TRIGGER_TYPE_VALUE,
                      "tMpv955Board::SetTriggerType", ttTriggerTypeToUse);
  }
  return OK;
}

/*******************************************************************************
* SS Sorry, no spec */
ERROR_CODE tMpv955Board::SetScanMode(SCAN_MODE smScanModeToUse) {
  switch(smScanModeToUse) {
   case(SM_CONTINUOUS):
     *(this->vuwCtrlStatusReg) &= 0xFFFB;
     TraceMsg(tmt_INFO, OK, "MPV955 Scan Mode set to SM_CONTINUOUS");
     break;
   case(SM_ONE_SHOT):
     *(this->vuwCtrlStatusReg) |= 0x0004;
     TraceMsg(tmt_INFO, OK, "MPV955 Scan Mode set to SM_ONE_SHOT");
     break;
   default:
     return Traceback(S_mpv955_INVALID_SCAN_MODE_VALUE,
                      "tMpv955Board::SetTriggerType");
  }
  return OK;
}

/*******************************************************************************
* SS Sorry, no spec */
ERROR_CODE tMpv955Board::SetTimeoutEnable(BOOL bTimeoutToUse) {
  switch(bTimeoutToUse) {
   case(TRUE):
     *(this->vuwCtrlStatusReg) &= 0xFFF7;
     break;
   case(FALSE):
     *(this->vuwCtrlStatusReg) |= 0x0008;
     break;
   default:
     return Traceback(S_mpv955_INVALID_TIMEOUT_VALUE,
                      "tMpv955Board::SetTimeoutEnable");
  }
  return OK;
}


/*******************************************************************************
* 
* tMpv955Board::SetNumChannels(void)
*
* Description: Sets the number of channels to output.  This function is
*              protected because it is only called at initialize time.
[[D tMpv955Board::SetNumChannels(in: this;)]
[ this is invalid => UNDEFINED
| this->iNumActiveChannels is not valid for board=> 
  return appropriate error
| true=>
  set control/status register to this->iNumActiveChannels.
]]
*
******/
ERROR_CODE tMpv955Board::SetNumChannels(void)
{
  char sTraceBuf[512];

  if(this->iNumActiveChannels < 1 ||
     this->iNumActiveChannels > this->ChannelBank.CiMaxNumChannels) {
    return Traceback(S_mpv955_INVALID_NUMBER_OF_CHANNELS,
                     "tMpv955Board::SetNumChannels");
  }
  *(this->vuwCtrlStatusReg) = (*(this->vuwCtrlStatusReg) & 0xff8f) +
                            ((this->iNumActiveChannels - 1) << 4);
  sprintf(sTraceBuf, "MPV955 Set to %d channels, register value = 0x%x",
          this->iNumActiveChannels, *(this->vuwCtrlStatusReg));
  TraceMsg(tmt_INFO, OK, sTraceBuf);
  return OK;
}



/*******************************************************************************
*
* class tMpv955WaveformChannelBank - BB->SB mapping
* 
*  BB Representation           SB Representation
*  --------------------        -----------------
*
*/


/*******************************************************************************
*
* tMpv955WaveformChannelBank::tMpv955WaveformChannelBank
*
[[U tMpv955WaveformChannelBank::tMpv955WaveformChannelBank()]]
*
******/
tMpv955WaveformChannelBank::tMpv955WaveformChannelBank(tMpv955Board *pBoardToUse)
{
  this->pBoard = pBoardToUse;
  this->bOutputOn = FALSE;
  this->u8CachedRateTimerValue = 0;
}

/*******************************************************************************
*
* ERROR_CODE tMpv955WaveformChannelBank::Initialize(...)
* Description: Initialization method for a bank of MPV955 channels.  Sets state
*              variables, and sets bInit to TRUE.
[[U tMpv955WaveformChannelBank::Initialize()]]
*
******/
ERROR_CODE tMpv955WaveformChannelBank::Initialize(volatile int16_t *vwDataToUse,
                         volatile uint16_t *vuwStartOutputToUse,
                         volatile uint16_t *vuwStopOutputToUse,
                         volatile uint16_t *vuwStartAddressRegToUse,
                         volatile uint16_t *vuwStopAddressRegToUse,
                         volatile uint16_t *vuwRateTimerCtrlRegToUse,
                         int        iResetValueToUse,
                         CONV_CODE  ccOutputCodeToUse,
                         int        iVoltageRangeToUse,
                         uint16_t   uwNumChannelsToUse)
{
  ERROR_CODE ec;
  int i;

  TraceMark(S_mpv955_ENTERING_CHANNELBANK_INITIALIZE);

  this->vwStartOfData   = vwDataToUse;
  this->vuwStartOutput   = vuwStartOutputToUse;
  this->vuwStopOutput    = vuwStopOutputToUse;
  this->vuwStartAddressReg = vuwStartAddressRegToUse;
  this->vuwStopAddressReg  = vuwStopAddressRegToUse;
  this->vuwRateTimerCtrlReg = vuwRateTimerCtrlRegToUse;
  this->iTotalMemoryLoc  = this->CiTotalMemoryLoc;
  this->iWaveformSize    = this->CiTotalMemoryLoc / 
                           this->GetNumChannels();

  this->DumpRegs();
                      
  for(i=0; i < this->GetNumChannels(); i++) {
    ec=ChannelArray[i].Initialize(this, NULL, iResetValueToUse, vwDataToUse + i,
                               ccOutputCodeToUse, iVoltageRangeToUse);
    if(ec != OK) {
      return Traceback(ec,"tMpv955WaveformChannelBank::Initialize",
                     *vwDataToUse, *vuwStartOutputToUse,
                     *vuwStopOutputToUse, *vuwStartAddressRegToUse,
                     *vuwStopAddressRegToUse, *vuwRateTimerCtrlRegToUse,
                     iResetValueToUse, ccOutputCodeToUse, iVoltageRangeToUse,
                     uwNumChannelsToUse);
    }
  }

  this->SetInit(TRUE);

  TraceMark(S_mpv955_CHANNELBANK_INITIALIZED);
  return OK;
}

/*******************************************************************************
*
* void tMpv955WaveformChannelBank::DumpRegs(void)
*    AND
* void tMpv955WaveformChannelBank::DumpBoardRegs(void)
* Description: Register Dump
*
******/
void tMpv955WaveformChannelBank::DumpRegs(void)
{
  char sTraceBuf[1024];

  sprintf(sTraceBuf, "MPV955 Channel Bank Data \n \
                      vwStartOfData       = 0x%x, value = 0x%x \n \
                      vuwStartOutput      = 0x%x, value = 0x%x \n \
                      vuwStopOutput       = 0x%x, value = 0x%x \n \
                      vuwStartAddressReg  = 0x%x, value = 0x%x \n \
                      vuwStopAddressReg   = 0x%x, value = 0x%x \n \
                      vuwRateTimerCtrlReg = 0x%x, value = 0x%x \n \
                      iTotalMemoryLoc     = %d \n \
                      iWaveformSize       = %d \n",
          (int) this->vwStartOfData, *this->vwStartOfData,
          (int) this->vuwStartOutput, *this->vuwStartOutput,
          (int) this->vuwStopOutput, *this->vuwStopOutput,
          (int) this->vuwStartAddressReg, *this->vuwStartAddressReg, 
          (int) this->vuwStopAddressReg, *this->vuwStopAddressReg,
          (int) this->vuwRateTimerCtrlReg, *this->vuwRateTimerCtrlReg,
          this->iTotalMemoryLoc, this->iWaveformSize);
  TraceMsg(tmt_INFO, OK, sTraceBuf);
}

void tMpv955WaveformChannelBank::DumpBoardRegs(void)
{
  this->pBoard->DumpRegs();
}


/******************************************************************************
*
* tMpv955WaveformChannelBank::CreateChannel - gets channel specified by iNumChannel
*                             (global board numbering)
*
[[U tMpv955WaveformChannelBank::CreateChannel()]]
*/
tChannel *tMpv955WaveformChannelBank::CreateChannel(int iChannelNum)
{
  if(iChannelNum > this->GetNumChannels()) {
    TraceMsg(tmt_ERROR, S_mpv955_RETURNING_NULL_CHANNEL_POINTER,
             "ERROR: Returning NULL channel pointer for MPV955 - channel #%d\n",
           iChannelNum); 
    return NULL;
  }  
  return (&(ChannelArray[iChannelNum]));
}


/*******************************************************************************
*
* ERROR_CODE tMpv955WaveformChannelBank::ZeroMemory(...)
* Description: Zeros all data memory locations on the board.
[[D tMpv955WaveformChannelBank::ZeroMemory(in: this;)]
[ this is invalid => UNDEFINED
| this->bInit == FALSE =>
  return S_mpv955_BOARD_NOT_INITIALIZED
| true =>
  this->all data memory locations := 0
]]
*
******/
ERROR_CODE tMpv955WaveformChannelBank::ZeroMemory(void)
{
  int16_t wOutputValue;
  int     i;
  char sTraceBuf[512];

  if(this->IsInit() == FALSE) {
    return Traceback(S_mpv955_CHANNEL_BANK_NOT_INITIALIZED,
                     "tMpv955WaveformChannelBank::ZeroMemory");
  }

  wOutputValue = (int16_t) this->ChannelArray[0].GetOutputValue(0);

  sprintf(sTraceBuf, "Zero value = 0x%x",wOutputValue);
  TraceMsg(tmt_INFO,OK, sTraceBuf);

  for(i=0; i<this->iTotalMemoryLoc; i++) {
    this->vwStartOfData[i] = wOutputValue;
  }

  return OK;
}

/*******************************************************************************
*
* ERROR_CODE tMpv955WaveformChannelBank::SetStartStopAddresses(void)
* Description: Sets StartAddressReg and StopAddressReg so they correspond to
*              the beginning and end of the actual data.  The start and stop
*              registers are especially useful because if the number of
*              channels is not evenly divisible by the total number of memory
*              locations, there would be some garbage at the end of the 
*              waveform.
*
[[U tMpv955WaveformChannelBank::SetStartStopAddresses(void)]]
*
******/
ERROR_CODE tMpv955WaveformChannelBank::SetStartStopAddresses(void)
{
  if(!this->IsInit()) {
    return Traceback(S_mpv955_CHANNEL_BANK_NOT_INITIALIZED,
                     "tMpv955WaveformChannelBank::SetStartStopAddresses");
  }

  *this->vuwStartAddressReg = 0;
  *this->vuwStopAddressReg  = 
               (this->iWaveformSize * this->GetNumChannels()) - 1;

  TraceMark(S_mpv955_START_STOP_ADDRESSES_SET);
  
  return OK;
}



/*******************************************************************************
*
* ERROR_CODE tMpv955WaveformChannelBank::StartOutput(void)
* Description: Start "pulse" for the board to begin outputting voltages to all
*              active channels.
[[D tMpv955WaveformChannelBank::StartOutputSB(inout: this;)]
[ this is invalid => UNDEFINED
| this->bInit == FALSE =>
  return S_mpv955_CHANNEL_BANK_NOT_INITIALIZED
| this->bOutputOn == TRUE =>
  return OK
| true =>
  access the start register
  : this->bOutputOn := TRUE
]]
*
******/
ERROR_CODE tMpv955WaveformChannelBank::StartOutput(void)
{
  int16_t wDummy;

  if(this->IsInit() == FALSE) {
    return Traceback(S_mpv955_CHANNEL_BANK_NOT_INITIALIZED,
                     "tMpv955WaveformChannelBank::StartOutput");
  }

  if(this->bOutputOn == TRUE) {
    return OK;
  }
  
  this->pBoard->DisableOutput(FALSE);

  /* accessing mpv955 Start Register starts output */
  wDummy = *vuwStartOutput;
  this->bOutputOn = TRUE;

  return OK;
}


/*******************************************************************************
*
* ERROR_CODE tMpv955WaveformChannelBank::StopOutput(void)
* Description: Stop "pulse" for the board to stop outputting voltages to its
*              channels.
[[D tMpv955WaveformChannelBank::StopOutput(inout: this;)]
[ this is invalid || this->vuwStopOutput is not R/W reg => UNDEFINED
| this->vuwStopOutput is not an Area1 register =>
  output will continue with this->bOutputOn set to FALSE
| this->bInit == FALSE =>
  return S_mpv955_CHANNEL_BANK_NOT_INITIALIZED
| this->bOutputOn == FALSE =>
  return := OK
| true =>
  write to this->vuwStopOutput
  : this->bOutputOn := FALSE
  : return := OK
]] 
*
*/
ERROR_CODE tMpv955WaveformChannelBank::StopOutput(void)
{
  if(this->IsInit() == FALSE) {
    return Traceback(S_mpv955_CHANNEL_BANK_NOT_INITIALIZED,
                     "tMpv955WaveformChannelBank::StartOutput");
  }

  if(this->bOutputOn == FALSE) {
    return OK;
  }

  /* writing to an "Area 1 register" Stops output cycling */
  /* This code assumes the Stop Output register is the Area 1 version
     of the CtrlStatusReg (which has no info in MS nibble.  I didn't
     want to &= with 0xffff because it might be optimized out */
  *(this->vuwStopOutput) &= 0x0fff;
  this->bOutputOn = FALSE;

  return OK;
}

/*******************************************************************************
*
* int tMpv955WaveformChannelBank::GetNumChannels(void)
* Description: Sets the number of active channels on the board.
*
[[U tMpv955WaveformChannelBank::GetNumChannels()]]
*
******/
int tMpv955WaveformChannelBank::GetNumChannels(void)
{
  return pBoard->GetNumChannels();
}

/*******************************************************************************
*
* ERROR_CODE tMpv955WaveformChannelBank::SetWaveformSize(int iSize)
*
* Description: Sets the size of the waveform (in number of samples)
*
[[U tMpv955WaveformChannelBank::SetWaveformSize(iSize)]]
*
******/
ERROR_CODE tMpv955WaveformChannelBank::SetWaveformSize(int iSize)
{
  if(this->IsInit() == FALSE) {
    return Traceback(S_mpv955_CHANNEL_BANK_NOT_INITIALIZED,
                     "tMpv955WaveformChannelBank::SetWaveformSize", iSize);
  }

  if((iSize <= 0) ||
     (iSize > (this->iTotalMemoryLoc/this->GetNumChannels()))) {
    return Traceback(S_mpv955_INVALID_WAVEFORM_SIZE,
                     "tMpv955WaveformChannelBank::SetWaveformSize", iSize);
  }

  this->iWaveformSize = iSize;

  this->SetStartStopAddresses();

  return OK;
}


/*******************************************************************************
*
* ERROR_CODE tMpv955WaveformChannelBank::SetSamplePeriod(...)
*
* Description: Sets the sample period (in nanoseconds)
*
[[U tMpv955WaveformChannelBank::SetSamplePeriod()]]
*
******/
ERROR_CODE tMpv955WaveformChannelBank::SetSamplePeriod(int iNewPeriod)
{
  TraceMark(S_mpv955_SETTING_SAMPLE_PERIOD);

  if(this->IsInit() == FALSE) {
    return Traceback(S_mpv955_CHANNEL_BANK_NOT_INITIALIZED,
                     "tMpv955WaveformChannelBank::SetSamplePeriod", iNewPeriod);
  }

  if(iNewPeriod <= 0) {
    return Traceback(S_mpv955_ZERO_OR_NEGATIVE_PERIOD,
                   "tMpv955WaveformChannelBank::SetSamplePeriod", iNewPeriod);
  }

  if(iNewPeriod < this->CiSampleClockPeriod) {
    TraceMsg(tmt_WARN, S_mpv955_PERIOD_ADJUSTED,
           "SetSamplePeriod: Period too small; adjusted appropriately");
    iNewPeriod = this->CiSampleClockPeriod;
  }

  if(iNewPeriod > (this->CiSampleClockPeriod * this->CiRateTimerMaxValue)) {
    TraceMsg(tmt_WARN, S_mpv955_PERIOD_ADJUSTED,
           "SetSamplePeriod: Period too large; adjusted appropriately");
    iNewPeriod = this->CiSampleClockPeriod * this->CiRateTimerMaxValue;
  }

  if((iNewPeriod % this->CiSampleClockPeriod) != 0) {
    TraceMsg(tmt_WARN, S_mpv955_PERIOD_ADJUSTED,
             "SetSamplePeriod: Period not even multiple; adjusted appropriately");
  }

  this->u8CachedRateTimerValue = this->CiRateTimerMaxValue -
                                   (iNewPeriod / this->CiSampleClockPeriod);
  *(this->vuwRateTimerCtrlReg) = this->u8CachedRateTimerValue;
  
  TraceMark(S_mpv955_SAMPLE_PERIOD_SET);
  return OK;
}



/*******************************************************************************
*
* class tMpv955WaveformChannel
*
******/

/*******************************************************************************
*
* tMpv955WaveformChannel::Initialize(...)
*
[[U tMpv955WaveformChannel::Initialize()]]
*
******/
ERROR_CODE tMpv955WaveformChannel::Initialize(
                                tMpv955WaveformChannelBank  *pChannelBankToUse,
                                tInputChannel               *pReadbackChannelToUse,
                                int                         iResetValueToUse,
                                volatile int16_t           *vwDataToUse,
                                CONV_CODE                   ccOutputCodeToUse,
                                int                         iVoltageRangeToUse)
{
  TraceMark(S_mpv955_ENTERING_CHANNEL_INITIALIZE);

  this->pChannelBank     = pChannelBankToUse;
  this->pReadbackChannel = pReadbackChannelToUse;
  this->iResetValue      = iResetValueToUse;
  this->ccConvCode       = ccOutputCodeToUse;
  this->dChannelRange    = iVoltageRangeToUse;
  this->iBitWidth        = this->CiBitWidth;
  this->vwData           = vwDataToUse;
  this->SetInit(TRUE);

  TraceMark(S_mpv955_CHANNEL_INITIALIZED);

  return OK;
}


/*******************************************************************************
*
* int tMpv955WaveformChannel::GetMemSize(void)
* Description:  
*
[[U tMpv955WaveformChannel::GetMemSize()]]
* 
******/
int tMpv955WaveformChannel::GetMemSize(void)
{
  return this->pChannelBank->GetWaveformSize();
}

/*******************************************************************************
*
* ERROR_CODE tMpv955WaveformChannel::Output(int iValue)
* Description:  Simple Output routine for the MPV955 -- primarily written for
*               conformity to the tOutputChannel pure virtual methods.  This
*               method simply loads all values of the channel with iValue.
*               Remaining memory values associated with this channel are set
*               to zero in case the waveform size is later set to a larger value.
*
[[U tMpv955WaveformChannel::Output(iValue)]]
* 
******/
ERROR_CODE tMpv955WaveformChannel::Output(int iValue)
{
  int16_t wOutputValue, wZeroValue;
  int     i, iEndWaveformLoc, iTotalMemory, iNumChannels;

  // this->pChannelBank->DumpRegs();
  // this->pChannelBank->DumpBoardRegs();

  wOutputValue = (int16_t) this->GetOutputValue(iValue);
  wZeroValue   = (int16_t) this->GetOutputValue(0);

  iTotalMemory     = this->pChannelBank->GetTotalMemory();
  iNumChannels     = this->pChannelBank->GetNumChannels();
  iEndWaveformLoc  = this->pChannelBank->GetWaveformSize() * iNumChannels;

  for(i=0; i<(iTotalMemory-iNumChannels); i+=iNumChannels) {
    if(i<iEndWaveformLoc) {
      *(vwData + i) = wOutputValue;
    }
    else {
      *(vwData + i) = wZeroValue;
    }
  }
  // this->pChannelBank->DumpRegs();
  // this->pChannelBank->DumpBoardRegs();

  return OK;
}
    
/*******************************************************************************
*
* ERROR_CODE tMpv955WaveformChannel::LoadMemory(int *piValues)
* Description:  Routine to load MPV955's memory. Loads all values of the channel
*               with values pointed to by piValues.  Assumes that there are
*               this->GetMemSize values iValue.
*               Remaining memory values associated with this channel are set
*               to zero in case the waveform size is later set to a larger value.
[[U tMpv955WaveformChannel::LoadMemory()]]
* 
******/
ERROR_CODE tMpv955WaveformChannel::LoadMemory(int *piValues)
{
  int16_t wZeroValue;
  int     i, iEndWaveformLoc, iTotalMemory, iNumChannels;

  wZeroValue   = (int16_t) this->GetOutputValue(0);

  iTotalMemory     = this->pChannelBank->GetTotalMemory();
  iNumChannels     = this->pChannelBank->GetNumChannels();
  iEndWaveformLoc  = this->pChannelBank->GetWaveformSize() * iNumChannels;

  for(i=0; i<(iTotalMemory-2); i+=iNumChannels) {
    if(i<iEndWaveformLoc) {
      *(this->vwData + i) =
          (int16_t) this->GetOutputValue(*(piValues+i));
    }
    else {
      *(this->vwData + i) = wZeroValue;
    }
  }

  // this->pChannelBank->DumpRegs();
  // this->pChannelBank->DumpBoardRegs();

  return OK;
}


/*******************************************************************************
*
* ERROR_CODE tMpv955WaveformChannel::LoadMemory(double *pdValues)
* Description:  Routine to load MPV955's memory. Loads all values of the channel
*               with values pointed to by pdValues.  Assumes that there are
*               this->GetMemSize values iValue.
*               Remaining memory values associated with this channel are set
*               to zero in case the waveform size is later set to a larger value.
[[U tMpv955WaveformChannel::LoadMemory()]]
* 
******/
ERROR_CODE tMpv955WaveformChannel::LoadMemory(double *pdValues)
{
  int16_t wZeroValue;
  int     i, iWaveformSize, iChannelSize, iNumChannels;
  char    sTraceBuf[512];

  wZeroValue   = (int16_t) this->GetOutputValue(0);

  iNumChannels     = this->pChannelBank->GetNumChannels();
  iChannelSize     = this->GetMemSize();
  iWaveformSize    = this->pChannelBank->GetWaveformSize();
 
  for(i=0; i<iChannelSize; i++) {

    if(i<iWaveformSize) {
      *(vwData + i*iNumChannels) =
          (int16_t) this->GetOutputValue((int) *(pdValues+i));
    }
    else {
      *(vwData + i*iNumChannels) = wZeroValue;
    }

  }

  // this->pChannelBank->DumpRegs();
  // this->pChannelBank->DumpBoardRegs();

  return OK;
}


/*******************************************************************************
*
* ERROR_CODE tMpv955WaveformChannel::GetChannelBank(void)
*
* Description: Returns channel bank that owns this channel
* 
[[U tMpv955WaveformChannel::GetChannelBank()]]
* 
******/
tWaveformChannelBank *tMpv955WaveformChannel::GetChannelBank(void)
{
  return this->pChannelBank;
}

