/* $RCSfile: pztMotor.cc,v $ - class to access the JPL pzt and motor control board.  */
/*
$Header: /home/chara/cvsroot/ople/sys/hw/drivers/pztMotor.cc,v 1.2 1998/08/17 21:01:08 phil Exp $
*/

/*
modification history
--------------------
03d,pci,18May97     Changed ramp channel to inherit from daChannel
03c,pci,16May97     Added GetChannel
03b,pci,29Apr97     modified for new hardware architecture
03a,pci,22Oct96     modified for ISIT
02c,hs, 16Mar94     adding support for the paddle select line
02b,hs, 22Feb94     added encoder support routines
02a,beh,18Nov93     C++ version created
01e,dlp,04Oct93     added pztPositionSet for pzt_D/A chip variation
01d,mmc,02feb92     added pztSetHigh, pztSetLow
01c,mmc,31aug91     include sysLib.h
01b,dlp,08Nov90     Changed Lint16_t types to int16_t in pointer declarations
01a,dlp,05Oct90     written, debugged
*/

/*
DESCRIPTION
This module contains all the routines necessary to manipulate the pzt/motor
  controller board hardware.  
*/

#ifdef VW_502
#include <vxProto.h>
#endif   /* VW_502 */

/*#include <sysLib.h>*/
/*#include <errnoLib.h>*/
#include "isit.h"
#include "hwTypesP.h"
#include "board.h"
#include "inputChannel.h"
#include "outputChannel.h"
#include "pztMotor.h"
#include "hwError.h"


LIMIT_TYPE const CltLimitTypeArray[PZT_MOTOR_MAX_NUM_LIMITS] = {
    LT_CCW_SOFT, LT_CW_SOFT, LT_CCW_HARD, LT_CW_HARD, LT_HOME,
    LT_CCW_SOFT, LT_CW_SOFT, LT_CCW_HARD, LT_CW_HARD, LT_HOME,
    LT_CCW_SOFT, LT_CW_SOFT, LT_CCW_HARD, LT_CW_HARD, LT_HOME,
    LT_CCW_SOFT, LT_CW_SOFT, LT_CCW_HARD, LT_CW_HARD, LT_HOME,
    LT_CCW_SOFT, LT_CW_SOFT, LT_CCW_HARD, LT_CW_HARD, LT_HOME,
    LT_CCW_SOFT, LT_CW_SOFT, LT_CCW_HARD, LT_CW_HARD, LT_HOME,
};

BOOL const CbDefaultRampArray[PZT_MOTOR_MAX_NUM_PZTS] = {
    TRUE, TRUE, TRUE, TRUE, TRUE, TRUE
};

CONV_CODE const CccDefaultConvCodeArray[PZT_MOTOR_MAX_NUM_PZTS] = {
                                CC_TWOS_COMPLEMENT, CC_STRAIGHT_BINARY,
                                CC_TWOS_COMPLEMENT, CC_TWOS_COMPLEMENT,
                                CC_TWOS_COMPLEMENT, CC_TWOS_COMPLEMENT
};

#define PZT_RATE     0.0019  /* rate constant in fract_of_full_scale/s/lsb */
                                                   /* JPL card: 8MHz/65k^2 */
double const CdDefaultPztRangeArray[PZT_MOTOR_MAX_NUM_PZTS] = {
                                (100.0/65536.0*4.0e6), (100.0/65536.0*4.0e6),
                                (100.0/65536.0*4.0e6), (100.0/65536.0*4.0e6),
                                (100.0/65536.0*4.0e6), (100.0/65536.0*4.0e6)
};       /* Volts/sec at full scale output, assuming 10X amplifier */

double const CdDefaultMotorRangeArray[PZT_MOTOR_MAX_NUM_MOTORS] = {  
                                2.0e6, 2.0e6, 2.0e6, 2.0e6, 2.0e6, 2.0e6
};       /* pulses/sec output at full scale output */


/*******************************************************
*
* tPztMotorBoard::tPztMotorBoard - constructor
*
* This is the constructor for the PZT/Motor class.
*
[[U tPztMotorBoard::tPztMotorBoard()]]
* 
******/
tPztMotorBoard::tPztMotorBoard(int iBoardNum)
{
  int i;

  TraceMark(S_pztMotor_ENTERING_CONSTRUCTOR);

  switch(iBoardNum) {
   case 0:
    this->vmeAddress = tVmeAddress(VME_AM_SUP_SHORT_IO, (uint8_t *) 0x8100);
    break;
   default:
    this->vmeAddress = tVmeAddress();
    break;
  }
  this->boardName  = PZT_MOTOR_BOARD_NAME + iBoardNum;

  this->iNumPztChannels = 0;
  this->iNumMotorChannels = 0;
  this->iNumLimitChannels = 0;

  for(i=0; i < PZT_MOTOR_MAX_NUM_PZTS; i++) {
    if(this->bRampArray[i]) {
      this->pPztChannelArray[i] = new tPztMotorRampChannel;
    }
    else {
      this->pPztChannelArray[i] = new tPztMotorDaChannel;
    }
  }

  TraceMark(S_pztMotor_DONE_CONSTRUCTING);
}


/*******************************************************
*
* ERROR_CODE tPztMotorBoard::SoftInit - Initialize pzt/motor board "softly"
*
* Initializes the software driver by setting up pointers to the various
* pieces of hardware on the pzt/motor board, but it does not reset the 
* hardware.  This routine is useful if, for example, you have had to reboot
* The argument numPzts is the number of pztcontrollers on the board to init.
* You may specify a number smaller than the number actually installed.
* If numPzts=0 then the board is not initialized.
*
[[U tPztMotorBoard::SoftInit(iNumPztsToUse, bRampArray, ccConvCodeArray,
                             dPztRangeArray, iNumMotorsToUse, dMotorRangeArray)]]
*
******/
ERROR_CODE tPztMotorBoard::SoftInit(const int         iNumPztsToUse,
                                    const BOOL        *bRampArrayToUse,
                                    const CONV_CODE   *ccConvCodeArray,
                                    const double      *dPztRangeArray,
                                    const int         iNumMotorsToUse, 
                                    const double      *dMotorRangeArray,
                                    const int         iNumLimitsToUse)
{
  int i;
  ERROR_CODE ec;

  TraceMark(S_pztMotor_ENTERING_SOFTINIT);

  if ((iNumPztsToUse > PZT_MOTOR_MAX_NUM_PZTS) || (iNumPztsToUse < 1)) {
    return Traceback(S_pztMotor_E_INVALID_NUMBER_OF_PZTS,
                     "tPztMotorBoard::SoftInit",
                      iNumPztsToUse, bRampArray,
                      ccConvCodeArray, dPztRangeArray,
                      iNumMotorsToUse, dMotorRangeArray);
  }
  
  if ((iNumMotorsToUse > PZT_MOTOR_MAX_NUM_MOTORS) || (iNumMotorsToUse < 1)) {
    return Traceback(S_pztMotor_E_INVALID_NUMBER_OF_MOTORS,
                     "tPztMotorBoard::SoftInit",
                       iNumPztsToUse, bRampArray,
                      ccConvCodeArray, dPztRangeArray,
                      iNumMotorsToUse, dMotorRangeArray);
  }

  if ((iNumLimitsToUse > PZT_MOTOR_MAX_NUM_LIMITS) || (iNumLimitsToUse < 1)) {
    return Traceback(S_pztMotor_E_INVALID_NUMBER_OF_LIMITS,
                     "tPztMotorBoard::SoftInit",
                       iNumPztsToUse, bRampArray,
                      ccConvCodeArray, dPztRangeArray,
                      iNumMotorsToUse, dMotorRangeArray);
  }

  ec = tBoard::Initialize();
  if(ec!=OK) {
    return Traceback(ec, "tPztMotorBoard::SoftInit", iNumPztsToUse,
                     bRampArray, ccConvCodeArray, dPztRangeArray,
                      iNumMotorsToUse, dMotorRangeArray);
  }

  this->iNumPztChannels   = MAX(this->iNumPztChannels,   iNumPztsToUse);
  this->iNumMotorChannels = MAX(this->iNumMotorChannels, iNumMotorsToUse);
  this->iNumLimitChannels = MAX(this->iNumLimitChannels, iNumLimitsToUse);

  for(i=0; i<PZT_MOTOR_MAX_NUM_PZTS; i++) {
    this->bRampArray[i] = bRampArrayToUse[i];
  } 

  TraceMark(S_pztMotor_INITIALIZING_DA_CHANNEL_ARRAY);

  for(i=0; i < this->iNumPztChannels; i++) {
    ec = PztReadbackChannelArray[i].Initialize((volatile int16_t *)
                                     (vu8LocalAddress + GetPztReadbackOffset(i)),
                                  CC_TWOS_COMPLEMENT,
                                  dPztRangeArray[i]); 
    if(ec != OK) {
        return Traceback(ec, "tPztMotorBoard::SoftInit", iNumPztsToUse,
                     bRampArray, ccConvCodeArray, dPztRangeArray,
                      iNumMotorsToUse, dMotorRangeArray); 
    }
                      
    if(this->bRampArray[i]) {
      ec = pPztChannelArray[i]->Initialize((volatile int16_t *)
                                            (vu8LocalAddress + GetPztOffset(i)),
                                        ccConvCodeArray[i],
                                        dPztRangeArray[i],
                                        CiRampResetValue,
                                        &(this->PztReadbackChannelArray[i]));
      if(ec != OK) {
        return Traceback(ec, "tPztMotorBoard::SoftInit", iNumPztsToUse,
                     bRampArray, ccConvCodeArray, dPztRangeArray,
                      iNumMotorsToUse, dMotorRangeArray);
      }
    }
    else { /* Not a ramp channel, must be a D/A channel */
      ec = pPztChannelArray[i]->Initialize((volatile int16_t *)
                                          (vu8LocalAddress + GetPztOffset(i)),
                                        ccConvCodeArray[i],
                                        dPztRangeArray[i],
                                        CiDaResetValue, 0);
      if(ec != OK) {
        return Traceback(ec, "tPztMotorBoard::SoftInit", iNumPztsToUse,
                     bRampArray, ccConvCodeArray, dPztRangeArray,
                      iNumMotorsToUse, dMotorRangeArray);
      }
    }
  }

  TraceMark(S_pztMotor_INITIALIZING_MOTOR_CHANNEL_ARRAY);

  for(i=0; i < this->iNumMotorChannels; i++) {
    ec = MotorReadbackChannelArray[i].Initialize((volatile int16_t *)
                                    (vu8LocalAddress + GetMotorReadbackOffset(i)),
                                 CC_TWOS_COMPLEMENT,
                                 dMotorRangeArray[i]);

    if(ec != OK) {
        return Traceback(ec, "tPztMotorBoard::SoftInit", iNumPztsToUse,
                     bRampArray, ccConvCodeArray, dPztRangeArray,
                      iNumMotorsToUse, dMotorRangeArray);
    }
     
    ec = MotorChannelArray[i].Initialize((volatile int16_t *)
                                            (vu8LocalAddress + GetMotorOffset(i)),
                                        ccConvCodeArray[i],
                                        dMotorRangeArray[i],
                                        &(this->MotorReadbackChannelArray[i]));
    if(ec != OK) {
      return Traceback(ec, "tPztMotorBoard::SoftInit", iNumPztsToUse,
                     bRampArray, ccConvCodeArray, dPztRangeArray,
                      iNumMotorsToUse, dMotorRangeArray); 
    }
  }    

  TraceMark(S_pztMotor_INITIALIZING_LIMIT_CHANNEL_ARRAY);

  for(i=0; i < this->iNumLimitChannels; i++) {
    ec = LimitChannelArray[i].Initialize((volatile uint16_t *)
                                         (vu8LocalAddress + GetLimitOffset(i)),
                                         CltLimitTypeArray[i]);
    if(ec!=OK) {
      return Traceback(ec, "tPztMotorBoard::SoftInit", iNumPztsToUse,
                     bRampArray, ccConvCodeArray, dPztRangeArray,
                      iNumMotorsToUse, dMotorRangeArray);
    }
  }

  TraceMark(S_pztMotor_INITIALIZED_SOFTLY);
  this->SetInit(TRUE);
  return OK;
} 


/*******************************************************
*
* ERROR_CODE tPztMotorBoard::SoftInit - void version -
*                                     Initialize pzt/motor board "softly"
*
*
*
****/
ERROR_CODE tPztMotorBoard::SoftInit(void)
{
  return (this->SoftInit(PZT_MOTOR_MAX_NUM_PZTS,
                 CbDefaultRampArray,
                 CccDefaultConvCodeArray,
                 CdDefaultPztRangeArray,
                 PZT_MOTOR_MAX_NUM_MOTORS,
                 CdDefaultMotorRangeArray,
                 PZT_MOTOR_MAX_NUM_LIMITS));
}


/************************************************************
*
* HardInit - Initialize pzt/motor board software and hardware
*
* Initializes the pzt velocities and counters to zero.  Note
* that a pzt position of zero is mid-scale, or 5 volts. 
* The argument iNumPztChannels is the number of pzt controllers
* on the board to init.
* You may specify a number smaller than the number actually installed.
* If iNumPztChannels = 0, then the board is not initialized.
*
IN :
  this

INOUT :
  this->bInit

OUT :
  this->vu8LocalAddress
  this->boardName
  this->iNumPztChannels
  this->pztVelocity[1..iNumPztChannels]
  this->pztClearCmd[1..iNumPztChannels]
  this->pztPosition[1..iNumPztChannels]
  this->pztStatus[1..iNumPztChannels]

SPEC :
  [this is invalid  -> UNDEFINED
  | this->vmeAddress not reachable by CPU ->
      return VxWorks error code from sysBusToLocalAdrs
  | numCounters is not in range of board ->
      return S_pztMotor_INIT_INVALID_ARGUMENT
  | this->bInit == TRUE -> return S_pztMotor_BOARD_ALREADY_INITIALIZED
  | true -> 
    : this->vu8LocalAddress := local address corresponding to this->vmeAddress
    : this->boardName := boardName
    : this->bInit := TRUE
    : initialize board register pointers and protected variables
    : set 1..iNumPztChannels channels velocity to zero
    : clear 1..iNumPztChannels channels
    : return OK
  ]
LOCAL :
  int        i
  ERROR_CODE ec

RETURNS: ERROR_CODE
S_pztMotor_BOARD_ALREADY_INITIALIZED
S_pztMotor_INVALID_NUMBER_OF_PZTS
OK
*/
ERROR_CODE tPztMotorBoard::HardInit(const int         iNumPztsToUse,
                                    const BOOL        *bRampArray,
                                    const CONV_CODE   *ccConvCodeArray,
                                    const double      *dPztRangeArray,
                                    const int         iNumMotorsToUse, 
                                    const double      *dMotorRangeArray,
                                    const int         iNumLimitsToUse)
{
  int        i;
  ERROR_CODE ec;

  TraceMark(S_pztMotor_ENTERING_HARDINIT);

  /*
  [[U tPztMotorBoard::SoftInitParams()]]
  */
  ec = this->SoftInit(iNumPztsToUse,
                      bRampArray, ccConvCodeArray, dPztRangeArray,
                      iNumMotorsToUse, dMotorRangeArray, iNumLimitsToUse);
  if (ec!=OK) {
    return Traceback(ec, "tPztMotorBoard::HardInit", iNumPztsToUse,
                     bRampArray, ccConvCodeArray, dPztRangeArray,
                      iNumMotorsToUse, dMotorRangeArray); 
  }

  /*
  [ true -> set 0..iNumPztChannels channels velocity to zero
    : clear 0..iNumPztPztChannels channels
    : return OK
  ] */
  for (i=0; i<this->iNumPztChannels; i++) {
    TraceMsg(tmt_INFO, S_pztMotor_CALLING_SOFT_INIT,
             "Calling tPztMotorDaChannel::SoftInit\n");
    pPztChannelArray[i]->Reset();
  }

  /*
  [ true -> set 0..iNumMotorChannels channels velocity to zero
    : clear 0..iNumMotorChannels channels
    : return OK
  ] */
  for (i=0; i<this->iNumMotorChannels; i++) {
    TraceMsg(tmt_INFO, S_pztMotor_CALLING_SOFT_INIT,
             "Calling tPztMotorStepMotorChannel::SoftInit");
    MotorChannelArray[i].Reset();
  }

  TraceMark(S_pztMotor_INITIALIZED_HARDLY);

  return OK;
}

/*******************************************************
*
* ERROR_CODE tPztMotorBoard::HardInit - void version -
*                                     Initialize pzt/motor board 
*                                     including board registers
*
[[U tPztMotorBoard::HardInitVoid()]]
*
****/
ERROR_CODE tPztMotorBoard::HardInit(void)
{
  return (this->HardInit(PZT_MOTOR_MAX_NUM_PZTS,
                 CbDefaultRampArray,
                 CccDefaultConvCodeArray,
                 CdDefaultPztRangeArray,
                 PZT_MOTOR_MAX_NUM_MOTORS, 
                 CdDefaultMotorRangeArray,
                 PZT_MOTOR_MAX_NUM_LIMITS));
}

/******************************************************************************
*
* tPztMotorBoard::CreateChannel - gets channel specified by iNumChannel (global
*                                 channel numbering)
*
[[U tPztMotorBoard::CreateChannel()]]
*/
tChannel *tPztMotorBoard::CreateChannel(int iChannelNum)
{
  int iMaxLocalPzt           = PZT_MOTOR_MAX_NUM_PZTS,
      iMaxLocalMotor         = (PZT_MOTOR_MAX_NUM_PZTS + PZT_MOTOR_MAX_NUM_MOTORS),
      iMaxLocalPztReadback   = (iMaxLocalMotor + PZT_MOTOR_MAX_NUM_PZTS),
      iMaxLocalMotorReadback = (iMaxLocalPztReadback + PZT_MOTOR_MAX_NUM_MOTORS),
      iMaxLocalLimits        = (iMaxLocalMotorReadback + PZT_MOTOR_MAX_NUM_LIMITS),
      iMaxChannel            = iMaxLocalLimits,
      iLocalChannel;

  if(iChannelNum < 0 || iChannelNum >= iMaxChannel) {
    TraceMsg(tmt_ERROR, S_pztMotor_CHANNEL_NUM_OUT_OF_RANGE,
             "NULL pointer returned for pztMotor board\n");
    return NULL;
  }
  
  if(iChannelNum<iMaxLocalPzt) {
    iLocalChannel = iChannelNum;
    if( (this->IsInit() && (iLocalChannel >= this->iNumPztChannels)) ||
                             iLocalChannel >= PZT_MOTOR_MAX_NUM_PZTS ) {
      TraceMsg(tmt_ERROR, S_pztMotor_PZT_CHANNEL_NUM_OUT_OF_RANGE,
               "Returning NULL channel for pzt channel\n");
      return NULL;
    }
    this->iNumPztChannels = MAX(this->iNumPztChannels, iLocalChannel+1);
    return (pPztChannelArray[iLocalChannel]);
  }

  else if(iChannelNum<(iMaxLocalMotor)) {
    iLocalChannel = iChannelNum-iMaxLocalPzt;
    if( (this->IsInit() && (iLocalChannel >= this->iNumMotorChannels)) ||
                             iLocalChannel >= PZT_MOTOR_MAX_NUM_MOTORS ) {
      TraceMsg(tmt_ERROR, S_pztMotor_MOTOR_CHANNEL_NUM_OUT_OF_RANGE,
               "Returning NULL channel for motor channel\n");
      return NULL;
    }
    this->iNumMotorChannels = MAX(this->iNumMotorChannels, iLocalChannel+1);
    return (&(MotorChannelArray[iLocalChannel]));
  }

  else if(iChannelNum<(iMaxLocalPztReadback)) {
    iLocalChannel = iChannelNum-iMaxLocalMotor;
    if( (this->IsInit() && (iLocalChannel >= this->iNumPztChannels)) ||
                             iLocalChannel >= PZT_MOTOR_MAX_NUM_PZTS ) {
      TraceMsg(tmt_ERROR, S_pztMotor_PZT_READBACK_CHANNEL_NUM_OUT_OF_RANGE,
               "Returning NULL channel for pzt readback channel\n");
      return NULL;
    }
    this->iNumPztChannels = MAX(this->iNumPztChannels, iLocalChannel+1);
    return (&(PztReadbackChannelArray[iLocalChannel]));
  }

  else if(iChannelNum<(iMaxLocalMotorReadback)) {
    iLocalChannel = iChannelNum-iMaxLocalPztReadback;
    if( (this->IsInit() && (iLocalChannel >= this->iNumMotorChannels)) ||
                             iLocalChannel >= PZT_MOTOR_MAX_NUM_MOTORS ) {
      TraceMsg(tmt_ERROR, S_pztMotor_MOTOR_READBACK_CHANNEL_NUM_OUT_OF_RANGE,
               "Returning NULL channel for pzt readback channel\n");
      return NULL;
    }
    this->iNumMotorChannels = MAX(this->iNumMotorChannels, iLocalChannel+1);
    return (&(PztReadbackChannelArray[iLocalChannel]));
  }

  else if(iChannelNum<(iMaxLocalLimits)) {
    iLocalChannel = iChannelNum-iMaxLocalMotorReadback;
    if( (this->IsInit() && (iLocalChannel >= this->iNumLimitChannels)) ||
                             iLocalChannel >= PZT_MOTOR_MAX_NUM_LIMITS ) {
      TraceMsg(tmt_ERROR, S_pztMotor_LIMIT_CHANNEL_NUM_OUT_OF_RANGE,
               "Returning NULL channel for limit channel\n");
      return NULL;
    }
    this->iNumLimitChannels = MAX(this->iNumLimitChannels, iLocalChannel+1);
    return (&(LimitChannelArray[iLocalChannel]));
  }
  TraceMsg(tmt_ERROR, S_pztMotor_INVALID_CHANNEL_NUM,
               "Should never reach this point.  Returning NULL channel pointer\n");
  return NULL;
}



/*****************************************************************
*
* tPztMotorRampChannel::tPztMotorRampChannel()  - constructor
*
[[U tPztMotorRampChannel::tPztMotorRampChannel()]]
*
******/
tPztMotorRampChannel::tPztMotorRampChannel(void)
{
  TraceMark(S_pztMotor_ENTERING_RAMP_CHANNEL_CONSTRUCTOR);

  this->vwDataOut   = NULL;
  this->vuwClearCmd = NULL;
  this->SetInit(FALSE);

  TraceMark(S_pztMotor_DONE_CONSTRUCTING_RAMP_CHANNEL);
}
  

/*****************************************************************
*
* tPztMotorRampChannel::Initialize(...)  - class initializer for
*                                       a pzt "ramp" channel (IOW. integrating)
*
[[U tPztMotorRampChannel::Initialize()]]
*
******/
ERROR_CODE tPztMotorRampChannel::Initialize(volatile int16_t *vwDataOutToUse,
                                     CONV_CODE        ccConvCodeToUse,
                                     double           dVoltageRampRangeToUse,
                                     int              iResetValueToUse,
                                     tInputChannel    *pReadbackChannelToUse)
{
  TraceMark(S_pztMotor_ENTERING_RAMP_CHANNEL_INITIALIZE);

  TraceMark(S_pztMotor_CALLING_DA_CHANNEL_INITIALIZE);
  tPztMotorDaChannel::Initialize(vwDataOutToUse, ccConvCodeToUse,
                                 dVoltageRampRangeToUse, iResetValueToUse,
                                 pReadbackChannelToUse);

  this->vuwClearCmd      = (volatile uint16_t *) (vwDataOutToUse + 0x04);

  TraceMark(S_pztMotor_DONE_INITIALIZING_RAMP_CHANNEL);

  return OK;
}


/*****************************************************************
*
* tPztMotorRampChannel::Output(...)  - Set pzt velocity
*            <formerly pztVelocitySet 
*
* Sets the velocity for this pzt controller.  
*
SPEC :
  [ this is invalid -> UNDEFINED
  | this->bInit == FALSE || whichPzt<1 || whichPzt>this->iNumPztChannels ||
    newVelocity<MIN_VELOCITY || newVelocity>MAX_VELOCITY ->
      UNDEFINED
  | true -> this->*pztVelocity[whichPzt] := newVelocity
    : return OK
  ]

RETURNS:
S_pztMotor_BOARD_NOT_INITIALIZED
S_pztMotor_INVALID_PZT
S_pztMotor_INVALID_VELOCITY
OK
*/
ERROR_CODE tPztMotorRampChannel::Output(int iValue)
{
  int16_t wOutputValue;
  /*
  [ this is invalid -> UNDEFINED
  | this->bInit == FALSE -> UNDEFINED
  | true -> NOP
  ] */
#ifdef DEBUG
  if (!this->bInit){
    return S_pztMotor_BOARD_NOT_INITIALIZED;
  }
#endif

  /*
  [ iValue<MIN_VELOCITY || iValue>MAX_VELOCITY ->
      UNDEFINED
  | true -> NOP
  ] */
#ifdef DEBUG
  if (iValue < MIN_VELOCITY || iValue > MAX_VELOCITY) {
    return S_pztMotor_INVALID_VELOCITY;
  }
#endif

  /*
  [ true -> this->*pztVelocity[whichPzt] := iValue
    : return OK
  ] */
  wOutputValue = (int16_t) (GetOutputValue(iValue));
  *(this->vwDataOut) = wOutputValue;
  return OK;
}

/***********************************************************
*
* tPztMotorRampChannel::Reset - Reset pzt velocity and counter
*
* Writes a velocity of zero to the pzt controller and resets the
* counter to zero (mid-scale or 5 volts).
*
*
LOCAL :
ERROR_CODE ec
*
RETURNS:
S_pztMotor_BOARD_NOT_INITIALIZED
S_pztMotor_INVALID_PZT
S_pztMotor_INVALID_VELOCITY
OK
*
*/
ERROR_CODE tPztMotorRampChannel::Reset(void)
{
  ERROR_CODE ec;

  /*
  [ this is invalid -> UNDEFINED
  | this->bInit == FALSE || whichPzt is out of range for board ->
      UNDEFINED
  | true ->  set whichPzt channel velocity to zero
    : clear whichPzt channel
  ] */
  ec = this->Output(this->CiZeroVelocity);
#ifdef DEBUG
  if(ec!=OK) {
    return Traceback(ec, "tPztMotorRampChannel::Reset");
  }
#endif


  /*
  [ this is invalid -> UNDEFINED
  | this->bInit == FALSE || whichPzt is out of range for the board ->
      UNDEFINED
  | true -> clear whichPzt channel 
    : return OK
  ] */
  ec = this->ClearCounter();
#ifdef DEBUG
  if(ec!=OK) {
    return Traceback(ec, "tPztMotorRampChannel::Reset");
  } 
#endif
  return OK;
}

/***************************************************************
*
* pztMotorRampChannel::ClearCounter - Clear pzt counter
*
* Uses the hardware reset command to clear the motor counter to zero.
*
SPEC :
  [ this is invalid -> UNDEFINED
  | this->bInit == FALSE ->
      UNDEFINED
  | true -> this->*pztClearCmd[whichPzt] := DUMMY
    : return OK
  ]
*
RETURNS: ERROR_CODE
S_pztMotor_BOARD_NOT_INITIALIZED
S_pztMotor_INVALID_PZT
OK
*/
ERROR_CODE tPztMotorRampChannel::ClearCounter(void)
{
  uint16_t uwDummy = 0;

  /*
  [ this is invalid -> UNDEFINED
  | this->bInit == FALSE -> UNDEFINED
  | true -> NOP
  ] */
#ifdef DEBUG
  if (!this->bInit) {
    return S_pztMotor_BOARD_NOT_INITIALIZED;
  }
#endif

  /*
  [ true -> this->*pztClearCmd[whichPzt] := DUMMY
    : return OK
  ] */
  *(this->vuwClearCmd) = uwDummy;
  return OK;
}


/*****************************************************************
*
* tPztMotorDaChannel::tPztMotorDaChannel()  - constructor
*
[[U tPztMotorDaChannel::tPztMotorDaChannel()]]
*
******/
tPztMotorDaChannel::tPztMotorDaChannel(void)
{
  TraceMark(S_pztMotor_ENTERING_DA_CHANNEL_CONSTRUCTOR);

  this->vwDataOut   = NULL;
  this->SetInit(FALSE);

  TraceMark(S_pztMotor_DONE_CONSTRUCTING_DA_CHANNEL);
}


/*****************************************************************
*
* tPztMotorDaChannel::Initialize(...)  - class initializer for
*                                a pzt "D/A" channel (IOW. non-integrating)
*
[[U tPztMotorDaChannel::Initialize()]]
*
******/
ERROR_CODE tPztMotorDaChannel::Initialize(volatile int16_t *vwDataOutToUse,
                                     CONV_CODE        ccConvCodeToUse,
                                     double           dVoltageRampRangeToUse,
                                     int              iResetValueToUse,
                                     tInputChannel    *pReadbackChannelToUse)
{
  TraceMark(S_pztMotor_ENTERING_DA_CHANNEL_INITIALIZE);

  this->vwDataOut        = vwDataOutToUse;
  this->ccConvCode       = ccConvCodeToUse;
  this->dChannelRange    = dVoltageRampRangeToUse;
  this->pReadbackChannel = pReadbackChannelToUse;
  this->iResetValue      = iResetValueToUse;
  this->iBitWidth        = this->CiBitWidth;
  this->SetInit(TRUE);
  
  TraceMark(S_pztMotor_DONE_INITIALIZING_DA_CHANNEL);
  return OK;
}


/*****************************************************************
*
* tPztMotorDaChannel::Output(...) - sets output voltage on a D/A channel
*            <formerly pztPositionSet>
*
* Sets the position (D/A value) for a given pzt_d2a controller.  Note that
* this is a special command for use with the pzt_d2a version of the pzt
* chip.  It is equivalent to pztVelocitySet and in fact uses the same
* register.  So this command is just provided for clarity.  Also,
* note that the only usable commands for this version of the chip
* are pztPositionSet (or pztVelocitySet) and pztPositionRead.  To set
* the D/A output to 0 volts, call pztPositionSet(whichPzt, MIN_VELOCITY).  To
* set the D/A output to midrange,
* call pztPositionSet(whichPzt, ZERO_VELOCITY).
* At power-up, the output is set to 0 volts.  System reset has no effect
* on the output.
*
LOCAL :
ERROR_CODE ec

RETURNS: ERROR_CODE
S_pztMotor_BOARD_NOT_INITIALIZED
S_pztMotor_INVALID_PZT
S_pztMotor_INVALID_PZT_POSITION
OK
*/


ERROR_CODE tPztMotorDaChannel::Output(int iValue)
{
  int16_t wOutputValue;
  /*
  [ this is invalid -> UNDEFINED
  | this->bInit == FALSE -> UNDEFINED
  | true -> NOP
  ] */
#ifdef DEBUG
  if (!this->bInit){
    return S_pztMotor_BOARD_NOT_INITIALIZED;
  }
#endif

  /*
  [ iValue<MIN_VELOCITY || iValue>MAX_VELOCITY ->
      UNDEFINED
  | true -> NOP
  ] */
#ifdef DEBUG
  if (iValue < MIN_VELOCITY || iValue > MAX_VELOCITY) {
    return S_pztMotor_INVALID_VELOCITY;
  }
#endif

  /*
  [ true -> this->*pztVelocity[whichPzt] := iValue
    : return OK
  ] */
  wOutputValue = (int16_t) (GetOutputValue(iValue));
  *(this->vwDataOut) = wOutputValue;
  return OK;
}
/*****************************************************************
*
* tPztMotorReadbackChannel::tPztMotorReadbackChannel()  - constructor
*
[[U tPztMotorReadbackChannel::tPztMotorReadbackChannel()]]
*
******/
tPztMotorReadbackChannel::tPztMotorReadbackChannel(void)
{
  TraceMark(S_pztMotor_ENTERING_MOTOR_READBACK_CHANNEL_CONSTRUCTOR);

  this->vwDataIn    = NULL;
  this->SetInit(FALSE);

  TraceMark(S_pztMotor_DONE_CONSTRUCTING_MOTOR_READBACK_CHANNEL);
}


/***************************************************************
*
* tPztMotorReadbackChannel::Initialize(...) class initializer for
*                                a readback channel
*
*  This class works for both a pzt readback channel and a
*  motor readback channel, as long as the proper vwDataToUse
*  is passed in.
* 
[[U tPztMotorReadbackChannel::Initialize()]]
*
******/
ERROR_CODE tPztMotorReadbackChannel::Initialize(volatile int16_t *vwDataInToUse,
                                       CONV_CODE        ccConvCodeToUse,
                                       double           dVoltageRangeToUse)
{
  TraceMark(S_pztMotor_ENTERING_MOTOR_READBACK_CHANNEL_INITIALIZE);
  
  this->vwDataIn         = vwDataInToUse;
  this->ccConvCode       = ccConvCodeToUse;
  this->dChannelRange    = dVoltageRangeToUse;
  this->iBitWidth        = this->CiBitWidth;
  this->SetInit(TRUE);

  TraceMark(S_pztMotor_DONE_INITIALIZING_MOTOR_READBACK_CHANNEL);
  return OK;
}
   
   
/***************************************************************
*
* tPztMotorReadbackChannel::Input(...)
*            - Reads the current position of the pzt.
*              <formerly pztPositionRead>
*
* Reads the position from the counter of the selected pzt controller.
*
[[U tPztMotorReadbackChannel::Input()]]
*
******/
ERROR_CODE tPztMotorReadbackChannel::Input(int *piValue) const
{
  int iInputValue; 
  /*
  [ this is invalid -> UNDEFINED
  | this->bInit == FALSE -> UNDEFINED
  | true -> NOP
  ] */
#ifdef DEBUG
  if (!this->bInit) {
    return S_pztMotor_BOARD_NOT_INITIALIZED;
  }
#endif

  /* read counter to postion pointer */
  /*
  [ position is invalid -> UNDEFINED
  | true -> *position := this->*pztPosition[whichPzt]
    : return OK
  */
  iInputValue = (int) *(this->vwDataIn);
  *piValue = GetInputValue(iInputValue);
  return OK;
}


/*****************************************************************
*
* tPztMotorStepMotorChannel::tPztMotorStepMotorChannel()  - constructor
*
[[U tPztMotorStepMotorChannel::tPztMotorStepMotorChannel()]]
*
******/
tPztMotorStepMotorChannel::tPztMotorStepMotorChannel(void)
{
  TraceMark(S_pztMotor_ENTERING_MOTOR_CHANNEL_CONSTRUCTOR);

  this->vwDataOut   = NULL;
  this->vuwClearCmd = NULL;
  this->SetInit(FALSE);

  TraceMark(S_pztMotor_DONE_CONSTRUCTING_MOTOR_CHANNEL);
}


/*****************************************************************
* 
* ERROR_CODE tPztMotorStepMotorChannel::Initialize(...)
*
*****/
ERROR_CODE tPztMotorStepMotorChannel::Initialize
                       (volatile int16_t *vwDataOutToUse,
                        CONV_CODE        ccConvCodeToUse,
                        double           dPulseSpeedRangeToUse,
                        tInputChannel    *pReadbackChannelToUse)
{
  TraceMark(S_pztMotor_ENTERING_MOTOR_CHANNEL_INITIALIZE);

  this->vwDataOut        = vwDataOutToUse;
  this->vuwClearCmd      = (volatile uint16_t *) (vwDataOutToUse + 0x04);
  this->ccConvCode       = ccConvCodeToUse;
  this->dChannelRange    = dPulseSpeedRangeToUse;
  this->pReadbackChannel = pReadbackChannelToUse;
  this->iResetValue      = this->CiResetValue;
  this->iBitWidth        = this->CiBitWidth;
  this->SetInit(TRUE);

  TraceMark(S_pztMotor_DONE_INITIALIZING_MOTOR_CHANNEL);
  return OK;
}

/*****************************************************************
* 
* ERROR_CODE tPztMotorStepMotorChannel::Output(...)
*                 <formerly motorVelocitySet>
*
*Sets the velocity for a given motor controller.  
pztStop(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid motor controller
*or an invalid velocity.
******/
ERROR_CODE tPztMotorStepMotorChannel::Output(int iValue)
{
  int16_t wOutputValue;
  
#if (DEBUG>=30)
  if (!this->bInit) {
    return(S_pztMotor_BOARD_NOT_INITIALIZED);
  }
#endif
#ifdef DEBUG
  if (iValue < MIN_VELOCITY || newVelocity > MAX_VELOCITY){
    errnoSet(S_pztMotor_INVALID_VELOCITY);
    return(ERROR);
  }
#endif

  wOutputValue = (int16_t) (GetOutputValue(iValue));
  /* Write new velocity */
  *(this->vwDataOut) = wOutputValue;
  return(OK);
}


/***********************************************************
*
* tPztMotorStepMotorChannel::Reset - Reset motor velocity and counter
*
* Writes a velocity of zero to the pzt controller and resets the
* counter to zero (mid-scale or 5 volts).
*
*
LOCAL :
ERROR_CODE ec
*
RETURNS:
S_pztMotor_BOARD_NOT_INITIALIZED
S_pztMotor_INVALID_PZT
S_pztMotor_INVALID_VELOCITY
OK
*
*/
ERROR_CODE tPztMotorStepMotorChannel::Reset(void)
{
  ERROR_CODE ec;

  /*
  [ this is invalid -> UNDEFINED
  | this->bInit == FALSE || whichPzt is out of range for board ->
      UNDEFINED
  | true ->  set whichPzt channel velocity to zero
    : clear whichPzt channel
  ] */
  ec = this->Output(this->CiZeroVelocity);
#ifdef DEBUG
  if(ec!=OK) {
    return Traceback(ec, "tPztMotorRampChannel::Reset");
  }
#endif


  /*
  [ this is invalid -> UNDEFINED
  | this->bInit == FALSE || whichPzt is out of range for the board ->
      UNDEFINED
  | true -> clear whichPzt channel 
    : return OK
  ] */
  ec = this->ClearCounter();
#ifdef DEBUG
  if(ec!=OK) {
    return Traceback(ec, "tPztMotorRampChannel::Reset");
  } 
#endif
  return OK;
}


/***************************************************************
*
* pztMotorStepMotorChannel::ClearCounter - Clear pzt counter
*
* Uses the hardware reset command to clear the motor counter to zero.
*
SPEC :
  [ this is invalid -> UNDEFINED
  | this->bInit == FALSE -> UNDEFINED
  | true -> this->*vuwClearCmd := DUMMY
    : return OK
  ]
*
RETURNS: ERROR_CODE
S_pztMotor_BOARD_NOT_INITIALIZED
S_pztMotor_INVALID_PZT
OK
*/
ERROR_CODE tPztMotorStepMotorChannel::ClearCounter(void)
{
  uint16_t uwDummy = 0;
  /*
  [ this is invalid -> UNDEFINED
  | this->bInit == FALSE -> UNDEFINED
  | true -> NOP
  ] */
#ifdef DEBUG
  if (!this->bInit) {
    return S_pztMotor_BOARD_NOT_INITIALIZED;
  }
#endif

  /*
  [ true -> this->*pztClearCmd[whichPzt] := DUMMY
    : return OK
  ] */
  *(this->vuwClearCmd) = uwDummy;
  return OK;
}

/*****************************************************************
*
* tPztMotorDigitalInChannel::tPztMotorDigitalInChannel()  - constructor
*
[[U tPztMotorDigitalInChannel::tPztMotorDigitalInChannel()]]
*
******/
tPztMotorDigitalInChannel::tPztMotorDigitalInChannel(void)
{
  TraceMark(S_pztMotor_ENTERING_DI_CHANNEL_CONSTRUCTOR);

  this->vuwDataIn     = NULL;
  this->iStatusBitNum = 0;
  this->SetInit(FALSE);

  TraceMark(S_pztMotor_DONE_CONSTRUCTING_DI_CHANNEL);
}

ERROR_CODE tPztMotorDigitalInChannel::Initialize(volatile uint16_t *vuwDataToUse,
                                                 LIMIT_TYPE        ltLimitTypeToUse)
{
  TraceMark(S_pztMotor_ENTERING_DI_CHANNEL_INITIALIZE);

  this->vuwDataIn     = vuwDataToUse;
  this->iBitWidth     = this->CiBitWidth;
  this->dChannelRange = this->CdChannelRange;
  this->ccConvCode    = this->CccConvCode;

  switch(ltLimitTypeToUse) {
   case LT_CW_SOFT:
     this->iStatusBitNum = 1;
     break;
   case LT_CCW_SOFT:
     this->iStatusBitNum = 2;
     break;
   case LT_CW_HARD:
     this->iStatusBitNum = 5;
     break;
   case LT_CCW_HARD:
     this->iStatusBitNum = 6;
     break;
   case LT_HOME:
     this->iStatusBitNum = 9;
     break;
   default:
     return Traceback(S_pztMotor_INVALID_LIMIT_TYPE,
                      "tPztMotorDigitalInChannel::Initialize",
                      vuwDataToUse, ltLimitTypeToUse);
  }
  this->SetInit(TRUE);
  
  TraceMark(S_pztMotor_DONE_INITIALIZING_DI_CHANNEL);
  return OK;
}


ERROR_CODE tPztMotorDigitalInChannel::Input(int *piValue) const
{
#ifdef DEBUG
  if(!this->bInit) {
    return Traceback(S_pztMotor_CHANNEL_NOT_INITIALIZED,
                     "tPztMotorDigitalInChannel::Input", piValue);
  }
#endif

  *piValue = (0x01 & (*(this->vuwDataIn) >> this->iStatusBitNum));

  return(OK);
}




/* End of file.  Everything beyond this point is comments */

///////////////////////////////////////
//  DELETED FUNCTIONS FROM INCREMENT 1 
///////////////////////////////////////
/***********************************************************
*
* pztSetHigh - set pzt to max output
*
* writes a positive velocity to the pzt controller to ramp it
* to the plus rail
*
LOCAL :
ERROR_CODE ec
*
* RETURNS:
S_pztMotor_BOARD_NOT_INITIALIZED
S_pztMotor_INVALID_PZT
S_pztMotor_INVALID_VELOCITY
OK
*
*/
//ERROR_CODE tPztMotor::pztSetHigh(int whichPzt)
//{
//  ERROR_CODE ec;
//
//  /*
//  [ this is invalid -> UNDEFINED
//  | this->bInit == FALSE || whichPzt is out of range for board
//      UNDEFINED
//  | true ->  set whichPzt channel velocity to MAX_VELOCITY/10
//    : return OK
//  ] */
//  ec = pztVelocitySet(whichPzt,MAX_VELOCITY/10);
//#ifdef DEBUG
//  if(ec!=OK) {
//    return Traceback(ec, "tPztMotor::pztSetHigh", whichPzt);
//  } 
//#endif
//  return OK;
//}

/***********************************************************
*
* pztSetLow - set pzt to min output
*
* writes a negative velocity to the pzt controller to ramp it
* to zero
*
LOCAL :
ERROR_CODE ec
*
* RETURNS:
S_pztMotor_BOARD_NOT_INITIALIZED
S_pztMotor_INVALID_PZT
S_pztMotor_INVALID_VELOCITY
* OK
*/
//ERROR_CODE tPztMotor::pztSetLow(int whichPzt)
//{
//  ERROR_CODE ec;
//
//  /*
//  [ this is invalid -> UNDEFINED
//  | this->bInit == FALSE || whichPzt is out of range for board ->
//      UNDEFINED
//  | true ->  set whichPzt channel velocity to -MAX_VELOCITY/10
//    : return OK
//  ] */
//  ec = pztVelocitySet(whichPzt,-MAX_VELOCITY/10);
//#ifdef DEBUG
//  if(ec!=OK) {
//    return Traceback(ec, "tPztMotor::pztSetLow", whichPzt);
//  } 
//#endif
//  return OK;
//}



/******************************************************************
*
* pztStop - Stops current pzt motion
*
* Writes a velicity of zero to the velocity register to stop the 
* current pzt motion.
*
RETURNS:
OK
*/
//ERROR_CODE tPztMotor::pztStop(int whichPzt)
//{
//  
//
//  /*
//  [ this is invalid -> UNDEFINED
//  | this->bInit == FALSE -> UNDEFINED
//  | true -> NOP
//  ] */
//#ifdef DEBUG
//  if(!this->bInit) {
//    return S_pztMotor_BOARD_NOT_INITIALIZED;
//  }
//#endif
//
//  /*
//  [ whichPzt<1 || whichPzt>this->iNumPztChannels ->
//    UNDEFINED
//  | true -> NOP
//  ] */
//#ifdef DEBUG
//  if (whichPzt < 1 || whichPzt > iNumPztChannels) {
//    return S_pztMotor_INVALID_PZT;
//  }
//#endif
//
//  /*
//  [ ZERO_VELOCITY<MIN_VELOCITY || ZERO_VELOCITY>MAX_VELOCITY ->
//      UNDEFINED
//  | true -> NOP
//  ] */
//#ifdef DEBUG
//  if (ZERO_VELOCITY < MIN_VELOCITY || ZERO_VELOCITY > MAX_VELOCITY) {
//    return S_pztMotor_INVALID_VELOCITY;
//  }
//#endif
// 
//  /* 
//  [ true -> this->*pztVelocity[whichPzt] = ZERO_VELOCITY
//    : return OK
//  ] */
//  *pztVelocity[whichPzt] = ZERO_VELOCITY;
//  return OK;
//}

