/* Description:
 * This file contains the TIM class methods.  The TIM board is a JPL-designed
 * card containing an Altera FPGA which contains interrupt and clock capabilities.
 */

/* $Log: tim.cc,v $
 * Revision 1.9  1999/05/04 20:57:12  theo
 * Clock fixed, jitter still broken
 *
 * Revision 1.5  1999/03/30 20:23:20  chara
 * Some bug fixes from Phil and using external clock
 *
 * Revision 1.4  1999/03/29 20:03:43  theo
 * Got external clock working
 *
 * Revision 1.3  1999/01/12 21:42:49  chara
 * Updated tim.cc module from Phil, should fix time set probelm
 *
 * Revision 1.2  1998/08/17 21:01:08  phil
 * Laser Ignore, Laser board works, Modulation works
 *
 * Revision 1.1.1.1  1998/06/11 21:36:34  chara
 * First import of RICST Code
 *
 * Revision 1.1.2.4  1998/05/15 19:02:18  tgl
 * Pass CSR to channel routines to allow resetting sync.
 * Calculate time divider value to match actual board behavior.
 * Reserve and use channels 0 and 1 for board-level sync functions.
 * Implement and use new SYNC_INTERNAL channel timing mode.
 * Set (unused) fields in underlying channel class including bit width.
 * Implement channel reset, enable and disable functions.
 * Fix calculation of minimum allowed rate for channel outputs.
 *
 * Revision 1.1.2.2  1998/05/12 20:26:16  phil
 * Interrupt and tTimBoard changes for I5b
 *
 * Revision 1.1.2.1  1998/05/08 01:21:30  phil
 * interrupt functionality for I5b
 *
 * Jan 12 99, Fix by Phil
 * */

/*
 * The following line added by THEO to start External Clock
 */

#define EXTERNAL_SYNC

#include "isit.h"

#ifndef  SPECTOOL
#include <sysLib.h>
#include <iv.h>
#include <intLib.h>
#endif

#include "hwTypesP.h"
#include "hardware.h"
#include "board.h"
#include "tim.h"
#include "hwError.h"

/*********************************************************************
*
* tTimBoard::tTimBoard - TIM card constructor
*
  [[U tTimBoard::tTimBoard(iBoardNum)]]
*
*/
tTimBoard::tTimBoard(int iBoardNum)
{
  TraceMark(S_tim_ENTERING_CONSTRUCTOR);

  switch(iBoardNum) {
   case 0:
    this->vmeAddress = tVmeAddress(VME_AM_STD_SUP_DATA, (uint8_t *) TIM_BOARD_ADDRESS_0);
    break;
   case 1:
    this->vmeAddress = tVmeAddress(VME_AM_STD_SUP_DATA, (uint8_t *) TIM_BOARD_ADDRESS_1);
    break;
   default:
    this->vmeAddress = tVmeAddress();
    break;
  }
  this->boardName  = TIM_BOARD_NAME + iBoardNum;

  TraceMark(S_tim_DONE_CONSTRUCTING);
}


/*********************************************************************
*
* tTimBoard::SoftInit - TIM card Soft Initialize method
*
  [[U tTimBoard::SoftInit()]]
*
*/
ERROR_CODE tTimBoard::SoftInit(void)
{
  ERROR_CODE ec;
  volatile uint32_t   *ITRbase, *ICRbase, *CDRbase;
  int        i;

  TraceMark(S_tim_ENTERING_SOFTINIT);

  ec = this->Initialize();
  if(ec != OK) {
    return Traceback(ec, "tTimBoard::SoftInit");
  }

  this->bfFunctions = BF_CLOCK | BF_INTERRUPT;
  this->irMasterClockRate = (TIM_MASTER_CLOCK_SPEED);

  this->vCSR          = (uint32_t *) (vu8LocalAddress + 0x00);
  this->vRolloverReg  = (uint32_t *) (vu8LocalAddress + 0x04);
  this->vTimeStampReg = (uint32_t *) (vu8LocalAddress + 0x08);
  this->vTimeDivReg   = (uint32_t *) (vu8LocalAddress + 0x0C);
  this->vWatchdogReg  = (uint32_t *) (vu8LocalAddress + 0x10);
  this->vTimeSetReg   = (uint32_t *) (vu8LocalAddress + 0x14);

  ITRbase = (uint32_t *) (vu8LocalAddress + 0x20);
  ICRbase = (uint32_t *) (vu8LocalAddress + 0x40);
  CDRbase = (uint32_t *) (vu8LocalAddress + 0x80);

  for(i = 0; i < this->CiMaxChannels; i++) {
    this->ChannelArray[i].Initialize(this->vCSR,
                                     (CDRbase + i),
                                     this->irMasterClockRate);
  }

  /*
   * Interrupts are mapped to channels 1-7, so offset by one...
   * Actually, better to use the maximum number defined for the class instead.
   * That is what we do below.
   */
  for(i = 0; i < this->CiMaxInterrupts; i++) {
    this->InterruptArray[i].Initialize((ICRbase + i + 1),
                                       (ITRbase + i + 1),
                                       &(this->ChannelArray[i+1]),
                                       this->irMasterClockRate);
  }

  /****
    [[U tBaseObject::SetInit(TRUE)]]
  ****/
  this->SetInit(TRUE);

  TraceMark(S_tim_INITIALIZED_SOFTLY);

  return OK;
}


/*********************************************************************
*
* tTimBoard::HardInit - TIM card Hard Initialize method
*
  [[U tTimBoard::HardInit()]]
*
*/
ERROR_CODE tTimBoard::HardInit(void)
{
  ERROR_CODE ec;
  uint32_t iDivider,
           iRollover;
  int      i;
  INT_RATE pSyncRate;

  TraceMark(S_tim_ENTERING_HARDINIT);
  /****
    [[U tTimBoard::SoftInit()]]
  ****/
  ec = this->SoftInit();
  if(ec != OK) {
    return Traceback(ec, "tTimBoard::HardInit");
  }

  iDivider = ((TIM_MASTER_CLOCK_SPEED / TIME_TICKS_PER_SECOND)-1);
  iRollover = (86400 * TIME_TICKS_PER_SECOND);

  *(this->vTimeDivReg)  = iDivider;
  *(this->vRolloverReg) = iRollover;
  *(this->vTimeSetReg)  = TIM_STARTUP_TIME;

  *(this->vCSR)         = BITMASK( TIM_ENABLE_OUTPUTS_V );
  this->bOutputEnabled  = TRUE;

  for(i = 0; i < this->CiMaxChannels; i++) {
    this->ChannelArray[i].Reset();
  }

  /*
   * Actually, better to use the maximum number defined for the class instead.
   * That is what we do below.
   */
  for(i = 0; i < this->CiMaxInterrupts; i++) {
    this->InterruptArray[i].Reset();
  }

// ***** set up to use channel 0 for internal sync    - eamck
// *****    (this should go in higher-level code later...)

  pSyncRate = (INT_RATE)(1);              // 1PPS 
  this->ChannelArray[0].RateSet(&pSyncRate);            
  this->ChannelArray[0].Enable();            

// ***** set up to use channel 1 for dither sync    - eamck
// *****    (this should go in higher-level code later...)

  pSyncRate = (INT_RATE)(500000);          // 500kHz 
  this->ChannelArray[1].RateSet(&pSyncRate);            
  this->ChannelArray[1].Enable();            

// ***** end set up channels 0 and 1  - eamck

#ifndef EXTERNAL_SYNC
/*
 * The following line sets up for an internal clock.
 */

  *(this->vCSR)  |=  BITMASK( TIM_SYNC_ONCE_V ) |
                     BITMASK( TIM_SYNC_INTERNAL_V );
#else

/* 
 * Use the following line instead of above to setup for an external clock 
 */

/* Brackets added by THEO */

  *(this->vCSR)  |=  (BITMASK( TIM_SYNC_ONCE_V ) | 
                     BITMASK( TIM_CLOCK_SELECT_V ) |
		     BITMASK(TIM_SYNC_ENABLE_V));
#endif

  TraceMark(S_tim_INITIALIZED_HARDLY);
  return OK;
}


/*********************************************************************
*
* tTimBoard::SetTime - Sets time for next second, and enables one-shot
*                      sync if sync is not already being used.
*                      Virtual method for tClockBoard.
*
  [[U tTimBoard::SetTime(iSeconds)]]
*
*/
ERROR_CODE tTimBoard::SetTime(int iSeconds)
{
  ERROR_CODE ec;
  const char* CsMethodName = "tTimBoard::SetTime";
  
  ec = this->SetTimeNextSecond((uint32_t) (iSeconds * TIME_TICKS_PER_SECOND));
  if(ec!=OK)     return Traceback(ec, CsMethodName);

  if( !(*(this->vCSR) & BITMASK(TIM_SYNC_ENABLE_V)) ) {
    ec = this->SetOneShotSync();
    if(ec!=OK)   return Traceback(ec, CsMethodName);
  }

  return OK;
}


/*********************************************************************
*
* tTimBoard::CreateChannel -  access to non-interrupt timing channels
*
  [[U tTimBoard::CreateChannel(iChannelToUse)]]
*
*/
tTimChannel *tTimBoard::CreateChannel(int iChannelToUse)
{
  if ((iChannelToUse < 0) || (iChannelToUse >= CiMaxChannels)) {
    TraceMsg(tmt_ERROR, S_tim_RETURNING_NULL_CHANNEL_POINTER,
             "Request for channel %d outside of allowed range [0-%d]",
             iChannelToUse, (CiMaxChannels-1));
    return NULL;
  }

  return (&(ChannelArray[iChannelToUse]));
}


/*********************************************************************
*
* tTimBoard::CreateInterrupt -  access to non-interrupt timing channels
*
  [[U tTimBoard::CreateInterrupt(iInterruptToUse)]]
*
*/
tTimInterrupt *tTimBoard::CreateInterrupt(int iInterruptToUse)
{
  if ((iInterruptToUse < 0) || (iInterruptToUse >= CiMaxInterrupts)) {
    TraceMsg(tmt_ERROR, S_tim_CREATE_CHANNEL_ATTEMPTED,
             "Request for channel %d outside of allowed range [0-%d]",
             iInterruptToUse, (CiMaxInterrupts-1));
    return NULL;
  }

  return (&(InterruptArray[iInterruptToUse]));
}


/*********************************************************************
*
* tTimBoard::SetExternalSync - Enables external sync function.
*
  [[U tTimBoard::SetExternalSync()]]
*
*/
ERROR_CODE tTimBoard::SetExternalSync(void)
{
  *(this->vCSR) |= BITMASK(TIM_SYNC_ENABLE_V);

  return OK;
}


/*********************************************************************
*
* tTimBoard::SetOneShotSync - Sets time for next second.
*
  [[U tTimBoard::SetOneShotSync()]]
*
*/
ERROR_CODE tTimBoard::SetOneShotSync(void)
{
  *(this->vCSR) |= BITMASK(TIM_SYNC_ENABLE_V) |
                   BITMASK(TIM_SYNC_ONCE_V);
  
  return OK;
}


/*********************************************************************
*
* tTimBoard::SetTimeNextSecond - Loads SetTime Register.
*
  [[U tTimBoard::SetTimeNextSecond(uiTime)]]
*
*/
ERROR_CODE tTimBoard::SetTimeNextSecond(uint32_t uiTime)
{
  ERROR_CODE ec;

  *(this->vTimeSetReg) = uiTime;

  return OK;
}


/*********************************************************************
*
* tTimChannel::Initialize - (re)set channel-specific parameters
*
  [[U tTimChannel::Initialize(volatile uint32_t *pCSRToUse,
                              volatile uint32_t *pInterruptCtlRegToUse,
                              INT_RATE irMasterClockRateToUse)]]
*
*/
ERROR_CODE tTimChannel::Initialize(volatile uint32_t *pCSRToUse,
                                   volatile uint32_t *pChannelCtlRegToUse,
                                   INT_RATE irMasterClockRateToUse)
{
  this->vCSR = pCSRToUse;
  this->vChannelCtlReg = pChannelCtlRegToUse;
  this->irMasterClockRate = irMasterClockRateToUse;

  this->iBitWidth     = this->CiBitWidth;
  this->dChannelRange = (((double) (1 << this->iBitWidth))-1);
  this->ccConvCode    = CC_NO_CONV_CODE;

  return OK;
}


/*********************************************************************
*
* tTimChannel::Reset - (re)set channel-specific parameters
*
  [[U tTimChannel::Reset()]]
*
*/
ERROR_CODE tTimChannel::Reset(void)
{
  *(this->vChannelCtlReg) = 0x00000000;

  return OK;
}


/*********************************************************************
*
* tTimChannel::RateSet - set rate of a clock-driven interrupt source
*
  [[U tTimChannel::RateSet(INT_RATE *pRate)]]
*
*/
ERROR_CODE tTimChannel::RateSet(INT_RATE *pRate)
{
  double dDivider;
  int    iDivider;
  
  if(!this->IsInit()) {
    return S_tim_CHANNEL_NOT_INITIALIZED;
  }

  dDivider = ((irMasterClockRate / (*pRate * 2)) - 1);
  if (dDivider < 0)
    iDivider = 0;
  else if (dDivider > INT_MAX)
    iDivider = INT_MAX;
  else
    iDivider = (int) dDivider;

  this->Output(iDivider);

  *pRate = this->irRate = (irMasterClockRate / (2*iDivider+1));
  
  return OK;
}

/*********************************************************************
*
* tTimChannel::RateGet - get current rate setting on TIM board
*
  [[U tTimChannel::RateGet()]]
*
*/
INT_RATE tTimChannel::RateGet(void)
{
  return this->irRate;
}

/*********************************************************************
*
* tTimChannel::Enable - Enable an interrupt
*
  [[U tTimChannel::Enable()]]
*
*/

ERROR_CODE tTimChannel::Enable(void)
{
  ERROR_CODE ec;

  *(this->vChannelCtlReg) = (((uint32_t) BITMASK(TIM_CHANNEL_ENABLE_V)) | this->iDivider);

  #ifdef EXTERNAL_SYNC
  /*
   * Problem found on the GSU system.  When using external clock, 
   * the sync_once command must be given after the last channel
   * is running.  This may not be the best place for this code,
   * but it seems the most thorough at this point   -PcI 4/2/1999
   **** WARNING: If Enable is called after the system is initialized,
   ****  the time clock will reset and screw everything up until the
   ****  next SetTime command is given.
   */

  TraceMsg(tmt_WARN, OK, "Setting CSR (0x%x) bits %d and %d",
           this->vCSR, TIM_SYNC_ONCE_V, TIM_SYNC_ENABLE_V);

  *(this->vCSR)  |=  BITMASK( TIM_SYNC_ONCE_V ) |
                     BITMASK( TIM_SYNC_ENABLE_V);
  #endif


  /* Get this channel time synced by rezeroing the timestamp register
   * Note that this will affect all other channels and interrupts!
   * Next rev of board will likely have channel zero reserved for sync...
   * - thomas 1998-05-13
   */
// testing rev of board with channel zero reserved for sync...
//  *(this->vCSR)             |= BITMASK(TIM_SYNC_ONCE_V);

  return OK;
}

 
/*********************************************************************
*
*tTimChannel::Disable - Disable an interrupt
*
  [[U tTimChannel::Disable()]]
*
*/
ERROR_CODE tTimChannel::Disable(void)
{
  ERROR_CODE ec;

  *(this->vChannelCtlReg) &= ~((uint32_t) BITMASK(TIM_CHANNEL_ENABLE_V));

  return OK;
}


/*********************************************************************
*
* tTimChannel::Output - set current rate on TIM board
*
  [[U tTimChannel::Output(int iDivider)]]
*
*/
ERROR_CODE tTimChannel::Output(int iDivider)
{
  this->iDivider = iDivider;
  *(this->vChannelCtlReg) = ((*(this->vChannelCtlReg) & BITMASK(TIM_CHANNEL_ENABLE_V)) | this->iDivider);

  return OK;
}


/*********************************************************************
*
* tTimInterrupt::Initialize - (re)set channel-specific parameters
*
  [[U tTimInterrupt::Initialize(volatile uint32_t *pInterruptCtlRegToUse,
                                volatile uint32_t *pInterruptTimeRegToUse,
                                tTimChannel       *pChannelToUse,
                                INT_RATE          irMasterClockRateToUse)]]
*
*/
ERROR_CODE tTimInterrupt::Initialize(volatile uint32_t *pInterruptCtlRegToUse,
                                     volatile uint32_t *pInterruptTimeRegToUse,
                                     tTimChannel       *pChannelToUse,
                                     INT_RATE          irMasterClockRateToUse)
{
  ERROR_CODE ec;
  double     dMinRateToUse;
  INT_RATE   irMinRateToUse;

  dMinRateToUse = (irMasterClockRateToUse / (2*(pow(2.0, 31.0)-1)+1));
  irMinRateToUse = (INT_RATE) ((dMinRateToUse >= 1)? dMinRateToUse: 1);

  /****
    [[U tPeriodicTimeInterrupt::Initialize(irMinRateToUse,
                                           irMasterClockRateToUse)]]
  ****/
  ec = tPeriodicTimeInterrupt::Initialize(irMinRateToUse,
                                          irMasterClockRateToUse);
  if(ec != OK) {
    return Traceback(ec, "tTimInterrupt::Initialize");
  }

  this->irMasterClockRate = irMasterClockRateToUse;
  this->pChannel          = pChannelToUse;
  this->vInterruptTimeReg = pInterruptTimeRegToUse;
  this->vInterruptCtlReg  = pInterruptCtlRegToUse;

  return OK;
}


/*********************************************************************
*
* tTimInterrupt::Reset - (re)set interrupt-specific parameters
*
  [[U tTimInterrupt::Reset()]]
*
*/
ERROR_CODE tTimInterrupt::Reset()
{
  /****
    [[U tTimChannel::Reset()]]
  ****/
  this->pChannel->Reset();
  *(this->vInterruptCtlReg) &= ~(BITMASK(TIM_INTERRUPT_ENABLE_V));

  return OK;
}


/*********************************************************************
*
* tTimInterrupt::RateSet - set rate of a clock-driven interrupt source
*
  [[U tTimInterrupt::RateSet(INT_RATE *pRate)]]
*
*/
ERROR_CODE tTimInterrupt::RateSet(INT_RATE *pRate)
{
  ERROR_CODE ec;
  double     dDivider;
  int        iDivider;
  
  /****
    [[U tTimChannel::RateSet(pRate)]]
  ****/
  ec = this->pChannel->RateSet(pRate);
  if(ec != OK) {
    return Traceback(ec, "tTimInterrupt::RateSet");
  }

  this->irRate = this->pChannel->RateGet();
  this->bRateSetYet = TRUE;

  return OK;
}


/*********************************************************************
*
* tTimInterrupt::Enable - Enable an interrupt
*
  [[U tTimInterrupt::Enable()]]
*
*/

ERROR_CODE tTimInterrupt::Enable(void)
{
  ERROR_CODE ec;

  /****
    [[U tPeriodicTimeInterrupt::Enable()]]
  ****/
  ec = tPeriodicTimeInterrupt::Enable();

  /****
    [[U tTimChannel::Enable()]]
  ****/
  this->pChannel->Enable();

  *(this->vInterruptCtlReg) = (((uint32_t) BITMASK(TIM_INTERRUPT_ENABLE_V)) | this->uiVMEVector);

  return OK;
}

 
/*********************************************************************
*
*tTimInterrupt::Disable - Disable an interrupt
*
  [[U tTimInterrupt::Disable()]]
*
*/
ERROR_CODE tTimInterrupt::Disable(void)
{
  ERROR_CODE ec;

  /****
    [[U tPeriodicTimeInterrupt::Disable()]]
  ****/
  ec = tPeriodicTimeInterrupt::Disable();

  *(this->vInterruptCtlReg) &= ~((uint32_t) BITMASK(TIM_INTERRUPT_ENABLE_V));

  return OK;
}


/*********************************************************************
*
*tTimInterrupt::GetTimeLastInt - Gets time of last interrupt in 25's of usec
*
  [[U tTimInterrupt::GetTimeLastInt()]]
*
*/
TIME tTimInterrupt::GetTimeLastInt()
{
  return ((TIME) *(this->vInterruptTimeReg));
}

/*********************************************************************
*
* tTimBoard::IsHbSynced - Returns TRUE if the Heartbeat is synced to 1PPS
*
  [[U tTimBoard::IsHbSynced(void)]]
*
*/
BOOL tTimBoard::IsHbSynced(void)
{
  if(*(this->vCSR) & BITMASK(TIM_SYNC_ENABLE_V)) {
    return TRUE;
  }
  else {
    return FALSE;
  }
}
