/* $RCSfile: laserLib.h,v $  -  header file for laserLib.c library */
/*
$Header: /home/chara/cvsroot/ople/sys/hw/include/laserLib.h,v 1.4 1998/09/13 19:43:46 phil Exp $
*/

/*
modification history
--------------------
02a,pci,23Oct96     Added C++ support for ISIT (removed non-Inc1 functions)
01f,beh,23May90     Added 64-bit stuff, locking, laserFullInit
01e,beh,09May90     Added function prototypes, changed from int to int32_t
01d,beh,05May90     Added LASER_ prefix to identifiers
01c,beh,19Feb90     Moved user-configuration info off to laserConfig.h
01b,beh,19Feb90     Put source code in standard format
01a,beh,24Jan90     written, debugged
*/

#ifndef INC_laserLibh
#define INC_laserLibh

#include "channel.h"


/***************************************************************
* Pointers to various addresses & registers on the laser card  *
* The various laser counters                                   *
*
*/
const int CiLaserTestModuleOffset   = 0x00;
const int CiLaserStatusModuleOffset = 0xf0;

/* Healthy clock values */
#define LASER_HEALTHY_REF     0x01
#define LASER_HEALTHY_2_MHZ   0x02
#define LASER_HEALTHY_16_MHZ  0x04
#define LASER_HEALTHY_64_MHZ  0x10
#define LASER_HEALTHY_EXT     0x20
#define LASER_HEALTHY_PLL     0x40
#define LASER_HEALTHY_CLOCKS  (LASER_HEALTHY_REF | LASER_HEALTHY_2_MHZ | \
                               LASER_HEALTHY_16_MHZ | LASER_HEALTHY_EXT | \
                               LASER_HEALTHY_PLL)

/* Testing constants */
#define LASER_LONG_TEST_POS 0x05a5a5a5
#define LASER_LONG_TEST_NEG 0xfa5a5a5a
#define LASER_FAST_TEST_POS 0x000005a5
#define LASER_FAST_TEST_NEG 0xfffffa5a
#define LASER_TEST_FRAC_BITS 0x7
#define LASER_OFFSET_FRAC 0x2

/******************************************************************
* 
* tFastLaserChannel Class - Fast Laser Board driver class.  Public of tBoard
*
* State Variables:
* ----------------
*
* 
* METHODS:  
[[D tFastLaserChannel::Initialize(volatile uint8_t* vu8BoardBase,
                                  volatile uint8_t* vu8ChannelBase,
                                  int               iChannelNumToUse)]
[ this is invalid || vu8BoardBase is invalid || vu8ChannelBase is invalid=>
    UNDEFINED
| true =>
  set class variables to corresponding parameter values
  : this->bInit := TRUE
  : return OK
]]
*
*
[[D tFastLaserChannel::LoadOnFiducial(void) ]
[ this is invalid || this->bInit == FALSE =>
  UNDEFINED
  | this->lcChannel.bOnFiducialR is never TRUE after =>
  this->lcChannel.smLaserSyncModeR := SYNC_TO_FIDUCIAL
  : this->lcChannel.bWaitingOnFiducialR := TRUE
  : return OK
| true =>
  this->lcChannel.dWholeFiducialR := this->lcChannel.dWholeCounterR
  : return OK
]]
*
*
[[D tFastLaserChannel::ClearCounterOnFiducial(void) ]
[ this is invalid || this->bInit == FALSE =>
  UNDEFINED
| this->lcChannel.bOnFiducialR is never TRUE  =>
  this->lcChannel.smLaserSyncModeR := SYNC_TO_FIDUCIAL
  : this->lcChannel.bWaitingOnFiducialR := TRUE
  : return OK
| true =>
  this->lcChannel.dWholeFiducialR := this->lcChannel.dWholeCounterR
  : this->lcChannel.dWholeCounterR := ~0 (when home sensor is crossed)
  : return OK
]]
*
*
[[D tFastLaserChannel::ResetFiducialWait(void) ]
[ this is invalid || this->bInit == FALSE =>
  UNDEFINED
| true =>
  this->lcChannel.bWaitingOnFiducialR := FALSE
  : return OK
]]
*
*
[[D tFastLaserChannel::FiducialWaitState(void) ]
[ this is invalid || this->bInit == FALSE =>
  UNDEFINED
| true =>
  return this->lcChannel.bWaitingOnFiducialR
]]
*
*
[[D tFastLaserChannel::IsFiducialOn(void) ]
[ this is invalid || this->bInit == FALSE =>
  UNDEFINED
| true =>
  return this->lcChannel.bOnFiducialR
]]
*
*
[[D tFastLaserChannel::TestModeEnable(void)]
[ this is invalid =>
    UNDEFINED
| *this.bInit == FALSE || counter is out of range =>
    return appropriate error
| true => set counter channel to enter test mode
  : return OK
]]
*
*
[[D tFastLaserChannel::TestModeDisable(void)]
[ this is invalid => UNDEFINED
| *this.bInit == FALSE || counter is out of range =>
    return appropriate error
| true => set counter channel to exit test mode (enter operational mode)
  : return OK
]]
*
*
[[D tFastLaserChannel::SetSyncMode(void)]
[ this is invalid => UNDEFINED
| *this.bInit == FALSE || counter is out of range ||
  newMode is out of range =>
    return appropriate error
| true => Set Sync mode of counter to newMode
  : return OK
]]
*
*
[[D tFastLaserChannel::ClearCounter(void)]
[ this is invalid => UNDEFINED
| *this.bInit == FALSE || counter is out of range => 
    UNDEFINED
| true => Clear specified laser counter
  : return OK
]]
*
*
[[D tFastLaserChannel::InputDouble(double* pdValue)]
[ this is invalid || pdValue is invalid => UNDEFINED
| *this.bInit == FALSE || pdValue == NULL =>
    UNDEFINED
|true =>
  set *pdValue to the value of the counter
  : return OK
]]
*
*
[[D tFastLaserChannel::InputInt(int* piValue)]
[ this is invalid || piValue is invalid => UNDEFINED
| *this.bInit == FALSE || piValue == NULL =>
    UNDEFINED
|true =>
  set *piValue to the value of the counter
  : return OK
]]
*
*
[[D tFastLaserChannel::UnknownStatus(void)]
[ this is invalid => UNDEFINED
| *this.bInit == FALSE || counter is out of range ||
  unknown signal is not present for counter => 
    return appropriate error
| true => return OK
]]
*
*
*/
class tFastLaserChannel : public tLaserMetrologyChannel {
protected:
  /***************************************************
  * tChannel base class member constants
  * The range units of this channel are wavelengths. *
  * Since there are 2^32 wavelengths of dynamic      *
  * range, the dChannelRange will be set during      *
  * initialization.                                  *
  */
  const int       CiBitWidth  = 40;
  const CONV_CODE CccConvCode = CC_TWOS_COMPLEMENT; 

  /* The Current Position Register(R), Fiducial Position Register(R),
  *  Mode(W), and Command(W) Offsets (from channel base)  */
  const int CiCprMswOffset  = 0x0;
  const int CiCprLswOffset  = 0x4;
  const int CiFprMswOffset  = 0x8;
  const int CiFprLswOffset  = 0xc;
  const int CiModeOffset    =  0x0;
  const int CiCommandOffset =  0x8;
  /* Offsets (from board base) for fiducial states */
  const int CiFidWaitStatesOffset = 0x6 + CiLaserStatusModuleOffset;
  const int CiFidStatesOffset     = 0xe + CiLaserStatusModuleOffset;
  const int CiUnknownStatusOffset = 0x6 + CiLaserTestModuleOffset;

  /* Various values for command and control bits */
  /* Mode word:    */
  const int CiModeLowestValue        =  0x0;
  const int CiModeCprFlowthrough     =  0x0;
  const int CiModeCprSyncToRef       =  0x1;
  const int CiModeCprSyncToExt       =  0x2;
  const int CiModeFprLoadOnFiducial  =  0x3;
  const int CiModeHighestValue       =  0x3;
  /* Clear counter command -- meaning of various modes for clearing
   * the counters       */
  const int CiClearImmediate  = CiModeCprFlowthrough;
  const int CiClearOnRef      = CiModeCprSyncToRef;
  const int CiClearOnExt      = CiModeCprSyncToExt;
  const int CiClearOnFiducial = CiModeFprLoadOnFiducial;   
  /* Command Word  */
  const int CiCmdLeaveTestMode       =  0x0;
  const int CiCmdEnterTestMode       =  0x1;
  const int CiCmdClearCounter        =  0x2;
  const int CiCmdResetWaitStates     =  0x3;

  /* Default Values */
  const int CiDefaultSyncMode = CiModeCprSyncToExt;

  /* Channel Register pointers */  
  volatile int32_t *viCprHi;
  volatile int32_t *viFprHi;
  volatile int32_t *viCprLo;
  volatile int32_t *viFprLo;
  volatile int16_t *vwUnknownStatus;
  volatile int32_t *viCommand;
  volatile int32_t *viMode;
  /* Board Register pointers that contain channel info */
  volatile int16_t *vwFidWaitStates;
  volatile int16_t *vwFidStates;

  // For channels that are used in pairs  -- PCI 7/15/98
  tFastLaserChannel *pRefChannel;
  int               iChannelNum;

public :
  tFastLaserChannel(void);
  ERROR_CODE Initialize(volatile uint8_t *vu8BoardBase,
                        volatile uint8_t *vu8ChannelBase,
                        int              iChannelNumToUse);
  ERROR_CODE Initialize(volatile uint8_t *vu8BoardBase,
                        volatile uint8_t *vu8ChannelBase,
                        tFastLaserChannel *RefChannelToUse,
                        int              iChannelNumToUse);
  ERROR_CODE Input(int *piValue) const;
  ERROR_CODE Input(double *pdValue) const;
//  ERROR_CODE ReadCounter(int64_t *p64Value) const;
//  ERROR_CODE ReadFiducial(int64_t *p64Value) const;
  ERROR_CODE TestModeEnable(void);
  ERROR_CODE TestModeDisable(void);
  ERROR_CODE SetSyncMode(int newMode=CiDefaultSyncMode);
  ERROR_CODE ClearCounter(void);
  int GetSyncMode(void) { return *viMode; }
  ERROR_CODE LoadOnFiducial(void);
  ERROR_CODE ClearCounterOnFiducial(void);
  ERROR_CODE ResetFiducialWait(void);
  BOOL FiducialWaitState(void);
  BOOL IsFiducialOn(void);
  BOOL UnknownStatus(void);
  
/*
  STATUS FastTest(int counter);
  STATUS LongTest(int counter);
  ERROR_CODE CprHi(int counter,int32_t *high);
  ERROR_CODE CprLo(int counter,int32_t *low);
  ERROR_CODE CprInt(int counter, int32_t *ival);
  ERROR_CODE CprFrac(int counter,int32_t *fval);
  ERROR_CODE Fpr(int counter, Qint16_t *contents);
  ERROR_CODE FprHi(int counter, int32_t *high);
  ERROR_CODE FprLo(int counter, int32_t *low);
  ERROR_CODE FprInt(int counter, int32_t *ival);
  ERROR_CODE FprFrac(int counter,int32_t *fval);
*/

};


/******************************************************************
* 
* tFastLaserBoard Class - Fast Laser Board driver class.  Public of tBoard
*
* State Variables:
* ----------------
*
* METHODS:
[[D tFastLaserBoard::tFastLaserBoard(void)]
[ this is invalid => UNDEFINED
| true => set all register pointer variables to NULL
  : *this.bInit := FALSE
]]
*
*
[[D tFastLaserBoard::SoftInit(int numCounters)]
[this is invalid || vmeBoardAddress is invalid => UNDEFINED
| vmeBoardAddress not reachable by CPU =>
    return VxWorks error code from sysBusToLocalAdrs
| numCounters is not in range of board =>
    return S_laser_INIT_INVALID_ARGUMENT
| *this.bInit == TRUE => return S_laser_BOARD_ALREADY_INITIALIZED
| true => *this.vmeAddress := vmeBoardAddress
  : *this.v8LocalAddress := local address corresponding to vmeBoardAddress
  : *this.boardName := boardName
  : *this.bInit := TRUE
  : initialize board register pointers and protected variables
  : return OK
]]
*
*
[[D tFastLaserBoard::HardInit(int numCounters)]
[this is invalid || vmeBoardAddress is invalid =>
    UNDEFINED
| vmeBoardAddress not reachable by CPU =>
    return VxWorks error code from sysBusToLocalAdrs
| *this.bInit == TRUE => return S_laser_BOARD_ALREADY_INITIALIZED
| board clock signals are not healthy || channel unknowns are not present ||
  numCounters is not in range of board => print error message on console
         and proceed with true case
| true => *this.vmeAddress := vmeBoardAddress
  : *this.v8LocalAddress := local address corresponding to vmeBoardAddress
  : *this.boardName := boardName
  : *this.bInit := TRUE
  : Initialize board register pointers and protected variables
  : Set all channels to exit test mode (enter operational mode)
  : Set Sync mode to external sync for all installed counters
  : return OK
]]
*
* ERROR_CODE tFastLaserBoard::GetChannel(iChannelNum)
* Description:  Gets channel specified by iChannelNum.
*        Fast Laser Global board numbering:
*     tFastLaserMetrologyChannel[0-12] ==> Global Channels 0-11
*     FiducialInputChannel[0..12]      ==> Global Channels 12-23
*
[[D tFastLaserBoard::GetChannel(int iChannelNum)]
[ this is invalid => UNDEFINED
| this->bInit == FALSE =>
  return := NULL
| iChannelNum<0 || iChannelNum>Max Global Channel ||
  iChannelNum is not valid for current configuration =>
  return := NULL
| true =>
  return := D/A channel corresponding to iChannelNum
]]
*
*
[[D tFastLaserBoard::CheckClocks(void)]
[ this is invalid =>
  UNDEFINED
| *this.bInit == FALSE || no reference signal || PLL is out of lock ||
  at least one board clock is not present || external clock not present ||
  *this.*clockStatus has invalid value =>
    return appropriate error
| true => return OK
]]
*
*
[[D tFastLaserChannel::TestModeEnable(void)]
[ this is invalid =>
    UNDEFINED
| *this.bInit == FALSE || counter is out of range =>
    return appropriate error
| true => set counter channel to enter test mode
  : return OK
]]
*
*
[[D tFastLaserChannel::TestModeDisable(void)]
[ this is invalid =>
    UNDEFINED
| *this.bInit == FALSE || counter is out of range =>
    return appropriate error
| true => set counter channel to exit test mode (enter operational mode)
  : return OK
]]
*
*
[[D tFastLaserChannel::SetSyncMode(int newMode)]
[ this is invalid =>
   UNDEFINED
| *this.bInit == FALSE || counter is out of range ||
  newMode is out of range =>
    return appropriate error
| true => Set Sync mode of counter to newMode
  : return OK
]]
*
*
[[D tFastLaserChannel::ClearCounter(void)]
[ this is invalid || *this.bInit == FALSE || counter is out of range => 
    UNDEFINED
| true => Clear specified laser counter
  : return OK
]]
*
*
[[D tFastLaserChannel::InputDouble(double* pdValue)]
[ this is invalid || pdValue is invalid ||
  *this.bInit == FALSE || pdValue == NULL =>
    UNDEFINED
| true =>
  set *pdValue to the value of the counter
  : return OK
]]
*
*
[[D tFastLaserChannel::InputInt(int* piValue)]
[ this is invalid || piValue is invalid || *this.bInit == FALSE ||
  piValue == NULL =>
    UNDEFINED
| true =>
  set *piValue to the value of the counter
  : return OK
]]
*
*
[[D tFastLaserChannel::UnknownStatus(void)]
[ this is invalid => UNDEFINED
| *this.bInit == FALSE || counter is out of range ||
  unknown signal is not present for counter => 
    return appropriate error
| true => return OK
]]
*
*
*/
class tFastLaserBoard : public tBoard {
public:
  tFastLaserBoard(int iBoardNum);
  ERROR_CODE SoftInit(int iNumCountersToUse);
  ERROR_CODE SoftInit(void);
  ERROR_CODE HardInit(int iNumCountersToUse);
  ERROR_CODE HardInit(void);
  int GetRefChannelNum(int iUnkChannelNum);
  tChannel *CreateChannel(int iChannelNum);
  int GetChannelOffset(int iNum) { return ((iNum+1)<<4); }
  ERROR_CODE CheckClocks(void);
  int TestSignalStatus(void);
  BOOL TestDone(void);
  void SetTestCount(int numCounts); 
/*
  void Read64Unlocked(volatile int32_t *fromAddress, int32_t *p64bitNum);
  volatile int32_t *CprAddr(int counter);
  volatile int32_t *FprAddr(int counter);
  void TestWait(void);
  STATUS FastTestAll(void);
  STATUS LongTestAll(void);
*/

protected:
  const int CiMaxNumChannels     = 24;
  const int CiMaxNumCounters     = 12;
  const int CiDefaultNumCounters = 4;
  static const int iChannelPair2DArray[][];

  /* Test and status module registers */
  /* Status registers: For reading    */
  const int CiTestSignalStatusIntOffset     = 0x4 + CiLaserTestModuleOffset;
  const int CiTestSignalStatusFracIntOffset = 0xc + CiLaserTestModuleOffset;
  const int CiClockStatusOffset             = 0xe + CiLaserTestModuleOffset;
  const int CiSyncModeHiOffset              = 0x4 + CiLaserStatusModuleOffset;
  const int CiSyncModeLoOffset              = 0xc + CiLaserStatusModuleOffset;

  /* Test module control registers: for writing */
  const int CiTestCntHiOffset               = 0x6 + CiLaserStatusModuleOffset;
  const int CiTestCntLoOffset               = 0xe + CiLaserStatusModuleOffset;
  const int CiTestSigStatusHiOffset         = 0x4 + CiLaserTestModuleOffset;
  const int CiTestSigStatusLoOffset         = 0xc + CiLaserTestModuleOffset;
  
  volatile int32_t  *viTestSigStatusInt;
  volatile int32_t  *viTestSigStatusFracInt;
  volatile int16_t  *vwClockStatus;
  volatile int32_t  *viSyncModeHi;
  volatile int32_t  *viSyncModeLo;
  volatile uint16_t *vuwTestCntHi;
  volatile uint16_t *vuwTestCntLo;
  volatile int32_t  *viTestSigStatHi;
  volatile int32_t  *viTestSigStatLo;
  tFastLaserChannel ChannelArray[CiMaxNumCounters];
  tFiducialInputChannel FidChannelArray[CiMaxNumCounters];
  int               iNumActiveCounters;
/*  int              laserFailedTest;  */
};

#endif /* INC_laserLibh */

