/* $RCSfile: tim.h,v $ - Hardware manager classes */

/* $Header: /home/chara/cvsroot/ople/sys/hw/include/tim.h,v 1.2 1998/08/17 21:01:10 phil Exp $ */

/* Description:
 * This file contains the TIM card class definition.
 * The TIM card is a timing and interrupt generating board
 *  designed by Phil Irwin and Beth McKenney for the RICST and MAM
 *  testbeds.
 * The tTimBoard class is a subclass of tBoard and is used mostly
 *  by the hardware manager.
 */

/* $Log: tim.h,v $
 * Revision 1.2  1998/08/17 21:01:10  phil
 * Laser Ignore, Laser board works, Modulation works
 *
 * Revision 1.1.1.1  1998/06/11 21:36:35  chara
 * First import of RICST Code
 *
 * Revision 1.1.2.4  1998/05/15 19:57:53  tgl
 * Add new CSR bit definitions.
 * Include spec for interrupt Reset().
 * Remove spec for interrupt Connect() and Disconnect()
 *  since this is now done from base classes.
 * Add spec for channel Reset(), Enable() and Disable().
 * Add bit width constant to channel class.
 * Add pointer to board CSR for channel class.
 *
 * Revision 1.1.2.3  1998/05/12 22:59:06  phil
 * changes before build_5b_001
 *
 * Revision 1.1.2.2  1998/05/12 20:26:17  phil
 * Interrupt and tTimBoard changes for I5b
 *
 * Revision 1.1.2.1  1998/05/08 01:20:55  phil
 * interrupt functionality for I5b
 * */

#ifndef INC_TIM_H
#define INC_TIM_H

#include <limits.h>
#include "interrupt.h"

/*
 * Board configuration
 */
#define TIM_BOARD_ADDRESS_0 (0x0220000)
#define TIM_BOARD_ADDRESS_1 (0x0220100)

/*
 * Fundamental configuration
 */
#define TIM_MASTER_CLOCK_SPEED 16000000
#define TIM_EXTERNAL_SYNC      FALSE
#define TIM_MAX_CHANNELS       8
#define TIM_MAX_INTERRUPTS     7

/*
 * CSR status/control bits
 */
#define TIM_CLOCK_SELECT_V      0
#define TIM_SYNC_ENABLE_V       1
#define TIM_SYNC_ONCE_V         2
#define TIM_SYNC_INTERNAL_V     3     // eamck
#define TIM_ENABLE_OUTPUTS_V    4
#define TIM_WD_ASSERTED_V       5
#define TIM_ERROR_V             6     // eamck

#define TIM_CHANNEL_ENABLE_V   31
#define TIM_INTERRUPT_ENABLE_V 15
#define TIM_STARTUP_TIME        0

#define BITMASK(v) (1 << (v))


/****************************************
*
* class tTimInterrupt : public tInterrupt
*
* State Variables:
*
* Methods:
* void tTimInterrupt::Initialize(...)
* Description:
*
[[D tTimInterrupt::Initialize(volatile uint32_t *pInterruptCtlRegToUse,
                              volatile uint32_t *pInterruptTimeRegToUse,
                              tTimChannel       *pChannelToUse,
                              INT_RATE          irMasterClockRateToUse)]
[ invalid_pointer(this) ||
  invalid_pointer(pInterruptCtlRegToUse) ||
  invalid_pointer(pInterruptTimeRegToUse) ||
  invalid_pointer(pChannelToUse) ||
  irMasterClockRateToUse out of range ||
  pInterruptCtlRegToUse does not point to valid timing board register ||
  pInterruptTimeRegToUse does not point to valid timing board register => UNDEFINED
| true =>
    set pointers and constants for board
    : tBaseObject::ecInitError = OK
    : tBaseObject::bInit = TRUE
]]
*
*
* ERROR_CODE tTimInterrupt::Reset(...)
* Description:  Reset board interrupt and interrupt handler
*
[[D tTimInterrupt::Reset()]
[ invalid_pointer(this) => UNDEFINED
| this->bInit == FALSE =>
  return S_tim_BOARD_NOT_INITIALIZED
| More than one CPU responds to this interrupt => UNDEFINED
| true =>
  reset board registers to disable interrupt
  : this->bEnabled := FALSE
  : return OK
]]
*
*
* ERROR_CODE tTimInterrupt::RateSet(...)
* Description:  Sets interrupt rate.
*
[[D tTimInterrupt::RateSet(INT_RATE *pRate)]
[ invalid_pointer(this) => UNDEFINED
| this->bInit == FALSE ||
  pRate is invalid => UNDEFINED
| *pRate > highest possible rate =>
  *pRate := highest possible rate for *this
  : this->pRate := highest possible rate for *this
  : call TraceMsg(S_tim_ADJUSTED_CLOCK_RATE)
  : return OK
| *pRate in range but not in set of allowed values => 
  *pRate := valid rate nearest and higher to *pRate
  : this->pRate := valid rate nearest and higher to *pRate
  : call TraceMsg(S_tim_ADJUSTED_CLOCK_RATE)
  : return OK
| true =>
  this->pRate := *pRate
  :  return OK 
]]
*
*
* ERROR_CODE tTimInterrupt::RateGet(...)
* Description:  Gets current interrupt rate.
*
[[D tTimInterrupt::RateGet()]
[ invalid_pointer(this) => UNDEFINED
| this->bInit == FALSE => UNDEFINED
| this->pRate is out of sync with actual rate =>
  return incorrect result
| true =>
  return current rate
]]
*
*
* ERROR_CODE tTimInterrupt::Enable(...)
* Description:  Enables board interrupt and interrupt handler
*
[[D tTimInterrupt::Enable()]
[ invalid_pointer(this) => UNDEFINED
| this->bInit == FALSE =>
  return S_tim_BOARD_NOT_INITIALIZED
| More than one CPU responds to this interrupt => UNDEFINED
| true =>
  enable interrupt
  : Begin sending interrupt number this->puwIntNum to the
      VME bus.  The ISR on the CPU associated with this
      will be called at a rate of this->pRate
      until a Disable command is given.
  : this->bEnabled := TRUE
  : return OK
]]
*
*
* ERROR_CODE tTimInterrupt::Disable(...)
* Description:  Disables board interrupt and interrupt handler
*
[[D tTimInterrupt::Disable()]
[ invalid_pointer(this) => UNDEFINED
| this->bInit == FALSE =>
  return S_tim_BOARD_NOT_INITIALIZED
| true =>
  this->bEnabled := FALSE
  : disable interrupt this->interrupt[iIntNum]
  : Stop sending interrupt number this->puwIntNum
      to the VME bus. The ISR on the CPU connected to
      this->puwIntNum will no longer be called periodically
  : return OK
]]
*
*
* TIME tTimInterrupt::GetTimeLastInt(...)
* Description:  Retrieve last time interrupt was fired
*
[[D tTimInterrupt::GetTimeLastInt(void)]
[ invalid_pointer(this) => UNDEFINED
| this->bInit == FALSE =>
  return S_tim_CLOCK_NOT_INITIALIZED
| true =>
  return time of last interrupt in 25us since midnight
]]
*
*/


/****************************************
*
* class tTimChannel : public tClockChannel
*
* State Variables:
*
* Methods:
* void tTimChannel::Initialize(...)
* Description:
*
[[D tTimChannel::Initialize(volatile uint32_t *pCSRToUse,
                            volatile uint32_t *pChannelCtlRegToUse,
                            INT_RATE          irMasterClockToUse)]
[ invalid_pointer(this) ||
  invalid_pointer(pCSRToUse) ||
  invalid_pointer(pChannelCtlRegToUse) ||
  irMasterClockToUse out of range ||
  pCSRToUse does not point to valid timing board register => UNDEFINED
| true =>
    tBaseObject::bInit := FALSE
    : this->vmeAddress := default VME address for this board
    : tBaseObject::ecInitError = OK
    : tBaseObject::bInit = TRUE
]]
*
*
* ERROR_CODE tTimChannel::Reset(...)
* Description:  Reset channel configuration registers.
*
[[D tTimChannel::Reset()]
[ invalid_pointer(this) => UNDEFINED
| this->bInit == FALSE =>
  return S_tim_BOARD_NOT_INITIALIZED
| More than one CPU responds to this interrupt => UNDEFINED
| true =>
  reset board registers to disable channel
  : this->bEnabled := FALSE
  : return OK
]]
*
*
* ERROR_CODE tTimChannel::RateSet(...)
* Description:  Sets output rate for specified channel.
*
[[D tTimChannel::RateSet(INT_RATE *pRate)]
[ invalid_pointer(this) => UNDEFINED
| this->bInit == FALSE ||
  pRate is invalid => UNDEFINED
| *pRate > highest possible rate =>
  *pRate := highest possible rate for *this
  : this->pRate := highest possible rate for *this
  : call TraceMsg(S_tim_ADJUSTED_CLOCK_RATE)
  : return OK
| *pRate in range but not in set of allowed values => 
  *pRate := valid rate nearest and higher to *pRate
  : this->pRate := valid rate nearest and higher to *pRate
  : call TraceMsg(S_tim_ADJUSTED_CLOCK_RATE)
  : return OK
| true =>
  this->pRate := *pRate
  :  return OK 
]]
*
*
* ERROR_CODE tTimChannel::RateGet(...)
* Description:  Gets current interrupt rate for specified channel.
*
[[D tTimChannel::RateGet()]
[ invalid_pointer(this) => UNDEFINED
| this->bInit == FALSE => UNDEFINED
| this->pRate is out of sync with actual rate =>
  return incorrect result
| true =>
  return current rate
]]
*
*
* ERROR_CODE tTimChannel::Enable(...)
* Description:  Enables channel output.
*
[[D tTimChannel::Enable()]
[ invalid_pointer(this) => UNDEFINED
| this->bInit == FALSE =>
  return S_tim_BOARD_NOT_INITIALIZED
| More than one CPU responds to this interrupt => UNDEFINED
| true =>
  enable channel output
  : this->bEnabled := TRUE
  : return OK
]]
*
*
* ERROR_CODE tTimChannel::Disable(...)
* Description:  Disables channel output.
*
[[D tTimChannel::Disable()]
[ invalid_pointer(this) => UNDEFINED
| this->bInit == FALSE =>
  return S_tim_BOARD_NOT_INITIALIZED
| true =>
  disable channel output
  : this->bEnabled := FALSE
  : return OK
]]
*
*
* ERROR_CODE tTimChannel::Output(...)
* Description:  Sets current interrupt rate for specified channel.
*
[[D tTimChannel::Output(int iRate)]
[ invalid_pointer(this) => UNDEFINED
| this->bInit == FALSE => UNDEFINED
| true =>
  set output rate for this channel
  :  return OK 
]]
*
*
* ERROR_CODE tTimChannel::OutputRate(...)
* Description:  Sets current interrupt rate for specified channel.
*
[[D tTimChannel::OutputRate(INT_RATE *pRate)]
[ invalid_pointer(this) => UNDEFINED
| this->bInit == FALSE => UNDEFINED
| true =>
  set output rate for this channel
  :  return OK 
]]
*
*/


/****************************************
*
* class tTimBoard : public tBoard
*
* State Variables:
*
* Methods:
* void tTimBoard::tTimBoard(...)
* Description: Constructor - sets all default values for variables.
*
[[D tTimBoard::tTimBoard(int iBoardNum)]
[ invalid_pointer(this) => UNDEFINED
  | iBoardNum is out of range =>
      tBaseObject::bInit := FALSE
      : tBaseObject::ecInitError := S_Tim_INVALID_BOARD_NUM
  | true =>
      tBaseObject::bInit := FALSE
      : this->vmeAddress := default VME address for this board
      : this->boardName  := TIM_BOARD_NAME + iBoardNum
]]
*
*
* ERROR_CODE tTimBoard::SoftInit(...)
* Description:  Sets board addresses, initializes the pointers to board variables,
*
*               and sets class variables.
[[D tTimBoard::SoftInit(void)]
[ invalid_pointer(this) => UNDEFINED
| true =>
  this->bInit              := TRUE
  : set tBoard characteristics this->bfFunctions to include BF_CLOCK and BF_INTERRUPT
  : set this global board addresses to correct values
  : return OK
]]
*
*
* ERROR_CODE tTimBoard::HardInit(...)
* Description:  Sets board addresses, initializes the pointers to board variables,
*               sets class variables, and sets board to initial state.
*
[[D tTimBoard::HardInit(void)]
[ invalid_pointer(this) => UNDEFINED
| true =>
  this->bInit              := TRUE
  : set tBoard characteristics this->bfFunctions to include BF_CLOCK and BF_INTERRUPT
  : set this global board addresses to correct values
  : disable all interrupts and outputs from this board
  : set all configuration registers to a known state
  : return OK
]]
*
*
* ERROR_CODE tTimBoard::VerifyId(...)
* Description:  Verify presence of TIM board.
[[D tTimBoard::VerifyId()]
[ invalid_pointer(this) => UNDEFINED
| true =>
  return := board matches driver?
]]
*
*
* tTimChannel tTimBoard::CreateChannel(...)
* Description:  Gets channel specified by iChannelToUse.
*        TIM board channel numbered 0-31
*        TIM board interrupts reserved for channels 1-7
*
[[D tTimBoard::CreateChannel(int iChannelToUse)]
[ invalid_pointer(this) => UNDEFINED
| first interrupt channel is not the second clock channel => UNDEFINED
| (iChannelToUse < 0) || (iChannelToUse > this->CiMaxChannels) ||
  ((iChannelToUse >= 1) && (iChannelToUse <= this->CiMaxInterrupts)) =>
  return := NULL
| true =>
  return := clock channel corresponding to iChannelToUse
]]
*
*
* ERROR_CODE tTimBoard::CreateInterrupt(...)
* Description:  Gets interrupt specified by iInterruptToUse.
*        TIM board interrupts numbered 0-6
*        TIM board interrupts reserved for channels 1-7
*
[[D tTimBoard::CreateInterrupt(int iInterruptToUse)]
[ invalid_pointer(this) => UNDEFINED
| first interrupt channel is not the second clock channel => UNDEFINED
| (iInterruptToUse < 0) || (iInterruptToUse > this->CiMaxInterrupts) =>
  return := NULL
| true =>
  return := interrupt instance corresponding to iInterruptToUse
]]
*
*
* TIME tTimBoard::GetTimeNow(...)
* Description:  Retrieve current time
*
[[D tTimBoard::GetTimeNow()]
[ invalid_pointer(this) => UNDEFINED
| this->bInit == FALSE =>
  return S_tim_CLOCK_NOT_INITIALIZED
| no clock function available => UNDEFINED
| true =>
  return instantaneous time
]]
*
*
* BOOL tTimBoard::IsHbSynced(...)
* Description:  Determine if board has external 1PPS synchronization
*
[[D tTimBoard::IsHbSynced()]
[ invalid_pointer(this) => UNDEFINED
| this->bInit == FALSE =>
  return S_tim_BOARD_NOT_INITIALIZED
| *(this->CSR) indicates external sync is enabled =>
  return TRUE
| true =>
  return FALSE
]]
*
*/

class tTimChannel : public tClockChannel {
  public:
    ERROR_CODE      Initialize(volatile uint32_t *pCSRToUse,
                               volatile uint32_t *pChannelCtlRegToUse,
                               INT_RATE          irMasterClockRateToUse);
    ERROR_CODE      Reset(void);

    ERROR_CODE      RateSet(INT_RATE *pRate);
    INT_RATE        RateGet(void);
    ERROR_CODE      Enable(void);
    ERROR_CODE      Disable(void);
    ERROR_CODE      Output(int iRate);
    ERROR_CODE      Output(INT_RATE *pRate);

  protected:
    const int       CiBitWidth = 31;
    INT_RATE        irMasterClockRate;
    INT_RATE        irRate;
    uint32_t        iDivider;

  private:
    volatile uint32_t *vCSR;
    volatile uint32_t *vChannelCtlReg;  /* Channel Control Register */
};

class tTimInterrupt : public tPeriodicTimeInterrupt {
  public:
    ERROR_CODE      Initialize(volatile uint32_t *pInterruptCtlRegToUse,
                               volatile uint32_t *pInterruptTimeRegToUse,
                               tTimChannel       *pChannelToUse,
                               INT_RATE          irMasterClockToUse);
    ERROR_CODE      Reset(void);

    ERROR_CODE      RateSet(INT_RATE *pRate);
    ERROR_CODE      Connect(FUNCPTR      pFunc,
                            int          iArg,
                            unsigned int uiIntNum,
                            unsigned int uiIntVector);
    //ERROR_CODE      Disconnect(void);
    ERROR_CODE      Enable(void);
    ERROR_CODE      Disable(void);
    TIME            GetTimeLastInt(void);

  protected:
    tTimChannel     *pChannel;
    INT_RATE        irMasterClockRate;
    uint32_t        iDivider;

    volatile uint32_t *vInterruptTimeReg;  /* Interrupt Time Register */
    volatile uint32_t *vInterruptCtlReg;   /* Interrupt Control Register */
};

class tTimBoard : public tBoard {
  public:
    tTimBoard(int iBoardNum);

    virtual ERROR_CODE SoftInit(void);
    virtual ERROR_CODE HardInit(void);
    virtual ERROR_CODE SetTime(int iSeconds);
    ERROR_CODE         VerifyId(void);
    ERROR_CODE         ResetAllInterrupts(void);
    ERROR_CODE         ResetAllChannels(void);
    ERROR_CODE         SetExternalSync(void);
    ERROR_CODE         SetOneShotSync(void);
    ERROR_CODE         SetTimeNextSecond(uint32_t uiTime);
    tTimInterrupt      *CreateInterrupt(int iInterruptToUse);
    tTimChannel        *CreateChannel(int iChannelToUse);
    
    //TIME               GetTimeNow(void);

  private:
    const int       CiMaxChannels = TIM_MAX_CHANNELS;
    const int       CiMaxInterrupts = TIM_MAX_INTERRUPTS;

    BOOL            IsHbSynced(void);       /* checks if HB is synced to 1PPS */

    INT_RATE        irMasterClockRate;
    BOOL            bOutputEnabled;

    /* Board register address variables */
    volatile uint32_t *vCSR;
    volatile uint32_t *vRolloverReg;
    volatile uint32_t *vTimeStampReg;
    volatile uint32_t *vTimeDivReg;
    volatile uint32_t *vWatchdogReg;
    volatile uint32_t *vTimeSetReg;

    tTimChannel       ChannelArray[CiMaxChannels];
    tTimInterrupt     InterruptArray[CiMaxInterrupts];
};

#endif
