#include "isit.h"
#include <iostream.h>


/*****

AVL HYBRID TREE's 

This file contains an hybrid implementation of an Adel'son Vel'shii and Landis (AVL)
binary tree. This type of tree garantees O(log(N)) insertion, deletion, and 
search by maintaining the tree balance such that the difference in height 
between the left and right subtree is never greater than 1.

In addition to it's AVL-ness, this tree also supports linear scanning in both directions.
The result of the scan is to return a pointer to the current "scan" node value and move
the scan node to the next (greater or lesser) tree node.  The scan is circular so that 
once the maximum tree node is reached the next scanned node will be the minimum node and 
and vice versa.  Scan starting point can be set to the minimum, maximum, or root tree 
node.

This particular implementation  parameterizes both the indexing and data 
components of the tree.  This allows the user to specify any orderable data type
(types with well defined <,>, and == operations) as the index by which to sort
associated data.  Both index and data types  must have a well defined
copy contructor and = operator.

For example a set of employee records could ordered by employee id or by employee name.

tAvlHybridTree <EMPLOYEE_ID, tEmployeeRecord*> IdSortedTree;
tAvlHybridTree <EMPLOYEE_NAME, tEmployeeRecord*> NameSortedTree;

Note 1: The data parameter tEmployeeRecord* could also be tEmployeeRecord if it was 
desirable to have the tree own copies of all the employee records.

Note 2: Both the index and data type classes must have default constructors.
In the example above.  EMPLOYEE_ID is typedef of a primitive type so it is OK, and
tEmployeeRecord* is a pointer type so it has a default constructor as well. 
(If the parameter were tEmployeeRecord, then there must be a default (no arg) 
constructor for tEmployeeRecord)

This implementation also fully encapsulates the concept of a tree 
node such that the user of the tree never has access to a node, only to an 
instance of the parameterized type stored within the node. To insert a new 
tree node, the user only specifies the index and the value rather
than being responsible for creating a new tree node with those values.
To delete or search for an tree node the user specifies only the index. 

Example:

// Insertion:

IdSortedTree.Insert (24601, &JeanValJeanRecord)

// Search;
tEmployeeRecord* Record = IdSortedTree.Search (24601);

// Deletion:
IdSortedTree.Delete (24601);

// Scan

tEmployeeRecord* Record;

IdSortedTree.ResetScan (AVL_MIN); // start at minimum record

for (i = 0; i < iNumEmployees; i++ {
  Record = IdSortedTree.ScanForward (); // get record and set scan to next record
  if (Record->GoodPerformance () ) Record->GiveRaise (BIG_BUCKS); // do operation
}


************************ tAvlHybridTree Specification ******************************

[[D tAvlHybridTree::tAvlHybridTree (void) ]
Constructs AVL Hybrid tree with no nodes.
[ invalid_pointer (this) => UNDEFINED
| true => 
    this->Tree := empty AVL tree
    : this->List := empty doubly linked list
]]

[[D tAvlHybridTree::~tAvlHybridTree (void) ]
Destroys AVL hybrid tree freeing all resources consumed.
[ invalid_pointer (this) => UNDEFINED
| true => 
    this->Tree := empty AVL tree
    : this->List := empty doubly linked list
]]

[[D tAvlHybridTree::Insert (const TINDEX& NodeIndex, const TDATA& NodeValue) ]
Inserts new node with specified index and value into AVL tree if node with 
NodeIndex is not already in the tree. If insertion makes the tree unbalanced 
appropriate rotations are performed to balance the tree.
[ invalid_pointer (this) ||
  invalid_reference (NodeValue) ||
  invalid_tree (this->Tree) ||
  invalid_list (this->List) ||
  There is not a one to one mapping of nodes in this->Tree 
    and this->List => UNDEFINED
| Node with NodeIndex exists in this->Tree => no response
| true => 
    this->Tree := AVL tree with new node having NodeIndex and NodeValue
    : this->List := doubly linked list with new node inserted such that
      for all list nodes Node.PreviousNode.Index < Node.Index < Node.NextNode.Index
]]

[[D tAvlHybridTree::Delete (const TINDEX& NodeIndex) ]
Deletes node with specified index from AVL tree if it exists. 
If deletion makes the tree unbalanced appropriate 
rotations are performed to balance the tree.
[ invalid_pointer (this) ||
  invalid_reference (NodeValue) ||
  invalid_tree (this->Tree) ||
  invalid_list (this->List) ||
  There is not a one to one mapping of nodes in this->Tree 
    and this->List => UNDEFINED
| Node with NodeIndex does not exist in this->Tree => no response
| true => 
    this->Tree := AVL tree with node having NodeIndex deleted
    : this->List := doubly linked list with node having NodeIndex deleted
]]

[[D tAvlHybridTree::Search (const TINDEX& NodeIndex) ]
Search for node with specified index from AVL tree. If node exists returns
pointer to the node's data. If it does not exist, returns NULL.
[ invalid_pointer (this) ||
  invalid_reference (NodeValue) ||
  invalid_tree (this->Tree) => UNDEFINED
| true => return pointer to the data in node of this->Tree with index NodeIndex.
]]


[[D tAvlHybridTree::ScanForward (void) ]
Return pointer to the data of the current scan list element and moves to the next
element.
[ invalid_pointer (this) ||
  invalid_list (this->List) => UNDEFINED
| this->List is an empty list => return := NULL
| this->ScanNode == this->List.LastNode => 
    this->ScanNode = this->List.FirstNode
    : return pointer to the data in this->ScanNode
| true => 
    this->ScanNode = this->ScanNode.NextNode
    : return pointer to the data in this->ScanNode
]]



[[D tAvlHybridTree::ScanBackward (void) ]
Return pointer to the data of the current scan list element and moves to the next
element.
[ invalid_pointer (this) ||
  invalid_list (this->List) => UNDEFINED
| this->List is an empty list => return := NULL
| this->ScanNode == this->List.FirstNode => 
    this->ScanNode = this->List.LastNode
    : return pointer to the data in this->ScanNode
| true => 
    this->ScanNode = this->ScanNode.PreviousNode
    : return pointer to the data in this->ScanNode
]]

[[D tAvlHybridTree::ResetScan (AVL_SCAN_RESET ResetType) ]
Reset list scan setting the current scan node to either the mimimun, maximum, or
root node of the tree.
[ invalid_pointer (this) ||
  invalid_tree (this->Tree) ||
  invalid_list (this->List) ||
  There is not a one to one mapping of nodes in this->Tree 
    and this->List => UNDEFINED
| ResetType == AVL_MIN => this->ScanNode = node in this->List that is the minimum
    node in this->TREE
| ResetType == AVL_MAX => this->ScanNode = node in this->List that is the maximum
    node in this->TREE
| ResetType == AVL_ROOT => this->ScanNode = node in this->List that is the root
    node in this->TREE
| true => this->ScanNode := any node in this->List
]]


******************* Private Specification ****************************************



[[D tAvlHybridTree::Balance (tAvlHybridNode<TINDEX, TDATA>*& pNode) ]
Balances tree using right, left, right-left, or left-right rotations about pNode.
[ invalid_pointer (this, pNode) ||
  invalid_reference (pNode) ||
  invalid_tree (pNode) => UNDEFINED
| pNode->iBalance < -1 && 
  pNode->pSubTree [AVL_LEFT]->iBalance == 1 =>
    pNode := pNode->pSubTree [AVL_LEFT]
    : pNode->pSubTree [AVL_LEFT] := pNode->pSubTree [AVL_LEFT]->pSubTree [AVL_RIGHT]
    : pNode->pSubTree [AVL_LEFT]->pSubTree [AVL_RIGHT] := pNode
    : pNode->pSubTree [AVL_LEFT]->pSubTree [AVL_LEFT] := 
      pNode->pSubTree [AVL_LEFT]->pSubTree [AVL_LEFT]->pSubTree [AVL_RIGHT]
    : return := 1
| pNode->iBalance < -1 =>
    pNode := pNode->pSubTree [AVL_LEFT]
    : pNode->pSubTree [AVL_LEFT] := pNode->pSubTree [AVL_LEFT]->pSubTree [AVL_RIGHT]
    : pNode->pSubTree [AVL_LEFT]->pSubTree [AVL_RIGHT] := pNode
    : return := (pNode->pSubTree [AVL_LEFT]->iBalance == 0)
| pNode->iBalance > 1 && 
  pNode->pSubTree [AVL_LEFT]->iBalance == -1 =>
    pNode := pNode->pSubTree [AVL_RIGHT]
    : pNode->pSubTree [AVL_RIGHT] := pNode->pSubTree [AVL_RIGHT]->pSubTree [AVL_LEFT]
    : pNode->pSubTree [AVL_RIGHT]->pSubTree [AVL_LEFT] := pNode
    : pNode->pSubTree [AVL_RIGHT]->pSubTree [AVL_RIGHT] := 
      pNode->pSubTree [AVL_RIGHT]->pSubTree [AVL_RIGHT]->pSubTree [AVL_LEFT]
    : return := 1
| pNode->iBalance > 1 =>
    pNode := pNode->pSubTree [AVL_RIGHT]
    : pNode->pSubTree [AVL_RIGHT] := pNode->pSubTree [AVL_RIGHT]->pSubTree [AVL_LEFT]
    : pNode->pSubTree [AVL_RIGHT]->pSubTree [AVL_LEFT] := pNode
    : return := (pNode->pSubTree [AVL_RIGHT]->iBalance == 0)
| true => return := 0
]]  
  


[[D tAvlHybridTree::RotateRight (tAvlHybridNode<TINDEX, TDATA>*& pNode) ]
Performs single right rotation around pNode.
[ invalid_pointer (this, pNode) ||
  invalid_reference (pNode) ||
  invalid_tree (pNode) => UNDEFINED
| true =>
    pNode := pNode->pSubTree [AVL_LEFT]
    : pNode->pSubTree [AVL_LEFT] := pNode->pSubTree [AVL_LEFT]->pSubTree [AVL_RIGHT]
    : pNode->pSubTree [AVL_LEFT]->pSubTree [AVL_RIGHT] := pNode
    : return := (pNode->pSubTree [AVL_LEFT]->iBalance == 0)
]]


[[D tAvlHybridTree::RotateLeft (tAvlHybridNode<TINDEX, TDATA>*& pNode) ]
Performs single left rotation around pNode.
[ invalid_pointer (this, pNode) ||
  invalid_reference (pNode) ||
  invalid_tree (pNode) => UNDEFINED
| true =>
    pNode := pNode->pSubTree [AVL_RIGHT]
    : pNode->pSubTree [AVL_RIGHT] := pNode->pSubTree [AVL_RIGHT]->pSubTree [AVL_LEFT]
    : pNode->pSubTree [AVL_RIGHT]->pSubTree [AVL_LEFT] := pNode
    : return := (pNode->pSubTree [AVL_RIGHT]->iBalance == 0)
]]

[[D tAvlHybridTree::RotateRightLeft (tAvlHybridNode<TINDEX, TDATA>*& pNode) ]
Performs left rotation about pNode->pSubTree (AVL_RIGHT) and
right rotation about pNode.
[ invalid_pointer (this, pNode) ||
  invalid_reference (pNode) ||
  invalid_tree (pNode) => UNDEFINED
| true =>
    pNode := pNode->pSubTree [AVL_LEFT]
    : pNode->pSubTree [AVL_LEFT] := pNode->pSubTree [AVL_LEFT]->pSubTree [AVL_RIGHT]
    : pNode->pSubTree [AVL_LEFT]->pSubTree [AVL_RIGHT] := pNode
    : pNode->pSubTree [AVL_LEFT]->pSubTree [AVL_LEFT] := 
      pNode->pSubTree [AVL_LEFT]->pSubTree [AVL_LEFT]->pSubTree [AVL_RIGHT]
    : return := 1
]]


[[D tAvlHybridTree::RotateLeftRight (tAvlHybridNode<TINDEX, TDATA>*& pNode) ]
Performs right rotation about pNode->pSubTree (AVL_LEFT) and
left rotation about pNode.
[ invalid_pointer (this, pNode) ||
  invalid_reference (pNode) ||
  invalid_tree (pNode) => UNDEFINED
| true =>
    pNode := pNode->pSubTree [AVL_RIGHT]
    : pNode->pSubTree [AVL_RIGHT] := pNode->pSubTree [AVL_RIGHT]->pSubTree [AVL_LEFT]
    : pNode->pSubTree [AVL_RIGHT]->pSubTree [AVL_LEFT] := pNode
    : pNode->pSubTree [AVL_RIGHT]->pSubTree [AVL_RIGHT] := 
      pNode->pSubTree [AVL_RIGHT]->pSubTree [AVL_RIGHT]->pSubTree [AVL_LEFT]
    : return := 1
]]


[[D tAvlHybridTree::LeftHeavy (int iBalance) ]
Tests if the balance too left heavy.
[ invalid_pointer (this) => UNDEFINED 
| iBalance < -1 => return := TRUE
| true => return := FALSE
]]

[[D tAvlHybridTree::RightHeavy (int iBalance) ]
Tests if the balance too right heavy.
[ invalid_pointer (this) => UNDEFINED 
| iBalance > 1 => return := TRUE
| true => return := FALSE
]]

[[D tAvlHybridTree::Compare (const tAvlHybridNode<TINDEX, TDATA>* pNode, 
        const TINDEX& NodeIndex, AVL_COMPARE CompareType) ]
Compares pNode to NodeIndex and returns tree traversal direction based on CompareType.
AVL_EQUAL comparision traverses in search of the node with index NodeIndex.
AVL_MINIMUM traverses in search of the minimum node.
AVL_MAXIMUM traverses in search of the maximum node.
[ invalid_pointer (this, pNode) => UNDEFINED 
| CompareType == AVL_EQUAL  && pNode->Index == NodeIndex => return := AVL_NO_GO
| CompareType == AVL_EQUAL  && pNode->Index < NodeIndex => return := AVL_GO_RIGHT
| CompareType == AVL_EQUAL  && pNode->Index > NodeIndex => return := AVL_GO_LEFT
| CompareType == AVL_MINIMUM && pNode->pSubTree [AVL_LEFT] != NULL => 
    return := AVL_GO_LEFT
| CompareType == AVL_MAXIMUM && pNode->pSubTree [AVL_RIGHT] != NULL => 
    return := AVL_GO_RIGHT
| true => return := AVL_NO_GO
]]


********************** tAvlHybridNode Specification *******************************

tAvlHybridNode is a private object used only by tAvlHybridTree to form AVL tree nodes.

Index store the ordinal value associated with the data
Value stores the data portion of the node.
iBalance is the diference in height between the right and left subtree
  iBalance < 0 => tree is left heavy
  iBalance > 0 => tree is right heavy
  iBalance = 0 => tree is balanced
pSubTree [0] points to the Left sub-tree.
pSubTree [1] points to the Right sub-tree.

State Invariants (UNDEFINED results if conditions not true)

1) iBalance must always be the true diference between right and left subtree
   heights.

************************************************************

[[D tAvlHybridNode::tAvlHybridNode (const TINDEX& InitialIndex, const TDATA& InitialValue) ]
Constructs AVL tree node with initial value and no sub-trees.
[ invalid_pointer (this) ||
  invalid_reference (InitialValue) => UNDEFINED
| true => 
    this->Index := InitialIndex
    : this->Value := InitialValue
    : this->iBalance := 0
    : this->pSubTree [0] := NULL
    : this->pSubTree [1] := NULL
]]

[[D tAvlHybridNode::~tAvlHybridNode (void) ]
Destroys AVL tree node and all sub-tree nodes.
[ invalid_pointer (this) ||
  invalid_tree (this->pSubTree [0], this->pSubTree [1]) => UNDEFINED
| true => 
   *this->pSubTree [0] := empty tree
   : *this->pSubTree [1] := empty tree
   : *this := anything
]]

**********/

template <class TINDEX, class TDATA> class tAvlHybridNode {
  friend class tAvlHybridTree<TINDEX, TDATA>;
  private: 
    tAvlHybridNode (const TINDEX& InitialIndex, 
              const TDATA& InitialValue) : 
      Index (InitialIndex), Value (InitialValue), 
      iBalance (0) {
        this->pSubTree [0] = NULL;
        this->pSubTree [1] = NULL;
        this->pSubList [0] = NULL;
        this->pSubList [1] = NULL;
      }
    ~tAvlHybridNode () { delete this->pSubTree [0]; delete this->pSubTree [1];}
    friend ostream& operator<< (ostream& Stream, 
      const tAvlHybridNode<TINDEX, TDATA>* pNode);
    TINDEX Index;    
    TDATA Value;
    int iBalance;
    tAvlHybridNode<TINDEX, TDATA>* pSubTree [2];
    tAvlHybridNode<TINDEX, TDATA>* pSubList [2];
}; 

template <class TINDEX, class TDATA> ostream& operator<< (ostream& Stream, const tAvlHybridNode<TINDEX, TDATA>* pNode) 
{
  if (pNode == NULL) return Stream;
  else {
    Stream << pNode->pSubTree [0] ;
    Stream << pNode->Index << " " << pNode->Value << " " << pNode->iBalance ; 
    if (pNode->pSubList[0] == NULL) Stream << " NULL" ;
    else Stream << " " << pNode->pSubList[0]->Value;
    if (pNode->pSubList[1] == NULL) Stream << " NULL" << endl;
    else Stream << " " << pNode->pSubList[1]->Value << endl;
    Stream << pNode->pSubTree [1];
    return Stream;
  }
}



enum AVL_SCAN_RESET { AVL_MIN = -1, AVL_ROOT = 0, AVL_MAX = 1};
enum AVL_COMPARE {AVL_MINIMUM = -1, AVL_EQUAL = 0, AVL_MAXIMUM = 1};
enum AVL_COMPARE_RESULT { AVL_GO_LEFT = -1, AVL_NO_GO = 0, AVL_GO_RIGHT};

template <class TINDEX, class TDATA> class tAvlHybridTree {
  public:
    tAvlHybridTree () : pRoot (NULL), pMin (NULL), pMax (NULL), pScanNode (NULL) {}
    ~tAvlHybridTree () { delete this->pRoot;}
    void Insert (const TINDEX& NodeIndex, const TDATA& NodeValue);
    void Delete (const TINDEX& NodeIndex);
    TDATA* Search (const TINDEX& NodeIndex);
    TDATA* Max () { return &this->pMax->Value;}
    TDATA* Min () { return &this->pMin->Value;}
    TDATA* ScanForward ();
    TDATA* ScanBackward ();
    void ResetScan (AVL_SCAN_RESET ResetType);
    friend ostream& operator<< (ostream& Stream, const tAvlHybridTree<TINDEX, TDATA>& Tree) {
      return Stream << Tree.pRoot;
    }
    
        
  private:
    enum AVL_HEIGHT {AVL_HEIGHT_NO_CHANGE = 0, AVL_HEIGHT_CHANGE = 1};
    enum AVL_BALANCE { AVL_LEFT_HEAVY = -1, AVL_BALANCED = 0, 
      AVL_RIGHT_HEAVY = 1 };
    enum AVL_SUBTREE {AVL_LEFT = 0, AVL_RIGHT = 1};
    enum AVL_SUBLIST {AVL_PREV = 0, AVL_NEXT = 1};
    
    int Insert (tAvlHybridNode<TINDEX, TDATA>* pPrev, tAvlHybridNode<TINDEX, TDATA>* pNext,
                tAvlHybridNode<TINDEX, TDATA>*& pNode,
                const TINDEX& NodeIndex, const TDATA& NodeValue);
    tAvlHybridNode<TINDEX, TDATA> Delete (tAvlHybridNode<TINDEX, TDATA>*& pNode, const TINDEX& NodeIndex, 
                int& iHeightChange, AVL_COMPARE CompareType);
    TDATA* Search (tAvlHybridNode<TINDEX, TDATA>* pNode, const TINDEX& NodeIndex);
    int Balance (tAvlHybridNode<TINDEX, TDATA>*& pNode);
    int RotateRight (tAvlHybridNode<TINDEX, TDATA>*& pNode);
    int RotateLeft (tAvlHybridNode<TINDEX, TDATA>*& pNode);
    int RotateRightLeft (tAvlHybridNode<TINDEX, TDATA>*& pNode);
    int RotateLeftRight (tAvlHybridNode<TINDEX, TDATA>*& pNode);
    
    
    BOOL LeftHeavy (int iBalance) const { return iBalance < AVL_LEFT_HEAVY;}
    BOOL RightHeavy (int iBalance) const { return iBalance > AVL_RIGHT_HEAVY;}
    AVL_COMPARE_RESULT Compare (const tAvlHybridNode<TINDEX, TDATA>* pNode, 
        const TINDEX& NodeIndex, AVL_COMPARE CompareType) const;

    tAvlHybridNode<TINDEX, TDATA>* pRoot;
    tAvlHybridNode<TINDEX, TDATA>* pMin;
    tAvlHybridNode<TINDEX, TDATA>* pMax;
    tAvlHybridNode<TINDEX, TDATA>* pScanNode;
};


template <class TINDEX, class TDATA> void tAvlHybridTree<TINDEX, TDATA>::
  Insert (const TINDEX& NodeIndex, const TDATA& NodeValue)
{
  Insert (NULL, NULL, this->pRoot, NodeIndex, NodeValue);
} 

template <class TINDEX, class TDATA> void tAvlHybridTree<TINDEX, TDATA>::
  Delete (const TINDEX& NodeIndex) 
{
  int iHeightChange;
  
  this->Delete (this->pRoot, NodeIndex, iHeightChange, AVL_EQUAL);
}

template <class TINDEX, class TDATA> TDATA* tAvlHybridTree<TINDEX, TDATA>::
  Search (const TINDEX& NodeIndex)  
{
  return this->Search (this->pRoot, NodeIndex);
}

template <class TINDEX, class TDATA> TDATA* tAvlHybridTree<TINDEX, TDATA>::ScanForward ()
{
  TDATA* pData = &this->pScanNode->Value;
  
  if (this->pScanNode == this->pMax) this->pScanNode = this->pMin;
  else this->pScanNode = this->pScanNode->pSubList [AVL_NEXT];
  return pData;
}

template <class TINDEX, class TDATA> TDATA* tAvlHybridTree<TINDEX, TDATA>::ScanBackward ()
{
  TDATA* pData = &this->pScanNode->Value;
  
  if (this->pScanNode == this->pMin) this->pScanNode = this->pMax;
  else this->pScanNode = this->pScanNode->pSubList [AVL_PREV];
  return pData;
}

template <class TINDEX, class TDATA> void tAvlHybridTree<TINDEX, TDATA>::
  ResetScan (AVL_SCAN_RESET ResetType)
{
  switch (ResetType) {
    case AVL_MIN : this->pScanNode = this->pMin; break;
    case AVL_ROOT : this->pScanNode = this->pRoot; break;
    case AVL_MAX : this->pScanNode = this->pMax; break;
  }
}



template <class TINDEX, class TDATA> int tAvlHybridTree<TINDEX, TDATA>::
  Insert (tAvlHybridNode<TINDEX, TDATA>* pPrev, tAvlHybridNode<TINDEX, TDATA>* pNext,
          tAvlHybridNode<TINDEX, TDATA>*& pNode,
          const TINDEX& NodeIndex, const TDATA& NodeValue)
{
  int iIncrease = 0;

  if (pNode == NULL) {
    pNode = new tAvlHybridNode<TINDEX, TDATA> (NodeIndex, NodeValue);

    pNode->pSubList [AVL_PREV] = pPrev;
    pNode->pSubList [AVL_NEXT] = pNext;

    if (pPrev) pPrev->pSubList [AVL_NEXT] = pNode;
    else pMin = pNode;
   
    if (pNext) pNext->pSubList [AVL_PREV] = pNode;
    else pMax = pNode;
    
    return AVL_HEIGHT_CHANGE;
  }
  
  if (NodeIndex > pNode->Index) {
    iIncrease = this->Insert (pNode, pNode->pSubList [AVL_NEXT], 
                              pNode->pSubTree [AVL_RIGHT], NodeIndex, NodeValue);
  }
  else if (NodeIndex < pNode->Index) {
    iIncrease = -1 * this->Insert (pNode->pSubList [AVL_PREV], pNode, 
                                   pNode->pSubTree [AVL_LEFT], NodeIndex, NodeValue);
  }
  
  else iIncrease = AVL_HEIGHT_NO_CHANGE;
  
  pNode->iBalance += iIncrease;
  

  if (iIncrease != 0 || pNode->iBalance != 0)  {
    return 1 - this->Balance ( pNode);
  }
  
  else return iIncrease;
  
}


template <class TINDEX, class TDATA> tAvlHybridNode<TINDEX, TDATA> 
  tAvlHybridTree<TINDEX, TDATA>::Delete (tAvlHybridNode<TINDEX, TDATA>*& pNode, 
    const TINDEX& NodeIndex, int& iHeightChange, AVL_COMPARE CompareType)
{
 
 tAvlHybridNode<TINDEX, TDATA>* pNodeToDelete;
 int iDecrease;
 TINDEX FoundIndex;
 TDATA FoundValue;
 tAvlHybridNode <TINDEX, TDATA> FoundNode (FoundIndex, FoundValue);
  
 if (pNode == NULL) {
   iHeightChange = AVL_HEIGHT_NO_CHANGE;
   return  tAvlHybridNode <TINDEX, TDATA> (FoundIndex, FoundValue);
 }

 
 switch (this->Compare (pNode, NodeIndex, CompareType) ) {
   case AVL_GO_RIGHT :
     
     FoundNode = this->Delete (pNode->pSubTree [AVL_RIGHT], NodeIndex, 
       iHeightChange, CompareType);  
     iDecrease = iHeightChange;
     break;
   
   case AVL_GO_LEFT :
  
     FoundNode = this->Delete (pNode->pSubTree [AVL_LEFT], NodeIndex, 
       iHeightChange, CompareType);  
     iDecrease = -iHeightChange;
     break;
     
   case AVL_NO_GO :
 
     if ( pNode->pSubTree [AVL_LEFT] == NULL &&
          pNode->pSubTree [AVL_RIGHT] == NULL ) {

       if (pNode->pSubList [AVL_NEXT] != NULL)
         pNode->pSubList [AVL_NEXT]->pSubList [AVL_PREV] = 
           pNode->pSubList [AVL_PREV];
       if (pNode->pSubList [AVL_PREV] != NULL)
         pNode->pSubList [AVL_PREV]->pSubList [AVL_NEXT] = 
           pNode->pSubList [AVL_NEXT];
      
       FoundIndex = pNode->Index;
       FoundValue = pNode->Value;
       delete pNode;
       pNode = NULL;
       iHeightChange = AVL_HEIGHT_CHANGE;
       return tAvlHybridNode <TINDEX, TDATA> (FoundIndex, FoundValue);
     }
    
     else if ( pNode->pSubTree [AVL_LEFT] == NULL) {
       pNodeToDelete = pNode;
       pNode = pNode->pSubTree [AVL_RIGHT];
       pNodeToDelete->pSubTree [AVL_LEFT] = NULL;
       pNodeToDelete->pSubTree [AVL_RIGHT] = NULL;


       if (pNodeToDelete->pSubList [AVL_NEXT] != NULL)
         pNodeToDelete->pSubList [AVL_NEXT]->pSubList [AVL_PREV] = 
           pNodeToDelete->pSubList [AVL_PREV];
       if (pNodeToDelete->pSubList [AVL_PREV] != NULL)
         pNodeToDelete->pSubList [AVL_PREV]->pSubList [AVL_NEXT] = 
           pNodeToDelete->pSubList [AVL_NEXT];

       FoundIndex = pNode->Index;
       FoundValue = pNodeToDelete->Value;
       delete pNodeToDelete;
       iHeightChange = AVL_HEIGHT_CHANGE;
       return tAvlHybridNode <TINDEX, TDATA> (FoundIndex, FoundValue);
     }
    
     else if ( pNode->pSubTree [AVL_RIGHT] == NULL) {
       pNodeToDelete = pNode;
       pNode = pNode->pSubTree [AVL_LEFT];
       pNodeToDelete->pSubTree [AVL_LEFT] = NULL;
       pNodeToDelete->pSubTree [AVL_RIGHT] = NULL;
       
       if (pNodeToDelete->pSubList [AVL_NEXT] != NULL)
         pNodeToDelete->pSubList [AVL_NEXT]->pSubList [AVL_PREV] = 
           pNodeToDelete->pSubList [AVL_PREV];
       if (pNodeToDelete->pSubList [AVL_PREV] != NULL)
         pNodeToDelete->pSubList [AVL_PREV]->pSubList [AVL_NEXT] = 
           pNodeToDelete->pSubList [AVL_NEXT];
        
       FoundIndex = pNodeToDelete->Index;
       FoundValue = pNodeToDelete->Value;
       delete pNodeToDelete;
       iHeightChange = AVL_HEIGHT_CHANGE;
       return tAvlHybridNode <TINDEX, TDATA> (FoundIndex, FoundValue);
     }
    
     else { 
       FoundNode = this->Delete (pNode->pSubTree [AVL_RIGHT],
         NodeIndex, iDecrease, AVL_MINIMUM);  
       pNode->Index = FoundNode.Index;
       pNode->Value = FoundNode.Value;
     }
     break; 
  }
  
  pNode->iBalance -= iDecrease;  

  if (iDecrease != 0) {
    if (pNode->iBalance != 0) iHeightChange = this->Balance (pNode);
    else iHeightChange = AVL_HEIGHT_CHANGE;
  }
  else iHeightChange = AVL_HEIGHT_NO_CHANGE;

  return FoundNode;
}


template <class TINDEX, class TDATA> TDATA* tAvlHybridTree<TINDEX, TDATA>::
  Search (tAvlHybridNode<TINDEX, TDATA>* pNode, const TINDEX& NodeIndex) 
{
  if (pNode->Index > NodeIndex) 
    return this->Search (pNode->pSubTree [AVL_LEFT], NodeIndex);
  else if (pNode->Index < NodeIndex)
    return this->Search (pNode->pSubTree [AVL_RIGHT], NodeIndex);
  else return &(pNode->Value);
}

    

template <class TINDEX, class TDATA> int tAvlHybridTree<TINDEX, TDATA>::
  Balance (tAvlHybridNode<TINDEX, TDATA>*& pNode)
{
  
  if (this->LeftHeavy (pNode->iBalance) ) {
    if (pNode->pSubTree [AVL_LEFT]->iBalance == AVL_RIGHT_HEAVY)
      return this->RotateRightLeft (pNode);
    else return this->RotateRight (pNode);
  }
  else if (this->RightHeavy (pNode->iBalance)) {
    if (pNode->pSubTree [AVL_RIGHT]->iBalance == AVL_LEFT_HEAVY)
      return this->RotateLeftRight (pNode);
    else return this->RotateLeft (pNode);
  }

  return AVL_HEIGHT_NO_CHANGE;
} 



template <class TINDEX, class TDATA> int tAvlHybridTree<TINDEX, TDATA>::
  RotateRight (tAvlHybridNode<TINDEX, TDATA>*& pNode)
{
  tAvlHybridNode<TINDEX, TDATA>* pOldNode = pNode;
  int iHeightChange;
  
  if (pNode->pSubTree [AVL_LEFT]->iBalance == 0) iHeightChange = 0;
  else iHeightChange = AVL_HEIGHT_CHANGE;
  
  pNode = pNode->pSubTree [AVL_LEFT];
  pOldNode->pSubTree [AVL_LEFT] = pNode->pSubTree [AVL_RIGHT];
  pNode->pSubTree [AVL_RIGHT] = pOldNode;

  // Upate balances
  pNode->iBalance++;
  pOldNode->iBalance = -pNode->iBalance; 
  
  return iHeightChange;
}


template <class TINDEX, class TDATA> int tAvlHybridTree<TINDEX, TDATA>::
  RotateLeft (tAvlHybridNode<TINDEX, TDATA>*& pNode)
{
  tAvlHybridNode<TINDEX, TDATA>* pOldNode = pNode;
  int iHeightChange;
  
  if (pNode->pSubTree [AVL_RIGHT]->iBalance == 0) iHeightChange = 0;
  else iHeightChange = AVL_HEIGHT_CHANGE;
  
  pNode = pNode->pSubTree [AVL_RIGHT];
  pOldNode->pSubTree [AVL_RIGHT] = pNode->pSubTree [AVL_LEFT];
  pNode->pSubTree [AVL_LEFT] = pOldNode;
  
  // Upate balances
  pNode->iBalance--;
  pOldNode->iBalance = -pNode->iBalance; 
  
  return iHeightChange;
}

template <class TINDEX, class TDATA> int tAvlHybridTree<TINDEX, TDATA>::
  RotateRightLeft (tAvlHybridNode<TINDEX, TDATA>*& pNode)
{
  tAvlHybridNode<TINDEX, TDATA>* pOldNode = pNode;
  tAvlHybridNode<TINDEX, TDATA>* pOldLeftSubTree = pNode->pSubTree [AVL_LEFT];
  
  pNode = pOldNode->pSubTree [AVL_LEFT]->pSubTree [AVL_RIGHT];
  pNode->pSubTree [AVL_RIGHT] = pOldNode;
  
  pOldLeftSubTree->pSubTree [AVL_RIGHT] = pNode->pSubTree [AVL_LEFT];
  pNode->pSubTree [AVL_LEFT] = pOldLeftSubTree;
  pOldNode->pSubTree [AVL_LEFT] = NULL;
  
  
  // Update balances
  pNode->pSubTree [AVL_LEFT]->iBalance = - max (pNode->iBalance, 0);
  pNode->pSubTree [AVL_RIGHT]->iBalance = - min (pNode->iBalance, 0);
  pNode->iBalance = 0;
  
  return AVL_HEIGHT_CHANGE;
}
 
template <class TINDEX, class TDATA> int tAvlHybridTree<TINDEX, TDATA>::
  RotateLeftRight (tAvlHybridNode<TINDEX, TDATA>*& pNode)
{
  tAvlHybridNode<TINDEX, TDATA>* pOldNode = pNode;
  tAvlHybridNode<TINDEX, TDATA>* pOldRightSubTree = pNode->pSubTree [AVL_RIGHT];
  
  pNode = pOldNode->pSubTree [AVL_RIGHT]->pSubTree [AVL_LEFT];
  pNode->pSubTree [AVL_LEFT] = pOldNode;
  
  pOldRightSubTree->pSubTree [AVL_LEFT] = pNode->pSubTree [AVL_RIGHT];
  pNode->pSubTree [AVL_RIGHT] = pOldRightSubTree;
  pOldNode->pSubTree [AVL_RIGHT] = NULL;
  
  
  // Update balances
  pNode->pSubTree [AVL_LEFT]->iBalance = - max (pNode->iBalance, 0);
  pNode->pSubTree [AVL_RIGHT]->iBalance = - min (pNode->iBalance, 0);
  pNode->iBalance = 0;
  
  return AVL_HEIGHT_CHANGE;
}

template <class TINDEX, class TDATA> AVL_COMPARE_RESULT tAvlHybridTree<TINDEX, TDATA>::
  Compare (const tAvlHybridNode<TINDEX, TDATA>* pNode, const TINDEX& NodeIndex, 
  AVL_COMPARE CompareType) const
{

  switch (CompareType) {
    case AVL_EQUAL : 
      if (pNode->Index > NodeIndex) return AVL_GO_LEFT;
      else if (pNode->Index < NodeIndex) return AVL_GO_RIGHT;
      else return AVL_NO_GO;
    case AVL_MINIMUM : 
      if (pNode->pSubTree [AVL_LEFT] == NULL) return AVL_NO_GO;
      else return AVL_GO_LEFT;
    case AVL_MAXIMUM :
      if (pNode->pSubTree [AVL_RIGHT] == NULL) return AVL_NO_GO;
      else return AVL_GO_RIGHT;
  }
}





