/* $RCSfile: dataStore.h,v $ */
/* CM Info:   
$Header $
*/
/* Richard Johnson - Jet Propulsion Laboratory */

/*
modification history
--------------------
04a,6Sept97,rlj    written
*/

#ifndef INC_dataStoreh
#define INC_dataStoreh

#include "isit.h"
#include "typeTag.h"
#include "gizmo.h"
#include <semLib.h>
#include <taskLib.h>
#include <errnoLib.h>
#include "ppcFpExceptions.h"
#include "image.h"

#define S_datastore_UNABLE_TO_CREATE_DATA_BUFFERS (M_datastore | 0x01)
#define S_dataStore_DATA_SEND_INCOMPLETE          (M_datastore | 0x02)
#define DATA_DUMP_STACK_SIZE  (16384)
#define DATA_DUMP_OPTIONS  VX_FP_TASK


/**** DESCRIPTION ******

The dataStore class is designed to provide a data storage and sending mechanism for data types with well 
defined assignment operators and copy contructors.  Since the data type could be sent over shared memory, 
data types with pointers or virtual functions are not allowed.

******/

/****
Black Box State Variables

tDataStore <tDataType>

BAG <tTypeTag + tDataType> DataBag
int                        iMaxItems
int                        iDataDumpPriority  // should add an error priority...
tTypeTag                   DataTag


tBufferDataStore

BAG <tTypeTag + tBuffer>   DataBag
int                        iMaxItems
int                        iDataDumpPriority // should add an error priority... 
tTypeTag                   DataTag


[[D tDataStore::tDataStore (int iDataDumpPriorityToUse, int iMaxItemsToUse, 
                const tTypeTag& TypeTagToUse) ]
[ invalid_pointer (this) ||
  invalid_reference (TypeTagToUse) => UNDEFINED
| iDataDumpPriority < priority st. Time needed to empty this->DataBag < 
  iMaxItemsToUse / (average rate at which tDataStore::StoreData 
  will be called) => 
    ::ErrorStreamG (::pThisRteG) := ::ErrorStreamG (::pThisRteG) + appropriate 
    error messages whenever storing data would cause this->DataBag to overflow
| true =>
     this->DataBag := EMPTY
    : this->iMaxItems := iMaxItemsToUse
    : this->iDataDumpPriority := iDataDumpPriorityToUse
    : this->DataTag := TypeTagToUse
]]

[[D tDataStore::~tDataStore (void) ]
[ invalid_pointer (this) => UNDEFINED
| true => *this := anything
]]

[[D tDataStore::StoreData (tDataType& DataIn) ]
[ invalid_pointer (this) ||
  invalid_pointer (::pThisRteG) ||
  ::DataStream (::pThisRteG).bInit == FALSE => UNDEFINED
| NumberOfItems (this->DataBag)== this->iMaxItems - 1 &&
  ::DataStream (::pThisRteg).bFull == TRUE at some point during the dump => 
    this->DataBag := EMPTY
    : ::DataStream (::pThisRteG) := ::DataStream (::pThisRteG) + anything
    : console := console + appropriate error messages
| NumberOfItems (this->DataBag)== this->iMaxItems - 1 =>
    this->DataBag := EMPTY
    : ::DataStream (::pThisRteG) := ::DataStream (::pThisRteG) + 
      this->DataBag at priority this->iDataDumpPriority
| true => this->DataBag := this->DataBag + (this->DataTag + DataIn)
]]





[[D tImageDataStore::tImageDataStore (int iDataDumpPriorityToUse, 
                                        int iMaxItemsToUse, 
                                        const tTypeTag& TypeTagToUse) ]
[ invalid_pointer (this) ||
  invalid_reference (TypeTagToUse) => UNDEFINED
| iDataDumpPriority < priority st. Time needed to empty this->DataBag < 
  iMaxItemsToUse / (average rate at which tDataStore::StoreData 
  will be called) => 
    ::ErrorStreamG (::pThisRteG) := ::ErrorStreamG (::pThisRteG) + appropriate 
    error messages whenever storing data would cause this->DataBag to overflow
| true =>
     this->DataBag := EMPTY
    : this->iMaxItems := iMaxItemsToUse
    : this->iDataDumpPriority := iDataDumpPriorityToUse
    : this->DataTag := TypeTagToUse
]]

[[D tImageDataStore::~tImageDataStore (void) ]
[ invalid_pointer (this) => UNDEFINED
| true => *this := anything
]]

[[D tImageDataStore::StoreData (tImage& DataIn) ]
[ invalid_pointer (this) ||
  invalid_pointer (::pThisRteG) ||
  ::DataStream (::pThisRteG).bInit == FALSE => UNDEFINED
| NumberOfItems (this->DataBag)== this->iMaxItems - 1 &&
  ::DataStream (::pThisRteg).bFull == TRUE at some point during the dump => 
    this->DataBag := EMPTY
    : ::DataStream (::pThisRteG) := ::DataStream (::pThisRteG) + anything
    : console := console + appropriate error messages
| NumberOfItems (this->DataBag)== this->iMaxItems - 1 =>
    this->DataBag := EMPTY
    : ::DataStream (::pThisRteG) := ::DataStream (::pThisRteG) + 
      this->DataBag at priority this->iDataDumpPriority
| true => this->DataBag := this->DataBag + (this->DataTag + DataIn)
]]
*****/


template <class tDataType> class tDataStore {
  public:
    tDataStore (int iDataDumpTaskPriority, int iBufferSize, 
                const tTypeTag& TypeTagToUse);
    virtual ~tDataStore ();
    void StoreData (const tDataType& DataIn);
  protected:
    virtual void DumpToDataStream ();
 
    int iNumberOfDataElements;
    tTypeTag DataTag;
    int iCurrentWriteBuffer;
    int iCurrentReadBuffer;
    int iCurrentElement;
    BOOL bSendDone;
    ERROR_CODE ecErrorFlag;
    tDataType* DataBufferArrayArray[2];
    SEM_ID siDataDumpBinarySemaphore;
    SEM_ID siErrorEvent;
    int iDataDumpTaskId;
    int iHandleErrorTaskId;   
  private:
    void ProcessDumpToDataStream ();
    void HandleError ();
};


class tImageDataStore : public tDataStore<tImage> {
  public:
    tImageDataStore (int iDataDumpTaskPriority, int iBufferSize, 
                      const tTypeTag& TypeTagToUse);
    ~tImageDataStore ();
  protected:
    void DumpToDataStream ();

}; 
    


template <class tDataType> tDataStore<tDataType>::
  tDataStore (int iDataDumpTaskPriority, int iNumberOfDataElementsToUse, 
              const tTypeTag& TypeTagToUse) :

  iNumberOfDataElements (iNumberOfDataElementsToUse),
  DataTag (TypeTagToUse),
  iCurrentWriteBuffer (0),
  iCurrentReadBuffer (1),
  iCurrentElement (0),
  bSendDone (TRUE),
  ecErrorFlag (OK)
{
  DataBufferArrayArray[0] = new tDataType [iNumberOfDataElementsToUse]; 
  DataBufferArrayArray[1] = new tDataType [iNumberOfDataElementsToUse]; 

  this->siDataDumpBinarySemaphore = semBCreate (SEM_Q_FIFO,SEM_EMPTY);
 
  if (this->siDataDumpBinarySemaphore == 0 )  { 
     TraceMsg (tmt_ERROR, errnoGet (), "tDataStore::tDataStore");
     return;    
  }
    
  this->siErrorEvent = semBCreate (SEM_Q_FIFO,SEM_EMPTY);
 
  if (this->siErrorEvent == 0 )  { 
     TraceMsg (tmt_ERROR, errnoGet (), "tDataStore::tDataStore");
     return;    
  }


  this->iDataDumpTaskId = taskSpawn ("DataStoreDump", iDataDumpTaskPriority,
        DATA_DUMP_OPTIONS, DATA_DUMP_STACK_SIZE, 
        (FUNCPTR) this->ProcessDumpToDataStream, (int) this,
        0,0,0,0,0,0,0,0,0);

  if (this->iDataDumpTaskId == ERROR)   { 
     TraceMsg (tmt_ERROR, errnoGet (), "tDataStore::tDataStore");
     return;    
  }
   
  this->iHandleErrorTaskId = taskSpawn ("DataStoreDump", iDataDumpTaskPriority,
        DATA_DUMP_OPTIONS, DATA_DUMP_STACK_SIZE, 
        (FUNCPTR) this->HandleError, (int) this,
        0,0,0,0,0,0,0,0,0);

 
 }


template <class tDataType> tDataStore<tDataType>::~tDataStore () {

  taskDelete (this->iDataDumpTaskId);
//  taskDelete (this->iHandleErrorTaskId);
  
  semDelete (this->siDataDumpBinarySemaphore);
//  semDelete (this->siErrorEvent);

  delete[] this->DataBufferArrayArray[0];
  
  delete[] this->DataBufferArrayArray[1];

}


// Someday this should be made more robust by (e.g.) checking for
// weird member values ( >= instead of == ) and maybe return code 
// from semGive...

template <class tDataType> void tDataStore<tDataType>::StoreData (
    const tDataType& DataIn) {

  if (this->iCurrentElement == this->iNumberOfDataElements) {
    if (this->bSendDone == FALSE) {
      this->ecErrorFlag = S_dataStore_DATA_SEND_INCOMPLETE;
      semGive (this->siErrorEvent);
      return;
    }
    this->bSendDone = FALSE;  
    this->iCurrentElement = 0;
     
    if (this->iCurrentWriteBuffer == 0) {
      this->iCurrentWriteBuffer = 1;
      this->iCurrentReadBuffer = 0;
    }
     
    else { 
      this->iCurrentWriteBuffer = 0;
      this->iCurrentReadBuffer = 1;
    }
     
    semGive (this->siDataDumpBinarySemaphore);
  }

  this->DataBufferArrayArray [this->iCurrentWriteBuffer][this->iCurrentElement] =
           DataIn;
  
  this->iCurrentElement++;

}


template <class tDataType> void tDataStore<tDataType>::ProcessDumpToDataStream () {

  int i;

  // Added 5/17/99 PcI
  // Underflows keep happening in the RMS jitter calculations
  DisableFpUnderflow();
  DisableFpOverflow();

  FOREVER {
     
    if (this->ecErrorFlag != OK) { 
      /****
      [[U tGizmo::SendError (this->ecErrorFlag) ]]
      ****/    
      tGizmo::SendError (this->ecErrorFlag);
      
      this->ecErrorFlag = OK;

      taskDelay (1);
    } 
    else {

      semTake (this->siDataDumpBinarySemaphore,WAIT_FOREVER);
 
      this->DumpToDataStream ();
    
      this->bSendDone = TRUE;
     
    }
  }
}




template <class tDataType> void tDataStore<tDataType>::DumpToDataStream ()
{
   /****
   [[U tGizmo::SendData (this->DataTag, 
         &this->DataBufferArrayArray [this->iCurrentReadBuffer][0] ) ]]
   ****/
   tGizmo::SendData (this->DataTag, (BYTE*)
     &this->DataBufferArrayArray [this->iCurrentReadBuffer][0] );
}



template <class tDataType> void tDataStore<tDataType>::HandleError () 
{

  FOREVER {
   
    semTake (this->siErrorEvent,WAIT_FOREVER);

    ::tGizmo::SendError (this->ecErrorFlag);
  }
}


#endif INC_dataStoreh





