/* $RCSfile: dlDither.h,v $ - Delay Line Dither Controller */
/* CM Info:
$Header: /home/chara/cvsroot/ople/sys/include/dlDither.h,v 1.1.1.1 1998/06/11 21:36:11 chara Exp $
*/
/* Beth McKenney - Jet Propulsion Laboratory */
 
/*
modification history
--------------------
01a,eamck,17Apr97    written 
*/
 
/*
DESCRIPTION

*/

#ifndef INC_dlDitherh
#define INC_dlDitherh

#include "dither.h"
#include "outputChannel.h"
#include "channelBank.h"


 /* these values may be changed if performance is not satisfactory */
#define NUM_WAVES_TO_AVG (100)       /* prob. should be 1000 or so */
#define MAX_SWEEP_ERRORS (1000)       /* max acceptable # failed sweeps */


/*********************************************************************
   NOTE - BIG ASSUMPTION:  the Dither software is not responsible for
   turning the waveform board scan "off" or "on", in fact the Dither
   software doesn't even know that the board exists (and only knows
   about the ChannelBank to get the size info).  The tDlDither object
   only knows about one channel on the board, so all of this code is
   written with the assumption that the waveform generator board is
   already scanning by the first time we want to start dithering.
 *********************************************************************/

 
/******************************************************************

 Dither waveform ramblings:

      When computing the dither waveform, the *top* priority is that
    the period be an even multiple of the servo period.  This is
    necessary in order to ensure that the measured waveform array
    gets filled properly.  If the user requests a period that is
    not an even multiple, the period will be changed internally so
    that it is.
      Next, the actual stored waveform size must represent an even
    multiple of dither ticks; since there may not be an integer
    number of dither ticks in a servo period, the dither period
    may not be exactly the same as the storage buffer size.
      We considered a number of methods that would make the system
    capable of handling arbitrary waveform periods, but none of 
    them was reasonable for implementing in Increment 2.  Thus for
    Increment 2, the user MUST enter a waveform period that is an
    integer multiple of both the servo period and the dither sample
    period.  Inputting any other number will produce an error.

      Another issue is the fact that, in Increment 2 at least, all 
    delay lines will be receiving their dither signals from the 
    same waveform generator board, so there are board-wide properties 
    and state variables to consider.  The sample output frequency 
    and the dither cycle period will be the same for all delay lines 
    in the system; changing the dither cycle period for any one delay 
    line changes it for all of them.  (The action is permitted, but 
    the user should be aware of the ramifications.) Starting and 
    stopping output from the channel bank are other functions that 
    affect all the delay lines.  An individual channel can be turned
    "off" or "on" by loading that channel's portion of the RAM with
    all zeroes or a signal waveform, respectively; however, the
    board itself needs to be commanded to start output at some point.
    The dither object(s) should probably not be given the capability
    of turning on and off the scanning since it affects all the other
    DLs as well; better to have the board simply come up already 
    scanning but with its RAM filled with all zeroes.


 ******************************************************************/



/******************************************************************
*
* tDlDither Black Box
*
* BB State Variables
* ------------------
*
*
*******************************************************************
*
* ERROR_CODE tDlDither::GetStoreIndex(int tTheTime)
*
* SPEC:
   [[D tDlDither::GetStoreIndex(int tTheTime)]
   [ this invalid => UNDEFINED
   | this->sdDitherSamplePeriod == 0 => UNDEFINED
   | this->sdDitherSamplePeriod or this->iWaveOffset or
         this->siDitherBufSize not appropriate for current conditions =>
		                           return := anything
   | true =>
        return := index into waveform measurement table that
                     corresponds to the output waveform sample
                     that was output at time tTheTime
   ]]
*
* RETURNS:
*    OK
*
*******************************************************************
*
* double tDlDither::GetDitherValue(TIME tCurrTime)
*
* SPEC:
   [[D tDlDither::GetDitherValue(TIME tCurrTime)]
   [ this invalid => UNDEFINED
   | this->bDithering == FALSE => 
       return := 0.0
   | true => 
       return := dither correction value corresponding to time tCurrTime
   ]]

* RETURNS:
*    OK
*
*******************************************************************
*
* ERROR_CODE tDlDither::SetWaveform(TIME tPeriod, 
                                    double dAmplitude,
                                    double dAmplFactor, 
                                    int iPhaseOffset, 
                                    BOOL bPreDistortion)
*     SetWaveform draws a triangle wave in this->dDitherOutputTable[]
*     It should also take the metrics necessary for modifying a perfect
*     triangle wave so that outputting the modified wave will produce
*     a more perfect triangle wave modulation in the metrology.
*     For Incr 2, we just subtract the diff betw the measured wave and
*     the output wave from the output wave and iterate. (from mmc)
*
*     Currently, we ignore iPhaseOffset completely.
*
*     SetWaveform also sets the board-wide dither period; note this affects 
*     ALL DLs using this waveform generator board (in incr 2, ALL DLs).
*
* SPEC:
   [[D tDlDither::SetWaveform(TIME tPeriod, double dAmplitude,
			        double dAmplFactor, int iPhaseOffset,
				BOOL bPreDistortion)]
   [ this invalid => UNDEFINED
   | this->pDitherBank is invalid => UNDEFINED
   | dAmplFactor <= 0 => return := S_dlDither_INVALID_AMPL_FACTOR
   | tPeriod not an integer multiple of this->sdDitherSamplePeriod ||
            tPeriod not an integer multiple of this->tDlServoTaskPeriod =>
        return := S_dlDither_INVALID_DITHER_PERIOD
   | tPeriod <=0  || ROUND( tPeriod / (1.0*TIME_TICKS_PER_SECOND) * 
	(1/(0.000001 * sample period of each channel)) )  > 
	                            this->pDitherBank->iTotalMemoryLoc =>
               this->siDitherBufSize := waveform size value from channel bank
   | true =>
       this->stDitherPeriod := tPeriod
       : this->siDitherBufSize := bufsize needed to store waveform
       : waveform size for this->pDitherBank := bufsize needed to store waveform
       : this->siDoubleDitherBufSize := 2 * this->siDitherBufSize
       : this->dDitherAmplitude := dAmplitude
       : this->dIdealTriangleWave[] gets a triangle wave of amplitude
           this->dDitherAmplitude and period this->tDitherPeriod
       : this->dDitherOutputTable[] gets two cycles of a triangle wave 
	    of amplitude this->dDitherAmplitude and period this->tDitherPeriod,
               maybe modified according to dAmplFactor, iPhaseOffset, and 
	       predistortion (if bPreDistortion==TRUE)
       : return := OK
   ]]

* RETURNS:
*    OK
*
*******************************************************************
*
* ERROR_CODE tDlDither::LoadDitherWaveform(int iOffset)
*
*     LoadDitherWaveform takes the current waveform stored in
*     this->dDitherOutputTable[], converts it to "channel units"
*     and outputs it to the waveform board's RAM.
*     iOffset changes the starting point.  (Note that a positive
*     offset means the wave shifts *earlier* in time.) Always
*     loads this->siDitherBufSize samples to include a full cycle.
*     NOTE that tWaveformChannel::LoadMemory will set the period
*     for the whole board to iSize, i.e. this->siDitherBufSize. 
*
* SPEC:
   [[D tDlDither::LoadDitherWaveform(int iOffset)]
   [ this invalid || this->pDitherChannel invalid => UNDEFINED
   | iOffset out of range => return := S_dlDither_INVALID_OFFSET
   | max index will be out of range => return := S_dlDither_INVALID_BUFSIZE
   | this->pDitherChannel->bInit == FALSE => 
       this->dDitherOutputTableInChannelUnits[] :=
	     this->dDitherChannelUnits * (this->dDitherOutputTable+iOffset)[]
       : return := S_tWaveformChannel_CHANNEL_NOT_INITIALIZED
   | ABS(any element(s) in this->dDitherOutputTableInChannelUnits) >=
	                                  2^this->pDitherChannel->iBitWidth =>
       this->dDitherOutputTableInChannelUnits[] :=
	     this->dDitherChannelUnits * (this->dDitherOutputTable+iOffset)[]
       : this->pDitherChannel RAM gets loaded with anything
		(loaded values will reflect relevant bits)
       : return := OK
   | true =>
       this->dDitherOutputTableInChannelUnits[] :=
 	      this->dDitherChannelUnits * (this->dDitherOutputTable+iOffset)[]
       : this->pDitherChannel RAM gets loaded with appropriate values from 
	      (this->dDitherOutputTable+iOffset)[], converted to channel units
       : return := OK
   ]]
*
* RETURNS:
*    OK
*
*******************************************************************
*
* ERROR_CODE tDlDither::DitherOff(void)
*
* SPEC:
   [[D tDlDither::DitherOff(void)]
   [ this invalid || pDitherChannel invalid => UNDEFINED
   | true =>
       [[U tOutputChannel::Reset(this=this.*pDitherChannel)]]
       : this->bDithering := FALSE
       : return := OK
   ]]

* RETURNS:
*    OK
*    tOutputChannel::Reset error msgs?
*
*******************************************************************
*
* ERROR_CODE tDlDither::DitherOn(void)
*
* SPEC:
   [[D tDlDither::DitherOn(void)]
   [ this invalid => UNDEFINED
   | true =>
       [[U tDlDither::LoadDitherWaveform(iOffset=0)]]
       : this->bDithering := TRUE
       : return := OK
   ]]
 
* RETURNS:
*    OK
*    tDlDither::LoadDitherWaveform error msgs?
*
*******************************************************************
*
* ERROR_CODE tDlDither::Initialize(tDlGizmo         *pGizmo, 
                                   TIME              tDlServoPeriodToUse, 
                                   tWaveformChannel *pDitherPztChannel,
                                   tActuator        *pWaveformActuatorToUse,
                                   tLaserMetrologySensor *pLaserSensorToUse,
                                   TIME              tDitherPeriod,
                                   double            dDitherAmplitude)
*
* SPEC:
   [[D tDlDither::Initialize(tDlGizmo         *pGizmo,
		             TIME              tDlServoPeriodToUse,
			     tWaveformChannel *pDitherPztChannel,
			     TIME              tDitherPeriod,
			     double            dDitherAmplitude)]
   [ this invalid || pGizmo invalid || pDitherPztChannel invalid ||
     pWaveformActuatorToUse invalid || pLaserSensorToUse invalid => UNDEFINED
   | true =>
       tDlDither object initialized appropriately, with dither waveform
          amplitude dDitherAmplitude, period tDitherPeriod, and dithering OFF
       : return := OK
   ]]
*
* RETURNS:
*    OK
*
*******************************************************************
*
* double tDlDither::DoDither(TIME tCurrTime, double dLaserReading)
*    Called by any task that wants to remove dither from laser reading
*
*    See Dither Calibration document for explanation of dither 
*           measurement algorithm, plus examples.
*
* SPEC:
   [[D tDlDither::DoDither(TIME tCurrTime, double dLaserReading)]
   [ this invalid => UNDEFINED
   | (storing dither values for dither check or dither calib command) &&
               (done with measurements for this value of this->iWaveOffset) =>
        reset flags and give sem back to MeasureDither()
        : return := dither correction value for time tCurrTime
   | (storing dither values for dither check or dither calib command) =>
        store dLaserReading at appropriate place in waveform storage buffer, 
                                                   setting flags as appropriate
        : return := dither correction value for time tCurrTime
   | (not storing dither values) && (dithering on) =>
        return := dither correction value for time tCurrTime
   | true => 
       return := 0.0
   ]]

* RETURNS:
*    appropriate dither value for time tCurrTime
*
*******************************************************************
*
* ERROR_CODE tDlDither::MeasureDither(int iNumWaves)
*     MeasureDither fills up dCollectDitherTable[] with laser readings
*     from DoDither, offsetting the dither waveform in between (this 
*     requires reloading the waveform starting at a different point 
*     each time), then calculates the average waveform, subtracts the
*     mean, and stores the avg waveform's amplitude, phase, and 
*     distortion in the DL Status block.
*
* SPEC:
   [[D tDlDither::MeasureDither(int iNumWaves)]
   [ this invalid => UNDEFINED
   | true =>
       this->iWavesToAvg := iNumWaves
       :\ reset measurement buffer to all elements := 0.0
       : [ while (current dither wave offset < max range) &&
             (not done filling all elements of measurement buffer) =>
                  load our dither waveform at new offset
                    : start storing laser values
                    : take semaphore and wait until DoDither() says we're done
                  ]
       : [ (current dither wave offset < max range) && 
                             (not done filling measuremnet buffer) =>
               return := S_dlDither_DITHER_MEASUREMENT_ERROR ]
       :\ divide our accumulated wave by number of measurements to get avg wave
       :\ calculate and subtract off mean value of avg waveform
       :\ calculate metrics
       :\ update metrics in DL status block
       : return := OK
   ]]
*
* RETURNS:
*    OK
*
*******************************************************************
*
* ERROR_CODE tDlDither::CheckDither(void)
*    Called by tDlDitherCheckCommandEngine::ExecuteMethod();
*      assumes that the command engine calls tDlDither::CleanUp()
*
* SPEC:
   [[D tDlDither::CheckDither(void)]
   [ this invalid => UNDEFINED
   | true =>
       save our current dithering state (i.e. off or on)
       :\ turn dithering on  (reset in CleanUp() )
       :\ save current dither output and correction tables
       :\ set new waveform with perfect triangle wave
       : [[U tDlDither::MeasureDither(iNumWaves=NUM_WAVES_TO_AVG) ]]
       : [ this->MeasureDither barks =>
            return := S_dlDither_DITHER_MEASUREMENT_ERROR ]
       :\ return := OK
   ]]

* RETURNS:
*    OK
*    error msgs from these tDlDither methods
*
*******************************************************************
*
* ERROR_CODE tDlDither::CalibrateDither(void)
*    Called by tDlDitherCalibrateCommandEngine::ExecuteMethod();
*      assumes that the command engine calls tDlDither::CleanUp()
*
* SPEC:
   [[D tDlDither::CalibrateDither(void)]
   [ this invalid => UNDEFINED
   | other bad things I haven't thought of yet => UNDEFINED
   | true => 
       this->bCalibrating := TRUE  (set to FALSE in CleanUp() )
       :\ save our current dithering state (i.e. off or on)
       :\ turn dithering on
       :\ save current dither output and correction tables 
       :\ set new waveform with perfect triangle wave
       : [[U tDlDither::MeasureDither(iNumWaves=NUM_WAVES_TO_AVG)]]
       : [ if this->MeasureDither barks =>
            return := S_dlDither_DITHER_MEASUREMENT_ERROR ] 
       : [ while (this->dMetrics > GOOD_METRIC_VALUE) =>
             set new waveform with triangle wave modified according to metrics
             : [[U tDlDither::MeasureDither(iNumWaves=NUM_WAVES_TO_AVG)]]
             : [ if this->MeasureDither barks =>
                 return := S_dlDither_DITHER_MEASUREMENT_ERROR ]
         ]
       :\ copy the latest measured wave into our dither correction table
       :\ return := OK
   ]]
*
* RETURNS:
*    OK
*    error msgs from these tDlDither methods
*
*******************************************************************
*
* ERROR_CODE tDlDither::CleanUp(ERROR_CODE ecCondition)
*     CleanUp is called by CalibrateDither and CheckDither command
*     engines' Execute and Abort methods, to make sure things get back
*     to normal after the calibration or check.  ecCondition is our
*     way of telling whether to reset to the previous settings or just
*     close up shop.
*
* SPEC:
   [[D tDlDither::CleanUp(ERROR_CODE ecCondition)]
   [ this invalid => UNDEFINED
   | (current command is calibrate dither) && (ecCondition != OK) =>
       stop storing dither measurements
       :\ give semaphore back
       :\ restore original dither output and correction tables
       :\ load original output wave into RAM (offset == 0)
       :\ turn dithering off or on, as it was before command
       :\ this->bCalibrating := FALSE
       :\ return := ecCondition
   | true =>
       stop storing dither measurements
       :\ give semaphore back
       :\ load current output wave into RAM (offset == 0)
       :\ turn dithering off or on, as it was before command
       :\ this->bCalibrating := FALSE
       :\ return := ecCondition
   ]]
*
* RETURNS:
*    OK
*


[[D tDlDither::GetCalibInfo (void) ]
Return calibration information.
[ invalid_pointer (this) => UNDEFINED
| true => return := this->CalibInfo
]]


[[D tDlDither::DoingCheckOrCalib (void) ]
Returns TRUE if this is calibrating or checking.
[ invalid_pointer (this) => UNDEFINED
| true => return := bDoingCheckOrCalib
]]

[[D tDlDither::SetDitherParameters(double dNewAmplitude, TIME tInitPeriod) ]
Sets the dither waveform parameters.
[ invalid_pointer (this) => UNDEFINED
| true => 
    this->dAmplitude := dNewAmplitude
    : this->tPeriod := tInitPeriod
    : return := OK
]]

      
  

*******************************************************************/

class tWaveformChannel;
class tWaveformChannelBank;

class tDitherCalibInfo {
  public:
    tDitherCalibInfo () {};
    tDitherCalibInfo (double dAmp, double dPhase, double dMetricIn) :
      dAmplitude (dAmp), dPhaseShift(dPhase), dMetric (dMetricIn) {}
  double dAmplitude;
  double dPhaseShift;
  double dMetric;
};
  

class tDlDither : public tDither {       
  public:
    tDlDither (tWaveformChannel& DitherPztChannel, double dDitherChannelUnitsPerCount,
               TIME tDlFastServoPeriodIn, TIME tDlDitherPeriodIn, 
               double dDitherAmplitudeIn);
/*          
    ERROR_CODE        Initialize(tDlGizmo         *pGizmo, 
				   TIME              tDlServoPeriodToUse,
                                   tWaveformChannel *pDitherPztChannel,
                                   tActuator        *pWaveformActuatorToUse,
                                   tLaserMetrologySensor *pLaserSensorToUse,
                                   TIME              tDitherPeriod,
                                   double            dDitherAmplitude);
*/
      double            DoDither(TIME tCurrTime, double dLaserReading);
      double            GetDitherValue(TIME tCurrTime); 
      ERROR_CODE        DitherOff(void);
      ERROR_CODE        DitherOn(void);
      BOOL              IsDitherOn () const { return this->bDithering;}
      ERROR_CODE        CalibrateDither(void);
      ERROR_CODE        CheckDither(void);
      ERROR_CODE        CleanUp(ERROR_CODE ecCondition); 
      ERROR_CODE        SpoutMeasurementDiags(void);
      double            DitherAmplitude () const { return this->dDitherAmplitude;}
      TIME              DitherPeriod () const { return this->stDitherPeriod;}
      const tDitherCalibInfo &GetCalibInfo(void) const { return CalibInfo; }
      BOOL DoingCheckOrCalib() const { return bDoingCheckOrCalib; }
      ERROR_CODE        SetDitherParameters(double dNewAmplitude,
                                            TIME tInitPeriod); 
      
   protected:
      BOOL              bDoingCheckOrCalib;
      tWaveformChannelBank      *pDitherBank;
      tWaveformChannel  *pDitherChannel;
      double		dDitherChannelUnits;
      TIME               tDlServoTaskPeriod;

      BOOL               bDithering;  

         /* these vars are static since they reflect board properties */
      static TIME       stDitherPeriod;          /* wave board scanning rate */
      static double     sdDitherSamplePeriod;  /* TIME units but floating pt */
      static int        siDitherBufSize;        /* buf size to store 1 cycle */
      static int        siDoubleDitherBufSize;      /* = 2 * siDitherBufSize */
      static int        siMaxDitherBufSize;          /* = total RAM / #chans */
      static int        siDoubleMaxDitherBufSize;  /* = 2*siMaxDitherBufSize */
      static int        siMaxWaveOffset;    /* max offset shift during meas. */
      static int        siSweepBufSize;    /* # measurements in single sweep */

         /* amplitude however can be different for different DLs */
      double            dDitherAmplitude;             /* same units as laser */


            /* vars used during dither calibrate operation */

      BOOL              bCalibrating;      /* only true for calib, NOT check */
      BOOL              bWasDitherOn;      /* save prev state to reset later */
      BOOL              bStoreDither;
      SEM_ID            siNextOffsetSem;
      int               iWaveOffset;
      int               iWavesToAvg;
      int               iSweepSampleCount;
      int               iSweepErrorCount;
      int               iTotalSweepCount;
      BOOL		*bGotSampleArray;
      double            *dSweepSampleArray;
      double            *dCollectDitherTable;
      double            *dCollectSqDitherTable;      /* for diagnostics only */
      double            *dWaveDistortionTable;    /* diff betw output & meas */
      double            *dIdealTriangleWave;

      double            dCurrentAmplFactor;
      double            dProposedAmplFactor;
      int               iCurrentPhaseShiftIndex;
      int               iProposedPhaseShiftIndex;
      tDitherCalibInfo  CalibInfo;
      double            dAmpError;
      double            dPhaseError;
      double            dDistortionError;

      double            *dDitherCorrTable;
      double            *dDitherOutputTable;
      double            *dDitherOutputTableInChannelUnits;
      double            *dSavedDitherCorrTable;
      double            *dSavedDitherOutputTable;

      ERROR_CODE        LoadDitherWaveform(int iOffset);
      ERROR_CODE        MeasureDither(int iNumWaves);
      ERROR_CODE        SetWaveform(void);

     /* this GetStoreIndex uses all integers, avoids floating point errors */

      inline int        GetStoreIndex(TIME tTheTime)
          { return ( (int)((tTheTime % this->stDitherPeriod)  *
                                          this->siDitherBufSize  / 
					     this->stDitherPeriod) +
                                                this->iWaveOffset )  %
                                                   this->siDitherBufSize ; }

      /* This is the same as GetStoreIndex, but ignores iWaveOffset */
      inline int        GetCorrectionIndex(TIME tTheTime)
          { return ( (tTheTime % this->stDitherPeriod)  *
                           this->siDitherBufSize  / this->stDitherPeriod ); }

};

 
extern "C"
{
   ERROR_CODE DitherDiags( int iDlNumber );    
}
 
#endif  /* INC_dlDitherh */


