/* $RCSfile: integratorSensor.h,v $ - Implements abstract hardware and software input classes */
/* CM Info:
$Header: /home/chara/cvsroot/ople/sys/include/integratorSensor.h,v 1.1.1.1 1998/06/11 21:36:12 chara Exp $
*/
/* Ken Starr - Jet Propulsion Laboratory */
 
/*
modification history
--------------------
01a,kms,11Nov97    written
*/
 
/*
DESCRIPTION

*/

#ifndef INC_ftInputSourceh
#define INC_ftInputSourceh


#include "isit.h"
#include "inputSource.h"

class tInputChannel;
class tInputSource;
class tSensor;

/*******************************************************************
*******************  tIntegratorSensor Black Box  ******************
********************************************************************
*
* tIntegratorSensor
*
* A tIntegratorSensor is a type of tSensor that deals with an
* integrator as a data source. The value of the integrator will continue
* increasing until it is commanded to reset. Each time Update is called, 
* the new value the integrator diffrenced with the last value (in the case 
* of wraparound the value is corrected). The corrected difference is
* returned.
*
* Currently, tIntegratorSensor's ignore the TIME input. 
*
* BB State Variables
*
* A tIntegratorSensor has all the state of a tInputSource and tSensor, plus...
*
* Variable                   Initial Value
* ----------------------------------------
* dLastIntegratorCount	    0.0
*
*******************************************************************
*
* tIntegratorSensor::tIntegratorSensor
*
* DESCRIPTION: Constructor
*
* RETURNS: Nothing
*

[[D tIntegratorSensor::tIntegratorSensor (const tInputChannel &ChannelToUse,
		    		 const double dElectricalToPhysicalScaleFactorToUse) ]
  [ invalid_pointer(this) or invalid_reference(ChannelToUse)  => UNDEFINED
  | true =>
     this->Channel := ChannelToUse
     : this->dElectricalToPhysicalScaleFactor :=
                               dElectricalToPhysicalScaleFactorToUse
     : this->dValue := 0.0
     : this->dOffset := 0.0
     : this->dLastIntegratorCount := 0.0
]]

*
*******************************************************************
*
* tIntegratorSensor::~tIntegratorSensor
*
* DESCRIPTION: Compiler-supplied destructor
*
* RETURNS: Nothing
*

[[D tIntegratorSensor::~tIntegratorSensor (void) ]
  [ true => *this := anything
]]


*
*******************************************************************
*
* tIntegratorSensor::Update
*
* DESCRIPTION: Computes new Integrated value using current and last
*		   values.
*
* Currently, tIntegratorSensor's ignore the TIME input.
*
* Note that the TIME input is ignored by a base tSensor class.
*
* ASSUMPTIONS:  There is a bottom-up condition that comes up from
* tInputChannel::Input that says that the board must be receiving
* input properly.  For now, this condition is considered to be handled at
* the system level via a statement like "All hardware must be connected
* and receiving input properly".  Therefore, we'll use that assumption
* here to make the bottom-up condition disappear.
*
* RETURNS: double
*

 [[D tIntegratorSensor::Update (const TIME tUpdateTime) ]
  [ invalid_pointer(this) or invalid_reference(ChannelToUse) or
         this->Channel not properly initialized => UNDEFINED
  | if sensor is saturated =>   
     this->dLastIntegratorCount := anything
     : return := anything      
  | (new value read from this->Channel) < (dLastIntegratorCount) =>
     this->dValue := (2^24 + new value read from this->Channel) 
                           - (dLastIntegratorCount)
     : this->dLastIntegratorCount := new value read from this->Channel
     : return := this->dValue
  | true =>
     this->dValue := (new value read from this->Channel) 
                       - (dLastIntegratorCount)
     : this->dLastIntegratorCount := new value read from this->Channel
     : return := this->dValue
]]


*
*******************************************************************
*
* tIntegratorSensor::ResetIntegrator
*
* DESCRIPTION: Resets the Integrator state on the board
*
* RETURNS: ERROR_CODE
*

[[D tIntegratorSensor::ResetIntegrator (void) ]
  [ invalid_pointer(this) or invalid_reference(ChannelToUse) or
         this->Channel not properly initialized => UNDEFINED
  | true =>
     Reset Channel counter
     : this->dLastIntegratorCount := 0.0
     : this->dValue := 0.0	 
     : return := OK
]]

*
*/


/*******************************************************************/
/*******************************************************************/
/********************                     **************************/
/********************  Class Definitions  **************************/
/********************                     **************************/
/*******************************************************************/
/*******************************************************************/


/*******************************************************************/
/***********************  tIntegratorSensor  ************************/
/*******************************************************************/

#define MAX_INTEGRATOR_COUNT 16777216 /* number where the integrator count */
                             /* will wrap.  24 bit unsigned count = 2^24 */

class tIntegratorSensor : public tSensor {
public:
  tIntegratorSensor(const tInputChannel &ChannelToUse,
          const double dElectricalToPhysicalScaleFactorToUse);
  double Update(const TIME tUpdateTime);
  ERROR_CODE ResetIntegrator(void);
  
private:
  double dLastIntegratorCount;
};



#endif  /* ftINC_InputSourceh */

