/* $RCSfile: ipcElement.h,v $ */
/* $Header: /home/chara/cvsroot/ople/sys/include/ipcElement.h,v 1.1.1.1 1998/06/11 21:36:12 chara Exp $ */


#ifndef INC_ipcElementh
#define INC_ipcElementh

#include "isit.h"
#include "ipcid.h"
#include "ipc.h"
#include "ipcError.h"


/*********************************************************
tIpcElement<tDataType> Black Box State

tDataType dtElementData

________________________________________________________________________

Specification Functions:

CanAlocateIpcObject (..) - see ipcObjectAccessor.h
NewObjectSet (...)       - see ipcObjectAccessor.h
AllocationSize (...)     - see ipcObjectAccessor.h
________________________________________________________________________

Stimuli:

Default constructor: Sets element value to default tDataType value
Usage: 
  tIpcElement<int> MyIntElement;
  
[[D tIpcElement::tIpcElement (void) ]
[ invalid_pointer (this) => UNDEFINED
| true => this->dtDataType := default value
]]

Default copy contructor: Copies element contents into another element
Usage:
  tIpcElement<int> MyIntElement (YourIntElement); or
  MyIntElement = YourIntElement;
  
[[D tIpcElement::tIpcElement_copy (const tIpcElement& Element) ]
[ invalid_pointer (this) => UNDEFINED
| true => this->dtDataType := Element.dtDataType
]]


Non-default constructor: initializes element data with user specified value.
Usage: 
  tIpcElement<int> MyIntElement (1); or
  MyIntElement = 1;

[[D tIpcElement::tIpcElement_DT (tDataType dtInit) ]
[ invalid_pointer (this) => UNDEFINED
| true => this->dtElementData := dtInit
]]

Default destructor: destructs element object
Usage: occurs when object is no longer in current scope

[[D tIpcElement::~tIpcElement () ]
[ invalid_pointer (this) => UNDEFINED
| true => *this := anything
]]

Method Get: Data access to element  with error handling. 
  Get places element data into DataOut and returns error if any.
Usage:
  errorcode = MyIntElement.Get (MyInt);
  
[[D tIpcElement::Get (tDataType &DataOut) ]
[ invalid_pointer (this) ||
  simultaneous put on this => UNDEFINED
| true => 
    DataOut := this->dtElementData
    : return := OK 
]]

Method Put: Data asignment to element with error handling. 
  Put assigns element data to value of DataIn and returns error if any. 
  DataIn is not altered by Put.
Usage:
  errorcode = MyIntElement.Put (MyInt);

[[D tIpcElement::Put (const tDataType &DataIn) ]
[ invalid_pointer (this) || 
  simultaneous put on this => UNDEFINED
| true => 
    this->dtElementData := DataIn
    : return := OK 
]]
 
Operator new: If element with iiId does not exist: new allocates storage for a single element in local or shared memory 
  as specified by the iiId. (see ipcid.h). new elementatically provides szSize
  parameter and automatically calls the constructor (default or other 
  if specified) and returns a pointer to the newly constructed element.  
  If new is unable to allocate the element it returns NULL and sets ecErrCode
  to an appropriate error code. If element with iiId already exists, new returns
  a pointer to the pre-existing element. 

Usage: 
  tIpcElement<int> *pMyIntElement;
  pMyIntElement = new (MyIntElementId,ecErrCode) tIpcElement<int>; // using default constructor
  pMyIntElement = new (MyIntElementId,ecErrCode) tIpcElement<int> (10); // using non default constructor
    
[[D tIpcElement::operator_new (const tIpcAtomId &iiId, 
                        ERROR_CODE &ecErrCode) ]
[ invalid_reference (iiId) ||
  ::IpcMemoryG.bInit == FALSE => UNDEFINED
| CanAllocateIpcObject (::IpcMemoryG, iiId) == FALSE => 
    ecErrCode := appropriate error code
    : console := console + appropriate error messages
    : return := NULL
| true =>
    this->ecInitError := OK
    : this->bInit := TRUE
    : ::IpcMemoryG.ObjectSet := ::IpcMemoryG.ObjectSet U 
      NewObjectSet (::IpcMemoryG, iiId)
    : return := address of appropriate tIpcElement
]]

Operator new[]: If element array with iiId does not already exist, new 
  allocates storage for an array of elements in local or shared memory 
  as specified by the iiId. (see ipcid.h). new elementatically provides szSize
  parameter which is the aggregate size of the array and automatically calls 
  the constructor (default or other if specified) for each array element. 
  new[] returns a pointer to the newly constructed array of elements.  
  If new is unable to allocate the array it returns NULL and sets ecErrCode
  to an appropriate error code. 
  If element array with iiId already exists, new returns a pointer to the first
  element of the array. Note: the tIpcAtomId is the Id of the entire
  array.

Usage: 
  tIpcElement<int> *MyIntElementArray;
  // 10 element array using default constructor
  pMyIntElementArray = new (MyIntElementId,ecErrCode) tIpcElement<int> [10]; 
  // 10 element array using non default constructor
  pMyIntElement = new (MyIntElementId,ecErrCode) tIpcElement<int> (10) [10]; 
    
[[D tIpcElement::operator_array_new (const tIpcAtomId &iiId, 
                        ERROR_CODE &ecErrCode) ]
[ invalid_reference (iiId) ||
  ::IpcMemoryG.bInit == FALSE => UNDEFINED
| CanAllocateObject (::IpcMemoryG, iiId) == FALSE => 
    ecErrCode := appropriate error code
    : console := console + appropriate error messages
    : return := NULL
| true =>
    this->ecInitError := OK
    : this->bInit := TRUE
    : ::IpcMemoryG.ObjectSet := ::IpcMemoryG.ObjectSet U 
      NewObjectSet (::IpcMemoryG, iiId)
    : return := address of appropriate tIpcElement
]]

Operator tDataType(): Casts tIpcElement to tDataType.
Usage:
  int MyInt;
  tIpcElement<int> MyIntElement (10);
  // explicit type cast
  MyInt = (int) MyIntElement); 
  // implicit type cast
  MyInt = MyIntElement;
  
[[D tIpcElement::operator_tDataType_cast (void) ]
[ invalid_pointer (this) => UNDEFINED
| true => return := this->dtElementData
]]


*****************************************************/

template <class tDataType> class tIpcElement {
  public:
    tIpcElement ();
    tIpcElement (tDataType dtInit);
    ERROR_CODE Get (tDataType &DataOut) const volatile;
    ERROR_CODE Put (const tDataType &DataIn) volatile;
    void *operator new (size_t szSize,const tIpcAtomId &iiId, 
                        ERROR_CODE &ecErrCode);
    void *operator new[] (size_t szSize,const tIpcAtomId &iiId, 
                        ERROR_CODE &ecErrCode);
    operator tDataType() { return dtElementData;}                      
  protected:
    tDataType dtElementData;
  };
  
  
/***************************
Template Method Definitions
****************************/
/******************
Representation Mapping

::IpcMemoryG.bInit == FALSE maps to invalid_pointer (::pThisRteG) ||
  invalid_pointer (::pThisRteG->Ipc.pLocalMemory) ||
  invalid_pointer (::pThisRteG->Ipc.pSharedMemory) ||
  invalid_pointer (::pThisRteG->Ipc.pTokenClient) || 
  ::pThisRteG->Ipc.pSharedMemory->bStaticInit == FALSE  => UNDEFINED

****************/

/******************************************************************************
			tIpcElement Method Definitions
******************************************************************************/

/****
[[U tIpcElement::tIpcElement(void) ]]
****/  
template <class tDataType> tIpcElement<tDataType>::tIpcElement() {};

/****
[[U tIpcElement::Get(DataOut) ]]
****/  

template <class tDataType> ERROR_CODE tIpcElement<tDataType>::
  Get(tDataType &DataOut) const volatile {
    DataOut = this->dtElementData;
    return OK;
}


/****
[[U tIpcElement::Put(DataIn) ]]
****/  
template <class tDataType> ERROR_CODE tIpcElement<tDataType>::
  Put(const tDataType &DataIn) volatile {
    this->dtElementData = DataIn;
    return OK;
}

/****
[[U tIpcElement::operator_new(szSize,&iiId,ecErrCode) ]]
****/  
template <class tDataType> void *tIpcElement<tDataType>::
  operator new (size_t szSize,const tIpcAtomId &iiId, ERROR_CODE &ecErrCode) {

  TraceMark (S_object_CREATING_ATOM);

  /****
  [[U return := (::pThisRteG)->pIpc.tIpc::Allocate (iiId,szSize,ecErrCode) ]]
  ****/
  return GetIpc().Allocate(iiId,szSize,ecErrCode);
}

/****
[[U tIpcElement::operator_new(szSize,&iiId,ecErrCode) ]]
****/  
template <class tDataType> void *tIpcElement<tDataType>::
  operator new[] (size_t szSize,const tIpcAtomId &iiId, ERROR_CODE &ecErrCode) {

  TraceMark (S_object_CREATING_ATOM);

  /****
  [[U return := (::pThisRteG)->pIpc.tIpc::Allocate (iiId,szSize,ecErrCode) ]]
  ****/
  return GetIpc().Allocate(iiId,szSize,ecErrCode);
}



#endif INC_ipcElementh
