/* $RCSfile: lagLib.h,v $ - lag compensator */
/*
$Header: /home/chara/cvsroot/ople/sys/include/lagLib.h,v 1.1.1.1 1998/06/11 21:36:13 chara Exp $
*/
/* Brad Hines - Jet Propulsion Laboratory */

/*
modification history
--------------------

05a,13Feb98,rlj   added specification and additional accessors.
01a,21May97,beh   adapted from Palomar code
*/

#ifndef INC_lagLibh
#define INC_lagLibh

/*********** Specification Functions ***************************

double LimitValue (double x, double dLimit) :==
[ dLimit < 0 => return := anything
| x > dLimit => return := dLimit
| x < -dLimit => return := -dLimit
| true => return := x
]

double LAG (tLagCompensator Lag, double x) :=
[ invalid_pointer (this) ||
 Lag.dTauZ == 0 ||
 Lag.dTauP == 0 ||
 Lag.dSamplePeriod / Lag.dTauP == -1 => UNDEFINED
| true => 
    return := (1 + Lag.dSamplePeriod / Lag.dTauZ) / 
      (1 + Lag.dSamplePeriod / this->tauP * x + 
      (Lag.yLast - Lag.xLast) / (1 + dSamplePeriod / Lag.dTauZ) )
]

tLagCompensator UpdateLag (tLagCompensator Lag, double x) :==
[ invalid_pointer (this) ||
 Lag.dTauZ == 0 ||
 Lag.dTauP == 0 ||
 Lag.dSamplePeriod / Lag.dTauP == -1 => UNDEFINED
| true => 
    return := tLagCompensator with 
      xLast := x,
      Lag.yLast := LimitValue (
        (1 + Lag.dSamplePeriod / Lag.dTauZ) / 
        (1 + Lag.dSamplePeriod / Lag.tauP * x + 
        (Lag.yLast - Lag.xLast) / (1 + dSamplePeriod / Lag.dTauZ) ), Lag.limit)
]


**************** tLagCompensator Stimuli ****************************


[[D tLagCompensator::tLagCompensator (void) ]
Default constructor.
[ invalid_pointer (this) => UNDEFINED
| true => *this := anything
]]        


[[D tLagCompensator::tLagCompensator_ddd (double taup, double tauz, 
    double tsamp) ]
Constructs this with default limit. Computes lag coeficients from specifed pole, 
zero, and sample period.
[ invalid_pointer (this) ||
  tauz == 0 ||
  taup == 0 ||
  tsamp / taup == -1 => UNDEFINED
| true =>
    this->xLast := 0
    : this->yLast := 0
    : this->lim := 32766.99
    : this->dTauP := taup
    : this->dTauZ := tauz
    : this->dSamplePeriod := tsamp
]]        

[[D tLagCompensator::tLagCompensator_dddd (double taup, double tauz, 
    double tsamp, double lim) ]
Constructs this with a limit. Computes lag coeficients from specifed pole, 
zero, and sample period.
[ invalid_pointer (this) ||
  tauz == 0 ||
  taup == 0 ||
  tsamp / taup == -1 => UNDEFINED
| true =>
    this->xLast := 0
    : this->yLast := 0
    : this->lim := lim
    : this->dTauP := taup
    : this->dTauZ := tauz
    : this->dSamplePeriod := tsamp
]]      


[[D tLagCompensator::~tLagCompensator (void) ]
Default destructor.
[ invalid_pointer (this) => UNDEFINED
| true => *this := anything
]] 


[[D tLagCompensator::Initialize_ddd (double taup, double tauz, 
    double tsamp) ]
Constructs this with default limit. Computes lag coeficients from specifed pole, 
zero, and sample period. 
[ invalid_pointer (this) ||
  tauz == 0 ||
  taup == 0 ||
  tsamp / taup == -1 => UNDEFINED
| true =>
    this->xLast := 0
    : this->yLast := 0
    : this->lim := 32766.99
    : this->dTauP := taup
    : this->dTauZ := tauz
    : this->dSamplePeriod := tsamp
]]        


[[D tLagCompensator::Initialize_dddd (double taup, double tauz, 
    double tsamp, double lim) ]
Constructs this with a limit. Computes lag coeficients from specifed pole, 
zero, and sample period. 
[ invalid_pointer (this) ||
  tauz == 0 ||
  taup == 0 ||
  tsamp / taup == -1 => UNDEFINED
| true =>
    this->xLast := 0
    : this->yLast := 0
    : this->lim := lim
    : this->dTauP := taup
    : this->dTauZ := tauz
    : this->dSamplePeriod := tsamp
]]        
  

[[D tLagCompensator::TauP (void) ]
Returns lag compensator pole.
[ invalid_pointer (this) => UNDEFINED
| true => return := this->dTauP
]]   

[[D tLagCompensator::TauZ (void) ]
Returns lag compensator zero.
[ invalid_pointer (this) => UNDEFINED
| true => return := this->dTauZ
]] 
  
[[D tLagCompensator::SamplePeriod (void) ]
Returns lag compensator sample period.
[ invalid_pointer (this) => UNDEFINED
| true => return := this->dSamplePeriod
]] 

[[D tLagCompensator::Reset (double x) ]
Resets lag compensator memory.
[ invalid_pointer (this) => UNDEFNED
| true => 
    this->xLast := x
    : this->yLast := 0
]] 


[[D tLagCompensator::Calculate (double x) ]
Returns lag compensated value computed for x based on the current state of this.
[ invalid_pointer (this) ||
 this->dTauZ == 0 ||
 this->dTauP == 0 ||
 this->dSamplePeriod / this->dTauP == -1 => UNDEFINED
| true => 
    *this := UpdateLag (*this, x)
    : return:= LAG (*this, x)
]] 


[[D ::CopyLagMemory (tLagCompensator& Destination, const tLagCompensator& Source) ]
Copies the memory from on lag compensator to another to give continuity between
lag compensator with different pole/zero that control the same entity.
[ invalid_reference (Destination, Source) => UNDEFINED
| true => 
    Destination.xLast := Source.xLast
    : Destination.yLast := Source.yLast
]] 

*******/

class tLagCompensator {
  public:
    tLagCompensator () {} // left for back compatibility
    tLagCompensator (double taup, double tauz, double tsamp, double lim);
    tLagCompensator (double taup, double tauz, double tsamp);
    void Initialize (double taup, double tauz, double tsamp, double lim); // left for back compatibility
    void Initialize (double taup, double tauz, double tsamp); // left for back compatibility
    double TauP () const;
    double TauZ () const;
    double SamplePeriod () const;
    void Reset (double x);
    double Calculate (double x);
    friend void CopyLagMemory (tLagCompensator& Destination, const tLagCompensator& Source);
  private:
    double xLast;
    double yLast;
    double a; // maps to (1 + tsamp/tauz) / (1 + tsamp/taup)
    double b; // currently not used
    double c; // maps to 1 + tsamp/tauz
    double lim;
    double dSamplePeriod;
};

#endif  /* INC_lagLibh */
