#ifndef INC_matrixh

#define INC_matrixh

#include "tVector.h"

/****

The class tMatrix represents a matrix.  It supports

	matrix multiplication, addition, subtraction 
		(usual linear algebra definitions)
	multiplication of a matrix by a double
	multiplication of a matrix by a vector
	matrix inverse, transpose, determinant
	
The class also supports various techniques for solving simultaneous 
linear equations:
	
	svdcmp			singular value decomposition
	svbksb			singular value back substitution
	ludcmp			LU decomposition
	lubksb			LU back substitution
	LeastSquares	least-square solution
	
These methods encapsulate routines from "Numerical Recipes."  See that
book for more information on these techniques.

A matrix element is referenced by

	m[row][column]
	
where "m" is a tMatrix, and "row" and "column" are 0-based integers, and 
the result is a NUMBER (double).  That expression may be used on either 
side of the equal sign for a non-const tMatrix, but only on the right 
for a const tMatrix.

Other operations are also supported.  Whenever a size problem is 
detected, or whenever one of the linear-equation-solving methods fails 
to find a solution, a call is made to TraceMsg().  That behavior is 
localized to inline function,error(), in order that it may be easily 
modified.  

This class was derived (in the English sense) from a matrix 
class by Brad Hines.  It makes extensive use of class tVector.


Black Box State Variables
-------------------------------------- 
iSize             int                        number of rows
iColSize          int                        number of columns
pData             type*                      pointer to data


[[D tMatrix::tMatrix(void) ]
default constructor
[ invalid_pointer (this) => UNDEFINED
| true =>
   create an empty matrix
]]

[[D tMatrix::tMatrix_i(const int numRows) ]
constructor taking number of rows
[ invalid_pointer (this) => UNDEFINED
| numRows < 1 => UNDEFINED
| true =>
   create a matrix of numRows empty rows
]]

[[D tMatrix::tMatrix_ii(const int numRows, const int numCols) ]
constructor taking number of rows and columns
[ invalid_pointer (this) => UNDEFINED
| numRows < 1 => UNDEFINED
| numCols < 1 => UNDEFINED
| true =>
   create a numRows x numCols matrix of all zeros
]]

[[D tMatrix::tMatrix_m(const tMatrix& Matrix) ]
copy constructor
[ invalid_pointer (this) => UNDEFINED
| invalid_reference(Matrix) => UNDEFINED
| true =>
   create a deep copy of Matrix
]]

[[D tMatrix::tMatrix_v(const tVector& Vector) ]
type conversion from vector to matrix
[ invalid_pointer (this) => UNDEFINED
| Vector.iSize == 0 =>
   create an empty matrix
| true =>
   create a matrix whose single-column is Vector
]]

[[D tMatrix::~tMatrix() ]
destructor
[ invalid_pointer (this) => UNDEFINED
| true =>
   delete[] pData
]]

[[D tMatrix::addrow(void) ]
add a zero row
[ invalid_pointer (this) => UNDEFINED
| true =>
   add a zero row
   : return := *this
]]

[[D tMatrix::operator= (const tMatrix& Matrix) ]
assignment operator
[ invalid_pointer (this) => UNDEFINED
| invalid_reference(Matrix) => UNDEFINED
| true =>
   *this := deep copy of Matrix
   : return := *this
]]

[[D tMatrix::print(void) ]
format data for stdout
[ invalid_pointer (this) => UNDEFINED
| true =>
   format data for stdout
]]

[[D tMatrix::input(void) ]
read data from stdin
[ invalid_pointer (this) => UNDEFINED
| true =>
   read data from stdin
]]

[[D tMatrix::clear(void) ]
zero all elements
[ invalid_pointer (this) => UNDEFINED
| true =>
   zero all elements
]]

[[D tMatrix::operator[] (const int element) ]
index operator for non-const matrices: returns a vector
[ invalid_pointer (this) => UNDEFINED
| element < 0 || element >= this->numRows => ERROR
| true =>
   return := this->pData[element]
]]

[[D tMatrix::operator[]_c (const int element) ]
index operator for const matrices: returns a const vector
[ invalid_pointer (this) => UNDEFINED
| element < 0 || element >= this->numRows => ERROR
| true =>
   return := this->pData[element]
]]

[[D tMatrix::transpose(void) ]
returns the transpose of the matrix
[ invalid_pointer (this) => UNDEFINED
| true =>
   return := transpose of *this
]]

[[D tMatrix::inverse(void) ]
returns the inverse of the matrix
[ invalid_pointer (this) => UNDEFINED
| this->iSize == 0 || this->iSize != this->iColSize => ERROR
| true =>
   return := inverse of *this
]]

[[D tMatrix::operator*_m (const tMatrix& Matrix)]
returns the product of the matrices
[ invalid_pointer (this) => UNDEFINED
| invalid_reference(Matrix) => UNDEFINED
| this->iSize ==  0  || Matrix.iSize == 0 => ERROR
| this->ColSize != Matrix.iSize => ERROR
| true =>
   return := product of *this and Matrix
]]

[[D tMatrix::operator*_v (const tVector& Vector) ]
returns the product of the matrix and the row vector
[ invalid_pointer (this) => UNDEFINED
| invalid_reference(Vector) => UNDEFINED
| this->iSize ==  0  || Vector.iSize == 0 => ERROR
| this->ColSize != Vector.iSize => ERROR
| true =>
   return := product of *this and Vector
]]

[[D tMatrix::operator+ (const tMatrix& Matrix) ]
returns the sum of the matrices
[ invalid_pointer (this) => UNDEFINED
| invalid_reference(Matrix) => UNDEFINED
| this->iSize != Matrix.iSize || 
  this->iColSize != Matrix.iColSize => ERROR
| true =>
   return := *this + Matrix
]]

[[D tMatrix::operator+= (const tMatrix& Matrix) ]
add the specified matrix to *this
[ invalid_pointer (this) => UNDEFINED
| invalid_reference(Matrix) => UNDEFINED
| this->iSize != Matrix.iSize || 
  this->iColSize != Matrix.iColSize => ERROR
| true =>
   add Matrix to *this
   : return := *this
]]

[[D tMatrix::operator- (const tMatrix& Matrix) ]
returns the difference of the matrices
[ invalid_pointer (this) => UNDEFINED
| invalid_reference(Matrix) => UNDEFINED
| this->iSize != Matrix.iSize || 
  this->iColSize != Matrix.iColSize => ERROR
| true =>
   return := *this - Matrix
]]

[[D tMatrix::operator-= (const tMatrix& Matrix) ]
subtract the specified matrix from *this
[ invalid_pointer (this) => UNDEFINED
| invalid_reference(Matrix) => UNDEFINED
| this->iSize != Matrix.iSize || 
  this->iColSize != Matrix.iColSize => ERROR
| true =>
   subtract Matrix from *this
   : return := *this
]]

[[D tMatrix::operator-() ]
returns the negative of the matrix
[ invalid_pointer (this) => UNDEFINED
| true =>
   return := -(*this)
]]

[[D tMatrix::operator+(const NUMBER r) ]
returns a matrix with the specified number added to each element 
[ invalid_pointer (this) => UNDEFINED
| true =>
   return := a matrix with r added to each element
]]

[[D tMatrix::operator-(const NUMBER r) ]
returns a matrix with the specified number subtracted from each element 
[ invalid_pointer (this) => UNDEFINED
| true =>
   return := a matrix with r subtracted from each element

]]

[[D tMatrix::operator*(const NUMBER r) ]
returns a matrix with each element multiplied by the specified number
[ invalid_pointer (this) => UNDEFINED
| true =>
   return := a matrix with each element multiplied by r

]]

[[D tMatrix::operator/(const NUMBER r) ]
returns a matrix with each element divided by the specified number
[ invalid_pointer (this) => UNDEFINED
| r == 0 => UNDEFINED
| true =>
   return := a matrix with each element divided by r

]]

[[D tMatrix::operator+=(const NUMBER r) ]
adds the specified number to each element
[ invalid_pointer (this) => UNDEFINED
| true =>
   add r to each element of *this
   : return := *this
]]

[[D tMatrix::operator-=(const NUMBER r) ]
subtracts the specified number from each element
[ invalid_pointer (this) => UNDEFINED
| true =>
   subtract r from each element of *this
   : return := *this
]]

[[D tMatrix::operator*=(const NUMBER r) ]
multiplies each element by the specified number
[ invalid_pointer (this) => UNDEFINED
| true =>
   multiply each element of *this by r
   : return := *this
]]

[[D tMatrix::operator/=(const NUMBER r) ]
divides each element by the specified number
[ invalid_pointer (this) => UNDEFINED
| r == 0 => UNDEFINED
| true =>
   divide each element of *this by r
   : return := *this
]]

[[D tMatrix::Dimension(void) ]
returns the number of rows
[ invalid_pointer (this) => UNDEFINED
| true =>
   return := this->iSize
]]

[[D tMatrix::rows(void) ]
returns the number of rows
[ invalid_pointer (this) => UNDEFINED
| true =>
   return := this->iSize
]]

[[D tMatrix::columns(void) ]
returns the number of columns
[ invalid_pointer (this) => UNDEFINED
| true =>
   return := this->iColSize
]]

[[D tMatrix::ludcmp(tVector& indx, NUMBER& d) ]
performs LU decomposition on *this
[ invalid_pointer (this) => UNDEFINED
| invalid_reference(indx) || invalid_reference(d) => UNDEFINED
| indx.iSize != this->iSize => UNDEFINED
| result is singular => ERROR
| true =>
   perform LU decomposition on *this   
]]

[[D tMatrix::lubksb(tVector& indx, tVector& b) ]
perform LU back-substitution
[ invalid_pointer (this) => UNDEFINED
| invalid_reference(indx) || invalid_reference(b) => UNDEFINED
| true =>
   perform LU back-subsitution on *this
]]

[[D tMatrix::svdcmp(tVector& w, tMatrix& v) ]
performs singular value decomposition on *this
[ invalid_pointer (this) => UNDEFINED
| invalid_reference(w) || invalid_reference(v) => UNDEFINED
| this->iSize < this->iColSize => ERROR
| answer does not converge => ERROR
| true =>
   perform singular value decomposition on *this
]]

[[D tMatrix::svbksb(const tVector& w, const tMatrix& v, const tVector& b) ]
perform singular value back-substitution
[ invalid_pointer (this) => UNDEFINED
| invalid_reference(w) ||
  invalid_reference(v) ||
  invalid_reference(b) => UNDEFINED
| true =>
   perform singular value back-substitution
   : return := vector of coefficients  
]]

[[D tMatrix::LeastSquares(tVector& b, tVector& errorVec) ]
Returns x, where x is the best solution to: *this * x == b
[ invalid_pointer (this) => UNDEFINED
| invalid_reference(b) || invalid_reference(errorVec) => UNDEFINED
| this->iSize != b.iSize => ERROR
| true =>
   errorVec := error vector of solution to *this * x == b
   : return := solution to *this * x == b
]]

[[D tMatrix::det(void) ]
returns the determinant
[ invalid_pointer (this) => UNDEFINED
| this->iSize != this->iColSize || this->iSize < 1 => ERROR
| true =>
   return := determinant of *this
]]

[[D tMatrix::make_tVector(void) ]
explicit conversion from tVector to tMatrix
[ invalid_pointer (this) => UNDEFINED
| this->iSize != 1 && this->iColSize != 1 => ERROR
| true =>
   return := tVector whose elements are the single row or column of *this
]]

[[D ::operator*(const NUMBER r, const tMatrix& Matrix) ]
returns a matrix with each element multiplied by the specified number
[ invalid_pointer (this) => UNDEFINED
| invalid_reference(Matrix) => UNDEFINED
| true =>
   return := r * (*this)
]]

[[D ::Identity(const int dimension) ]
returns an identity matrix of the specified size
[ invalid_pointer (this) => UNDEFINED
| dimension <= 0 => UNDEFINED
| true =>
   returns := the (dimension x dimension) identity matrix
]]

****/


class tMatrix {

   public:
   
     tMatrix(void);
     explicit tMatrix(const int numRows);
     tMatrix(const int numRows, const int numCols);
     tMatrix(const tMatrix& Matrix);
     tMatrix(const tVector& Vector);
     ~tMatrix();
   
     tMatrix& addrow(void);
     tMatrix& operator= (const tMatrix&Matrix);
     void print(void) const;
     void input(void);
     void clear(void);
     tVector& operator[] (const int element);
     const tVector& operator[] (const int element) const;
     tVector make_tVector(void) const;   //intentionally explicit
     tMatrix transpose(void) const;
     tMatrix inverse(void) const;
     tMatrix operator* (const tMatrix& Matrix) const;
     tVector operator* (const tVector& tVector) const;
     tMatrix operator+ (const tMatrix& Matrix) const; 
     tMatrix& operator+= (const tMatrix& Matrix);
     tMatrix operator- (const tMatrix& Matrix) const;
     tMatrix& operator-= (const tMatrix& Matrix);
     tMatrix operator- (); 
     tMatrix operator+(const NUMBER r) const;
     tMatrix operator-(const NUMBER r) const;
     tMatrix operator*(const NUMBER r) const;
     tMatrix operator/(const NUMBER r) const; 
     tMatrix& operator+=(const NUMBER r); 
     tMatrix& operator-=(const NUMBER r); 
     tMatrix& operator*=(const NUMBER r); 
     tMatrix& operator/=(const NUMBER r); 
     int Dimension(void) const {return size;}
     int rows(void) const {return size;}
     int columns(void) const {return colSize;}
     void ludcmp(tVector& indx, NUMBER& d);
     void lubksb(tVector& indx, tVector& b);
     void svdcmp(tVector& w, tMatrix& v);
     tVector&svbksb(const tVector& w, const tMatrix& v, const tVector& b) const;
     tVector& LeastSquares(tVector& b, tVector& errorVec = tVector() );
     NUMBER det(void) const;
   
   private:
     int size, colSize;
     tVector *pointer;
};

tMatrix operator*(const NUMBER r, const tMatrix& Matrix);
tMatrix Identity(const int dimension);

#endif

