/* $RCSfile: modLib.h,v $ - modulation class header file */
/* CM Info:
$Header: /home/chara/cvsroot/ople/sys/include/modLib.h,v 1.1.1.1 1998/06/11 21:36:13 chara Exp $
*/
/* Beth McKenney - Jet Propulsion Laboratory */

/*
modification history
--------------------
01b,08Jul97,eamck   changed ints to doubles
01a,22Oct96,eamck   written
*/
      
/*
DESCRIPTION
  ISIT modulation library definitions
*/


#ifndef INC_modlibh
#define INC_modlibh

#include "isitExt.h"

#define S_modLib_INVALID_MOD_TYPE               (M_modLib | 0x1)
#define S_modLib_INVALID_MOD_FREQ               (M_modLib | 0x2)

typedef enum MOD_TYPE {
  MT_NONE, MT_SQUARE, MT_TRIANGLE, MT_SINE, MT_RANDOM, MT_WOBBLE
} MOD_TYPE ;


/*************Specification Functions****************************************

double SQUARE (double dAmp, double dMedian, TIME tPeriod, TIME tPhase,
               double dDuty ,TIME tTime) :==
Returns value of a square wave with specified parameters for time tTime.
[ 0 < tTime % tPeriod - tPhase <= dDuty * tPeriod =>
    return := dMedian + dAmp
| true => return := dMedian - dAmp
]

double SINE (double dAmp, double dMedian, TIME tPeriod, TIME tPhase,
             TIME tTime) :==
Returns value of a sine wave with specified parameters for time tTime.
[ true => 
    return := dMedian + dAmp * sin (2 *  PI * ( tTime - tPhase ) / tPeriod)
]

double TRIANGLE (double dAmp, double dMedian, TIME tPhase,
                 TIME tTau0, TIME tTau1, TIME tTime) :==
Returns value of triangle wave with specified parameters at time tTime.
[ 0 < tTime % (2 * (tTau0 + tTau1) ) - tPhase < tTau0 =>
    return := dMedian + dAmp / tTau0 * (tTime % (2 * (tTau0 + tTau1)) - tPhase)
  tTau0 <= tTime % (2 * (tTau0 + tTau1) ) - tPhase < tTau0 + 2 * tTau1 =>
    return := dMedian - dAmp / tTau1 * (tTime % (2 * (tTau0 + tTau1)) - tPhase - (tTau0 + tTau1) )
| true =>
    return := dMedian + dAmp / (tTau0) * 
      ( tTime % (tTau0 + tTau1) - tPhase - 2 * (tTau0 + tTau1) )
]


double WOBBLE (double dAmp, double dMedian, TIME tPhase,
                 TIME tTau0, TIME tTau1, TIME tTime) :==
Returns value of wobble wave with specified parameters at time tTime.
[ 0 < tTime % (2 * (tTau0 + tTau1)) - tPhase <= tTau0 =>
    return := dMedian + dAmp 
| tTau0 < tTime % (2 * (tTau0 + tTau1)) - tPhase <= tTau0 + tTau1 =>
    return := dMedian
| tTau0 + tTau1 < tTime % (2 * (tTau0 + tTau1)) - tPhase <= 2*tTau0 + tTau1 =>
    return := dMedian - dAmp
| 2*tTau0 + tTau1 < tTime % (2 * (tTau0 + tTau1)) - tPhase <= 2 * (tTau0 + tTau1) =>
    return := dMedian
]


TIME TRIANGLE_TAU0 (double dFreq, double dDuty) :==
Returns the rise time of a triangle wave with frequence dFreq and Duty Cycle dDuty
[ dFreq == 0 => return := 0
| true => return := TIME_TICKS_PER_SECOND * dDuty / (2 * dFreq)
]


TIME TRIANGLE_TAU1 (double dFreq, double dDuty) :==
Returns the fall time of a triangle wave with frequence dFreq and Duty Cycle dDuty
[ dFreq == 0 => return := 0
| true => return := TIME_TICKS_PER_SECOND * (1 - dDuty) / (2 * dFreq)
]


TIME WOBBLE_TAU0 (double dFreq, double dDuty) :==
Returns the time that a wobble wave has positive values relative to the median.
[ dFreq == 0 => return := 0
| true => return := TIME_TICKS_PER_SECOND * dDuty / dFreq / 2
]


TIME WOBBLE_TAU1 (double dFreq, double dDuty) :==
Returns the time that a wobble wave has negative values relative to the median.
[ dFreq == 0 => return := 0
| true => return := TIME_TICKS_PER_SECOND * (1 - dDuty) / dFreq / 2
]

double MODULATION (tModulationParams mpParams, TIME tUpdateTime) :==
Returns the value of an enabled periodic function with mpParams at time tUpdateTime.
[ mpParams.mtModulationType == MT_NONE => return := 0
| mpParams.mtModulationType == MT_SQUARE => 
    return := SQUARE (mpParms.dAmplitude, mpParams.dMeanVal,
      TIME_TICKS_PER_SECOND/mpParams.dFrequency, mpParams.tSyncTime,
      mpParams.dDutyCycle, tUpdateTime)
| mpParams.mtModulationType == MT_SINE => 
    return := SINE (mpParms.dAmplitude, mpParams.dMeanVal,
      TIME_TICKS_PER_SECOND/mpParams.dFrequency, mpParams.tSyncTime,
      tUpdateTime)
| mpParams.mtModulationType == MT_TRIANGLE => 
    return := TRIANGLE (mpParms.dAmplitude, mpParams.dMeanVal,
      TIME_TICKS_PER_SECOND/mpParams.dFrequency, mpParams.tSyncTime,
      TRIANGLE_TAU0 (mpParams.dFrequency, mpParams.dDutyCycle),
      TRIANGLE_TAU1 (mpParams.dFrequency, mpParams.dDutyCycle),
      tUpdateTime )
| mpParams.mtModulationType == MT_WOBBLE => 
    return := WOBBLE (mpParms.dAmplitude, mpParams.dMeanVal,
      TIME_TICKS_PER_SECOND/mpParams.dFrequency, mpParams.tSyncTime,
      WOBBLE_TAU0 (mpParams.dFrequency, mpParams.dDutyCycle),
      WOBBLE_TAU1 (mpParams.dFrequency, mpParams.dDutyCycle),
      tUpdateTime )
| true => return := 0
]

*************Stimuli *************************************************************


[[D tModulationParams::tModulationParams (void)]
Constructs this with default parameters.
[ invalid_pointer (this) => UNDEFINED
| true =>
    this->mtModulationType := MT_NONE
    : this->dAmplitude := 0
    : this->dFrequency := 0
    : this->dMeanVal := 0
    : this->tSyncTime := 0
    : this->dDutyCycle := 0
]]

[[D tModulationParams::tModulationParams_MTdd (MOD_TYPE mtType, 
      double dFrequencyIn, double dAmplitudeIn) ]
Constructs this with specified and default parameters.
[ invalid_pointer (this) => UNDEFINED
| true =>
    this->mtModulationType := mtType
    : this->dAmplitude := dAmplitudeIn
    : this->dFrequency := dFrequencyIn
    : this->dMeanVal := 0
    : this->tSyncTime := 0
    : this->dDutyCycle := 0.5
]]


[[D tModulationParams::tModulationParams_MTddd  (MOD_TYPE mtType, 
      double dFrequencyIn, double dAmplitudeIn, double dMeanValIn) ]
Constructs this with specified and default parameters.
[ invalid_pointer (this) => UNDEFINED
| true =>
    this->mtModulationType := mtType
    : this->dFrequency := dFrequencyIn
    : this->dAmplitude := dAmplitudeIn
    : this->dMeanVal := dMeanValIn
    : this->tSyncTime := 0
    : this->dDutyCycle := 0.5
]]


[[D tModulationParams::tModulationParams_MTdddt (MOD_TYPE mtType,  
      double dFrequencyIn, double dAmplitudeIn, double dMeanValIn, 
      TIME tSyncTimeIn) ]
Constructs this with specified and default parameters.
[ invalid_pointer (this) => UNDEFINED
| true =>
    this->mtModulationType := mtType
    : this->dFrequency := dFrequencyIn
    : this->dAmplitude := dAmplitudeIn
    : this->dMeanVal := dMeanValIn
    : this->tSyncTime := tSyncTimeIn
    : this->dDutyCycle := 0.5
]]


[[D tModulationParams::tModulationParams_MTdddtd (MOD_TYPE mtType,  
      double dFrequencyIn, double dAmplitudeIn, double dMeanValIn, 
      TIME tSyncTimeIn, double dDutyCycleIn) ]
Constructs this with specified parameters.
[ invalid_pointer (this) => UNDEFINED
| true =>
    this->mtModulationType := mtType
    : this->dFrequency := dFrequencyIn
    : this->dAmplitude := dAmplitudeIn
    : this->dMeanVal := dMeanValIn
    : this->tSyncTime := tSyncTimeIn
    : this->dDutyCycle := dDutyCycleIn
]]


[[D tModulationParams::~tModulationParams (void)]
[ invalid_pointer (this) => UNDEFINED
| true => *this := anything
]]

*************************tModulation****************************************


[[D tModulation::tModulation (void) ]
Constructs this with default modulation parameters.
[ invalid_pointer (this) => UNDEFINED
| true =>
    this->mpParams.mtModulationType := MT_NONE
    : this->mpParams.dAmplitude := 0
    : this->mpParams.dFrequency := 0
    : this->mpParams.dMeanVal := 0
    : this->mpParams.tSyncTime := 0
    : this->mpParams.dDutyCycle := 0
    : this->bActive := FALSE
]]

[[D tModulation::tModulation_t (const tModulationParams& InitialParams) ]
Constructs this with initial parameters.
[ invalid_pointer (this) ||
  invalid_reference (InitialParams) => UNDEFINED
| true =>
    this->mpParams := InitialParams
    : this->bActive := FALSE
]]


[[D tModulation::tModulation_MTdddtd (MOD_TYPE mtType, 
      double dFrequencyIn, double dAmplitudeIn, double dMeanValIn, 
      TIME tSyncTimeIn, double dDutyCycleIn)]
Constructs this with initial parameters.
[ invalid_pointer (this) => UNDEFINED
| true =>
    this->mpParams.mtModulationType := mtType
    : this->mpParams.dAmplitude := dAmplitudeIn
    : this->mpParams.dFrequency := dFrequencyIn
    : this->mpParams.dMeanVal := dMandValIn
    : this->mpParams.tSyncTime := tSyncTimeIn
    : this->mpParams.dDutyCycle := dDutyCycleIn
    : this->bActive := FALSE
]]


[[D tModulation::~tModulation (void)]
[ invalid_pointer (this) => UNDEFINED
| true => *this := anything
]] 


[[D tModulation::SetParameters_MTdddtd (MOD_TYPE mtType,
    double dFrequency, mdouble dAmplitude, double dMeanVal,
    TIME tSyncTime, double dDutyCycle) ]
Set this->mpParams with specified parameters.
[ invalid_pointer (this) => UNDEFINED
| true =>
    this->mpParams.mtModulationType := mtType
    : this->mpParams.dAmplitude := dAmplitudeIn
    : this->mpParams.dFrequency := dFrequencyIn
    : this->mpParams.dMeanVal := dMandValIn
    : this->mpParams.tSyncTime := tSyncTimeIn
    : this->mpParams.dDutyCycle := dDutyCycleIn
    : return := OK
]] 

[[D tModulation::SetParameters_t (const tModulationParams& NewModParams)]
Sets this->mpParms with specified parameters
[ invalid_pointer (this) || 
  invalid_reference (NewModParams) => UNDEFINED
| true =>
    this->mpParams := NewModParams
    : return := OK
]]


[[D tModulation::GetParameters (void)]
[ this invalid => UNDEFINED
| true =>
    return := this->mpParams
]]


[[D tModulation::ComputeModulation (TIME tUpdateTime)]
Compute modulation value for given time
[ invalid_pointer (this) => UNDEFINED
| this->bActive == FALSE => return := 0
| true => return := MODULATION (this->mpParams, tUpdateTime) 
]]


[[D tModulation::Enable (void)]
Turn on modulation
[ invalid_pointer (this) => UNDEFINED
| true => this->bActive := TRUE 
]]

[[D tModulation::Disable (void)]
Turn off modulation
[ invalid_pointer (this) => UNDEFINED
| true => this->bActive := FALSE 
]]

[[D tModulation::IsEnabled (void)]
[ invalid_pointer (this) => UNDEFINED
| true => return := this->bActive
]]


[[D tModulation::Reset (void)]
Resets modulation to default values and disables this.
[ invalid_pointer (this) => UNDEFINED
| true =>
    this->mpParams.mtModulationType := MT_NONE
    : this->mpParams.dAmplitude := 0
    : this->mpParams.dFrequency := 0
    : this->mpParams.dMeanVal := 0
    : this->mpParams.tSyncTime := 0
    : this->mpParams.dDutyCycle := 0
    : this->bActive := FALSE
]]


[[D tModulation::SyncNow (void) ]
Offsets modulation waveform so that the phase shift is set to the current time
[ invalid_pointer (this) => UNDEFINED
| system clock not valid => this->mpParams.tSyncTime := anything
| true => this->mpParams.tSyncTime := current system time
]]

[[D tModulation::SyncAt (TIME tTime) ]
Offsets modulation waveform so that the phase shift is set to tTime
[ invalid_pointer (this) => UNDEFINED
| true => this->mpParams.tSyncTime := tTime
]]


*****************************Private method spec**************************************

[[D tModulation::ComputePeriod (double dFreq) ]
[ invalid_pointer (this) => UNDEFINED
| dFreq == 0 => this->tPeriod := 1
| true => this->tPeriod := TIME_TICKS_PER_SECOND / dFreq
]]


[[D tModulation::ComputeTaus (void) ]
[ invalid_pointer (this) => UNDEFINED
| this->mtModulationType == MT_SQUARE =>
    this->tTau1 := this->dDutyCycle * this->tPeriod
    : this->tTau0 := this->tPeriod - (this->dDutyCycle * this->tPeriod)
| this->mtModulationType == MT_WOBBLE =>
    this->tTau1 := this->dDutyCycle * this->tPeriod / 2
    : this->tTau0 := (this->tPeriod - * this->dDutyCycle * this->tPeriod) / 2
| true =>
    this->tTau1 := 0
    : this->tTau0 := this->tPeriod
]]
*****/


/**************** tModulationParams ******************************************/

// Inlined so that the client code can declare and use...  

class tModulationParams {
  public:
    tModulationParams (void) {
      this->mtModulationType = MT_NONE;
      this->dFrequency = 1;
      this->dAmplitude = 0;
      this->dMeanVal = 0;
      this->tSyncTime = 0;
      this->dDutyCycle = 0.5;
    }
     tModulationParams (MOD_TYPE mtType, double dFrequencyIn, double dAmplitudeIn) {
        this->mtModulationType = mtType;  // no error checking here
        this->dFrequency = dFrequencyIn;
        this->dAmplitude = dAmplitudeIn;
        this->dMeanVal = 0.0;
        this->tSyncTime = 0;
        this->dDutyCycle = 0.5;
      }
     tModulationParams (MOD_TYPE mtType, double dFrequency, 
           double dAmplitude, double dMeanVal) {
        this->mtModulationType = mtType;  // no error checking here
        this->dFrequency = dFrequency;
        this->dAmplitude = dAmplitude;
        this->dMeanVal = dMeanVal;
        this->tSyncTime = 0;
        this->dDutyCycle = 0.5;
      }
      tModulationParams (MOD_TYPE mtType, double dFrequency, 
          double dAmplitude,  double dMeanVal, TIME tSyncTime) {
        this->mtModulationType = mtType;  // no error checking here
        this->dFrequency = dFrequency;
        this->dAmplitude = dAmplitude;
        this->dMeanVal = dMeanVal;
        this->tSyncTime = tSyncTime;
        this->dDutyCycle = 0.5;
      }
     tModulationParams (MOD_TYPE mtType, double dFrequency, double dAmplitude, 
         double dMeanVal, TIME tSyncTime, double dDutyCycle) {
        this->mtModulationType = mtType;  // no error checking here
        this->dFrequency = dFrequency;
        this->dAmplitude = dAmplitude;
        this->dMeanVal = dMeanVal;
        this->tSyncTime = tSyncTime;
        this->dDutyCycle = dDutyCycle;
      }
    MOD_TYPE mtModulationType;
    double   dFrequency;
    double   dAmplitude;
    double   dMeanVal;
    TIME     tSyncTime;
    double   dDutyCycle;
};  


/***************************** tModulation *****************************************/

class tModulation {
  public:
    tModulation (void);
    tModulation (const tModulationParams& InitialParams);
    tModulation (MOD_TYPE mtType, double dFrequency, 
                 double dAmplitude, double dMeanVal=0, TIME tSyncTime=0,
                 double dDutyCycle = 0.5);
    ERROR_CODE SetParameters (MOD_TYPE mtType, double dFrequency,
	                      double dAmplitude, double dMeanVal=0,
	                      TIME tSyncTime=0, double dDutyCycle = 0.5);
    ERROR_CODE SetParameters (const tModulationParams& NewModParams);
    void Reset (void);
    const tModulationParams& GetParameters () const;
    void Enable ();
    void Disable ();
    BOOL IsEnabled () const;
    void SyncNow ();
    void SyncAt (TIME tNewSyncTime);
    double ComputeModulation (TIME tUpdateTime);
  
  private:
    void ComputePeriod (double dFreq);
    void ComputeTaus (); // compute Taus from current duty cycle

    tModulationParams mpParams;  
    BOOL bActive;
    TIME tPeriod; 
    TIME tTau0; // used for duty cycling
    TIME tTau1; // used for duty cycling 
}; 

#endif INC_modlibh

