/* $RCSfile: mutex.h,v $ - implements mutually exclusive access to a resource */
/* CM Info:
$Header: /home/chara/cvsroot/ople/sys/include/mutex.h,v 1.1.1.1 1998/06/11 21:36:13 chara Exp $
*/
/* Brad Hines - Jet Propulsion Laboratory */
 
/*
modification history
--------------------
01a,beh,28Sep96    written
*/
 
/*
DESCRIPTION
When several tasks want to use the same resource (such as various
DL commands like Front and Track, which both want to use the
physical delay line), a means of ensuring mutually exclusive
access to that resource is required.

It is desirable that this mechanism be flexible, allowing a
task to pre-empt another task (causing it to cleanly abort)
and to implement timeouts for operations.
*/

#ifndef INC_mutexh
#define INC_mutexh

#ifdef VXWORKS 
 #ifndef SPECTOOL
   #include <sysLib.h>
   #include "isit.h"
   #include "watchdog.h"
 #endif
#endif

#include "errorStream.h"

#define S_mutexUser_MUTEX_TASK_PRIORITY_TOO_HIGH                  (M_mutex | 0x1)
#define S_mutexUser_ABORT_ATTEMPTED_ON_INACTIVE_MUTEX_USER_OBJECT (M_mutex | 0x2)


/******************************************************************
*
* tMutex Black Box
*
* You don't do much with tMutex's other than create them and
* pass them in to tMutexUsers when creating them.  tMutex's
* are essentially mutual exclusion semaphores which are used
* to prevent more than one tMutexUser from simultaneously
* accessing the same shared resource.
*
* BB State Variables
*
* Variable                   Initial Value
* ----------------------------------------
* tMutexUser *pOwner         NULL
*
*******************************************************************
*
* tMutex::tMutex
*
* SPEC: 

[[D tMutex::tMutex (void) ] 
[ invalid_pointer (this) => UNDEFINED
| true => this->pOwner := NULL 
]]
 
*
*******************************************************************
*
* tMutex::~tMutex
*
* SPEC: (if called via delete)

[[D tMutex::~tMutex (void) ]
[ invalid_pointer (this) => UNDEFINED
| true =>
    Terminate the ExecuteMethod of and abort any current this->pOwner
    : *this := anything
]]
*       
*
*******************************************************************
*
* BOOL IsOwner(tMutexUser *pPossibleOwner)
*
* SPEC:
    [ this invalid -> UNDEFINED
    | this.pOwner == NULL -> return FALSE
    | this.pOwner != pPossibleOwner -> return FALSE
    | true -> return TRUE
    ]
*
*******************************************************************
*
*/

class tMutexUser;

class tMutex {
public:
  tMutex(void);
  ~tMutex(void);
  BOOL IsOwner(tMutexUser *pPossibleOwner);

protected:
  friend class tMutexUser;
  ERROR_CODE Take(tMutexUser *pNewOwner, int iTimeoutInTicks,
                  BOOL bKillCurrentOwnerIfTimeout,
                  int iTimeToAllowForOperation);
  ERROR_CODE Give(void);
  
  /************
    State Box state:

    Representational mappings:
      1. [ siSemaphore == NULL <-> bInit == FALSE
         | siSemaphore != NULL <-> bInit == TRUE
         ]
   ************/

  SEM_ID siSemaphore;
  tMutexUser *pOwner;
  tTaskLevelWatchDog WatchDog;

  ERROR_CODE WatchDogHandler(void);
};




/******************************************************************
*
* tMutexUser Black Box
*
*  State Variable Dictionary
*  Variable                   Initial Value
*  ----------------------------------------
*  pMutex                     Unknown
*  pErrorClient               Unknown
*  sTaskName                  NULL
*  iTaskOptions               Unknown
*  iTaskStackSize             Unknown
*  iTaskPriority              Unknown
*  
*******************************************************************
* 
* tMutexUser::tMutexUser(void) - default constructor
*
* NOT FOR GENERAL USE - provided merely to allow an tMutexUser
* object to be declared if need be prior to actual initialization.
* You must call the other constructor before beginning to use
* the tMutexuser object.
*
* IN:   this
* OUT:  this.sTaskName
* SPEC: 

[[D tMutexUser::tMutexUser (void) ]
[ true => this->sTaskName := NULL 
]]

* RETURNS: Nothing
*
*******************************************************************
*
* tMutexUser::tMutexUser(tMutex *pMutexToUse, tGizmo *pGizmoToUse,
*                        char *sTaskNameToUse)
*
* Constructor - prepares a tMutexUser for operation
*
* Note!  In general, you should always use new to create tMutexUsers.
* If a tMutexUser is deleted while an Execute is in progress,
* scary things happen.  Also, in general, you should never delete a
* tMutexUser.  They delete themselves when Execute completes.  For
* this reason, you should also never do anything else with the
* tMutexUser once you call Execute, since the object can disappear
* at any time.
*
* IN:
*   pMutexToUse
*   pGizmoToUse
*   sTaskNameToUse
*   *sTaskNameToUse (if sTaskName not NULL)
* INOUT:
*   system heap
* OUT:
*   this.pMutex
*   this.pErrorClient
*   this.sTaskName
*   this.iTaskOptions
*   this.iTaskStackSize
*   this.iTaskPriority
*
* SPEC:
*

[[D tMutexUser::tMutexUser_tts (tMutex *pMutexToUse, tGizmo *pGizmoToUse,
                            char *sTaskNameToUse) ]
[ this or pMutexToUse or pGizmoToUse invalid => UNDEFINED
| sTaskNameToUse != NULL && sTaskName invalid or not a pointer to a
                         valid (null-terminated) string => UNDEFINED
| true =>
       this->pMutex := pMutexToUse
       : this->pErrorClient := pGizmoToUse->pRteError
       : sTaskName := copy of sTaskNameToUse if non-NULL, else "MutexTask"
       : this->iTaskOptions     := MUTEX_TASK_OPTIONS
       : this->iTaskStackSize   := MUTEX_TASK_STACK_SIZE
       : this->iTaskPriority    := MUTEX_TASK_PRIORITY
]]
 
*
* RETURNS:
*   Nothing
*
*******************************************************************
*
* tMutexUser::~tMutexUser(void) - destructor
*
* FOR INFORMATIONAL PURPOSES ONLY!  Well, sort of.  You generally
* should not call delete on a tMutexUser.  tMutexUsers delete
* themselves once ExecuteMethod (and possibly AbortMethod) complete.
*
* The only time you would do a delete is if you create a tMutexUser
* and then decide not to use it without ever calling Execute().
* I recommend you don't do that, though.
*
* IN:
*   this
*   this.sTaskName
* INOUT:
*   system heap
*
* SPEC: (if called via delete)
*
*  
[[D tMutexUser::~tMutexUser (void) ]
[ this invalid => UNDEFINED
| this->sTaskName != NULL =>
    return *this->sTaskName storage to system heap
    : return *this storage to system heap
    : *this := anything
| true =>
    return *this storage to system heap
    : *this := anything
]]
*
* RETURNS:
*   Nothing
*
*******************************************************************
*
* tMutexUser::Execute(int iTicksToWaitForMutex,
*                     BOOL bKillCurrentOwnerIfTimeout,
*                     int iTicksToAllowForOperation)
*
* IN:
*   iTicksToWaitForMutex       (defaults to NO_WAIT)
*   bKillCurrentOwnerIfTimeout (defaults to TRUE)
*   iTicksToAllowForOperation  (defaults to MUTEX_DEFAULT_TIMEOUT)
*   this
*   this.iTaskOptions
*   this.iTaskStackSize
*   this.iTaskPriority
*   this.pMutex
*
* INOUT:
*   this.*pMutex
*   OS.SpawnedTasks
*
* SPEC:  (a bit hand-wavy in the true case, but this is what it does)
*
[[D tMutexUser::Execute(int iTicksToWaitForMutex,
                        BOOL bKillCurrentOwnerIfTimeout,
                        int iTicksToAllowForOperation)  ]
[ this invalid ||
  bKillCurrentOwnerIfTimeout != FALSE ||
  this.pMutex invalid ||
  *this.pMutex cannot be acquired in iTicksToWaitForMutex ticks ||
  somebody deletes this object while ExecuteMethod is running => UNDEFINED
| true =>
     task is spawned to run this.ExecuteMethod() with
       options this.iTaskOptions,
       name this.sTaskname,
       stack size this.iTaskStackSize,
       and priority this.iTaskPriority
     : [ this.ExecuteMethod() does not complete in
         iTicksToAllowForOperation ticks =>
           ExecuteMethod is halted after iTicksToAllowForOperation ticks
           : this.AbortMethod() is called from a *different* task context
       ]
]]
*
*   Note: While this->ExecuteMethod() (and this->AbortMethod()) are running,
*   this.*pMutex.pOwner := this, and this.*pMutex cannot be acquired by
*   other tMutexUsers.
*
*******************************************************************
*
* protected tMutexUser::SetTaskOptions(int options, int stackSize,
*                                      int priority)
*
* Intended to be called by a derived class's constructor if that
* class wants to modify the default properties of the task
* spawned to run ExecuteMethod.
*
* This function sets the options that will be used by taskSpawn
* when a tMutexUser ExecuteMethod is spawned.  It is normally
* called automatically and is not required to be called by the
* user.  However, the user may wish to call this method if he
* wants something other than the default task options.
*
* IN:
*   this
*   options    (defaults to MUTEX_TASK_OPTIONS)
*   stackSize  (defaults to MUTEX_TASK_STACK_SIZE)
*   priority   (defaults to MUTEX_TASK_PRIORITY)
* INOUT:
* OUT:
*   this.iTaskOptions
*   this.iTaskStackSize
*   this.iTaskPriority
*
* SPEC:
*
   [ this invalid -> UNDEFINED
   | true ->
       this.iTaskOptions     := options
       : this.iTaskStackSize := stackSize
       : this.iTaskPriority  := priority
       : return OK
   ] 
*
* RETURNS:
*   OK
*
*******************************************************************
*
* protected tMutexUser::ExecuteMethod(void)
* protected tMutexUser::AbortMethod(void)
*
* ExecutedMethod is a virtual method, which you are intended to
* overload, which is essentially spawned off as a task for you
* when Execute is called.  This routine MUST be prepared for
* the fact that it may be terminated at any time without warning.
* In the general case, this requires that a derived class keep
* information on the progress of ExecuteMethod so that things can
* be cleaned up if it is terminated.  If it is terminated, AbortMethod
* will be called (although in a *different* task context from
* ExecuteMethod).  AbortMethod is then responsible for restoring
* to an acceptable state all system resources that were used by the
* derived class's ExecuteMethod.
*
* If there is a region of code that absolutely cannot tolerate
* task deletion, use paired calls to taskSafe/taskUnsafe to
* protect the critical region.
*/


class tGizmo;

class tMutexUser {
public:
  tMutexUser(void) { sTaskName = NULL; }
  tMutexUser(tMutex *pMutexToUse, tGizmo *pGizmoToUse,
             char *sTaskNameToUse);

  ERROR_CODE Execute(int iTicksToWaitForMutex = NO_WAIT,
                     BOOL bKillCurrentOwnerIfTimeout = TRUE,
                     int iTicksToAllowForOperation = MUTEX_DEFAULT_TIMEOUT);

protected:
  virtual ERROR_CODE ExecuteMethod(void) { return OK; }
  virtual ERROR_CODE AbortMethod(void) { return OK; }
  
  SetTaskOptions(int options = MUTEX_TASK_OPTIONS,
                 int stackSize = MUTEX_TASK_STACK_SIZE,
                 int priority = MUTEX_TASK_PRIORITY);
  
  tMutex *pMutex;
  char *sTaskName;
  int iTaskOptions;
  int iTaskStackSize;
  int iTaskPriority;
  int iTaskId;
  SEM_ID siAbortSemaphore;
  tErrorStreamSender *pErrorClient;

private:
  ERROR_CODE DoTheOperation(int iTicksToWaitForMutex,
                            BOOL bKillCurrentOwnerIfTimeout,
                            int iTicksToAllowForOperation);
  ERROR_CODE CleanUp(void);
  public: // But should be considered private!
  ERROR_CODE Abort(int iTicksToWaitForAbortSemaphore = MUTEX_DEFAULT_TIMEOUT);
  virtual ~tMutexUser(void);
};


#endif   /* INC_mutexh */
