/*
$RCSfile: objectId.h,v $
$Header: /home/chara/cvsroot/ople/sys/include/objectId.h,v 1.1.1.1 1998/06/11 21:36:13 chara Exp $

$log$
****/


#ifndef INC_objectIdh
#define INC_objectIdh

#include "isit.h"

/*************** Description **********************************************



tObjectId provides an object instance identification hierarchy.  

Each instance of a tObjectId has a string sub-name and a pointer to a parent
tObjectId node.  If the object id has no parent the parent node must be set to 
NULL.  

The sub-name can be any valid null terminated string but should uniquely 
identify the type of the object to which it belongs (class names are probably
the best name to use).

The object id is defined as a concatenation of all subnames from the parent to
the tObjectId node of interest delimited by '.'

For example:  "tRte.tDlGizmo.tDlCompleteControllerSet.tDlPztController"

It may be useful for the subnames to indicate instance numbers .
For example:  "tRte[0].tDlGizmo[2].tDlCompleteIoSet.tDlPztController"
where '[n]' denotes the instance number.  '[n]' is added by the creator of the
tObjectId as part of the specified subname, however the '.' is automatically 
inserted between sub-names.


**************** Specification Functions *************************************

BOOL invalid_string (char* sString) :==
[ invalid_pointer (sString) ||
  sString not a null terminated array of char => return := TRUE
| true => return := FALSE
]

int StringLength  (char* sString) :==
[ invalid_string (sString) => UNDEFINED
| true => return := length of sString excluding null terminator
]

int ObjectIdDepth (tObjectId* pObjectId) :==
Returns the number of subnames of the object id.
[ wild_ponter (pObjectId, pObjectId->pParentId, ...) => UNDEFINED
| pObjectId == NULL => return := 0
| true => return := 1 + ObjectIdDepth (pObjectId->pParentId)
] 

int ObjectIdLength (tObjectId* pObjectId) :==
Returns length of complete object id string from the parent to pObjectId.
[ wild_ponter (pObjectId, pObjectId->pParentId, ...) ||
  invalid_string (pObjectId->sSubName, 
    pObjectId->pParentId->sSubName, ...) => UNDEFINED
| pObjectId == NULL => return := 0
| true => return := StringLength (pObjectId->sSubName) + 
             ObjectIdLength (pObjectId->pParentId)
]


string ParentName (tObjectId* pObjectId) :==
Returns the full object name of the parent.
[ invalid_pointer (pObjectId) ||
  wild_ponter (pObjectId->pParentId, pObjectId->pParentId->pParentId, ...) ||
  invalid_string (pObjectId->sSubName, 
    pObjectId->pParentId->sSubName, ...) => UNDEFINED
| pObjectId->pParentId == NULL => return := this->sSubName
| true => return := ParentName (this->pParentId) + *this->sSubName
]

****************** tObjectId Specification **********************************


[[D tObjectId::tObjectId (char* sSubNameIn, const tObjectId* pParentIdIn) ]
Construct this with specified subname and parent.
[ invalid_pointer (this) || 
  wild_pointer (pParentIdIn) ||
  invalid_string (sSubNameIn) => UNDEFINED
| pParentId == NULL =>
    this->sSubName := new string initialized to *sSubNameIn
    : this->pParentId := NULL
| true =>
    this->sSubName := new string initialized to '.' + *sSubNameIn
    : this->pParentId := pParentIdIn
]]

[[D tObjectId::~tObjectId () ]
Destruct this freeing memory used to store subname.
[ invalid_pointer (this) ||
  invalid_string (this->sSubNameIn) => UNDEFINED
| true =>
    Resources consumed by this->sSubName freed
    : *this := anything
]]


[[D tObjectId::IdDepth (void) ]
Returns the depth of the parent hierarchy.
[ invalid_pointer (this)
  wild_ponter (this->pParentId, this->pParentId->pParentId,...) => UNDEFINED
| true => return := ObjectIdDepth (this)
]]

[[D tObjectId::IdLength (void) ]
Returns the length of the complete object id string. The additional character 
accounts for the null terminator.
[ invalid_pointer (this)
  wild_ponter (this->pParentId, this->pParentId->pParentId, ...) ||
  invalid_string (this->sSubName, this->pParentId->sSubName, ...) => UNDEFINED
| true => return := ObjectIdLength (this) + 1
]]

[[D tObjectId::ObjectId (char* sObjectId, int iMaxChar) ]
Returns a copy of the complete object id string up to iMaxChar characters 
including the null terminator.
[ invalid_pointer (this, sObjectId) ||
  sObjectId doesn't have iMaxChar char reserved ||
  wild_ponter (this->pParentId, this->pParentId->pParentId, ...) ||
  invalid_string (this->sSubName, this->pParentId->sSubName, ...) => UNDEFINED
| iMaxChar < ObjectIdLength (this) => *sObjectId := anything
| true => 
    sObjectId := deep copy of ParentName (this, iMaxChar - 1) + null terminator
]]

********/

class tObjectId {
  public:
    tObjectId (const char* sSubNameIn, const tObjectId* pParentIdIn);
    ~tObjectId ();
    int IdDepth () const;
    int IdLength () const;
    void ObjectId (char* sObjectId, size_t iMaxChar) const;
  private:
    char* sSubName;
    const tObjectId* pParentId;
};

#endif INC_objectIdh 
