/* $RCSfile: outputDest.h,v $ - implements generic output destination classes */
/* CM Info:
$Header: /home/chara/cvsroot/ople/sys/include/outputDest.h,v 1.1.1.1 1998/06/11 21:36:13 chara Exp $
*/
/* Richard Johnson - Jet Propulsion Laboratory */
 
/*
modification history
--------------------
05a,rlj,9Feb98     Updated specification etc.
01a,beh,23Apr97    written
*/
 
/*
DESCRIPTION

*/


#ifndef INC_OutputDesth
#define INC_OutputDesth


#include "isit.h"
#include "modLib.h"
#include "outputChannel.h"

/****************************Specification Functions *****************************************

double OutputValue (tOutputDestination& Output, double x, TIME tTime) :==
Return updated output value base on x and time tTime
[ invalid_reference (Output) ||
  invalid_output (Output) => UNDEFINED
| Output is a tActuator => return := x + Output.dOffset
| (Output is a tModulatedActuator || 
  Output is a tModulatedAccelerationActuator || 
  Output is a tModulatedVelocityActuator ||
  Output is a tModulatedPositionActuator ) && this->bEnabled == TRUE =>
    return := x + MODULATION (Output.ModulationParameters, tTime) + Output.dOffset
| (Output is a tModulatedActuator || 
  Output is a tModulatedAccelerationActuator || 
  Output is a tModulatedVelocityActuator ||
  Output is a tModulatedPositionActuator ) && this->bEnabled == FALSE =>
    return := x + Output.dOffset
| true => return := appropriate output value for time tTime
]

tOutputDestination UpdateOutput (tOutputDestination& Output, double x, TIME tTime) :==
Returns output destination with updated state based on x and time tTime
[ invalid_reference (Output) ||
  invalid_output (Output) => UNDEFINED
| Output is a tActuator => return := copy of Output with
    dValue = x + Output.dOffset
| (Output is a tModulatedActuator || 
  Output is a tModulatedAccelerationActuator || 
  Output is a tModulatedVelocityActuator ||
  Output is a tModulatedPositionActuator ) && this->bEnabled == TRUE =>
    return := copy of Output with
      dValue = x,
      dModulationValue = MODULATION (Output.ModulationParameters, tTime)
| (Output is a tModulatedActuator || 
  Output is a tModulatedAccelerationActuator || 
  Output is a tModulatedVelocityActuator ||
  Output is a tModulatedPositionActuator ) && this->bEnabled == FALSE =>
    return := copy of Output with
      dValue = x,
      dModulationValue = 0.0
| true => return := appropriate copy of Output
]

double LimitValue (double dValue, double dLimit) :==
[ dLimit < 0 => return := anything 
| -dLimit < dValue < dLimit => return := dValue 
| true => return := SIGN (dValue) * dLimit
]

double LimitValue2 (double dValue, double dUpperLimit, double dLowerLimit) :==
[ dUpperLimit < dLowerLimit => return := anything
| dValue > dUpperLimit => return := dUpperLimit
| dValue < dLowerLimit => return := dLowerLimit
| true => return := dValue
]

double LimitVelocityAcceleration (double dOldVel, double dNewVel, 
                                     double dVelLimit, double dAccelLimit, double dRate) :==
Returns velocity limited by max acceleration and velocity.                                   
[ return := LimitValue (dOldVelocity + (LimitValue ((dNewVel - dOldVel) * dRate, dAccelLimit),
               dVelLimit) 


double LimitPositionVelocityAcceleration (double dOldPos, dNewPos, double dOldVel, 
          double dMaxPos, double dMinPos, double dVelLimit, double dAccelLimit, double dRate) :==
Returns position limited by max acceleration and velocity, and position.                                   
[ return := LimitValue2 (dOldPos + 
    LimitValue ( ( LimitValue ((dNewPos - dOldPos) * dRate, dVelLimit), dVelLimit) - dOldVel) * dRate, dAccLimit), 
    dMaxPos, dMinPos)
]


*******************************************************************
*************************  tOutputDestination **********************
********************************************************************
*
* A tOutputDestination is a thing which has a value, a method for
* updating that value, and a method for writing the current value
* to some output (most often a piece of hardware).  Derived classes can
* overload the Output method to change the way in which the value is
* Output.  Some derived classes of tOutputDestination use the current TIME
* in their Output methods, so the TIME must be passed in whenever
* you do an Output.  For classes that ignore the TIME, you can pass
* in anything for the TIME. 


[[D tOutputDestination::tOutputDestination (void) ]
Constructs this with default offset.
[ invalid_pointer (this) => UNDEFINED
| true =>
    this->dValue := 0.0
    : this->dOffset := 0.0
]]

[[D tOutputDestination::tOutputDestination_d (double dOffsetIn) ]
Constructs this with specified offset
[ invalid_pointer (this) => UNDEFINED
| true =>
    this->dValue := 0.0
    : this->dOffset := dOffsetIn
]]


[[D tOutputDestination::~tOutputDestination (void) ]
Destructs this.
[ invalid_pointer (this) => UNDEFINED
| true => *this := anything
]]


[[D tOutputDestination::SetValue (double dNewValue) ]
Virtual: Sets output value to dNewValue limited by the specific output class.
[ invalid_pointer (this) => UNDEFINED
| true => this->dValue := this->dNewValue limited to a specified range
]]
  
[[D tOutputDestination::Value (void)]
Returns current non-offset value of this.
[ invalid_pointer (this) => UNDEFINED
| true => return := this->dValue
]]
  
[[D tOutputDestination::SetOffset (double dNewOffset) ]
Sets offset value to dNewOffset.
[ invalid_pointer (this) => UNDEFINED
| true => this->dOffset := this->dNewOffset
]]
  
[[D tOutputDestination::Offset (void) ]
Returns offset value of this.
[ invalid_pointer (this) => UNDEFINED
| true => return := this->dOffset
]]
  
  
[[D tOutputDestination::Output (TIME tUpdateTime) ]
Pure virtual: Outputs value appropriately.
[ invalid_pointer (this) => UNDEFINED
| true =>
   Output this->dValue + this->dOffset
   : return := OK
]]


[[D tOutputDestination::OutputValue (TIME tServoTime, double dNewValue)]
Sets value to dNewValue and outputs dNewValue appropriately.
[ this invalid => UNDEFINED
| true =>
   this->dValue := dNewValue
   : Appropriate hardware output := this->dNewValue + this->dOffset
   : return := OK
]]


[[D tOutputDestination::Reset (void) ]
Resets this to zero value.
[ this invalid => UNDEFINED
| true =>
   this->dValue := 0.0
   : return := OK
]]

**************** protected specification **********************************

[[D tOutputDestination::LimitValue_dd (double dValue, double dLimit) ]
[ invalid_pointer (this) => UNDEFINED
| true => return := LimitValue (dValue,dLimit)
]]


[[D tOutputDestination::LimitValue_ddd (double dValue, double dLowerLimit, 
    double dUpperLimit) ]
[ invalid_pointer (this) => UNDEFINED
| true => return := LimitValue2 (dValue,dLowerLimit,dUpperLimit)
]]

*****************************  tActuator  **************************
*
* A tActuator is a tOutputDestination that writes its output
* values to a piece of hardware connected to a tOutputChannel.
*
* It contains all the state of a tOutputDestination, plus a
* reference to a channel and a conversion factor from electrical
* to physical units.
*

[[D tActuator::tActuator_td (tOutputChannel& ChannelIn, 
                          double dPhysicalUnitsPerElectricalUnit)]
Contructs this as an actuator connected to ChannelIn with appropriate scale factor and
default offset.                          
[ invalid_pointer (this) ||
  invalid_reference (ChannelIn) ||
  invalid_channel (ChannelIn) => UNDEFINED
| true =>
   this->dValue := 0.0
   : this->dOffset := 0.0
   : this->Channel := reference to ChannelIn
   : this->dElectricalToPhysicalScaleFactor :=
     dPhysicalUnitsPerElectricalUnit
   : CHANNEL_OUTPUT (ChannelIn) :=  0.0
   : return := OK
]]

[[D tActuator::tActuator_tdd (tOutputChannel& ChannelIn, 
                          double dPhysicalUnitsPerElectricalUnit,
                          double dOffsetIn)]
Contructs this as an actuator connected to ChannelIn with appropriate scale factor and
specified offset.                          
[ invalid_pointer (this) ||
  invalid_reference (ChannelIn) ||
  invalid_channel (ChannelIn) => UNDEFINED
| true =>
   this->dValue := 0.0
   : this->dOffset := dOffsetIn
   : this->Channel := reference to ChannelIn
   : this->dElectricalToPhysicalScaleFactor :=
     dPhysicalUnitsPerElectricalUnit
   : CHANNEL_OUTPUT (ChannelIn) :=  dOffsetIn
   : return := OK
]]


[[D tActuator::~tActuator (void) ]
Destructs this.
[ invalid_pointer (this) => UNDEFINED
| true => *this := anything
]]

[[D tActuator::SetValue (double dNewValue) ] = tOutputDestination::SetValue ]]

[[D tActuator::Value (void) ] = tOutputDestination::Value ]] 

[[D tActuator::SetOffset (double dNewValue) ] = tOutputDestination::SetOffset ]]

[[D tActuator::Offset (void) ] = tOutputDestination::Offset ]] 

[[D tActuator::Reset (void) ] = tOutputDestination::Reset ]]

[[D tActuator::CountsToPhysicalUnits (double dCounts) ]
Convert channel counts to physical value 
[ invalid_pointer (this) => UNDEFINED
| true => return := dPhysicalValue * this->dPhysicalUnitsPerCount
]]

[[D tActuator::PhysicalUnitsToCounts (double dPhysicalValue) ]
Convert physical value to channel counts
[ invalid_pointer (this) ||
  this->dPhsyicalUnitsPerCount == 0 => UNDEFINED
| true => return := dPhysicalValue / this->dPhysicalUnitsPerCount
]]


[[D tActuator::MaxChannelUnits (void) ]
Return the maximum value that can be represented in channel units for this->Channel.
[ invlid_pointer (this) ||
  invalid_reference (this->Channel)
  invalid_channel (this->Channel) => UNDEFINED
| true =>
    return := maximum value in channel units for this->Channel
]]


[[D tActuator::Output (TIME tUpdateTime) ]
Send offseted value to actuator connected to this->Channel.
[ invalid_pointer (this) ||
  invalid_reference (this->Channel) ||
  invalid_channel (this->Channel) => UNDEFINED
| true =>
    CHANNEL_OUTPUT (this->Channel) := this->dValue + this->dOffset
   : return := OK
]]


[[D tActuator::OutputValue (TIME tUpdateTime, double dNewValue) ]
Send dNewValue with offset to actuator connected to this->Channel.
[ invalid_pointer (this) ||
  invalid_reference (this->Channel) ||
  invalid_channel (this->Channel) => UNDEFINED
| true =>
   this->dValue := dNewValue
   : CHANNEL_OUTPUT (this->Channel) := this->dNewValue + this->dOffset
   : return := OK
]]

*************************  protected specification *********************

[[D tActuator::LimitValue_dd (double dValue, double dLimit) ] =
    tOutputDestination::LimitValue_dd ]]
    
[[D tActuator::LimitValue_ddd (double dValue, double dLimit) ] =
    tOutputDestination::LimitValue_ddd ]]
    
[[D tActuator::ChannelOutput (double dOutputValue) ]
[ invalid_pointer (this) ||
  invalid_reference (this->Channel) ||
  invalid_channel (this->Channel) => UNDEFINED
| true =>
   CHANNEL_OUTPUT (this->Channel) := this->dOutputValue
   : return := OK
]]
 
*************************  tModulatedActuator  *********************
*
* A tModulatedActuator is a tActuator that can add a modulation
* to its output for diagnostic purposes.
*
* This allows a waveform to be modulated onto an actuator
* output, while maintaining the same interface as an
* ordinary tActuator has.
*


[[D tModulatedActuator::tModulatedActuator_td (tOutputChannel& ChannelIn,
       double dPhysicalUnitsPerElectricalUnit) ]
Constructs this with disabled default modualtion
[ invalid_pointer (this) ||
  invalid_reference (ChannelIn) ||
  invalid_channel (ChannelIn) => UNDEFINED
  | true =>
     this->dValue := 0.0
     : this->dOffset := 0.0
     : this->ModulatioParameters := ZERO_MODULATION
     : this->bEnabled := FALSE
     : this->dModulationValue := 0.0
     : this->Channel := reference to ChannelIn
     : this->dElectricalToPhysicalScaleFactor :=
       dPhysicalUnitsPerElectricalUnit
     : CHANNEL_OUTPUT (ChannelIn) := 0.0
]]


[[D tModulatedActuator::tModulatedActuator_tdt (tOutputChannel& ChannelIn,
       double dPhysicalUnitsPerElectricalUnit, const tModulationParameters&
       InitialParameters) ]
Constructs this with disabled modualtion specified by InitialParameters.
[ invalid_pointer (this) ||
  invalid_reference (ChannelIn) ||
  invalid_channel (ChannelIn) => UNDEFINED
  | true =>
     this->dValue := 0.0
     : this->dOffset := 0.0
     : this->ModulationParameters := InitialParameters
     : this->bEnabled := FALSE
     : this->dModulationValue := 0.0
     : this->Channel := reference to ChannelIn
     : this->dElectricalToPhysicalScaleFactor :=
       dPhysicalUnitsPerElectricalUnit
     : CHANNEL_OUTPUT (ChannelIn) := 0.0
]]

[[D tModulatedActuator::tModulatedActuator_tdtd (tOutputChannel& ChannelIn,
       double dPhysicalUnitsPerElectricalUnit, const tModulationParameters&
       InitialParameters, double dOffsetIn) ]
Constructs this with disabled default modualtion
[ invalid_pointer (this) ||
  invalid_reference (ChannelIn) ||
  invalid_channel (ChannelIn) => UNDEFINED
  | true =>
     this->dValue := 0.0
     : this->dOffset := dOffsetIn
     : this->ModulatioParameters := InitialParameters
     : this->bEnabled := FALSE 
     : this->dModulationValue := 0.0
     : this->Channel := reference to ChannelIn
     : this->dElectricalToPhysicalScaleFactor :=
       dPhysicalUnitsPerElectricalUnit
     : CHANNEL_OUTPUT (ChannelIn) := dOffsetIn
]]

[[D tModulatedActuator::~tModulatedActuator (void) ]
Destructs this.
[ invalid_pointer (this) => UNDEFINED
| true => *this := anything
]]


[[D tModulatedActuator::SetValue (double dNewValue) ] = tActuator::SetValue ]]

[[D tModulatedActuator::Value (void) ] = tActuator::Value ]] 

[[D tModulatedActuator::SetOffset (double dNewOffset) ] = tActuator::SetOffset ]]

[[D tModulatedActuator::Offset (void) ] = tActuator::Offset ]]


[[D tModulatedActuator::MaxChannelUnits (void) ] = tActuator::MaxChannelUnits ]]


[[D tModulatedActuator::CountsPhysicalUnits (void) ] =
    tActuator::CountsPhysicalUnits  ]]

[[D tModulatedActuator::PhysicalUnitsToCounts (void) ] =
    tActuator::PhysicalUnitsToCounts ]]

[[D tModulatedActuator::Offset (void) ] = tActuator::Offset ]] 



[[D tModulatedActuator::SetModulationParameters (
      const tModulationParameters& NewParameters) ]
Updates modulation parameters using NewParameters.
[  invalid_pointer (this) ||
   invalid_reference (NewParameters) => UNDEFINED
| true =>
    this->ModulationParameters := NewParameters
    : return := OK
]]

[[D tModulatedActuator::ModulationParameters (void) ]
Returns modulation parameters.
[  invalid_pointer (this) => UNDEFINED
| true => return := this->ModulationParameters
]]

[[D tModulatedActuator::ModulationValue (void) ]
Returns the current value of the modulation that's being
added to the raw actuator output 
[ invalid_pointer (this) => UNDEFINED
| true => return := this->dModulationValue
]]

[[D tModulatedActuator::EnableModulation (void) ]
Enables modulation so that the modulation is superimposed on the 
output value.
[ invalid_pointer (this) => UNDEFINED
| true =>
    this->bEnable := TRUE
    : return := OK
]]

[[D tModulatedActuator::DisableModulation (void) ]
Diables modulation so that the modulation is not superimposed on the 
output value.
[ invalid_pointer (this) => UNDEFINED
| true =>
    this->bEnable := FALSE
    : this->dModulationValue := 0.0
    : return := OK
]]


[[D tModulatedActuator::IsModulationEnabled (void) ]
Returns status of the modulation enable.
[ invalid_pointer (this) => UNDEFINED
| true => return := this->bEnabled
]]


[[D tModulatedActuator::Reset (void) ] 
Resets this.
[ invalid_pointer (this) ||
  invalid_reference (this->Channel) ||
  invalid_channel (this->Channel) => UNDEFINED
| true => 
    this->dModulationValue := 0
    : this->dValue := 0.0
    : CHANNEL_VALUE (this->Channel) := 0.0
    : return := OK
]]


[[D tModulatedActuator::Output (TIME tUpdateTime) ]
Output cached value with modulation if enabled to actuator connected to this->Channel.
[ invalid_pointer (this) ||
  invalid_reference (this->Channel) ||
  invalid_channel (this->Channel) => UNDEFINED
| this->bEnable == FALSE =>
    this->dModulationValue := 0.0
    : CHANNEL_OUTPUT (this->Channel) := this->dValue + this->dOffset
    : return := OK 
| true =>
    this->dModulationValue := MODULATION (this->ModulationParameters, tUpdateTime)
    : CHANNEL_OUTPUT (this->Channel) := dNewValue + this->dOffset + 
      MODULATION (this->ModulationParameters, tUpdateTime)
    : return := OK
]]

[[D tModulatedActuator::OutputValue (TIME tUpdateTime, double dNewValue) ]
Output dNewValue with modulation if enabled to actuator connected to this->Channel.
[ invalid_pointer (this) ||
  invalid_reference (this->Channel) ||
  invalid_channel (this->Channel) => UNDEFINED
| this->bEnable == FALSE =>
    this->dValue := dNewValue
    : this->dModulationValue := 0.0
    : CHANNEL_OUTPUT (this->Channel) := dNewValue + this->dOffset
    : return := OK 
| true =>
    this->dValue := dNewValue
    : this->dModulationValue := MODULATION (this->ModulationParameters, tUpdateTime)
    : CHANNEL_OUTPUT (this->Channel) := dNewValue + this->dOffset + 
      MODULATION (this->ModulationParameters, tUpdateTime)
    : return := OK
]]



[[D tModulatedActuator::UnmodulatedOutput (TIME tUpdateTime) ]
Output cached value without modulation to actuator connected to this->Channel.
[ invalid_pointer (this) ||
  invalid_reference (this->Channel) ||
  invalid_channel (this->Channel) => UNDEFINED
| true =>
    CHANNEL_OUTPUT (this->Channel) := this->dValue + this->dOffset
    : return := OK 
    
]]


[[D tModulatedActuator::UnmodulatedOutputValue (TIME tUpdateTime,
      double dNewValue) ]
Output dNewValue without modulation to actuator connected to this->Channel.
[ invalid_pointer (this) ||
  invalid_reference (this->Channel) ||
  invalid_channel (this->Channel) => UNDEFINED
| true =>
    this->dValue := dNewValue
    : CHANNEL_OUTPUT (this->Channel) := dNewValue + this->dOffset
    : return := OK 
    
]]


[[D tModulatedActuator::Acceleration (void) ]
Return last acceleration value sent to actuator.
[ invalid_pointer (this) || => UNDEFINED
| true => return : this->dValue
]]


****************** protected specification **************************

[[D tModulatedActuator::LimitValue_dd (double dValue, double dLimit) ] =
    tActuator::LimitValue_dd ]]
    
[[D tModulatedActuator::LimitValue_ddd (double dValue, double dLimit) ] =
    tActuator::LimitValue_ddd ]]
    
[[D tModulatedActuator::ChannelOutput (double dOutputValue) ] = tActuator::ChannelOutput ]]



*******************  tModulatedAccelerationActuator  ***************
*
* A tModulatedAccelerationActuator is a tModulatedActuator with
* an acceleration limit applied to the unmodulated output.
*


[[D tModulatedAccelerationActuator::SetValue (double dNewValue) ]
[ invalid_pointer (this) => UNDEFINED
| true => this->dValue := LimitValue (dNewValue, this->CdMaxAcceleration) 
]]
 
[[D tModulatedAccelerationActuator::Value (void) ] = 
    tModulatedActuator::Value ]]

[[D tModulatedAccelerationActuator::Reset (void) ] =
    tModulatedActuator::Reset ]]
    
[[D tModulatedAccelerationActuator::SetModulationParameters (
      const tModulationParams& NewParameters) ] =
        tModulatedActuator::SetModulationParameters ]]

[[D tModulatedAccelerationActuator::ModulationParameters (void) ] =
    tModulatedActuator::ModulationParameters ]]
        
[[D tModulatedAccelerationActuator::EnableModulaion (void) ] =
    tModulatedActuator::EnableModulation ]]

[[D tModulatedAccelerationActuator::DisableModulation (void) ] =
    tModulatedActuator::DisableModulation ]]

[[D tModulatedAccelerationActuator::IsModulationEnabled (void) ] =
    tModulatedActuator::IsModulationEnabled ]]

    
[[D tModulatedAccelerationActuator::Output (TIME tUpdateTime) ] =
    tModulatedActuator::Output ]]

[[D tModulatedAccelerationActuator::OutputValue (TIME tUpdateTime, dNewValue) ] =
    tModulatedActuator::OutputValue ]]

[[D tModulatedAccelerationActuator::UnmodulatedOutput (TIME tUpdateTime) ] =
    tModulatedActuator::UnmodulatedOutput ]]

[[D tModulatedAccelerationActuator::UnmodulatedOutputValue (TIME tUpdateTime, dNewValue) ] =
    tModulatedActuator::UnmodulatedOutputValue ]]
    
[[D tModulatedAccelerationActuator::MaxChannelUnits (void) ] = 
    tModulatedActuator::MaxChannelUnits ]]


[[D tModulatedAccelerationActuator::CountsPhysicalUnits (void) ] =
    tModulatedActuator::CountsPhysicalUnits  ]]

[[D tModulatedAccelerationActuator::PhysicalUnitsToCounts (void) ] =
    tModulatedActuator::PhysicalUnitsToCounts ]]


[[D tModulatedAccelerationActuator::tModulatedAccelerationActuator_tdd (
       tOutputChannel& ChannelIn,
       double dPhysicalUnitsPerElectricalUnit,
       double dMaxAccelerationIn, double dOutputRateIn) ]
Constructs this with disabled default modualtion
[ invalid_pointer (this) ||
  invalid_reference (ChannelIn) ||
  invalid_channel (ChannelIn) => UNDEFINED
  | true =>
     this->dValue := 0.0
     : this->dOffset := 0.0
     : this->ModulationParameters := ZERO_MODULATION
     : this->bEnabled := FALSE 
     : this->dModulationValue := 0.0
     : this->Channel := reference to ChannelIn
     : this->dElectricalToPhysicalScaleFactor :=
       dPhysicalUnitsPerElectricalUnit
     : CHANNEL_OUTPUT (ChannelIn) := 0.0
     : this->dMaxAcceleration := dMaxAccelerationIn
     : this->dOutputRate := dOutputRateIn
]]


[[D tModulatedAccelerationActuator::tModulatedAccelerationActuator_tddt (
       tOutputChannel& ChannelIn,
       double dPhysicalUnitsPerElectricalUnit, double dMaxAccelerationIn, 
       double dOutputRateIn, const tModulationParameters& InitialParameters) ]
Constructs this with disabled default modualtion
[ invalid_pointer (this) ||
  invalid_reference (ChannelIn) ||
  invalid_channel (ChannelIn) => UNDEFINED
  | true =>
     this->dValue := 0.0
     : this->dOffset := 0.0
     : this->ModulationParameters := InitialParameters
     : this->bEnabled := FALSE
     : this->dModulationValue := 0.0
     : this->Channel := reference to ChannelIn
     : this->dElectricalToPhysicalScaleFactor :=
       dPhysicalUnitsPerElectricalUnit
     : CHANNEL_OUTPUT (ChannelIn) := 0.0
     : this->dMaxAcceleration := dMaxAccelerationIn
     : this->dOutputRate := dOutputRateIn
]]

[[D tModulatedAccelerationActuator::tModulatedAccelerationActuator_tddtd (
       tOutputChannel& ChannelIn,
       double dPhysicalUnitsPerElectricalUnit, double dMaxAccelerationIn,
       double dOutputRateIn, const tModulationParameters& InitialParameters, double dOffsetIn) ]
Constructs this with disabled default modualtion
[ invalid_pointer (this) ||
  invalid_reference (ChannelIn) ||
  invalid_channel (ChannelIn) => UNDEFINED
  | true =>
     this->dValue := 0.0
     : this->dOffset := dOffsetIn
     : this->ModulatioParameters := InitialParameters
     : this->bEnabled := FALSE 
     : this->dModulationValue := 0.0
     : this->Channel := reference to ChannelIn
     : this->dElectricalToPhysicalScaleFactor :=
       dPhysicalUnitsPerElectricalUnit
     : CHANNEL_OUTPUT (ChannelIn) := dOffsetIn
     : this->dMaxAcceleration := dMaxAccelerationIn
     : this->dOutputRate := dOutputRateIn
]]



[[D tModulatedAccelerationActuator::~tModulatedAccelerationActuator (void) ]
Destructs this.
[ invalid_pointer (this) => UNDEFINED
| true => *this := anything
]]


[[D tModulatedAccelerationActuator::MaxAcceleration (void) ]
Returns max acceleration allowed, in units of the scale factor
[ invalid_pointer (this) => UNDEFINED
| true => return := this->CdMaxAcceleration
]]

[[D tModulatedAccelerationActuator::MaxAccelerationInChannelUnits (void) ]
Returns max acceleration allowed in channel units
[ invalid_pointer (this) => UNDEFINED
| true => return := this->CdMaxAcceleration in channel units
]]

[[D tModulatedAccelerationActuator::OutputRate (void) ]
Returns nomimal actuation rate in Hz
[ invalid_pointer (this) => UNDEFINED
| true => return := this->dOutputRate
]]



[[D tModulatedAccelerationActuator::Output (TIME tUpdateTime) ] = tModulatedActuator::Output ]]

[[D tModulatedAccelerationActuator::OutputValue (TIME tUpdateTime, double dNewValue) ]
Output dNewValue with modulation if enabled to actuator connected to this->Channel with 
acceleration limits applied.
[ invalid_pointer (this) ||
  invalid_reference (this->Channel) ||
  invalid_channel (this->Channel) => UNDEFINED
| this->bEnable == FALSE =>
    this->dValue := LimitValue (dNewValue, this->CdMaxAcceleration)
    : this->dModulationValue := 0.0
    : CHANNEL_OUTPUT (this->Channel) := dNewValue + this->dOffset
    : return := OK 
| true =>
    this->dValue := LimitValue (dNewValue, this->CdMaxAcceleration)
    : this->dModulationValue := MODULATION (this->ModulationParameters, tUpdateTime)
    : CHANNEL_OUTPUT (this->Channel) := LimitValue (dNewValue, this->CdMaxAcceleration) + this->dOffset + MODULATION (this->ModulationParameters, tUpdateTime)
    : return := OK
]]



[[D tModulatedAccelerationActuator::UnmodulatedOutput (TIME tUpdateTime) ] = tModulatedActuator::UnmodulatedOutput ]]



[[D tModulatedAccelerationActuator::UnmodulatedOutputValue (TIME tUpdateTime,
      double dNewValue) ]
Output dNewValue without modulation to actuator connected to this->Channel with acceleration limits applied
[ invalid_pointer (this) ||
  invalid_reference (this->Channel) ||
  invalid_channel (this->Channel) => UNDEFINED
| true =>
    this->dValue := LimitValue (dNewValue, this->CdMaxAcceleration)
    : CHANNEL_OUTPUT (this->Channel) := LimitValue (dNewValue, this->CdMaxAcceleration) + this->dOffset
    : return := OK 
    
]]


*******************protected specification**************************************************

[[D tModulatedAccelerationActuator::LimitAcceleration (double dAcceleration) ]
[ invalid_pointer (this) => UNDEFINED
| true => return := LimitValue (dAcceleration, this->CdMaxAcceleration in channel units)
]]

[[D tModulatedAccelerationActuator::SetMaxAcceleration (double dMaxAccInPhysicalUnits) ]
[ invalid_pointer (this) ||
  this->Channel.dPhysicalUnitsPerCount == 0 => UNDEFINED
| true => this->dMaxVelocity := dMaxAccInPhysicalUnits / this->Channel.dPhysicalUnitsPerCount
]]

[[D tModulatedAccelerationActuator::LimitValue_dd (double dValue, double dLimit) ] =
    tModulatedActuator::LimitValue_dd ]]
    
[[D tModulatedAccelerationActuator::LimitValue_ddd (double dValue, double dLimit) ] =
    tModulatedActuator::LimitValue_ddd ]]
    
[[D tModulatedAccelerationActuator::ChannelOutput (double dOutputValue) ] = tModulatedActuator::ChannelOutput ]]


*******************  tModulatedVelocityActuator  ***************
*
* A tModulatedVelocityActuator is a tModulatedAccelerationActuator with
* a velocity limit applied to the unmodulated output.
*

[[D tModulatedVelocityActuator::tModulatedVelocityActuator_tddd (
       tOutputChannel& ChannelIn,
       double dPhysicalUnitsPerElectricalUnit,
       double dMaxAccelerationIn, double dMaxVelocityIn, double dOutputRateIn) ]
Constructs this with disabled default modualtion
[ invalid_pointer (this) ||
  invalid_reference (ChannelIn) ||
  invalid_channel (ChannelIn) => UNDEFINED
  | true =>
     this->dValue := 0.0
     : this->dOffset := 0.0
     : this->ModulationParameters := ZERO_MODULATION
     : this->bEnabled := FALSE 
     : this->dModulationValue := 0.0
     : this->Channel := reference to ChannelIn
     : this->dElectricalToPhysicalScaleFactor :=
       dPhysicalUnitsPerElectricalUnit
     : CHANNEL_OUTPUT (ChannelIn) := 0.0
     : this->dMaxAcceleration := dMaxAccelerationIn
     : this->dMaxVelocity := dMaxVelocityIn
     : this->dLastVelocity := 0.0
     : this->dOutputRate := dOutputRateIn
]]


[[D tModulatedVelocityActuator::tModulatedVelocityActuator_tdddt (
       tOutputChannel& ChannelIn,
       double dPhysicalUnitsPerElectricalUnit, double dMaxAccelerationIn,
       double dMaxVelocityIn, double dOutputRateIn, const tModulationParameters& InitialParameters) ]
Constructs this with disabled default modualtion
[ invalid_pointer (this) ||
  invalid_reference (ChannelIn) ||
  invalid_channel (ChannelIn) => UNDEFINED
  | true =>
     this->dValue := 0.0
     : this->dOffset := 0.0
     : this->ModulationParameters := InitialParameters
     : this->bEnabled := FALSE
     : this->dModulationValue := 0.0
     : this->Channel := reference to ChannelIn
     : this->dElectricalToPhysicalScaleFactor :=
       dPhysicalUnitsPerElectricalUnit
     : CHANNEL_OUTPUT (ChannelIn) := 0.0
     : this->dMaxAcceleration := dMaxAccelerationIn
     : this->dMaxVelocity := dMaxVelocityIn
     : this->dLastVelocity := 0.0
     : this->dOutputRate := dOutputRateIn
]]

[[D tModulatedVelocityActuator::tModulatedVelocityActuator_tdddtd (
       tOutputChannel& ChannelIn,
       double dPhysicalUnitsPerElectricalUnit, double dMaxAccelerationIn,
       double dMaxVelocityIn, double dOutputRateIn, const tModulationParameters& InitialParameters, 
       double dOffsetIn) ]
Constructs this with disabled default modualtion
[ invalid_pointer (this) ||
  invalid_reference (ChannelIn) ||
  invalid_channel (ChannelIn) => UNDEFINED
  | true =>
     this->dValue := 0.0
     : this->dOffset := dOffsetIn
     : this->ModulatioParameters := InitialParameters
     : this->bEnabled := FALSE 
     : this->dModulationValue := 0.0
     : this->Channel := reference to ChannelIn
     : this->dElectricalToPhysicalScaleFactor :=
       dPhysicalUnitsPerElectricalUnit
     : CHANNEL_OUTPUT (ChannelIn) := dOffsetIn
     : this->dMaxAcceleration := dMaxAccelerationIn
     : this->dMaxVelocity := dMaxVelocityIn
     : this->dLastVelocity := 0.0
     : this->dOutputRate := dOutputRateIn
]]



[[D tModulatedVelocityActuator::~tModulatedVelocityActuator (void) ]
Destructs this.
[ invalid_pointer (this) => UNDEFINED
| true => *this := anything
]]


[[D tModulatedVelocityActuator::SetValue (double dNewValue) ] 
[ invalid_pointer (this) => UNDEFINED
| true => 
    this->dValue := LimitVelocityAcceleration (this->dLastVelocity, dNewValue, this->CdMaxVelocity, this->CdMaxAcceleration) 
    : this->dLastVelocity := this->dValue
]]

[[D tModulatedVelocityActuator::Reset (void) ] 
[ invalid_pointer (this) ||
  invalid_reference (this->Channel) ||
  invalid_channel (this->Channel) => UNDEFINED
    this->dValue := 0.O 
| true => 
    this->dModulationValue := 0
    : this->dValue := 0.0
    : CHANNEL_VALUE (this->Channel) := 0.0
    : this->dLastVelocity := 0.0
    : return := OK
]]


    
[[D tModulatedVelocityActuator::Value (void)] =
    tModulatedAccelerationActuator::Value ]]
    
[[D tModulatedVelocityActuator::SetOffset (double dNewOffset) ] =
    tModulatedAccelerationActuator::SetOffset ]]
    
[[D tModulatedVelocityActuator::Offset (void)] =
    tModulatedAccelerationActuator::Offset ]]

    
[[D tModulatedVelocityActuator::SetModulationParameters (
      const tModulationParams& NewParameters) ] =
        tModulatedAccelerationActuator::SetModulationParameters ]]

[[D tModulatedVelocityActuator::ModulationParameters (void) ] =
    tModulatedAccelerationActuator::ModulationParameters ]]
        
[[D tModulatedVelocityActuator::EnableModulaion (void) ] =
    tModulatedAccelerationActuator::EnableModulation ]]

[[D tModulatedVelocityActuator::DisableModulation (void) ] =
    tModulatedAccelerationActuator::DisableModulation ]]

[[D tModulatedVelocityActuator::IsModulationEnabled (void) ] =
    tModulatedAccelerationActuator::IsModulationEnabled ]]
    
[[D tModulatedVelocityActuator::MaxChannelUnits (void) ] = 
    tModulatedAccelerationActuator::MaxChannelUnits ]]


[[D tModulatedVelocityActuator::CountsPhysicalUnits (void) ] =
    tModulatedAccelerationActuator::CountsPhysicalUnits  ]]

[[D tModulatedVelocityActuator::PhysicalUnitsToCounts (void) ] =
    tModulatedAccelerationActuator::PhysicalUnitsToCounts ]]
    
[[D tModulatedVelocityActuator::MaxAcceleration (void) ] =
    tModulatedAccelerationActuator::MaxAcceleration ]]

[[D tModulatedVelocityActuator::MaxAccelerationInChannelUnits (void) ] =
    tModulatedAccelerationActuator::MaxAccelerationInChannelUnits ]]

[[D tModulatedVelocityActuator::OutputRate (void) ] = 
    tModulatedAccelerationActuator::OutputRate ]] ]

[[D tModulatedVelocityActuator::MaxVelocity (void)]
Returns max velocity allowed, in physical units
[ invalid_pointer (this) => UNDEFINED
| true => return := this->CdMaxVelocity
]]

[[D tModulatedVelocityActuator::MaxVelocityInChannelUnits (void) ]
Returns max velocity allowed in channel units
[ invalid_pointer (this) => UNDEFINED
| true => return := this->CdMaxVelocity in channel units
]]

[[D tModulatedVelocityActuator::Velocity (void) ]
Returns last velocity 
[ invalid_pointer (this) => UNDEFINED
| true => return := this->dLastVelocity
]]


[[D tModulatedVelocityActuator::Output (TIME tUpdateTime) ] = tModulatedAccelerationActuator::Output ]]

[[D tModulatedVelocityActuator::OutputValue (TIME tUpdateTime, double dNewValue) ]
Output dNewValue with modulation if enabled to actuator connected to this->Channel with accel and velocity limits applied 
[ invalid_pointer (this) ||
  invalid_reference (this->Channel) ||
  invalid_channel (this->Channel) => UNDEFINED
| this->bEnable == FALSE =>
    this->dValue := LimitVelocityAcceleration (this->dLastVelocity, dNewValue, this->CdMaxVelocity, this->CdMaxAcceleration)
    : this->dModulationValue := 0.0
    : CHANNEL_OUTPUT (this->Channel) := 
      LimitVelocityAcceleration (this->dLastVelocity, dNewValue, this->CdMaxVelocity, this->CdMaxAcceleration) + this->dOffset
    : this->dLastVelocity := this->dValue
    : return := OK 
| true =>
    this->dValue := LimitVelocityAcceleration (this->dLastVelocity, dNewValue, this->CdMaxVelocity, this->CdMaxAcceleration)
    : this->dModulationValue := MODULATION (this->ModulationParameters, tUpdateTime)
    : CHANNEL_OUTPUT (this->Channel) := 
      LimitVelocityAcceleration (this->dLastVelocity, dNewValue, this->CdMaxVelocity, this->CdMaxAcceleration) +
      this->dOffset + MODULATION (this->ModulationParameters, tUpdateTime)
    : this->dLastVelocity := this->dValue
    : return := OK
]]



[[D tModulatedVelocityActuator::UnmodulatedOutput (TIME tUpdateTime) ] = tModulatedAccelerationActuator::UnmodulatedOutput ]]



[[D tModulatedVelocityActuator::UnmodulatedOutputValue (TIME tUpdateTime,
      double dNewValue) ]
Output dNewValue without modulation to actuator connected to this->Channel limited
by maximum velocity and acceleraton.
[ invalid_pointer (this) ||
  invalid_reference (this->Channel) ||
  invalid_channel (this->Channel) => UNDEFINED
| true =>
    this->dValue := LimitVelocityAcceleration (this->dLastVelocity, dNewValue, this->CdMaxVelocity, this->CdMaxAcceleration)
    : CHANNEL_OUTPUT (this->Channel) := 
      LimitVelocityAcceleration (this->dLastVelocity, dNewValue, this->CdMaxVelocity, this->CdMaxAcceleration) + this->dOffset
    : this->dLastVelocity := this->dValue
    : return := OK    
]]


[[D tModulatedVelocityActuator::Acceleration (void) ]
Compute average acceleration assuming constant actuator velocity update rate.
[ invalid_pointer (this) => UNDEFINED 
| true => return := (this->dValue - this->dLastVelocity) * this->dOutputRate
]]

[[D tModulatedVelocityActuator::Velocity (void) ]
Returns last velocity sent to actuator
[ invalid_pointer (this) => UNDEFINED 
| true => return := this->dValue
]]

*******************protected specification**************************************************

[[D tModulatedVelocityActuator::LimitVelocity (double dOldVelocity, double dNewVelocity) ]
[ invalid_pointer (this) => UNDEFINED
| true => return := LimitVelocityAcceleration (dOldVelocity, dNewVelocity, 
    this->CdMaxVelocity in channel units, this->CdMaxAcceleration in channel units)
]]

[[D tModulatedVelocityActuator::SetMaxAcceleration (double dMaxAccInPhysicalUnits) ] =
    tModulatedAccelerationActuator::SetMaxAcceleration ]]

[[D tModulatedVelocityActuator::SetMaxVelocity (double dMaxVelInPhysicalUnits) ]
[ invalid_pointer (this) ||
  this->Channel.dPhysicalUnitsPerCount == 0 => UNDEFINED
| true => this->dMaxVelocity := dMaxVelInPhysicalUnits / this->Channel.dPhysicalUnitsPerCount
]]

[[D tModulatedVelocityActuator::SetLastVelocity (double dVelocity) ]
Sets last velocity 
[ invalid_pointer (this) => UNDEFINED
| true =>  this->dLastVelocity := dVelocity
]]


[[D tModulatedVelocityActuator::LastVelocity (void) ]
Returns last velocity 
[ invalid_pointer (this) => UNDEFINED
| true =>  return := this->dLastVelocity
]]


[[D tModulatedVelocityActuator::LimitAcceleration (
      double dOldVelocity, double dNewVelocity) ] =
    tModulatedAccelerationActuator::LimitAcceleration ]]

[[D tModulatedVelocityActuator::LimitValue_dd (double dValue, double dLimit) ] =
    tModulatedAccelerationActuator::LimitValue_dd ]]
    
[[D tModulatedVelocityActuator::LimitValue_ddd (double dValue, double dLimit) ] =
    tModulatedAccelerationActuator::LimitValue_ddd ]]
    
[[D tModulatedVelocityActuator::ChannelOutput (double dOutputValue) ] = tModulatedAccelerationActuator::ChannelOutput ]]

*******************  tModulatedPositionActuator  ***************
*
* A tModulatedPositionActuator is a tModulatedAccelerationActuator with
* a Position limit applied to the unmodulated output.
*




[[D tModulatedPositionActuator::tModulatedPositionActuator_tddddd (
       tOutputChannel& ChannelIn,
       double dPhysicalUnitsPerElectricalUnit,
       double dMaxAccelerationIn, double dMaxVelocityIn,
       double dMinPositionIn, double dMaxPositionIn, double dOutputRateIn) ]
Constructs this with disabled default modualtion
[ invalid_pointer (this) ||
  invalid_reference (ChannelIn) ||
  invalid_channel (ChannelIn) => UNDEFINED
  | true =>
     this->dValue := 0.0
     : this->dOffset := 0.0
     : this->ModulationParameters := ZERO_MODULATION
     : this->bEnabled := FALSE 
     : this->dModulationValue := 0.0
     : this->Channel := reference to ChannelIn
     : this->dElectricalToPhysicalScaleFactor :=
       dPhysicalUnitsPerElectricalUnit
     : CHANNEL_OUTPUT (ChannelIn) := 0.0
     : this->dMaxAcceleration := dMaxAccelerationIn
     : this->dMaxVelocity := dMaxVelocityIn
     : this->dLastVelocity := 0.0
     : this->dMinPosition := dMinPositionIn
     : this->dMaxPosition := dMaxPositionIn
     : this->dLastPosition := 0.0
     : this->dOutputRate := dOutputRateIn
]]


[[D tModulatedPositionActuator::tModulatedPositionActuator_tdddddt (
       tOutputChannel& ChannelIn,
       double dPhysicalUnitsPerElectricalUnit, double dMaxAccelerationIn,
       double dMaxVelocityIn, double dMinPositionIn, 
       double dMaxPositionIn, double dOutputRateIn, const tModulationParameters& InitialParameters) ]
Constructs this with disabled default modualtion
[ invalid_pointer (this) ||
  invalid_reference (ChannelIn) ||
  invalid_channel (ChannelIn) => UNDEFINED
  | true =>
     this->dValue := 0.0
     : this->dOffset := 0.0
     : this->ModulationParameters := InitialParameters
     : this->bEnabled := FALSE
     : this->dModulationValue := 0.0
     : this->Channel := reference to ChannelIn
     : this->dElectricalToPhysicalScaleFactor :=
       dPhysicalUnitsPerElectricalUnit
     : CHANNEL_OUTPUT (ChannelIn) := 0.0
     : this->dMaxAcceleration := dMaxAccelerationIn
     : this->dMaxVelocity := dMaxVelocityIn
     : this->dLastVelocity := 0.0
     : this->dMinPosition := dMinPositionIn
     : this->dMaxPosition := dMaxPositionIn
     : this->dLastPosition := 0.0
     : this->dOutputRate := dOutputRateIn
]]

[[D tModulatedPositionActuator::tModulatedPositionActuator_tdddddtd (
       tOutputChannel& ChannelIn,
       double dPhysicalUnitsPerElectricalUnit, double dMaxAccelerationIn,
       double dMaxVelocityIn,
       double dMinPositionIn, double dMaxPositionIn, double dOutputRateIn,
       const tModulationParameters& InitialParameters, 
       double dOffsetIn) ]
Constructs this with disabled default modualtion
[ invalid_pointer (this) ||
  invalid_reference (ChannelIn) ||
  invalid_channel (ChannelIn) => UNDEFINED
  | true =>
     this->dValue := 0.0
     : this->dOffset := dOffsetIn
     : this->ModulatioParameters := InitialParameters
     : this->bEnabled := FALSE 
     : this->dModulationValue := 0.0
     : this->Channel := reference to ChannelIn
     : this->dElectricalToPhysicalScaleFactor :=
       dPhysicalUnitsPerElectricalUnit
     : CHANNEL_OUTPUT (ChannelIn) := dOffsetIn
     : this->dMaxAcceleration := dMaxAccelerationIn
     : this->dMaxVelocity := dMaxVelocityIn
     : this->dLastVelocity := 0.0
     : this->dMinPosition := dMinPositionIn
     : this->dMaxPosition := dMaxPositionIn
     : this->dLastPosition := 0.0
     : this->dOutputRate := dOutputRateIn
]]


[[D tModulatedPositionActuator::~tModulatedPositionActuator (void) ]
Destructs this.
[ invalid_pointer (this) => UNDEFINED
| true => *this := anything
]]

[[D tModulatedPositionActuator::SetValue (double dNewValue) ] 
[ invalid_point (this) => UNDEFINED
| true =>
  
    this->dValue := LimitPositionVelocityAcceleration (
      dNewValue, this->dValue, dNewValue, this->dLastVelocity, 
      this->CdMaxPosition, this->CdMinPosition, 
      this->CdMaxVelocity, this->CdMaxAcceleration)
    : this->dLastPosition := this->dValue
    : this->dLastVelocity := (this->dValue - this->LastPosition) * this->dOutputRate
]]

[[D tModulatedPositionActuator::Value (void)] =
    tModulatedVelocityActuator::Value ]]
    
[[D tModulatedPositionActuator::SetOffset (double dNewOffset) ] =
    tModulatedVelocityActuator::SetOffset ]]
    
[[D tModulatedPositionActuator::Offset (void)] =
    tModulatedVelocityActuator::Offset ]]

[[D tModulatedVelocityActuator::Reset (void) ] 
[ invalid_pointer (this) ||
  invalid_reference (this->Channel) ||
  invalid_channel (this->Channel) => UNDEFINED
    this->dValue := 0.O 
| true => 
    this->dModulationValue := 0
    : this->dValue := 0.0
    : CHANNEL_VALUE (this->Channel) := 0.0
    : this->dLastVelocity := 0.0
    : this->dLastPosition := 0.0
    : return := OK
]]
    
[[D tModulatedPositionActuator::SetModulationParameters (
      const tModulationParams& NewParameters) ] =
        tModulatedVelocityActuator::SetModulationParameters ]]

[[D ttModulatedPositionActuator::ModulationParameters (void) ] =
    tModulatedVelocityActuator::ModulationParameters ]]
        
[[D tModulatedPositionActuator::EnableModulaion (void) ] =
    tModulatedVelocityActuator::EnableModulation ]]

[[D tModulatedPositionActuator::DisableModulation (void) ] =
    tModulatedVelocityActuator::DisableModulation ]]

[[D tModulatedPositionActuator::IsModulationEnabled (void) ] =
    tModulatedVelocityActuator::IsModulationEnabled ]]

[[D tModulatedPositionActuator::MaxChannelUnits (void) ] = 
    tModulatedVelocityActuator::MaxChannelUnits ]]

    
[[D tModulatedPositionActuator::CountsPhysicalUnits (void) ] =
    tModulatedVelocityActuator::CountsPhysicalUnits  ]]

[[D tModulatedPositionActuator::PhysicalUnitsToCounts (void) ] =
    tModulatedVelocityActuator::PhysicalUnitsToCounts ]]
    
[[D tModulatedPositionActuator::MaxAcceleration (void) ] =
    tModulatedVelocityActuator::MaxAcceleration ]]

[[D tModulatedPositionActuator::MaxAccelerationInChannelUnits (void) ] =
    tModulatedVelocityActuator::MaxAccelerationInChannelUnits ]]

 
[[D tModulatedPositionActuator::MaxVelocity (void) ] =
    tModulatedVelocityActuator::MaxVelocity ]]

[[D tModulatedPositionActuator::MaxVelocityInChannelUnits (void) ] =
    tModulatedVelocityActuator::MaxVelocityInChannelUnits ]]


[[D tModulatedPositionActuator::OutputRate (void) ] = 
    tModulatedVelocityActuator::OutputRate ]] ]

[[D tModulatedPositionActuator::Acceleration (void) ]
Compute average acceleration of actuator assuming constant position update rate.
[ invalid_pointer (this) => UNDEFINED 
| true => return := ((this->dValue - this->dLastPosition) * this->dOutputRate - this->dLastVelocity) * this->dOutputRate
]]


[[D tModulatedPositionActuator::Velocity (void) ]
Compute average velocity of actuator assuming constant position update rate.
[ invalid_pointer (this) => UNDEFINED 
| true => return := (this->dValue - this->dLastPosition) * this->dOutputRate
]]


[[D tModulatedPositionActuator::Position (void) ]
Return last posiiton sent to actuator
[ invalid_pointer (this) => UNDEFINED 
| true => return := this->dValue
]]


[[D tModulatedPositionActuator::MinPosition (void) ] 
[ invalid_pointer (this) => UNDEFINED 
| true => return := this->dMinPosition
]]


[[D tModulatedPositionActuator::MinPositionInChannelUnits (void)]
Returns min position allowed in channel units
[ invalid_pointer (this) => UNDEFINED
| true => return := this->CdMinPosition in channel units
]]


[[D tModulatedPositionActuator::MaxPosition (void) ] 
[ invalid_pointer (this) => UNDEFINED 
| true => return := this->dMaxPosition
]]


[[D tModulatedPositionActuator::MaxPositionInChannelUnits (void)]
Returns max position allowed in channel units
[ invalid_pointer (this) => UNDEFINED
| true => return := this->CdMaxPosition in channel units
]]


[[D tModulatedPositionActuator::Output (TIME tUpdateTime) ] = tModulatedVelocityActuator::Output ]]


[[D tModulatedPositionActuator::OutputValue (TIME tUpdateTime, double dNewValue) ]
Output dNewValue with modulation if enabled to actuator connected to this->Channel limited
by by max/min position, velocity, and acceleration.
[ invalid_pointer (this) ||
  invalid_reference (this->Channel) ||
  invalid_channel (this->Channel) => UNDEFINED
| this->bEnable == FALSE =>
    this->dValue := LimitPositionVelocityAcceleration (
      dNewValue + this->dOffset, this->dValue, dNewValue, this->dLastVelocity, 
      this->CdMaxPosition, this->CdMinPosition,
      this->CdMaxVelocity, this->CdMaxAcceleration)
    : this->dModulationValue := 0.0
    : CHANNEL_OUTPUT (this->Channel) := LimitPositionVelocityAcceleration (
      dNewValue + this->dOffset, this->dValue, dNewValue, this->dLastVelocity, 
      this->CdMaxPosition, this->CdMinPosition, this->CdMaxVelocity, this->CdMaxAcceleration) + this->dOffset
    : this->dLastPosition := this->dValue
    : this->dLastVelocity := (this->dValue - this->dLastPosition) * this->dOutputRate
    : return := OK 
| true =>
    this->dValue := LimitPositionVelocityAcceleration (
      dNewValue + this->dOffset, this->dValue, dNewValue, this->dLastVelocity, 
      this->CdMaxPosition, this->CdMinPosition,
      this->CdMaxVelocity, this->CdMaxAcceleration)
    : this->dModulationValue := MODULATION (this->ModulationParameters, tUpdateTime)
    : CHANNEL_OUTPUT (this->Channel) := LimitPositionVelocityAcceleration (
      dNewValue + this->dOffset, this->dValue, dNewValue, this->dLastVelocity, 
      this->CdMaxPosition, this->CdMinPosition,
      this->CdMaxVelocity, this->CdMaxAcceleration) + this->dOffset + MODULATION (this->ModulationParameters, tUpdateTime)
    : this->dLastPosition := this->dValue
    : this->dLastVelocity := (this->dValue - this->dLastPosition) * this->dOutputRate
    : return := OK
]]


[[D tModulatedPositionActuator::UnmodulatedOutput (TIME tUpdateTime) ] = tModulatedVelocityActuator::UnmodulatedOutput ]]



[[D tModulatedPositionActuator::UnmodulatedOutputValue (TIME tUpdateTime,
      double dNewValue) ]
Output dNewValue without modulation to actuator connected to this->Channel limited
by by max/min position, velocity, and acceleration.
[ invalid_pointer (this) ||
  invalid_reference (this->Channel) ||
  invalid_channel (this->Channel) => UNDEFINED
| true =>
    this->dValue := LimitPositionVelocityAcceleration (
      dNewValue + this->dOffset, this->dValue, dNewValue, this->dLastVelocity, 
      this->CdMaxPosition, this->CdMinPosition,
      this->CdMaxVelocity, this->CdMaxAcceleration)
    : CHANNEL_OUTPUT (this->Channel) := LimitPositionVelocityAcceleration (
      dNewValue + this->dOffset, this->dValue, dNewValue, this->dLastVelocity, 
      this->CdMaxPosition, this->CdMinPosition,
      this->CdMaxVelocity, this->CdMaxAcceleration) + this->dOffset
    : this->dLastPosition := this->dValue
    : this->dLastVelocity := (this->dValue - this->dLastPosition) * this->dOutputRate
    : return := OK
]]




*******************protected specification**************************************************

[[D tModulatedPositionActuator::LimitPosition (double dOldPosition, double dNewPosition,
      doubled dOldVelocity) ]
[ invalid_pointer (this) => UNDEFINED
| true => return := LimitPositionVelocityAcceleration (dOldPosition, dNewPosition,
    dOldVelocity, this->CdMaxPosition in channel units, 
    this->CdMinPosition in channel units, this->CdMaxVelocity in channel units, 
    this->CdMaxAcceleration in channel units )
]]

[[D tModulatedPositionActuator::SetMaxAcceleration (double dMaxAccInPhysicalUnits) ] =
    tModulatedVelocityActuator::SetMaxAcceleration ]]

[[D tModulatedPositionActuator::SetMaxVelocity (double dMaxVelInPhysicalUnits) ] =
    tModulatedVelocityActuator::SetMaxVeloicty ]]

[[D tModulatedPositionActuator::SetLastVelocity (double dVelocity) ] = tModulatedVelocityActuator::SetLastVelocity ]]


[[D tModulatedPositionActuator::LastVelocity (void) ] = tModulatedVelocityActuator::LastVelocity ]]

[[D tModulatedPositionActuator::LimitVelocity (
      double dOldVelocity, double dNewVelocity) ] =
    tModulatedVelocityActuator::LimitVelocity ]]

[[D tModulatedPositionActuator::LimitValue_dd (double dValue, double dLimit) ] =
    tModulatedVelocityActuator::LimitValue_dd ]]
    
[[D tModulatedPositionActuator::LimitValue_ddd (double dValue, double dLimit) ] =
    tModulatedVelocityActuator::LimitValue_ddd ]]

[[D tPositionModulatedActuator::ChannelOutput (double dOutputValue) ] = tModulatedVelocityActuator::ChannelOutput ]]

*****************/


class tOutputDestination {
  public:
    tOutputDestination ();
    tOutputDestination (double dOffsetIn);
    virtual void SetValue (double dNewValue);
    inline double Value (void) const;
    inline void SetOffset (double dNewOffset);
    inline double Offset () const {return this->dOffset;}
    virtual ERROR_CODE Output (TIME tUpdateTime) = 0;
    ERROR_CODE OutputValue (TIME tUpdateTime, double dNewValue);  
    virtual ERROR_CODE Reset (void);
  protected:
    double LimitValue (double dNewValue, double dLimit) const;
    double LimitValue (double dNewValue, double dLowerLimit, double dUpperLimit) const;
  private:
    double dValue;
    double dOffset;
};



class tActuator : public tOutputDestination {
  public:
    tActuator (tOutputChannel &ChannelIn,
            double dPhysicalUnitsPerElectricalUnit);
    tActuator (tOutputChannel &ChannelIn,
               double dPhysicalUnitsPerElectricalUnit,
               double dOffsetIn);
    virtual ERROR_CODE Reset ();
    double PhysicalUnitsToCounts (double dPhysicalValue) const {
      return dPhysicalValue / this->dPhysicalUnitsPerCount; }
     double CountsToPhysicalUnits (double dCounts) const {
      return dCounts * this->dPhysicalUnitsPerCount; }
    virtual double MaxValueInChannelUnits () const;
    virtual double MinValueInChannelUnits () const;
    virtual ERROR_CODE Output (TIME tUpdateTime);
  protected:
    ERROR_CODE ChannelOutput (double dOutputValue);
  private:
    tOutputChannel &Channel;
    double dPhysicalUnitsPerCount;  
};


class tModulatedActuator : public tActuator {
 public:
   tModulatedActuator (tOutputChannel &ChannelIn,
              double dPhysicalUnitsPerElectricalUnit);
   tModulatedActuator (tOutputChannel &ChannelIn,
              double dPhysicalUnitsPerElectricalUnit,
              const tModulationParams& InitialParameters);
   tModulatedActuator (tOutputChannel &ChannelIn,
              double dPhysicalUnitsPerElectricalUnit,
              const tModulationParams& InitialParameters,
              double dOffsetIn);
    ERROR_CODE SetModulationParameters (const tModulationParams &);
    const tModulationParams& ModulationParameters () const;
    double ModulationValue () const;
    ERROR_CODE EnableModulation ();
    ERROR_CODE DisableModulation ();
    BOOL IsModulationEnabled () const;
    ERROR_CODE Reset ();
    virtual ERROR_CODE Output (TIME tUpdateTime);
    virtual ERROR_CODE UnmodulatedOutput (TIME tUpdateTime);
    ERROR_CODE UnmodulatedOutputValue (TIME tUpdateTime, double);
  
  private:
    tModulation Modulation;
    double dModulationValue;
};


class tModulatedAccelerationActuator : public tModulatedActuator {
  public:
    tModulatedAccelerationActuator (tOutputChannel &ChannelIn,
              double dPhysicalUnitsPerElectricalUnit,
              double dMaxAccelerationIn, double dOutputRateIn);
    tModulatedAccelerationActuator (tOutputChannel &ChannelIn,
              double dPhysicalUnitsPerElectricalUnit,
              double dMaxAccelerationIn, double dOutputRateIn,
              const tModulationParams& InitialParameters);
    tModulatedAccelerationActuator (tOutputChannel &ChannelIn,
              double dPhysicalUnitsPerElectricalUnit,
              double dMaxAccelerationIn, double dOutputRateIn,
              const tModulationParams& InitialParameters,
              double dOffsetIn);
    double MaxValueInChannelUnits () const { return MaxAccelerationInChannelUnits ();}
    double MinValueInChannelUnits () const { return -MaxAccelerationInChannelUnits ();}
    double MaxAcceleration () const;
    double MaxAccelerationInChannelUnits () const;
    double OutputRate () const { return this->dOutputRate; }
    void SetValue (double dNewValue);
    double Acceleration () const { return this->Value (); }
  protected:
    double LimitAcceleration (double dAcceleration) const;
    void SetMaxAcceleration (double dMaxAccelInPhysicalUnits) { 
      this->dMaxAcceleration = this->PhysicalUnitsToCounts (dMaxAccelInPhysicalUnits); }
  private: 
    double dMaxAcceleration; // In channel units 
    double dOutputRate; // In channel units 
    
};


class tModulatedVelocityActuator : public tModulatedAccelerationActuator {
  public:
    tModulatedVelocityActuator (tOutputChannel &ChannelIn,
          double dPhysicalUnitsPerElectricalUnit,
          double dMaxAccelerationIn, double dMaxVelocityIn, double dOutputRateIn);
    tModulatedVelocityActuator (tOutputChannel &ChannelIn,
          double dPhysicalUnitsPerElectricalUnit,
          double dMaxAccelerationIn, double dMaxVelocityIn, double dOutputRateIn,
          const tModulationParams& ModulationParameters);
    tModulatedVelocityActuator (tOutputChannel &ChannelIn,
          double dPhysicalUnitsPerElectricalUnit,
          double dMaxAccelerationIn, double dMaxVelocityIn, double dOutputRateIn,
          const tModulationParams& ModulationParameters,
          double dOffsetIn);
    double MaxValueChannelUnits () const { return MaxVelocityInChannelUnits ();}
    double MinValueChannelUnits () const { return -MaxVelocityInChannelUnits ();}
    double MaxVelocity () const;       
    double MaxVelocityInChannelUnits () const;
    void SetValue (double dNewValue);
    ERROR_CODE Reset ();
    double Acceleration () const;
    double Velocity () const { return this->Value ();}
  protected: 
    void SetMaxVelocity (double dMaxVelInPhysicalUnits) { 
      this->dMaxVelocity = this->PhysicalUnitsToCounts (dMaxVelInPhysicalUnits); }
    double LimitVelocity (double dOldVelocity, double dNewVelocity) const;
    void SetLastVelocity (double dVelocity) { this->dLastVelocity = dVelocity;}
    double LastVelocity () const { return this->dLastVelocity;}
  private:
    double dMaxVelocity;    // In channel units
    double dLastVelocity;
};



class tModulatedPositionActuator : public tModulatedVelocityActuator {
  public:
    tModulatedPositionActuator (tOutputChannel &ChannelIn,
          double dPhysicalUnitsPerElectricalUnit,
          double dMaxAccelerationIn, double dMaxVelocityIn,
          double dMinPositionAllowed, double dMaxPositionAllowed, double dOutputRateIn);
    tModulatedPositionActuator (tOutputChannel &ChannelIn,
          double dPhysicalUnitsPerElectricalUnit,
          double dMaxAccelerationIn, double dMaxVelocityIn,
          double dMinPositionAllowed, double dMaxPositionAllowed, double dOutputRateIn,
          const tModulationParams& InitialParameters);
    tModulatedPositionActuator (tOutputChannel &ChannelIn,
          double dPhysicalUnitsPerElectricalUnit,
          double dMaxAccelerationIn, double dMaxVelocityIn,
          double dMinPositionAllowed, double dMaxPositionAllowed, double dOutputRateIn,
          const tModulationParams& InitialParameters,
          double dOffsetIn);
    double MaxValueChannelUnits () const { return MaxVelocityInChannelUnits ();}
    double MinValueChannelUnits () const { return MinPositionInChannelUnits ();}
    double MinPosition () const;
    double MinPositionInChannelUnits () const;
    double MaxPosition () const;
    double MaxPositionInChannelUnits () const;
    void SetValue (double dNewValue);
    ERROR_CODE Reset ();
    double Acceleration () const;
    double Velocity () const;
    double Position () const { return this->Value ();}

  protected: 
    double LimitPosition (double dOldPosition, double dNewPosition, 
                          double dOldVelocity) const;
  
  private:
    const double CdMinPosition;
    const double CdMaxPosition;    // In physical units, the same as scale factor
    double dLastPosition;
};



/************ERROR CODES THAT SHOULD GO ELSEWHERE ************************/

#define S_outputDest_TACTUATOR_CONSTRUCTOR              (M_outputDest | 0x1001)
#define S_outputDest_TACTUATOR_PHYSICAL_UNITS_PER_COUNT (M_outputDest | 0x1002)
#define S_outputDest_TMODULATEDACTUATOR_CONSTRUCTOR     (M_outputDest | 0x1003)
#define S_outputDest_TMODULATEDACTUATOR_RESET_MODULATION   (M_outputDest | 0x1004)
#define S_outputDest_TMODULATEDACTUATOR_SET_MODULATION  (M_outputDest | 0x1005)
#define S_outputDest_TMOD_ACCEL_ACTUATOR_CONSTRUCTOR    (M_outputDest | 0x1006)
#define S_outputDest_GET_MAX_ACCEL                      (M_outputDest | 0x1007)
#define S_outputDest_TMOD_VEL_ACTUATOR_CONSTRUCTOR      (M_outputDest | 0x1008)
#define S_outputDest_GET_MAX_VEL                        (M_outputDest | 0x1009)
#define S_outputDest_TMOD_POSITION_ACTUATOR_CONSTRUCTOR (M_outputDest | 0x100A)
#define S_outputDest_GET_MIN_POSITION                   (M_outputDest | 0x100B)
#define S_outputDest_GET_MAX_POSITION                   (M_outputDest | 0x100C)


/******************************* INLINE METHODS *****************************/

/******************************************************************
* tOutputDest::tOutputDest (void)
* 
* DESCRIPTION: 
[[U2 tOutputDestination::tOutputDestination (void) ]]
*
[[U3 tOutputDestination::tOutputDestination (void) ]]
*/
inline tOutputDestination::tOutputDestination ()
{ 
  this->dValue = 0.0; 
  this->dOffset = 0.0; 
}


/******************************************************************
* tOutputDestination::tOutputDestination (double dOffsetIn)
* 
* DESCRIPTION: 
[[U2 tOutputDestination::tOutputDestination_d (dOffsetIn) ]]
*
[[U3 tOutputDestination::tOutputDestination_d (dOffsetIn) ]]
*/
inline tOutputDestination::tOutputDestination (double dOffsetIn)
{ 
  this->dValue = 0.0; 
  this->dOffset = dOffsetIn;
}


/******************************************************************
* tOutputDestination::SetValue (double dNewValue)
* 
* DESCRIPTION: 
[[U2 tOutputDestination::SetValue (dNewValue) ]]
*
[[U3 tOutputDestination::SetValue (dNewValue) ]]
*/
inline void tOutputDestination::SetValue (double dNewValue)
{ 
  this->dValue = dNewValue;
}


/******************************************************************
* tOutputDestination::Value ()
* 
* DESCRIPTION: 
[[U2 tOutputDesinationt::Value () ]]
*
[[U3 tOutputDestination::Value () ]]
*/
inline double tOutputDestination::Value () const
{ 
  return this->dValue;
}

/******************************************************************
* tOutputDestination::SetOffset (double dNewOffset)
* 
* DESCRIPTION: 
[[U2 tOutputDestination::SetOffset (dNewOffset) ]]
*
[[U3 tOutputDestination::SetOffset (dNewOffset) ]]
*/
inline void tOutputDestination::SetOffset (double dNewOffset) 
{ 
  this->dOffset = dNewOffset;
}


/******************************************************************
*
* tOutputDestination::OutputValue (double dNewValue)
*
* DESCRIPTION: 
[[U2 tOutputDestination::OutputValue (dNewValue) ]]
*
[[U3 tOutputDestination::OutputValue (dNewValue) ]]
*/
inline ERROR_CODE tOutputDestination::OutputValue (TIME tUpdateTime, double dNewValue)
{

   // Virtual call which may limit the value..
   
   /****
   [[U tOutputDestination::SetValue (dNewValue) ]]
   ****/
   this->SetValue (dNewValue);

   /****
   [[U tOutputDestination::Output (tUpdateTime) ]]
   ****/
   return this->Output (tUpdateTime);
}

/******************************************************************
*
* tOutputDestination::Reset ()
*
* DESCRIPTION: 
[[U2 tOutputDestination::Reset () ]]
*
[[U3 tOutputDestination::Reset () ]]
*/
inline ERROR_CODE tOutputDestination::Reset ()
{

   this->dValue = 0.0;
   return OK;
}




#endif  /* INC_OutputDesth */
