/* $RCSfile: pidLibF.h,v $  -- structure header for pidLibF.c */
/*
$Header: /home/chara/cvsroot/ople/sys/include/pidLibF.h,v 1.1.1.1 1998/06/11 21:36:13 chara Exp $
*/
/* Richard Johnson - Jet Propulsion Laboratory */

/*
modification history
--------------------
04a,22Oct96,beh   Converted from old pidLibF.c
03c,12jun90,mmc
*/

/*
DESCRIPTION
This module contains a PID (proportional,integral,derivative)
compensator, plus routines to calculate the compensator
coefficients.
*/


#ifndef INC_pidLibFh
#define INC_pidLibFh

/********************** Specification Functions ***********************

double PID (tPidDouble Pid, double x) :==
[ return := LimitValue (Pid.p * x + Pid.i * x + Pid.imem  + 
              Pid.d * (x - Pid.xmem) + Pid.dpole * Pid.dmem, Pid.dLimit)
]

tPidDouble UpdatePid (tPidDouble Pid, double x) :==
[ return := tPidDouble with
    xmem  := x
    imem := LimitValue (Pid.imem + Pid.i * x, Pid.dLimit)
    dmem := LimitValue (Pid.d * (x - Pid.xmem) + Pid.dpole * Pid.dmem, Pid.dLimit)
]

***************** Stimuli *********************************************

[[D ::limit (double dValue, double dLimit) ]
[ dValue > dLimit => return := dLimit
| dValue < -dLimit => return := -dLimit
| true => return := dValue 
]]

[[D tPidDouble::tPidDouble (void) ]
[ invalid_pointer (this) => UNDEFINED
| true => *this := anything
]]


[[D tPidDouble::tPidDouble_ddddd (double gain, double tau1, 
                           double tau2, double taud, double tsamp) ]
Initialize pid coeficients using poles and zeros
[ invalid_pointer (this) ||
  tau1 == 0 ||
  tau2 == 0 ||
  tsamp + taud == 0 => UNDEFINED
|true =>
   this->p := gain * (tau1 + tau2)
   : this->i := gain * tsamp
   : this->d  := gain * (tau1 * tau2) / (tsamp + taud)
   : this->dpole := taud / (tsamp + taud)
   : this->dLimit := 32766.99   
   : this->imem := 0.0
   : this->dmem := 0.0
   : this->xmem := 0.0
]]


[[D tPidDouble::tPidDouble_dddddd (double dGainIn, double dPin, double dIIn,
                double dDIn, double dTauD, double dSamplePeriod) ]
Initialize pid coeficients using p i d coeficients.
[ invalid_pointer (this) ||
  tau1 == 0 ||
  tau2 == 0 ||
  dSamplePeriod + dTauD == 0 => UNDEFINED
|true =>
   this->p := dGainIn * dPin
   : this->i := dGainIn * dIIn * dSamplePeriod
   : this->d := dGainIn * dDIn / (dSamplePeriod + dTauD)
   : this->dpole := dTauD / (dSamplePeriod + dTauD)
   : this->dLimit := 32766.99  
   : this->imem := 0.0
   : this->dmem := 0.0
   : this->xmem := 0.0
]]

[[D tPidDouble::InitializeUsingPolesAndZeros (double gain, double tau1, 
                           double tau2, double taud, double tsamp) ]
Initialize pid coeficients using poles and zeros
[ invalid_pointer (this) ||
  tau1 == 0 ||
  tau2 == 0 ||
  tsamp + taud == 0 => UNDEFINED
|true =>
   this->p := gain * (tau1 + tau2)
   : this->i := gain * tsamp
   : this->d  := gain * (tau1 * tau2) / (tsamp + taud)
   : this->dpole := taud / (tsamp + taud)
   : this->dLimit := 32766.99      
   : this->imem := 0.0
   : this->dmem := 0.0
   : this->xmem := 0.0
]]


[[D tPidDouble::InitializeUsingPids (double dGainIn, double dPin, double dIIn,
                double dDIn, double dTauD, double dSamplePeriod) ]
Initialize pid coeficients using poles and zeros
[ invalid_pointer (this) ||
  tau1 == 0 ||
  tau2 == 0 ||
  dSamplePeriod + dTauD == 0 => UNDEFINED
|true =>
   this->p := dGainIn * dPin
   : this->i := dGainIn * dIIn * dSamplePeriod
   : this->d := dGainIn * dDIn / (dSamplePeriod + dTauD)
   : this->dpole := dTauD / (dSamplePeriod + dTauD)
   : this->dLimit := 32766.99   
   : this->imem := 0.0
   : this->dmem := 0.0
   : this->xmem := 0.0
]]

[[D tPidDouble::Reset (double dInitialX) ]
Resets pid to an initial state dInitialX
[ invalid_pointer (this) => UNDEFINED
| true =>
    this->imem := 0.0
    : this->dmem := 0.0
    : this->xmem := dInitialX
]]

[[D tPidDouble::Calculate (double x) ]
Computes the PID control law for x using state of this.
[ invalid_pointer (this) => UNDEFINED
| true =>
    *this := PidUpdate (*this, x)
    : return := PID (*this, x)
]] 

[[D tPidDouble::SetLimit (double dLimitToUse) ]
[ invalid_pointer (this) => UNDEFINED
| true => this->dLimit := dLimitToUse
]] 


[[D tPidDouble::P (void) ]
[ invalid_pointer (this) => UNDEFINED
| true => return := this->p
]] 


[[D tPidDouble::I (void) ]
[ invalid_pointer (this) => UNDEFINED
| true => return := this->i
]] 


[[D tPidDouble::D (void) ]
[ invalid_pointer (this) => UNDEFINED
| true => return := this->d
]] 


[[D tPidDouble::DPole (void) ]
[ invalid_pointer (this) => UNDEFINED
| true => return := this->dpole
]] 



[[D ::CopyPidMemory (tPidDouble& Desitination, const tPidDouble& Source) ]
Copy memory values of pid allowing smooth transition betwee pids
[ invalid_reference (Desitination, Source) => UNDEFINED
| true =>
    Desitination.xmem := Source.xmem
    : Destination.dmem := Source.dmem
    : Destination.imem :=  Source.imem
]]
    
[[D tPidDouble::Display (void) ]
[ invalid_pointer (this) => UNDEFINED
| true => display contents of pid on console
]]

******************/



inline double limit (double dValue, double dLimit) {
  if (dValue > dLimit)   return dLimit;
  if (dValue < -dLimit)  return -dLimit;
  return dValue;
}

class tPidDouble {
  public:
    tPidDouble () {}
    tPidDouble (double gain, double tau1, double tau2, 
                double taud, double tsamp);
    tPidDouble (double gain, double p, double i,
                double d, double taud, double tsamp);
    void InitializeUsingPolesAndZeros (double gain, double tau1, 
                           double tau2, double taud, double tsamp);
    void InitializeUsingPids (double gain, double p, double i,
                           double d, double taud, double tsamp);
    void SetLimit (double dLimitToUse) { this->dLimit = dLimitToUse; }
    double P () const { return this->p;}
    double I () const { return this->i;}
    double D () const { return this->d;}
    double DPole () const { return this->dpole;}
    void Display (void);
    void Reset (double dInitialX);
    double Calculate (double dX);
    friend void CopyPidMemory (tPidDouble& Destination, const tPidDouble& Source);
  
  private:
    double xmem;  /* x memory */
    double imem;  /* integral memory */
    double dmem;  /* derivative memory */
    double p; 	/* proportional coeff */
    double i; 	/* integral coeff */
    double d; 	/* derivative coeff */
    double dpole; /* derivative-term pole */
    double dLimit; /* Value to limit output to */
};


#endif  /* INC_pidLibFh */
