/* $RCSfile: ringBuffer.h,v $ */
/* $Header: /home/chara/cvsroot/ople/sys/include/ringBuffer.h,v 1.1.1.1 1998/06/11 21:36:14 chara Exp $ */

/*
modification history
--------------------
04c,4Dec97, rlj    written
*/

#ifndef INC_ringBufferh
#define INC_ringBufferh


/*****
[[D tRingBuffer::tRingBuffer (int iRingSizeToUse) ]
[ invalid_pointer (this) ||
  unable to allocate memory for iRingSizeToUse Elements => UNDEFINED
| iRingSizeToUse < = 0 => *this := anything
| true =>
    *this := a RING_BUFFER with iRingSizeToUse contents are garbage
]]

[[D tRingBuffer::tRingBuffer (int iRingSizeToUse, const tDataType& FillValue) ]
[ invalid_pointer (this) ||
  unable to allocate memory for iRingSizeToUse Elements => UNDEFINED
| iRingSizeToUse < = 0 => *this := anything
| true =>
    *this := a RING_BUFFER with iRingSizeToUse Elements contents are FillValue
]]

[[D tRingBuffer::~tRingBuffer (void) ]
[ invalid_pointer (this)  => UNDEFINED
| true =>
    this->iRingSize Element freed to heap
    : *this := anything
]]

[[D tRingBuffer::operator_index (int iRingIndex) ]
[ invalid_pointer (this)  => UNDEFINED
| true =>
    return := const reference to ring element that is iRingIndex away from the last
    Appended element ( negative indices give previous values)
]]

[[D tRingBuffer::Append (const tDataType& NewElement) ]
[ invalid_pointer (this) ||
  invalid_reference (NewElement) => UNDEFINED
| true =>
   *this := *this + NewElement
   : return := return const reference to NewElement
]]


[[D tRingBuffer::Fill (const tDataType& FillValue) ]
[ invalid_pointer (this) ||
  invalid_reference (FillValue) => UNDEFINED
| true => All elements of *this := FillValue
]]


*****/

template <class tDataType> class tRingBuffer {
  public:
    tRingBuffer (int iRingSizeToUse);
    tRingBuffer (int iRingSizeToUse, const tDataType& FillValue);
    ~tRingBuffer ();
    const tDataType& operator[] (int iRingIndex) const;
    const tDataType& Append (const tDataType& NewElement);
    void Fill (const tDataType& FillValue);
    int Size () const  { return this->iRingSize;} 

  private:
    tRingBuffer (const tRingBuffer<tDataType>& RingBuffer) {} // not used
    void operator= (const tRingBuffer<tDataType>& RingBuffer) {} // not used
    int iRingSize;
    int iCurrentElement;
    tDataType *RingElementArray;
};


/******************************************************************
*
* tRingBuffer<tDataType>::tRingBuffer (int iRingSizeToUse) 
*
* DESCRIPTION: 
*
[[U tRingBuffer<tDataType>::tRingBuffer (iRingSizeToUse) ]]
****/
template <class tDataType> tRingBuffer<tDataType>::tRingBuffer (int iRingSizeToUse) :
  iRingSize (MAX (0,iRingSizeToUse) ),
  iCurrentElement (0),
  RingElementArray (new tDataType [iRingSize])
{}


/******************************************************************
*
* tRingBuffer<tDataType>::tRingBuffer (int iRingSizeToUse, const tDataType& FillValue) 
*
* DESCRIPTION: 
*
[[U tRingBuffer<tDataType>::tRingBuffer (iRingSizeToUse, FillValue) ]]
****/
template <class tDataType> tRingBuffer<tDataType>::tRingBuffer (int iRingSizeToUse, 
    const tDataType& FillValue) :
  iRingSize (MAX (0,iRingSizeToUse) ),
  iCurrentElement (0),
  RingElementArray (new tDataType [iRingSize])
{
  this->Fill (FillValue);
}


/******************************************************************
*
* tRingBuffer<tDataType>::~tRingBuffer () 
*
* DESCRIPTION: 
*
[[U tRingBuffer<tDataType>::~tRingBuffer (void) ]]
****/
template <class tDataType> tRingBuffer<tDataType>::~tRingBuffer ()
{
  delete[] this->RingElementArray;
}



/******************************************************************
*
* tRingBuffer<tDataType>::operator[] (int iRingIndex)
*
* DESCRIPTION: 
*
[[U tRingBuffer<tDataType>::operator_index (iRingIndex) ]]
****/
template <class tDataType> const tDataType& tRingBuffer<tDataType>::
  operator[] (int iRingIndex) const 
{
  int iArrayIndex = (this->iCurrentElement + iRingIndex) % this->iRingSize;

  if (iArrayIndex < 0) iArrayIndex += this->iRingSize;
  
  return this->RingElementArray [iArrayIndex];
}


/******************************************************************
*
* tRingBuffer<tDataType>::Append (const tDataType& NewElement)
*
* DESCRIPTION: 
*
[[U tRingBuffer<tDataType>::Append (NewElement) ]]
****/
template <class tDataType> const tDataType& tRingBuffer<tDataType>::
  Append (const tDataType& NewElement) 
{

  this->iCurrentElement = (this->iCurrentElement + 1) % this->iRingSize;

  this->RingElementArray [this->iCurrentElement] = NewElement;
  return NewElement;
}


/******************************************************************
*
* tRingBuffer<tDataType>::Fill (const tDataType& FillValue)
*
* DESCRIPTION: 
*
[[U tRingBuffer<tDataType>::Fill (FillValue) ]]
****/
template <class tDataType> void tRingBuffer<tDataType>::Fill (const tDataType& FillValue)
{
  int i;

  for (i = 0; i < this->iRingSize; i++)  this->RingElementArray [i] = FillValue;
 
  this->iCurrentElement = 0;

} 

        
#endif INC_ringBufferh
