/* $RCSfile: search.h,v $ - Implements abstract hardware and software input classes */
/* CM Info:
$Header: /home/chara/cvsroot/ople/sys/include/search.h,v 1.1.1.1 1998/06/11 21:36:14 chara Exp $
*/
/* Ken Starr - Jet Propulsion Laboratory */
 
/*
modification history
--------------------
01a,kms,14Nov97    written
*/
 
/*
DESCRIPTION

*/

#ifndef INC_searchh
#define INC_searchh

#include "ftTypes.h"
#include "isit.h"
#include "modLib.h"
 



/*******************************************************************
************************  tSearch Black Box  ***********************
********************************************************************
*
* tSearch
*
* A tSearch object is an objectwhich will provide a moving search pattern 
* target for initial fringe search. It will start at zero and
* return the next location each time UpdateTarget is called. The search 
* pattern will start over starting again at zero when
* Reset is called.  
*
* Currently, two search patterns are supported SPIRAL_GEOMETRIC (0,1,2,4,8...)
* and SPIRAL_ARITHMETRIC. (0,1,2,3,4...). 
*
* BB State Variables
*
* Variable                    	Initial Value
* ------------------------------------------------------------
* dSpiralRadiusInitial		Passed in during construction
* dSpiralRadiusLimit		Passed in during construction
* dSpiralStep			Passed in during construction
* dProgressionConstant		Passed in during construction
* stSearchType			Passed in during construction
*				(SPIRAL_GEOMETRIC or SPIRAL_ARITHMETRIC) 
*
* dLastStepValue	        0.0
* iCurrentStepDirection		0 
* dCurrentSpiralMaximum		dSpiralRadiusInitial
* iCurrentSpiralCycle		0
*
*******************************************************************
*
* tSearch::tSearch
*
* DESCRIPTION: Constructor
*
* RETURNS: Nothing
*

[[D tSearch::tSearch_ddddst (const double dSpiralRadiusInitialToUse,		 
			 const double dSpiralRadiusLimitToUse,		 
			 const double dSpiralStepToUse,
			 const double dProgressionConstantToUse,		 
			 const SEARCH_TYPE stSearchTypeToUse ) ]
  [ invalid_pointer (this) => UNDEFINED
  | true =>    
	dSpiralRadiusInitial := dSpiralRadiusInitialToUse			   
	: this->dSpiralRadiusLimit := dSpiralRadiusLimitToUse
	: this->dSpiralStep := dSpiralStepToUse
	: this->dProgressionConstant := dProgressionConstantToUse
	: this->stSearchType := stSearchTypeToUse
	: this->dLastStepValue := 0.0
	: this->iCurrentStepDirection := 0 
	: this->dCurrentSpiralMaximum := dSpiralRadiusInitial
	: this->iCurrentSpiralCycle := 0
 ]]


*
*******************************************************************
*
* tSearch::tSearch
*
* DESCRIPTION: Constructor
*
* RETURNS: Nothing
*

[[D tSearch::tSearch_t (const tFtSearchParameters& FtSearchParametersToUse) ]
  [ invalid_pointer (this) || invalid_reference(FtSearchParametersToUse)
                                                               => UNDEFINED
  | true =>    
	dSpiralRadiusInitial := FtSearchParametersToUse.dSpiralRadiusInitial 			   
	: this->dSpiralRadiusLimit := FtSearchParametersToUse.dSpiralRadiusLimit 
	: this->dSpiralStep := FtSearchParametersToUse.dSpiralStep 
	: this->dProgressionConstant := FtSearchParametersToUse.dProgressionConstant 
	: this->stSearchType := FtSearchParametersToUse.stSearchType 
	: this->dLastStepValue := 0.0
	: this->iCurrentStepDirection := 0  
	: this->dCurrentSpiralMaximum := FtSearchParametersToUse.dSpiralRadiusInitial
	: this->iCurrentSpiralCycle := 0
 ]]

*
*******************************************************************
*
* tSearch::~tSearch
*
* DESCRIPTION: Compiler-supplied destructor
*
* RETURNS: Nothing
*

[[D tSearch::~tSearch (void) ]
  [ invalid_pointer (this) => UNDEFINED
   | true => *this := anything
]]


*
*******************************************************************
*
* tSearch::UpdateTarget
*
* DESCRIPTION: Computes new Search Target value using last values.
*		   
* New Search Target is calulated from last value and stSearchType, 
	dSpiralRadiusInitial, dSpiralRadiusLimit, dSpiralStep, 
	dProgressionConstant, dLastStepValue, iCurrentStepDirection, 
	dCurrentSpiralMaximum, iCurrentSpiralCycle.

  Update will return the next value in the search pattern taking steps of size
  dSpiralStep until the dCurrentSpiralMaximum is reached. Once the 
  dCurrentSpiralMaximum	is reached a new value for dCurrentSpiralMaximum is set 
  based on the Search Type and the  CurrentStepDirection is reversed.
	
  Currently, two search patterns are supported SPIRAL_GEOMETRIC (0,1,2,4,8...)
  and SPIRAL_ARITHMETRIC (0,1,2,3,4...).
 
* RETURNS: double
*

 [[D tSearch::UpdateTarget (void) ]
  [ invalid_pointer (this) || 
    dProgressionConstant < 0 ||
    dSpiralRadiusInitial < 0 ||
    dSpiralStep < 0  ||
    stSearchType is not SPIRAL_GEOMETRIC or SPIRAL_ARITHMETRIC 
                                => UNDEFINED
    
  |  iCurrentStepDirection == 0 =>
     dLastStepValue := 0
     : iCurrentStepDirection := 1 
     : dCurrentSpiralMaximum := dSpiralRadiusInitial
     : iCurrentSpiralCycle := 0
     : return := 0
     
  |  (dLastStepValue + (iCurrentStepDirection * dSpiralStep)) >= dSpiralRadiusLimit =>
     dLastStepValue := dSpiralRadiusLimit 
     : return := dSpiralRadiusLimit
        
  |  (dLastStepValue + (iCurrentStepDirection * dSpiralStep)) <= dSpiralRadiusLimit =>
     dLastStepValue := -dSpiralRadiusLimit 
     : return := dSpiralRadiusLimit
        
  | abs((dLastStepValue + (iCurrentStepDirection * dSpiralStep)) 
                     > dCurrentSpiralMaximum) && (stSearchType == SPIRAL_GEOMETRIC) =>
     dLastStepValue := dLastStepValue + (iCurrentStepDirection * dSpiralStep)
     : dCurrentSpiralMaximum :=  (dProgressionConstant^(iCurrentSpiralCycle+1)) *
     				   dSpiralRadiusInitial
     : dCurrentStepDirection := - (dCurrentStepDirection)
     : iCurrentSpiralCycle := iCurrentSpiralCycle + 1	
     : return := dLastStepValue + (iCurrentStepDirection * dSpiralStep)
     
  | abs((dLastStepValue + (iCurrentStepDirection * dSpiralStep)) 
                   > dCurrentSpiralMaximum) && (stSearchType == SPIRAL_ARITHMETRIC)  =>
     dLastStepValue := dLastStepValue + (iCurrentStepDirection * dSpiralStep)
     : dCurrentSpiralMaximum := (iCurrentSpiralCycle + 1) *
     				   dProgressionConstant *
     				   dSpiralRadiusInitial
     : dCurrentStepDirection := - (dCurrentStepDirection)
     : iCurrentSpiralCycle := iCurrentSpiralCycle + 1			    
     : return := dLastStepValue + (iCurrentStepDirection * dSpiralStep)
     
  | true =>
     dLastStepValue := dLastStepValue + (iCurrentStepDirection * dSpiralStep)
     : return := dLastStepValue + (iCurrentStepDirection * dSpiralStep)
]]


*
*******************************************************************
*
* tSearch::Reset
*
* DESCRIPTION: Resets the Search pattern
*
* RETURNS: ERROR_CODE
*

[[D tSearch::Reset (void) ]
  [ invalid_pointer (this)  => UNDEFINED
  | true =>
     iCurrentStepDirection := 0
     :  dLastStepValue := 0
     : dCurrentSpiralMaximum := dSpiralRadiusInitial
     : iCurrentSpiralCycle := 0
     : return := OK
]]

*
*

*
*******************************************************************
*
* tSearch::GetValue
*
* DESCRIPTION: Returns Last Search Target
*
* RETURNS: double
*

[[D tSearch::GetValue (void) ]
  [ invalid_pointer (this)  => UNDEFINED
  | true => return := dLastStepValue
]]

*
*


*
*******************************************************************
*
* tSearch::tSetParameters
*
* DESCRIPTION: Set search's controlling parameters
*
* RETURNS: ERROR_CODE
*
*

[[D tSearch::SetParameters_ddddst (const double dSpiralRadiusInitialToUse,		 
			 const double dSpiralRadiusLimitToUse,		 
			 const double dSpiralStepToUse,
			 const double dProgressionConstantToUse,		 
			 const SEARCH_TYPE stSearchTypeToUse ) ]
  [ invalid_pointer (this) => UNDEFINED
  | true =>    
	dSpiralRadiusInitial := dSpiralRadiusInitialToUse			   
	: dSpiralRadiusLimit := dSpiralRadiusLimitToUse
	: dSpiralStep := dSpiralStepToUse
	: dProgressionConstant := dProgressionConstantToUse
	: stSearchType := stSearchTypeToUse
	: return := OK
 ]]
*

*
*******************************************************************
*
* tSearch::SetParameters
*
* DESCRIPTION: Set search's controlling parameters
*
* RETURNS: ERROR_CODE
*
*

[[D tSearch::SetParameters_t (const tFtSearchParameters& FtSearchParametersToUse) ]
  [ invalid_pointer (this) => UNDEFINED
  | true =>    
	dSpiralRadiusInitial := FtSearchParametersToUse.dSpiralRadiusInitial 			   
	: dSpiralRadiusLimit := FtSearchParametersToUse.dSpiralRadiusLimit 
	: dSpiralStep := FtSearchParametersToUse.dSpiralStep 
	: dProgressionConstant := FtSearchParametersToUse.dProgressionConstant 
	: stSearchType := FtSearchParametersToUse.stSearchType 
	: return := OK
 ]]
*
*
*******************************************************************
*
* tSearch::SearchParameters
*
* DESCRIPTION: Return current values of search parameters
*
* RETURNS: tFtSearchParameters 
*
*

[[D tSearch::SearchParameters_t () ]
  [ invalid_pointer (this) => UNDEFINED
  | true =>    
	Temp_FtSearchParameters.dSpiralRadiusInitial := dSpiralRadiusInitial 			   
	: Temp_FtSearchParameters.dSpiralRadiusLimit := dSpiralRadiusLimit 
	: Temp_FtSearchParameters.dSpiralStep := dSpiralStep 
	: Temp_FtSearchParameters.dProgressionConstant := dProgressionConstant 
	: Temp_FtSearchParameters.stSearchType := stSearchType 
	: return := Temp_FtSearchParameters
 ]]
*/




/*******************************************************************/
/*******************************************************************/
/********************                     **************************/
/********************  Class Definitions  **************************/
/********************                     **************************/
/*******************************************************************/
/*******************************************************************/


/*******************************************************************/
/***********************  tSearch  ************************/
/*******************************************************************/


class tSearch {
public:
  tSearch(const double dSpiralRadiusInitialToUse,		 
	   const double dSpiralRadiusLimitToUse,		 
	   const double dSpiralStepToUse,
	   const double dProgressionConstantToUse,		 
	   const SEARCH_TYPE stSearchTypeToUse);
	   
  tSearch(const tFtSearchParameters& FtSearchParametersToUse);
  	   
  double UpdateTarget(void);
  
  ERROR_CODE Reset(void);
  
  double GetValue(void) const { return(dLastStepValue); }
   
  ERROR_CODE SetParameters(const double dSpiralRadiusInitialToUse,
                const double dSpiralRadiusLimitToUse,
                const double dSpiralStepToUse,
                const double dProgressionConstantToUse,
                const SEARCH_TYPE stSearchTypeToUse );
	   	
   ERROR_CODE SetParameters(const tFtSearchParameters& FtSearchParametersToUse);
   
   tFtSearchParameters SearchParameters () const;
	   		  
private:
  double dSpiralRadiusInitial;		
  double dSpiralRadiusLimit;	
  double dSpiralStep;		
  double dProgressionConstant;	
  SEARCH_TYPE stSearchType;		

  double dLastStepValue	;       
  int iCurrentStepDirection;	
  double dCurrentSpiralMaximum;		
  int iCurrentSpiralCycle;		
};

#endif  /* ftINC_searchh */

