/* $RCSfile: telemetryManager.h,v $*/
/* CM Info:
$Header: /home/chara/cvsroot/ople/sys/include/telemetryManager.h,v 1.1.1.1 1998/06/11 21:36:15 chara Exp $
*/
/* Richard L. Johnson - Jet Propulsion Laboratory */
 

/*
DESCRIPTION

Class definition for the telemetry manager.
*/


#ifndef INC_telemetryManagerh
#define INC_telemetryManagerh

#include "isit.h"
#include "telemetryEvent.h"
#include "telemetryItem.h"

#include "avlHybridTree.h"
#include <msgQLib.h>
#include "relayStream.h"



/************

Specification Functions 



************************* tTelemetryManager Specification ************************************

[[D tTelemetryManager::tTelemetryManager (const tTelemetryManagerInfo& Info) ]
[ invalid_pointer (this) ||
  invalid_reference (Info) => UNDEFINED
| true =>
    this->CtMinimumSampleTime := Info.tMinimumSampleTime
    : this->CommandStream := IPC_FIFO (Info.iiCommandStreamId)
    : this->TelemetryRegistry := Empty
    : this->EventQueue := VxWorks msg queue of size Info.iEventQueueSize etc.
    : this->BufferedQueue :=  VxWorks msg queue of size Info.iEventQueueSize etc.
    : this->EventTelemetryHandler := EventDrivenTask ("EventTelemetryHandler",..., 
      Info.iEventDrivenPriority, this->ProcessEventTelemetry)
    : this->BufferedTelemetryHandler := EventDrivenTask ("BufferedTelemetryHandler",..., 
      Info.iBufferedPriority, this->ProcessBufferedTelemetry)
    : this->TelemetrySampler := enabled PeriodicTask ("TelemetrySampler",..., Info.iSamplerPriority,
      Info.tSamplerPeriod, Info.tSamplerOffset, this->ProcessBufferedTelemetry)

]]


[[D tTelemetryManager::RegisterEventDrivenTelemetry (
      const tTelemetryId& Id, tEventDrivenTelemetryBase& Item) ]
[ invalid_pointer (this) ||
  invalid_reference (Id, Item) => UNDEFINED
| Entry exisits for Id in this->TelemetryRegistry => return NullEvent
| true =>
    this->TelemetryRegistry := this->TelemetryRegistry with new entry {Id, &Item}
    : return := TelemetryEvent (this->miEventDrivenQueue, &Item)
]]


[[D tTelemetryManager::RegisterBufferedTelemetry (
      const tTelemetryId& Id, tEventDrivenTelemetryBase& Item) ]
[ invalid_pointer (this) ||
  invalid_reference (Id, Item) => UNDEFINED
| Entry exisits for Id in this->TelemetryRegistry => return NullEvent
| true =>
    this->TelemetryRegistry := this->TelemetryRegistry with new entry {Id, &Item}
    : return := TelemetryEvent (this->miBufferedQueue, &Item)
]]


[[D tTelemetryManager::RegisterSampledTelemetry (
      const tTelemetryId& Id, tEventDrivenTelemetryBase& Item) ]
[ invalid_pointer (this) ||
  invalid_reference (Id, Item) => UNDEFINED
| Entry exisits for Id in this->TelemetryRegistry => return NullEvent
| true =>
    this->TelemetryRegistry := this->TelemetryRegistry with new entry {Id, &Item}
    : return := TelemetryEvent (this->miBufferedQueue, &Item)
]]


*************************** Private Specification ***********************************************



    
****/


class tTelemetryManagerInfo {
  public:
    const tIpcFifoId& CommandStreamId () const;
    const tIpcFifoId& TelemetryStreamId () const;
    TIME MinimumSamplePeriod () const;
    int EventDrivenPriority () const;
    int BufferedPriority () const;
    int SamplerPriority () const;
    TIME SamplerPeriod () const;
    TIME SamplerOffset () const;
    int CommandProcessorPriority () const;
    TIME CommandProcessorPeriod () const;
    TIME CommandProcessorOffset () const;
    int EventQueueSize () const;
    int EventDrivenTelemetryBufferSize () const;
    int BufferedTelemetryBufferSize () const;
    int SampledTelemetryBufferSize () const;
};

typedef int TELEMETRY_ID;
    
class tTelemetryCommand {
  public:
    TELEMETRY_ID tiTelemetryId;
    BOOL bSelected;
    int iPriority;
    TIME tSamplePeriod;
    TIME tSampleOffset;
};


TELEMETRY_TYPE TelemetryType (TELEMETRY_ID tiTelemetryId);

#define S_telemetry_UNKNOWN_ITEM 1
#define S_telemetry_UNKNOWN_TELEMETRY_TYPE 2

class tTelemetryManager {
  public:
    tTelemetryManager (const tTelemetryManagerInfo& Info);
    ~tTelemetryManager ();
    ERROR_CODE Restart ();
    ERROR_CODE Shutdown ();
    BOOL IsOk ();
    tTelemetryEvent RegisterTelemetryItem (tEventDrivenTelemetryItemBase& Item);
    tTelemetryEvent RegisterTelemetryItem (tBufferedTelemetryItemBase& Item);
    void RegisterTelemetryItem (tSampledTelemetryItemBase& Item);
    
    private:
      ERROR_CODE ProcessCommand (TIME tSchedTime);
      ERROR_CODE ConfigureEventDrivenItem (TELEMETRY_ID Id, int iPriorityIn, BOOL bSelectedIn);
      ERROR_CODE ConfigureBufferedItem (TELEMETRY_ID Id, int iPriorityIn, BOOL bSelectedIn);
      ERROR_CODE ConfigureSampledItem (TELEMETRY_ID Id, int iPriorityIn, TIME tSamplePeriod, 
                                 TIME tSampleOffset, BOOL bSelectedIn);
 
      void ProcessEventTelemetry ();
      void ProcessBufferedTelemetry ();
      ERROR_CODE ProcessSampledTelemetry (TIME tSampleTime);
      
      ERROR_CODE SendTelemetry (tEventDrivenTelemetryItemBase* Item, TIME tTime);
      ERROR_CODE SendTelemetry (tBufferedTelemetryItemBase* Item, TIME tTime);
      ERROR_CODE SendTelemetry (tSampledTelemetryItemBase* Item, TIME tTime);
      
      TIME RoundTime (TIME tTimeToRound);
      
      tIpcFifoAccessor<tTelemetryCommand> CommandStream;
      tRelayStreamSender TelemetryStream;
      
      tAvlHybridTree <TELEMETRY_ID, tEventDrivenTelemetryItemBase*> EventDrivenRegistry;
      tAvlHybridTree <TELEMETRY_ID, tBufferedTelemetryItemBase*> BufferedRegistry;
      tAvlHybridTree <TELEMETRY_ID, tSampledTelemetryItemBase*> SampledRegistry;
      tAvlHybridTree <int, tSampledTelemetryItemBase*> PrioritizedSampled;
        
      const TIME CtMinimumSamplePeriod;
      const TIME CtTimeStep;

      MSG_Q_ID miEventQueue;
      MSG_Q_ID miBufferedQueue;
      SEM_ID siSampledMutex;
    
      int iCommandProcessorTaskId;
      int iEventHandlerTaskId;
      int iBufferHandlerTaskId;
      int iSamplerTaskId;
      
      int iEventDrivenTelemetryBufferSize;
      int iBufferedTelemetryBufferSize;
      int iSampledTelemetryBufferSize;

      BYTE* pEventDrivenTelemetryBuffer;
      BYTE* pBufferedTelemetryBuffer;
      BYTE* pSampledTelemetryBuffer;
      
      
};     
      
#endif INC_telemetryManagerh      
      
