/****************************************************************

NAME
     msgQLib - message queue library

SYNOPSIS
     msgQCreate() - create and initialize a message queue
     msgQDelete() - delete a message queue
     msgQSend() - send a message to a message queue
     msgQReceive() - receive a message from a message queue
     msgQNumMsgs() - get the number of messages queued to a message queue

     MSG_Q_ID msgQCreate (int maxMsgs, int maxMsgLength, int options)
     STATUS msgQDelete (MSG_Q_ID msgQId)
     STATUS msgQSend (MSG_Q_ID msgQId, char * buffer, 
                      UINT nBytes, int timeout, in...
     int msgQReceive (MSG_Q_ID msgQId, char * buffer, 
                      UINT maxNBytes, int timeout)
     int msgQNumMsgs (MSG_Q_ID msgQId)


DESCRIPTION
     This library contains routines  for creating and using  mes-
     sage  queues,  the primary intertask communication mechanism
     within a single CPU.  Message queues allow a variable number
     of  messages  (varying  in length) to be queued in first-in-
     first-out (FIFO) order.  Any task or interrupt service  rou-
     tine  can  send  messages  to a message queue.  Any task can
     receive messages from a message queue.  Multiple  tasks  can
     send  to  and  receive  from  the same message queue.  Full-
     duplex communication between two  tasks  generally  requires
     two message queues, one for each direction.

CREATING AND USING MESSAGE QUEUES
     A message queue is created with msgQCreate().   Its  parame-
     ters  specify  the  maximum  number  of messages that can be
     queued to that message queue and the maximum length in bytes
     of  each message.  Enough buffer space will be pre-allocated
     to accommodate the specified number of messages of specified
     length.

     A task or interrupt service routine sends  a  message  to  a
     message  queue with msgQSend().  If no tasks are waiting for
     messages on the message queue, the message is  simply  added
     to  the buffer of messages for that queue.  If any tasks are
     already waiting to receive a message from the message queue,
     the  message  is  immediately delivered to the first waiting
     task.

     A task receives a message from a message queue with  msgQRe-
     ceive().   If any messages are already available in the mes-
     sage  queue's  buffer,  the  first  message  is  immediately
     dequeued  and  returned  to  the caller.  If no messages are
     available, the calling task will block and  be  added  to  a
     queue  of tasks waiting for messages.  This queue of waiting
     tasks can be ordered either by task  priority  or  FIFO,  as
     specified in an option parameter when the queue is created.

TIMEOUTS
     Both msgQSend() and msgQReceive() take  timeout  parameters.
     When  sending  a message, if no buffer space is available to
     queue the message, the timeout specifies how many  ticks  to
     wait  for  space to become available.  When receiving a mes-
     sage, the timeout specifies how many ticks  to  wait  if  no
     message  is  immediately available.  The timeout parameter
     can have the special values NO_WAIT (0) or  WAIT_FOREVER  (-
     1).   NO_WAIT  means  the routine should return immediately;
     WAIT_FOREVER means the routine should never time out.

URGENT MESSAGES
     The msgQSend() routine allows the priority of a  message  to
     be  specified  as  either  normal (MSG_PRI_NORMAL) or urgent
     (MSG_PRI_URGENT).  Normal priority messages are added to the
     tail  of  the list of queued messages, while urgent priority
     messages are added to the head of the list.

INCLUDE FILES
     msgQLib.h

SEE ALSO
     pipeDrv, msgQSmLib

*/

/****************************************************************************
[[D msgQLib::msgQCreate 
   (in: maxMsgs, maxMsgLength, options;
    inout: system;
    out: return, OS.Task.errno;)]
 [ system is unable to allocate enough memory for message queue 
   and message buffers => 
     return := NULL
   : OS.Task.errno := S_memLib_NOT_ENOUGH_MEMORY
 | method is called from interrupt sevice routine => 
     return := NULL
   : OS.Task.errno := S_intLib_NOT_ISR_CALLABLE
 | options == MSG_Q_FIFO => 
     system := system with message queue created that is capable 
     of  holding  up to <maxMsgs> messages, each up to maxMsgLength
     bytes long, that will queue pended tasks in FIFO order
   : return := message queue ID that identifies the message queue created
 | options == MSG_Q_PRIORITY => 
     system := system with message queue created that is capable 
     of  holding  up to <maxMsgs> messages, each up to maxMsgLength
     bytes long, that will queue pended tasks in priority order
   : return := message queue ID that identifies the message queue created
 | true => unknown what happens if options not one of FIFO or PRIORITY
]] 
*****************************************************************************/

/****************************************************************************
[[D msgQLib::msgQSend
      (in: msgQId, buffer, nBytes, timeout, priority;
       inout: system;
       out: return;)]
[ msgQId is invalid => 
    return := ERROR
  : OS.Task.errno := S_objLib_OBJ_ID_ERROR
| buffer invalid or (size of object pointed to by buffer) < nBytes => UNDEFINED 
| message queue msgQId was deleted while this method is executing => 
    return := ERROR
  : OS.Task.errno := S_objLib_OBJ_DELETED
| timeout is set to NO_WAIT, and message queue msgQId is full =>
    return := ERROR
  : OS.Task.errno := S_objLib_OBJ_UNAVAILABLE
| message queue msgQId is full for timeout ticks =>
    return := ERROR
  : OS.Task.errno := S_objLib_OBJ_TIMEOUT
| <nBytes> is larger than the maxMsgLength set for message queue 
    msgQId at time of its creation => 
    return := ERROR
  : OS.Task.errno := S_msgQLib_INVALID_MSG_LENGTH
| called from an ISR, with timeout not set to NO_WAIT =>
    return := ERROR
  : OS.Task.errno := S_msgQLib_NON_ZERO_TIMEOUT_AT_INT_LEVEL
| priority == MSG_PRI_NORMAL => 
    system message queue msgQId := system message queue msgQId 
      with first nBytes of *buffer appended as last message 
  : return := OK
| priority == MSG_PRI_URGENT => 
    system message queue msgQId := system message queue msgQId 
      with first nBytes of *buffer appended as first message 
  : return := OK
| true => ?? what happens if priority not NORMAL or URGENT??
]]
*****************************************************************************/


/****************************************************************************
[[D msgQLib::msgQReceive 
      (in: msgQId, buffer, maxNBytes, timeout;
       inout: system;
       out: return, OS.Task.errno;)]
[ method called from interrupt service routine => UNDEFINED
| sizeof(*buffer) < maxNBytes => UNDEFINED
| msgQId is invalid =>
     return := ERROR
  : OS.Task.errno := S_objLib_OBJ_ID_ERROR
| the message queue was deleted while this method is executing =>
     return := ERROR
  : OS.Task.errno := S_objLib_OBJ_DELETED
| timeout is set  to  NO_WAIT,  and  no  messages  are available =>
     return := ERROR
  : OS.Task.errno := S_objLib_OBJ_UNAVAILABLE
| no messages were received in timeout ticks =>
     return := ERROR
  : OS.Task.errno :=  S_objLib_OBJ_TIMEOUT
| true => system message queue msgQId := system message queue msgQId
    with oldest message removed
  : *buffer := the first maxNBytes of oldest message in system message queue 
    msgQId or the number of bytes in the oldest message in the queue, whichever 
    is smaller, PLUS any bytes from the end of *buffer needed to fill up the 
    rest of *buffer

  : return := number of bytes in oldest message in system message queue msgQId, 
     or maxNBytes, whichever is smaller
]]
*****************************************************************************/

/****************************************************************************
[[D msgQLib::msgQDelete 
   (in: msgQId; 
    inout: OS; 
    out: return, OS.Task.errno;)]
[ msgQId is invalid => 
    return := ERROR
  : OS.Task.errno := S_objLib_OBJ_ID_ERROR
| called from interrupt service routine => 
    return := ERROR
  : OS.Task.errno := S_intLib_NOT_ISR_CALLABLE
| true => 
    return := OK
  : OS := OS with message queue msgQId deleted
]]
*****************************************************************************/







