/* $RCSfile: watchdog.h,v $ - Watchdog timer class */
/* CM Info:
$Header: /home/chara/cvsroot/ople/sys/include/watchdog.h,v 1.1.1.1 1998/06/11 21:36:15 chara Exp $
*/
/* Brad Hines - Jet Propulsion Laboratory */
 
/*
modification history
--------------------
01a,beh,28Sep96    written
*/
 
/*
DESCRIPTION
This file contains a C++ wrapper for VxWorks watchdog timers.
*/

#ifndef INC_watchdogh
#define INC_watchdogh

#include <wdLib.h>

#define S_watchDog_ALREADY_INITIALIZED     (M_watchDog | 0x1)
#define S_watchDog_NULL_FUNCTION_POINTER   (M_watchDog | 0x2)
#define S_watchDog_INVALID_TIMEOUT_VALUE   (M_watchDog | 0x3)
#define S_watchDog_NOT_INITIALIZED         (M_watchDog | 0x4)


/******************************************************************
*
* tWatchDog Black Box and tTaskLevelWatchDog Black Box
*
* The two classes are identical except that the handler for
* a tWatchDog timeout is called at interrupt level (with all
* its associated restrictions) while the handler for a tTaskLevelWatchDog
* is called at task level, at the priority and with the options
* specified during Initialize.
*
* Note that the functionality of tTaskLevelWatchDog cannot be
* guaranteed if there is a task, with priority higher than (i.e.,
* lower numeric value than) that specified during
* tTaskLevelWatchDog::Initialize, that does not ever release
* the CPU.
*
* USAGE!  If you use tWatchDog with a handler that is a method of
* a C++ object, you must ENSURE that you never allow the watchdog
* to timeout after the object has been deleted (i.e., after its
* this pointer becomes invalid).  Your usage pattern should include
* deletion of the tWatchDog object prior to deletion of the object
* containing the method.  Similar restrictions exist on data objects
* of any sort used by a handler, whether C or C++.
* 
*
* BB State Variable Dictionary (collectively referred to as *this.state)
* Variable                   Initial Value
* ----------------------------------------
* BOOL bInit                 FALSE
* FUNCPTR fpHandler          NULL
* int iHandlerArg            0
* int iTimeout               0
* BOOL bActive               FALSE
* int iTimeLeft              unknown
*
* Conceptual States
* -----------------
* WatchDog.Armed :== { WatchDog.bActive = TRUE,
*                      WatchDog.iTimeLeft = WatchDog.iTimeout }
*
*******************************************************************
*
* SPEC FUNCTIONS
*
*********
*
*
*******************************************************************
*
* tWatchDog(void) - constructor
*
* IN:
*   tWatchDog   *this
* 
* INOUT:
* OUT:
*   *this.fpHandler
*   *this.iHandlerArg
*   *this.iTimeout
*   *this.bInit
*   
*
* SPEC:
*
  [ this invalid -> UNDEFINED
  | true ->
      *this.fpHandler := NULL
      : *this.iHandlerArg := 0
      : *this.iTimeout := 0
      : *this.bInit := FALSE
  ] 
*
* RETURNS:
*   Nothing.
*
*******************************************************************
*
* ERROR_CODE Initialize(FUNCPTR fpHandler, int iHandlerArg,
*                       int iDefaultTimeout)
*
* IN:
*   this
*   fpHandler
*   iHandlerArg
*   iDefaultTimeout
*   list of methods of tWatchDog
*   
*   
* INOUT:
*   *this.bInit
*   VxWorks.WdTimersAvailable
* OUT:
* SPEC:
*
    [ this invalid -> UNDEFINED
    | *this.bInit == TRUE -> return S_watchDog_ALREADY_INITIALIZED
    | fpHandler == NULL -> return S_watchDog_NULL_FUNCTION_POINTER
    | ! (fpHandler is a pointer to valid C function ||
            (fpHandler is a valid pointer to a C++ method &&
            iHandlerArg is a valid pointer to a C++ object
            with method *fpHandler)) -> UNDEFINED
    | fpHandler a method of tWatchDog -> UNDEFINED
    | iDefaultTimeout < 1 -> return S_watchDog_INVALID_TIMEOUT_VALUE
    | VxWorks.WdTimersAvailable <= 0 -> return VxWorks error code (wdCreate)
    | true ->
        *this.fpHandler := fpHandler
        : *this.iHandlerArg := iHandlerArg
        : *this.iTimeout := iDefaultTimeout
        : *this.bActive := FALSE
        : *this.bInit := TRUE
        : VxWorks.WdTimersAvailable :-= 1
        : return OK
    ]
*
* RETURNS:
*   OK or S_watchDog_ALREADY_INITIALIZED or S_watchDog_NULL_FUNCTION_POINTER
*      or S_watchDog_INVALID_TIMEOUT_VALUE
*
*******************************************************************
*
* ERROR_CODE Start(void)
*
* IN:
*   this
*   *this.bInit
*   *this.fpHandler
*   *this.iHandlerArg
*   *this.iTimeout
* 
* INOUT:
*
* OUT:
*   *this.bActive
*   *this.iTimeLeft
*
* SPEC:
*
    [ this invalid -> UNDEFINED
    | *this.bInit == FALSE -> return S_watchDog_NOT_INITIALIZED
    | *this.iDefaultTimeoutInTicks < 1 ->
         return S_watchDog_INVALID_TIMEOUT_VALUE
    | !(*this.fpHandler not a pointer to a valid C function ||
         (*this.fpHandler a pointer to a valid C++ method &&
          *this.iHandlerArg is a pointer to an object of the type for
                             which the method is defined)) -> UNDEFINED
    | true ->
        *this.iTimeLeft := *this.iDefaultTimeoutinTicks
        *this.bActive := TRUE
        : return OK
    ]
*
* RETURNS:
*   OK or S_watchDog_NOT_INITIALIZED
*
*******************************************************************
*
* ERROR_CODE StartUsingTIME(TIME tTimeout)
*
* IN:
*   this
*   *this.bInit
*   *this.fpHandler
*   *this.iHandlerArg
*   tTimeout
* INOUT:
* OUT:
*   *this.bActive
*   *this.iTimeLeft
*
* SPEC:
*
    [ this invalid -> UNDEFINED
    | *this.bInit == FALSE -> return S_watchDog_NOT_INITIALIZED
    | TIME_TO_VXWORKS_TICKS(tTimeout) < 1 ->
         return S_watchDog_INVALID_TIMEOUT_VALUE
    | !(*this.fpHandler not a pointer to a valid C function ||
         (*this.fpHandler a pointer to a valid C++ method &&
          *this.iHandlerArg is a pointer to an object of the type for
                             which the method is defined)) -> UNDEFINED
    | true ->
        *this.iTimeLeft = TIME_TO_VXWORKS_TICKS(tTimeout)
        : *this.bActive := TRUE
        : return OK
    ]
*
* RETURNS:
*   OK or S_watchDog_NOT_INITIALIZED or S_watchDog_INVALID_TIMEOUT_VALUE
*
*******************************************************************
*
* ERROR_CODE StartUsingTicks(int iTimeoutInTicks)
*
* IN:
*   this
*   *this.bInit
*   *this.fpHandler
*   *this.iHandlerArg
*   iTimeoutInTicks
* INOUT:
* OUT:
*   *this.bActive
*   *this.iTimeLeft
*
* SPEC:
*
    [ this invalid -> UNDEFINED
    | *this.bInit == FALSE -> return S_watchDog_NOT_INITIALIZED
    | iTimeoutInTicks < 1 -> return S_watchDog_INVALID_TIMEOUT_VALUE
    | !(*this.fpHandler not a pointer to a valid C function ||
         (*this.fpHandler a pointer to a valid C++ method &&
          *this.iHandlerArg is a pointer to an object of the type for
                             which the method is defined)) -> UNDEFINED
    | true ->
        *this.iTimeLeft = iTimeoutInTicks
        : *this.bActive := TRUE
        : return OK
    ]
*
* RETURNS:
*   OK or S_watchDog_NOT_INITIALIZED or S_watchDog_INVALID_TIMEOUT_VALUE
*
*******************************************************************
*
* ERROR_CODE Reset(void)
*
* IN:
*   this
*   *this.bInit
* INOUT:
* OUT:
*   *this.bActive
*
* SPEC:
*
    [ this invalid -> UNDEFINED
    | *this.bInit == FALSE -> return S_watchDog_NOT_INITIALIZED
    | true ->
        *this.bActive := FALSE
        : return OK
    ]
*
* RETURNS:
*   OK or S_watchDog_NOT_INITIALIZED
*
*******************************************************************
*
* tWatchDog::~tWatchDog - destroy a tWatchDog object
*
* IN:
*   this 
* INOUT:
*   VxWorks.WdTimersAvailable
* OUT:
*   *this.bInit
*   
* SPEC:
*
  [ this invalid -> UNDEFINED
  | *this.bInit != FALSE ->
      VxWorks.WdTimersAvailable :+= 1
      : *this.bInit := FALSE
  | true ->
      *this.bInit := FALSE
  ] 
*
* RETURNS:
*   Nothing.
*/



class tWatchDog {
public:
  tWatchDog(void);
  virtual ~tWatchDog(void);
  ERROR_CODE Initialize(FUNCPTR fpHandler, int iHandlerArg,
                        int iDefaultTimeout);
  ERROR_CODE Start(void);
  ERROR_CODE StartUsingTIME(TIME tTimeout);
  ERROR_CODE StartUsingTicks(int iTimeoutInTicks);
  ERROR_CODE Reset(void);

protected:
  void WdTimeoutHandler(void);
  virtual void Handler(void);

  /************
    State Box state:

    Representational mappings:
      1. [ wdId == NULL <-> bInit == FALSE
         | wdId != NULL <-> bInit == TRUE
         ]
   ************/

  FUNCPTR fpHandler;
  int iHandlerArg;
  int iDefaultTimeoutInTicks;
  // BOOL bActive;    // Conceptual piece of state box state
  // int iTimeLeft;   // Conceptual piece of state box state
  WDOG_ID wdId;
};


/*******************************************************************
*
* tTaskLevelWatchDog Black Box
*
* GENERAL:
*   See comments for tWatchDog above.
*
*******************************************************************
* tTaskLevelWatchDog(void) - constructor
*
* IN:
*   tTaskLevelWatchDog   *this
* 
* INOUT:
* OUT:
*   *this.fpHandler
*   *this.iHandlerArg
*   *this.iTimeout
*   *this.bInit
*
* SPEC:
*
  [ this invalid -> UNDEFINED
  | true ->
      *this.fpHandler := NULL
      : *this.iHandlerArg := 0
      : *this.iTimeout := 0
      : *this.bInit := FALSE
  ] 
*
* RETURNS:
*   Nothing.
*
*******************************************************************
*
* ERROR_CODE Initialize(FUNCPTR fpHandler, int iHandlerArg,
*                       int iDefaultTimeout, int iHandlerTaskPriority,
*                       int iHandlerTaskOptions, int iHandlerTaskStackSize)
*
* This function has default arguments.  Default values are:
*   iHandlerTaskPriority  = WATCHDOG_HANDLER_DEFAULT_PRIORITY
*   iHandlerTaskOptions   = WATCHDOG_HANDLER_DEFAULT_TASK_OPTIONS
*   iHandlerTaskStackSize = WATCHDOG_HANDLER_DEFAULT_STACK_SIZE
*
* IN:
*   this
*   fpHandler
*   iHandlerArg
*   iDefaultTimeout
*   iHandlerTaskPriority
*   iHandlerTaskOptions
*   iHandlerTaskStackSize
*   list of methods of tTaskLevelWatchDog
*   
* INOUT:
*   *this.bInit
*   VxWorks.WdTimersAvailable
* OUT:
* SPEC:
*
    [ this invalid -> UNDEFINED
    | *this.bInit == TRUE -> return S_watchDog_ALREADY_INITIALIZED
    | fpHandler == NULL -> return S_watchDog_NULL_FUNCTION_POINTER
    | ! (fpHandler is a pointer to valid C function ||
            (fpHandler is a valid pointer to a C++ method &&
            iHandlerArg is a valid pointer to a C++ object
            with method *fpHandler)) -> UNDEFINED
    | fpHandler a method of tTaskLevelWatchDog -> UNDEFINED
    | iDefaultTimeout < 1 -> return S_watchDog_INVALID_TIMEOUT_VALUE
    | VxWorks.WdTimersAvailable <= 0 -> return VxWorks error code (wdCreate)
    | Task cannot be created || Insufficient memory to spawn task ->
           *this.iHandlerTaskId := ERROR
           : return VxWorks errno (taskSpawn)
    | true ->
        *this.fpHandler := fpHandler
        : *this.iHandlerArg := iHandlerArg
        : *this.iTimeout := iDefaultTimeout
        : *this.bActive := FALSE
        : *this.bInit := TRUE
        : VxWorks.WdTimersAvailable :-= 1
        : OS.SpawnedTasks := OS.SpawnedTasks U
                    { *this.TaskLevelHandler(), with
                      priority iHandlerTaskPriority,
                      options iHandlerTaskOptions,
                      stack size iHandlerTaskStackSize,
                      and name "wdHandlerTask" }
        : return OK
    ]
*
* RETURNS:
*   OK or S_watchDog_ALREADY_INITIALIZED or S_watchDog_NULL_FUNCTION_POINTER
*      or S_watchDog_INVALID_TIMEOUT_VALUE
*
*******************************************************************
*
* ERROR_CODE Start(void)
* ERROR_CODE StartUsingTIME(TIME tTimeout)
* ERROR_CODE StartUsingTicks(int iTimeoutInTicks)
* ERROR_CODE Reset(void)
*
* Identical to specs for tWatchDog::...
*
*******************************************************************
*
* tTaskLevelWatchDog::~tTaskLevelWatchDog - destroy a tTaskLevelWatchDog object
*
* IN:
*   this 
* INOUT:
*   VxWorks.WdTimersAvailable
* OUT:
*   *this.bInit
*   
* SPEC:
*
  [ this invalid -> UNDEFINED
  | *this.bInit != FALSE ->
      VxWorks.WdTimersAvailable :+= 1
      : *this.bInit := FALSE
  | true ->
      *this.bInit := FALSE
  ] 
*
* RETURNS:
*   Nothing.
*/

class tTaskLevelWatchDog : public tWatchDog {
public:
  tTaskLevelWatchDog(void) : tWatchDog() { iHandlerTaskId = ERROR; }
  virtual ~tTaskLevelWatchDog(void);
  ERROR_CODE Initialize(FUNCPTR fpHandler, int iHandlerArg,
                        int iDefaultTimeout,
            int iHandlerPriority = WATCHDOG_HANDLER_DEFAULT_PRIORITY,
            int iHandlerTaskOptions = WATCHDOG_HANDLER_DEFAULT_TASK_OPTIONS,
            int iHandlerTaskStackSize = WATCHDOG_HANDLER_DEFAULT_STACK_SIZE);
protected:
  /************
    State Box state:

    Representational mappings:
      1. [ wdId == NULL || *this.iHandlerTaskId
             is not the id of a task executing the code
             in tTaskLevelWatchDog::TaskLevelHandler<-> bInit == FALSE
         | wdId != NULL && *this.iHandlerTaskId
             is the id of a task executing the code
             in tTaskLevelWatchDog::TaskLevelHandler<-> bInit == TRUE
         ]
   ************/

  virtual void Handler(void);
  void TaskLevelHandler(void);

  int iHandlerTaskId;
};


#endif  /* INC_watchdogh */
