/* adcLib.c - Setup and interface to the Burr-Brown MPV906 adc board */
/* sccs info:	@(#)adcLib.c	1.3	17 Apr 1995 */

/*
modification history
--------------------
02k,17Apr95,beh  ANSI-fied
02j,31Aug91,mmc  added more includes
02g,30Jul91,mmc  changed includes for 5.0
02f,16Jul91,grw  added function adcFullInit
02e,15Oct90,arg  changed subroutine comments for mangen
02d,20Sep90,arg  initial release
02c,19Sep90,arg  adding comments and DEBUG levels
02b,17Sep90,arg  touching up and adding comments
02a,12Sep90,arg  revamping all code and adding standard format
01a,09Mar90,arg  written
*/

/*
DESCRIPTION
Provides setup and access to the Burr-Brown MPV906 analog to digital board.
The board is set up to handle 32 channels of differential input that are
scanned continuously.  The scan rate is set in adcInit.  Since the board
makes 64 samples per cycle (we only use 32), there is a delay of 64*(scan
period) between each read of a certain channel (currently 3.2 msec).

The function adcInit must be called before any other function.  It calculates
addresses and initilizes the boards under certain conditions.  The function
adcRead(channel) is used to obtain the latest data from the specified channel.
The function adcSetGain(channel, gain) can be used to set the gain applied
to a channel at 1, 10, or 100.  The gain is initially set to 1.  Channels
are specified by a number from 0 to ADC_CHANNELS-1.
*/

#include <vxWorks.h>
#include <vme.h>

#ifdef VW_502
#include <68k/iv.h>
#include <vxProto.h>
#endif  /* VW_502 */

#include <sysLib.h>
#include <errnoLib.h>
#include "adcLib.h"
#include "adcConfig.h"

static volatile WORD  *adcData[ADC_BOARDS];	/* word 0 of converted data */
static volatile UBYTE *adcScan[ADC_BOARDS];	/* byte 0 of the scan list */
static volatile UWORD *adcIntr[ADC_BOARDS];	/* interrupt control reg */
static volatile UWORD *adcSync[ADC_BOARDS];	/* sync/status register */
static volatile UWORD *adcTrig[ADC_BOARDS];	/* trigger control register */
static volatile UWORD *adcTime[ADC_BOARDS];	/* internal trigger period */ 


/******************************************************************************
*
* adcListAddr - local function which returns the scan list address of a channel
*
* This local function returns the address of the scan list element for
* the channel number that is passed.  Normally one would expect that the
* scan list elements would be addressed in the same order as there rank in
* the list.  This is not so.  The elements are byte reversed.  The actual
* byte order in memory is 1, 0, 3, 2, 5, 4, etc.  This explains the need
* for the following function.  It adds one to the even offsets and subtracts
* one from the odd offsets.
*
* RETURNS:
* The address of the desired scan element.
* Sets an error if channel is invalid but returns the address of channel 0.
*/

LOCAL volatile UBYTE *adcListAddr(int channel)
{

    if ((channel >= ADC_CHANNELS) || (channel < 0)) {
      errnoSet(S_adc_INVALID_CHANNEL);
      return(adcScan[0]);
    }

  if ((channel & 1) == 0)
    return((UBYTE *) (adcScan[channel/ADC_CHANNELS_PER_BOARD]+channel+1));
  else
    return((UBYTE *) (adcScan[channel/ADC_CHANNELS_PER_BOARD]+channel-1));
}


/******************************************************************************
*
* adcFullInit - Initialize boards and adcLib
*
* This function must be called before the boards can be used.  It initializes
* ADC boards and calls adcInit.
*
* RETURNS:
*  echos error status of adcInit
*/
STATUS adcFullInit(void)
{
  UBYTE num;
  int board;

  if (ERROR == adcInit())		 /* call address translation*/
    return ERROR;

  for (board = 0; board < ADC_BOARDS; board++) {
    for(num = 0; num < ADC_CHANNELS_PER_BOARD; num++)
      /* set the scan list for channels 0-31, gain 1 */
      *adcListAddr(board*ADC_CHANNELS_PER_BOARD+num) = (UBYTE) num;
    for (; num<64; num++)                       /* set last 32 bytes of scan */
      *(adcScan[board]+num) = (UBYTE) 0;	/* list to channel 0, gain 1 */
    *adcIntr[board] = (UWORD) 0x0800;		/* enable continuous mode */
    *adcTime[board] = (UWORD) 0x0009;		/* convert every 50 usec */
    *adcTrig[board] = (UWORD) 0x0001;		/* enable internal trigger */
    *adcSync[board] = (UWORD) 0x0000;		/* start sampling */
  }
  return OK;
}


/******************************************************************************
*
* adcInit - initialize adcLib
*
* This function initializes the adcLib pointers; it
* must be called before the boards can be used;
* automatically called by adcFullInit
*
* RETURNS:
* OK if addresses were calculated properly, otherwise ERROR.
* The addresses are returned in adcData, adcScan, adcIntr, adcSync, adcTrig,
* and adcTime.  
*/

STATUS adcInit(void)
{
  int board;
  /*UBYTE num;*/
  char *tmp;

  for (board = 0; board < ADC_BOARDS; board++) {
    if (sysBusToLocalAdrs((int) VME_AM_STD_SUP_DATA,
			  (char *) adcBases[board], &tmp) == ERROR) {
      /*get board's local address from its offset in the standard space*/
      errnoSet(S_adc_ADDRESS_UNAVAILABLE);
      return(ERROR);
    }
    /* find the addresses of the registers on each board */
    adcData[board] = (WORD *) tmp;
    adcScan[board] = (UBYTE *) (adcData[board] + 0x40);
    adcIntr[board] = (UWORD *) (adcData[board] + 0x80);
    adcSync[board] = (UWORD *) (adcData[board] + 0x81);
    adcTrig[board] = (UWORD *) (adcData[board] + 0x7f);
    adcTime[board] = (UWORD *) (adcData[board] + 0x7e);
  }
  return OK;
}


/******************************************************************************
*
* adcRead - get the converted data from a channel
*
* Reads the converted digital value from the specified channel.  The data
* is being converted continuously so adcRead returns the most recent value.
* The 12 bit data is shifted left by 4 bits.  This is done so that the data
* resembles a 16 bit conversion.  Channel numbers start at 0.
*
* RETURNS:
* The digital data from the channel.
* Sets an error if the channel is invalid, but returns a 0 value.
*/

int adcRead(int channel)
{

    if ((channel >= ADC_CHANNELS) || (channel < 0)) {
      errnoSet(S_adc_INVALID_CHANNEL);
      return(0);
    }

  return ((int) *(adcData[channel/ADC_CHANNELS_PER_BOARD] + channel) << 4);
}


/******************************************************************************
*
* adcSetGain - sets the gain applied to a channel to 1, 10, or 100
*
* The gain applied to the specified channel can be set to 1, 10, or 100.  No
* other value for gain is legal.  Channel numbers start at 0.
*
* RETURNS:
* Sets an error if the channel or gain are invalid, but has no return.
*/

void adcSetGain(int channel, int gain)
{

    if ((channel >= ADC_CHANNELS) || (channel < 0)) {
      errnoSet(S_adc_INVALID_CHANNEL);
      return;
    }

  switch(gain) {
  case 1:
    *adcListAddr(channel) =
      *adcListAddr(channel) & 0x3F;		/* Set top 2 bits to 00 */
    break;
  case 10:
    *adcListAddr(channel) =
      (*adcListAddr(channel) & 0x3F) | 0x40;	/* Set top 2 bits to 01 */
    break;
  case 100:
    *adcListAddr(channel) =
      (*adcListAddr(channel) & 0x3F) | 0x80;	/* Set top 2 bits to 10 */
    break;
  default:
    errnoSet(S_adc_INVALID_GAIN);
  }
}    


#if (DEBUG >= 40)
/******************************************************************************
*
* adcStatus - prints out the status of options and operations for the board
*
* This function is for debugging only.
*
* RETURNS:
* Prints out information about the status of options and operations.
* Sets an error if board is invalid but does not return anything.
*/

void adcStatus(int board)
{
  UWORD status;

  if ((board < ADC_BOARDS) && (board >= 0)) {
    status = *adcSync[board];
    printf("Status Register: %xH\n", status);
    printf("Channel %d (list number %xH) will be read next.\n",
	   (*adcListAddr(board*ADC_CHANNELS_PER_BOARD
			 + (status & 0x3f)) & 0x3f),
	   (status & 0x3f));
    if (status & 0x200)
      printf("Single-ended\n");
    else
      printf("Differential\n");
    if (status & 0x400)
      printf("Not converting\n");
    else
      printf("Converting\n");
    if (status & 0x800)
      printf("No new data\n");
    else
      printf("New data\n");
    if ((status & 0x1000)==0)
      printf("Oversampling... hey man slow down!\n");
    printf("Data ready: ");
    if ((status & 0x2000)==0)
      printf("All\n");
    else
      if ((status & 0x4000)==0)
	printf("More than half\n");
      else
	printf("Less than half\n");
    printf("Trigger: %xH\n", *adcIntr[board] & 3);
  }
  else
    errnoSet(S_adc_INVALID_BOARD);
}


/******************************************************************************
*
* adcShowData - prints out the contents of the scan list and the data scanned
*
* This function is for debugging only.  Prints out in physical order the
* channel numbers contained in the scan list and the data available from that
* channel.
*
* RETURNS:
* Sets an error if board is invalid but does not return anything.
*/

void adcShowData(int board)
{
  int a;

  if ((board >= ADC_BOARDS) || (board < 0)) {
    errnoSet(S_adc_INVALID_BOARD);
    return;
  }
  for (a=0;a<64;a++) {
    printf("%d-%d: %d  \011", a,
	   *adcListAddr(board*ADC_CHANNELS_PER_BOARD + a),
	   *(adcData[board]+a));
    if ((a % 5) == 4)
      printf("\n");
  }
}

#endif
