/* irigLib.c - routines to access Bancomm PC03V Time Code Processor */
/*    sccs info:  @(#)irigLib.c	1.3  27 Apr 1995  */

/*
modification history
--------------------
01p,27Apr95,beh    Moved pirigIntAck assignment from FullInit to Init
01o,14Sep93,beh    added "static" to local functions, made 5.1-compatible
01n,09sep91,mmc    *******irigGetTime and irigGetTimeNow now return time
                   *******in units of 50 usec, instead of 100 usec
01m,31aug91,mmc    added 2 includes, added VOIDFUNCPTR casts
01l,08aug91,mmc    added defines for irig A, B, and G
01k,30Jul91,mmc    changed includes for 5.0
01j,16Jul91,grw    added function irigFullInit, instead of INIT_PROC_NUM
01i,07jun91,mmc    added list of allowable clock rates
01h,29may91,mmc    sped up irigGetTimeUSec (12 usec on v30xe)
01g,13nov90,mmc    sped up irigGetTime significantly (36 usec on v30xe)
01f,08jun90,mmc    made irigGetTime faster
01e,08May90,beh    added use of function prototypes, etc.
01d,05may90,mmc    added IRIG_INIT_PROC_NUM to irigConfig.h to allow use
                   in a multiprocessing environment
01c,06feb90,mmc    made documentation SCCS compliant
01b,11jan90,mmc    documentation changes
01a,08jan90,mmc    initial release 
*/

/*
DESCRIPTION

This module contains subroutines to interface with the
Bancomm PC03V Time Code Processor (IRIG time code reader)

The initialization routine irigInit should be called before any other
subroutine.  The routines irigClkConnect, irigClkDisconnect,
irigClkRateGet, irigClkRateSet, irigClkEnable, irigClkDisable are
similar to the system clock routines.

The irig clock interrupt is synchronized to 1 Hz, i.e., an interrupt
will occur at a second transition.  Note that the clock rate should
divide evenly into 1 MHz in order to avoid problems.

The routines irigGetTime, irigGetTimeUSec, and irigGetTimeStruct read
the time at the last interrupt. An interrupt routine must be connected
for these to work. The routines irigGetTimeNow, irigGetTimeUSecNow, and
and irigGetTimeStructNow read the time when invoked, destroying
the strobed time; these should be used with care when interrupts are on.

irigGetTime returns the time in 50's of microseconds since 0 hrs,
which provides a high resolution system clock which fits in an int.
irigGetTimeUSec provides time at the resolution of the clock, but
with a dynamic range of only one minute.  irigGetTimeStruct provides
all available time information in a structure, defined in irigLib.h. 
*/

#include <vxWorks.h>
#include <vme.h>

#ifdef VW_502
#include <68k/iv.h>
#include <vxProto.h>
#else
#include <iv.h>
#endif   /* VW_502 */

#include <logLib.h>
#include <stdioLib.h>
#include <sysLib.h>
#include <intLib.h>

#include "rtTypes.h"
#include "irigLib.h"    /* structure definition and prototypes */
#include "irigConfig.h" /* board configuration header */

/* Local function prototypes */
static void irigClkPeriodSet(int period);
static void irigIntOn(void);
static void irigIntOff(void);
static void irigClkInt(void);
static void irigReadTime(void);
static int useconds (void);
static int mseconds (void);
static int seconds (void);
static int minutes (void);
static int hours (void);
static int days (void);
static int status (void);

static volatile UWORD *pirigBase; /* pointer to base address */
static volatile UWORD *pirigIntAck; /* pointer to interrupt acknowledge addr */
static int irigPeriod = 1000;   /* period of clock interrupt (microsec) */
static UWORD timeArray0;  /* time array */
static UWORD timeArray1;
static UWORD timeArray2;
static UWORD timeArray3;

#define NIBBLE(A,B) (0x000f & ((A) >> (4*(B))))

/* register (short) offsets from base */
#define CAPTR (0x00/2)
#define TIME0 (0x02/2)
#define TIME1 (0x04/2)
#define TIME2 (0x06/2)
#define TIME3 (0x08/2)
#define CLRIG (0x0a/2)
#define CR0   (0x0c/2)
#define CR1   (0x0e/2)
#define BIMCR0  (0x20/2)
#define BIMCR1  (0x22/2)
#define BIMCR2  (0x24/2)
#define BIMCR3  (0x26/2)
#define BIMV0 (0x28/2)
#define BIMV1 (0x2a/2)
#define BIMV2 (0x2c/2)
#define BIMV3 (0x2e/2)
#define RELFR (0x30/2)
#define STR1W3  (0x40/2)
#define STR1W2  (0x42/2)
#define STR1W1  (0x44/2)
#define STR1W0  (0x46/2)
#define STR2W3  (0x60/2)
#define STR2W2  (0x62/2)
#define STR2W1  (0x64/2)
#define STR2W0  (0x66/2)
#define CTC0  (0x80/2)
#define CTC1  (0x82/2)
#define CTC2  (0x84/2)
#define CTCCR (0x86/2)


/*********************************************
*
* irigFullInit - initialize PC03V card and irigLib.
*
* DESCRIPTION
* Initialize Bancomm card and calculate pointers for irigLib.
*
* RETURNS
* ERROR if problem with clock, or with irigInit;
* OK otherwise.
*/

STATUS irigFullInit (void)
{
  /* address translation */
  if (ERROR == irigInit())
    return ERROR;

  /* setup control register 0 */
#define IRIG_B  0x0000  /* IRIG B, forward, PLL, channel 1 */
#define IRIG_A  0x0001  /* IRIG A ... */
#define IRIG_G  0x0002  /* IRIG G ... */

  *(pirigBase + CR0) = IRIG_B;
  
  /* turn off interrupt */
  irigIntOff();
  
  /* set interrupt level and vector*/
  *(pirigBase + BIMCR1) = 0x0010 + IRIG_INT;
  *(pirigBase + BIMV1)  = 0x0000 + IRIG_VECTOR;
  
  /* set initial clock rate */
  irigClkPeriodSet(irigPeriod);
  
  /* check clock */
  {
    IrigTime tim;
    STATUS status;
    
    status = irigGetTimeStructNow (&tim);
    if (status == ERROR)
      return ERROR;
    printf(" %3d::%2d:%2d:%2d.%3d%3d\n",
     tim.day, tim.hour, tim.minute, tim.second,
     tim.msecond, tim.usecond);
  }
  return OK;
}

  
/*********************************************
*
* irigInit - initialize irigLib
*
* DESCRIPTION
* Calculate pointers for irigLib.  
*
* RETURNS
* ERROR if problems converting bus address
* OK otherwise.
*/

STATUS irigInit (void)
{
  char *localAdrs;

  /* convert bus to local addresses */
  if (sysBusToLocalAdrs((int) IRIG_AM, (char *)IRIG_ADRS, &localAdrs)== ERROR)
    return ERROR;
  /*
  printf ("\n irig: busAdrs: %lxH, localAdrs: %lxH \n", IRIG_ADRS, localAdrs);
  */
  pirigBase = (UWORD *) localAdrs;

  /* Interrupt Acknowledge address */
  pirigIntAck = pirigBase + RELFR;
  
  return OK;
}



static void irigClkPeriodSet(int period) /* set clock_period to period usecs */
{
  /* set CTC counter to mode 5, write LSB first */
  *(pirigBase + CTCCR) = 0x003a;
  /* set lsb and msb of count (in order). Count = period - 1 */
  *(pirigBase + CTC0) = 0x00ff & (period - 1);
  *(pirigBase + CTC0) = 0x00ff & ((period-1) >> 8);
}

static void irigIntOn(void)   /* enable interrupts */
{
/* CR1 = 0x2005 = sync heartbeat to 1pps, enable secondary capture on heartbea*/
#define CR1_VALUE 0x2005 
#define CR1_INTEN0 0x0010
  *(pirigBase + CR1) = CR1_VALUE | CR1_INTEN0; 
}

static void irigIntOff(void)  /* disable interrupts */
{
  /* disable interrupts */
  *(pirigBase + CR1) = CR1_VALUE;
}

/*************************************************************
*
* irigClkRateSet - set frequency of irig clock
*
* DESCRIPTION
* The interrupt period will be set to 1MHz / rate, rate >=16.
* Since the interrupt resyncs on the second, the interrupt rate
* should be evenly divisible into 1 MHz to avoid problems. 
*
* Allowable rates: 16, 20, 25, 32, 40, 50, 64, 80,
* 100, 125, 160, 200, 250, 320, 400, 500, 625, 800,
* 1k, 1.25k, 1.6k, 2k, 2.5k, 3.125k, 4k, 5k, 6.25k, 8k,
* 10k, 12.5k, 15.625k, 20k, 25k, 31.25k, 40k, 50k, 62.5k,
* 100k, 125k, 200k, 250k, 500k, 1M.
*
* RETURNS
* rate to which clock set (Hz)
* or ERROR if board not initialized
*
*/
int irigClkRateSet(int rate)
{
#ifdef DEBUG
  /* return error if board not initialized */
  if(pirigBase == 0)
  {
    errnoSet(S_irig_BOARD_NOT_INTIALIZED);
    return ERROR;
  }
#endif
  if (rate < 16) rate = 16; /* minimum rate */
  irigClkPeriodSet(irigPeriod = 1000000/rate );
  return 1000000 / irigPeriod;
}
  

/****************************************************************
*
* irigClkRateGet - get frequency of irig clock
*
* RETURNS
* rate of clock (Hz) or ERROR if board not initialized
*
*/
int irigClkRateGet(void)
{
#ifdef DEBUG
  /* return error if board not initialized */
  if(pirigBase == 0)
  {
    errnoSet(S_irig_BOARD_NOT_INTIALIZED);
    return ERROR;
  }
#endif
  return 1000000 / irigPeriod;
}


/******************************************************************
*
* irigClkEnable - enable irig clock
*
* RETURNS
* OK or ERROR if board not initialized
*
*/
STATUS irigClkEnable(void)
{
#ifdef DEBUG
  /* return error if board not initialized */
  if(pirigBase == 0)
  {
    errnoSet(S_irig_BOARD_NOT_INTIALIZED);
    return ERROR;
  }
#endif
  irigIntOn();
  return OK;
}


/****************************************************************
*
* irigClkDisable - disable irig clock
*
* RETURNS
* OK or ERROR if board not initialized
*
*/
STATUS irigClkDisable(void)
{
#ifdef DEBUG
  /* return error if board not initialized */
  if(pirigBase == 0)
  {
    errnoSet(S_irig_BOARD_NOT_INTIALIZED);
    return ERROR;
  }
#endif
  irigIntOff();
  return OK;
}




static FUNCPTR irigClkRoutine = (FUNCPTR) logMsg;
static int irigClkArg = (int) "irig clock interrupt\n";

static void irigClkInt(void) /* handle an irig clock interrupt */
{
  FAST UWORD tmp;

  tmp = *pirigIntAck;            /* release capture lockout*/
  (*irigClkRoutine)(irigClkArg); /* call user routine */
}


/****************************************************************
*
* irigClkConnect - connect a user routine to the irig clock interrupt
*
* RETURNS
* ERROR if invalid routine, problems connecting the routine
* or enabling the interrupt level, or board not initialized;
* OK otherwise
*
*/
STATUS irigClkConnect(FUNCPTR routine, int arg)
{
  STATUS status;

#ifdef DEBUG
  /* return error if board not initialized */
  if(pirigBase == 0)
  {
    errnoSet(S_irig_BOARD_NOT_INTIALIZED);
    return ERROR;
  }
  /* check for invalid address */
  if( ((int)routine == 0) || ((int)routine == -1) )
  {
    errnoSet(S_irig_NULL_FUNC_POINTER);
    return ERROR;
  }
#endif
  irigClkRoutine = routine; /* set clock interrupt routine */
  irigClkArg = arg;
  status = intConnect ((VOIDFUNCPTR *)INUM_TO_IVEC(IRIG_VECTOR),
              (VOIDFUNCPTR) irigClkInt, 0);
  if(status == ERROR)
    return ERROR;
  return (sysIntEnable ((int) IRIG_INT)); /*enable this bus interrupt level */
}


/****************************************************************
*
* irigClkDisconnect - disconnect a user routine from the irig clk interrupt
*
* RETURNS
* ERROR if problem disabling an interrupt level or board
* not initialized; OK otherwise
*
*/
STATUS irigClkDisconnect(void)
{
  irigIntOff(); /* turn off irig interrupts */
  irigClkRoutine = (FUNCPTR) logMsg;  /* connect dummy routine */
  irigClkArg = (int) "irig clock interrupt\n";
  return (sysIntDisable ((int) IRIG_INT));  /* disable bus interrupt */
}


#define IRIG_OK 5

/*********************************************************
*
* irigGetTime - get irig time at last strobe, 50's usec
*
* RETURNS
* time in units of 50 usec since 0 hours at last strobe
*
*/
int irigGetTime (void)
{
  timeArray1 = *(pirigBase + TIME0 + 1);  /* read time */
  timeArray2 = *(pirigBase + TIME0 + 2);  /* read time */
  timeArray3 = *(pirigBase + TIME0 + 3);  /* read time */
  return    NIBBLE(timeArray3,1) / 5
        + NIBBLE(timeArray3,2) * 2

      + NIBBLE(timeArray3,3) * 20
      + NIBBLE(timeArray2,0) * 200
      + NIBBLE(timeArray2,1) * 2000

      + NIBBLE(timeArray2,2) * 20000
      + NIBBLE(timeArray2,3) * 200000

      + NIBBLE(timeArray1,0) * 1200000
      + NIBBLE(timeArray1,1) * 12000000

      + NIBBLE(timeArray1,2) * 72000000
      + NIBBLE(timeArray1,3) * 720000000;
}

/*********************************************************
*
* irigGetTimeNow - get irig time immediately, 50's usec
*
* RETURNS
* time in units of 50 usec since 0 hours immediately, 
* destroying the value stored at the last stroble
*
*/
int irigGetTimeNow (void)
{
  FAST UWORD tmp;

  tmp = *pirigIntAck; /* release capture lockout */
  tmp = *(pirigBase + CAPTR); /* capture time */
  return irigGetTime();
}


/*********************************************************
*
* irigGetTimeUSec - get irig time at last strobe, usecs
*
* RETURNS
* time in units of usec since the last minute at the last strobe 
*
*/
int irigGetTimeUSec (void)
{
  timeArray2 = *(pirigBase + TIME0 + 2);  /* read time */
  timeArray3 = *(pirigBase + TIME0 + 3);  /* read time */
  return    NIBBLE(timeArray3,0)
      + NIBBLE(timeArray3,1) * 10
      + NIBBLE(timeArray3,2) * 100
  
      + NIBBLE(timeArray3,3) * 1000
      + NIBBLE(timeArray2,0) * 10000
      + NIBBLE(timeArray2,1) * 100000

      + NIBBLE(timeArray2,2) * 1000000
      + NIBBLE(timeArray2,3) * 10000000;
}

/**********************************************************
*
* irigGetTimeUSecNow - get irig time immediately, usecs
*
* RETURNS
* time in units of usec since last minute, immediately, 
* destroying the value stored at the last strobe
*
*/
int irigGetTimeUSecNow (void)
{
  FAST UWORD tmp;

  tmp = *pirigIntAck; /* release capture lockout */
  tmp = *(pirigBase + CAPTR); /* capture time */
  return irigGetTimeUSec();
}


/**********************************************************
*
* irigGetTimeStruct - get irig time at strobe, structure
*
* DESCRIPTION
* Get the time at the last strobe. 
* Complete time information (days -- usec) is returned
* in the structure.  
*
* RETURNS
* OK if clock ok; ERROR otherwise
*
*/
STATUS irigGetTimeStruct (IrigTime *ptime)
{
#if (DEBUG>=20)
  /* Check for NULL pointer */
  if (ptime == NULL)
  {
    errnoSet(S_irig_NULL_POINTER_ARGUMENT);
    return(ERROR);
  }
#endif
  irigReadTime();
  ptime->day = days();
  ptime->hour = hours();
  ptime->minute = minutes();
  ptime->second = seconds();
  ptime->msecond = mseconds();
  ptime->usecond = useconds();
  if(status() != IRIG_OK)
    return ERROR;
  else
    return OK;
}


/**********************************************************
*
* irigGetTimeStructNow - get irig time immediately, structure
*
* DESCRIPTION
* Get the time immediately, destroying the value stored at the
* last strobe. Complete time information (days -- usec) is returned
* in the structure.  
*
* RETURNS
* OK if clock ok; ERROR otherwise
*
*/
STATUS irigGetTimeStructNow (IrigTime *ptime)
{
  FAST UWORD tmp;

  tmp = *pirigIntAck; /* release capture lockout */
  tmp = *(pirigBase + CAPTR); /* capture time */
  return irigGetTimeStruct (ptime);
}


static void irigReadTime(void)
{
  timeArray0 = *(pirigBase + TIME0 + 0);  /* read time */
  timeArray1 = *(pirigBase + TIME0 + 1);  /* read time */
  timeArray2 = *(pirigBase + TIME0 + 2);  /* read time */
  timeArray3 = *(pirigBase + TIME0 + 3);  /* read time */
  /*printf("%x, %x, %x, %x \n",timeArray0,timeArray1,timeArray2,timeArray3);*/
}

static int useconds (void)
{
  return    NIBBLE(timeArray3,0) +
            NIBBLE(timeArray3,1) * 10 +
            NIBBLE(timeArray3,2) * 100;
}

static int mseconds (void)
{
  return    NIBBLE(timeArray3,3) +
            NIBBLE(timeArray2,0) * 10 +
            NIBBLE(timeArray2,1) * 100;
}

static int seconds (void)
{
  return    NIBBLE(timeArray2,2) +
            NIBBLE(timeArray2,3) * 10;
}

static int minutes (void)
{
  return    NIBBLE(timeArray1,0) +
            NIBBLE(timeArray1,1) * 10;
}

static int hours (void)
{
  return    NIBBLE(timeArray1,2) +
            NIBBLE(timeArray1,3) * 10;
}

static int days (void)
{
  return    NIBBLE(timeArray0,0) +
            NIBBLE(timeArray0,1) * 10 +
            NIBBLE(timeArray0,2) * 100;
}

static int status (void)
{
  return    NIBBLE(timeArray0,3);
}


