/* laserLib.c - routines to access the JPL laser fringe counter board.  */
/* sccs info :     @(#)laserLib.c	1.2  5/4/94  */

/*
modification history
--------------------
01k,beh,25Sep91     LaserFullInit no longer tests or clears laser counter
01j,mmc,31Aug91     Added 3 vxWorks includes
01i,beh,29Aug91     Made changes to lboardTemp stuff to avoid gcc warnings
01h,beh,06Jul90     Added small delays in a couple of spots to laserTest.
01g,beh,06Jul90     Changed a couple of things to work with DEBUG=30
01f,beh,05Jul90     Changed laserTest to print diagnostics in hex, clean up properly
01e,beh,01Jun90     Fixed laserTest to be robust, added 64-bit locked read stuff
01d,beh,11May90     Completed cosmetics
01c,beh,03May90     Put in final form - changes to init, test routines
01b,beh,19Feb90     Put source code in standard format
01a,beh,24Jan90     written, debugged
*/

/*
DESCRIPTION
This module contains all the routines necessary to manipulate the laser counter
board hardware.  These routines use D32 accesses to the board.  If you do D16 or D8
reads or writes, you will need to make a lot of revisions to this library.  The
detailed operation of the laser board hardware is explained in detail in the
"Laser Counter Board Hardware Manual."  The use of these drivers is described
in the "Laser Counter Board Software User's Guide."
*/

#include <vxWorks.h>
#include <vme.h>

#ifdef VW_502
#include <vxProto.h>
#endif

#include <errnoLib.h>
#include <stdio.h>
#include <sysLib.h>
#include "rtTypes.h"
#include "laserConfig.h"
#include "laserLib.h"

/*---------------------------------------------------------------
* These are Pointers to various on-board registers initialized by
* laserInit.
* "STATIC" is used simply to hide these declarations from the
* user; it may be removed by changing the STATIC define below.
* 
* Note that two of the pointers are defined as UWORD.  This
* is because they are used to pick 16-bit values directly out of a
* 32-bit field, and this eliminates the ambiguity that can arise in
* a case where the 32-bit number is positive but bit 15 is a 1.  By using
* an UWORD, there's no danger of having something funny happen
* with the sign or two's complement stuff.
*/

#define STATIC static
STATIC int lboardTemp;  /* Initial destination of pointers -- used to keep pointers
      		   from pointing into never-never land when we start up */
STATIC volatile char *lboardAddress 		= (char *) &lboardTemp ;
STATIC volatile char *testModule    		= (char *) &lboardTemp ;
STATIC volatile char *statusModule  		= (char *) &lboardTemp ;

#define pLWORDINIT ((LWORD *) &lboardTemp)
#define pWORDINIT ((WORD *) &lboardTemp)
#define pUWORDINIT ((UWORD *) &lboardTemp)

STATIC volatile LWORD *cpr[LASER_NUM_COUNTERS + 1]  = { pLWORDINIT,
				                    pLWORDINIT, pLWORDINIT, pLWORDINIT, pLWORDINIT,
				                    pLWORDINIT, pLWORDINIT, pLWORDINIT, pLWORDINIT,
				                    pLWORDINIT, pLWORDINIT, pLWORDINIT, pLWORDINIT } ;
STATIC volatile LWORD *fpr[LASER_NUM_COUNTERS + 1]	= { pLWORDINIT,
				                    pLWORDINIT, pLWORDINIT, pLWORDINIT, pLWORDINIT,
				                    pLWORDINIT, pLWORDINIT, pLWORDINIT, pLWORDINIT,
				                    pLWORDINIT, pLWORDINIT, pLWORDINIT, pLWORDINIT } ;
STATIC volatile LWORD *cprLo[LASER_NUM_COUNTERS + 1]  = { pLWORDINIT,
				                    pLWORDINIT, pLWORDINIT, pLWORDINIT, pLWORDINIT,
				                    pLWORDINIT, pLWORDINIT, pLWORDINIT, pLWORDINIT,
				                    pLWORDINIT, pLWORDINIT, pLWORDINIT, pLWORDINIT } ;
STATIC volatile LWORD *fprLo[LASER_NUM_COUNTERS + 1]	= { pLWORDINIT,
				                    pLWORDINIT, pLWORDINIT, pLWORDINIT, pLWORDINIT,
				                    pLWORDINIT, pLWORDINIT, pLWORDINIT, pLWORDINIT,
				                    pLWORDINIT, pLWORDINIT, pLWORDINIT, pLWORDINIT } ;
STATIC volatile LWORD *command[LASER_NUM_COUNTERS + 1] = { pLWORDINIT,
				                    pLWORDINIT, pLWORDINIT, pLWORDINIT, pLWORDINIT,
			                            pLWORDINIT, pLWORDINIT, pLWORDINIT, pLWORDINIT,
				                    pLWORDINIT, pLWORDINIT, pLWORDINIT, pLWORDINIT } ;
STATIC volatile LWORD *mode[LASER_NUM_COUNTERS + 1]	= { pLWORDINIT,
				                    pLWORDINIT, pLWORDINIT, pLWORDINIT, pLWORDINIT,
				                    pLWORDINIT, pLWORDINIT, pLWORDINIT, pLWORDINIT,
			           	            pLWORDINIT, pLWORDINIT, pLWORDINIT, pLWORDINIT } ;
STATIC volatile LWORD *testSigStatusInt		= pLWORDINIT;
STATIC volatile LWORD *testSigStatusFracint	= pLWORDINIT;
STATIC volatile WORD *pUnknownStatus		= pWORDINIT;
STATIC volatile WORD *clockStatus		= pWORDINIT;
STATIC volatile LWORD *syncModeHi			= pLWORDINIT;
STATIC volatile LWORD *syncModeLo			= pLWORDINIT;
STATIC volatile WORD *fidWaitStates		= pWORDINIT;
STATIC volatile WORD *fidStates			= pWORDINIT;
STATIC volatile UWORD *testCntHi		= pUWORDINIT;
STATIC volatile UWORD *testCntLo		= pUWORDINIT;
STATIC volatile LWORD *testSigStatHi		= pLWORDINIT;
STATIC volatile LWORD *testSigStatLo		= pLWORDINIT;

int laserNumInstalled = 0;
int laserFailedTest = 0;		/* Used to report which counter failed test */
BOOL laserInitYet = FALSE;		/* TRUE once laserInit has been run */

/**************************************************
*
* laserInit  -  Initialize laser counter board software
*
* Initializes the software driver by setting up pointers to the various
* pieces of hardware on the laser board, but it does not reset or test the
* hardware.  This routine is useful if, for example, you have had to reboot
* your CPU but don't want to clear the laser counters when you initialize the
* board.  The
* argument numCounters is the number of counters on the board to initialize;
* you may specify a number smaller than the number actually installed if you
* wish.
* SEE ALSO:
* laserFullInit(2), laserCheckClocks(2), laserFastTestAll(2)
*
* RETURNS:
* OK if successful, or ERROR if there is some problem with the hardware or if
* you specify an invalid number of counters to initialize.
*/

STATUS laserInit(int numCounters)
/*
  int numCounters;	// The number of counters on the laser board
*/
{
  int i;
  STATUS retVal = OK;
	
#ifdef DEBUG
  /* Check for error */
  if (numCounters > LASER_NUM_COUNTERS || numCounters < 1) {
    errnoSet(S_laser_INIT_INVALID_ARGUMENT);
    return(ERROR);
  }
#endif

  /* Record number of counters in a global variable */
  laserNumInstalled = numCounters;
	
  /* Convert bus address to local address */
  if (sysBusToLocalAdrs ((int)LASER_AM, (char *)LASER_ADDRESS,
			 (char **) &lboardAddress) == ERROR)
    return (ERROR);
  /* Set up pointers */
  testModule   = lboardAddress + LASER_TEST_MODULE;
  statusModule = lboardAddress + LASER_STATUS_MODULE;
  for (i = 1; i<=laserNumInstalled; i++) {
    cpr[i]     = (LWORD *) (lboardAddress + LASER_COUNTER(i) + LASER_CPR_MSW);
    fpr[i]     = (LWORD *) (lboardAddress + LASER_COUNTER(i) + LASER_FPR_MSW);
    cprLo[i]   = (LWORD *) (lboardAddress + LASER_COUNTER(i) + LASER_CPR_LSW);
    fprLo[i]   = (LWORD *) (lboardAddress + LASER_COUNTER(i) + LASER_FPR_LSW);
    command[i] = (LWORD *) (lboardAddress + LASER_COUNTER(i) + LASER_COMMAND_WORD);
    mode[i]    = (LWORD *) (lboardAddress + LASER_COUNTER(i) + LASER_MODE_WORD);
  }
  testSigStatusInt = (LWORD *) (lboardAddress + LASER_TEST_SIGNAL_STATUS_INT);
  testSigStatusFracint = (LWORD *) (lboardAddress + LASER_TEST_SIGNAL_STATUS_FRACINT);
  pUnknownStatus = (WORD *) (lboardAddress + LASER_UNKNOWN_STATUS);
  clockStatus = (WORD *) (lboardAddress + LASER_CLOCK_STATUS);
  syncModeHi = (LWORD *) (lboardAddress + LASER_SYNC_MODE_HI);
  syncModeLo = (LWORD *) (lboardAddress + LASER_SYNC_MODE_LO);
  fidWaitStates = (WORD *) (lboardAddress + LASER_FID_WAIT_STATES);
  fidStates = (WORD *) (lboardAddress + LASER_FID_STATES);
  testCntHi = (UWORD *) (lboardAddress + LASER_TEST_CNT_HI);
  testCntLo = (UWORD *) (lboardAddress + LASER_TEST_CNT_LO);
  testSigStatHi = (LWORD *) (lboardAddress + LASER_TEST_SIG_STATUS_HI);
  testSigStatLo = (LWORD *) (lboardAddress + LASER_TEST_SIG_STATUS_LO);

  laserInitYet = TRUE;
  return (retVal);
}


/**************************************************
*
* laserFullInit  -  Initialize and test laser counter board hardware and software
*
* Initializes the hardware and performs some simple self-tests by calling
* laserCheckClocks and laserFastTestAll.  The
* argument numCounters is the number of counters on the board to initialize;
* you may specify a number smaller than the number actually installed if you
* wish.
* Modified 9/25/91 so that it no longer tests or clears the laser counters.
* SEE ALSO:
* laserInit(2), laserCheckClocks(2), laserFastTestAll(2)
*
* RETURNS:
* OK if successful, or ERROR if there is some problem with the hardware or if
* you specify an invalid number of counters to initialize.
*/

STATUS laserFullInit(int numCounters)
/*
	int numCounters;	// The number of counters on the laser board
*/
{
	int i;
	STATUS retVal = OK;
	
	if (laserInit(numCounters) == ERROR)
		return(ERROR);

	/* Check for presence of onboard clocks */
	if (laserCheckClocks() == ERROR /* || laserFastTestAll() == ERROR */)
		retVal = ERROR;

	/* Initialize counters to a reasonable state */
	for (i=1; i<=laserNumInstalled; i++) {
		/* Disable test mode (select normal operation) */
 		laserTestModeDisable(i);
#if 0
		  /* Removed: Clear the counter on the next reference edge */
		  laserSetSyncMode(i,LASER_CLEAR_ON_EXT);
		  laserClearCounter(i);
#endif
		/* Set the sync mode to external */
		laserSetSyncMode(i,LASER_MODE_CPR_SYNC_TO_EXT);
		/* Verify that the clocks are present */
/*		if (retVal == OK && laserUnknownStatus(i) == ERROR)
			retVal = ERROR;
*/
	}
	
	return (retVal);
}

/*****************************************************
*
* laserTestModeEnable - Turn on test mode for a counter
*
* Allows you to turn on test mode for each counter.
* In test mode, the counter will use a pair of internally
* generated test signals rather than the externally supplied
* reference and unknown.
* SEE ALSO:
* laserTestModeDisable(2)
*
* RETURNS:
* OK, or ERROR if counter is an invalid counter number.
*/

STATUS laserTestModeEnable(int counter)
/*
	int counter;       // Which counter's test mode to enable/disable 
*/
{
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	/* Check for error */
	if (counter<1 || counter>laserNumInstalled) {
		errnoSet(S_laser_INVALID_COUNTER);
		return(ERROR);
	}
#endif
    
	*command[counter] = LASER_CMD_ENTER_TEST_MODE;
	return(OK);
}

/*****************************************************
*
* laserTestModeDisable - Turn off test mode for a counter
*
* Allows you to turn off test mode for each counter.
* In test mode, the counter will use a pair of internally
* generated test signals rather than the externally supplied
* reference and unknown.
* SEE ALSO:
* laserTestModeEnable(2)
*
* RETURNS:
* OK, or ERROR if counter is an invalid counter number.
*/

STATUS laserTestModeDisable(int counter)
/*
	int counter;       // Which counter's test mode to enable/disable
*/
{
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	/* Check for error */
	if (counter<1 || counter>laserNumInstalled) {
		errnoSet(S_laser_INVALID_COUNTER);
		return(ERROR);
	}
#endif
    
	*command[counter] = LASER_CMD_LEAVE_TEST_MODE;
	return(OK);
}

/*****************************************************
*
* laserSetSyncMode - sets the register synchronization mode for a counter.
*
* In addition to setting the synchronization mode for the counter,
* laserSetSyncMode is also used to set the sync mode prior to
* issuing a clearCounter command in order to select the
* desired clear mode.
* The desired mode for the counter is specified by the argument
* newMode.  For sync mode selection, it must be one of LASER_MODE_CPR_FLOWTHROUGH,
* LASER_MODE_CPR_SYNC_TO_REF, LASER_MODE_CPR_SYNC_TO_EXT, or
* LASER_MODE_FPR_LOAD_ON_FIDUCIAL, or, for selecting a clear mode, it must be
* one of: LASER_CLEAR_IMMED, LASER_CLEAR_ON_REF, LASER_CLEAR_ON_EXT, or
* LASER_CLEAR_ON_FIDUCIAL. Note that the various LASER_MODE_...
* and LASER_CLEAR_... modes are
* actually the same.  For example, calling laserSetSyncMode with
* an argument of LASER_CLEAR_IMMED will prepare the counter to be
* cleared immediately on the next clearCounter call, but it
* will also cause the counter to switch to LASER_MODE_CPR_FLOWTHROUGH
* register synchronization mode.
* See the source code for laserInit for an example of how to use this procedure.
* SEE ALSO:
* laserClearCounter(2), laserLoadFprOnFiducial(2), laserClearCounterOnFiducial(2),
* laserFiducialWaitState(2), laserResetFiducialWait(2), laserSyncMode(2)
*
* RETURNS:
* OK, or ERROR if counter is an invalid counter number or if
* newMode is an invalid counter mode.
*/

STATUS laserSetSyncMode(int counter, int newMode)
/*
	int counter;        // Which counter
	int newMode;        // New synchronization mode
*/
{
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	/* Check for error */
	if (counter<1 || counter>laserNumInstalled) {
		errnoSet(S_laser_INVALID_COUNTER);
		return(ERROR);
	}
	
	if (newMode<LASER_MODE_CPR_FLOWTHROUGH || newMode>LASER_MODE_FPR_LOAD_ON_FIDUCIAL) {
		errnoSet(S_laser_INVALID_SYNC_MODE);
		return(ERROR);
	}
#endif
	
	*mode[counter] = (LWORD) newMode;
	return(OK);
}

/************************************************************
*
* laserClearCounter - clears (zeros) the selected counter.
*
* Clears the counter using the
* clear mode selected in the most recent laserSetSyncMode call.
* SEE ALSO:
* laserLoadFprOnFiducial(2), laserClearCounterOnFiducial(2), laserFiducialWaitState(2),
* laserResetFiducialWait(2), laserSetSyncMode(2), laserSyncMode(2)
*
* RETURNS:
* OK, or ERROR if counter is an invalid counter number.
*/

STATUS laserClearCounter(int counter)
/*
	int counter;		// Which counter to clear
*/
{
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	/* Check for error */
	if (counter<1 || counter>laserNumInstalled) {
		errnoSet(S_laser_INVALID_COUNTER);
		return(ERROR);
	}
#endif
	
	*command[counter] = LASER_CMD_CLEAR_COUNTER;
	return(OK);
}

/************************************************************
*
* laserCheckClocks - examines the Status of Reference and Clock Signals
*
* laserCheckClocks checks for the presence of valid clock signals, reference,
* external, and internal.  In order for the hardware (in its default configuration)
* to reliably detect the presence of these signals, the external sync signal
* must be at least 2kHz, and all others (reference, unknowns, and clocks) must be
* above 300 kHz.
* SEE ALSO:
* laserInit(2)
*
* RETURNS:
* OK if all the signals are present and healthy
* or ERROR if there is a problem.  If there is an error, detailed information about
* the problem can be ascertained using the errnoGet and printErrno functions.
*/

STATUS laserCheckClocks(void)
{
	LWORD curStatus;

#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
	if (((curStatus = *clockStatus) & LASER_HEALTHY_CLOCKS) == LASER_HEALTHY_CLOCKS)
						return(OK);

	if ( (LASER_HEALTHY_CLOCKS & LASER_HEALTHY_REF) &&
	     (curStatus & LASER_HEALTHY_REF) != LASER_HEALTHY_REF ) {
		errnoSet(S_laser_NO_REFERENCE);
		return(ERROR);
	}

	if ( (LASER_HEALTHY_CLOCKS & LASER_HEALTHY_PLL) &&
	     (curStatus & LASER_HEALTHY_PLL) != LASER_HEALTHY_PLL ) {
		errnoSet(S_laser_PLL_OUT_OF_LOCK);
		return(ERROR);
	}

	if ( (LASER_HEALTHY_CLOCKS & LASER_HEALTHY_64_MHZ) &&
	     (curStatus & LASER_HEALTHY_64_MHZ) != LASER_HEALTHY_64_MHZ ) {
		errnoSet(S_laser_NO_64_MHZ);
		return(ERROR);
	}

	if ( (LASER_HEALTHY_CLOCKS & LASER_HEALTHY_16_MHZ) &&
	     (curStatus & LASER_HEALTHY_16_MHZ) != LASER_HEALTHY_16_MHZ ) {
		errnoSet(S_laser_NO_16_MHZ);
		return(ERROR);
	}

	if ( (LASER_HEALTHY_CLOCKS & LASER_HEALTHY_2_MHZ) &&
	     (curStatus & LASER_HEALTHY_2_MHZ) != LASER_HEALTHY_2_MHZ ) {
		errnoSet(S_laser_NO_2_MHZ);
		return(ERROR);
	}

#ifndef LASER_NO_EXT_SYNC
	if ( (LASER_HEALTHY_CLOCKS & LASER_HEALTHY_EXT) &&
	     (curStatus & LASER_HEALTHY_EXT) != LASER_HEALTHY_EXT ) {
		errnoSet(S_laser_NO_EXTERNAL_CLOCK);
		return(ERROR);
	}
#endif

	errnoSet(S_laser_UNKNOWN_CLOCK_ERROR);
	return(ERROR);
}

/*---------------------------------------------------------------
 *  The following  are a series of routines for reading the laser
 *  counters.  The one you use depends on your requirements.
 *  The first, laserCpr, is the most general, and the others are
 *  more specialized.  For efficiency, many of these could be
 *  written as preprocessor macros rather than subroutines, but
 *  they're presented as subroutines here for clarity.
*/


/****************************************************************
* laserRead64Unlocked - performs a 64-bit read
*
* Reads two longwords from the address specified by fromAddress, copying the
* results to the address specified by p64bitNum.  This routine does not
* implement bus locking and, in the general case, you should replace it by
* one that does.  See laserConfig.h and "Bus Locking" in the User's Guide
* for more information.
*
* SEE ALSO:
* laserConfig.h, Laser Counter Board User's Guide
*
* RETURNS:
* Has no return value, but places the contents found at fromAddress into
* the memory location specified by p64bitNum.
*/

void laserRead64Unlocked(volatile LWORD *fromAddress, LWORD *p64bitNum)
/*
	LWORD *fromAddress;		// Address to read from (usually on VMEbus)
	LWORD *p64bitNum;		// Address to copy data to
*/
{
	*(p64bitNum+1) = *(fromAddress+1);
	*p64bitNum = *fromAddress;
        /* Added this line to eliminate glitches around 0 */
	*(p64bitNum+1) = *(fromAddress+1);
}


/****************************************************************
* laserCprAddr - returns a pointer to the laser board's CPR
*
* Returns the address of the CPR in case you need to access it
* yourself (don't do this unless you need absolute speed).
* SEE ALSO:
* laserFprAddr(2)
*
* RETURNS:
* An LWORD * to the counter's CPR, or NULL if there's an error.
*/

volatile LWORD *laserCprAddr(int counter)
/*
  int counter;             // which counter's address to return
*/
{
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return NULL;
	}
#endif
#if (DEBUG>=20)
	/* Check for error */
	if (counter<1 || counter>laserNumInstalled) {
		errnoSet(S_laser_INVALID_COUNTER);
		return NULL;
	}
#endif

	return cpr[counter];
}


/****************************************************************
* laserFprAddr - returns a pointer to the laser board's FPR
*
* Returns the address of the FPR in case you need to access it
* yourself (don't do this unless you need absolute speed).
* SEE ALSO:
* laserCprAddr(2)
*
* RETURNS:
* An LWORD * to the counter's FPR, or NULL if there's an error.
*/

volatile LWORD *laserFprAddr(int counter)
/*
  int counter;             // which counter's address to return
*/
{
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return NULL;
	}
#endif
#if (DEBUG>=20)
	/* Check for error */
	if (counter<1 || counter>laserNumInstalled) {
		errnoSet(S_laser_INVALID_COUNTER);
		return NULL;
	}
#endif

	return fpr[counter];
}


/****************************************************************
* laserCpr - reads all 40 bits of the laser counter
*
* Reads all 40 bits of a laser counter, returning the
* results in one 8-byte quadword.
* SEE ALSO:
* laserCprHi(2), laserCprLo(2), laserCprInt(2), laserCprFrac(2)
*
* RETURNS:
* OK, or ERROR if the counter number is invalid.
* *contents is a pointer to a quadword.  This will contain
* the bytes SSSI IIIF on return, MSB first, where IIII is the 32
* bits of the integer count and the F is the 8 fraction bits, and
* SSS are just sign bits (either all 1's or all 0's).  That is, the
* result is sign-extended from 40 bits to 64 bits.
*/

STATUS laserCpr(int counter, QWORD *contents)
/*
	int counter;		// Which counter to read
	QWORD *contents;	// 64 bits of counter
*/
{
	FAST volatile LWORD *cprAddr;
	
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
	if (contents == NULL) {
		errnoSet(S_laser_NULL_POINTER_ARGUMENT);
		return(ERROR);
	}
#endif
#if (DEBUG>=20)
	/* Check for error */
	if (counter<1 || counter>laserNumInstalled) {
		errnoSet(S_laser_INVALID_COUNTER);
		return(ERROR);
	}
#endif

	cprAddr = cpr[counter];
	LASER_READ64LOCKED(cprAddr, (LWORD *) contents);
	/* Sign extend */
	if (contents->high & 0x00000080)
	   contents->high |= 0xffffff00;
	else
	   contents->high &= 0x000000ff;
	  
	return(OK);
}

/****************************************************************
*
* laserCprHi - returns the high-order longword of the laser counter
*
* Returns the 8 most significant bits of the integer count.  Note that this
* routine does NOT latch a new value.  That is, the value returned for these 8
* high bits is whatever the value of the CPR was the last time a laserCprLo
* read was performed.  This is a designed-in feature.  Since the full 40-bit
* CPR cannot be read in one VME cycle, there must be a method for "freezing"
* the outputs until you've read all 40 bits; the hardware therefore freezes
* (latches) the value of the 40-bit internal bus whenever the 8 least significant
* bits are accessed (e.g., via laserCprLo).
* In practice, you should probably not call this routine directly but should
* use laserCprInt or laserCpr if you need to see these bits.
* SEE ALSO:
* laserCpr(2), laserCprLo(2), laserCprInt(2), laserCprFrac(2)
*
* RETURNS:
* OK, or ERROR if counter is an invalid counter number.
* On return, *high contains the high-order 8 bits of the integer count.
*/

STATUS laserCprHi(int counter, LWORD *high)
/*
	int counter;	// Which counter to read 
	LWORD *high;		// High-order 8 bits of counter (returned)
*/
{
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
	if (high == NULL) {
		errnoSet(S_laser_NULL_POINTER_ARGUMENT);
		return(ERROR);
	}
#endif
#if (DEBUG>=20)
	/* Check for error */
	if (counter<1 || counter>laserNumInstalled) {
		errnoSet(S_laser_INVALID_COUNTER);
		return(ERROR);
	}
#endif

	*high = (*cpr[counter]) & 0x000000ff;
	if (*high & 0x00000080)
	   *high |= 0xffffff00;
	return(OK);
}

/****************************************************************
*
* laserCprLo - returns the low-order longword of the laser counter
*
* This is the complement of laserCprHi, as it returns the low-order
* bits of the laser counter, in other words, the 24
* least significant integer bits and the 8 fraction bits.
* SEE ALSO:
* laserCpr(2), laserCprHi(2), laserCprInt(2), laserCprFrac(2)
*
* RETURNS:
* OK, or ERROR if counter is an invalid counter number.
* On return, *low contains the low-order 32 bits of the laser counter.
*/

STATUS laserCprLo(int counter,LWORD *low)
/*
	int counter;	// Which counter to read
	LWORD *low;		// Low-order 32 bits of counter (returned)
*/
{
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
	if (low == NULL) {
		errnoSet(S_laser_NULL_POINTER_ARGUMENT);
		return(ERROR);
	}
#endif
#if (DEBUG>=20)
	/* Check for error */
	if (counter<1 || counter>laserNumInstalled) {
		errnoSet(S_laser_INVALID_COUNTER);
		return(ERROR);
	}
#endif

	*low = *cprLo[counter];
	return(OK);
}

/****************************************************************
*
* laserCprInt -  returns the 32 integer bits from the laser counter
*
* laserCprInt picks off the 32 integer bits of the laser counter and
* returns them.  It is a higher-level routine than any of the
* previous three.
* SEE ALSO:
* laserCpr(2), laserCprHi(2), laserCprLo(2), laserCprFrac(2)
*
* RETURNS:
* OK, or ERROR if counter is an invalid counter number.
* On return, *ival contains the 32 bit integer value of the laser counter.
*/

STATUS laserCprInt(int counter, LWORD *ival)
/*
	int counter;	// which counter
	LWORD *ival;	// integer value of counter (returned)
*/
{
	QWORD contents;
			
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
	if (ival == NULL) {
		errnoSet(S_laser_NULL_POINTER_ARGUMENT);
		return(ERROR);
	}
#endif
	if (laserCpr(counter,&contents) == ERROR)
		return(ERROR);
	*ival = (contents.high << 24) | ((contents.low >> 8) & 0x00ffffff);
	return(OK);
}

/****************************************************************
*
* laserCprFrac - returns the fractional count of a laser counter
*
* laserCprFrac picks off just the 8 fraction bits from the laser
* counter and returns them.  Be careful - sign errors are lurking
* here.  The full 40-bit laser counter is a two's complement
* quantity, and you're picking off the least significant 8 bits.
* That's fine if you're just monitoring those eight, but if you
* choose to combine them somehow with other bits from the counter,
* be sure to use the correct technique.
* SEE ALSO:
* laserCpr(2), laserCprHi(2), laserCprLo(2), laserCprInt(2)
* 
* RETURNS: 
* OK, or ERROR if the counter number is invalid.  On return,
* *fval will contain a 32-bit longword with four bytes in the form 000F,
* where F is the 8 fractional bits.
*/

STATUS laserCprFrac(int counter,LWORD *fval)
/*
	int counter;	// Which counter
	LWORD *fval;	// Fractional value (returned)
*/
{
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
	if (fval == NULL) {
		errnoSet(S_laser_NULL_POINTER_ARGUMENT);
		return(ERROR);
	}
#endif
#if (DEBUG>=20)
	/* Check for error */
	if (counter<1 || counter>laserNumInstalled) {
		errnoSet(S_laser_INVALID_COUNTER);
		return(ERROR);
	}
#endif
	
	*fval = *cprLo[counter] & 0x000000ff;
	return(OK);
}

/*-----------------------------------------------------------------
 * The following routines are the exact analogues to the previous
 * routines for reading the CPR, but these are for reading the FPR.
 * See the analogous CPR routines above for a functional
 * description.
 */

/****************************************************************
* laserFpr - reads all 40 bits of the laser counter's FPR
*
* Reads all 40 bits of a laser counter's FPR, returning the
* results in one 8-byte quadword.
* SEE ALSO:
* laserFprHi(2), laserFprLo(2), laserFprInt(2), laserFprFrac(2)
*
* RETURNS:
* OK, or ERROR if the counter number is invalid.
* *contents is a pointer to a quadword.  This will contain
* the bytes 000I IIIF on return, MSB first, where IIII is the 32
* bits of the integer count and the F is the 8 fraction bits.
* If a sign-extended result is desired, use
* laserFprInt or left-shift by 24 and divide by 2^24.
*/

STATUS laserFpr(int counter, QWORD *contents)
/*
	int counter;		// Which counter
	QWORD *contents;		// 32 bits of FPR
*/
{
	FAST volatile LWORD *fprAddr;
	
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
	if (contents == NULL) {
		errnoSet(S_laser_NULL_POINTER_ARGUMENT);
		return(ERROR);
	}
#endif
#if (DEBUG>=20)
	/* Check for error */
	if (counter<1 || counter>laserNumInstalled) {
		errnoSet(S_laser_INVALID_COUNTER);
		return(ERROR);
	}
#endif
	
	fprAddr = fpr[counter];
	LASER_READ64LOCKED(fprAddr, (LWORD *) contents);
	contents->high &= 0x000000ff;
	return(OK);
}

/****************************************************************
*
* laserFprHi - returns the high-order longword of the FPR
*
* Returns the 8 most significant bits of the integer part of the FPR.  Note that this
* routine does NOT latch a new value.  That is, the value returned for these 8
* high bits is whatever the value of the FPR was the last time a laserFprLo
* read was performed.  This is a designed-in feature.  Since the full 40-bit
* FPR cannot be read in one VME cycle, there must be a method for "freezing"
* the outputs until you've read all 40 bits; the hardware therefore freezes
* (latches) the value of the internal 40-bit bus whenever the 8 least significant
* bits are accessed (e.g., via laserFprLo).
* In practice, you should probably not call this routine directly but should
* use laserFprInt or laserFpr if you need to see these bits.
* SEE ALSO:
* laserFpr(2), laserFprLo(2), laserFprInt(2), laserFprFrac(2)
*
* RETURNS:
* OK, or ERROR if counter is an invalid counter number.
* On return, *high contains the high-order 8 bits of the integer count.
*/

STATUS laserFprHi(int counter, LWORD *high)
/*
	int counter;	// Which counter
	LWORD *high;		// High-order 8 bits of FPR (returned)
*/
{
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
	if (high == NULL) {
		errnoSet(S_laser_NULL_POINTER_ARGUMENT);
		return(ERROR);
	}
#endif
#if (DEBUG>=20)
	/* Check for error */
	if (counter<1 || counter>laserNumInstalled) {
		errnoSet(S_laser_INVALID_COUNTER);
		return(ERROR);
	}
#endif
	
	*high = (*fpr[counter]) & 0x000000ff;
	return(OK);
}

/****************************************************************
*
* laserFprLo - returns the low-order longword of the FPR
*
* This is the complement of laserFprHi, as it returns the low-order
* bits of the FPR, in other words, the 24
* least significant integer bits and the 8 fraction bits.
* SEE ALSO:
* laserFpr(2), laserFprHi(2), laserFprInt(2), laserFprFrac(2)
*
* RETURNS:
* OK, or ERROR if counter is an invalid counter number.
* On return, *low contains the low-order 32 bits of the FPR.
*/

STATUS laserFprLo(int counter, LWORD *low)
/*
	int counter;	// Which counter
	LWORD *low;		// Low-order 32 bits of counter (returned)
*/
{
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
	if (low == NULL) {
		errnoSet(S_laser_NULL_POINTER_ARGUMENT);
		return(ERROR);
	}
#endif
#if (DEBUG>=20)
	/* Check for error */
	if (counter<1 || counter>laserNumInstalled) {
		errnoSet(S_laser_INVALID_COUNTER);
		return(ERROR);
	}
#endif
	
	*low = *fprLo[counter];
	return(OK);
}

/****************************************************************
*
* laserFprInt -  returns the 32 integer bits from the FPR
*
* laserFprInt picks off the 32 integer bits of the FPR and
* returns them.  It is a higher-level routine than any of the
* previous three.
* SEE ALSO:
* laserFpr(2), laserFprHi(2), laserFprLo(2), laserFprFrac(2)
*
* RETURNS:
* OK, or ERROR if counter is an invalid counter number.
* On return, *ival contains the 32 bit integer value of the FPR.
*/

STATUS laserFprInt(int counter, LWORD *ival)
/*
	int counter;	// which counter
	LWORD *ival;	// integer value of FPR (returned)
*/
{
	QWORD contents;
	
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
	if (ival == NULL) {
		errnoSet(S_laser_NULL_POINTER_ARGUMENT);
		return(ERROR);
	}
#endif
	if (laserFpr(counter,&contents) == ERROR)
		return(ERROR);
	*ival = (contents.high << 24) | ((contents.low >> 8) & 0x00ffffff);
	return(OK);
}

/****************************************************************
*
* laserFprFrac - returns the fractional count of the FPR
*
* laserFprFrac picks off just the 8 fraction bits from the FPR
* and returns them.  Be careful - sign errors are lurking
* here.  The full 40-bit FPR is a two's complement
* quantity, and you're picking off the least significant 8 bits.
* That's fine if you're just monitoring those eight, but if you
* choose to combine them somehow with other bits from the counter,
* be sure to use the correct technique.
* SEE ALSO:
* laserFpr(2), laserFprHi(2), laserFprLo(2), laserFprInt(2)
* 
* RETURNS: 
* OK, or ERROR if the counter number is invalid.  On return,
* *fval will contain a 32-bit longword with four bytes in the form 000F,
* where F is the 8 fractional bits.
*/

STATUS laserFprFrac(int counter,LWORD *fval)
/*
	int counter;	// Which counter
	LWORD *fval;		// Fractional value (returned)
*/
{
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
	if (fval == NULL) {
		errnoSet(S_laser_NULL_POINTER_ARGUMENT);
		return(ERROR);
	}
#endif
#if (DEBUG>=20)
	/* Check for error */
	if (counter<1 || counter>laserNumInstalled) {
		errnoSet(S_laser_INVALID_COUNTER);
		return(ERROR);
	}
#endif

	*fval = *fprLo[counter] & 0x000000ff;
	return(OK);
}

/*****************************************************************
*
* laserLoadFprOnFiducial - copies CPR to FPR on next fiducial crossing
*
* This routine simply commands the counter to load the value
* of the CPR into the FPR the next time the fiducial is crossed
* (effectively storing the fiducial position in the FPR).  Note that the
* synchronization mode of the CPR is unaffected by this routine.
* If you desire to also clear the CPR on the next fiducial crossing, use
* the laserClearCounterOnFiducial procedure.
* SEE ALSO:
* laserClearCounter(2), laserClearCounterOnFiducial(2), laserFiducialWaitState(2),
* laserResetFiducialWait(2), laserSetSyncMode(2), laserSyncMode(2)
*
* RETURNS:
* OK, or ERROR if given an invalid counter number.
*/

STATUS laserLoadFprOnFiducial(int counter)
/*
	int counter;	// Which counter
*/
{
	return (laserSetSyncMode(counter,LASER_MODE_FPR_LOAD_ON_FIDUCIAL));
}

/*****************************************************************
*
* laserClearCounterOnFiducial - causes counter to be zeroed on next fiducial crossing
*
* This routine sets up the counter so that the next fiducial
* crossing will cause the CPR to be loaded into the FPR and then the
* counter will be cleared.  Thus, after the next fiducial crossing, the
* FPR contains the position of the fiducial, and the CPR will contain
* zero.  Note that the CPR synchronization mode will continue to be
* the same as it was before this call.  The only difference is that,
* when the fiducial is crossed, the CPR will suddenly jump to zero and
* then continue as before.
* SEE ALSO:
* laserClearCounter(2), laserLoadFprOnFiducial(2), laserFiducialWaitState(2),
* laserResetFiducialWait(2), laserSetSyncMode(2), laserSyncMode(2)
*
* RETURNS:
* OK, or ERROR if given an invalid counter number.
*/

STATUS laserClearCounterOnFiducial(int counter)
/*
	int counter;	// Which counter
*/
{	
	if (laserLoadFprOnFiducial(counter) == ERROR || laserClearCounter(counter) == ERROR)
		return(ERROR);
	return(OK);
}

/******************************************************************
*
* laserResetFiducialWait - clears any pending wait-for-fiducial operations
*
* This routine will cancel the effect of any
* previous call to laserLoadFprOnFiducial or laserClearCounterOnFiducial (assuming
* the fiducial hasn't been crossed since that call).
* SEE ALSO:
* laserClearCounter(2), laserLoadFprOnFiducial(2), laserClearCounterOnFiducial(2),
* laserFiducialWaitState(2), laserSetSyncMode(2), laserSyncMode(2)
*
* RETURNS:
* OK, or ERROR if given an invalid counter number.
*/

STATUS laserResetFiducialWait(int counter)
/*
	int counter;	// Which counter
*/
{
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	/* Check for error */
	if (counter<1 || counter>laserNumInstalled) {
		errnoSet(S_laser_INVALID_COUNTER);
		return(ERROR);
	}
#endif
	
	*command[counter] = LASER_CMD_RESET_WAIT_STATES;
	return(OK);
}

/*****************************************************************
*
* laserTestSignalStatus - check progress of a simulated motion
*
* This is a function that allows you to check on the
* status of a simulated slew that was initiated with a
* laserSetTestCounts call.  The value returned is the number of counts 
* remaining in the slew in units of lambda/8.  Note the fraction
* part (the 3 least significant bits) may never actually reach
* zero but will remain at the value you programmed in
* laserSetTestCounts so the the phase of the test signals will remain
* as you have commanded.  Also note that due do pinout limitations of some
* of the programmable logic chips, not all the bits in the count are actually
* available for inspection.  This routine fills in the missing low-order bits
* with zeros, so that the number you read back is the right number rounded to
* the nearest available value (the fraction is always correct, though).
* This is a lower-level interface to the test module; simpler access to the
* self-test features is available via laserFastTestAll and its relatives.
* SEE ALSO:
* laserTestDone(2), laserSetTestCount(2), laserTestWait(2), laserFastTestAll(2)
*
* RETURNS:
* The number of integer counts remaining in the simulated motion (they decrease
* at a rate of 1,000,000 per second).
*/

int laserTestSignalStatus(void)
{
	WORD TstStatHi,TstStatLo;
	LWORD statHi,statMid,statLo;
	
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
	TstStatHi = (WORD) ( (*testSigStatHi) >> 16 );
	TstStatLo = (WORD) ( (*testSigStatLo) >> 16 );
	/* Pick off the FFF bits */
	statLo = ((LWORD) (TstStatLo & 0x0e00)) >> 9 ;
	/* Pick off the MMMM LLLL LLLL bits */
	statMid = ((LWORD) (TstStatHi & 0x0fff)) << 12;
	/* Pick off the xUUU UUUU bits */
	statHi = ((LWORD) (TstStatLo & 0x007f)) << 24;
	/* Now assemble it all */
	return( (int) (statHi | statMid | statLo) );
}

/******************************************************************
*
* laserTestDone - Checks whether the test module is busy performing a simulated motion
*
* This routine checks the status of the test module by returning the state
* of the Done pin on the test module.
* This is a lower-level interface to the test module; simpler access to the
* self-test features is available via laserFastTestAll and its relatives.
* SEE ALSO:
* laserTestSignalStatus(2), laserSetTestCount(2), laserTestWait(2), laserFastTestAll(2)
*
* RETURNS:
* FALSE if testmodule is still busy or TRUE if it's done, or ERROR if the
* laser board is not yet initialized.
*/

BOOL laserTestDone(void)
{
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
	if ( (*testSigStatLo & 0x00800000) == 0)
		return(FALSE);
	else
		return(TRUE);
}

/******************************************************************
*
* laserSetTestCount - programs the test module to do a simulated slew
*
* This routine allows you to program the test
* module to perform a simulated slew of a size specified by
* numCounts.  numCounts is expected to be an ordinary two's
* complement integer in units of lambda/8.
* This is a lower-level interface to the test module; simpler access to the
* self-test features is available via laserFastTestAll and its relatives.
* SEE ALSO:
* laserTestDone(2), laserTestSignalStatus(2), laserTestWait(2), laserFastTestAll(2)
*
* RETURNS:
* Nothing.
*/

void laserSetTestCount(int numCounts)
/*
	int numCounts;	// Distance to slew in units of lambda/8.
*/
{
	int oldFrac, newFrac;
	int i;
	
	/* if numCounts is positive, then everything is set to go,
	 * but negative numbers require a little manipulation due to
	 * the way the laser board interprets negative numbers.   */
	if (numCounts < 0)  numCounts = (-numCounts) | 0x80000000;
	
	/* First, let's be careful with the fraction...  If the fraction
	 * changes by more than one count, let's step it in units of one
	 * count.  This avoids spurious integer counts that can result from
	 * abrupt changes in the fraction. */
	oldFrac = (*testSigStatLo >> 25) & LASER_TEST_FRAC_BITS;
	newFrac = (numCounts & LASER_TEST_FRAC_BITS);
	if (oldFrac > newFrac) {
		for ( i=oldFrac; i>newFrac; i-- )
			*testCntLo = (UWORD) i;
	}
	else {
		for ( i=oldFrac; i<newFrac; i++ )
			*testCntLo = (UWORD) i;
	}
	
	/* The order of the writes here is important */
	*testCntLo = (UWORD) (numCounts & 0x0000ffff);
	*testCntHi = (UWORD) ((numCounts >> 16) & 0x0000ffff);
}

/*************************************************************
*
* laserSyncMode - returns current counter synchronization mode
*
* This function reports the current counter
* synchronization mode for the specified counter. 
* SEE ALSO:
* laserClearCounter(2), laserLoadFprOnFiducial(2), laserClearCounterOnFiducial(2),
* laserResetFiducialWait(2), laserSetSyncMode(2)
*
* RETURNS:
* One of LASER_MODE_CPR_FLOWTHROUGH, LASER_MODE_CPR_SYNC_TO_REF,
* or LASER_MODE_CPR_SYNC_TO_EXT, or ERROR if given an invalid counter number.
*/

int laserSyncMode(int counter)
/*
	int counter;	// Which counter
*/
{
	WORD statWord;

#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#if (DEBUG>=20)
	/* Check for error */
	if (counter<1 || counter>laserNumInstalled) {
		errnoSet(S_laser_INVALID_COUNTER);
		return(ERROR);
	}
#endif
	
	if (counter > LASER_HALF_COUNTERS) {
		statWord =  (WORD) ( (*syncModeHi) >> 16 );
		counter = counter - LASER_HALF_COUNTERS;
	}
	else
		statWord = (WORD) ( (*syncModeLo) >> 16 );

	/* Shift the result the appropriate number of bits and 
	 * pick off the two lsb's.   */
	return( 0x03 & (int) (statWord >> ( (counter-1) << 1)) );
}

/**************************************************************
*
* laserFiducialWaitState - check if counter is waiting on fiducial
*
* This routine returns TRUE if any part of
* the specified counter is waiting for a fiducial crossing (i.e.,
* if there is a pending Load-FPR-on-fiducial-crossing or Clear-
* CPR-on-fiducial-crossing condition).  You can use this function
* to monitor whether your commanded action has occurred yet (i.e.,
* whether the fiducial has been crossed since you issued the
* command).  If there is no pending wait-for-fiducial state, this
* function will return FALSE.
* SEE ALSO:
* laserSetSyncMode(2), laserClearCounter(2), laserLoadFprOnFiducial(2),
* laserClearCounterOnFiducial(2), laserResetFiducialWait(2), laserFiducialState(2)
*
* RETURNS:
* TRUE, or ERROR if given an invalid counter number.
*/

BOOL laserFiducialWaitState(int counter)
/*
	int counter;	// Which counter
*/
{
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#if (DEBUG>=20)
	/* Check for error */
	if (counter<1 || counter>laserNumInstalled) {
		errnoSet(S_laser_INVALID_COUNTER);
		return(ERROR);
	}
#endif

	if ( ( 0x01 & ( (*fidWaitStates) >> (counter-1) ) ) == 0 )
		return(FALSE);
	else 
	    return(TRUE);
}

/****************************************************************
*
* laserFiducialState - check whether fiducial indicator is on
*
* This routine returns TRUE if the specified
* counter's fiducial is currently tripped (in other
* words, if the cart is at the fiducial) and returns FALSE
* otherwise.
* SEE ALSO:
* laserFiducialWaitState(2)
*
* RETURNS:
* TRUE or FALSE, or ERROR if given an invalid counter number.
*/

BOOL laserFiducialState(int counter)
/*
	int counter;	// Which counter
*/
{
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#if (DEBUG>=20)
	/* Check for error */
	if (counter<1 || counter>laserNumInstalled) {
		errnoSet(S_laser_INVALID_COUNTER);
		return(ERROR);
	}
#endif
	
	if ( ( 0x01 & ( (*fidStates) >> (counter-1) ) ) == 0 )
		return(FALSE);
	else
	     return(TRUE);
}

/****************************************************************
*
* laserUnknownStatus - check if Unknown signal is present
*
* This routine returns OK if the specified
* counter's Unknown signal is alive and healthy, or returns
* ERROR otherwise.
* SEE ALSO:
* laserInit(2)
*
* RETURNS:
* OK if signal is present, or ERROR otherwise.  Also returns ERROR if counter
* number is invalid.
*/

STATUS laserUnknownStatus(int counter)
/*
	int counter;	// Which counter
*/
{
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#if (DEBUG>=20)
	/* Check for error */
	if (counter<1 || counter>laserNumInstalled) {
		errnoSet(S_laser_INVALID_COUNTER);
		return(ERROR);
	}
#endif
	
	if ( ( 0x01 & ( (*pUnknownStatus) >> (counter-1) ) ) == 0 ) {
		errnoSet(S_laser_UNKNOWN_NOT_PRESENT);
		return(ERROR);
	}
	else
		return(OK);
}

/*****************************************************************
*
* laserTestWait - Wait for a simulated motion to complete
*
* This routine waits until the test module is idle
* and ready to begin a new simulated motion.  It does this by
* polling the test signal status done bit.
* This is a lower-level interface to the test module; simpler access to the
* self-test features is available via laserFastTestAll and its relatives.
* SEE ALSO:
* laserTestDone(2), laserTestSignalStatus(2), laserSetTestCount(2), laserFastTestAll(2)
*
* RETURNS:
* Nothing.
*/

void laserTestWait(void)
{
	do ;
	while ( !laserTestDone() );
	/* Do one more read just to add a little delay here before returning */
	do ;
	while ( !laserTestDone() );
}

/*****************************************************************
*
* laserTest - test laser counters
*
* DESCRIPTION
* NO MANUAL
* This is a general-purpose testing routine that is used internally only.
* It provides a general interface that is called by the various testing
* functions.  Depending on the input arguments, it will test one or all of
* the installed laser counters, and it will perform a long, slow test or a
* short, fast test.  Whatever the test, each counter indicated is exercised
* with a slew up, a slew back to zero, a slew down, a clear counter call,
* and a slew through the fraction bits.
* SEE ALSO:
* laserFastTest(2), laserLongTest(2), laserFastTestAll(2), laserLongTestAll(2)
*
* RETURNS:
* OK if the counter(s) are working properly, ERROR otherwise.  Also returns
* ERROR if given invalid arguments or if some subroutine returns an error.
*/

STATIC STATUS laserTest(int counter, BOOL longTest, BOOL testAll)
/*
	int counter;	// Which counter to test
	BOOL longTest;  // TRUE if this is a long test
	BOOL testAll;	// TRUE if we are to test all the counters
*/
{
	int testMotion;
	LWORD intMotion, fracMotion, motion;
	LWORD originalPosition[LASER_NUM_COUNTERS+1];
	int counterToTest;
	
#if (DEBUG>=30)
	if (!laserInitYet) {
		errnoSet(S_laser_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	/* Check for error */
	if (counter<1 || counter>laserNumInstalled) {
		errnoSet(S_laser_INVALID_COUNTER);
		return(ERROR);
	}
	if ((longTest != TRUE && longTest != FALSE) || (testAll != TRUE && testAll != FALSE)) {
		errnoSet(S_laser_INVALID_BOOL_VALUE);
		return(ERROR);
	}
#endif
	
	/* Reset the test module (terminate any present slews) */
	laserSetTestCount(0);

	/* Set end of for loops if using testAll */
	if (testAll) counter = laserNumInstalled;
	
	/* Set up hardware for testing */
	for (counterToTest = testAll ? 1 : counter; counterToTest<=counter; counterToTest++) {
		/* Place the counter in test mode */
		if (laserTestModeEnable(counterToTest) == ERROR)
			return(ERROR);
		/* Clear the counter */
		if (laserSetSyncMode(counterToTest,LASER_CLEAR_ON_REF) == ERROR)
			return(ERROR);
		if (laserClearCounter(counterToTest) == ERROR)
			return(ERROR);
		/* Set the sync mode to sync on reference */
		if (laserSetSyncMode(counterToTest,LASER_MODE_CPR_SYNC_TO_REF) == ERROR)
			return(ERROR);
		/* Record the current reading for each counter */
		if (laserCprInt(counterToTest,&originalPosition[counterToTest]) == ERROR)
			return(ERROR);
	}
	
	/* Exercise integer part of counters */
	
	/* Begin a positive simulated motion */
	testMotion = (longTest ? LASER_LONG_TEST_POS : LASER_FAST_TEST_POS)<<3 &
						~LASER_TEST_FRAC_BITS;					
	laserSetTestCount(testMotion);
	/* Wait for the motion to complete */
	laserTestWait();
	
	/* Compare the value in the laser counter(s) */
	for (counterToTest = testAll ? 1 : counter; counterToTest<=counter; counterToTest++) {
		if (laserCprInt(counterToTest,&intMotion) == ERROR)
			return(ERROR);
		if ((int) (intMotion-originalPosition[counterToTest]) != testMotion>>3) {
			laserFailedTest = counterToTest;
			printf("Counter %d failed positive slew test: actual: %lx  original: %lx  test: %x\n",laserFailedTest, intMotion,originalPosition[counterToTest],testMotion);
			errnoSet(S_laser_COUNTER_FAILED_POS_SLEW_TEST);
			return(ERROR);
		}
	}
	
	/* Slew back to zero */
	testMotion = -testMotion;
	laserSetTestCount(testMotion);
	/* Wait for the motion to complete */
	laserTestWait();
	
	/* Compare the value in the laser counter(s) */
	for (counterToTest = testAll ? 1 : counter; counterToTest<=counter; counterToTest++) {
		if (laserCprInt(counterToTest,&intMotion) == ERROR)
			return(ERROR);
		if (intMotion != originalPosition[counterToTest]) {
			laserFailedTest = counterToTest;
			printf("Counter %d failed return from positive to zero test: actual: %lx  original: %lx  test: %x\n",laserFailedTest,intMotion,originalPosition[counterToTest],testMotion);
			errnoSet(S_laser_COUNTER_FAILED_RET_TO_ZERO_TEST);
			return(ERROR);
		}
	}

	/* Slew negative */
	testMotion = (longTest ? LASER_LONG_TEST_NEG : LASER_FAST_TEST_NEG)<<3 &
						~LASER_TEST_FRAC_BITS;
	laserSetTestCount(testMotion);
	/* Wait for the motion to complete */
	laserTestWait();
	
	/* Compare the value in the laser counter(s) */
	for (counterToTest = testAll ? 1 : counter; counterToTest<=counter; counterToTest++) {
		if (laserCprInt(counterToTest,&intMotion) == ERROR)
			return(ERROR);
		if ((int) (intMotion-originalPosition[counterToTest]) != testMotion>>3) {
			laserFailedTest = counterToTest;
			printf("Counter %d failed negative slew test: actual: %lx  original: %lx  test: %x\n",laserFailedTest,intMotion,originalPosition[counterToTest],testMotion);
			errnoSet(S_laser_COUNTER_FAILED_NEG_SLEW_TEST);
			return(ERROR);
		}
	}
	/* Slew back to zero */
	testMotion = -testMotion;
	laserSetTestCount(testMotion);
	/* Wait for the motion to complete */
	laserTestWait();
	
	/* Compare the value in the laser counter(s) */
	for (counterToTest = testAll ? 1 : counter; counterToTest<=counter; counterToTest++) {
		if (laserCprInt(counterToTest,&intMotion) == ERROR)
			return(ERROR);
		if (intMotion != originalPosition[counterToTest]) {
			laserFailedTest = counterToTest;
			printf("Counter %d failed return from negative to zero test: actual: %lx  original: %lx  test: %x\n",laserFailedTest,intMotion,originalPosition[counterToTest],testMotion);
			errnoSet(S_laser_COUNTER_FAILED_RET_TO_ZERO_TEST);
			return(ERROR);
		}
	}

	/* Slew a small amount negative */
	testMotion = LASER_FAST_TEST_NEG<<3 & ~LASER_TEST_FRAC_BITS;
	laserSetTestCount(testMotion);
	/* Wait for the motion to complete */
	laserTestWait();
	
	/* Compare the value in the laser counter(s) */
	for (counterToTest = testAll ? 1 : counter; counterToTest<=counter; counterToTest++) {
		if (laserCprInt(counterToTest,&intMotion) == ERROR)
			return(ERROR);
		if ((int) (intMotion-originalPosition[counterToTest]) != testMotion>>3) {
			laserFailedTest = counterToTest;
			printf("Counter %d failed negative slew test: actual: %lx  original: %lx  test: %x\n",laserFailedTest,intMotion,originalPosition[counterToTest],testMotion);
			errnoSet(S_laser_COUNTER_FAILED_NEG_SLEW_TEST);
			return(ERROR);
		}
	}
	
	/* Clear counter(s), first setting fraction to ensure we'll clear to 0 not -1 */
	laserSetTestCount(LASER_OFFSET_FRAC);
	for (counterToTest = testAll ? 1 : counter; counterToTest<=counter; counterToTest++)
		if (laserClearCounter(counterToTest) == ERROR)
			return(ERROR);

	/* Do a VMEbus read to give the counters time to latch new values */
	laserTestWait();

	/* Compare the value in the laser counter(s) */
	for (counterToTest = testAll ? 1 : counter; counterToTest<=counter; counterToTest++) {
		if (laserCprInt(counterToTest,&intMotion) == ERROR)
			return(ERROR);
		if (intMotion != 0) {
			laserFailedTest = counterToTest;
			printf("Counter %d failed clear test: actual: %lx\n",laserFailedTest,intMotion);
			errnoSet(S_laser_COUNTER_FAILED_CLEAR_TEST);
			return(ERROR);
		}
	}
	
	/* Fraction part */
	/* First record what the fraction reads when set to 0 */
	laserSetTestCount(0);
	/* Do a VMEbus read to give the counters time to latch new values */
	laserTestWait();

	for (counterToTest = testAll ? 1 : counter; counterToTest<=counter; counterToTest++)
		if (laserCprFrac(counterToTest,&originalPosition[counterToTest]) == ERROR)
			return(ERROR);

	/* Now check that the reading changes in step with the setting */	
	for (testMotion = 0; testMotion<=LASER_TEST_FRAC_BITS; testMotion++) {
		laserSetTestCount(testMotion);
		/* Wait for the motion to complete */
		laserTestWait();
		/* Compare the value in the laser counter(s) */
		for (counterToTest = testAll ? 1 : counter; counterToTest<=counter; counterToTest++)
			{
			if (laserCprFrac(counterToTest,&motion) == ERROR)
				return(ERROR);
			/* The 16 here accomplishes rounding */
			fracMotion = ((motion-originalPosition[counterToTest])+16)>>5
			               % (LASER_TEST_FRAC_BITS+1);
			if (fracMotion<0) fracMotion+=8;
			/* If needed, read it again in case it didn't have time to settle */
			if (fracMotion != testMotion) {
				if (laserCprFrac(counterToTest,&motion) == ERROR)
					return(ERROR);
				fracMotion = ((motion-originalPosition[counterToTest])+16)>>5
				               % (LASER_TEST_FRAC_BITS+1);
				if (fracMotion<0) fracMotion+=8;
			}
			if (fracMotion != testMotion) {
				laserFailedTest = counterToTest;
				printf("Counter %d failed fraction test: original: %lx  reading: %lx  test: %x\n",laserFailedTest, originalPosition[counterToTest],motion,testMotion);
				errnoSet(S_laser_COUNTER_FAILED_FRACTION_TEST);
				return(ERROR);
			}
		}
	}
	

	/* Restore counter(s) to normal */
	laserSetTestCount(0);
	for (counterToTest = testAll ? 1 : counter; counterToTest<=counter; counterToTest++) {
		if (laserClearCounter(counterToTest) == ERROR)
			return(ERROR);
		if (laserTestModeDisable(counterToTest) == ERROR)
			return(ERROR);
		if (laserSetSyncMode(counterToTest,LASER_MODE_CPR_SYNC_TO_EXT) == ERROR)
			return(ERROR);
	}

	return(OK);
}

/*****************************************************************
*
* laserFastTest - quick test of a laser counter
*
* This a routine that exercises the laser counter and
* verifies that it is functioning correctly.  This routine will
* return OK if the counter is functional or ERROR if there is some
* problem.  This routine invalidates any information previously
* stored in the counter.
* The fast test differs from the long test in
* that it simply performs shorter slews when exercising the integer
* counting module; all the same tests are still performed.
* SEE ALSO:
* laserLongTest(2), laserFastTestAll(2), laserLongTestAll(2)
*
* RETURNS:
* OK if counter is working, ERROR otherwise, or ERROR if counter number is
* invalid.
*/

STATUS laserFastTest(int counter)
/*
	int counter;	// Which counter to test
*/
{
	return(laserTest(counter,FALSE,FALSE));
}

/*****************************************************************
*
* laserLongTest - Thorough test of a laser counter
*
* This a routine that exercises the laser counter and
* verifies that it is functioning correctly.  This routine will
* return OK if the counter is functional or ERROR if there is some
* problem.  This routine invalidates any information previously
* stored in the counter.
* The fast test differs from the long test in
* that it simply performs shorter slews when exercising the integer
* counting module; all the same tests are still performed.
* SEE ALSO:
* laserFastTest(2), laserFastTestAll(2), laserLongTestAll(2)
*
* RETURNS:
* OK if counter is working, ERROR otherwise, or ERROR if counter number is
* invalid.
*/

STATUS laserLongTest(int counter)
/*
	int counter;	// Which counter to test
*/
{
	return(laserTest(counter,TRUE,FALSE));
}

/*****************************************************************
*
* laserFastTestAll - quick test of the entire laser counter board
*
* This routine exercises all the laser counters on the board and
* verifies that they function correctly.  This routine will
* return OK if all counters are functional or ERROR if there is some
* problem.  This routine invalidates any information previously
* stored in the counters.
* Note that the number of counters selected for testing is the number
* that was specified in the call to laserInit.
* The fast test differs from the long test in
* that it simply performs shorter slews when exercising the integer
* counting module; all the same tests are still performed.
* SEE ALSO:
* SEE ALSO:
* laserFastTest(2), laserLongTest(2), laserLongTestAll(2),
* laserInit(2)
*
* RETURNS:
* OK if counter is working, ERROR otherwise, or ERROR if counter number is
* invalid.
*/

STATUS laserFastTestAll(void)
{
	return(laserTest(laserNumInstalled,FALSE,TRUE));
}

/*****************************************************************
*
* laserLongTestAll - thorough test of the entire laser counter board
*
* This routine exercises all the laser counters on the board and
* verifies that they function correctly.  This routine will
* return OK if all counters are functional or ERROR if there is some
* problem.  This routine invalidates any information previously
* stored in the counters.
* Note that the number of counters selected for testing is the number
* that was specified in the call to laserInit.
* The fast test differs from the long test in
* that it simply performs shorter slews when exercising the integer
* counting module; all the same tests are still performed.
* SEE ALSO:
* laserFastTest(2), laserLongTest(2), laserFastTestAll(2), laserLongTestAll(2)
*
* RETURNS:
* OK if counter is working, ERROR otherwise, or ERROR if counter number is
* invalid.
*/

STATUS laserLongTestAll(void)
{
	return(laserTest(laserNumInstalled,TRUE,TRUE));
}

