/* pztMotorLib.c - routines to access the JPL pzt and motor control board.  */
/* sccs info :     @(#)pztMotorLib.c	1.5  6/7/94  */

/*
modification history
--------------------
01g,hs, 16Mar94     adding support for an output bit for the paddle
01f,hs, 21Feb94     added encoderStatusRead
01e,dlp,04Oct93     added pztPositionSet for pzt_D/A chip variation
01d,mmc,02feb92     added pztSetHigh, pztSetLow
01c,mmc,31aug91     include sysLib.h
01b,dlp,08Nov90     Changed LWORD types to WORD in pointer declarations
01a,dlp,05Oct90     written, debugged
*/

/*
DESCRIPTION
This module contains all the routines necessary to manipulate the pzt/motor controller board hardware.  
*/

#include <vxWorks.h>
#include <vme.h>

#ifdef VW_502
#include <vxProto.h>
#endif   /* VW_502 */

#include "rtTypes.h"
#include "sysLib.h"
#include "errnoLib.h"
#include "pztMotorLib.h"

/*---------------------------------------------------------------
* These are Pointers to various on-board registers initialized by
* pztMotorInit.
* "STATIC" is used simply to hide these declarations from the
* user; it may be removed by changing the STATIC define below.
*/

#define STATIC static
STATIC WORD pmboardTemp;  /* Initial destination of pointers -- used to keep
		pointers from pointing into never-never land when we start up */
STATIC volatile char *pmboardAddress		= (char *) &pmboardTemp ;
STATIC volatile WORD *pztVelocity[MAX_NUM_PZTS + 1] = { &pmboardTemp,
				&pmboardTemp, &pmboardTemp, &pmboardTemp,
				&pmboardTemp, &pmboardTemp, &pmboardTemp };
STATIC volatile WORD *pztClearCmd[MAX_NUM_PZTS + 1] = { &pmboardTemp, 
				&pmboardTemp, &pmboardTemp, &pmboardTemp, 
				&pmboardTemp, &pmboardTemp, &pmboardTemp };
STATIC volatile WORD *pztPosition[MAX_NUM_PZTS + 1] = { &pmboardTemp, 
				&pmboardTemp, &pmboardTemp, &pmboardTemp, 
				&pmboardTemp, &pmboardTemp, &pmboardTemp };
STATIC volatile WORD *pztStatus[MAX_NUM_PZTS + 1] = { &pmboardTemp, 
				&pmboardTemp, &pmboardTemp, &pmboardTemp, 
				&pmboardTemp, &pmboardTemp, &pmboardTemp };
STATIC volatile WORD *motorVelocity[MAX_NUM_MOTORS + 1] = { &pmboardTemp, 
				&pmboardTemp, &pmboardTemp, &pmboardTemp, 
				&pmboardTemp, &pmboardTemp, &pmboardTemp };
STATIC volatile WORD *motorOdPre[MAX_NUM_MOTORS + 1] = { &pmboardTemp, 
				&pmboardTemp, &pmboardTemp, &pmboardTemp, 
				&pmboardTemp, &pmboardTemp, &pmboardTemp };
STATIC volatile WORD *motorClearCmd[MAX_NUM_MOTORS + 1] = { &pmboardTemp, 
				&pmboardTemp, &pmboardTemp, &pmboardTemp, 
				&pmboardTemp, &pmboardTemp, &pmboardTemp };
STATIC volatile WORD *motorLatchCmd[MAX_NUM_MOTORS + 1] = { &pmboardTemp, 
				&pmboardTemp, &pmboardTemp, &pmboardTemp, 
				&pmboardTemp, &pmboardTemp, &pmboardTemp };
STATIC volatile WORD *motorPosHi[MAX_NUM_MOTORS + 1] = { &pmboardTemp, 
				&pmboardTemp, &pmboardTemp, &pmboardTemp, 
				&pmboardTemp, &pmboardTemp, &pmboardTemp };
STATIC volatile WORD *motorPosLo[MAX_NUM_MOTORS + 1] = { &pmboardTemp, 
				&pmboardTemp, &pmboardTemp, &pmboardTemp, 
				&pmboardTemp, &pmboardTemp, &pmboardTemp };
STATIC volatile WORD *motorStatus[MAX_NUM_MOTORS + 1] = { &pmboardTemp, 
				&pmboardTemp, &pmboardTemp, &pmboardTemp, 
				&pmboardTemp, &pmboardTemp, &pmboardTemp };

int pztNumInstalled = 0;
int motorNumInstalled = 0;
BOOL pztMotorInitYet = FALSE;	/* TRUE once pztMotorInit has been run */

/*******************************************************
*
* pztMotorInit - Initialize pzt/motor board software
*
*Initializes the software driver by setting up pointers to the various
*pieces of hardware on the pzt/motor board, but it does not reset the 
*hardware.  This routine is useful if, for example, you have had to reboot
*your CPU but don't want to clear the pzt and motor position counters when
*you initialize the board.  The argument numPzts is the number of pzt
*controllers on the board to initialize and the argument numMotors is
*the number of motor controllers and encoder channels on the board to
*initialize; you may specify a number smaller than the number actually 
*installed if you wish.
*If numPzts or numMotors = 0, then that part of the board is not 
*initialized.
*SEE ALSO:
*pztMotorFullInit(2), pztFullInit(2), motorFullInit(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid number of pzt
*controllers or motor controllers to initialize.
*/

STATUS pztMotorInit(int numPzts,int numMotors)
/*
	int numPzts;	// 	The number of pzt controllers on the board
	int numMotors;	// 	The number of motor controllers on the board
*/
{
	int i;
	STATUS retVal = OK;

#ifdef DEBUG
	/* Check for error */
	if ((numPzts > MAX_NUM_PZTS || numPzts < 0) ||
		(numMotors > MAX_NUM_MOTORS || numMotors < 0)){
		errnoSet(S_pztMotor_INIT_INVALID_ARGUMENT);
		return(ERROR);
	}
#endif

	/* Record number of pzts in a global variable */
	pztNumInstalled = numPzts;

	/* Record number of motors in a global variable */
	motorNumInstalled = numMotors;

	/* Convert bus address to local address */
	if (sysBusToLocalAdrs ((int)PZTMOTOR_AM,(char *)PZTMOTOR_ADDRESS,
			(char **) &pmboardAddress) == ERROR)
		return (ERROR);

	/* Set up pointers */
	for (i = 1; i<=pztNumInstalled; i++){
	pztVelocity[i] = (WORD *) (pmboardAddress + PZT(i) + VELOCITY);
	pztClearCmd[i] = (WORD *) (pmboardAddress + PZT(i) + CLR_CMD);
	pztPosition[i] = (WORD *) (pmboardAddress + PZT(i) + PR_LSW);
 	pztStatus[i] = (WORD *) (pmboardAddress + PZT(i) + STAT);
	}
	for (i = 1; i<= motorNumInstalled; i++){
	motorVelocity[i] = (WORD *) (pmboardAddress + MOTOR(i) + VELOCITY);
	motorOdPre[i] = (WORD *) (pmboardAddress + MOTOR(i) + OD_PRE);
	motorClearCmd[i] = (WORD *) (pmboardAddress + MOTOR(i) + CLR_CMD);
	motorLatchCmd[i] = (WORD *) (pmboardAddress + MOTOR(i) + LTCH_CMD);
	motorPosLo[i] = (WORD *) (pmboardAddress + MOTOR(i) + PR_LSW);
	motorPosHi[i] = (WORD *) (pmboardAddress + MOTOR(i) + PR_MSW);
	motorStatus[i] = (WORD *) (pmboardAddress + MOTOR(i) + STAT);
	}

	pztMotorInitYet = TRUE;
	return (retVal);
}	

/************************************************************
*
*pztMotorFullInit - Initialize pzt/motor board software and hardware
*
*Initializes the pzt and motor velocities and counters to zero.  Note
*that a pzt position of zero is mid-scale, or 5 volts. 
*The argument numPzts is the number of pzt
*controllers on the board to initialize and the argument numMotors is
*the number of motor controllers and encoder channels on the board to 
*initialize; you may specify a number smaller than the number actually 
*installed if you wish.
*If numPzts or numMotors = 0, then that part of the board is not 
*initialized.
*SEE ALSO:
*pztMotorInit(2), pztReset(2), motorReset(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid number of pzt
*controllers or motor controllers to initialize.
*/

STATUS pztMotorFullInit(int numPzts,int numMotors)
/*
	int numPzts;	// 	The number of pzt controllers on the board
	int numMotors;	// 	The number of motor controllers on the board
*/
{
	int i;
	STATUS retVal = OK;

	/* Initialize software */
	if (pztMotorInit(numPzts,numMotors) == ERROR)
		return(ERROR);

	/* Initialize pzt velocities and counters */
	for (i=1; i<=numPzts; i++){
		pztReset(i);
	}

	/* Initialize motor velocities and counters */
	/* This also initializes the encoder chips */
	for (i=1; i<=numMotors; i++){
		motorReset(i);

	}

	return (retVal);
}
			 		
/***********************************************************
*
*pztReset - Reset pzt velocity and counter
*
*Writes a velocity of zero to the pzt controller and resets the
*counter to zero (mid-scale or 5 volts).
*SEE ALSO:
*motorReset(2), pztMotorInit(2), pztMotorFullInit(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify and invalid pzt to reset.
*/

STATUS pztReset(int whichPzt)
/*
	int whichPzt;	//	Which pzt controller to reset
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichPzt < 1 || whichPzt > pztNumInstalled){
		errnoSet(S_pztMotor_INVALID_PZT);
		return(ERROR);
	}
#endif

	/* Initialize velocity */
	pztVelocitySet(whichPzt,ZERO_VELOCITY);
	/* Reset Counter */
	pztClearCounter(whichPzt);
	return(OK);
}


/***********************************************************
*
*pztSetHigh - set pzt to max output
*
*writes a positive velocity to the pzt controller to ramp it
*to the plus rail
*
*SEE ALSO:
*pztReset(2),pztSetLow(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify and invalid pzt to reset.
*/

STATUS pztSetHigh(int whichPzt)
/*
	int whichPzt;	//	Which pzt controller to reset
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichPzt < 1 || whichPzt > pztNumInstalled){
		errnoSet(S_pztMotor_INVALID_PZT);
		return(ERROR);
	}
#endif

	/* set velocity */
	pztVelocitySet(whichPzt,MAX_VELOCITY/10);
	return(OK);
}

/***********************************************************
*
*pztSetLow - set pzt to min output
*
*writes a negative velocity to the pzt controller to ramp it
*to zero
*
*SEE ALSO:
*pztReset(2),pztSetHigh(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify and invalid pzt to set.
*/

STATUS pztSetLow(int whichPzt)
/*
	int whichPzt;	//	Which pzt controller to reset
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichPzt < 1 || whichPzt > pztNumInstalled){
		errnoSet(S_pztMotor_INVALID_PZT);
		return(ERROR);
	}
#endif

	/* set velocity */
	pztVelocitySet(whichPzt,-MAX_VELOCITY/10);
	return(OK);
}


/*************************************************************
*
*motorReset - Reset motor velocity and counter
*
*Writes a velocity of zero to the motor controller and resets the
*counter to zero.
*SEE ALSO:
*pztReset(2), pztMotorInit(2), pztMotorFullInit(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify and invalid motor to reset.
*/

STATUS motorReset(int whichMotor)
/*
	int whichMotor;		//	Which motor controller to reset
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichMotor < 1 || whichMotor > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_MOTOR);
		return(ERROR);
	}
#endif

	/* Initialize velocity */
	/* if whichMotor corresponds to an encoder channel, this command */
	/* does nothing */
	motorVelocitySet(whichMotor,ZERO_VELOCITY);
	/* Reset Counter motor or encoder counter as appropriate */
	motorClearCounter(whichMotor);
	return(OK);
}

/*****************************************************************
*
*pztVelocitySet - Set pzt velocity
*
*Sets the velocity for a given pzt controller.  
*SEE ALSO:
*pztStop(2), motorVelocitySet(2), motorStop(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid pzt controller
*or an invalid velocity.
*/

STATUS pztVelocitySet(int whichPzt,int newVelocity)
/*
	int whichPzt;		//	Which pzt to send velocity to
	int newVelocity;	//	New velocity to send
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichPzt < 1 || whichPzt > pztNumInstalled){
		errnoSet(S_pztMotor_INVALID_PZT);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (newVelocity < MIN_VELOCITY || newVelocity > MAX_VELOCITY){
		errnoSet(S_pztMotor_INVALID_VELOCITY);
		return(ERROR);
	}
#endif
	/* Write new velocity */
	*pztVelocity[whichPzt] = newVelocity;
	return(OK);
}

/*****************************************************************
*
*pztPositionSet - Set pzt position
*
*Sets the position (D/A value) for a given pzt_d2a controller.  Note that
*this is a special command for use with the pzt_d2a version of the pzt
*chip.  It is equivalent to pztVelocitySet and in fact uses the same
*register.  So this command is just provided for clarity.  Also,
*note that the only usable commands for this version of the chip
*are pztPositionSet (or pztVelocitySet) and pztPositionRead.  To set
*the D/A output to 0 volts, call pztPositionSet(whichPzt, MIN_VELOCITY).  To
*set the D/A output to midrange, call pztPositionSet(whichPzt, ZERO_VELOCITY).
*At power-up, the output is set to 0 volts.  System reset has no effect
*on the output.
*SEE ALSO:
*pztVelocitySet(2), pztPositionRead(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid pzt controller
*or an invalid position.
*/

STATUS pztPositionSet(int whichPzt,int newPosition)
/*
	int whichPzt;		//	Which pzt to send position to
	int newPosition;	//	New velocity to send
*/
{
  /* Since *pztVelocity[] is the same register as *pztPosition[]
     would be, call pztVelocitySet to set the postion. */
  if(pztVelocitySet(whichPzt, newPosition)==ERROR){
    if(errnoGet()==S_pztMotor_INVALID_VELOCITY)
      errnoSet(S_pztMotor_INVALID_PZT_POSITION);
    return(ERROR);
  }
  return(OK);
}

/*****************************************************************
*
*motorVelocitySet - Set motor velocity
*
*Sets the velocity for a given motor controller.  
*SEE ALSO:
*motorStop(2), motorPrescaleSet(2), pztVelocitySet(2), pztStop(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid motor controller
*or an invalid velocity.
*/

STATUS motorVelocitySet(int whichMotor,int newVelocity)
/*
	int whichMotor;		//	Which pzt to send velocity to
	int newVelocity;	//	New velocity to send
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichMotor < 1 || whichMotor > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_MOTOR);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (newVelocity < MIN_VELOCITY || newVelocity > MAX_VELOCITY){
		errnoSet(S_pztMotor_INVALID_VELOCITY);
		return(ERROR);
	}
#endif
	/* Write new velocity */
	*motorVelocity[whichMotor] = newVelocity;
	return(OK);
}

/******************************************************************
*
*pztStop - Stops current pzt motion
*
*Writes a velicity of zero to the velocity register to stop the 
*current pzt motion.
*SEE ALSO:
*pztVelocitySet(2), motorStop(2), motorVelocitySet(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid pzt controller.
*/

STATUS pztStop(int whichPzt)
/*
	int whichPzt;		//	which pzt to stop 
*/
{
#if (DEBUG>=30)
	if (! pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichPzt < 1 || whichPzt > pztNumInstalled){
		errnoSet(S_pztMotor_INVALID_PZT);
		return(ERROR);
	}
#endif
	/* Write a velocity of zero */
	*pztVelocity[whichPzt] = ZERO_VELOCITY;
	return(OK);
}

/******************************************************************
*
*motorStop - Stops current motor motion
*
*Writes a velicity of zero to the velocity register to stop the 
*current motor motion.
*SEE ALSO:
*motorVelocitySet(2), pztStop(2), pztVelocitySet(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid motor controller.
*/

STATUS motorStop(int whichMotor)
/*
	int whichMotor;		//	which motor to stop 
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichMotor < 1 || whichMotor > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_MOTOR);
		return(ERROR);
	}
#endif
	/* Write a velocity of zero */
	*motorVelocity[whichMotor] = ZERO_VELOCITY;
	return(OK);
}

/***************************************************************
*
*pztClearCounter - Reset pzt counter
*
*Uses the hardware reset command to clear the motor counter to zero.
*SEE ALSO:
*motorClearCounter(2)
*
*RETURNS:
*OK if successful, or error if you specify an invalid pzt counter.
*/

STATUS pztClearCounter(int whichPzt)
/*
	int whichPzt;		//	which pzt to clear
*/
{
#if (DEBUG>=30)
	if (! pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichPzt < 1 || whichPzt > pztNumInstalled){
		errnoSet(S_pztMotor_INVALID_PZT);
		return(ERROR);
	}
#endif
	/* Write a dummy argument to the clear command address */
	*pztClearCmd[whichPzt] = DUMMY;
	return(OK);
}
	
/***************************************************************
*
*motorClearCounter - Reset motor counter
*
*Uses the hardware reset command to clear the motor counter to zero.
*SEE ALSO:
*pztClearCounter(2)
*
*RETURNS:
*OK if successful, or error if you specify an invalid motor counter.
*/

STATUS motorClearCounter(int whichMotor)
/*
	int whichMotor;		//	which motor to clear
*/
{
#if (DEBUG>=30)
	if (! pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichMotor < 1 || whichMotor > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_MOTOR);
		return(ERROR);
	}
#endif
	/* Write a dummy argument to the clear command address */
	*motorClearCmd[whichMotor] = DUMMY;
	return(OK);
}
	
/***************************************************************
*
*pztPositionRead - Reads the current position of the pzt.
*
*Reads the position from the counter of the selected pzt controller.
*SEE ALSO:
*motorPositonRead(2), motorPostionLatch(2), motorPositionLo(2)
*motorPostionHi(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid pzt controller.
*/

STATUS pztPositionRead(int whichPzt, int *position)
/*
	int whichPzt;		//	Which pzt counter to read
	int *position		//	Pzt position (returned)
*/
{
#if (DEBUG>=30)
	if (! pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichPzt < 1 || whichPzt > pztNumInstalled){
		errnoSet(S_pztMotor_INVALID_PZT);
		return(ERROR);
	}
#endif
	/* read counter to postion pointer */
	*position = *pztPosition[whichPzt];
	return(OK);
}

/***************************************************************
*
*motorPositionLatch - Latch current motor position
*
*Command to latch the motor position counter.  This is required
*since the board really has a D16 interface while the counter is
*32 bits.  Once latched, the hi and lo words may be read and then 
*combined to form one longword.
*SEE ALSO:
*pztPositionRead(2), motorPositionRead(2), motorPositionLo(2)
*motorPositionHi(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid motor controller.
*/

STATUS motorPositionLatch(int whichMotor)
/*
	int whichMotor;		//	Which motor counter to latch
*/

{
#if (DEBUG>=30)
	if (! pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichMotor < 1 || whichMotor > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_MOTOR);
		return(ERROR);
	}
#endif
	/*Write a dummy argument to the latch command register.*/
	*motorLatchCmd[whichMotor] = DUMMY;
	return(OK);
}

/*****************************************************************
*
*motorPositionHi - Read high word of motor position
*
*This routine reads the high order word of the motor position register.
*Note that motorPositionLatch must be run prior to this in order to 
*latch the most current counter value into the position register.
*SEE ALSO:
*motorPostionLatch(2), motorPositionLo(2), motorPositionRead(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid motor controller.
*/

STATUS motorPositionHi(int whichMotor, int *high)
/*
	int whichMotor;		//	Which motor position to read
	int *high;		//	High order position (returned)
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichMotor < 1 || whichMotor > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_MOTOR);
		return(ERROR);
	}
#endif
	*high = *motorPosHi[whichMotor] & 0x0000ffff;
	return(OK);
}


/*****************************************************************
*
*motorPositionLo - Read Low word of motor position
*
*This routine reads the low order word of the motor position register.
*This value should not be used by itself since it is the low order
*word of a 32 bit signed quantity with the sign in the high order word.
*Note that motorPositionLatch must be run prior to this in order to 
*latch the most current counter value into the position register.
*SEE ALSO:
*motorPostionLatch(2), motorPositionHi(2), motorPositionRead(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid motor controller.
*/

STATUS motorPositionLo(int whichMotor, int *low)
/*
	int whichMotor;		//	Which motor position to read
	int *low;		//	High order position (returned)
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichMotor < 1 || whichMotor > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_MOTOR);
		return(ERROR);
	}
#endif
	*low = *motorPosLo[whichMotor] & 0x0000ffff;
	return(OK);
}

/**************************************************************
*
*motorPositionRead - Read motor position
*
*Latches and reads all 32 bits of the motor position counter.
*SEE ALSO:
*motorPositionLatch(2), motorPositionHi(2), motorPositionLo(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid motor controller.
*/

STATUS motorPositionRead(int whichMotor, int *position)
/*
	int whichMotor;		//	Which motors position to read
	int *position;  	//	Current motor position (returned)
*/
{
        int high,low;

#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichMotor < 1 || whichMotor > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_MOTOR);
		return(ERROR);
	}
#endif
	/* Latch counter into position register */
	if (motorPositionLatch(whichMotor) == ERROR)
		return(ERROR);

	/* Read high order word */	
	if (motorPositionHi(whichMotor, &high) == ERROR)
		return(ERROR);
 
	/* Read low order word */
	if (motorPositionLo(whichMotor, &low) == ERROR)
		return(ERROR);

	/* combine high and low order words */
	*position = ((high << 16) | low);
	return(OK);
}
		
/***************************************************************
*
*pztStatusRead - Read clip status of pzt
*
*Reads the status port and tests bits for clipping.
*SEE ALSO:
*pztStatusClipUp(2), pztStatusClipDown(2)
*
*RETURNS:
*One of PZT_NOT_CLIPPING, PZT_CLIPPING_UP, or PZT_CLIPPING_DOWN,
*or ERROR if you specify an invalid pzt controller.
*/

STATUS pztStatusRead(int whichPzt)
/*
	int whichPzt;		//	Which pzt status to read
*/
{
	char clipStatus;

#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichPzt < 1 || whichPzt > pztNumInstalled){
		errnoSet(S_pztMotor_INVALID_PZT);
		return(ERROR);
	}
#endif
	clipStatus = *pztStatus[whichPzt] & 0x0007;
	return(clipStatus);
}

/*******************************************************************
*
*pztStatusClipUp - Test the clip-up status
*
*Reads the status word and tests the clip-up bit.
*SEE ALSO:
*pztStatusRead(2), pztStatusClipDown(2)
*
*RETURNS:
*TRUE or FALSE, or ERROR if you specify an invalid pzt controller.
*/

BOOL pztStatusClipUp(int whichPzt)
/*
	int whichPzt;		//	Which pzt clip-up status to read
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichPzt < 1 || whichPzt > pztNumInstalled){
		errnoSet(S_pztMotor_INVALID_PZT);
		return(ERROR);
	}
#endif
	if (pztStatusRead(whichPzt) == PZT_CLIPPING_UP)
		return(TRUE);
	else
		return(FALSE);
}

/*******************************************************************
*
*pztStatusClipDown - Test the clip-down status
*
*Reads the status word and tests the clip-down bit.
*SEE ALSO:
*pztStatusRead(2), pztStatusClipUp(2)
*
*RETURNS:
*TRUE or FALSE, or ERROR if you specify an invalid pzt controller.
*/

BOOL pztStatusClipDown(int whichPzt)
/*
	int whichPzt;		//	Which pzt clip-down status to read
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichPzt < 1 || whichPzt > pztNumInstalled){
		errnoSet(S_pztMotor_INVALID_PZT);
		return(ERROR);
	}
#endif
	if (pztStatusRead(whichPzt) == PZT_CLIPPING_DOWN)
		return(TRUE);
	else
		return(FALSE);
}

/*****************************************************************
*
*motorStatusAnyLimit - Check whether "any" limit switch is made.
*
*Reads the motor status word and checks whether or not any limit
*switch is made.
*SEE ALSO:
*motorStatusCwHardLimit(2), motorStatusCwSoftLimit(2), motorStatusCCwHardLimit(2),
*motorStatusCCwSoftLimit(2), motorStatusHome(2), motorStatusOverDrive(2)
*
*RETURNS:
*TRUE if the switch is closed, FALSE if it is open, or ERROR if you
*specify and invalid motor controller.
*/

BOOL motorStatusAnyLimit(int whichMotor)
/*
	int whichMotor;		//	which motor status to read
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichMotor < 1 || whichMotor > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_MOTOR);
		return(ERROR);
	}
#endif
	/* read motor status word and test any limit bit */
	if ((*motorStatus[whichMotor] & MOTOR_ANY_LIMIT) == MOTOR_ANY_LIMIT)
		return(TRUE);
	else
		return(FALSE);
}

/*****************************************************************
*
*motorStatusCwHardLimit - Check the Clockwise hard limit switch
*
*Reads the motor status word and checks whether or not the
*clockwise hard limit switch is made.
*SEE ALSO:
*motorStatusAnyLimit(2), motorStatusCwSoftLimit(2), motorStatusCCwHardLimit(2),
*motorStatusCCwSoftLimit(2), motorStatusHome(2), motorStatusOverDrive(2)
*
*RETURNS:
*TRUE if the switch is closed, FALSE if it is open, or ERROR if you
*specify and invalid motor controller.
*/

BOOL motorStatusCwHardLimit(int whichMotor)
/*
	int whichMotor;		//	which motor status to read
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichMotor < 1 || whichMotor > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_MOTOR);
		return(ERROR);
	}
#endif
	/* read motor status word and test CW hard limit bit */
	if ((*motorStatus[whichMotor] & MOTOR_CW_HARD_LIMIT) == MOTOR_CW_HARD_LIMIT)
		return(TRUE);
	else
		return(FALSE);
}

/*****************************************************************
*
*motorStatusCwSoftLimit - Check the Clockwise soft limit switch.
*
*Reads the motor status word and checks whether the
*clockwise soft limit switch is made.
*SEE ALSO:
*motorStatusAnyLimit(2), motorStatusCwHardLimit(2), motorStatusCCwHardLimit(2), 
*motorStatusCCWSoftLimit(2), motorStatusHome(2), motorStatusOverDrive(2)
*
*RETURNS:
*TRUE if the switch is closed, FALSE if it is open, or ERROR if you
*specify and invalid motor controller.
*/

BOOL motorStatusCwSoftLimit(int whichMotor)
/*
	int whichMotor;		//	which motor status to read
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichMotor < 1 || whichMotor > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_MOTOR);
		return(ERROR);
	}
#endif
	/* read motor status word and test CW soft limit bit */
	if ((*motorStatus[whichMotor] & MOTOR_CW_SOFT_LIMIT) == MOTOR_CW_SOFT_LIMIT)
		return(TRUE);
	else
		return(FALSE);
}

/*****************************************************************
*
*motorStatusCCwHardLimit - Check the Counter-Clockwise hard limit switch
*
*Reads the motor status word and checks whether or not the
*counter-clockwise hard limit switch is made.
*SEE ALSO:
*motorStatusAnyLimit(2), motorStatusCwHardLimit(2), motorStatusCwSoftLimit(2),
*motorStatusCCwSoftLimit(2), motorStatusHome(2), motorStatusOverDrive(2)
*
*RETURNS:
*TRUE if the switch is closed, FALSE if it is open, or ERROR if you
*specify and invalid motor controller.
*/

BOOL motorStatusCCwHardLimit(int whichMotor)
/*
	int whichMotor;		//	which motor status to read
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichMotor < 1 || whichMotor > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_MOTOR);
		return(ERROR);
	}
#endif
	/* read motor status word and test CCW hard limit bit */
	if ((*motorStatus[whichMotor] & MOTOR_CCW_HARD_LIMIT) == MOTOR_CCW_HARD_LIMIT)
		return(TRUE);
	else
		return(FALSE);
}

/*****************************************************************
*
*motorStatusCCwSoftLimit - Check the Counter-Clockwise soft limit switch
*
*Reads the motor status word and checks whether or not the
*counter-clockwise soft limit switch is made.
*SEE ALSO:
*motorStatusAnyLimit(2), motorStatusCwHardLimit(2), motorStatusCwSoftLimit(2),
*motorStatusCCwHardLimit(2), motorStatusHome(2), motorStatusOverDrive(2)
*
*RETURNS:
*TRUE if the switch is closed, FALSE if it is open, or ERROR if you
*specify and invalid motor controller.
*/

BOOL motorStatusCCwSoftLimit(int whichMotor)
/*
	int whichMotor;		//	which motor status to read
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichMotor < 1 || whichMotor > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_MOTOR);
		return(ERROR);
	}
#endif
	/* read motor status word and test CCW soft limit bit */
	if ((*motorStatus[whichMotor] & MOTOR_CCW_SOFT_LIMIT) == MOTOR_CCW_SOFT_LIMIT)
		return(TRUE);
	else
		return(FALSE);
}

/*****************************************************************
*
*motorStatusHome - Check the home switch
*
*Reads the motor status word and checks whether or not the
*home switch is made.
*SEE ALSO:
*motorStatusAnyLimit(2), motorStatusCwHardLimit(2), motorStatusCwSoftLimit(2),
*motorStatusCCwHardLimit(2), motorStatusCCwSoftLimit(2), motorStatusOverDrive(2)
*
*RETURNS:
*TRUE if the switch is closed, FALSE if it is open, or ERROR if you
*specify and invalid motor controller.
*/

BOOL motorStatusHome(int whichMotor)
/*
	int whichMotor;		//	which motor status to read
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichMotor < 1 || whichMotor > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_MOTOR);
		return(ERROR);
	}
#endif
	/* read motor status word and test the home bit */
	if ((*motorStatus[whichMotor] & MOTOR_HOME) == MOTOR_HOME)
		return(TRUE);
	else
		return(FALSE);
}

/*****************************************************************
*
*motorStatusOverDrive - Check the over-drive status.
*
*Reads the motor status word and checks whether or not the
*over-drive bit is set.
*SEE ALSO:
*motorStatusAnyLimit(2), motorStatusCwHardLimit(2), motorStatusCwSoftLimit(2),
*motorStatusCCwHardLimit(2), motorStatusCCwSoftLimit(2), motorStatusOverDrive(2)
*
*RETURNS:
*TRUE if the bit is set, FALSE if it is not set, or ERROR if you
*specify an invalid motor controller.
*/

BOOL motorStatusOverDrive(int whichMotor)
/*
	int whichMotor;		//	which motor status to read
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichMotor < 1 || whichMotor > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_MOTOR);
		return(ERROR);
	}
#endif
	/* read motor status word and test the over-drive bit */
	if ((*motorStatus[whichMotor] & MOTOR_OVERDRIVE) == MOTOR_OVERDRIVE)
		return(TRUE);
	else
		return(FALSE);
}



/*********************************************************
*
*motorOverDriveOn - Turns on the motor over-drive
*
*Sets the over-drive bit in the motor controller
*SEE ALSO:
*motorOverDriveOff(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid motor controller
*/

STATUS motorOverDriveOn(int whichMotor)
/*
	int whichMotor;		//	which motor's overdrive to set
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichMotor < 1 || whichMotor > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_MOTOR);
		return(ERROR);
	}
#endif
	/* Set the overdrive bit in the motor controller */
	*motorOdPre[whichMotor] = MOTOR_OVERDRIVE_ON;
	return(OK);
}


/*********************************************************
*
*motorOverDriveOff - Turns off the motor over-drive
*
*Clears the over-drive bit in the motor controller
*SEE ALSO:
*motorOverDriveOn(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid motor controller
*/

STATUS motorOverDriveOff(int whichMotor)
/*
	int whichMotor;		//	which motor's overdrive to set
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichMotor < 1 || whichMotor > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_MOTOR);
		return(ERROR);
	}
#endif
	/* Clear the overdrive bit in the motor controller */
	*motorOdPre[whichMotor] = MOTOR_OVERDRIVE_OFF;
	return(OK);
}

/*************************************************************
*
*motorPrescaleSet - Sets a new prescale value
*
*Writes a new value to the motor controller prescale register.
*The clock is then divided by (n+1) where 0 <= n <= 15.
*In order to ensure a minimum pulse width of 500 nsec at the 
*output, this command should only be used when the velocity is
*set at zero.
*SEE ALSO:
*motorVelocitySet(2)
*
*RETURNS:
*OK if successful or ERROR if you specify an invalid prescale value
*or invalid motor controller.
*/

STATUS motorPrescaleSet(int whichMotor,int newPrescale)
/*
	int whichMotor;		//	which motor's prescale to set
	int newPrescale;	//	new prescale divisor
*/
{
#if (DEBUG>=30)
	if (! pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichMotor < 1 || whichMotor > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_MOTOR);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (newPrescale < MIN_PRESCALE || newPrescale > MAX_PRESCALE){
		errnoSet(S_pztMotor_INVALID_PRESCALE);
		return(ERROR);
	}
#endif
	/* Write new value to prescale register */
	*motorOdPre[whichMotor] = newPrescale;
	return(OK);
}

/*****************************************************************
*
*motorStatusRead - Read the motor status register
*
*Reads the motor status register.
*
*SEE ALSO:
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid motor controller
*/

STATUS motorStatusRead(int whichMotor, UWORD *StatusPtr)
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichMotor < 1 || whichMotor > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_MOTOR);
		return(ERROR);
	}
#endif
	/* read motor status word */
	*StatusPtr = *motorStatus[whichMotor];
  return(OK);
}

/*****************************************************************
*
*encoderStatusRead - Read the encoder status register
*
*Reads the encoder status register.
*
*SEE ALSO:
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid motor controller
*/

STATUS encoderStatusRead(int whichEncoder, UWORD *StatusPtr)
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichEncoder < 1 || whichEncoder > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_ENCODER);
		return(ERROR);
	}
#endif
	/* read encoder status word */
	*StatusPtr = *motorStatus[whichEncoder];
  return(OK);
}
	
/***************************************************************
*
*encoderClearCounter - Reset encoder counter
*
*Uses the hardware reset command to clear the encoder counter to zero.
*SEE ALSO:
*pztClearCounter(2)
*
*RETURNS:
*OK if successful, or error if you specify an invalid encoder counter.
*/

STATUS encoderClearCounter(int whichEncoder)
/*
	int whichEncoder;	//	which encoder to clear
*/
{
#if (DEBUG>=30)
	if (! pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichEncoder < 1 || whichEncoder > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_ENCODER);
		return(ERROR);
	}
#endif
	/* Write a dummy argument to the clear command address */
	*motorClearCmd[whichEncoder] = DUMMY;
	return(OK);
}

/***************************************************************
*
*encoderPositionLatch - Latch current encoder position
*
*Command to latch the encoder position counter.  This is required
*since the board really has a D16 interface while the counter is
*32 bits.  Once latched, the hi and lo words may be read and then 
*combined to form one longword.
*SEE ALSO:
*pztPositionRead(2), motorPositionRead(2), motorPositionLo(2)
*motorPositionHi(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid encoder controller.
*/

STATUS encoderPositionLatch(int whichEncoder)
/*
	int whichEncoder;	//	Which encoder counter to latch
*/

{
#if (DEBUG>=30)
	if (! pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichEncoder < 1 || whichEncoder > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_ENCODER);
		return(ERROR);
	}
#endif
	/*Write a dummy argument to the latch command register.*/
	*motorLatchCmd[whichEncoder] = DUMMY;
	return(OK);
}

/*****************************************************************
*
*encoderPositionHi - Read high word of encoder position
*
*This routine reads the high order word of the encoder position register.
*Note that encoderPositionLatch must be run prior to this in order to 
*latch the most current counter value into the position register.
*SEE ALSO:
*motorPostionLatch(2), motorPositionLo(2), motorPositionRead(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid encoder controller.
*/

STATUS encoderPositionHi(int whichEncoder, int *high)
/*
	int whichEncoder;	//	Which encoder position to read
	int *high;		//	High order position (returned)
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichEncoder < 1 || whichEncoder > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_ENCODER);
		return(ERROR);
	}
#endif
	*high = *motorPosHi[whichEncoder] & 0x0000ffff;
	return(OK);
}


/*****************************************************************
*
*encoderPositionLo - Read Low word of encoder position
*
*This routine reads the low order word of the encoder position register.
*This value should not be used by itself since it is the low order
*word of a 32 bit signed quantity with the sign in the high order word.
*Note that encoderPositionLatch must be run prior to this in order to 
*latch the most current counter value into the position register.
*SEE ALSO:
*motorPostionLatch(2), motorPositionHi(2), motorPositionRead(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid encoder controller.
*/

STATUS encoderPositionLo(int whichEncoder, int *low)
/*
	int whichEncoder;	//	Which encoder position to read
	int *low;		//	High order position (returned)
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichEncoder < 1 || whichEncoder > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_ENCODER);
		return(ERROR);
	}
#endif
	*low = *motorPosLo[whichEncoder] & 0x0000ffff;
	return(OK);
}

/**************************************************************
*
*encoderPositionRead - Read encoder position
*
*Latches and reads all 32 bits of the encoder position counter.
*SEE ALSO:
*motorPositionLatch(2), motorPositionHi(2), motorPositionLo(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid encoder controller.
*/

STATUS encoderPositionRead(int whichEncoder, int *position)
/*
	int whichEncoder;	//	Which encoder's position to read
	int *position;  	//	Current encoder position (returned)
*/
{
        int high,low;

#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichEncoder < 1 || whichEncoder > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_ENCODER);
		return(ERROR);
	}
#endif
	/* Latch counter into position register */
	if (encoderPositionLatch(whichEncoder) == ERROR)
		return(ERROR);

	/* Read high order word */	
	if (encoderPositionHi(whichEncoder, &high) == ERROR)
		return(ERROR);
 
	/* Read low order word */
	if (encoderPositionLo(whichEncoder, &low) == ERROR)
		return(ERROR);

	/* combine high and low order words */
	*position = ((high << 16) | low);
	return(OK);
}

/*****************************************************************
*
*encoderSelectState - set the state of the paddle select line
*
*This routine sets the state of the over drive line.  If the 
*channel is a motor channel, then this is a higher current
*select line.  If the channel is an encoder channel, then
*this line is used to switch the mux inside the paddle.  This
*routine does not and cannot check whether the channel it
*is changing is a paddle or a motor so the user beware!
*SEE ALSO:
*motorOverDriveOn(2) motorOverDriveOff(2)
*
*RETURNS:
*OK if successful, or ERROR if you specify an invalid channel number.
*/

STATUS encoderSelectState(int whichEncoder, int state)
/*
	int whichEncoder;	//	Which channel to change
	int state;		//	state to set the line
				//	0 = low, non zero = high
*/
{
#if (DEBUG>=30)
	if (!pztMotorInitYet){
		errnoSet(S_pztMotor_BOARD_NOT_INITIALIZED);
		return(ERROR);
	}
#endif
#ifdef DEBUG
	if (whichEncoder < 1 || whichEncoder > motorNumInstalled){
		errnoSet(S_pztMotor_INVALID_ENCODER);
		return(ERROR);
	}
#endif
	if(state)
	{
	  /* set the select bit in the encoder controller */
	  *motorOdPre[whichEncoder] = PADDLE_SELECT_ON;
	}
	else
	{
	  /* Clear the select bit in the encoder controller */
	  *motorOdPre[whichEncoder] = PADDLE_SELECT_OFF;
	}

	return(OK);
}



