/* pztMotorLib.h - header file for pztMotor.c library */
/* sccs info :     @(#)pztMotorLib.h	1.7  6/7/94  */

/*
modification history
--------------------
02a,beh,18Nov93     Modified to work with both C and C++
01d,dlp,04Oct93     added stuff for pztPostionSet function
01c,beh,13Dec91     moved some stuff off to pztMotorConfig.h
01b,dlp,13Nov90     changed LWORD type in motorPositionRead to int type
01a,dlp,05Oct90     written, debugged
*/

#ifndef INC_pztMotorLibh
#define INC_pztMotorLibh

#include "rtModNum.h"
#include "rtTypes.h"
#include "pztMotorConfig.h"

/* Maximum number of pzt, motors and encoder controllers */
/* These have to do with the architecture and remain the same even
 * if you have an underpopulated board.
*/
#define MAX_NUM_PZTS		0x6
#define MAX_NUM_MOTORS		0x6
#define MAX_NUM_ENCODERS	(MAX_NUM_MOTORS)
#define MAX_NUM_CHANNELS	(MAX_NUM_PZTS + MAX_NUM_MOTORS)

/****************************************************************/
/*Pointers to various addresses & registers on the pztmotor card*/
/*The various motor controllers			*/
#define MOTOR(N)	(N << 5)

/*The various pzt controllers			*/
#define PZT(N)		((N << 5) + 0x10)

/*The position registers read from these offsets	*/
#define PR_LSW 0x0
#define PR_MSW 0x4

/*The status words read from these offsets		*/
#define STAT 0xc

/*Velocity and overdrive/preset are written to*/
/*these offsets			*/
#define VELOCITY 0x0
#define OD_PRE 0x4

/*The clear command and latch command are written*/
/*to these offsets		*/
#define CLR_CMD 0x8
#define LTCH_CMD 0xc

/* Chip indentification mask and values */
#define CHIP_TYPE_MASK		(0x1c00)	/* bits 12 - 10 inclusive */
#define CHIP_TYPE_MOTOR		(0x0000)	/* motor chip */
#define CHIP_TYPE_UDMOTOR       (0x0800)        /* up/down motor chip */
#define CHIP_TYPE_ENCODER	(0x0400)	/* encoder chip */
#define CHIP_TYPE_PZT_INTEG	(0x1000)	/* pzt integrating chip */
#define CHIP_TYPE_PZT_DAC	(0x1400)	/* pzt dac chip */
#define CHIP_TYPE_NONE		(0x1c00)	/* no chip installed */

/***************************************************************/
/* Various constants				*/
#define MIN_VELOCITY		-0x8000
#define MAX_VELOCITY		0x7fff
#define ZERO_VELOCITY		0x0
#define MIN_PRESCALE		0x0
#define MAX_PRESCALE		0xf
#define DUMMY			0x0

/*Pzt status definitions			*/
#define PZT_NOT_CLIPPING	0x0
#define PZT_CLIPPING_UP		0x3
#define PZT_CLIPPING_DOWN	0x5

/*Motor status definitions			*/
#define MOTOR_ANY_LIMIT		0x8000
#define MOTOR_CW_HARD_LIMIT	0x8030
#define MOTOR_CW_SOFT_LIMIT	0x8003
#define MOTOR_CCW_HARD_LIMIT	0x8050
#define MOTOR_CCW_SOFT_LIMIT	0x8005
#define MOTOR_HOME		0x0200
#define MOTOR_OVERDRIVE		0x0100

/*Motor overdrive commands			*/
#define MOTOR_OVERDRIVE_OFF	0x8000
#define MOTOR_OVERDRIVE_ON	0x8001

/*Paddle select line command */
#define PADDLE_SELECT_OFF	0x0000
#define PADDLE_SELECT_ON	0x0001

/**************************************************************/

/* Error codes returned by pztMotorLib */
#define S_pztMotor_INIT_INVALID_ARGUMENT	(M_pztMotorLib | 1)
#define S_pztMotor_BOARD_NOT_INITIALIZED	(M_pztMotorLib | 2)
#define S_pztMotor_INVALID_PZT			(M_pztMotorLib | 3)
#define S_pztMotor_INVALID_MOTOR		(M_pztMotorLib | 4)
#define S_pztMotor_INVALID_VELOCITY		(M_pztMotorLib | 5)
#define S_pztMotor_INVALID_PRESCALE		(M_pztMotorLib | 6)
#define S_pztMotor_INVALID_PZT_POSITION         (M_pztMotorLib | 7)
#define S_pztMotor_INIT_PZT_CONFIGURATION_MISMATCH (M_pztMotorLib | 8)
#define S_pztMotor_INIT_MOTOR_CONFIGURATION_MISMATCH (M_pztMotorLib | 9)

/* This section is only required for the Gnu compiler */

#ifdef __cplusplus
/*********** C++ class and constructor declaration *************/
class tPztMotor {
public:
tPztMotor(void){};
tPztMotor(int iAdrsSpace, char *pBusAdrs);

STATUS pztMotorFullInit(int *boardConfig);
STATUS pztMotorGetBoardConfig(int *boardConfig);
#endif   /* __cplusplus */

STATUS pztMotorInit(int numPzts,int numMotors);
STATUS pztMotorFullInit(int numPzts,int numMotors);
STATUS pztReset(int whichPzt);
STATUS pztSetLow(int whichPzt);
STATUS pztSetHigh(int whichPzt);
STATUS motorReset(int whichMotor);
STATUS pztVelocitySet(int whichPzt,int newVelocity);
STATUS pztPositionSet(int whichPzt,int newPosition);
STATUS motorVelocitySet(int whichMotor,int newVelocity);
STATUS pztStop(int whichPzt);
STATUS motorStop(int whichMotor);
STATUS pztClearCounter(int whichPzt);
STATUS motorClearCounter(int whichMotor);
STATUS pztPositionRead(int whichPzt,int *position);
STATUS motorPositionLatch(int whichMotor);
STATUS motorPositionHi(int whichMotor,int *high);
STATUS motorPositionLo(int whichMotor,int *low);
STATUS motorPositionRead(int whichMotor,int *position);
STATUS pztStatusRead(int whichPzt);
BOOL pztStatusClipUp(int whichPzt);
BOOL pztStatusClipDown(int whichPzt);
BOOL motorStatusAnyLimit(int whichMotor);
BOOL motorStatusCwHardLimit(int whichMotor);
BOOL motorStatusCwSoftLimit(int whichMotor);
BOOL motorStatusCCwHardLimit(int whichMotor);
BOOL motorStatusCCwSoftLimit(int whichMotor);
BOOL motorStatusHome(int whichMotor);
BOOL motorStatusOverDrive(int whichMotor);
STATUS motorOverDriveOn(int whichMotor);
STATUS motorOverDriveOff(int whichMotor);
STATUS motorPrescaleSet(int whichMotor,int newPrescale);
STATUS motorStatusRead(int whichMotor, UWORD *StatusPtr);
STATUS encoderStatusRead(int whichEncoder, UWORD *StatusPtr);
STATUS encoderClearCounter(int whichEncoder);
STATUS encoderPositionLatch(int whichEncoder);
STATUS encoderPositionHi(int whichEncoder, int *high);
STATUS encoderPositionLo(int whichEncoder, int *low);
STATUS encoderPositionRead(int whichEncoder, int *position);
STATUS encoderSelectState(int whichEncoder, int state);
#ifdef __cplusplus

protected:

char *pBusAddress;
int iAMCode;

volatile char *pmboardAddress;
volatile WORD *pztVelocity[MAX_NUM_PZTS + 1];
volatile WORD *pztClearCmd[MAX_NUM_PZTS + 1];
volatile WORD *pztPosition[MAX_NUM_PZTS + 1];
volatile WORD *pztStatus[MAX_NUM_PZTS + 1];
volatile WORD *motorVelocity[MAX_NUM_MOTORS + 1];
volatile WORD *motorOdPre[MAX_NUM_MOTORS + 1];
volatile WORD *motorClearCmd[MAX_NUM_MOTORS + 1];
volatile WORD *motorLatchCmd[MAX_NUM_MOTORS + 1];
volatile WORD *motorPosHi[MAX_NUM_MOTORS + 1];
volatile WORD *motorPosLo[MAX_NUM_MOTORS + 1];
volatile WORD *motorStatus[MAX_NUM_MOTORS + 1];

int pztNumInstalled;
int motorNumInstalled;
BOOL pztMotorInitYet; 	/* TRUE once pztMotorInit has been run */

};
#endif   /* __cplusplus */

#endif  /* INC_pztMotorLibh */
